#include<math.h>
#include "gopifitting.h"
#define HUGE_NO  100000.0 
#define EPS 1.e-03
#define DO_PRINT if (0) printf
double find_distance();

/***
        double distanceQuadratic(FiveTuple ref, FiveTuple p[]);
        distanceCubic(FiveTuple ref, FiveTuple p[]);

***/

  /****
 
    p(u) = p[0] + u*(-3p[0]+3p[1]) + u^2*(3p[0] - 6p[1] + 3p[2])
                + u^3*(-p[0] + 3p[1] -3p[2] +p[3])

   The above coefficients are to be calculated for x,y and z coordinates. Each of the 
above coefficients are to be multiplied by the ref point coordinates and the corresponding
terms are to be added and submitted to the root finding routine.

  *********/


double distanceQuadratic(FiveTuple ref, FiveTuple p[])
{

  int i;
  int j, nr, new_deg;
  double coeffs[100], roots[100];
  double perpx[10], perpy[10], perpz[10];
  double dotx[10], doty[10], dotz[10];

  double coefx[4], coefy[4], coefz[4];

  coefx[0] =  p[0].x;
  perpx[0] =  p[0].x - ref.x;
  perpx[1] = coefx[1] =  -2*p[0].x + 2*p[1].x;
  perpx[2] = coefx[2] =  p[0].x - 2*p[1].x + p[2].x;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefx[i]);*/
  /*DO_PRINT("\n");*/

  coefy[0] =  p[0].y;
  perpy[0] =  p[0].y - ref.y;
  perpy[1] = coefy[1] =  -2*p[0].y + 2*p[1].y;
  perpy[2] = coefy[2] =  p[0].y - 2*p[1].y + p[2].y;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefy[i]);*/
  /*DO_PRINT("\n");*/

  coefz[0] =  p[0].z;
  perpz[0] =  p[0].z - ref.z;
  perpz[1] = coefz[1] =  -2*p[0].z + 2*p[1].z;
  perpz[2] = coefz[2] =  p[0].z - 2*p[1].z + p[2].z;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefz[i]);*/
  /*DO_PRINT("\n");*/

  return find_distance(ref, p, coefx, coefy, coefz,perpx,perpy,perpz, 2);
}

double distanceCubic(FiveTuple ref, FiveTuple p[])
{

  int i;
  int j, nr, new_deg;
  double coeffs[100], roots[100];
  double perpx[10], perpy[10], perpz[10];
  double dotx[10], doty[10], dotz[10];

  double coefx[4], coefy[4], coefz[4];

  coefx[0] =  p[0].x;
  perpx[0] =  p[0].x - ref.x;
  perpx[1] = coefx[1] =  -3*p[0].x + 3*p[1].x;
  perpx[2] = coefx[2] =  3*p[0].x - 6*p[1].x + 3*p[2].x;
  perpx[3] = coefx[3] =  -p[0].x + 3*p[1].x - 3*p[2].x + p[3].x;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefx[i]);*/
  /*DO_PRINT("\n");*/

  coefy[0] =  p[0].y;
  perpy[0] =  p[0].y - ref.y;
  perpy[1] = coefy[1] =  -3*p[0].y + 3*p[1].y;
  perpy[2] = coefy[2] =  3*p[0].y - 6*p[1].y + 3*p[2].y;
  perpy[3] = coefy[3] =  -p[0].y + 3*p[1].y - 3*p[2].y + p[3].y;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefy[i]);*/
  /*DO_PRINT("\n");*/

  coefz[0] =  p[0].z;
  perpz[0] =  p[0].z - ref.z;
  perpz[1] = coefz[1] =  -3*p[0].z + 3*p[1].z;
  perpz[2] = coefz[2] =  3*p[0].z - 6*p[1].z + 3*p[2].z;
  perpz[3] = coefz[3] =  -p[0].z + 3*p[1].z - 3*p[2].z + p[3].z;

  /*for (i=0;i<=deg;i++)DO_PRINT("%lf ", coefz[i]);*/
  /*DO_PRINT("\n");*/

  return find_distance(ref, p, coefx, coefy, coefz,perpx,perpy,perpz, 3);
}

double find_distance(FiveTuple ref, FiveTuple p[], double coefx[], double coefy[], double coefz[], double perpx[], double perpy[], double perpz[], int deg) 
{
  double tempx[10], tempy[10], tempz[10];
  double dotx[10], doty[10], dotz[10];
  double roots[100];
  double x,y,z,  power;
  double dist_square, min_dist_square;
  double dot_product[15];
  int i;
  int j, nr, new_deg;

/***  Finding the derivative  ***/
/***   coef[xyz] is not destroyed. These values are required while finding the 
       distance between the ref. point and curve  ***/

  for(i=0; i<deg; i++)
  {

   tempx[i] = coefx[i+1]*(double)(i+1);
   tempy[i] = coefy[i+1]*(double)(i+1);
   tempz[i] = coefz[i+1]*(double)(i+1);
   DO_PRINT(" derivative : temp [%d] = %f %f %f \n", i, tempx[i],tempy[i],tempz[i]);

  }

/***   Calculating  x*p'(u).x,  y*p'(u).y, z*p'(u).z  *****/
/***   coef[xyz] is not destroyed. These values are required while finding the 
       distance between the ref. point and curve  ***/

  for (i=0; i< (2*deg); i++)
   {
     dotx[i] =  doty[i] =  dotz[i] =  0.0;
   }

  for (i=0;i< deg; i++)
  for (j=0;j<=deg; j++)
  {

   dotx[i+j] += tempx[i]*perpx[j];
   doty[i+j] += tempy[i]*perpy[j];
   dotz[i+j] += tempz[i]*perpz[j];

  }

/***  Computing the dot product  ****/

  new_deg = 2*deg -1;

  for (i=0; i<= new_deg; i++)
  {
  
   dot_product[i] = dotx[i] + doty[i] + dotz[i] ;
   DO_PRINT("eqncoef[%d] = %f \n", i, dot_product[i]);

  }

  i=new_deg;
  while ((i>=0)&& (dot_product[i]<EPS)&&(dot_product[i]>(-EPS))) i--;

  if(i==0) {
     DO_PRINT("The curve is a straight line parallel to X-axis \n"); 
     return(0);
  }
  else  
     {
     DO_PRINT("%i\n", new_deg);
     nr = compute_roots(dot_product, i, roots);
     }


  if (nr == -1) {
    DO_PRINT("Error in solution\n");
    exit(-1);
  }

  min_dist_square = HUGE_NO;

  for (i=0; i<nr; i++)
  {
    DO_PRINT("%dth root is %g\n", i, roots[i]);

   if(roots[i]>=0.0)
   {
	 /*  Find the point (x,y,z) on the curve for u=roots[i]  */
	 x = y = z = 0.0;
	 power =  1.0;
         DO_PRINT("degree %d\n ", deg);
	 for (j=0;j<=deg; j++)
	 {
	    x = x + coefx[j]*power;
	    y = y + coefy[j]*power;
	    z = z + coefz[j]*power;

            DO_PRINT("the accumulated %f %f %f \n", x, y, z);

	    power = power * roots[i];
	 }


	  /* Find the distance between the ref point and (x,y,z) */

            DO_PRINT("the accumulated %f %f %f %f %f %f\n", x, ref.x,y,ref.y,z,ref.z);
	 dist_square = (x-ref.x)*(x-ref.x)+(y-ref.y)*(y-ref.y)+(z-ref.z)*(z-ref.z);
	min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;
            DO_PRINT("the min_dist %f \n", min_dist_square );
	 }
    }

  dist_square = (p[0].x-ref.x)*(p[0].x-ref.x)+(p[0].y-ref.y)*(p[0].y-ref.y)+(p[0].z-ref.z)*(p[0].z-ref.z);
  min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;

  dist_square = (p[deg].x-ref.x)*(p[deg].x-ref.x)+(p[deg].y-ref.y)*(p[deg].y-ref.y)+(p[deg].z-ref.z)*(p[deg].z-ref.z);
  min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;

  if (min_dist_square == HUGE_NO) return -1;
     else  return (min_dist_square);
}
