#include<stdio.h>
#include<math.h>
#include <fitting.h>

#define EPS 1.e-06
#define DO_PRINT if (0) printf
extern int global_checkDomain;

extern void scaleTuple( FiveTuple , double, FiveTuple* );
extern void subTuples(FiveTuple, FiveTuple, FiveTuple* );
extern void addTuples(FiveTuple, FiveTuple, FiveTuple* );

void printEqn( Eqn e )
{
 int i;
  DO_PRINT("%d :",e.degree);
  for(i=0;i<=e.degree;i++) DO_PRINT("%.12lf ", e.coef[i]);
  DO_PRINT("\n");
}

void printTupleEqn(Tuple_Eqn te)
{
  DO_PRINT("s: "); printEqn(te.s);
  DO_PRINT("t: "); printEqn(te.t);
  DO_PRINT("x: "); printEqn(te.x);
  DO_PRINT("y: "); printEqn(te.y);
  DO_PRINT("z: "); printEqn(te.z);
  DO_PRINT("\n");
}

#ifdef _STAND_ALONE
void scaleTuple( FiveTuple in, double s, FiveTuple* out )
{
  out->s = in.s * s;
  out->t = in.t * s;
  out->x = in.x * s;
  out->y = in.y * s;
  out->z = in.z * s;
}

void addTuples(FiveTuple in1, FiveTuple in2,
	       FiveTuple* out)
{
  out->s = in1.s + in2.s;
  out->t = in1.t + in2.t;
  out->x = in1.x + in2.x;
  out->y = in1.y + in2.y;
  out->z = in1.z + in2.z;
}

void subTuples(FiveTuple in1, FiveTuple in2,
	       FiveTuple* out)
{
  out->s = in1.s - in2.s;
  out->t = in1.t - in2.t;
  out->x = in1.x - in2.x;
  out->y = in1.y - in2.y;
  out->z = in1.z - in2.z;
}
#endif

void EqnCopy(Eqn *source, Eqn *dest)
{
  int i;

   (*dest).degree = (*source).degree;

   for(i=0;i<=(*source).degree;i++)
       (*dest).coef[i] = (*source).coef[i];
}
  
void EqnProd (Eqn eqn1, Eqn eqn2, Eqn *result)
{

  int i,j;
  Eqn temp;

  int deg1=eqn1.degree, deg2=eqn2.degree;

  temp.degree = deg1 + deg2;

  for (i=0; i<= deg1+deg2; i++)
   {
     temp.coef[i] =  0.0;
   }

  for (i=0;i<=deg1; i++)
  for (j=0;j<=deg2; j++)
  {
   temp.coef[i+j] += eqn1.coef[i]*eqn2.coef[j];
  }

  EqnCopy(&temp,result);
 
}

void EqnScale(Eqn eqn1, double eqn_sc, Eqn *result)
{

  Eqn eqn2;

  eqn2.degree = 0;
  eqn2.coef[0] = eqn_sc;
  EqnProd(eqn1, eqn2, result);
  return;
}
void EqnDerivative(Eqn eqn, Eqn *result)
{
  Eqn temp;
  int i;

  temp.degree = eqn.degree -1;
  for (i=0;i<=temp.degree;i++)
    {
      temp.coef[i] = eqn.coef[i+1]*(i+1);
    }

  EqnCopy(&temp, result);
  return;
     
}
void EqnSumDiff (Eqn eqn1, Eqn eqn2, Eqn *result, int sumdif)
{
                          /* sumdif=1 : Sum ; sumdif=-1 : Difference */
  int min_deg, max_deg, i;
  double *min_eqn;
  Eqn temp;

  if(eqn1.degree<eqn2.degree)
    {
       min_deg = eqn1.degree;
       min_eqn = eqn1.coef;
       max_deg = eqn2.degree;
    }
  else
    {
       min_deg = eqn2.degree;
       min_eqn = eqn2.coef;
       max_deg = eqn1.degree;
    }

   for( i= min_deg+1; i<= max_deg; i++)
    {
      min_eqn[i] =  0.0;
    }
   for( i= 0; i<= max_deg; i++)
    {
      temp.coef[i] = eqn1.coef[i] + sumdif*eqn2.coef[i];
    }

    temp.degree = max_deg;
  
    EqnCopy(&temp,result);
  
}
void TupleEqnSum(Tuple_Eqn te1, Tuple_Eqn te2, Tuple_Eqn *teresult)
{
    EqnSumDiff(te1.s, te2.s, &(*teresult).s,1);
    EqnSumDiff(te1.t, te2.t, &(*teresult).t,1);
    EqnSumDiff(te1.x, te2.x, &(*teresult).x,1);
    EqnSumDiff(te1.y, te2.y, &(*teresult).y,1);
    EqnSumDiff(te1.z, te2.z, &(*teresult).z,1);
}

void TupleEqnDiff(Tuple_Eqn te1, Tuple_Eqn te2, Tuple_Eqn *teresult)
{
    EqnSumDiff(te1.s, te2.s, &(*teresult).s,-1);
    EqnSumDiff(te1.t, te2.t, &(*teresult).t,-1);
    EqnSumDiff(te1.x, te2.x, &(*teresult).x,-1);
    EqnSumDiff(te1.y, te2.y, &(*teresult).y,-1);
    EqnSumDiff(te1.z, te2.z, &(*teresult).z,-1);
}
void TupleEqnCopy(Tuple_Eqn *source, Tuple_Eqn *dest)
{
    EqnCopy(&(source->s), &(dest->s));
    EqnCopy(&(source->t), &(dest->t));
    EqnCopy(&(source->x), &(dest->x));
    EqnCopy(&(source->y), &(dest->y));
    EqnCopy(&(source->z), &(dest->z));
   
    return;
}

void set_degree_to_fteqn(Tuple_Eqn *e, int degree)
{
   (*e).s.degree = degree;
   (*e).t.degree = degree;
   (*e).x.degree = degree;
   (*e).y.degree = degree;
   (*e).z.degree = degree;

   return;
}
void DotProduct(Tuple_Eqn *peqn1, Tuple_Eqn *peqn2, Eqn **dp_st, Eqn **dp_xyz)
{

   Tuple_Eqn eqn_product, dot_product;
   
   set_degree_to_fteqn(&eqn_product, peqn1->s.degree + peqn2->s.degree); 

   /**dp_st  = NULL;*/
   /**dp_xyz = NULL;*/
   switch(global_checkDomain)
   {

     case 2:
       EqnProd(peqn1->s,peqn2->s, &eqn_product.s);
       EqnProd(peqn1->t,peqn2->t, &eqn_product.t);
       EqnSumDiff(eqn_product.s, eqn_product.t, &dot_product.s,1);
       EqnCopy(&(dot_product.s), *dp_st);
       /**dp_st = &dot_product.s;*/
         /* No break here */
     case 0:
       EqnProd(peqn1->x,peqn2->x, &eqn_product.x);
       EqnProd(peqn1->y,peqn2->y, &eqn_product.y);
       EqnProd(peqn1->z,peqn2->z, &eqn_product.z);
       EqnSumDiff(eqn_product.x, eqn_product.y, &dot_product.y,1);
       EqnSumDiff(dot_product.y, eqn_product.z, &dot_product.x,1);
       /**dp_xyz = &dot_product.x;*/
       EqnCopy(&(dot_product.x), *dp_xyz);
       break;

     case 1:
       EqnProd(peqn1->s,peqn2->s, &eqn_product.s);
       EqnProd(peqn1->t,peqn2->t, &eqn_product.t);
       EqnSumDiff(eqn_product.s, eqn_product.t, &dot_product.s,1);
       /**dp_st = &dot_product.s;*/
       EqnCopy(&(dot_product.s), *dp_st);
       break;
    }

  return;
} 
   
double evaluate_eqn(Eqn *eqn1, double val)
{
double power=1.0, answer=0.0;
int i;

 for (i=0;i<=eqn1->degree; i++)
         {
            answer = answer + eqn1->coef[i]*power;
            power = power * val;
         }
 return answer;

}

double dist_square_for_root(Tuple_Eqn normal, double  roots, int chkDom)
{
	 /*  Find the point (x,y,z) on the curve for u=roots[i]  */
  double s,t,x,y,z,dist=0.0;


    if(chkDom==0)
      {
         x = evaluate_eqn(&(normal.x), roots);
         y = evaluate_eqn(&(normal.y), roots);
         z = evaluate_eqn(&(normal.z), roots);
         dist = x*x + y*y + z*z;
       }

    if(chkDom==1)
      {
         s = evaluate_eqn(&(normal.s), roots);
         t = evaluate_eqn(&(normal.t), roots);
         dist = s*s + t*t;
       }
    return dist;

}

double dist_square_bet_points (FiveTuple P1, FiveTuple P2, int ChkDomain)
{
   FiveTuple diff;

   subTuples(P1, P2, &diff);
   switch(ChkDomain)
   {
     case 0: return (diff.x*diff.x + diff.y*diff.y + diff.z*diff.z);
             break;
     case 1: return (diff.s*diff.s + diff.t*diff.t);
             break;
   }
}
double find_distance( Tuple_Eqn fd_tangent, Tuple_Eqn fd_normal, Eqn denom, FiveTuple *p, FiveTuple ref)
{
  double roots[10], divided_by;
  double dist_square, min_dist_square, max_min_val;
  int i, j, nr, ChkDomain;
   Eqn *dp[2];
  double *dp_coef;

  dp[0] = (Eqn *)malloc(sizeof(Eqn));
  dp[1] = (Eqn *)malloc(sizeof(Eqn));
  DotProduct(&fd_tangent, &fd_normal, &dp[0], &dp[1]);
               /* dp[0] for dp_st and dp[1] for dp_xyz */

  switch(global_checkDomain)
    {
    case 0:
      ChkDomain =1;  /* ChkDomain 0 when st to be done..
			1 when xyz to be done */
      break;
    case 1:
    case 2:
      ChkDomain = 0;
      break;
    default:
      printf("find_distance: unanticipated case\n");
      break;
   }

  max_min_val = 0.0; /* The output should be max distance of min distances
                        in xyz and st */

  do
   {
      dp_coef = dp[ChkDomain]->coef;
      i=dp[ChkDomain]->degree;
      while ((i>0)&& (dp_coef[i]<EPS)&&(dp_coef[i]>(-EPS))) i--;

      if(i==0) {
         printf("Curve Degree is ZERO.\n");
         return(0);
      }
      else {
         DO_PRINT("%d\n", i);
#ifdef _POWER
         nr = compute_all_pos_roots(dp_coef, i, roots);
#else
         nr = compute_roots(dp_coef, i, roots);
#endif
      }
     if (nr == -1) {
       DO_PRINT("Error in solution\n");
       exit(-1);
     }

     min_dist_square = HUGE_VAL;

     for (i=0; i<nr; i++)
     {
       DO_PRINT("%dth root is %.18lf\n", i, roots[i]);

       if(roots[i]>=0.0)
        {
          dist_square = dist_square_for_root(fd_normal, roots[i], 1-ChkDomain); 
               DO_PRINT("the dist_square %.12lf \n", dist_square );
#ifdef _POWER
          divided_by = evaluate_eqn(&denom,roots[i]); 
          dist_square = dist_square/(divided_by * divided_by);
#endif
          min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;
         }
      }

      dist_square = dist_square_bet_points(p[0], ref, 1-ChkDomain);
      min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;

      dist_square = dist_square_bet_points(p[fd_normal.s.degree], ref, 1-ChkDomain);
      min_dist_square = (dist_square < min_dist_square)? dist_square :min_dist_square;

     if (min_dist_square == HUGE_VAL) return -1;
      else  
       max_min_val = max_min_val>min_dist_square ? max_min_val:min_dist_square;

      ChkDomain++;
    }
   while((global_checkDomain-ChkDomain)>0);

  myfree(dp[0]); dp[0] = NULL;
  myfree(dp[1]); dp[1] = NULL;

   DO_PRINT("Distance is %.12lf\n", max_min_val);
   return max_min_val;
}
void put_in_equation(int Coef, FiveTuple p, Tuple_Eqn *e)
{
   (*e).s.coef[Coef] = p.s; 
   (*e).t.coef[Coef] = p.t; 
   (*e).x.coef[Coef] = p.x; 
   (*e).y.coef[Coef] = p.y; 
   (*e).z.coef[Coef] = p.z; 

  return;
}

ScaleCoefByTuple(int Coef, Tuple_Eqn *fteq, FiveTuple ft)
{
   (*fteq).s.coef[Coef] *= ft.s;
   (*fteq).t.coef[Coef] *= ft.t;
   (*fteq).x.coef[Coef] *= ft.x;
   (*fteq).y.coef[Coef] *= ft.y;
   (*fteq).z.coef[Coef] *= ft.z;

   return;
}

void set_coef_to_fteqn(int Coef, Tuple_Eqn *fteqn, double coefficient)
{
   (*fteqn).s.coef[Coef] = coefficient;
   (*fteqn).t.coef[Coef] = coefficient;
   (*fteqn).x.coef[Coef] = coefficient;
   (*fteqn).y.coef[Coef] = coefficient;
   (*fteqn).z.coef[Coef] = coefficient;

   return;
}
double distanceCubic(FiveTuple ref, FiveTuple *p)
{

   FiveTuple tmp;
   Tuple_Eqn tangent, normal;
   Eqn denominator;

#ifdef _POWER
  /* Finding Tangent Vectors */ 
  /* As the parameter `u' in bernstein polynomial has been 
     converted to `t' by dividing by (1-u)^3 in the power basis
     form, while differentiating the equation, the polynomial 
     changes..
   orig eqn: 1/(1+t)^3 (P[0] + 3tP[1] + 3t^2P[2] + t^3P[3] )
  When differentiated, we get: 
     (P[1]-P[0]) + 2t(P[2]-P[1]) + (P[3]-P[2])t^2 
   (Actually it has a common denominator, which never becomes zero
   and is unimportant for our solution) */

   set_degree_to_fteqn(&tangent, 2);

   subTuples (p[1], p[0], &tmp);
   put_in_equation(0,tmp,&tangent);

   subTuples (p[2], p[1], &tmp);
   scaleTuple( tmp, 2.0, &tmp);
   put_in_equation(1,tmp, &tangent);

   subTuples( p[3], p[2], &tmp);
   put_in_equation(2,tmp,&tangent);

   printTupleEqn(tangent);

  /* Finding Normal Vectors */
  /* Cubic     : (P[R] - P[0]) -3tP[1] - 3t^2P[2] - t^3P[3] */
  /*  Actual equation is : 
    1/(1+t)^3 ((1+t)^3P[R] - (P[0]+3tP[1]+3t^2P[2]-t^3P[3] */
  /* After removing the denominator: 
    (P[R] - P[0]) +3t(P[R] - P[1])+3t^2(P[R] - P[2])+t^3(P[R] - P[3])*/
    
   set_degree_to_fteqn(&normal, 3);
   set_coef_to_fteqn(0,&normal, 1.0);
   set_coef_to_fteqn(1,&normal, 3.0);
   set_coef_to_fteqn(2,&normal, 3.0);
   set_coef_to_fteqn(3,&normal, 1.0);

   EqnCopy(&(normal.x), &denominator);


    subTuples(ref,p[0],&tmp);
    ScaleCoefByTuple(0,&normal, tmp);

    subTuples(ref, p[1], &tmp);
    ScaleCoefByTuple(1,&normal, tmp);

    subTuples(ref, p[2], &tmp);
    ScaleCoefByTuple(2,&normal, tmp);

    subTuples(ref, p[3], &tmp);
    ScaleCoefByTuple(3,&normal, tmp);
   printTupleEqn(normal);
#else
   /* Tgt : (p1-p0) + 2(p2-2p1+p0)u + (p3-3p2+3p1-p0)u^2 */

   set_degree_to_fteqn(&tangent, 2);

   subTuples (p[1], p[0], &tmp);
   put_in_equation(0,tmp,&tangent);

   scaleTuple( p[1], -2.0, &tmp);
   addTuples (p[2], tmp, &tmp);
   addTuples (p[0], tmp, &tmp);
   scaleTuple( tmp, 2.0, &tmp);
   put_in_equation(1,tmp, &tangent);

   subTuples (p[1], p[2], &tmp);
   scaleTuple( tmp, 3.0, &tmp);
   addTuples (p[3], tmp, &tmp);
   subTuples (tmp, p[0], &tmp);
   put_in_equation(2,tmp,&tangent);

   printTupleEqn(tangent);

   /* Norm : (pr-p0) - 3(p1-p0)u - 3(p2-2p1+p0)u^2 - (p3-3p2+3p1-p0)u^3 */

   set_degree_to_fteqn(&normal, 3);
   
   subTuples(ref, p[0], &tmp);
   put_in_equation(0,tmp,&normal);

   subTuples (p[1], p[0], &tmp);
   scaleTuple( tmp, -3.0, &tmp);
   put_in_equation(1,tmp,&normal);

   scaleTuple( p[1], -2.0, &tmp);
   addTuples (p[2], tmp, &tmp);
   addTuples (p[0], tmp, &tmp);
   scaleTuple( tmp, -3.0, &tmp);
   put_in_equation(2,tmp, &normal);

   subTuples (p[1], p[2], &tmp);
   scaleTuple( tmp, 3.0, &tmp);
   addTuples (p[3], tmp, &tmp);
   subTuples (tmp, p[0], &tmp);
   scaleTuple( tmp, -1.0, &tmp);
   put_in_equation(3,tmp,&normal);

   printTupleEqn(normal);

#endif
   
/* Return the max(min distance between the curve and point in xyz and st 
       separately) */
    return find_distance(tangent, normal, denominator,  p, ref);
}



double distanceQuadratic(FiveTuple ref, FiveTuple *p)
{

   FiveTuple tmp;
   Tuple_Eqn tangent, normal;
   Eqn denominator;

#ifdef _POWER
  /* Finding Tangent Vectors */ 
  /*Quadratic     : 2P[1] + 2tP[2]*/
  /* As the parameter `u' in bernstein polynomial has been 
     converted to `t' by dividing by (1-u)^2 in the power basis
     form, while differentiating the equation, the polynomial 
     changes..
   orig eqn: 1/(1+t)^2 (P[0] + 2tP[1] + t^2P[2]) 
  When differentiated, we get: 
     (P[1]-P[0]) + t(P[2]-P[1]) 
   (Actually it has a common denominator, which never becomes zero
   and is unimportant for our solution) */


   set_degree_to_fteqn(&tangent, 1);

   subTuples(p[1],p[0], &tmp);
   put_in_equation(0,tmp,&tangent);

   subTuples(p[2],p[1], &tmp);
   put_in_equation(1,tmp, &tangent);

  /* Finding Normal Vectors */
  /* Quadratic : (P[R] - P[0]) + 2t (P[R] - P[1])) + t^2 (P[R] - P[2]) */
    

   set_degree_to_fteqn(&normal, 2);

   set_coef_to_fteqn(0,&normal, 1.0);
   set_coef_to_fteqn(1,&normal, 2.0);
   set_coef_to_fteqn(2,&normal, 1.0);

   EqnCopy(&(normal.x), &denominator);

   subTuples(ref,p[0],&tmp);
   ScaleCoefByTuple(0,&normal, tmp);

   subTuples(ref,p[1],&tmp);
   ScaleCoefByTuple(1,&normal, tmp);

   subTuples(ref,p[2],&tmp);
   ScaleCoefByTuple(2,&normal, tmp);
#else
   /* Tgt : (p1-p0) + (p2-2p1+p0)u */

   set_degree_to_fteqn(&tangent, 1);

   subTuples (p[1], p[0], &tmp);
   put_in_equation(0,tmp,&tangent);

   scaleTuple( p[1], -2.0, &tmp);
   addTuples (p[2], tmp, &tmp);
   addTuples (p[0], tmp, &tmp);
   put_in_equation(1,tmp, &tangent);

   printTupleEqn(tangent);

   /* Norm : (pr-p0) - 2(p1-p0)u - (p2-2p1+p0)u^2 */

   set_degree_to_fteqn(&normal, 3);
   
   subTuples(ref, p[0], &tmp);
   put_in_equation(0,tmp,&normal);

   subTuples (p[1], p[0], &tmp);
   scaleTuple( tmp, -2.0, &tmp);
   put_in_equation(1,tmp,&normal);

   scaleTuple( p[1], -2.0, &tmp);
   addTuples (p[2], tmp, &tmp);
   addTuples (p[0], tmp, &tmp);
   scaleTuple( tmp, -1.0, &tmp);
   put_in_equation(2,tmp, &normal);

   printTupleEqn(normal);

#endif
   
/* Return the max(min distance between the curve and point in xyz and st 
       separately) */
   return find_distance(tangent, normal, denominator,  p, ref);
}

#ifdef _STAND_ALONE
main()
{

  int i, degree;
  FiveTuple p[4], ref;
  
  scanf("%d",&degree);
  for(i=0;i<=degree;i++){
     scanf("%lf %lf %lf",&p[i].x, &p[i].y, &p[i].z);
     p[i].s = p[i].t = 0.0;
   }
  scanf("%lf %lf %lf",&ref.x, &ref.y, &ref.z);
  ref.s = ref.t = 0;
  global_checkDomain=0;

  for(i=0;i<=degree;i++) DO_PRINT("%lf %lf %lf\n",p[i].x, p[i].y, p[i].z);
  if(degree == 2)  distanceQuadratic(ref,p);
  if(degree == 3)  distanceCubic(ref,p);

  
}
#endif
