#include  <stdio.h>
#include <math.h>
#include "curve.h"
#include "fitting.h"
#include <time.h>

static struct M_ENTRY  IMPLICIT[DIM][DIM];	/* The global matrix with the
					 * implicit representation */

static struct M_ENTRY  IMPLICIT1[DIM][DIM];

static coordinate      Mats[ORDER][DIM][DIM];	/* Used for storing the matrices used
					 * in linearizing a matrix to an
					 * Eigenvalue problem */

static coordinate      BigMat[DIM1][DIM1];	/* The matrix obtained after
					 * linearization */

static coordinate      BigMat1[DIM1][DIM1];	/* The matrix obtained after
					 * linearization */

static coordinate      Kernel[DIM][DIM];

static float           factor;

static int             Degree,dim;		/* The degree of the implicit representation */
static coordinate      vr1[DIM1][DIM1],rconde[DIM1], rcondv[DIM1];
static coordinate      wr[DIM1], wi[DIM1], beta[DIM1], vl[DIM1][DIM1], vr[DIM1][DIM1];
static struct CURVE    curve;
static struct CURVE    curve1;
static int t1,t2,tm,iwork[DIM1];

static struct MONOMIAL Poly[500];	/* A global structure for */
					/* maintaining local polynomials */

static coordinate      Matrix[ORDER][ORDER];



check_self_intersect(ParamCurve inCurve1, ParamCurve inCurve2,
		     int *count, double **sArray, double **tArray)
{
	int             i, j;
	struct M_ENTRY  m_entry;
	coordinate      alpha, gamma;
	coordinate      s, t, Z[2],x2,y2;

	factor = 1.0;
	Set_Im_Curve_(inCurve1);
	Set_Par_Curve_(inCurve2);

	Curv_Curv_Inst_();
	*sArray = (double*) malloc(sizeof(double)*dim);
	*tArray = (double*) malloc(sizeof(double)*dim);
	*count = 0;
	for (i = 0; i < dim; i++) 
            if ((fabs(wi[i]) < 0.00001) && ( wr[i] >= 0.0)) {
		if (fabs(wr[i])> 1.0) {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][dim - Degree];
				gamma = vr[i][dim - Degree + 1];
			} else {
				alpha = vr[i][dim - 2];
				gamma = vr[i][dim - 1];
			}
		} else {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][0];
				gamma = vr[i][1];
			} else {
				alpha = vr[i][Degree - 2];
				gamma = vr[i][Degree - 1];
			}

		}
		t = gamma / (alpha + gamma);
		s = wr[i] / (1 + wr[i]);

                if ((t >= 0.0) && (t <= 1.0)) {
		  (*sArray)[*count] = s;
		  (*tArray)[*count] = t;
		  (*count)++;
		}
	    }
}

old_main()
{
	int             i, j;
	struct M_ENTRY  m_entry;
	coordinate      alpha, gamma;
	coordinate      s, t, Z[2],x2,y2;
        clock_t clock();



	scanf("%f", &factor);
	Read_Im_Curve_();	/* Reads the implicit form of the curve */

	Read_Par_Curve_();	/* Reads the parametric curve */


	Curv_Curv_Inst_();
	for (i = 0; i < dim; i++) 
            if ((fabs(wi[i]) < 0.00001) && ( wr[i] >= 0.0)) {
		if (fabs(wr[i])> 1.0) {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][dim - Degree];
				gamma = vr[i][dim - Degree + 1];
			} else {
				alpha = vr[i][dim - 2];
				gamma = vr[i][dim - 1];
			}
		} else {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][0];
				gamma = vr[i][1];
			} else {
				alpha = vr[i][Degree - 2];
				gamma = vr[i][Degree - 1];
			}

		}
		t = gamma / (alpha + gamma);
		s = wr[i] / (1 + wr[i]);

                if ((t >= 0.0) && (t <= 1.0))
		    fprintf(stderr, " s=%15.10f,   t=%15.10f,  \n", s, t);
	    }
}

/*
 * Bino_Coeff computes the binomial coefficient used in the representation of
 * Bernstein Basis
 */
Bino_Coeff_(n, d)
	int             n, d;

{
	int             prod = 1;
	int             i, j;

	for (i = n - d + 1; i <= n; i++)
		prod = prod * i;

	for (j = 1; j <= d; j++)
		prod = prod / j;

	return (prod);
}


/*
 * Bezier_Power: Converts a Bezier  Representation into a power basis by a
 * transformation of the form s = t / (1 - t).
 */

Bezier_Power_(pcurve)
	struct CURVE   *pcurve;
{
	int             i, j, d;
	struct POINT   *ppoint;
	int             coeff;
	struct CURVE    curve;

	d = pcurve->order - 1;
	ppoint = pcurve->point;
	for (i = 0; i <= d; i++) {
		coeff = Bino_Coeff_(d, i);
		ppoint->x *= coeff;
		ppoint->y *= coeff;
		ppoint->w *= coeff;
		ppoint++;
	}
}

Subdivide_(pcurve)		/* Does midpoint subdivision */
	struct CURVE   *pcurve;
{
	int             i, j;
	struct POINT   *ppoint;
	struct CURVE    curve;
	int             order, coeff;
	coordinate      prod;

	order = pcurve->order;

	prod = 1.0;

	for (i = 0; i < order; i++) {
		curve.point[i].x = 0.0;
		curve.point[i].y = 0.0;
		curve.point[i].w = 0.0;
		for (j = 0; j <= i; j++) {
			coeff = Bino_Coeff_(i, j);
			curve.point[i].x = curve.point[i].x + coeff * pcurve->point[j].x / prod;
			curve.point[i].y = curve.point[i].y + coeff * pcurve->point[j].y / prod;
			curve.point[i].w = curve.point[i].w + coeff * pcurve->point[j].w / prod;
		}
		prod = prod * 2.0;
	}

	ppoint = pcurve->point;

	for (i = 0; i < order; i++) {
		ppoint->x = curve.point[i].x;
		ppoint->y = curve.point[i].y;
		ppoint->w = curve.point[i].w;
		ppoint++;
	}
}


/*
 * Mat_Mul : Multiplies matrices A and B, and returns the result in B.
 */

Mat_Mul_(A, B, order)
	coordinate      A[DIM][DIM];
	coordinate      B[DIM][DIM];
	int             order;
{
	int             i, j, k;
	coordinate      sum;
	coordinate      C[DIM][DIM];

	for (i = 0; i < order; i++)
		for (j = 0; j < order; j++) {
			sum = 0.0;
			for (k = 0; k < order; k++)
				sum += A[i][k] * B[k][j];
			C[i][j] = sum;
		}
	for (i = 0; i < order; i++)
		for (j = 0; j < order; j++)
			B[i][j] = C[i][j];
}


/*
 * Curv_Curv_Inst_ : The main routine used for the intersection of curves and
 * curves. The curves's implicit representation is stored in the global
 * matrix M_ENTRY.
 * 
 * At the end it computes the entries of BigMat, whose eigenvalues correspond to
 * the roots.
 */

Curv_Curv_Inst_()
{
	int             order;
	int             i, j, j1, k, l;
	struct POINT    point;
	int             ipvt[DIM];
	coordinate      z[DIM];
	coordinate      work[DIM];
	coordinate      rcond;
	int             dim = DIM;
	int             job;
	coordinate      det[2];
	coordinate      Mat_Temp[DIM][DIM];
	coordinate      EPS = .0001;

	struct M_ENTRY  m_entry;

	/*
	 * Subdivide_(&(curve)); Subdivide_(&(curve));
	 */

/*
*/
	order = curve.order;
	point = curve.point[0];
	for (i = 0; i < order; i++) {
		for (j = 0; j < Degree; j++)
			for (k = 0; k < Degree; k++) {
				m_entry = IMPLICIT[j][k];
				Mats[i][j][k] = m_entry.x * point.x +
					m_entry.y * point.y +
					m_entry.w * point.w;
			}
		point = curve.point[i + 1];
	}

	/*
	 * Computing the Inverse of the Matrix Corresponding to the leading
	 * degree
	 */
	for (i = 0; i < DIM; i++)
		ipvt[i] = 0.0;

/*
        printf(" M = [ ");
*/
	for (i = 0; i < Degree; i++)
            {
		for (j = 0; j < Degree; j++)
                   {
	               Mat_Temp[i][j] = Mats[order - 1][i][j];
/*
	               Mat_Temp[i][j] = Mats[0][i][j];
*/
/*
                       printf(" %f ",Mat_Temp[i][j]);
*/
                   }
/*
                printf(" ; \n");
*/
            }
/*
        printf(" ] ");
*/

	dgeco(Mat_Temp, &dim, &Degree, ipvt, &rcond, z);
/*
        printf(" TIME = %d \n",t2 - t1);
	fprintf(stderr, "Condition Number = %15.10f \n", rcond);
*/
	if (rcond > EPS) {
		job = 11;
		dgedi(Mat_Temp, &dim, &Degree, ipvt, det, work, &job);

		/*
		 * Performing Matrix Multiplications with the rest of the
		 * matrices.
		 */

		for (i = 0; i < order - 1; i++)
		    Mat_Mul_(Mat_Temp, &(Mats[i][0][0]), Degree);

		/*
		 * Use the Linearization Operation.
		 */
		dim = Degree * (order - 1);
		for (i = 0; i < order - 2; i++) {
			j = i * Degree;
			j1 = j + Degree;
			for (k = j; k < j1; k++) {
				l = 0;
				while (l < Degree + k) {
					BigMat[l][k] = 0.0;
					l++;
				}
				BigMat[l][k] = 1.0;
				l++;
				while (l < dim) {
					BigMat[l][k] = 0.0;
					l++;
				}
			}
		}
		l = dim - Degree;
		for (i = l; i < dim; i++) {
			for (j = 0; j < order - 1; j++)
				for (k = 0; k < Degree; k++)
					BigMat[k + j * Degree][i] = -Mats[j][i - l][k];
		}
/*
                for (i = 0; i < dim; i++)
                {
                     for (j = 0; j < dim; j++)
                         printf("%15.8f \n  ",BigMat[i][j]);
                }
*/
		Compute_Roots_();
	} else {		/* Reduce to Generalized Eigenvalue Problem */
		/*
		 * The Generalized system is of the form
		 * 
		 * BigMat1 * s + BigMat
		 */
		dim = Degree * (order - 1);
		for (i = 0; i < order - 2; i++) {
			j = i * Degree;
			j1 = j + Degree;
			for (k = j; k < j1; k++) {
				l = 0;
				while (l < Degree + k) {
					BigMat[l][k] = 0.0;
					l++;
				}
				BigMat[l][k] = 1.0;
				l++;
				while (l < dim) {
					BigMat[l][k] = 0.0;
					l++;
				}
			}
		}
		l = dim - Degree;
		for (i = l; i < dim; i++) {
			for (j = 0; j < order - 1; j++)
				for (k = 0; k < Degree; k++)
					BigMat[k + j * Degree][i] = -Mats[j][i - l][k];
		}

		for (i = 0; i < dim; i++)
			for (j = 0; j < dim; j++)
				BigMat1[j][i] = 0.0;

		for (j = 0; j < l; j++)
			BigMat1[j][j] = 1.0;

		for (i = l; i < dim; i++)
			for (j = l; j < dim; j++)
				BigMat1[j][i] = Mats[order - 1][i - l][j - l];

		/*
		 * for (i = 0; i < dim; i++) { for (j = 0; j < dim; j++)
		 * printf("B(%d,%d) = %19.15f; \n ",i+1,j+1,BigMat[i][j]); }
		 * for (i = l; i < dim - 1; i++) { for (j = 0; j < dim; j++)
		 * printf("%19.15f ",BigMat[i][j]); printf("; \n"); } for (j
		 * = 0; j < dim; j++) printf("%19.15f ",BigMat[dim-1][j]);
		 * printf(" ] \n");
		 * 
		 * printf("\n\n\n"); printf(" A = [");
		 */
		/*
		 * for (i = 0; i < dim; i++) { for (j = 0; j < dim; j++)
		 * printf("A(%d,%d) = %19.15f; \n ",i+1,j+1,BigMat1[i][j]); }
		 * for (i = l; i < dim - 1; i++) { for (j = 0; j < dim; j++)
		 * printf("%19.15f ",BigMat1[i][j]); printf("; \n"); } for (j
		 * = 0; j < dim; j++) printf("%19.15f ",BigMat1[dim-1][j]);
		 * printf(" ] \n");
		 * 
		 */
		Compute_Roots_1_();
	}
}

/*
 * Compute_Roots_1_:   Uses the  Generalized Eigenvalue routines for
 * computing all the Relevant roots.
 */

Compute_Roots_1_()
{
	int             lda, ldvl, ldvr, ilo, ihi, lwork, info;
	coordinate      eps, temp;
	int             i, ierr;
	extern double   fabs();
	extern double   dlamch_();
	coordinate      alpha, gamma, s, t;

	lda = DIM1;
	ldvl = DIM1;
	ldvr = DIM1;
	lwork = DIM1;
	dim = Degree * (curve.order - 1);
	i = 10;
	/*
	dgegs_(&compvl, &compvr, &dim, BigMat, &lda, BigMat1, &lda, wr, wi, beta, vl, &ldvr, vr, &ldvr, work, &lwork, &info);
	 */
	rgg(&lda,&dim,BigMat,BigMat1,wr,wi,beta,&i,vr,&ierr);

/*
        tm = t2 - t1;
	printf("%d \n",tm);
*/
	for (i = 0; i < dim; i++)
            if ((fabs(wi[i]) < 0.0000001) && ( wr[i] >= 0.0))
            {
		if (fabs(wr[i] / beta[i]) > 1.0) {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				gamma = vr[i][dim - Degree + 1];
				alpha = vr[i][dim - Degree];
			} else {
				alpha = vr[i][dim - 2];
				gamma = vr[i][dim - 1];
			}
		} else {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][0];
				gamma = vr[i][1];
			} else {
				alpha = vr[i][Degree - 2];
				gamma = vr[i][Degree - 1];
			}

		}
		t = gamma / (alpha + gamma);
		s = wr[i] / (beta[i] + wr[i]);

	    }
}




/*
 * Compute_Roots:   Uses the Eigenvalue routines for computing all the
 * Relevant roots.
 */
Compute_Roots_()
{
	int             lda, ldvl, ldvr, ilo, ihi, lwork, info;
	coordinate      abnrm;
	coordinate      alpha, gamma, s, t;
	coordinate      eps, temp;
	int             i, i1, j, m, k;
	extern double   fabs();
	char            job, initv, eigsrc;
	int             ierr;
        int matz;

	lda = DIM1;
	ldvl = DIM1;
	ldvr = DIM1;
	dim = Degree * (curve.order - 1);


        matz = 1; 

/*
        t2 = clock();
*/
        rg(&lda,&dim,BigMat,wr,wi,&matz,vr,iwork,rconde,&ierr);
/*
	printf("%d \n",clock() - t2);
*/
/*
	for (i = 0; i < dim; i++)
            if ((fabs(wi[i]) < 0.0000001) && ( wr[i] >= 0.0))
            {
		if (fabs(wr[i]) > 1.0) {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				gamma = vr[i][dim - Degree + 1];
				alpha = vr[i][dim - Degree];
			} else {
				alpha = vr[i][dim - 2];
				gamma = vr[i][dim - 1];
			}
		} else {
			if (fabs(vr[i][0]) > fabs(vr[i][Degree - 1])) {
				alpha = vr[i][0];
				gamma = vr[i][1];
			} else {
				alpha = vr[i][Degree - 2];
				gamma = vr[i][Degree - 1];
			}

		}
		t = gamma / (alpha + gamma);
		s = wr[i] / (1.0 + wr[i]);
	    }
*/
}

/*
 * Read_Im_Curve: Reads the implicit representation of the curve. It reads
 * into the array IMPLICIT.
 */
Read_Im_Curve_()
{
	int             order;
	struct M_ENTRY  m_entry;
	double          var;
	int             i, j;
	char            s[30];
	extern double   atof();
	coordinate      A[ORDER], B[ORDER];

	scanf("%d", &order);	/* THE ORDER OF THE IMPLICIT FORM */
	Degree = order - 1;
	curve1.order = order;

	for (i = 0; i < order; i++) {
		scanf("%s", s);
		curve1.point[i].x = atof(s);
		scanf("%s", s);
		curve1.point[i].y = atof(s);
		scanf("%s", s);
		curve1.point[i].w = atof(s);

		curve1.point[i].x = curve1.point[i].x * curve1.point[i].w / ((double) factor);
		curve1.point[i].y = curve1.point[i].y * curve1.point[i].w / ((double) factor);
		curve1.point[i].w = curve1.point[i].w / ((double) factor);
	}

	/*
	 * Subdivide_(&(curve1)); Subdivide_(&(curve1));
	 */
	Bezier_Power_(&(curve1));	/* Bernstein to Power Conversion */

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].x;
		B[i] = curve1.point[i].y;
	}

	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].w = Matrix[i][j];

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].y;
		B[i] = curve1.point[i].w;
	}
	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].x = Matrix[i][j];

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].w;
		B[i] = curve1.point[i].x;
	}
	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].y = Matrix[i][j];
/*

*/
}

Set_Im_Curve_(ParamCurve inputCurve)
{
	int             order;
	struct M_ENTRY  m_entry;
	double          var;
	int             i, j;
	char            s[30];
	extern double   atof();
	coordinate      A[ORDER], B[ORDER];

	Degree = inputCurve.degree;
	order = Degree + 1;
	curve1.order = order;

	for (i = 0; i < order; i++) {
		curve1.point[i].x = inputCurve.control[i].s;
		curve1.point[i].y = inputCurve.control[i].t;
		curve1.point[i].w = 1.0;

		curve1.point[i].x = curve1.point[i].x * curve1.point[i].w / ((double) factor);
		curve1.point[i].y = curve1.point[i].y * curve1.point[i].w / ((double) factor);
		curve1.point[i].w = curve1.point[i].w / ((double) factor);
	}

	/*
	 * Subdivide_(&(curve1)); Subdivide_(&(curve1));
	 */
	Bezier_Power_(&(curve1));	/* Bernstein to Power Conversion */

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].x;
		B[i] = curve1.point[i].y;
	}

	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].w = Matrix[i][j];

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].y;
		B[i] = curve1.point[i].w;
	}
	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].x = Matrix[i][j];

	for (i = 0; i < order; i++) {
		A[i] = curve1.point[i].w;
		B[i] = curve1.point[i].x;
	}
	Compute_Matrix(A, B, Degree);

	for (i = 0; i < Degree; i++)
		for (j = 0; j < Degree; j++)
			IMPLICIT[i][j].y = Matrix[i][j];
}

/*
 * Read_Par_Curve_ : Reads the parametric curve into the array curve;
 */

Read_Par_Curve_()
{
	int             order, i;
	struct POINT    point;
	float           var;

	char            s[30];
	extern double   atof();

	scanf("%d", &order);	/* THE ORDER OF THE BEZIER CURVE */
	curve.order = order;

	for (i = 0; i < order; i++) {
		scanf("%s", s);
		curve.point[i].x = atof(s);
		scanf("%s", s);
		curve.point[i].y = atof(s);
		scanf("%s", s);
		curve.point[i].w = atof(s);

		curve.point[i].x = curve.point[i].x * curve.point[i].w / ((double) factor);
		curve.point[i].y = curve.point[i].y * curve.point[i].w / ((double) factor);
		curve.point[i].w = curve.point[i].w / ((double) factor);
	}
	Bezier_Power_(&(curve));
}

Set_Par_Curve_(ParamCurve inputCurve)
{
	int             order, i;
	struct POINT    point;
	float           var;

	char            s[30];
	extern double   atof();

	curve.order = inputCurve.degree + 1;

	for (i = 0; i < order; i++) {
		curve.point[i].x = inputCurve.control[i].s;
		curve.point[i].y = inputCurve.control[i].t;
		curve.point[i].w = 1.0;

		curve.point[i].x = curve.point[i].x * curve.point[i].w / ((double) factor);
		curve.point[i].y = curve.point[i].y * curve.point[i].w / ((double) factor);
		curve.point[i].w = curve.point[i].w / ((double) factor);
	}
	Bezier_Power_(&(curve));
}


/*
 * Compute_Matrix():- Uses Cayley resultant formulation on the polynomials of
 * degree n in arrays A and B. The arrays consist of coefficients expressed
 * in power basis. The result is returned in Matrix[][] as n*n output.
 */

Compute_Matrix(A, B, n)
	coordinate     *A, *B;
	int             n;

{
	int             i, j, j1, k1, k;
	coordinate      sum, temp, temp1;
	int             index;
	struct MONOMIAL monomial;

	index = 0;

	for (i = 0; i <= n; i++) {
		for (j = 0; j <= i; j++) {
			monomial.coeff = A[i] * B[j];
			if (monomial.coeff != 0.0) {
				k1 = i - j - 1;
				for (k = 0; k <= k1; k++) {
					monomial.s = i - 1 - k;
					monomial.t = k + j;
					Poly[index++] = monomial;
				}
			}
		}
		for (j = i + 1; j <= n; j++) {
			monomial.coeff = -A[i] * B[j];
			if (monomial.coeff != 0.0) {
				k1 = j - i - 1;
				for (k = 0; k <= k1; k++) {
					monomial.s = j - 1 - k;
					monomial.t = i + k;
					Poly[index++] = monomial;
				}
			}
		}
	}
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			Matrix[i][j] = 0.0;

	for (i = 0; i < index; i++) {
		monomial = Poly[i];
		j = monomial.s;
		k = monomial.t;
		Matrix[j][k] = Matrix[j][k] + monomial.coeff;
	}
}

#ifdef _TEST_ALONE
main()
{
  ParamCurve curve1, curve2;
  int i, n;
  double *s, *t;

/*
  curve1.degree = curve2.degree = 4;
  curve1.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
				(curve1.degree+1));
  curve2.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
				(curve2.degree+1));
  curve1.control[0].s = -2.5; curve1.control[0].t = 0.3;
  curve1.control[1].s = 4.0; curve1.control[1].t = 6.1;
  curve1.control[2].s = -1.0; curve1.control[2].t = -3.3;
  curve1.control[3].s = 5.0; curve1.control[3].t = 1.4;
  curve1.control[4].s = 7.2; curve1.control[4].t = 3.3;

  curve2.control[0].s = -2.0; curve2.control[0].t = 2.3;
  curve2.control[1].s = 3.0; curve2.control[1].t = 2.1;
  curve2.control[2].s = -4.0; curve2.control[2].t = -2.3;
  curve2.control[3].s = 2.0; curve2.control[3].t = 1.3;
  curve2.control[4].s = 3.0; curve2.control[4].t = 3.3;
*/

  curve1.degree = 2;
  curve2.degree = 2;
  curve1.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
				(curve1.degree+1));
  curve2.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
				(curve2.degree+1));
  curve1.control[0].s = -3.0; curve1.control[0].t = 0.0;
  curve1.control[1].s = 0.0; curve1.control[1].t = 18.0;
  curve1.control[2].s = 3.0; curve1.control[2].t = 0.0;

  curve2.control[0].s = -1.0; curve2.control[0].t = 8.0;
  curve2.control[1].s = 0.0; curve2.control[1].t = 10.0;
  curve2.control[2].s = 1.0; curve2.control[2].t = 8.0;

  check_self_intersect(curve1, curve2, &n, &s, &t);
  for (i=0; i<n; i++) {
    printf("#%d: s = %15.10lf, t = %15.10lf\n", i, s[i], t[i]);
  }
}
#endif
