#ifndef __FITTING_H__
#define __FITTING_H__

typedef struct _FiveTuple {
  double s, t, x, y, z;
} FiveTuple;

typedef struct {
   int degree;
   double coef[10];
} Eqn;

typedef struct {
  Eqn s,t,x,y,z;
} Tuple_Eqn;
  
typedef struct _ParamCurve {
  int degree;
  FiveTuple *control;  
  /* NOTE: free(control) after use to prevent memory leak */
} ParamCurve;

/*----------------------------------------------------------------------*/
/* Parameters for fitting():
   Input (pass by value)
       count: number of points in piecewise linear approximation
       p[]  : points in first patch
       q[]  : points in the neighboring patch
       simpleSplit: use 0 for monotonic subdivision
                        1 for bisection (to improve error bound)
		       -1 for no initial subdivision at all
       errorBound: fitting error must be less than this value
       domainError: if 1, check the fitting error in domain value
                    if 0, check in the (x,y,z) 3-space.

   Output (passed by address)
       curveCount: number of curves in Bezier fitting
       pCurves[] : curves description of the 1st patch
       qCurves[] : curves desfription of the 2nd patch
*/
void fitting(int count, FiveTuple *p, FiveTuple *q,
	     int *curveCount, ParamCurve **pCurves, ParamCurve **qCurves,
	     char simpleSplit, double errorBound, int checkDomain);

#endif
