#include <assert.h>
#include <stdio.h>
#include "fitting.h"

typedef struct {
  int nTrim;
  FiveTuple *trimPoints;
  int adjCount;
  int *adjKnots;
  int *adjIDs;
} ParamSurface;

test_main()
{
  ParamCurve curve1, curve2;
  int i, n;
  double *s, *t;

  curve1.degree = curve2.degree = 2;
  curve1.control = (FiveTuple*) malloc(sizeof(FiveTuple)*3);
  curve2.control = (FiveTuple*) malloc(sizeof(FiveTuple)*3);
  curve1.control[0].s = 0; curve1.control[0].t = 0; 
  curve1.control[1].s = 1; curve1.control[1].t = 2; 
  curve1.control[2].s = 2; curve1.control[2].t = 0; 
  curve2.control[0].s = 0; curve2.control[0].t = 1; 
  curve2.control[1].s = 1; curve2.control[1].t = 0; 
  curve2.control[2].s = 2; curve2.control[2].t = 1;
  
  check_self_intersect(curve1, curve2, &n, &s, &t);
  for (i=0; i<n; i++) {
    printf("#%d: s = %f, t = %f\n", i, s[i], t[i]);
  }
}

main(int argc, char** argv)
{
  FILE *fp, *fpAdj, *pOutFile, *qOutFile;
  int count, curveCount, patchCount=0, i, j;
  int degreeS, degreeT, nNeighbors;
  int index=0;

  ParamSurface patches[1000];
  FiveTuple *p, *q;
  ParamCurve *pCurves, *qCurves;

  char adjName[1000];

  fp = fopen(argv[1], "r");
  if (! fp) {
    fprintf(stderr, "Can't open %s\n", argv[1]);
    return 1;
  }
  sprintf(adjName, "%s.adj", argv[1]);
  fpAdj = fopen(adjName, "r");
  if (! fpAdj) {
    fprintf(stderr, "Can't open adj file: %s\n", adjName);
    return 1;
  }
  pOutFile = fopen( "testp.m", "w" );
  qOutFile = fopen( "testq.m", "w" );

  while (EOF != fscanf(fp, "%d %d\n", &degreeS, &degreeT)) {
    for (i=0; i<(degreeS+1)*(degreeT+1); i++)
      fscanf(fp, "%*f %*f %*f %*f\n");
    fscanf(fp, "%d\n", &i);
    fscanf(fp, "%d %d\n", &count, &i);
    count++;
    patches[patchCount].nTrim = count;
    patches[patchCount].trimPoints=(FiveTuple*)malloc(count*sizeof(FiveTuple));
    for (i=0; i<count; i++) {
      FiveTuple *tmp = &(patches[patchCount].trimPoints[i]);
      fscanf(fp, "%lf %lf %lf %lf %lf\n", 
	     &tmp->s, &tmp->t, &tmp->x, &tmp->y, &tmp->z);
    }
    patchCount++;
  }

  /*--- reading the .adj file ---*/
  while (EOF != fscanf(fpAdj, "%d %d\n", &i, &nNeighbors)) {
    patches[i].adjCount = nNeighbors;
    patches[i].adjKnots = (int*) malloc(sizeof(int)*nNeighbors);
    patches[i].adjIDs = (int*) malloc(sizeof(int)*nNeighbors);
    for (j=0; j<nNeighbors; j++) {
      int in1, in2;
      fscanf(fpAdj, "%d %d\n", &in1, &in2);
      patches[i].adjKnots[j] = in1;
      patches[i].adjIDs[j] = in2;
    }
  }

  for (i=0; i<patchCount; i++) {
    /* output the original PIECEWISE LINEAR points */
    fprintf(pOutFile, "origCount(%d)=%d;\n", i+1, patches[i].nTrim);
    fprintf(qOutFile, "origCount(%d)=%d;\n", i+1, patches[i].nTrim);
    for (j=0; j<patches[i].nTrim; j++) {
      fprintTuple(pOutFile, "orig", i+1, j+1, patches[i].trimPoints[j]);
      fprintTuple(qOutFile, "orig", i+1, j+1, patches[i].trimPoints[j]);
    }
    
    for (j=0; j<patches[i].adjCount; j++) {
      int pStart, pEnd, qStart, qEnd;
      int qID, k;

      pStart = patches[i].adjKnots[j];
      pEnd = (j+1>=patches[i].adjCount) ?
	     patches[i].nTrim-1 : patches[i].adjKnots[j+1];
      qID = patches[i].adjIDs[j];

      if (-1 == qID) { /* skip this segment, if no neighbor */
	printf("patch %d p[%d..%d] doesn't have a neighbor\n",
	       i, pStart, pEnd);
	continue;
      }

      for (k=0; k<patches[qID].adjCount; k++) {
	if (patches[qID].adjIDs[k] == i)
	  break;
      }
      if (k >= patches[qID].adjCount) {
	printf("weird error\n");
	continue;
      }

      assert(k < patches[qID].adjCount);  /* the search should never fail */
      qStart = patches[qID].adjKnots[k];
      qEnd = (k+1>=patches[qID].adjCount) ?
	     patches[qID].nTrim-1 : patches[qID].adjKnots[k+1];

      if (qEnd-qStart != pEnd-pStart) {
	printf("#points mismatch: patch %d p[%d..%d] & patch %d q[%d..%d]\n",
	       i, pStart, pEnd, qID, qStart, qEnd);
	continue;
      }

      assert( qEnd-qStart == pEnd-pStart );
      printf("\nfitting patch %d p[%d..%d] & patch %d q[%d..%d]\n",
	     i, pStart, pEnd, qID, qStart, qEnd);
      fitting(pEnd-pStart+1,
	      &(patches[i].trimPoints[pStart]),
	      &(patches[qID].trimPoints[qStart]),
	      &curveCount, &pCurves, &qCurves, 0);
      printf("checking self intersection...\n");
      for (k=0; k<curveCount-1; k++) {
	int interCount;
	double *interS, *interT;

	check_self_intersect(pCurves[k], pCurves[k+1],
			     &interCount, &interS, &interT);
	if (0 == interCount)
	  printf("Weird! curves %d and %d are separated\n", k, k+1);
	else if (interCount > 1)
	  printf("===> SELF INTERSECT DETECTED: %d %d\n", k, k+1);
	else if (interS[0] != 1.0 || interT[0] != 0.0) {
	  printf("===> weird! s = %d, t = %d\n", interS[0], interT[0]);
	}
	free(interS); free(interT);

	check_self_intersect(qCurves[k], qCurves[k+1],
			     &interCount, &interS, &interT);
	if (0 == interCount)
	  printf("Weird! curves %d and %d are separated\n", k, k+1);
	else if (interCount > 1)
	  printf("===> SELF INTERSECT DETECTED: %d %d\n", k, k+1);
	else if (interS[0] != 1.0 || interT[0] != 0.0) {
	  printf("===> weird! s = %f, t = %f\n", interS[0], interT[0]);
	}
	free(interS); free(interT);
      }

      index++;
      resetCount();
      for (k=0; k<curveCount; k++) {
	outputCurve( pOutFile, pCurves[k], index );
	free(pCurves[k].control);
      }
      resetCount();
      for (k=0; k<curveCount; k++) {
	outputCurve( qOutFile, qCurves[k], index );
	free(qCurves[k].control);
      }
      free(pCurves);
      free(qCurves);
    }
  }
}
