#include <errno.h>
#include <stdio.h>

#define MAX_DEGREE 100

#define FP_EQ_EPS( a, b, c )  ((((a) - (b)) < (c)) && (((a) - (b)) > -(c)))

#define  ALLOCN(PTR,TYPE,N) \
  { if (N<=0) PTR = (TYPE *) NULL; \
    else \
      if ((PTR = (TYPE *) malloc((unsigned) (N)*sizeof(TYPE)))==NULL) { \
        printf("Tried to allocate %d, %d bytes\n", N,sizeof(TYPE)); \
	printf("ERRONO is %d\n", errno); \
	perror("malloc failed");  \
	exit(-1);  \
      } \
  }

#define LINALGINIT		double LAmag,LAsum; int LAi,LAj,LAk;

#define ALLOCMAT(a, m, n, type) {ALLOCN(a, type*, m);  \
				 for(LAi=0; LAi<m; LAi++) \
				   ALLOCN(a[LAi], type, n); }
#define FREEMAT(a, m, n)   {for(LAi=0; LAi<m; LAi++) myfree(a[LAi]);  \
			      myfree(a); }

/*int d_sign(double x, double y)*/
/*{*/
/**/
  /*if (x<y) return (-1);*/
  /*else if (x>y) return (1);*/
  /*else return 0;*/
/**/
/*}*/

#ifdef _STAND_ALONE
void myfree(ptr)
void *ptr;
{
  if (ptr == NULL) return;
  free(ptr);
}

/* this assumes that the arrays wr, wi, and dEVectors are pre-allocated */
call_rg(mat, m, n, wr, wi, dEVectors, matzval)
double	 **mat;
int	 m, n, matzval;
double   *wr, *wi;
double   **dEVectors;
{ 
  int	 i, j, k, ierr;
  int	 nm, matz;
  double  *a, *z, *if1;
  int	  *iv1;

  /* allocate/initialize the matrices */
  ALLOCN(a, double, m*n);
  for(j = k = 0; j < n; j++)
    for(i = 0; i < m; i++)
      a[k++] = mat[i][j];	

  ALLOCN(z, double, m*n);
  ALLOCN(iv1, int, n);
  ALLOCN(if1, double, n);

  matz = matzval;

#ifdef __SGI
  rg_(&m, &n, a, wr, wi, &matz, z, iv1, if1, &ierr);
#else
  rg(&m, &n, a, wr, wi, &matz, z, iv1, if1, &ierr);
#endif

  if (ierr != 0)
  { printf("Error computing eigenvalues\n"); 
    exit(-1);
  }

  for(i = 0; i < m; i++)
  { k = i;
    for(j = 0; j < n; j++)
    { dEVectors[i][j] = z[k];
      k += m;
    }
  }

  /* free up the allocated space */
  myfree(a); myfree(z); myfree(iv1); myfree(if1);
}
#endif /* _STAND_ALONE */

/* This routines returns the real roots (roots) of a polynomial given */
/* its coefficients (coeffs) in power basis and its degree. The return */
/* value gives the number of real roots that lie between 0 and 1. If */
/* there is any error, returns -1. The array roots must be allocated */
/* before calling the routine. */
int compute_roots(coeffs, degree, roots)
double *coeffs;
int degree;
double *roots;
{
  int i, j, k;
  int act_deg, num_roots;
  double **mat, rec;
  double *Re, *Im, **Vec;

  LINALGINIT

  act_deg = degree;
  while ((act_deg > 0) && FP_EQ_EPS(coeffs[act_deg], 0.0, 1e-3))
    act_deg--;

  if (act_deg < 1) {
    printf("compute_roots: Not a regular polynomial\n");
    return -1;
  }

  if (act_deg >= MAX_DEGREE) {
    printf("compute_roots: Very high degree polynomial.\n");
    return -1;
  }

  ALLOCMAT(mat, act_deg, act_deg, double);
  ALLOCMAT(Vec, act_deg, act_deg, double);
  ALLOCN(Re, double, act_deg);
  ALLOCN(Im, double, act_deg);

  for (i=0; i<(act_deg-1); i++)
    for (j=0; j<act_deg; j++) {
      if (j == (i+1)) mat[i][j] = 1.0;
      else mat[i][j] = 0.0;
    }
  rec = 1.0 / coeffs[act_deg];
  for (j=0; j<act_deg; j++)
    mat[act_deg-1][j] = -coeffs[j] * rec;

  call_rg(mat, act_deg, act_deg, Re, Im, Vec, 0);
  num_roots = 0;

  for (i=0; i<act_deg; i++) 
    if (FP_EQ_EPS(Im[i],0.0,1e-6) && (Re[i] > -1e-3) && (Re[i] < (1+1e-3)))
      roots[num_roots++] = Re[i];

  myfree(Re); myfree(Im);
  FREEMAT(mat, act_deg, act_deg); FREEMAT(Vec, act_deg, act_deg);

  return num_roots;
}

#undef _TEST_ALONE 
#ifdef _TEST_ALONE
main()
{
  int deg, i, j, nr;
  double coeffs[100], roots[100];

  printf("Enter the degree of polynomial: ");
  scanf("%d", &deg);

  for (i=0; i<=deg; i++) {
    printf("Enter %dth coefficient: ", i);
    scanf("%lf", &coeffs[i]);
  }

  nr = compute_roots(coeffs, deg, roots);
  if (nr == -1) {
    printf("Error in solution\n");
    exit(-1);
  }

  for (i=0; i<nr; i++)
    printf("%dth root is %g\n", i, roots[i]);
}
#endif
