#include "defs.h"
#define MAX_TRIM_PORTIONS 3

static int adj_curve_num[MAX_PATCHES_PER_SOLID][MAX_SPLINES_PER_TRIM_REGION];
static int done[MAX_PATCHES_PER_SOLID][MAX_SPLINES_PER_TRIM_REGION];
static int zero_one[MAX_PATCHES_PER_SOLID][MAX_TRIM_PORTIONS];

int test_zero_one(tpst)
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  int i, j;
  int ncurves;
  int bool;
  
  ncurves = tpst->NumOfCurves;
  if (ncurves != 4) return 0;

  for (i=0; i<ncurves; i++) {
    if (tpst->UV_Splines[i].Deg != 1) return 0;
    for (j=0; j<=tpst->UV_Splines[i].Deg; j++)
      if (!FP_EQ_EPS(tpst->UV_Splines[i].CntrlPts[j].dCoord[2],
		     1.0, 1e-4))
	return 0;
    
    bool = ((i % 3) != 0);
    if (!FP_EQ_EPS(tpst->UV_Splines[i].CntrlPts[0].dCoord[0], 
		   (double) bool, 1e-4))
      return 0;
    
    bool = (i >= 2);
    if (!FP_EQ_EPS(tpst->UV_Splines[i].CntrlPts[0].dCoord[1], 
		   (double) bool, 1e-4))
      return 0;
  }
   
  return 1;
}

int read_solid_from_file(FILE *fp, TRIM_PATCH_INFO *solid)
{
  int r, p, q, s;
  int i, j;
  int ncurves, uvdeg, xyzdeg, tmp;

  for(i=0;
      fscanf(fp,"%d %d", &solid[i].orig_tp.iDeg0, &solid[i].orig_tp.iDeg1) != EOF;
      i++) {
    solid[i].tp.iDeg0 = solid[i].tp.iDeg1 = 0;
    solid[i].tp.coeffs = NULL;
    allocate_tensor_patch(&solid[i].orig_tp);
    for (p=0; p<=solid[i].orig_tp.iDeg0; p++)
      for (q=0; q<=solid[i].orig_tp.iDeg1; q++) {
	for (r=X; r<=W; r++)
	  fscanf(fp, "%lf", 
		  &solid[i].orig_tp.coeffs[p][q].dCoord[r]);
      }

    /* TP has been input. Now read the trim-curves */

    fscanf(fp, "%d", &solid[i].NumTrimPortions);
    for (r=0; r<solid[i].NumTrimPortions; r++) {
      fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.NumOfCurves);
      ncurves = solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines,
	     UV_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines,
	     XYZ_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.AdjInfo,
	     int, ncurves);
      for (p=0; p<ncurves; p++) {
	fscanf(fp, "%d", 
	       &solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg);
	uvdeg = solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg; 
	fscanf(fp, "%d", &tmp);
	solid[i].SplineReprn = tmp - 1;
	if (!solid[i].SplineReprn) {
	  printf("Cannot handle piecewise case for adjacency. Quitting after patch %d\n", i);
	  exit(-1);
	}
	ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts,
	       POINT3D, (uvdeg+1));
	for (q=0; q<=uvdeg; q++) {
	  fscanf(fp, "%lf %lf %lf",
		 &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[0],     
		 &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[1], 
		 &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[2]); 
	}
	fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].Deg);
	xyzdeg = solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].Deg; 
	ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts,
	       POINT4D, (xyzdeg+1));
	for (q=0; q<=xyzdeg; q++) {
	  for (s=0; s<4; s++)
	    fscanf(fp, "%lf", 
		   &solid[i].TrimPortion[r].SplinePortion.
		   XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	}
      }
      zero_one[i][r] =
	test_zero_one(&solid[i].TrimPortion[r].SplinePortion);
      /* zero_one[i][r] = 0; */
    }
  }

  return i;
}

int write_final_solid_to_file(FILE *fp, int npatches, TRIM_PATCH_INFO *solid)
{
  int r, p, q, s;
  int i, j;
  int ncurves, uvdeg, xyzdeg;
  double utmp, vtmp;
  
  for (i=0; i< npatches; i++) {
    fprintf(fp,"%d %d\n",
            solid[i].orig_tp.iDeg0, solid[i].orig_tp.iDeg1);
    for (p=0; p<=solid[i].orig_tp.iDeg0; p++)
      for (q=0; q<=solid[i].orig_tp.iDeg1; q++) {
	for (r=X; r<=W; r++)
	  fprintf(fp, "%lf ", 
		  solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	fprintf(fp, "\n");
      }
    
    /* TP has been output. Now print the trim-curves */
    
    fprintf(fp, "%d\n", solid[i].NumTrimPortions);
    if (!solid[i].SplineReprn) {
      for (r=0; r<solid[i].NumTrimPortions; r++)
	{
	  fprintf(fp, "%d %d\n", 1, -1);
	  fprintf(fp,"%d %d\n",
		  solid[i].TrimPortion[r].PiecePortion.NumOfVertices, 1);
	  for (p=0;
	       p<=solid[i].TrimPortion[r].PiecePortion.NumOfVertices;
	       p++) {
	    utmp = solid[i].umin +
	      solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][0]*
	      (solid[i].umax - solid[i].umin);
	    vtmp = solid[i].vmin +
	      solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][1]*
	      (solid[i].vmax - solid[i].vmin);
	    fprintf(fp,"%lf %lf\n", utmp, vtmp);
	  }
	}
    } else {
      for (r=0; r<solid[i].NumTrimPortions; r++) {
	if (!zero_one[i][r]) {
	  ncurves = solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
	  fprintf(fp,"%d\n", ncurves);
	  for (p=0; p<ncurves; p++) {
	    fprintf(fp, "%d %d\n", 
		    solid[i].TrimPortion[r].SplinePortion.AdjInfo[p],
		    adj_curve_num[i][p]);
	    uvdeg = solid[i].TrimPortion[r].SplinePortion.
	      UV_Splines[p].Deg; 
	    xyzdeg = solid[i].TrimPortion[r].SplinePortion.
	      XYZ_Splines[p].Deg; 
	    fprintf(fp, "%d\n", uvdeg);
	    fprintf(fp, "%d\n", 1+solid[i].SplineReprn);
	    for (q=0; q<=uvdeg; q++) {
	      fprintf(fp, "%g %g %g\n", 
		      solid[i].TrimPortion[r].SplinePortion.
		      UV_Splines[p].CntrlPts[q].dCoord[0], 
		      solid[i].TrimPortion[r].SplinePortion.
		      UV_Splines[p].CntrlPts[q].dCoord[1],
		      solid[i].TrimPortion[r].SplinePortion.
		      UV_Splines[p].CntrlPts[q].dCoord[2]); 
	    }
	    fprintf(fp, "%d\n", xyzdeg);
	    for (q=0; q<=xyzdeg; q++) {
	      for (s=0; s<4; s++)
		fprintf(fp, "%g ", 
			solid[i].TrimPortion[r].SplinePortion.
			XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	      fprintf(fp, "\n");
	    }
	  }
	} else {
	  fprintf(fp,"%d\n", 0);
	  for (p=0; p<4; p++)
	    fprintf(fp, "%d %d\n", 
		    solid[i].TrimPortion[r].SplinePortion.AdjInfo[p],
		    adj_curve_num[i][p]);
	}
      }
    }
    
  }

  fclose(fp);

  return 0;
}

int points_match(pt1, pt2, num, eps)
POINT4D *pt1, *pt2;
int num;
double eps;
{
  int i, reverse;

  reverse = -1;
  if (NEW_EQ_4D_EPS(pt1[0], pt2[0], eps)) 
    reverse = 0;
  if ((reverse == -1) && NEW_EQ_4D_EPS(pt1[0], pt2[num], eps))
    reverse = 1;
  if (reverse == -1)
    return 0;
  if (!NEW_EQ_4D_EPS(pt1[num], pt2[(1-reverse)*num], eps))
    return 0;
  
/*
  for (i=1; i<=num; i++)
    if (!NEW_EQ_4D_EPS(pt1[i], pt2[reverse*(num-i)+(1-reverse)*i], eps))
      return 0;
*/
  return 1;
}

GenerateAdjacencies(FILE* fpin, FILE* fpout)
{
  TRIM_PATCH_INFO* solid;
  int i, j, k, ncurves, num_trim_regions, adj_pnum;
  int curve_deg, nc2, l, m, npatches;
  POINT4D cntl_pts1[5], cntl_pts2[5];

  for (i=0; i<MAX_PATCHES_PER_SOLID; i++)
      for (k=0; k<MAX_SPLINES_PER_TRIM_REGION; k++) {
	adj_curve_num[i][k] = -1;
	done[i][k] = 0;
      }

  /* npatches = read_solid_from_file(fp, solid); */
  npatches = read_solid_culver(fpin, &solid);
  printf("Number of patches read = %d\n", npatches);

  if (npatches > MAX_PATCHES_PER_SOLID) {
    printf("Too many patches. Increase the constant and proceed\n");
    exit(-1);
  }

  for (i=0; i<npatches; i++) {
    num_trim_regions = solid[i].NumTrimPortions;
    if (num_trim_regions != 1) {
      printf("More than one trim region. \n");
      exit(-1);
    }
    zero_one[i][0] =
      test_zero_one(&solid[i].TrimPortion[0].SplinePortion);
    ncurves = solid[i].TrimPortion[0].SplinePortion.NumOfCurves;
    for (k=0; k<ncurves; k++) {
      if (done[i][k]) continue;
      adj_pnum = solid[i].TrimPortion[0].SplinePortion.AdjInfo[k];
      curve_deg =
	solid[i].TrimPortion[0].SplinePortion.XYZ_Splines[k].Deg;
      for (j=0; j<=curve_deg; j++)
	cntl_pts1[j] =
	  solid[i].TrimPortion[0].SplinePortion.XYZ_Splines[k].CntrlPts[j]; 
      if (solid[adj_pnum].NumTrimPortions != 1) {
	printf("Number of trim portions of adjacent patch not 1\n");
	exit(-1);
      }
      nc2 = solid[adj_pnum].TrimPortion[0].SplinePortion.NumOfCurves;
      for (l=0; l<nc2; l++) {
	if ((solid[adj_pnum].TrimPortion[0].SplinePortion.AdjInfo[l] != i) || 
	    (solid[adj_pnum].TrimPortion[0].SplinePortion.XYZ_Splines[l].Deg 
	     != curve_deg)) 
	  continue;
	for (m=0; m<=curve_deg; m++)
	  cntl_pts2[m] =
	    solid[adj_pnum].TrimPortion[0].SplinePortion.XYZ_Splines[l].CntrlPts[m];
	if (points_match(cntl_pts1, cntl_pts2, curve_deg, .001)) {
	  if (done[adj_pnum][l]) {
	    printf("Not a good sign. Bad adjacency\n");
	    exit(-1);
	  }
	  done[i][k] = done[adj_pnum][l] = 1;
	  adj_curve_num[i][k] = l;
	  adj_curve_num[adj_pnum][l] = k;
	  break;
	}
      }
      if (l == nc2) {
	printf("Could not find adjacency\n");
	exit(-1);
      }
    }
  }

  write_final_solid_to_file(fpout, npatches, solid);
  /*  for (i=0; i<npatches; i++)
    free_trim_patch(&solid[i]); */
}
