#include <stdio.h>
#include <defs.h>

int write_solid_culver(TRIM_PATCH_INFO* parent_solid, int npatches, 
		       FILE* fp)
{
  int r, p, q, s;
  int i, j;
  double utmp, vtmp, uval, vval, xval, yval, zval;
  double normX, normY, normZ;
  int is_facing_outside;
  int ncurves, uvdeg, xyzdeg;
  TENSOR_PATCH norm_tp;
  double Scale = 1.0;

#ifdef _LOTS_OF_HOLES
  int NumElmtsInHole;
  double tmpCurve1[1000][2];
#endif

  for (i=0; i<npatches; i++) {
    
    fprintf(fp, "%d %d\n",
            parent_solid[i].tp.iDeg0, parent_solid[i].tp.iDeg1);

    for (p=0; p<=parent_solid[i].orig_tp.iDeg0; p++)
      for (q=0; q<=parent_solid[i].orig_tp.iDeg1; q++) {
	for (r=0; r<=2; r++)
	  fprintf(fp, "%.20g ", 
		  Scale*parent_solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	fprintf(fp, "%.20g\n", 
		parent_solid[i].orig_tp.coeffs[p][q].dCoord[3]);
      }

    /* TP has been output. Now print the trim-curves */

    fprintf(fp, "%d\n", parent_solid[i].NumTrimPortions);
    if (!parent_solid[i].SplineReprn) {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  fprintf(fp, "%d %d\n", 1, -1);
	  fprintf(fp,"%d %d\n",
		  parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices, 1);
	  for (p=0;
	       p<=parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices;
	       p++) {
	    utmp = parent_solid[i].umin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][0]*
		(parent_solid[i].umax - parent_solid[i].umin);
	    vtmp = parent_solid[i].vmin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][1]*
		(parent_solid[i].vmax - parent_solid[i].vmin);
	    fprintf(fp,"%.20g %.20g\n", utmp, vtmp);
	  }
	  utmp = parent_solid[i].umin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.u*
	      (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.v*
	      (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp, "%.20g %.20g\n", utmp, vtmp);
	}
    } else {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  ncurves = parent_solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
	  fprintf(fp,"%d\n", ncurves);
	  for (p=0; p<ncurves; p++) {
	    fprintf(fp, "%d\n", 
		    parent_solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	    uvdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          UV_Splines[p].Deg; 
	    xyzdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          XYZ_Splines[p].Deg; 
	    fprintf(fp, "%d\n", uvdeg);
	    fprintf(fp, "%d\n", 1+parent_solid[i].SplineReprn);
	    for (q=0; q<=uvdeg; q++) {
	      utmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]) * 
		       parent_solid[i].umin) + (parent_solid[i].umax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]);
	      vtmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]) * 
		       parent_solid[i].vmin) + (parent_solid[i].vmax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]);
	      fprintf(fp, "%.20g %.20g %.20g\n", utmp, vtmp,  
			parent_solid[i].TrimPortion[r].SplinePortion.
			UV_Splines[p].CntrlPts[q].dCoord[2]); 
	    }
	    fprintf(fp, "%d\n", xyzdeg);
	    for (q=0; q<=xyzdeg; q++) {
	      for (s=0; s<4; s++)
		fprintf(fp, "%.20g ", 
			parent_solid[i].TrimPortion[r].SplinePortion.
			XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	      fprintf(fp, "\n");
	    }
	  }
/*
	  utmp = parent_solid[i].umin +
	    parent_solid[i].TrimPortion[r].SplinePortion.ptInsideRegion.u*
	      (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin +
	    parent_solid[i].TrimPortion[r].SplinePortion.ptInsideRegion.v*
	      (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp, "%.20g %.20g\n", utmp, vtmp);
*/
	}
    }
    
  }

  return 0;
}

int read_solid_culver(FILE *fp, TRIM_PATCH_INFO **solidp)
{
  int r, p, q, s;
  int i, j;
  int ncurves, uvdeg, xyzdeg, tmp;
  int unused, retval;
  int create_edge_splines = 0;
  TRIM_PATCH_INFO* solid;

  /* Try reading newer tagged format.  If this fails, it will return 0,
   * and the failure will happen trying to match "npatches", i.e.
   * immediately, and no data will be lost; so we can go ahead and
   * try the old format. */
  
  retval = read_solid_tagged(fp, solidp);
  if (retval > 0) return retval;

  printf("Reading untagged file.\n");
  
  solid = (TRIM_PATCH_INFO*)malloc(MAX_PATCHES_PER_SOLID*sizeof(TRIM_PATCH_INFO));
  *solidp = solid;
  
  for(i=0;
      (retval = fscanf(fp,"%d %d", &solid[i].orig_tp.iDeg0, &solid[i].orig_tp.iDeg1)) != EOF;
      i++) {
    if (retval != 2) printf("Only read %d in line %d\n", retval, __LINE__-1);

    if (i >= MAX_PATCHES_PER_SOLID) {
      printf("Stopped after %d patches--the hard-coded limit\n", MAX_PATCHES_PER_SOLID);
      break;
    }

    /* printf("\nTensor patch %d, %dx%d\n", i, solid[i].orig_tp.iDeg0, solid[i].orig_tp.iDeg1); */
    solid[i].tp.iDeg0 = solid[i].orig_tp.iDeg0;
    solid[i].tp.iDeg1 = solid[i].orig_tp.iDeg1;
    
    allocate_tensor_patch(&solid[i].orig_tp);
    allocate_tensor_patch(&solid[i].tp);

    solid[i].umin = 0.0;
    solid[i].umax = 1.0;
    solid[i].vmin = 0.0;
    solid[i].vmax = 1.0;

    /* printf("Control points "); */
    for (p=0; p<=solid[i].orig_tp.iDeg0; p++) {
      for (q=0; q<=solid[i].orig_tp.iDeg1; q++) {
	for (r=0; r<=3; r++) {
	  retval = fscanf(fp, "%lf", 
			  &solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	  if (retval != 1) printf("Only read %d in line %d\n", retval, __LINE__-1);
	  solid[i].tp.coeffs[p][q].dCoord[r] = solid[i].orig_tp.coeffs[p][q].dCoord[r];
	}
	/* printf("*"); */
      }
      /* printf(" "); */
    }
    /* printf("\n"); */
    
    /* TP has been input. Now read the trim-curves */
    
    retval = fscanf(fp, "%d", &solid[i].NumTrimPortions);
    if (retval != 1) printf("Only read %d in line %d\n", retval, __LINE__-1);
    /* printf("NumTrimPortions = %d\n", i, solid[i].NumTrimPortions); */
    for (r=0; r<solid[i].NumTrimPortions; r++) {
      /* printf("trimportion %d:\n", r); */
      retval = fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.NumOfCurves);
      if (retval != 1) printf("Only read %d in line %d\n", retval, __LINE__-1);
      ncurves = solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
      /* printf("  ncurves = %d\n", ncurves); */

      if (ncurves == 0) {
	ncurves = 4;
	create_edge_splines = 1;
      }

      ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines,
	     UV_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines,
	     XYZ_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.AdjInfo,
	     int, ncurves);
      
	for (p=0; p<ncurves; p++) {
	  
	  retval = fscanf(fp, "%d", 
			  &solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	  if (retval != 1) printf("Only read %d in line %d\n", retval, __LINE__-1);
	  
	  retval = fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg);
	  if (retval != 1) printf("Only read %d in line %d\n", retval, __LINE__-1);
	  uvdeg = solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg; 
	  fscanf(fp, "%d", &tmp);
	  solid[i].SplineReprn = tmp - 1;
	  if (!solid[i].SplineReprn) {
	    printf("Cannot handle piecewise case for adjacency. Quitting after patch %d\n", i);
	    exit(-1);
	  }
	  ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts,
		 POINT3D, (uvdeg+1));
	  for (q=0; q<=uvdeg; q++) {
	    fscanf(fp, "%lf %lf %lf",
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[0],     
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[1], 
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[2]); 
	  }
	  fscanf(fp, "%d", &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].Deg);
	  xyzdeg = solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].Deg; 
	  ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts,
		 POINT4D, (xyzdeg+1));
	  for (q=0; q<=xyzdeg; q++) {
	    for (s=0; s<4; s++)
	      fscanf(fp, "%lf", 
		     &solid[i].TrimPortion[r].SplinePortion.
		     XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	  }

      }
    }
  }
  return i;
}


/***** Tagged versions *****/

int write_solid_tagged(TRIM_PATCH_INFO* parent_solid, int npatches, 
		       FILE* fp)
{
  int r, p, q, s;
  int i, j;
  double utmp, vtmp, uval, vval, xval, yval, zval;
  double normX, normY, normZ;
  int is_facing_outside;
  int ncurves, uvdeg, xyzdeg;
  TENSOR_PATCH norm_tp;
  double Scale = 1.0;

#ifdef _LOTS_OF_HOLES
  int NumElmtsInHole;
  double tmpCurve1[1000][2];
#endif

  fprintf(fp, "npatches %d\n", npatches);
  
  for (i=0; i<npatches; i++) {
    
    fprintf(fp, "\npatchdegree %d %d\n",
            parent_solid[i].tp.iDeg0, parent_solid[i].tp.iDeg1);

    for (p=0; p<=parent_solid[i].orig_tp.iDeg0; p++)
      for (q=0; q<=parent_solid[i].orig_tp.iDeg1; q++) {
	fprintf(fp, "cp ");
	for (r=0; r<=2; r++)
	  fprintf(fp, "%.20g ", 
		  Scale*parent_solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	fprintf(fp, "%.20g\n", 
		parent_solid[i].orig_tp.coeffs[p][q].dCoord[3]);
      }

    /* TP has been output. Now print the trim-curves */

    fprintf(fp, "splinetrimming %d\n", parent_solid[i].SplineReprn);
    fprintf(fp, "ntrimregions %d\n", parent_solid[i].NumTrimPortions);
    if (!parent_solid[i].SplineReprn) {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  fprintf(fp, "\n  nvertices %d\n",
		  parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices);
	  for (p=0;
	       p<=parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices;
	       p++) {
	    utmp = parent_solid[i].umin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][0]*
		(parent_solid[i].umax - parent_solid[i].umin);
	    vtmp = parent_solid[i].vmin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][1]*
		(parent_solid[i].vmax - parent_solid[i].vmin);
	    fprintf(fp," vtx %.20g %.20g\n", utmp, vtmp);
	  }
	  utmp = parent_solid[i].umin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.u*
	      (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.v*
	      (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp, " pointinside %.20g %.20g\n", utmp, vtmp);
	}
    } else {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  ncurves = parent_solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
	  fprintf(fp,"\n  ncurves %d\n", ncurves);
	  for (p=0; p<ncurves; p++) {
	    fprintf(fp, "\n    adjacentpatch %d\n", 
		    parent_solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	    
	    uvdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          UV_Splines[p].Deg; 
	    xyzdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          XYZ_Splines[p].Deg; 
	    fprintf(fp, "    uvdegree %d\n", uvdeg);
	    for (q=0; q<=uvdeg; q++) {
	      utmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]) * 
		       parent_solid[i].umin) + (parent_solid[i].umax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]);
	      vtmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]) * 
		       parent_solid[i].vmin) + (parent_solid[i].vmax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]);
	      fprintf(fp, "    uv %.20g %.20g %.20g\n", utmp, vtmp,  
			parent_solid[i].TrimPortion[r].SplinePortion.
			UV_Splines[p].CntrlPts[q].dCoord[2]); 
	    }
	    fprintf(fp, "    xyzdegree %d\n", xyzdeg);
	    for (q=0; q<=xyzdeg; q++) {
	      fprintf(fp, "    xyz ");
	      for (s=0; s<4; s++)
		fprintf(fp, "%.20g ", 
			parent_solid[i].TrimPortion[r].SplinePortion.
			XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	      fprintf(fp, "\n");
	    }
	  }
	}
    }
    
  }

  return 1;
}

/* NB: this allocates the TRIM_PATCH_INFO array */

int read_solid_tagged(FILE *fp, TRIM_PATCH_INFO **solidp)
{
  int r, p, q, s;
  int i, j;
  int npatches, ncurves, uvdeg, xyzdeg, tmp;
  int unused, retval;
  int create_edge_splines = 0;
  TRIM_PATCH_INFO* solid;

  retval = fscanf(fp, " npatches %d", &npatches);
  if (retval != 1 || npatches < 1) return 0;

  solid = (TRIM_PATCH_INFO*)malloc(npatches*sizeof(TRIM_PATCH_INFO));
  *solidp = solid;

  for(i=0; i<npatches; i++) {

    retval = fscanf(fp," patchdegree %d %d", &solid[i].orig_tp.iDeg0, &solid[i].orig_tp.iDeg1);
    if (retval != 2) return 0;

    solid[i].tp.iDeg0 = solid[i].orig_tp.iDeg0;
    solid[i].tp.iDeg1 = solid[i].orig_tp.iDeg1;
    
    allocate_tensor_patch(&solid[i].orig_tp);
    allocate_tensor_patch(&solid[i].tp);

    solid[i].umin = 0.0;
    solid[i].umax = 1.0;
    solid[i].vmin = 0.0;
    solid[i].vmax = 1.0;

    for (p=0; p<=solid[i].orig_tp.iDeg0; p++) {
      for (q=0; q<=solid[i].orig_tp.iDeg1; q++) {
	retval = fscanf(fp, " cp %lf %lf %lf %lf", 
			  &solid[i].orig_tp.coeffs[p][q].dCoord[0],
			  &solid[i].orig_tp.coeffs[p][q].dCoord[1],
			  &solid[i].orig_tp.coeffs[p][q].dCoord[2],
			  &solid[i].orig_tp.coeffs[p][q].dCoord[3]);
	if (retval != 4) return 0;
	for (r=0; r<4; r++)
	  solid[i].tp.coeffs[p][q].dCoord[r] = solid[i].orig_tp.coeffs[p][q].dCoord[r];
      }
    }
    
    /* TP has been input. Now read the trim-curves */
    
    retval = fscanf(fp, " splinetrimming %d", &solid[i].SplineReprn);
    if (retval != 1) return 0;
    retval = fscanf(fp, " ntrimregions %d", &solid[i].NumTrimPortions);
    if (retval != 1) return 0;
    if (!(solid[i].SplineReprn)) fprintf(stderr, "Warning: SplineReprn=%d\n", solid[i].SplineReprn);
    
    for (r=0; r<solid[i].NumTrimPortions; r++) {
      retval = fscanf(fp, " ncurves %d", &solid[i].TrimPortion[r].SplinePortion.NumOfCurves);
      if (retval != 1) return 0;
      ncurves = solid[i].TrimPortion[r].SplinePortion.NumOfCurves;

      if (ncurves == 0) {
	ncurves = 4;
	create_edge_splines = 1;
      }

      ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines,
	     UV_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines,
	     XYZ_SPLINE_TYPE, ncurves);
      ALLOCN(solid[i].TrimPortion[r].SplinePortion.AdjInfo,
	     int, ncurves);
      
	for (p=0; p<ncurves; p++) {
	  
	  retval = fscanf(fp, " adjacentpatch %d", 
			  &solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	  if (retval != 1) return 0;
	  
	  retval = fscanf(fp, " uvdegree %d", &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg);
	  if (retval != 1) return 0;
	  uvdeg = solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].Deg; 
	  ALLOCN(solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts,
		 POINT3D, (uvdeg+1));
	  for (q=0; q<=uvdeg; q++) {
	    retval = fscanf(fp, " uv %lf %lf %lf",
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[0],     
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[1], 
		   &solid[i].TrimPortion[r].SplinePortion.UV_Splines[p].CntrlPts[q].dCoord[2]);
	    if (retval != 3) return 0;
	  }
	  retval = fscanf(fp, " xyzdegree %d", &xyzdeg);
	  if (retval != 1) return 0;
	  solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].Deg = xyzdeg;
	  ALLOCN(solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts,
		 POINT4D, (xyzdeg+1));
	  for (q=0; q<=xyzdeg; q++) {
	      retval = fscanf(fp, " xyz %lf %lf %lf %lf", 
		     &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts[q].dCoord[0],
		     &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts[q].dCoord[1], 
		     &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts[q].dCoord[2], 
		     &solid[i].TrimPortion[r].SplinePortion.XYZ_Splines[p].CntrlPts[q].dCoord[3]);
	      if (retval != 4) return 0;
	  }
      }
    }
  }
  return i;
}
