#include <tree.h>
#include <macro.h>
#include <math.h>

#define MAX_ISO_VERTICES 3000
#define MAX_ISO_FACES    3000
#define MAX_FACET_SIZE   200	/* #sides of the trimming polygon */
#define ISO_FACET_SCALE_FACTOR 1.0

/* This version is for this file only */
static int allocate_tensor_patch_of_degree(TRIM_PATCH_INFO *tpI, int udeg, int vdeg) 
{ 
  int j; 
  tpI->tp.iDeg0 = udeg; 
  tpI->tp.iDeg1 = vdeg; 

  allocate_tensor_patch(&tpI->tp);
#if 0
  tpI->tp.coeffs = (POINT4D **) new((udeg + 1)*sizeof(POINT4D *)); 
  for (j = 0; j < udeg + 1; j++) 
    tpI->tp.coeffs[j] = (POINT4D *) new((vdeg + 1)*sizeof(POINT4D)); 
#endif

  return 0; 
} 

  
static int generate_patch_from_polygon(vector_t ivert[], int poly[],
				       int nsides, vector_t vn,
				       int pnum, TRIM_PATCH_INFO *tp)
{
  register int i, j;
  vector_t v0, v1, v;
  matrix_t xmat, xmat2;
  vector_t curve[MAX_FACET_SIZE], corner[4];
  double xmax, xmin, ymax, ymin, z, xscale, yscale, len;
  double xmax_tmp, ymax_tmp, xmin_tmp, ymin_tmp;
  TRIM_PORTION_PIECEWISE_TYPE *tpt;
  point_typ pt;
  
  /* Choose the line joining the first two points to be the new X-axis */
  
  i = poly[0]; j = poly[1];
  v0[0] = ivert[j][0] - ivert[i][0];
  v0[1] = ivert[j][1] - ivert[i][1];
  v0[2] = ivert[j][2] - ivert[i][2];
  len = _LENGTH(v0);
  NORMALISE(v0, len);

  CROSS(vn, v0, v1);		/* Obtain new Y-axis */
  len = _LENGTH(v1);
  NORMALISE(v1, len);
  
  memcpy((void *) &xmat[0], (void *) v0, sizeof(vector_t)); /* X-axis */
  memcpy((void *) &xmat[4], (void *) v1, sizeof(vector_t)); /* Y-axis */
  memcpy((void *) &xmat[8], (void *) vn, sizeof(vector_t)); /* Z-axis */
  xmat[12] = xmat[13] = xmat[14] = 0.0;
  xmat[3] = xmat[7] = xmat[11] = 0.0;
  xmat[15] = 1.0;

  /* Transform all the points so that 'vn' becomes the new Z-axis */
  /* At the same time, compute the bounding box */

  xmax = ymax = -1.0e+10;
  xmin = ymin = 1.0e+10;
  
  for (i = 0; i <= nsides; i++)
    {
      matvec_mult(xmat, ivert[poly[i]], curve[i]);

      if (curve[i][0] > xmax)
	xmax = curve[i][0];
      else if (curve[i][0] < xmin)
	xmin = curve[i][0];

      if (curve[i][1] > ymax)
	ymax = curve[i][1];
      else if (curve[i][1] < ymin)
	ymin = curve[i][1];
    }
  
  z = curve[0][2];
  
  /* Extend the bounding-rectangle by a small amount on all the sides */
  /* to take care of the rendering-problems */
  
  xmin_tmp = xmin;
  ymin_tmp = ymin;
  xmax_tmp = xmax;
  ymax_tmp = ymax;

  xmin = (xmin_tmp + xmax_tmp)/2.0 -
    ISO_FACET_SCALE_FACTOR*(xmax_tmp - xmin_tmp)/2.0;
  xmax = (xmin_tmp + xmax_tmp)/2.0 +
    ISO_FACET_SCALE_FACTOR*(xmax_tmp - xmin_tmp)/2.0;
  ymin = (ymin_tmp + ymax_tmp)/2.0 -
    ISO_FACET_SCALE_FACTOR*(ymax_tmp - ymin_tmp)/2.0;
  ymax = (ymin_tmp + ymax_tmp)/2.0 + 
    ISO_FACET_SCALE_FACTOR*(ymax_tmp - ymin_tmp)/2.0;
  
  xscale = xmax - xmin;
  yscale = ymax - ymin;
  
  /* Get the four corners for the rectangular bounding box */
  corner[0][0] = xmin;
  corner[0][1] = ymin;
  corner[0][2] = z;
  corner[0][3] = 1.0;

  corner[1][0] = xmax;
  corner[1][1] = ymin;
  corner[1][2] = z;
  corner[1][3] = 1.0;

  corner[2][0] = xmax;
  corner[2][1] = ymax;
  corner[2][2] = z;
  corner[2][3] = 1.0;

  corner[3][0] = xmin;
  corner[3][1] = ymax;
  corner[3][2] = z;
  corner[3][3] = 1.0;

  /* project the corner points back to their proper positions to */
  /* get the control points for the patch*/
  
  v0[3] = v1[3] = vn[3] = 0.0;

  xmat2[0] = v0[0];
  xmat2[4] = v0[1];
  xmat2[8] = v0[2];		/* X-axis */
  xmat2[12] = v0[3];

  xmat2[1] = v1[0];
  xmat2[5] = v1[1];
  xmat2[9] = v1[2];		/* Y-axis */
  xmat2[13] = v1[3];

  xmat2[2] = vn[0];
  xmat2[6] = vn[1];
  xmat2[10] = vn[2];		/* Z-axis */
  xmat2[14] = vn[3];

  xmat2[3] = xmat2[7] = xmat2[11] = 0.0;
  xmat2[15] = 1.0;
  
  for (i = 0; i < 4; i++)
    {
      matvec_mult(xmat2, corner[i], v);
      for (j = 0; j < 4; j++)
	corner[i][j] = v[j];
    }
  
  /* Create the tensor patch */

  allocate_tensor_patch_of_degree(tp, 1, 1);
  mypt_to_yourpt(corner[0], tp->tp.coeffs[0][0]);
  mypt_to_yourpt(corner[1], tp->tp.coeffs[1][0]);
  mypt_to_yourpt(corner[3], tp->tp.coeffs[0][1]);
  mypt_to_yourpt(corner[2], tp->tp.coeffs[1][1]);

  /* Initialize other junk */
  
  copy_ten_p(&tp->tp, &tp->orig_tp);
  tp->umin = tp->vmin = 0.0;
  tp->umax = tp->vmax = 1.0;
  tp->primNum = pnum;
  
  tp->csg_status = IN_CSG;
  
  /* Now get the trimming curve */
  
  tp->NumTrimPortions = 1;
  tp->SplineReprn = 0;

  tpt = &(tp->TrimPortion[0].PiecePortion);

  tpt->AdjInfo = NULL;
  tpt->NumOfVertices = nsides;
  tpt->poly_vertex = (double (*)[2]) new((tpt->NumOfVertices+5)*2* 
					 sizeof(double));
  for (i = 0; i <= nsides; i++)
    {
      tpt->poly_vertex[i][0] = (curve[i][0] - xmin) / xscale;
      tpt->poly_vertex[i][1] = (curve[i][1] - ymin) / yscale;
    }
 
  if (!is_curve_anticlockwise(tpt->poly_vertex, nsides))
    {
      double tmp0[2];

      for (i = 0, j = nsides; i < j; i++, j--)
	{
	  tmp0[0] = tpt->poly_vertex[i][0];
	  tmp0[1] = tpt->poly_vertex[i][1];
	  
	  tpt->poly_vertex[i][0] = tpt->poly_vertex[j][0];
	  tpt->poly_vertex[i][1] = tpt->poly_vertex[j][1];

	  tpt->poly_vertex[j][0] = tmp0[0];
	  tpt->poly_vertex[j][1] = tmp0[1];
	}
    }
  
  find_point_inside_polygon(nsides, tpt->poly_vertex, &pt);
  tpt->ptInsideRegion.u = pt.x;
  tpt->ptInsideRegion.v = pt.y;

  return 0;
}


/* Read in isolated solid. Complete trimming information filled in for */
/* the node. No need to call add_trim_patch_info() afterwards. */

int read_isolated_solid(node_t *np, FILE *infile, int pnum)
{
  register int i, j;
  int nverts, nfacets, fcount, ncontours, nsides;
  vector_t ivert[MAX_ISO_VERTICES];
  int poly[MAX_FACET_SIZE];

  double dummy;
  vector_t vn;

  fscanf(infile, "%d%d%lf%lf%lf", &nverts, &nfacets, &dummy,
	 &dummy, &dummy);
  if (nverts < 4)		/* Invalid solid */
    return -1;
  
  /* Read vertices 1 .. nverts */
  for (i = 1; i <= nverts; i++)
    {
      fscanf(infile, "%lf%lf%lf", &ivert[i][0], 
	     &ivert[i][1], &ivert[i][2]);
      ivert[i][3] = 1.0;
    }

  /* Allocate space 'nfacets' number of patches. Each patch is a */
  /* rectangular planar patch, with the facet boundary acting as the */
  /* trimming polygon. */
  
  np->trim = (TRIM_PATCH_INFO *) new(nfacets *
				     sizeof(TRIM_PATCH_INFO));
  np->patch = (TENSOR_PATCH *) new(nfacets *
				     sizeof(TENSOR_PATCH)); 
  fcount = 0;			/* facet-count */

  for (i = 0; i < nfacets; i++)
    {
      fscanf(infile, "%d%lf%lf%lf", &ncontours, &vn[0], &vn[1], &vn[2]);
      if ((ncontours != 1) && (ncontours != -1))
	{
	  fprintf(stderr, "read_isolated_solid: Too many contours\n");
	  free(np->trim); 
	  free (np->patch);
	  return -1;
	}
      
      fscanf(infile, "%d", &nsides);
      for (j = 0; j < nsides; j++)
	{
	  fscanf(infile, "%d", &poly[j]);
	  poly[nsides] = poly[0]; /* first vert == last vert */
	}
      
      generate_patch_from_polygon(ivert, poly, nsides,
				  vn, pnum, &np->trim[fcount]);
      np->patch[fcount] = np->trim[fcount].tp;
      fcount++;
    }
  
  return fcount;
}

