#include <primitives.h>
#include <tree.h>
#include <macro.h>
#include <math.h>
/* added by culver 1/25/96 */
#include <malloc.h>

#ifdef NEW_CYL_REPRN
#define NPOINTS 36
#define CYL_LEEWAY_FACTOR       1.1
#endif

#ifdef NEW_CONE_REPRN
#define NPOINTS 36
#define CONE_LEEWAY_FACTOR       1.1
#endif

#define CLOSE(a, b)  ((fabs((a) - (b)) <= 5.0e-3))

#define MAX(a, b) ((a > b) ? a : b)
#define VASE_LEEWAY_FACTOR 1.1

#define SAMEPTTOL 0.000001

extern vector_t topbot_trimcurve[4][MAX_TRIM_CURVE_POINTS];
extern int top_is_zero;
extern int vase_no_lid[2];

extern int GetSplineFromPatch();
extern int IsPatchFacingOutside();
extern void ReverseCntrlPtsOf3DSpline();

extern struct {
  double x, y;
} topbot_inpoint[4];

int perror_exit(char *s)
{
  perror(s);
  exit(1);;
}


void *new(int size)
{
  void *p;
  
  if (!size)
    return NULL;
  if ((p = malloc(size)) == NULL)
    perror_exit("malloc");
  
  return p;
}

#ifdef __LINK

/* pdis is a function to compute the distance between 2 points */
double pdis(POINT4D p1, POINT4D p2)
{
   double distance;

   /* Note: assumes that w = 1.0 for both points */
   distance = (p1.dCoord[0]-p2.dCoord[0]) * (p1.dCoord[0]-p2.dCoord[0]) +
              (p1.dCoord[1]-p2.dCoord[1]) * (p1.dCoord[1]-p2.dCoord[1]) +
              (p1.dCoord[2]-p2.dCoord[2]) * (p1.dCoord[2]-p2.dCoord[2]);
   distance = sqrt(distance);
   return distance;
}


/* This routine is called whenever the primitive information is read */
/* into a leaf node in the tree */
int add_trim_patch_info(node_t *np,  int chewflag, int ndomains,
			int nl1, int nl2, int pnum)
{
  register int i, j, k, n;
  TRIM_PATCH_INFO *p;
  TRIM_PATCH_INFO *q;
  TRIM_PATCH_INFO *q2;
  TRIM_PORTION_SPLINE_TYPE *sptr;
  int IsItT, coeff1, coeff2, dummy, spl_rep;
  double Val, root3, rad;
  int adj1[6];
  int adj2[6];
  int OK00[6];
  int OK01[6];
  int OK10[6];
  int OK11[6];
  int numcur[6];
  int coloc10;
  XYZ_SPLINE_TYPE *Spl;

  np->trim = (TRIM_PATCH_INFO *) new(np->npatches *
				     sizeof(TRIM_PATCH_INFO));

  rad = 0.45454545454545454545;
  root3 = sqrt(3.0)/2.0;

  p = np->trim;
  for (i = 0; i < np->npatches; i++, p++)
    {
      p->tp = np->patch[i];	/* point in the patch array */
#ifdef PRIM_TEST
      p->orig_tp = np->patch[i];
#else
      copy_ten_p(&p->tp, &p->orig_tp);
#endif
      p->primNum = pnum;

      p->umin = p->vmin = 0.0;
      p->umax = p->vmax = 1.0;

      p->csg_status = IN_CSG;
      p->SplineReprn = 1;
      p->NumTrimPortions = 1;
      if (!p->SplineReprn) {
	p->TrimPortion[0].PiecePortion.NumOfVertices = 4;
	p->TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
	p->TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;
	
	p->TrimPortion[0].PiecePortion.poly_vertex = 
	  (double (*)[2]) malloc(MAX_VERTEX_COUNT * 2 * sizeof(double *));
	p->TrimPortion[0].PiecePortion.poly_vertex[0][0] = 0.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[0][1] = 0.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[1][1] = 0.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[3][0] = 0.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[4][0] = 0.0;
	p->TrimPortion[0].PiecePortion.poly_vertex[4][1] = 0.0;
      } else {
	sptr = &(p->TrimPortion[0].SplinePortion);
	sptr->UV_Splines = (UV_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION * 
						  sizeof(UV_SPLINE_TYPE));
	
	sptr->XYZ_Splines = (XYZ_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION * 
						    sizeof(XYZ_SPLINE_TYPE));
	
	sptr->AdjInfo = (int *) new(MAX_SPLINES_PER_TRIM_REGION * sizeof(int));
      }
    }

  spl_rep = 1;
  if (spl_rep) {
    switch (chewflag) {
    case P_CONE:
    case P_CYLINDER:
      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
	sptr = &(p->TrimPortion[0].SplinePortion);
	if (i < 3)
	  sptr->NumOfCurves = 4;

	else
	  sptr->NumOfCurves = 3;

	for (j=0; j<sptr->NumOfCurves; j++) {
	  if (i < 3)
	    sptr->UV_Splines[j].Deg = 1;
	  else
	    sptr->UV_Splines[j].Deg = 2;
	  sptr->UV_Splines[j].CntrlPts = (POINT3D *) 
	                                 new((sptr->UV_Splines[j].Deg + 1) * 
					     sizeof(POINT3D));
	}
	sptr->ptInsideRegion.u = 0.5;
	sptr->ptInsideRegion.v = 0.5;
      }
      
      p = np->trim;
      for (i=0; i<3; i++, p++) {
	sptr = &(p->TrimPortion[0].SplinePortion);
	for (j=0; j<sptr->NumOfCurves; j++) {
	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;

	  sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;

	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;

	  sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;

	  if (j % 2)
	    IsItT = 0;
	  else
	    IsItT = 1;

	  if ((j==1) || (j==2)) 
	    Val = 1.0;
	  else
	    Val = 0.0;
	  if (GetSplineFromPatch(&(p->tp), IsItT, Val,
				 &(sptr->XYZ_Splines[j]))) {
	    printf("Could not get spline properly\n");
	    exit(-1);
	  }
	  if (j > 1)
	    ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));

	  if ((j == 0) || (j == 2))
	    sptr->AdjInfo[j] = (i + 2 - (j>>1)) % 3;
	  if (j==1)
	    sptr->AdjInfo[j] = 4;
	  if (j==3)
	    sptr->AdjInfo[j] = 3;
	}
      }

      p = np->trim + 3;
      for (i=3; i<5; i++, p++) {
	sptr = &(p->TrimPortion[0].SplinePortion);
	/* curve 0 */
	sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.5;
	sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.5 - rad;
	sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;

	sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = root3*rad + 0.25;
	sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = -0.5*rad + 0.25;
	sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 0.5;

	sptr->UV_Splines[0].CntrlPts[2].dCoord[0] = root3*rad + 0.5;
	sptr->UV_Splines[0].CntrlPts[2].dCoord[1] = 0.5*rad + 0.5;
	sptr->UV_Splines[0].CntrlPts[2].dCoord[2] = 1.0;
	
	if (i == 3) 
	  sptr->AdjInfo[0] = 1;
	else
	  sptr->AdjInfo[0] = 0;

	/* curve 1 */
	sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = root3*rad + 0.5;
	sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 0.5*rad + 0.5;
	sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;

	sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 0.25;
	sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = rad + 0.25;
	sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 0.5;

	sptr->UV_Splines[1].CntrlPts[2].dCoord[0] = -root3*rad + 0.5;
	sptr->UV_Splines[1].CntrlPts[2].dCoord[1] = 0.5*rad + 0.5;
	sptr->UV_Splines[1].CntrlPts[2].dCoord[2] = 1.0;

	sptr->AdjInfo[1] = 2; /* both 3 and 4 is adjacent to patch 2 */

	/* curve 2 */
	sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = -root3*rad + 0.5;
	sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 0.5*rad + 0.5;
	sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;

	sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = -root3*rad + 0.25;
	sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = -0.5*rad + 0.25;
	sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 0.5;

	sptr->UV_Splines[2].CntrlPts[2].dCoord[0] = 0.5;
	sptr->UV_Splines[2].CntrlPts[2].dCoord[1] = -rad + 0.5;
	sptr->UV_Splines[2].CntrlPts[2].dCoord[2] = 1.0;

	if (i == 3) 
	  sptr->AdjInfo[2] = 0;
	else
	  sptr->AdjInfo[2] = 1;
      }
      /* xyz curves */
      p = np->trim + 3;
      sptr = &(p->TrimPortion[0].SplinePortion);
      for (j=0; j<sptr->NumOfCurves; j++)
	GetSplineFromPatch(&(np->trim[sptr->AdjInfo[j]].tp), 0, 0.0,
			   &(sptr->XYZ_Splines[j]));
      
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      for (j=0; j<sptr->NumOfCurves; j++) {
	GetSplineFromPatch(&(np->trim[sptr->AdjInfo[j]].tp), 0, 1.0,
			   &(sptr->XYZ_Splines[j]));
	ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));
      }
      break;
/* begin KEYSER */

case P_BRLTGC:
      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        if (i < 4)
          sptr->NumOfCurves = 4;

        else
          sptr->NumOfCurves = 4;

        for (j=0; j<sptr->NumOfCurves; j++) {
          if (i < 4)
            sptr->UV_Splines[j].Deg = 1;
          else
            sptr->UV_Splines[j].Deg = 2;
          sptr->UV_Splines[j].CntrlPts = (POINT3D *)
                                         new((sptr->UV_Splines[j].Deg + 1) *
                                             sizeof(POINT3D));
        }
        sptr->ptInsideRegion.u = 0.5;
        sptr->ptInsideRegion.v = 0.5;
      }

      p = np->trim;
      for (i=0; i<4; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        for (j=0; j<sptr->NumOfCurves; j++) {

          switch (j) {
            case 0:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 1;
              Val = 0.0;
              break;
            case 1:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 0;
              Val = 1.0;
              break;
            case 2:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 1;
              Val = 1.0;
              break;
            case 3:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 0;
              Val = 0.0;
              break;
          }

          if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                 &(sptr->XYZ_Splines[j]))) {
            printf("Could not get spline properly\n");
            exit(-1);
          }

          if (j>1)
            ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));

          /* Fill in adjacency info */
          if (j==0)
            sptr->AdjInfo[j] = 4;
          if (j==2)
            sptr->AdjInfo[j] = 5;
          if (j==1)
            sptr->AdjInfo[j] = (i+1) % 4;
          if (j==3)
            sptr->AdjInfo[j] = (i+3) % 4;
        }
      }

      p = np->trim + 4;
      for (i=4; i<6; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        /* curve 0 */
        sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.5;
        sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.5-rad;
        sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;

        sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = 0.5*(0.5+rad);
        sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = 0.5*(0.5-rad);
        sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 0.5;

        sptr->UV_Splines[0].CntrlPts[2].dCoord[0] = 0.5+rad;
        sptr->UV_Splines[0].CntrlPts[2].dCoord[1] = 0.5;
        sptr->UV_Splines[0].CntrlPts[2].dCoord[2] = 1.0;

        if (i == 4)
          sptr->AdjInfo[0] = 1;
        else
          sptr->AdjInfo[0] = 0;

        /* curve 1 */
        sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = 0.5+rad;
        sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 0.5;
        sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;

        sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 0.5*(0.5+rad);
        sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = 0.5*(0.5+rad);
        sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 0.5;

        sptr->UV_Splines[1].CntrlPts[2].dCoord[0] = 0.5;
        sptr->UV_Splines[1].CntrlPts[2].dCoord[1] = 0.5+rad;
        sptr->UV_Splines[1].CntrlPts[2].dCoord[2] = 1.0;

        if (i == 4)
          sptr->AdjInfo[1] = 0;
        else
          sptr->AdjInfo[1] = 1;


        /* curve 2 */
        sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = 0.5;
        sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 0.5+rad;
        sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;

        sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = 0.5*(0.5-rad);
        sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = 0.5*(0.5+rad);
        sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 0.5;

        sptr->UV_Splines[2].CntrlPts[2].dCoord[0] = 0.5-rad;
        sptr->UV_Splines[2].CntrlPts[2].dCoord[1] = 0.5;
        sptr->UV_Splines[2].CntrlPts[2].dCoord[2] = 1.0;

        if (i == 4)
          sptr->AdjInfo[2] = 3;
        else
          sptr->AdjInfo[2] = 2;

	/* curve 3 */
        sptr->UV_Splines[3].CntrlPts[0].dCoord[0] = 0.5-rad;
        sptr->UV_Splines[3].CntrlPts[0].dCoord[1] = 0.5;
        sptr->UV_Splines[3].CntrlPts[0].dCoord[2] = 1.0;

        sptr->UV_Splines[3].CntrlPts[1].dCoord[0] = 0.5*(0.5-rad);
        sptr->UV_Splines[3].CntrlPts[1].dCoord[1] = 0.5*(0.5-rad);
        sptr->UV_Splines[3].CntrlPts[1].dCoord[2] = 0.5;

        sptr->UV_Splines[3].CntrlPts[2].dCoord[0] = 0.5;
        sptr->UV_Splines[3].CntrlPts[2].dCoord[1] = 0.5-rad;
        sptr->UV_Splines[3].CntrlPts[2].dCoord[2] = 1.0;

        if (i == 4)
          sptr->AdjInfo[3] = 2;
        else
          sptr->AdjInfo[3] = 3;


      }
      /* xyz curves */
      p = np->trim + 4;
      sptr = &(p->TrimPortion[0].SplinePortion);
      for (j=0; j<sptr->NumOfCurves; j++) {
        GetSplineFromPatch(&(np->trim[sptr->AdjInfo[j]].tp), 1, 0.0,
                           &(sptr->XYZ_Splines[j]));
        ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));
      }
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      for (j=0; j<sptr->NumOfCurves; j++)
        GetSplineFromPatch(&(np->trim[sptr->AdjInfo[j]].tp), 1, 1.0,
                           &(sptr->XYZ_Splines[j]));
      break;

    case P_BRLARB8:

      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        OK00[i] = 1;
        OK01[i] = 0;
        OK10[i] = 0;
        OK11[i] = 0;
        if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[0][0]) > SAMEPTTOL) OK01[i]=1;
        if ((pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) &&
            ( !(OK01[i]) || 
              (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][1]) > SAMEPTTOL) 
            )
           ) OK10[i]=1;
        if ((pdis(p->tp.coeffs[1][1],p->tp.coeffs[0][0]) > SAMEPTTOL) &&
            ( !(OK01[i]) || 
              (pdis(p->tp.coeffs[1][1],p->tp.coeffs[0][1]) > SAMEPTTOL)
            ) &&
            ( !(OK10[i]) || 
              (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL)
            )
           ) OK11[i]=1;

        sptr->NumOfCurves = OK00[i]+OK01[i]+OK10[i]+OK11[i];
        numcur[i] = sptr->NumOfCurves;
        if (sptr->NumOfCurves < 3) sptr->NumOfCurves = 0;

        /* If the patch is an edge, determine two faces it is adjacent to */
        if (numcur[i] == 2) {
           switch(i) {
              case 0:
                 adj1[0] = -1;
                 adj2[0] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[0]=4;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) {
                    if (adj1[0]<0) adj1[0]=3;
                    else adj2[0]=3;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[0]<0) adj1[0]=2;
                    else adj2[0]=2;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[0]<0) adj1[0]=5;
                    else adj2[0]=5;
                 }
                 break;
              case 1:
                 adj1[1] = -1;
                 adj2[1] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[1]=2;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) { 
                    if (adj1[1]<0) adj1[1]=3;
                    else adj2[1]=3;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[1]<0) adj1[1]=4;
                    else adj2[1]=4;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[1]<0) adj1[1]=5;
                    else adj2[1]=5;
                 }
                 break;
              case 2:
                 adj1[2] = -1;
                 adj2[2] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[2]=0;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) { 
                    if (adj1[2]<0) adj1[2]=3;
                    else adj2[2]=3;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[2]<0) adj1[2]=1;
                    else adj2[2]=1;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[2]<0) adj1[2]=5;
                    else adj2[2]=5;
                 }
                 break;
              case 3:
                 adj1[3] = -1;
                 adj2[3] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[3]=0;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) { 
                    if (adj1[3]<0) adj1[3]=4;
                    else adj2[3]=4;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[3]<0) adj1[3]=1;
                    else adj2[3]=1;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[3]<0) adj1[3]=2;
                    else adj2[3]=2;
                 }
                 break;
              case 4:
                 adj1[4] = -1;
                 adj2[4] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[4]=0;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) { 
                    if (adj1[4]<0) adj1[4]=5;
                    else adj2[4]=5;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[4]<0) adj1[4]=1;
                    else adj2[4]=1;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[4]<0) adj1[4]=3;
                    else adj2[4]=3;
                 }
                 break;
              case 5:
                 adj1[5] = -1;
                 adj2[5] = -1;
                 if (pdis(p->tp.coeffs[1][0],p->tp.coeffs[0][0]) > SAMEPTTOL) {
                    adj1[5]=0;
                 }
                 if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) { 
                    if (adj1[5]<0) adj1[5]=2;
                    else adj2[5]=2;
                 }
                 if (pdis(p->tp.coeffs[0][1],p->tp.coeffs[1][1]) > SAMEPTTOL) { 
                    if (adj1[5]<0) adj1[5]=1;
                    else adj2[5]=1;
                 }
                 if (pdis(p->tp.coeffs[0][0],p->tp.coeffs[0][1]) > SAMEPTTOL) { 
                    if (adj1[5]<0) adj1[5]=4;
                    else adj2[5]=4;
                 }
                 break;
           }
        }

        for (j=0; j<sptr->NumOfCurves; j++) {
          sptr->UV_Splines[j].Deg = 1;
          sptr->UV_Splines[j].CntrlPts = (POINT3D *)
                                         new((sptr->UV_Splines[j].Deg + 1) *
                                             sizeof(POINT3D));
        }
      }

      p = np->trim;
      for (i=0; i<6; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        if (sptr->NumOfCurves == 4) {
           /* Make UV Spline Curve 0 */
           sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.0;
           sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.0;
           sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;
           sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = 1.0;
           sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = 0.0;
           sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 1.0;
           /* Make XYZ Spline Curve 0 */
           IsItT = 1;
           Val = 0.0;
           if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                  &(sptr->XYZ_Splines[0]))) {
              printf("Could not get spline properly\n");
              exit(-1);
           }
           /* Fill in adjacency information for Curve 0 */
           switch(i) {
              case 0:
                 n=4;
                 break;
              case 1:
                 n=2;
                 break;
              case 2:
                 n=0;
                 break;
              case 3:
                 n=0;
                 break;
              case 4:
                 n=0;
                 break;
              case 5:
                 n=0;
                 break;
           }
           if (numcur[n] > 2) sptr->AdjInfo[0]=n;
           else 
              if (adj1[n] == i) sptr->AdjInfo[0]=adj2[n];
              else sptr->AdjInfo[0]=adj1[n];

           /* Make UV Spline Curve 1 */
           sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = 1.0;
           sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 0.0;
           sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;
           sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 1.0;
           sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = 1.0;
           sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 1.0;
           /* Make XYZ Spline Curve 1 */
           IsItT = 0;
           Val = 1.0;
           if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                  &(sptr->XYZ_Splines[1]))) {
              printf("Could not get spline properly\n");
              exit(-1);
           }
           /* Fill in adjacency information for Curve 1 */
           switch(i) {
              case 0:
                 n=3;
                 break;
              case 1:
                 n=3;
                 break;
              case 2:
                 n=3;
                 break;
              case 3:
                 n=4;
                 break;
              case 4:
                 n=5;
                 break;
              case 5:
                 n=2;
                 break;
           }
           if (numcur[n] > 2) sptr->AdjInfo[1]=n;
           else 
              if (adj1[n] == i) sptr->AdjInfo[1]=adj2[n];
              else sptr->AdjInfo[1]=adj1[n];

           /* Make UV Spline Curve 2 */
           sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = 1.0;
           sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 1.0;
           sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;
           sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = 0.0;
           sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = 1.0;
           sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 1.0;
           /* Make XYZ Spline Curve 2 */
           IsItT = 1;
           Val = 1.0;
           if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                  &(sptr->XYZ_Splines[2]))) {
              printf("Could not get spline properly\n");
              exit(-1);
           }
           ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[2]));
           /* Fill in adjacency information for Curve 2 */
           switch(i) {
              case 0:
                 n=2;
                 break;
              case 1:
                 n=4;
                 break;
              case 2:
                 n=1;
                 break;
              case 3:
                 n=1;
                 break;
              case 4:
                 n=1;
                 break;
              case 5:
                 n=1;
                 break;
           }
           if (numcur[n] > 2) sptr->AdjInfo[2]=n;
           else
              if (adj1[n] == i) sptr->AdjInfo[2]=adj2[n];
              else sptr->AdjInfo[2]=adj1[n];

           /* Make UV Spline Curve 3 */
           sptr->UV_Splines[3].CntrlPts[0].dCoord[0] = 0.0;
           sptr->UV_Splines[3].CntrlPts[0].dCoord[1] = 1.0;
           sptr->UV_Splines[3].CntrlPts[0].dCoord[2] = 1.0;
           sptr->UV_Splines[3].CntrlPts[1].dCoord[0] = 0.0;
           sptr->UV_Splines[3].CntrlPts[1].dCoord[1] = 0.0;
           sptr->UV_Splines[3].CntrlPts[1].dCoord[2] = 1.0;
           /* Make XYZ Spline Curve 3 */
           IsItT = 0;
           Val = 0.0;
           if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                  &(sptr->XYZ_Splines[3]))) {
              printf("Could not get spline properly\n");
              exit(-1);
           }
           ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[3]));
           /* Fill in adjacency information for Curve 3 */
           switch(i) {
              case 0:
                 n=5;
                 break;
              case 1:
                 n=5;
                 break;
              case 2:
                 n=5;
                 break;
              case 3:
                 n=2;
                 break;
              case 4:
                 n=3;
                 break;
              case 5:
                 n=4;
                 break;
           }
           if (numcur[n] > 2) sptr->AdjInfo[3]=n;
           else
              if (adj1[n] == i) sptr->AdjInfo[3]=adj2[n];
              else sptr->AdjInfo[3]=adj1[n];

           /* Identify point inside region */
           sptr->ptInsideRegion.u = 0.5;
           sptr->ptInsideRegion.v = 0.5;
        } else if (sptr->NumOfCurves == 3) {
           if (!(OK11[i])) {
              /* Control point [1][1] is invalid */

              /* determine whether 11 is colocated with 10 or 01 */
              if (pdis(p->tp.coeffs[1][1],p->tp.coeffs[1][0]) > SAMEPTTOL) 
                   coloc10 = 0;
              else coloc10 = 1;

              /* first determine new position for control point */
              p->tp.coeffs[1][1].dCoord[0] = p->tp.coeffs[1][0].dCoord[0] +
                                             p->tp.coeffs[0][1].dCoord[0] -
                                             p->tp.coeffs[0][0].dCoord[0];
              p->tp.coeffs[1][1].dCoord[1] = p->tp.coeffs[1][0].dCoord[1] +
                                             p->tp.coeffs[0][1].dCoord[1] -
                                             p->tp.coeffs[0][0].dCoord[1];
              p->tp.coeffs[1][1].dCoord[2] = p->tp.coeffs[1][0].dCoord[2] +
                                             p->tp.coeffs[0][1].dCoord[2] -
                                             p->tp.coeffs[0][0].dCoord[2];
              p->tp.coeffs[1][1].dCoord[3] = 1.0;

              p->orig_tp.coeffs[1][1].dCoord[0] =
                       p->tp.coeffs[1][1].dCoord[0];
              p->orig_tp.coeffs[1][1].dCoord[1] =
                       p->tp.coeffs[1][1].dCoord[1];
              p->orig_tp.coeffs[1][1].dCoord[2] =
                       p->tp.coeffs[1][1].dCoord[2];
              p->orig_tp.coeffs[1][1].dCoord[3] =
                       p->tp.coeffs[1][1].dCoord[3];

              /* now set up Spline Curves */

              /* Set up UV Spline Curve 0 */
              sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 0 */
              Spl = &(sptr->XYZ_Splines[0]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[1][0].dCoord[j];
              }
              /* Fill in adjacency information for Curve 0 */
              switch(i) {
                 case 0:
                    n=4;
                    break;
                 case 1:
                    n=2;
                    break;
                 case 2:
                    n=0;
                    break;
                 case 3:
                    n=0;
                    break;
                 case 4:
                    n=0;
                    break;
                 case 5:
                    n=0;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[0]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[0]=adj2[n];
                 else sptr->AdjInfo[0]=adj1[n];

              /* Set up UV Spline Curve 1 */
              sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 1 */
              Spl = &(sptr->XYZ_Splines[1]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[1][0].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[0][1].dCoord[j];
              }
              /* Fill in adjacency information for Curve 1 */
              if (coloc10) {
                 /* Note: 11 and 10 are colocated, so Curve 2 must lie along
                           the original 01 -> 11 line */
                 switch(i) {
                    case 0:
                       n=2;
                       break;
                    case 1:
                       n=4;
                       break;
                    case 2:
                       n=1;
                       break;
                    case 3:
                       n=1;
                       break;
                    case 4:
                       n=1;
                       break;
                    case 5:
                       n=1;
                       break;
                 }
                 if (numcur[n] > 2) sptr->AdjInfo[1]=n;
                 else
                    if (adj1[n] == i) sptr->AdjInfo[1]=adj2[n];
                    else sptr->AdjInfo[1]=adj1[n];
              } else {
                 /* Note: 11 and 01 are colocated, so Curve 2 must lie along
                           the original 10 -> 11 line */
                 switch(i) {
                    case 0:
                       n=3;
                       break;
                    case 1:
                       n=3;
                       break;
                    case 2:
                       n=3;
                       break;
                    case 3:
                       n=4;
                       break;
                    case 4:
                       n=5;
                       break;
                    case 5:
                       n=2;
                       break;
                 }
                 if (numcur[n] > 2) sptr->AdjInfo[1]=n;
                 else
                    if (adj1[n] == i) sptr->AdjInfo[1]=adj2[n];
                    else sptr->AdjInfo[1]=adj1[n];
              }

              /* Set up UV Spline Curve 2 */
              sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 2 */
              Spl = &(sptr->XYZ_Splines[2]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[0][1].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
              }
              /* Fill in adjacency information for Curve 2 */
              switch(i) {
                 case 0:
                    n=5;
                    break;
                 case 1:
                    n=5;
                    break;
                 case 2:
                    n=5;
                    break;
                 case 3:
                    n=2;
                    break;
                 case 4:
                    n=3;
                    break;
                 case 5:
                    n=4;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[2]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[2]=adj2[n];
                 else sptr->AdjInfo[2]=adj1[n];

              /* Identify point inside region */
              sptr->ptInsideRegion.u = 0.25;
              sptr->ptInsideRegion.v = 0.25;
           } else if (!(OK10[i])) {
              /* Control point[1][0] is invalid */
              /* first determine new position for control point */
              p->tp.coeffs[1][0].dCoord[0] = p->tp.coeffs[1][1].dCoord[0] +
                                             p->tp.coeffs[0][0].dCoord[0] -
                                             p->tp.coeffs[0][1].dCoord[0];
              p->tp.coeffs[1][0].dCoord[1] = p->tp.coeffs[1][1].dCoord[1] +
                                             p->tp.coeffs[0][0].dCoord[1] -
                                             p->tp.coeffs[0][1].dCoord[1];
              p->tp.coeffs[1][0].dCoord[2] = p->tp.coeffs[1][1].dCoord[2] +
                                             p->tp.coeffs[0][0].dCoord[2] -
                                             p->tp.coeffs[0][1].dCoord[2];
              p->tp.coeffs[1][0].dCoord[3] = 1.0;

              p->orig_tp.coeffs[1][0].dCoord[0] =
                       p->tp.coeffs[1][0].dCoord[0];
              p->orig_tp.coeffs[1][0].dCoord[1] =
                       p->tp.coeffs[1][0].dCoord[1];
              p->orig_tp.coeffs[1][0].dCoord[2] =
                       p->tp.coeffs[1][0].dCoord[2];
              p->orig_tp.coeffs[1][0].dCoord[3] =
                       p->tp.coeffs[1][0].dCoord[3];

              /* now set up Spline Curves */

              /* Set up UV Spline Curve 0 */
              sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 0 */
              Spl = &(sptr->XYZ_Splines[0]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[1][1].dCoord[j];
              }
              /* Fill in adjacency information for Curve 0 */
              /* Note: 00 and 10 are colocated, so Curve 2 must lie along
                        the original 10 -> 11 line */
              switch(i) {
                 case 0:
                    n=3;
                    break;
                 case 1:
                    n=3;
                    break;
                 case 2:
                    n=3;
                    break;
                 case 3:
                    n=4;
                    break;
                 case 4:
                    n=5;
                    break;
                 case 5:
                    n=2;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[0]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[0]=adj2[n];
                 else sptr->AdjInfo[0]=adj1[n];

              /* Set up UV Spline Curve 1 */
              sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 1 */
              Spl = &(sptr->XYZ_Splines[1]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[1][1].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[0][1].dCoord[j];
              }
              /* Fill in adjacency information for Curve 1 */
              switch(i) {
                 case 0:
                    n=2;
                    break;
                 case 1:
                    n=4;
                    break;
                 case 2:
                    n=1;
                    break;
                 case 3:
                    n=1;
                    break;
                 case 4:
                    n=1;
                    break;
                 case 5:
                    n=1;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[1]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[1]=adj2[n];
                 else sptr->AdjInfo[1]=adj1[n];

              /* Set up UV Spline Curve 2 */
              sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 2 */
              Spl = &(sptr->XYZ_Splines[2]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[0][1].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
              }
              /* Fill in adjacency information for Curve 2 */
              switch(i) {
                 case 0:
                    n=5;
                    break;
                 case 1:
                    n=5;
                    break;
                 case 2:
                    n=5;
                    break;
                 case 3:
                    n=2;
                    break;
                 case 4:
                    n=3;
                    break;
                 case 5:
                    n=4;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[2]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[2]=adj2[n];
                 else sptr->AdjInfo[2]=adj1[n];

              /* Identify point inside region */
              sptr->ptInsideRegion.u = 0.25;
              sptr->ptInsideRegion.v = 0.75;
           } else {
              /* Control point[0][1] is invalid */
              /* first determine new position for control point */
              p->tp.coeffs[0][1].dCoord[0] = p->tp.coeffs[1][1].dCoord[0] +
                                             p->tp.coeffs[0][0].dCoord[0] -
                                             p->tp.coeffs[1][0].dCoord[0];
              p->tp.coeffs[0][1].dCoord[1] = p->tp.coeffs[1][1].dCoord[1] +
                                             p->tp.coeffs[0][0].dCoord[1] -
                                             p->tp.coeffs[1][0].dCoord[1];
              p->tp.coeffs[0][1].dCoord[2] = p->tp.coeffs[1][1].dCoord[2] +
                                             p->tp.coeffs[0][0].dCoord[2] -
                                             p->tp.coeffs[1][0].dCoord[2];
              p->tp.coeffs[0][1].dCoord[3] = 1.0;

              /* now set up Spline Curves */

              /* Set up UV Spline Curve 0 */
              sptr->UV_Splines[0].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[0].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[0].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 0 */
              Spl = &(sptr->XYZ_Splines[0]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[1][0].dCoord[j];
              }
              /* Fill in adjacency information for Curve 0 */
              switch(i) {
                 case 0:
                    n=4;
                    break;
                 case 1:
                    n=2;
                    break;
                 case 2:
                    n=0;
                    break;
                 case 3:
                    n=0;
                    break;
                 case 4:
                    n=0;
                    break;
                 case 5:
                    n=0;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[0]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[0]=adj2[n];
                 else sptr->AdjInfo[0]=adj1[n];

              /* Set up UV Spline Curve 1 */
              sptr->UV_Splines[1].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[1].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[1].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 1 */
              Spl = &(sptr->XYZ_Splines[1]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[1][0].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[1][1].dCoord[j];
              }
              /* Fill in adjacency information for Curve 1 */
              switch(i) {
                 case 0:
                    n=3;
                    break;
                 case 1:
                    n=3;
                    break;
                 case 2:
                    n=3;
                    break;
                 case 3:
                    n=4;
                    break;
                 case 4:
                    n=5;
                    break;
                 case 5:
                    n=2;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[1]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[1]=adj2[n];
                 else sptr->AdjInfo[1]=adj1[n];

              /* Set up UV Spline Curve 2 */
              sptr->UV_Splines[2].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[2].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[2].CntrlPts[1].dCoord[2] = 1.0;
              /* Set up XYZ Spline Curve 2 */
              Spl = &(sptr->XYZ_Splines[2]);
              Spl->Deg = 1;
              Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) *sizeof(POINT4D));
              for(j=0;j<4;j++) {
                 Spl->CntrlPts[0].dCoord[j] = p->tp.coeffs[1][1].dCoord[j];
                 Spl->CntrlPts[1].dCoord[j] = p->tp.coeffs[0][0].dCoord[j];
              }
              /* Fill in adjacency information for Curve 2 */
              /* Note: 00 and 01 are colocated, so Curve 2 must lie along
			the original 01 -> 11 line */ 
              switch(i) {
                 case 0:
                    n=2;
                    break;
                 case 1:
                    n=4;
                    break;
                 case 2:
                    n=1;
                    break;
                 case 3:
                    n=1;
                    break;
                 case 4:
                    n=1;
                    break;
                 case 5:
                    n=1;
                    break;
              }
              if (numcur[n] > 2) sptr->AdjInfo[2]=n;
              else
                 if (adj1[n] == i) sptr->AdjInfo[2]=adj2[n];
                 else sptr->AdjInfo[2]=adj1[n];

              /* Identify point inside region */
              sptr->ptInsideRegion.u = 0.75;
              sptr->ptInsideRegion.v = 0.25;
           }
        } else {
           /* No trim curves - not a true face so do nothing*/
        }
      }

      /* Now remove any patches which have fewer than 3 vertices */

      i=0;
      while (i<np->npatches) {
         if(numcur[i] > 2) {
            i++;
         } else {
            /* this patch needs to be removed */
            p = np->trim + i;
            q = np->trim + i + 1;
            for (j=i+1;j<np->npatches;j++) {
               numcur[j-1]=numcur[j];
               free_trim_patch(p);
               copy_trim_patch(q,p);
               p++;
               if (j<(np->npatches-1)) q++;
            }
            q2 = np->trim + (np->npatches-1);
            free_trim_patch(q2);
            np->npatches = np->npatches-1;
            for (j=0;j<np->npatches;j++) {
               q2=np->trim + j;
               sptr = &(q2->TrimPortion[0].SplinePortion);
               for (k=0;k<sptr->NumOfCurves;k++)
                  if (sptr->AdjInfo[k] > i) sptr->AdjInfo[k]--;
            }
         }
      }

      break;
/* end KEYSER */

    case P_PPPED:
      /* For parallelopiped, ndomains actually gives 'sense' which */
      /* indicates how the vectors are oriented for adjacency info */
      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
	sptr = &(p->TrimPortion[0].SplinePortion);
	sptr->NumOfCurves = 4;
	for (j=0; j<sptr->NumOfCurves; j++) {
	  sptr->UV_Splines[j].Deg = 1;
	  sptr->UV_Splines[j].CntrlPts = (POINT3D *) 
	                                 new((sptr->UV_Splines[j].Deg + 1) * 
					     sizeof(POINT3D));
	  sptr->XYZ_Splines[j].Deg = 1;
	  sptr->XYZ_Splines[j].CntrlPts = (POINT4D *) 
	                                 new((sptr->XYZ_Splines[j].Deg + 1) * 
					     sizeof(POINT4D));
	  /* Initialize UV splines */
	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;

	  sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;

	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;

	  sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;

	  /* Initialize XYZ splines */
	  for (k=0; k<2; k++) {
	    coeff1 = (int) sptr->UV_Splines[j].CntrlPts[k].dCoord[0];
	    coeff2 = (int) sptr->UV_Splines[j].CntrlPts[k].dCoord[1];
	    sptr->XYZ_Splines[j].CntrlPts[k] = p->tp.coeffs[coeff1][coeff2];
	  }
	}
	sptr->ptInsideRegion.u = 0.5;
	sptr->ptInsideRegion.v = 0.5;
      }

      /* Initialize adjacency info */
      /* Patch 0 */
      p = np->trim;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 5; sptr->AdjInfo[1] = 3;
	sptr->AdjInfo[2] = 1; sptr->AdjInfo[3] = 2;
      } else {
	sptr->AdjInfo[0] = 1; sptr->AdjInfo[1] = 3;
	sptr->AdjInfo[2] = 5; sptr->AdjInfo[3] = 2;
      }
      
      /* Patch 1 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 0; sptr->AdjInfo[1] = 3;
	sptr->AdjInfo[2] = 4; sptr->AdjInfo[3] = 2;
      } else {
	sptr->AdjInfo[0] = 2; sptr->AdjInfo[1] = 4;
	sptr->AdjInfo[2] = 3; sptr->AdjInfo[3] = 0;
      }
      
      /* Patch 2 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 5; sptr->AdjInfo[1] = 0;
	sptr->AdjInfo[2] = 1; sptr->AdjInfo[3] = 4;
      } else {
	sptr->AdjInfo[0] = 0; sptr->AdjInfo[1] = 5;
	sptr->AdjInfo[2] = 4; sptr->AdjInfo[3] = 1;
      }
      
      /* Patch 3 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 5; sptr->AdjInfo[1] = 4;
	sptr->AdjInfo[2] = 1; sptr->AdjInfo[3] = 0;
      } else {
	sptr->AdjInfo[0] = 0; sptr->AdjInfo[1] = 1;
	sptr->AdjInfo[2] = 4; sptr->AdjInfo[3] = 5;
      }
      
      /* Patch 4 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 5; sptr->AdjInfo[1] = 2;
	sptr->AdjInfo[2] = 1; sptr->AdjInfo[3] = 3;
      } else {
	sptr->AdjInfo[0] = 3; sptr->AdjInfo[1] = 1;
	sptr->AdjInfo[2] = 2; sptr->AdjInfo[3] = 5;
      }
      
      /* Patch 5 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      if (ndomains) {
	sptr->AdjInfo[0] = 4; sptr->AdjInfo[1] = 3;
	sptr->AdjInfo[2] = 0; sptr->AdjInfo[3] = 2;
      } else {
	sptr->AdjInfo[0] = 2; sptr->AdjInfo[1] = 0;
	sptr->AdjInfo[2] = 3; sptr->AdjInfo[3] = 4;
      }
      break;

    case P_BRLELL:
      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
        sptr = &(p->TrimPortion[0].SplinePortion);
        sptr->NumOfCurves = 4;
        for (j=0; j<sptr->NumOfCurves; j++) {
          sptr->UV_Splines[j].Deg = 1;
          sptr->UV_Splines[j].CntrlPts = (POINT3D *)
                                         new((sptr->UV_Splines[j].Deg + 1) *
                                             sizeof(POINT3D));
          sptr->XYZ_Splines[j].Deg = 2;
          sptr->XYZ_Splines[j].CntrlPts = (POINT4D *)
                                         new((sptr->XYZ_Splines[j].Deg + 1) *
                                             sizeof(POINT4D));
          /* Initialize UV splines */
          switch (j) {
            case 0:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 1;
              Val = 0.0;
              break;
            case 1:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 0;
              Val = 1.0;
              break;
            case 2:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 1;
              Val = 1.0;
              break;
            case 3:
              sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;
              sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;
              sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;
              IsItT = 0;
              Val = 0.0;
              break;
          }

          if (GetSplineFromPatch(&(p->tp), IsItT, Val,
                                 &(sptr->XYZ_Splines[j]))) {
            printf("Could not get spline properly\n");
            exit(-1);
          }

          if (j>1)
              ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));
        }
        sptr->ptInsideRegion.u = 0.5;
        sptr->ptInsideRegion.v = 0.5;
      }
      /* set adjacency info */
      /* patch 0 */
      p=np->trim;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=3;
      sptr->AdjInfo[1]=4;
      sptr->AdjInfo[2]=1;
      sptr->AdjInfo[3]=2;
      /* patch 1 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=0;
      sptr->AdjInfo[1]=5;
      sptr->AdjInfo[2]=2;
      sptr->AdjInfo[3]=3;
      /* patch 2 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=1;
      sptr->AdjInfo[1]=6;
      sptr->AdjInfo[2]=3;
      sptr->AdjInfo[3]=0;
      /* patch 3 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=2;
      sptr->AdjInfo[1]=7;
      sptr->AdjInfo[2]=0;
      sptr->AdjInfo[3]=1;
      /* patch 4 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=6;
      sptr->AdjInfo[1]=5;
      sptr->AdjInfo[2]=0;
      sptr->AdjInfo[3]=7;
      /* patch 5 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=7;
      sptr->AdjInfo[1]=6;
      sptr->AdjInfo[2]=1;
      sptr->AdjInfo[3]=4;
      /* patch 6 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=4;
      sptr->AdjInfo[1]=7;
      sptr->AdjInfo[2]=2;
      sptr->AdjInfo[3]=5;
      /* patch 7 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0]=5;
      sptr->AdjInfo[1]=4;
      sptr->AdjInfo[2]=3;
      sptr->AdjInfo[3]=6;

      break;

    case P_SPHERE:
      /* Initialize adjacency info */
      /* Patch 0 */
      p = np->trim;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 4; sptr->AdjInfo[1] = 1;
      sptr->AdjInfo[2] = 5; sptr->AdjInfo[3] = 3;
      
      /* Patch 1 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 4; sptr->AdjInfo[1] = 2;
      sptr->AdjInfo[2] = 5; sptr->AdjInfo[3] = 0;
      
      /* Patch 2 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 4; sptr->AdjInfo[1] = 3;
      sptr->AdjInfo[2] = 5; sptr->AdjInfo[3] = 1;
      
      /* Patch 3 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 4; sptr->AdjInfo[1] = 0;
      sptr->AdjInfo[2] = 5; sptr->AdjInfo[3] = 2;
      
      /* Patch 4 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 2; sptr->AdjInfo[1] = 1;
      sptr->AdjInfo[2] = 0; sptr->AdjInfo[3] = 3;
      
      /* Patch 5 */
      p++;
      sptr = &(p->TrimPortion[0].SplinePortion);
      sptr->AdjInfo[0] = 0; sptr->AdjInfo[1] = 1;
      sptr->AdjInfo[2] = 2; sptr->AdjInfo[3] = 3;
    case P_TORUS:
      p = np->trim;
      for (i=0; i<np->npatches; i++, p++) {
	sptr = &(p->TrimPortion[0].SplinePortion);
	sptr->NumOfCurves = 4;
	for (j=0; j<sptr->NumOfCurves; j++) {
	  sptr->UV_Splines[j].Deg = 1;
	  sptr->UV_Splines[j].CntrlPts = (POINT3D *) 
	                                 new((sptr->UV_Splines[j].Deg + 1) * 
					     sizeof(POINT3D));
	  sptr->XYZ_Splines[j].Deg = 3;
	  sptr->XYZ_Splines[j].CntrlPts = (POINT4D *) 
	                                 new((sptr->XYZ_Splines[j].Deg + 1) * 
					     sizeof(POINT4D));
	  /* Initialize UV splines */
	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[0] = 0.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 0.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[0].dCoord[1] = 1.0;

	  sptr->UV_Splines[j].CntrlPts[0].dCoord[2] = 1.0;

	  if (j<2)
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[0] = 0.0;

	  if ((j==1) || (j==2)) 
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 1.0;
	  else
	    sptr->UV_Splines[j].CntrlPts[1].dCoord[1] = 0.0;

	  sptr->UV_Splines[j].CntrlPts[1].dCoord[2] = 1.0;

	  /* Initialize XYZ splines */
	  if (j % 2)
	    IsItT = 0;
	  else
	    IsItT = 1;

	  if ((j==1) || (j==2)) 
	    Val = 1.0;
	  else
	    Val = 0.0;
	  if (GetSplineFromPatch(&(p->tp), IsItT, Val,
				 &(sptr->XYZ_Splines[j]))) {
	    printf("Could not get spline properly\n");
	    exit(-1);
	  }
	  if (j > 1)
	    ReverseCntrlPtsOf3DSpline(&(sptr->XYZ_Splines[j]));
	}
	sptr->ptInsideRegion.u = 0.5;
	sptr->ptInsideRegion.v = 0.5;
      }
      if (chewflag == P_TORUS) { /* I dont want to do this for sphere */
	/* initialize adjacency info */
	/* Patch 0 */
	p = np->trim;
	for (i=0; i<np->npatches; i++, p++) {
	  sptr = &(p->TrimPortion[0].SplinePortion);
	  for (j=0; j<sptr->NumOfCurves; j++) {
	    dummy = i/3;
	    switch (j) {
	    case 0: sptr->AdjInfo[j] = (i+6)%9;
	      break;
	    case 1: sptr->AdjInfo[j] = dummy*3 + (i+1)%3;
	      break;
	    case 2: sptr->AdjInfo[j] = (i+3)%9;
	      break;
	    case 3: sptr->AdjInfo[j] = dummy*3 + (i+2)%3;
	      break;
	    }
	  }
	}
      }
      break;
    }
  } else {
    printf("add_trim_patch_info: Entered uncharted waters!!!\n");
    if (chewflag == P_PRISM)		/* lid for prism */
      {
	if (ndomains == 1)
	  {
	    if (top_is_zero)
	      {
		i = np->npatches - 2;
		p = &np->trim[i];
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1 - 1;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      topbot_trimcurve[0][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      topbot_trimcurve[0][i][1];
		  }
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[0].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[0].y;      
	      
		i =  np->npatches - 1;
		p = &np->trim[i];
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1 - 1;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      topbot_trimcurve[1][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      topbot_trimcurve[1][i][1];
		  }
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[1].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v =
		  topbot_inpoint[1].y;   
	      }
	    else			/* top_bot is reversed.... */
	      {
		i = np->npatches - 2;
		p = &np->trim[i];
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1 - 1;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      topbot_trimcurve[1][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      topbot_trimcurve[1][i][1];
		  }
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[1].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[1].y;      
		
		i =  np->npatches - 1;
		p = &np->trim[i];
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1 - 1;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      topbot_trimcurve[0][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      topbot_trimcurve[0][i][1];
		  }
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[0].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[0].y;   
	      }
	    
	  }
	else			/* ndomains > 1 */
	  {
	    
	    if (top_is_zero) {
	      i = np->npatches - 2;
	      p = &np->trim[i];
	      p->NumTrimPortions = 2;
	      p->TrimPortion[1].PiecePortion.poly_vertex = 
		(double (*)[2]) malloc(MAX_VERTEX_COUNT * 2 * sizeof(double *));
	      p->TrimPortion[0].PiecePortion.NumOfVertices = nl1;
	      p->TrimPortion[1].PiecePortion.NumOfVertices = nl2;
	      for (i = 0; i < nl1; i++)
		{
		  p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[0][i][0];
		  p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[0][i][1];
		}
	      p->TrimPortion[0].PiecePortion.poly_vertex[nl1][0] = topbot_trimcurve[0][0][0];
	      p->TrimPortion[0].PiecePortion.poly_vertex[nl1][1] = topbot_trimcurve[0][0][1];
	      
	      p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[0].x;
	      p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[0].y;
	      
	      for (i = 0; i < nl2; i++)
		{
		  p->TrimPortion[1].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[2][i][0];
		  p->TrimPortion[1].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[2][i][1];
		}
	      p->TrimPortion[1].PiecePortion.poly_vertex[nl2][0] = topbot_trimcurve[2][0][0];
	      p->TrimPortion[1].PiecePortion.poly_vertex[nl2][1] = topbot_trimcurve[2][0][1];
	      
	      p->TrimPortion[1].PiecePortion.ptInsideRegion.u = topbot_inpoint[2].x;
	      p->TrimPortion[1].PiecePortion.ptInsideRegion.v = topbot_inpoint[2].y;      
	      
	      i =  np->npatches - 1;
	      p = &np->trim[i];
	      p->NumTrimPortions = 2;
	      p->TrimPortion[1].PiecePortion.poly_vertex = 
		(double (*)[2]) malloc(MAX_VERTEX_COUNT * 2 * sizeof(double *));
	      p->TrimPortion[0].PiecePortion.NumOfVertices = nl1;
	      p->TrimPortion[1].PiecePortion.NumOfVertices = nl2;
	      for (i = 0; i < nl1; i++)
		{
		  p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[1][i][0];
		  p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[1][i][1];
		}
	      p->TrimPortion[0].PiecePortion.poly_vertex[nl1][0] = topbot_trimcurve[1][0][0];
	      p->TrimPortion[0].PiecePortion.poly_vertex[nl1][1] = topbot_trimcurve[1][0][1];
	      
	      p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[1].x;
	      p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[1].y;   
	      
	      for (i = 0; i < nl2; i++)
		{
		  p->TrimPortion[1].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[3][i][0];
		  p->TrimPortion[1].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[3][i][1];
		}
	      p->TrimPortion[1].PiecePortion.poly_vertex[nl2][0] = topbot_trimcurve[3][0][0];
	      p->TrimPortion[1].PiecePortion.poly_vertex[nl2][1] = topbot_trimcurve[3][0][1];
	      
	      p->TrimPortion[1].PiecePortion.ptInsideRegion.u = topbot_inpoint[3].x;
	      p->TrimPortion[1].PiecePortion.ptInsideRegion.v = topbot_inpoint[3].y;   
	    }
	    else 
	      {
		i = np->npatches - 2;
		p = &np->trim[i];
		p->NumTrimPortions = 2;
		p->TrimPortion[1].PiecePortion.poly_vertex = 
		  (double (*)[2]) malloc(MAX_VERTEX_COUNT * 2 * sizeof(double *));
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1;
		p->TrimPortion[1].PiecePortion.NumOfVertices = nl2;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[1][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[1][i][1];
		  }
		p->TrimPortion[0].PiecePortion.poly_vertex[nl1][0] = topbot_trimcurve[1][0][0];
		p->TrimPortion[0].PiecePortion.poly_vertex[nl1][1] = topbot_trimcurve[1][0][1];
		
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[1].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[1].y;
		
		for (i = 0; i < nl2; i++)
		  {
		    p->TrimPortion[1].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[3][i][0];
		    p->TrimPortion[1].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[3][i][1];
		  }
		p->TrimPortion[1].PiecePortion.poly_vertex[nl2][0] = topbot_trimcurve[3][0][0];
		p->TrimPortion[1].PiecePortion.poly_vertex[nl2][1] = topbot_trimcurve[3][0][1];
		
		p->TrimPortion[1].PiecePortion.ptInsideRegion.u = topbot_inpoint[3].x;
		p->TrimPortion[1].PiecePortion.ptInsideRegion.v = topbot_inpoint[3].y;      
		
		i =  np->npatches - 1;
		p = &np->trim[i];
		p->NumTrimPortions = 2;
		p->TrimPortion[1].PiecePortion.poly_vertex = 
		  (double (*)[2]) malloc(MAX_VERTEX_COUNT * 2 * sizeof(double *));
		p->TrimPortion[0].PiecePortion.NumOfVertices = nl1;
		p->TrimPortion[1].PiecePortion.NumOfVertices = nl2;
		for (i = 0; i < nl1; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[0][i][0];
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[0][i][1];
		  }
		p->TrimPortion[0].PiecePortion.poly_vertex[nl1][0] = topbot_trimcurve[0][0][0];
		p->TrimPortion[0].PiecePortion.poly_vertex[nl1][1] = topbot_trimcurve[0][0][1];
		
		p->TrimPortion[0].PiecePortion.ptInsideRegion.u = topbot_inpoint[0].x;
		p->TrimPortion[0].PiecePortion.ptInsideRegion.v = topbot_inpoint[0].y;   
		
		for (i = 0; i < nl2; i++)
		  {
		    p->TrimPortion[1].PiecePortion.poly_vertex[i][0] = topbot_trimcurve[2][i][0];
		    p->TrimPortion[1].PiecePortion.poly_vertex[i][1] = topbot_trimcurve[2][i][1];
		  }
		p->TrimPortion[1].PiecePortion.poly_vertex[nl2][0] = topbot_trimcurve[2][0][0];
		p->TrimPortion[1].PiecePortion.poly_vertex[nl2][1] = topbot_trimcurve[2][0][1];
		
		p->TrimPortion[1].PiecePortion.ptInsideRegion.u = topbot_inpoint[2].x;
		p->TrimPortion[1].PiecePortion.ptInsideRegion.v = topbot_inpoint[2].y;   
	      }
	  }
      }
    
    else if (chewflag == P_VASE && 
	     (!vase_no_lid[0] || !vase_no_lid[1]))	
      {				/* Generate top/bot for VASE */
	/* Generate 36 points for the circle with centre (0.5, 0.5) and */
	/* radius 0.5 spaced 10 deg. apart */
	
	if (!vase_no_lid[0] && !vase_no_lid[1])
	  {
	    for (j = np->npatches - 2; j <= np->npatches - 1; j++)
	      {
		p = &np->trim[j];
		p->TrimPortion[0].PiecePortion.NumOfVertices = NPOINTS;
		for (i = 0; i <= NPOINTS; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      0.5 + 0.5*cos(10.0*i*M_PI/180.0)/VASE_LEEWAY_FACTOR;
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      0.5 + 0.5*sin(10.0*i*M_PI/180.0)/VASE_LEEWAY_FACTOR;
		  }
	      }
	  }
	else 
	  {
	    for (j = np->npatches - 1; j <= np->npatches - 1; j++)
	      {
		p = &np->trim[j];
		p->TrimPortion[0].PiecePortion.NumOfVertices = NPOINTS;
		for (i = 0; i <= NPOINTS; i++)
		  {
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		      0.5 + 0.5*cos(10.0*i*M_PI/180.0)/VASE_LEEWAY_FACTOR;
		    p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		      0.5 + 0.5*sin(10.0*i*M_PI/180.0)/VASE_LEEWAY_FACTOR;
		  }
	      }
	  }
      }
#ifdef NEW_CYL_REPRN		/* do this only for the new cylinder */
    /* representation  */
    else if (chewflag == P_CYLINDER)
      {
	/* Generate 36 points for the circle with centre (0.5, 0.5) and */
	/* radius 0.5 spaced 10 deg. apart */
	
	for (j = 4; j <= 5; j++)
	  {
	    p = &np->trim[j];
	    p->TrimPortion[0].PiecePortion.NumOfVertices = NPOINTS;
	    for (i = 0; i <= NPOINTS; i++)
	      {
		p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		  0.5 + 0.5*cos(10.0*i*M_PI/180.0)/CYL_LEEWAY_FACTOR;
		p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		  0.5 + 0.5*sin(10.0*i*M_PI/180.0)/CYL_LEEWAY_FACTOR;
	      }
	  }
      }
#endif
    
#ifdef NEW_CONE_REPRN		/* do this only for the new cone */
    /* representation  */
    else if (chewflag == P_CONE)
      {
	/* Generate 36 points for the circle with centre (0.5, 0.5) and */
	/* radius 0.5 spaced 10 deg. apart */
	
	for (j = 4; j <= 5; j++)
	  {
	    p = &np->trim[j];
	    p->TrimPortion[0].PiecePortion.NumOfVertices = NPOINTS;
	    for (i = 0; i <= NPOINTS; i++)
	      {
		p->TrimPortion[0].PiecePortion.poly_vertex[i][0] = 
		  0.5 + 0.5*cos(10.0*i*M_PI/180.0)/CONE_LEEWAY_FACTOR;
		p->TrimPortion[0].PiecePortion.poly_vertex[i][1] = 
		  0.5 + 0.5*sin(10.0*i*M_PI/180.0)/CONE_LEEWAY_FACTOR;
	      }
	  }
      }
#endif
  }

  return 0;
}


/* Free the node np */
int free_csgnode(node_t *np)
{
  register int i;
  int j, k;
  LIST_POINT2D *p, *q;

  for (i = 0; i < np->npatches; i++)
    {
      /* free the tensor patch first */

      for (j = 0; j <= np->trim[i].tp.iDeg0; j++) {
	free(np->trim[i].tp.coeffs[j]);
	free(np->trim[i].orig_tp.coeffs[j]);
      }
      free(np->trim[i].tp.coeffs);
      free(np->trim[i].orig_tp.coeffs);
      
      /* free poly_vertex if piecewise and splines otherwise */
      if (np->trim[i].SplineReprn) 
	for (j=0; j<np->trim[i].NumTrimPortions; j++)
	  free_trim_portion_spline_type(
                       &np->trim[i].TrimPortion[j].SplinePortion);
      else
	for (j=0; j<np->trim[i].NumTrimPortions; j++)
	  free_trim_portion_piece_type(
                       &np->trim[i].TrimPortion[j].PiecePortion);
    }
      
#ifdef __FREE_ALL
  free(np->trim);
#endif

  return 0;
}

	
/* Allocate space for the array to hold all the control points for the */
/* patch */

int alloc_tensor_patch(TENSOR_PATCH *tp,int udeg, int vdeg)
{
  int j;
  
  tp->iDeg0 = udeg;
  tp->iDeg1 = vdeg;

#if 0
  tp->coeffs = (POINT4D **) new((udeg + 1)*sizeof(POINT4D *));
  for (j = 0; j < udeg + 1; j++)
    tp->coeffs[j] = (POINT4D *) new((vdeg + 1)*sizeof(POINT4D));
#endif
  allocate_tensor_patch(tp);
}

/* Compute the bounding-box for the solid represented by the node in */
/* the CSG tree */
int compute_bbox_for_node(node_t *np)
{
  int m, n;
  int i, j, k;
  double minx, maxx, miny, maxy, minz, maxz;
  double wval;
  TENSOR_PATCH *tp;

  if (np->npatches == 0)
    return 0;

  tp = &np->trim[0].tp;
  minx = maxx = tp->coeffs[0][0].dCoord[0] / tp->coeffs[0][0].dCoord[3];
  miny = maxy = tp->coeffs[0][0].dCoord[1] / tp->coeffs[0][0].dCoord[3]; 
  minz = maxz = tp->coeffs[0][0].dCoord[2] / tp->coeffs[0][0].dCoord[3];
  
  for (k = 0; k < np->npatches; k++)
    {
      tp = &np->trim[k].tp;
      m = tp->iDeg0;
      n = tp->iDeg1;
/*      
      minx = maxx = tp->coeffs[0][0].dCoord[0] / tp->coeffs[0][0].dCoord[3];
      miny = maxy = tp->coeffs[0][0].dCoord[1] / tp->coeffs[0][0].dCoord[3]; 
      minz = maxz = tp->coeffs[0][0].dCoord[2] / tp->coeffs[0][0].dCoord[3];
*/      
      for (i=0; i<=m; i++)
	for (j=0; j<=n; j++)
	  {
	    wval = 1.0/tp->coeffs[i][j].dCoord[3]; /* computing */
						   /* reciprocal */
	    if ((tp->coeffs[i][j].dCoord[0]/wval) < minx) 
	      minx = tp->coeffs[i][j].dCoord[0] * wval;
	    if ((tp->coeffs[i][j].dCoord[0]/wval) > maxx) 
	      maxx = tp->coeffs[i][j].dCoord[0] * wval;
	    if ((tp->coeffs[i][j].dCoord[1]/wval) < miny) 
	      miny = tp->coeffs[i][j].dCoord[1] * wval;
	    if ((tp->coeffs[i][j].dCoord[1]/wval) > maxy) 
	      maxy = tp->coeffs[i][j].dCoord[1] * wval;
	    if ((tp->coeffs[i][j].dCoord[2]/wval) < minz) 
	      minz = tp->coeffs[i][j].dCoord[2] * wval;
	    if ((tp->coeffs[i][j].dCoord[2]/wval) > maxz) 
	      maxz = tp->coeffs[i][j].dCoord[2] * wval;
	  }
    }
  
  (np->bbox.i3Dbox[0]).low = minx;
  (np->bbox.i3Dbox[0]).high = maxx;
  (np->bbox.i3Dbox[1]).low = miny;
  (np->bbox.i3Dbox[1]).high = maxy;
  (np->bbox.i3Dbox[2]).low = minz;
  (np->bbox.i3Dbox[2]).high = maxz;  
}


/* Compute the bounding box for the parent given the two bounding */
/* boxes for the two child-nodes */
int compute_bbox_from_leaves(node_t *np, node_t *nl, node_t *nr)
{
  if ((np->op == O_SPLIT1) || (np->op == O_SPLIT2)) /* special case */
    {
      if (np->left->prim->ptype == P_PLANE) /* left node is a plane */
	np->bbox = np->right->bbox;
      else			/* right node is plane */
	np->bbox = np->left->bbox;	
    }
  else				/* normal op: get the max bounding-box */
    {
      np->bbox.i3Dbox[0].high = MAX(np->left->bbox.i3Dbox[0].high,
				    np->right->bbox.i3Dbox[0].high);
      np->bbox.i3Dbox[0].low = MAX(np->left->bbox.i3Dbox[0].low,
				    np->right->bbox.i3Dbox[0].low);
      np->bbox.i3Dbox[1].high = MAX(np->left->bbox.i3Dbox[1].high,
				    np->right->bbox.i3Dbox[1].high);
      np->bbox.i3Dbox[1].low = MAX(np->left->bbox.i3Dbox[1].low,
				    np->right->bbox.i3Dbox[1].low);
      np->bbox.i3Dbox[2].high = MAX(np->left->bbox.i3Dbox[2].high,
				    np->right->bbox.i3Dbox[2].high);
      np->bbox.i3Dbox[2].low = MAX(np->left->bbox.i3Dbox[2].low,
				    np->right->bbox.i3Dbox[2].low);
    }
  return 0;
}


/* Perform a verry simple check to see whether the two solids are */
/* identical by comparing their bounding boxes */
int are_solids_identical(node_t *np1, node_t *np2)
{
 return (FP_EQUAL(np1->bbox.i3Dbox[0].high, np2->bbox.i3Dbox[0].high) && 
	 FP_EQUAL(np1->bbox.i3Dbox[0].low, np2->bbox.i3Dbox[0].low) && 
	 FP_EQUAL(np1->bbox.i3Dbox[1].high, np2->bbox.i3Dbox[1].high) && 
	 FP_EQUAL(np1->bbox.i3Dbox[1].low, np2->bbox.i3Dbox[1].low) && 
	 FP_EQUAL(np1->bbox.i3Dbox[2].high, np2->bbox.i3Dbox[2].high) && 
	 FP_EQUAL(np1->bbox.i3Dbox[2].low, np2->bbox.i3Dbox[2].low));
}


int dump_patches(FILE *outfile, node_t *np)
{
  TENSOR_PATCH *patch = np->patch;
  register int i, j, k;
  int l;

  printf("\n\nDumping patches\n");
  for (k = 0; k < np->npatches; k++)
    {
      fprintf(outfile, "\n%d %d\n", patch[k].iDeg0, patch[k].iDeg1);
      
      for (i = 0; i <= patch[k].iDeg0; i++)
	for (j = 0; j <= patch[k].iDeg1; j++)
	  {
	    for (l = 0; l < 4; l++)
	      fprintf(outfile, "%lf ",
		      patch[k].coeffs[i][j].dCoord[l]);
	    fprintf(outfile, "\n");
	  }
    }

  fflush(outfile);
  return 0;
}

int dump_one_patch(outfile, trim, patchnum)
FILE *outfile;
TRIM_PATCH_INFO *trim;
int patchnum;
{
  register int i, j, k;
  int l;

  k = patchnum;
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);

      for (i = 0; i <= trim[k].tp.iDeg0; i++)
        for (j = 0; j <= trim[k].tp.iDeg1; j++)
          {
            for (l = 0; l < 4; l++)
              fprintf(outfile, "%lf ",
                      trim[k].tp.coeffs[i][j].dCoord[l]);
            fprintf(outfile, "\n");
          }

  fflush(outfile);
  return 0;
}

int dump_trim_patches(FILE *outfile, TRIM_PATCH_INFO *trim, 
		      int npatches) 
{
  register int i, j, k;
  int l;
  
  fprintf(outfile, "\n\n Printing trim_patch_info dump\n");
  for (k = 0; k < npatches; k++)
    {
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);
      
      for (i = 0; i <= trim[k].tp.iDeg0; i++)
	for (j = 0; j <= trim[k].tp.iDeg1; j++)
	  {
	    for (l = 0; l < 4; l++)
	      fprintf(outfile, "%lf ",
		      trim[k].tp.coeffs[i][j].dCoord[l]);
	    fprintf(outfile, "\n");
	  }
    }

  fflush(outfile);
  return 0;
}

extern double Scale;

double calc_scale(TENSOR_PATCH *tp, int npatches)
{
  register int i, j, k, l;
  
  for (k = 0; k < npatches; k++)
    for (i = 0; i <= tp[k].iDeg0; i++)
      for (j = 0; j <= tp[k].iDeg1; j++)
	for (l = 0; l < 3; l++)
	  if (Scale < fabs(tp[k].coeffs[i][j].dCoord[l]))
	    Scale = fabs(tp[k].coeffs[i][j].dCoord[l]);
  
  return Scale;
}

#endif 


/* Given a list of patches for a node and a tranformation matrix, */
/* transform all the patches. Do this before add_trim_patch_info  */
int transform_patches(node_t *np, matrix_t xmat)
{
  register int i, j, k;
  vector_t v0, v;
  TENSOR_PATCH *tp;

  for (k = 0; k < np->npatches; k++)
    {
      tp = &np->patch[k];
      for (i = 0; i <= tp->iDeg0; i++)
	for (j = 0; j <= tp->iDeg1; j++)
	  {
	    yourpt_to_mypt(tp->coeffs[i][j], v);
	    matvec_mult(xmat, v, v0);
	    mypt_to_yourpt_no_w(v0, tp->coeffs[i][j]);
	    tp->coeffs[i][j].dCoord[3] = v0[3];
	  }
    }
  
  return 0;
}

int transform_patches2(TENSOR_PATCH *tpt, int npatches, matrix_t xmat)
{
  register int i, j, k;
  vector_t v0, v;
  TENSOR_PATCH *tp;

  for (k = 0; k < npatches; k++)
    {
      tp = &tpt[k];
      for (i = 0; i <= tp->iDeg0; i++)
	for (j = 0; j <= tp->iDeg1; j++)
	  {
	    yourpt_to_mypt(tp->coeffs[i][j], v);
	    matvec_mult(xmat, v, v0);
	    mypt_to_yourpt_no_w(v0, tp->coeffs[i][j]);
	    tp->coeffs[i][j].dCoord[3] = v0[3];
	  }
    }
  
  return 0;
}

/* Precprocess the solid. Look at all the trimportions and remove all */
/* the points which are too close together. */

int preprocess_solid(node_t *np)
{
  int i, j, k, l;
  TRIM_PORTION_PIECEWISE_TYPE *tpor;
  point_typ pt;

  for (i = 0; i < np->npatches; i++)
    if (!np->trim[i].SplineReprn)
      for (j = 0; j < np->trim[i].NumTrimPortions; j++)
	{
	  tpor = &(np->trim[i].TrimPortion[j].PiecePortion);
	  
	  /* Do not change the 0th and last point */
	  k = 1;
	  while (k < tpor->NumOfVertices)
	    {
	      if (CLOSE(tpor->poly_vertex[k][0],
			tpor->poly_vertex[k-1][0]) &&
		  CLOSE(tpor->poly_vertex[k][1],
			tpor->poly_vertex[k-1][1]))
		{
		  /* delete pt[k] */
		  for (l = k; l < tpor->NumOfVertices; l++)
		    {
		      tpor->poly_vertex[l][0] =
			tpor->poly_vertex[l+1][0];
		      tpor->poly_vertex[l][1] =
			tpor->poly_vertex[l+1][1];
		    }
		  tpor->NumOfVertices--;
		  k--;
		}
	      
	      k++;
	    }
	  
	  /* Find point inside polygon */
	  
	  find_point_inside_polygon(tpor->NumOfVertices,
				    tpor->poly_vertex, &pt);
	  
	  tpor->ptInsideRegion.u = pt.x;
	  tpor->ptInsideRegion.v = pt.y;
	}
  
  return 0;
}

  
int validate_solid(node_t *np)
{
  int i, j, k, l, m, n;
  TRIM_PORTION_PIECEWISE_TYPE *tpor;
  double x, y;
  POINT4D pt;

  for (m = 0; m < np->npatches; m++)
    {
      /* Check if the patch is proper (facing out) or not */

      if (np->trim[m].SplineReprn) continue;

      if (!IsPatchFacingOutside(np->trim, np->npatches, m))
	{
	  /* First flip all the control points */

	  for (i = 0; i <= np->trim[m].tp.iDeg0; i++)
	    {
	      for (k = 0, l = np->trim[m].tp.iDeg1; 
		   k < l; k++, l--)
		{
		  pt = np->trim[m].tp.coeffs[i][k];
		  np->trim[m].tp.coeffs[i][k] =
		    np->trim[m].tp.coeffs[i][l];
		  np->trim[m].tp.coeffs[i][l] = pt;
		}
	    }

	  for (i = 0; i <= np->trim[m].orig_tp.iDeg0; i++)
	    {
	      for (k = 0, l = np->trim[m].orig_tp.iDeg1; 
		   k < l; k++, l--)
		{
		  pt = np->trim[m].orig_tp.coeffs[i][k];
		  np->trim[m].orig_tp.coeffs[i][k] =
		    np->trim[m].orig_tp.coeffs[i][l];
		  np->trim[m].orig_tp.coeffs[i][l] = pt;
		}
	    }
	  

	  /* Flip all the trimPortions in the patch */
	  
	  for (j = 0; j < np->trim[m].NumTrimPortions; j++)
	    {
	      tpor = &(np->trim[m].TrimPortion[j].PiecePortion);
	      
	      for (k = 0; k <= tpor->NumOfVertices; k++)
		tpor->poly_vertex[k][1] = 1.0 - tpor->poly_vertex[k][1];
	      
	      for (k = 0, l = tpor->NumOfVertices; k < l; k++, l--)
		{
		  x = tpor->poly_vertex[k][0];
		  y = tpor->poly_vertex[k][1];
	
		  tpor->poly_vertex[k][0] = tpor->poly_vertex[l][0];
		  tpor->poly_vertex[k][1] = tpor->poly_vertex[l][1];

		  tpor->poly_vertex[l][0] = x;
		  tpor->poly_vertex[l][1] = y;
		}
	      
	      /* Flip point inside region */

	      tpor->ptInsideRegion.v = 1.0 - tpor->ptInsideRegion.v;

	    }
	  
	} /* end-flip patch */
    }
  
  return 0;
}
		  
