#include <defs.h>
#include "tree.h"
#include <malloc.h>

/* This stuff is in parse/kother.c */

int add_trim_patch_info(node_t *np,  int chewflag, int ndomains,
			int nl1, int nl2, int pnum);

    
/* This stuff was in parse/trees.c */
int is_planar_overlap = FALSE;

/* This stuff is in primitives/___.c */

int gen_cone(POINT4D, POINT4D, double, double, double,
	     TENSOR_PATCH **);
int gen_cylinder(POINT4D, POINT4D, double, double, TENSOR_PATCH **);
int gen_brlell(POINT4D, POINT4D,  POINT4D,  POINT4D, double,double,double,TENSOR_PATCH**);
int gen_sphere(POINT4D, double, TENSOR_PATCH **);
int gen_torus(POINT4D, POINT4D, double, double, TENSOR_PATCH **);


/* This is in surface/perf_csg.c */

int perform_CSG(TRIM_PATCH_INFO**, int,
		TRIM_PATCH_INFO**, int,
		int, int, int /* operation */, int,
		POINT4D*, double /* step size */,
		TRIM_PATCH_INFO** /* parent */);


void operate(TRIM_PATCH_INFO** red, int nred,
	     TRIM_PATCH_INFO** blue, int nblue,
	     int operation,
	     TRIM_PATCH_INFO** result, int* nresult,
	     double stepsize)
{
  *nresult = perform_CSG(red, nred, blue, nblue, 1, 1,
			 operation, 0, NULL, stepsize, result);
}

	     
/* Primitive generation */

void new_cone(POINT4D center, POINT4D axis_vec,
	     double cone_length, double cone_Lradius, double cone_Uradius,
	     TRIM_PATCH_INFO** patches, int* npatches)
{
  /* Allocate space for TRIM_PATCH_INFO* array */
  node_t* np = (node_t*)malloc(sizeof(node_t));
  np->left = np->right = NULL;
  np->nodetype = T_LEAF;
  np->prim = NULL;
  /* np->prim = (primitive_t *) new(sizeof(primitive_t)); */
  np->npatches = *npatches = gen_cone(center, axis_vec,
				      cone_length, cone_Lradius, cone_Uradius,
				      &(np->patch));
  add_trim_patch_info(np, P_CONE /* see primitives.h */,
		      0, 0, 0, 0 /*pnum*/);
  *patches = np->trim;
  
}

void new_cylinder(POINT4D center, POINT4D axis_vec,
	     double cyl_length, double cyl_radius,
	     TRIM_PATCH_INFO** patches, int* npatches)
{
  /* Allocate space for TRIM_PATCH_INFO* array */
  node_t* np = (node_t*)malloc(sizeof(node_t));
  np->left = np->right = NULL;
  np->nodetype = T_LEAF;
  np->prim = NULL;
  /* np->prim = (primitive_t *) new(sizeof(primitive_t)); */
  np->npatches = *npatches = gen_cylinder(center, axis_vec,
				      cyl_length, cyl_radius, 
				      &(np->patch));
  add_trim_patch_info(np, P_CYLINDER /* see primitives.h */,
		      0, 0, 0, 0 /*pnum*/);
  *patches = np->trim;
  
}

void new_ellipsoid(POINT4D center, POINT4D a_vec, POINT4D b_vec, POINT4D c_vec,
	     double a, double b, double c,
	     TRIM_PATCH_INFO** patches, int* npatches)
{
  /* Allocate space for TRIM_PATCH_INFO* array */
  node_t* np = (node_t*)malloc(sizeof(node_t));
  np->left = np->right = NULL;
  np->nodetype = T_LEAF;
  np->prim = NULL;
  /* np->prim = (primitive_t *) new(sizeof(primitive_t)); */
  np->npatches = *npatches = gen_brlell(center,a_vec,b_vec,c_vec,
					a,b,c,
					&(np->patch));
  add_trim_patch_info(np, P_BRLELL /* see primitives.h */,
		      0, 0, 0, 0 /*pnum*/);
  *patches = np->trim;
  
}

void new_sphere(POINT4D center, double r, 
	     TRIM_PATCH_INFO** patches, int* npatches)
{
  /* Allocate space for TRIM_PATCH_INFO* array */
  node_t* np = (node_t*)malloc(sizeof(node_t));
  np->left = np->right = NULL;
  np->nodetype = T_LEAF;
  np->prim = NULL;
  /* np->prim = (primitive_t *) new(sizeof(primitive_t)); */
  np->npatches = *npatches = gen_sphere(center, r,
					&(np->patch));
  add_trim_patch_info(np, P_SPHERE /* see primitives.h */,
		      0, 0, 0, 0 /*pnum*/);
  *patches = np->trim;
}

  
void new_torus(POINT4D center, POINT4D axis_vec, 
	     double iradius, double oradius,
	     TRIM_PATCH_INFO** patches, int* npatches)
{
  /* Allocate space for TRIM_PATCH_INFO* array */
  node_t* np = (node_t*)malloc(sizeof(node_t));
  np->left = np->right = NULL;
  np->nodetype = T_LEAF;
  np->prim = NULL;
  /* np->prim = (primitive_t *) new(sizeof(primitive_t)); */
  np->npatches = *npatches = gen_torus(center, axis_vec,
					iradius, oradius,
					&(np->patch));
  add_trim_patch_info(np, P_TORUS /* see primitives.h */,
		      0, 0, 0, 0 /*pnum*/);
  *patches = np->trim;
  
}
