#ifndef LIBINTERF_H
#define LIBINTERF_H

/* Data structures.  These are replicated from surface/defs.h.
 * If you are wondering why they are replicated, go look at defs.h
 * and tell me if you'd rather include that. */

#define MAX_NUM_PORTIONS  15

typedef struct _POINT2D {
    double  u,v;
} POINT2D;

typedef struct { 
    double 	dCoord[3];
} POINT3D;

typedef struct { 
    double 	dCoord[4];
} POINT4D;

typedef struct {
    double low;
    double high;
} INTERVAL;

typedef struct {
    INTERVAL i3Dbox[3];
} BOUNDINGBOX;

typedef struct {
    INTERVAL i2Dbox[2];
} RECT_BOX;

typedef struct {
    int Npts;
    POINT4D *ptList;
    RECT_BOX domain_def;
} TRACE_INFO;

typedef struct {
    int patchNum0;
    int patchNum1;
} OVERLAPBOXES;

typedef struct {
    OVERLAPBOXES ptches;
    int          same_side;
} OVR_BOXES;

typedef struct {
    double 	Re;
    double 	Im;
} COMPLEX;

typedef struct {
    COMPLEX	cpxCoord[4];
} POINT4CPX;

typedef struct _LIST_POINT4D {
    POINT4D               p4D;
    struct _LIST_POINT4D  *plp4DNext;
} LIST_POINT4D;

typedef struct _LIST_POINT2D {
    POINT2D               p2D;
    struct _LIST_POINT2D  *plp2DNext;
} LIST_POINT2D;

typedef struct {
    LIST_POINT4D	*Curve;
    POINT4D		endPoints[2];
} INT_CURVE_STRUCT;

typedef struct {
    LIST_POINT2D	*planeCurve;
    POINT2D		StEndPoints[2];
} DOMAIN_INT_CURVE;

typedef struct _LIST_POINT4CPX {
    POINT4CPX               p4CPX;
    struct _LIST_POINT4CPX  *plp4CPXNext;
} LIST_POINT4CPX;

typedef struct 
{ int		iDeg0, iDeg1;	/* degrees of the tensor product patch */
  POINT4D	**coeffs;	/* coefficients of the tensor product patch */
} TENSOR_PATCH;

typedef struct {
  int Deg;
  POINT3D *CntrlPts;
} UV_SPLINE_TYPE;

typedef struct {
  int Deg;
  POINT4D *CntrlPts;
} XYZ_SPLINE_TYPE;

typedef struct _INTERSECTION_CURVE {
    UV_SPLINE_TYPE		uv_spline;
    XYZ_SPLINE_TYPE		xyz_spline;
    int				neighbor;
    struct _INTERSECTION_CURVE	*other_solid_ptr;
    struct _INTERSECTION_CURVE	*next; 
} INTERSECTION_CURVE;

typedef struct {
    short   which_solid; /* is 1 or 2 depending on first or second */
                         /* solid */
    short   same_solid;
    short   patch_num;
    double  *param;
    short   param_length; /* gives length of the param array */
    POINT4D *st_pt, *end_pt;
    short   do_twice;
    short   other_patch_num, other_region_num;
} SPLIT_TPST_STRUCT;

typedef struct {
    INTERSECTION_CURVE	*int_curve;
} MERGED_CURVE;

typedef struct {
    int entering_from, going_into;
} CURVE_ADJCY;

typedef struct 
{ int		iDeg0, iDeg1;	/* degrees of the tensor product patch */
  double	**coeffs;	/* coefficients of the tensor product patch */
} TENSOR_PATCH_1D;

typedef struct {
  int NumOfCurves; /* Number of curves defining the trim curve */
  UV_SPLINE_TYPE *UV_Splines; /* spline coefficients in domain of patch */
  XYZ_SPLINE_TYPE *XYZ_Splines; /* spline coefficients in space */
  int *AdjInfo; /* Adjacency info for each spline */
  POINT2D ptInsideRegion; /* one point inside the trim curve */
} TRIM_PORTION_SPLINE_TYPE;

typedef struct {
  int NumOfVertices;
  double (*poly_vertex)[2];
  int *AdjInfo;   
  POINT2D ptInsideRegion;
} TRIM_PORTION_PIECEWISE_TYPE;

typedef union {
    TRIM_PORTION_SPLINE_TYPE	SplinePortion;
    TRIM_PORTION_PIECEWISE_TYPE PiecePortion;
} TRIM_PORTION_TYPE;

typedef struct {
    TENSOR_PATCH	tp;	/* tensor patch */
    TENSOR_PATCH	orig_tp; /* original, undivided tpatch */
    int			primNum;
    int			SplineReprn;
    int			NumTrimPortions;
    TRIM_PORTION_TYPE	TrimPortion[MAX_NUM_PORTIONS];
    int			csg_status;
    double		umin, umax, vmin, vmax;
} TRIM_PATCH_INFO;


/* Perform a boolean operation. */

void operate(TRIM_PATCH_INFO** red, int nred,
	     TRIM_PATCH_INFO** blue, int nblue,
	     int operation,
	     TRIM_PATCH_INFO** result, int* nresult,
	     double stepsize);

/* Generate primitive solids. */

int new_cone(POINT4D center, POINT4D axis_vec,
	     double cone_length, double cone_Lradius, double cone_Uradius,
	     TRIM_PATCH_INFO** patches, int* npatches);
void new_cylinder(POINT4D center, POINT4D axis_vec,
		  double cyl_length, double cyl_radius,
		  TRIM_PATCH_INFO **patches, int *npatches);
void new_ellipsoid(POINT4D center,
		   POINT4D a_vec, POINT4D b_vec, POINT4D c_vec,
		   double a, double b, double c,
		   TRIM_PATCH_INFO **patches, int *npatches);
void new_sphere(POINT4D center, double r,
		TRIM_PATCH_INFO **patches, int *npatches);
void new_torus(POINT4D center, POINT4D axis_vec,
	       double iradius, double oradius,
	       TRIM_PATCH_INFO **patches, int *npatches);

#endif
