#ifndef _primitives_h
#define _primitives_h

#include <basic.h>

typedef struct cone_s {
  point_typ c0, c1;		/* upper/lower centres */
  double r0, r1;		/* upper/lower radii */
} cone_t;

typedef struct cylinder_s {
  point_typ c0, c1;
  double r0, r1;
} cylinder_t;

typedef struct ppped {
  point_typ c0;
  vector_t v0, v1, v2;
  double len0, len1, len2;
} ppped_t;

typedef struct pipe_s {
  double r, ncrv;
  curve_t *crv;
} pipe_t;


typedef struct prism_s {
  vector_t v;
  plane_typ pl1, pl2;
  int ncrv, ndomains;
  domain_t *dom;
} prism_t;

typedef struct pyramid_s {
  vector_t v;
  int ncrv;
  curve_t *crv;
} pyramid_t;

typedef struct sphere_s {
  point_typ c;			/* centre */
  double rin, rout;		/* inner and outer radii */
} sphere_t;

typedef struct torus_s {
  point_typ c;
  vector_t v;
  double rg, rs;
} torus_t;

typedef struct vase_s {
  int ncrv;
  vector_t pt1;
  vector_t v;
  double a0, a1;
  curve_t *crv;
} vase_t;

typedef struct primplane_s {
  plane_typ pl;
  vector_t v0, v1;
} primplane_t;

typedef struct primitive_s {
  int ptype;
  double r, g, b;
  union {
    cone_t *con_p;
    cylinder_t *cyl_p;
    ppped_t *ppd_p;
    pipe_t *pip_p;
    primplane_t *prpl_p;
    prism_t *pri_p;
    pyramid_t *pyr_p;
    sphere_t *sph_p;
    torus_t *tor_p;
   } u;
} primitive_t;


/* primitive types */

#define P_CONE       18
#define P_CYLINDER   17
#define P_ISOLATED   0
#define P_PATCH      41
#define P_PPPED      9
#define P_PIPE       32
#define P_PLANE      1
#define P_PRISM      24
#define P_PYRAMID    25
#define P_SPHERE     16
#define P_TORUS      19
#define P_VASE       40
#define P_BRLELL     21
#define P_BRLTGC     22
#define P_BRLARB8    23

/* Maximum degree of any curve used in the model */

#define MAX_CURVEDEG  15

/* Inner or outer domain (used for generating normals) */
#define DOM_INNER 1
#define DOM_OUTER 2

/* The primitive SPACE elements */
#define SP_POINT   1
#define SP_LINE    2
#define SP_CURVE   3
#define SP_PLANE   4
#define SP_CIRCLE  20
#define SP_ELLIPSE 21


#endif /* _primitives_h */
