#ifndef _tree_h
#define _tree_h

#include <global.h>

/* Include this only if linked with Shankar's code */
#include <defs.h>


/* Operation performed at the node */
#define O_UNION 1
#define O_INTER 2
#define O_SUB   3
#define O_SPLIT1 4
#define O_SPLIT2 5
#define O_ASSEM 6

/* Node type */
#define T_LEAF    1
#define T_NONLEAF 2

/* The size of the tensor-patch table used to store the generated */
/* patches for the solid before allocating space to hold all the */
/* patches */
#define MAX_TPNUM 110
#define MAX_TRIM_CURVE_POINTS 400 /* maximum #points in a trim_curve */

#define is_leaf(p) ((p)->nodetype == T_LEAF)

/* The structure of each node in the CSG tree. Initially, only the */
/* leaf nodes contain the tensor-patch information. This is propagated */
/* upwards as the operations are performed */

typedef struct node_s {
  int nodetype;			/* whether leaf/non-leaf */
  int op;			/* Boolean operation */
  solinfo_t *info;

  int npatches;

  TENSOR_PATCH *patch;		/* patches which make up the solid */
				/* surface.  */
  TRIM_PATCH_INFO *trim;        /* Array of trim_patch information */
				/* one for each of the patches */
  BOUNDINGBOX bbox;		/* bounding-box */

  primitive_t *prim;		/* primitive type */

  struct node_s *left, *right;	/* pointer to children */
				/* (in case non-leaf) */
} node_t;


#endif /* _tree_h */
