#include <tree.h>
#include <macro.h>

#ifdef __LINK
extern double Sphradius;
int is_planar_overlap = FALSE;
#define MAX_CRITICAL_POINTS_IN_CHAIN	100
#define REQ_CRITICAL_ANGLE		0.173648

#ifdef __GL
extern int ip;
#endif

#endif

#ifdef __LINK

/* Routine to handle the split-by-plane case separately. This is */
/* required so that a different routine is called rather than */
/* perform_CSG */
#ifdef PRIM_TEST /* I changed this from ifndef to ifdef because I want */
		 /* to treat a PLANE_CUT as a DIFFERENCE. So I am */
		 /* putting a box with the plane and calling */
		 /* perform_CSG. Ideally I compile with -UPRIM_TEST, */
		 /* so this change */
static int handle_plane_case(node_t *np, double ssize)
{
  int whichop;
  POINT4D cutting_plane;
  primplane_t *pp;
  POINT4D v0, v1;
  
  if (np->op == O_SPLIT1)
    whichop = 1;
  else
    whichop = -1;

  if (np->left->prim->ptype == P_PLANE) /* left node is plane */
    {
      pp = np->left->prim->u.prpl_p;
      cutting_plane.dCoord[0] = pp->pl.coeff[0];
      cutting_plane.dCoord[1] = pp->pl.coeff[1];
      cutting_plane.dCoord[2] = pp->pl.coeff[2];
      cutting_plane.dCoord[3] = pp->pl.coeff[3];
      mypt_to_yourpt(pp->v0, v0);
      mypt_to_yourpt(pp->v1, v1);
      np->npatches = plane_cut_op(&(np->right->trim),
				  np->right->npatches, &cutting_plane,
				  &v0, &v1, whichop, 
				  ssize, &np->trim);
    }
  else				/* right node is plane */
    {
      pp = np->right->prim->u.prpl_p;
      cutting_plane.dCoord[0] = pp->pl.coeff[0];
      cutting_plane.dCoord[1] = pp->pl.coeff[1];
      cutting_plane.dCoord[2] = pp->pl.coeff[2];
      cutting_plane.dCoord[3] = pp->pl.coeff[3];
      mypt_to_yourpt(pp->v0, v0);
      mypt_to_yourpt(pp->v1, v1);
      np->npatches = plane_cut_op(&(np->left->trim),
				  np->left->npatches, &cutting_plane,
				  &v0, &v1, whichop, 
				  ssize, &np->trim);
    }      
  
  return 0;
}
#else
static int handle_plane_case(node_t *np, double ssize)
{
  int whichop, i;
  POINT4D cutting_plane;
  primplane_t *pp;
  POINT4D v0, v1;
  
  if (np->op == O_SPLIT1)
    whichop = 1;
  else
    whichop = -1;

  if (np->left->prim->ptype == P_PLANE) /* left node is plane */
    {
      pp = np->left->prim->u.prpl_p;

      /* I am multiplying v0 and pl with whichop so that while reversing */
      /* normals, I also preserve the right-handed coord. system */
      for (i=0; i<4; i++) 
	pp->v0[i] = whichop*pp->v0[i];
      
      cutting_plane.dCoord[0] = whichop*pp->pl.coeff[0];
      cutting_plane.dCoord[1] = whichop*pp->pl.coeff[1];
      cutting_plane.dCoord[2] = whichop*pp->pl.coeff[2];
      cutting_plane.dCoord[3] = whichop*pp->pl.coeff[3];
      mypt_to_yourpt(pp->v0, v0);
      mypt_to_yourpt(pp->v1, v1);
      np->npatches = plane_cut_op(&(np->right->trim),
				  np->right->npatches, &cutting_plane,
				  &v0, &v1, whichop, 
				  ssize, &np->trim);
    }
  else				/* right node is plane */
    {
      pp = np->right->prim->u.prpl_p;

      /* I am multiplying v0 and pl with whichop so that while reversing */
      /* normals, I also preserve the right-handed coord. system */
      for (i=0; i<4; i++) 
	pp->v0[i] = whichop*pp->v0[i];

      cutting_plane.dCoord[0] = whichop*pp->pl.coeff[0];
      cutting_plane.dCoord[1] = whichop*pp->pl.coeff[1];
      cutting_plane.dCoord[2] = whichop*pp->pl.coeff[2];
      cutting_plane.dCoord[3] = whichop*pp->pl.coeff[3];
      mypt_to_yourpt(pp->v0, v0);
      mypt_to_yourpt(pp->v1, v1);
      np->npatches = plane_cut_op(&(np->left->trim),
				  np->left->npatches, &cutting_plane,
				  &v0, &v1, whichop, 
				  ssize, &np->trim);
    }      
  
  return 0;
}
#endif
#endif 


/* Perform a depth first traversal of the tree. At each intermediate */
/* node, obtain the surface representation of the two children first, */
/* and perform the Boolean operation on them */

static int traverse_tree(node_t *np, double ssize, int is_root, char *fl_name)
{
  int frep, srep;

  if (np == NULL)
    return -1;

  if (np->nodetype == T_LEAF)
    return 0;
  
  /* Non-leaf node. traverse the two leaves and perform the boolean */
  /* operation on them */
  
  traverse_tree(np->left, ssize, FALSE, fl_name);
  traverse_tree(np->right, ssize, FALSE, fl_name);

#ifdef __LINK
#ifdef __GL
  ip = 0;
#endif

  if ((np->op == O_SPLIT1) || (np->op == O_SPLIT2)) {
#ifndef PRIM_TEST
    handle_plane_case(np, ssize);
#endif
  } else {
    preprocess_solid(np->left);
    preprocess_solid(np->right);
    frep = srep = 1;
    /* changing call to perform_CSG to accomodate splines */
    np->npatches = perform_CSG(&(np->left->trim), np->left->npatches,
			       &(np->right->trim), np->right->npatches,
			       frep, srep, np->op, 0, NULL, ssize,
			       &(np->trim));
  }

#ifndef _LOTS_OF_HOLES /* This ifndef statement is used to disable */
  /* the following lines of code only when a  */
  /* number of holes are dug into a single    */
  /* slab. Otherwise these lines are activated. */
  
  compute_bbox_from_leaves(np, np->left, np->right);
  free_csgnode(np->left);
  free_csgnode(np->right);
  
#ifdef _SUBDIVIDE_PATCHES
  subdivide_trimmed_patches(np);
#ifdef DONT_DUMP_PATCHES
  dump_subdivided_patches(np);
#endif
#endif

#endif
  
  /* Before returning, check if this was the root-level invocation, If */
  /* so, write out the output to the file */
  
  if (is_root)
    write_solid_to_file(np->npatches, np->trim, fl_name);
  
#endif /* LOTS_OF_HOLES */
  
  return 0;
}


/* Construct CSG trees for each solid in the file... and store them in */
/* the table csgtree[] */
/* If solnum is non-zero, read only the that solid and produce the */
/* output file. Skip all the solids before that */

int process_file(FILE *infile, int solnum, char *fl_name)
{
  int nsol, nisosol, nstruc, dummy;
  int i, j;
  node_t **csgtree, *np;
  double ssize;

  read_header(infile, &nsol, &nisosol, &nstruc);
  
#ifdef DEBUG
  printf("nsol = %d, niso = %d, nstruc = %d\n", nsol, nisosol, nstruc);
#endif

  /* Allocate space to hold the CSG trees for all the solids and */
  /* construct the tree for each solid in the file */

  if (!solnum)
    {				/* read all the solids */

      csgtree = (node_t **) new(nsol*sizeof(node_t *));
      for (i = 0; i < nsol; i++)
	csgtree[i] = (node_t *) read_solid(infile);
    }
  else
    {
      nsol = 1;
      csgtree = (node_t **) new(sizeof(node_t *));
      for (i = 0; i < solnum - 1; i++)
	skip_lines_until(infile, "solid:");
      csgtree[0] = (node_t *) read_solid(infile);
    }

  /* Now traverse all the CSG trees */

#ifdef __GL
  printf("Input the radius of traced spheres: ");
  fscanf(stdin, "%lf", &Sphradius);
#endif

  printf("Input the Step Size: ");
  fscanf(stdin, "%lf", &ssize);

#ifdef READ_SUBDIVIDED_PATCHES
  test_traverse(ssize);
#else
  
  for (i = 0; i < nsol; i++)
    {
      np = csgtree[i];
      if ((np->left == NULL) && (np->right == NULL)) /* single-node */
	write_solid_to_file(np->npatches, np->trim, fl_name);
      else
	traverse_tree(csgtree[i], ssize, TRUE, fl_name);
    }
  
#endif /* READ_SUBDIVIDED_PATCHES */

  return 0;
}

#ifdef READ_SUBDIVIDED_PATCHES

int test_traverse(double ssize)
{
  node_t npl, npr, np;
  int frep, srep;

  printf("test_traverse: enter operation code: ");
  scanf("%d", &np.op);
  read_subdivided_patches(&npl, "p0.in");
  read_subdivided_patches(&npr, "p1.in");

#ifdef __GL
  ip = 0;
#endif

  frep = srep = 1;
  /* changing call to perform_CSG to accomodate splines */
  np.npatches = perform_CSG(&(npl.trim), npl.npatches,
	      &(npr.trim), npr.npatches, frep, srep, 
	      np.op, ssize, &(np.trim));  

  subdivide_trimmed_patches(&np);
  dump_subdivided_patches(&np);
}

#endif  /* READ_SUBDIVIDED_PATCHES */

#define SMALL_DISPLACEMENT 0.01

double angle_between_consecutive_segments(p0, p1, p2) 
POINT2D *p0, *p1, *p2;
{
  POINT2D v1, v2;
  double norm_v1, norm_v2;

  v1.u = ((*p1).u - (*p0).u);
  v1.v = ((*p1).v - (*p0).v);

  v2.u = ((*p2).u - (*p1).u);
  v2.v = ((*p2).v - (*p1).v);

  norm_v1 = sqrt( v1.u*v1.u + v1.v*v1.v );
  norm_v2 = sqrt( v2.u*v2.u + v2.v*v2.v );

  v1.u /= norm_v1;
  v1.v /= norm_v1;

  v2.u /= norm_v2;
  v2.v /= norm_v2;

  return (v1.u*v2.u + v1.v*v2.v);
}

#ifdef SUBODH_REQUEST
int write_solid_to_file(int npatches, TRIM_PATCH_INFO *parent_solid,
			char *fl_name)
{
  int r, p, q;
  int i, j;
  double utmp, vtmp, uval, vval, xval, yval, zval;
  FILE *fp;
  double normX, normY, normZ;
  int is_facing_outside;
  TENSOR_PATCH norm_tp;
  double Scale = 1.0;
  POINT2D old_pt, pt, new_pt;
  int critical_pt_array[MAX_CRITICAL_POINTS_IN_CHAIN];
  int critical_count;
  double actual_chain_array[2*MAX_CRITICAL_POINTS_IN_CHAIN][2];
  double angle;
  int chain_count;
  int is_patch_planar;

  printf("write_solid_to_file: SUBODH_REQUEST is ON. Routine should change\n");
  exit(-1);

  fp = fopen(fl_name, "w");
  for (i=0; i< npatches; i++) {
    
    is_patch_planar = (parent_solid[i].tp.iDeg0 == 1 &&
		       parent_solid[i].tp.iDeg1 == 1);
#ifdef _PRINT_DOTS
    printf(". ");
    fflush(stdout);
#endif
    
#ifdef _DO_PROPER_NORMALS

    DetGaussMap(parent_solid[i].tp, &norm_tp);
    uval = parent_solid[i].TrimPortion[0].ptInsideRegion.u;
    vval = parent_solid[i].TrimPortion[0].ptInsideRegion.v;
    EvalPointOnPatch(parent_solid[i].tp, uval, vval, &xval, &yval, &zval);
    EvalUnitNormal(&norm_tp, uval, vval, &normX, &normY, &normZ);
    xval += normX * SMALL_DISPLACEMENT;
    yval += normY * SMALL_DISPLACEMENT;
    zval += normZ * SMALL_DISPLACEMENT;
    if (RobustInOutTest(xval, yval, zval, parent_solid, npatches) == 0)
      is_facing_outside = 1;
    else is_facing_outside = 0;

#else
    is_facing_outside = 1;
#endif
    
    fprintf(fp,"%d %d\n",
	    parent_solid[i].tp.iDeg0, parent_solid[i].tp.iDeg1);
    for (p=0; p<=parent_solid[i].orig_tp.iDeg0; p++) {
      if (is_facing_outside) 
	{
	  for (q=0; q<=parent_solid[i].orig_tp.iDeg1; q++) {
	    for (r=X; r<=Z; r++)
	      fprintf(fp, "%lf ", 
		      Scale*parent_solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	    fprintf(fp, "%lf\n", 
		    parent_solid[i].orig_tp.coeffs[p][q].dCoord[W]);
	  }
	}
      else 
	{
	  for (q=parent_solid[i].orig_tp.iDeg1; q>=0; q--) {
	    for (r=X; r<=Z; r++)
	      fprintf(fp, "%lf ", 
		      Scale*parent_solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	    fprintf(fp, "%lf\n", 
		    parent_solid[i].orig_tp.coeffs[p][q].dCoord[W]);
	  }
	}
    }

    /* TP has been output. Now print the trim-curves */

    fprintf(fp, "%d\n", parent_solid[i].NumTrimPortions);

    for (r=0; r<parent_solid[i].NumTrimPortions; r++) {
      critical_count = chain_count = 0;
      actual_chain_array[chain_count][0] = old_pt.u = 
		parent_solid[i].TrimPortion[r].poly_vertex[0][0];
      actual_chain_array[chain_count++][1] = old_pt.v = 
		parent_solid[i].TrimPortion[r].poly_vertex[0][1];
      actual_chain_array[chain_count][0] = pt.u = 
		parent_solid[i].TrimPortion[r].poly_vertex[1][0];
      actual_chain_array[chain_count++][1] = pt.v = 
		parent_solid[i].TrimPortion[r].poly_vertex[1][1];
      for (p=2;p<parent_solid[i].TrimPortion[r].NumOfVertices;p++) {
	new_pt.u = parent_solid[i].TrimPortion[r].poly_vertex[p][0];
	new_pt.v = parent_solid[i].TrimPortion[r].poly_vertex[p][1];
	angle = angle_between_consecutive_segments(&old_pt, &pt, &new_pt);
	if (angle < REQ_CRITICAL_ANGLE) 
	  critical_pt_array[critical_count++] = chain_count - 1;
	if (!is_patch_planar || !FP_EQ_EPS(angle, 1.0, 1e-3)) {
	  old_pt = pt;
	  pt = new_pt;
	  actual_chain_array[chain_count][0] = new_pt.u;
	  actual_chain_array[chain_count++][1] = new_pt.v;
	}
      }
      actual_chain_array[chain_count][0] = 
		 parent_solid[i].TrimPortion[r].poly_vertex[p][0];
      actual_chain_array[chain_count][1] = 
		 parent_solid[i].TrimPortion[r].poly_vertex[p][1];
    }
    for (r=0; r<parent_solid[i].NumTrimPortions; r++) 
      {
	fprintf(fp,"%d %d\n", chain_count, 1);
#ifdef _DO_PROPER_NORMALS
	if (is_facing_outside) 
	  {
	    for (p=0; p<=chain_count; p++) 
		{
		  utmp = parent_solid[i].umin + 
		    	 actual_chain_array[p][0]*
		         (parent_solid[i].umax - parent_solid[i].umin);
		  vtmp = parent_solid[i].vmin + 
		    	 actual_chain_array[p][1]*
		         (parent_solid[i].vmax - parent_solid[i].vmin);
		  fprintf(fp,"%lf %lf\n", utmp, vtmp);
		}
	    fprintf(fp,"%d\n", critical_count);
	    for (p=0; p<critical_count; p++) 
	      fprintf(fp,"%d\n", critical_pt_array[p]);
	  }
	else 
	  {
	    for (p=chain_count; p>=0; p--)
	      {
		utmp = parent_solid[i].umin + 
		       actual_chain_array[p][0]*
		       (parent_solid[i].umax - parent_solid[i].umin);
		vtmp = parent_solid[i].vmin + 
		       actual_chain_array[p][1]*
		       (parent_solid[i].vmax - parent_solid[i].vmin);
		fprintf(fp,"%lf %lf\n", utmp, 1.0 - vtmp);
	      }
	    fprintf(fp,"%d\n", critical_count);
	    for (p=critical_count-1; p>=0; p--) 
	      fprintf(fp,"%d\n", chain_count - critical_pt_array[p]);
	  }
#else
	for (p=0; p<=chain_count; p++) 
	{
	  utmp = parent_solid[i].umin + 
		 actual_chain_array[p][0]*
		 (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin + 
		 actual_chain_array[p][1]*
		 (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp,"%lf %lf\n", utmp, vtmp);
	}
	fprintf(fp,"%d\n", critical_count);
	for (p=0; p<critical_count; p++) 
	  fprintf(fp,"%d\n", critical_pt_array[p]);
#endif
      }
  }

  fclose(fp);

  return 0;
}

#else /* no SUBODH_REQUEST */

int write_solid_to_file(int npatches, TRIM_PATCH_INFO *parent_solid,
			char *fl_name)
{
  int r, p, q, s;
  int i, j;
  double utmp, vtmp, uval, vval, xval, yval, zval;
  FILE *fp;
  double normX, normY, normZ;
  int is_facing_outside;
  int ncurves, uvdeg, xyzdeg;
  TENSOR_PATCH norm_tp;
  double Scale = 1.0;

#ifdef _LOTS_OF_HOLES
  int NumElmtsInHole;
  double tmpCurve1[1000][2];
#endif

  fp = fopen(fl_name, "w");

  for (i=0; i< npatches; i++) {
    
#ifdef _PRINT_DOTS
    printf(". ");
    fflush(stdout);
#endif
    
    fprintf(fp,"%d %d\n",
            parent_solid[i].tp.iDeg0, parent_solid[i].tp.iDeg1);
    for (p=0; p<=parent_solid[i].orig_tp.iDeg0; p++)
      for (q=0; q<=parent_solid[i].orig_tp.iDeg1; q++) {
	for (r=X; r<=Z; r++)
	  fprintf(fp, "%lf ", 
		  Scale*parent_solid[i].orig_tp.coeffs[p][q].dCoord[r]);
	fprintf(fp, "%lf\n", 
		parent_solid[i].orig_tp.coeffs[p][q].dCoord[W]);
      }

    /* TP has been output. Now print the trim-curves */

    fprintf(fp, "%d\n", parent_solid[i].NumTrimPortions);
    if (!parent_solid[i].SplineReprn) {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  fprintf(fp, "%d %d\n", 1, -1);
	  fprintf(fp,"%d %d\n",
		  parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices, 1);
	  for (p=0;
	       p<=parent_solid[i].TrimPortion[r].PiecePortion.NumOfVertices;
	       p++) {
	    utmp = parent_solid[i].umin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][0]*
		(parent_solid[i].umax - parent_solid[i].umin);
	    vtmp = parent_solid[i].vmin +
	      parent_solid[i].TrimPortion[r].PiecePortion.poly_vertex[p][1]*
		(parent_solid[i].vmax - parent_solid[i].vmin);
	    fprintf(fp,"%lf %lf\n", utmp, vtmp);
	  }
	  utmp = parent_solid[i].umin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.u*
	      (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin +
	    parent_solid[i].TrimPortion[r].PiecePortion.ptInsideRegion.v*
	      (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp, "%lf %lf\n", utmp, vtmp);
	}
    } else {
      for (r=0; r<parent_solid[i].NumTrimPortions; r++)
	{
	  ncurves = parent_solid[i].TrimPortion[r].SplinePortion.NumOfCurves;
	  fprintf(fp,"%d\n", ncurves);
	  for (p=0; p<ncurves; p++) {
	    fprintf(fp, "%d\n", 
		    parent_solid[i].TrimPortion[r].SplinePortion.AdjInfo[p]);
	    uvdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          UV_Splines[p].Deg; 
	    xyzdeg = parent_solid[i].TrimPortion[r].SplinePortion.
	                                          XYZ_Splines[p].Deg; 
	    fprintf(fp, "%d\n", uvdeg);
	    fprintf(fp, "%d\n", 1+parent_solid[i].SplineReprn);
	    for (q=0; q<=uvdeg; q++) {
	      utmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]) * 
		       parent_solid[i].umin) + (parent_solid[i].umax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[0]);
	      vtmp = ((parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[2] - 
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]) * 
		       parent_solid[i].vmin) + (parent_solid[i].vmax *
		       parent_solid[i].TrimPortion[r].SplinePortion.
		       UV_Splines[p].CntrlPts[q].dCoord[1]);
	      fprintf(fp, "%g %g %g\n", utmp, vtmp,  
			parent_solid[i].TrimPortion[r].SplinePortion.
			UV_Splines[p].CntrlPts[q].dCoord[2]); 
	    }
	    fprintf(fp, "%d\n", xyzdeg);
	    for (q=0; q<=xyzdeg; q++) {
	      for (s=0; s<4; s++)
		fprintf(fp, "%g ", 
			parent_solid[i].TrimPortion[r].SplinePortion.
			XYZ_Splines[p].CntrlPts[q].dCoord[s]); 
	      fprintf(fp, "\n");
	    }
	  }
/*
	  utmp = parent_solid[i].umin +
	    parent_solid[i].TrimPortion[r].SplinePortion.ptInsideRegion.u*
	      (parent_solid[i].umax - parent_solid[i].umin);
	  vtmp = parent_solid[i].vmin +
	    parent_solid[i].TrimPortion[r].SplinePortion.ptInsideRegion.v*
	      (parent_solid[i].vmax - parent_solid[i].vmin);
	  fprintf(fp, "%lf %lf\n", utmp, vtmp);
*/
	}
    }
    
  }

  fclose(fp);

  return 0;
}
#endif
