#ifndef _intersect_h
#define _intersect_h

#include <triangulate.h>

#define MAX_INTERSECTIONS_PER_EDGE 8
#define POLYCHAIN_SIZE  1000
#define MAX_POLYCHAINS 100
#define MAX_INTERSECTIONS_PER_SEGMENT 5
#define MAX_CHAIN_LENGTH 300
#define MAX_STACKSIZE 30
#define MAX_PARTITIONS 60
#define MAX_PARTITION_SIZE 300
#define MAX_NEIGHBOURS 10

/* Structure to hold intersection info */
typedef struct {	
  int nch;			/* chain number */
  int nseg;			/* segment number within the chain*/
  int inum;			/* Intersection number (w.r.t. the */
				/* chain) */
  double s, t;
} iinfo_t;

typedef struct {
  point_typ v0, v1;               /* endpoints of the edge */
  int nintr;                    /* Intersection info per edge */
  iinfo_t intr[MAX_INTERSECTIONS_PER_EDGE];
  int epos;                     /* position in the poly[] structure */
} einfo_t;

typedef struct {
  point_typ v;                    /* intersection point */
  int nedge;                    /* which edge do I intersect ? */
  double s, t;                  /* Intrn. info for chain segment */
  int h;                        /* hack-factor */
} sinfo_t;


typedef struct {
  point_typ v;                    /* point-value */
  int valid;                    /* chain still going !! */
} pchain_t;


typedef struct {
  point_typ v;                    /* point-value */
  int edgenum;                  /* segment number of which the point */
                                /* is the tail */
  int next;
} polychain_t;


typedef struct {
  int chnum;			/* which chain ? */
  int segnum;			/* segment number in the chain */
  int edgenum;			/* edge in the polygon */
  int epos;			/* psn of the edge (at the time of */
				/* pushing in the stack) */
  int inum;			/* intersection number */
  int einum;			/* the posn. of this intersection */
				/* amongst al the intrns.. of the edge */
  int itype;			/* not used */

  int adj[MAX_NEIGHBOURS];	/* adjacency list for neighbours */
  int nidx;			/* neighbour index */

  double s, t;
} stack_t;


typedef struct {
  int npoints;			/* #points in the partition */
  int whichside;		/* what side of the chain the */
				/* partition lies!! */
  point_typ *pc;		/* array of points */
  
  int adj[MAX_NEIGHBOURS];	/* neighbouring partitions */
  int nnb;			/* #neighbours */
} part_t;


#define I_TOLERANCE 1.0e-6	/* was 1.0e-8 previously */

#define CROSS_VEC(v0, v1) ((v0).x*(v1).y - (v1).x*(v0).y)
#define FP_EQUAL(s, t) (fabs(s - t) <= I_TOLERANCE)

#define PT_TOLERANCE 1.0e-3
#define FP_EQUAL_POINT(s, t) (fabs(s - t) <= PT_TOLERANCE)

#define FP_EQUAL_PAR(a, b, tol)  (fabs((a) - (b)) <= tol)

#define I_S_ON    1
#define I_T_ON    2
#define I_NONE_ON 3		/* Segment intersection types */
#define I_ST_ON   4
#define I_PAR     5
#define I_ANTIPAR 6

#define I_FIRSTPT 1
#define I_LASTPT  2
#define I_NONEPT  3


/* Externally visible functions */

extern int segments_intersect(point_typ *u0, point_typ *u1, point_typ *v0,
                              point_typ *v1, int *intersect_type,
                              double *ss0, double *ss1,
                              double *tt0, double *tt1,
                              point_typ *p0, point_typ *p1)  ;

extern int shady_segments_intersect(point_typ *u0, point_typ *u1,
                                    point_typ *v0,
                                    point_typ *v1, int *intersect_type,
                                    double *ss0, double *ss1,
                                    double *tt0, double *tt1,
                                    point_typ *p0, point_typ *p1,
                                    int is_first_last);

#endif /* _intersect_h */
