#include <partition.h>
#include <math.h>

#define PAR_EDGE 2		/* Used for determining whether */
#define PAR_SEG  3		/* polygon is simple or not */

extern node_t qs[];		/* Query structure */
extern trap_t tr[];		/* Trapezoid structure */
extern segment_t seg[SEGSIZE];	/* Segment table */

extern einfo_t edge[];
extern pchain_t *pchain[];
struct global_s global;

#define SQR_LENGTH(v0, v1) (((v0).x - (v1).x)*((v0).x - (v1).x) + \
			     ((v0).y - (v1).y)*((v0).y - (v1).y))

#ifdef STAND_ALONE
main()
{
  point_typ u0, u1, v0, v1, p0, p1;
  double ss0, ss1, tt0, tt1;
  int n, itype, i, j;
  part_t *part;
  
  n = read_data();
  find_chain_intersections(&n);
  printf("n = %d\n", n);
  n = trace_around_edges(&part);

#ifdef DEBUG1
  printf("Partitions:\n");
  for (i = 0; i < n; i++)
    {
      printf("\n\npartition %d:\n\t", i);
      for (j = 0; j < part[i].npoints; j++)
	printf("(%.2lf, %.2lf) ", part[i].pc[j].x, part[i].pc[j].y);
    }
  printf("\n");
#endif
}

#endif

#ifdef __LINK

static int initialise(int nseg)
{
  register int i;

  for (i = 1; i <= nseg; i++)
    seg[i].is_inserted = FALSE;
  generate_random_ordering(nseg);
  
  return 0;
}

static int find_centroid(int triangle[], point_typ *p)
{
  p->x = (seg[triangle[0]].v0.x + seg[triangle[1]].v0.x +
	  seg[triangle[2]].v0.x)/3.0;
  p->y = (seg[triangle[0]].v0.y + seg[triangle[1]].v0.y +
	  seg[triangle[2]].v0.y)/3.0;
  
  return 0;
}


/* Determine whether the polygon is simple or not */
int is_simple_polygon(int param_eos, int npoints)
{
  register int i, j;
  point_typ p0, p1;
  double s0, s1, t0, t1;
  int itype;

  switch(param_eos)
    {
    case PAR_EDGE:
      for (i = 0; i < npoints; i++)
	for (j = (i+2)%npoints; j != (i + npoints - 1)%npoints; 
	     j = (j + 1)%npoints)
	  if (segments_intersect(&edge[i].v0, &edge[i].v1,
				 &edge[j].v0, &edge[j].v1, &itype,
				 &s0, &s1, &t0, &t1, &p0, &p1) > 0)
	    return FALSE;
      
      return TRUE;
      break;
      
    case PAR_SEG:
      for (i = 0; i < npoints; i++)
	for (j = (i+2)%npoints; j != (i + npoints - 1)%npoints; 
	     j = (j + 1)%npoints)
	  if (segments_intersect(&seg[i+1].v0, &seg[i+1].v1,
				 &seg[j+1].v0, &seg[j+1].v1, &itype,
				 &s0, &s1, &t0, &t1, &p0, &p1) > 0)
	    return FALSE;
      
      return TRUE;
      break;
      
    }

  return TRUE;
}



int find_point_inside_polygon(int npoints, double vertices[][2],
			      point_typ *p) 
{
  register int i;
  int nmonpoly, max;
  int triangles[SEGSIZE][3];
  double max_area, tmp;


#ifndef TRY_STUNT
  for (i = 0; i < npoints - 1; i++)
    if (FP_EQUAL_PAR(vertices[i][1], vertices[i+1][1], 1e-4))
      vertices[i+1][1] = vertices[i][1];

#endif

  memset((void *)seg, 0, sizeof(seg));
  for (i = 1; i <= npoints; i++)
    {
      seg[i].is_inserted = FALSE;
      
      seg[i].v0.x = vertices[i-1][0]; /* x-cood */
      seg[i].v0.y = vertices[i-1][1]; /* y-cood */
      if (i == 1)
        {
          seg[npoints].v1.x = seg[i].v0.x;
          seg[npoints].v1.y = seg[i].v0.y;
        }
      else
        {
          seg[i - 1].v1.x = seg[i].v0.x;
          seg[i - 1].v1.y = seg[i].v0.y;
        }       
#ifdef DEBUG_PRINT_POLY
      printf("%.17lf %.17lf\n", seg[i].v0.x, seg[i].v0.y);
      fflush(stdout);
#endif
    }

  global.nseg = npoints;
  initialise(npoints);
  construct_trapezoids(npoints, seg);
  nmonpoly = monotonate_trapezoids(npoints);
  triangulate_monotone_polygons(nmonpoly, triangles);

  /* Find the triangle with the maximum area hoping that we do not get */
  /* a sliver and return a point which is too close to the boundary of */
  /* the polygon */ 

  max = 0;
  max_area = CROSS(seg[triangles[0][0]].v0, seg[triangles[0][1]].v0,
		       seg[triangles[0][2]].v0);
  for (i = 1; i < npoints - 2; i++)
    if ((tmp =  CROSS(seg[triangles[i][0]].v0, seg[triangles[i][1]].v0,
		       seg[triangles[i][2]].v0)) > max_area)
      {
	max_area = tmp;
	max = i;
      }
  
  find_centroid(triangles[max], p);
  
  return 0;
}


int partition_polygon(TRIM_PORTION_TYPE *domain,
		      DOMAIN_INT_CURVE *chains, int nchain,
		      TRIM_PATCH_INFO *tpatch, neighbour_t *n_info)
{
  register int i, k;
  int j, n, npartitions, npts, tcount;
  LIST_POINT2D *lp;
  TRIM_PORTION_TYPE *trimp;
  part_t part[MAX_PARTITIONS];
  point_typ p, tarray[MAX_CHAIN_LENGTH];


  global.nseg = n = domain->PiecePortion.NumOfVertices; /* #points */
  for (i = 0; i < n; i++)
    {
      edge[i].v0.x = domain->PiecePortion.poly_vertex[i][0];
      edge[i].v0.y = domain->PiecePortion.poly_vertex[i][1];
      if (i == 0)
        {
          edge[n-1].v1.x = edge[i].v0.x;
          edge[n-1].v1.y = edge[i].v0.y;
        }
      else
        {
          edge[i - 1].v1.x = edge[i].v0.x;
          edge[i - 1].v1.y = edge[i].v0.y;
        }       
#ifdef DEBUG_PRINT_POLY
      printf("%lf %lf\n", edge[i].v0.x, edge[i].v0.y);
      fflush(stdout);
#endif
    }

#ifdef CHECK_SIMPLICITY
  if (!is_simple_polygon(PAR_EDGE, n))
    {
      fprintf(stderr, "partition_polygon: Polygon not simple\n");
      return -1;
    }
#endif

  /* Now read in the chains */

  for (i = 0; i < nchain; i++)
    {
      pchain[i] = (pchain_t *)
	malloc(MAX_CHAIN_LENGTH*sizeof(pchain_t));
      
#ifdef DEBUG_PRINT_POLY
	printf("\nChain number %d\n", i);
#endif

      lp = chains[i].planeCurve;
      j = 0;
      while (lp != NULL)
	{
	  pchain[i][j].v.x = lp->p2D.u;
	  pchain[i][j].v.y = lp->p2D.v;
	  pchain[i][j].valid = TRUE;
	  
#ifdef DEBUG_PRINT_POLY
         printf("%.13lf %.13lf\n", pchain[i][j].v.x, pchain[i][j].v.y);
         fflush(stdout);
#endif
	  j++;
	  lp = lp->plp2DNext;
	}

#ifdef DETECT_CHAIN_KINKS
      npts = j;
      straighten_chain(pchain[i], &npts);
      j = npts;

      while (remove_kinks_from_chain(pchain[i], &j));

#ifndef TRY_STUNT

      /* Code modified 17/4/95
	 
	 for (k = 0; k < j - 1; k++)
	 if (FP_EQUAL_PAR(pchain[i][k].v.y, pchain[i][k+1].v.y, 1e-4))
	 pchain[i][k+1].v.y = pchain[i][k].v.y;
	 */
      
      tarray[0] = pchain[i][0].v;
      tcount = 1;
      for (k = 0; k < j - 2; k++)
	{
	  if (FP_EQUAL_PAR(pchain[i][k].v.x, pchain[i][k+1].v.x, 1e-3)
	      &&
	      FP_EQUAL_PAR(pchain[i][k].v.y, pchain[i][k+1].v.y,
			   1e-3))
	    continue;
	  else
	    tarray[tcount++] = pchain[i][k+1].v;
	}
      tarray[tcount++] = pchain[i][j-1].v;

      j = tcount;
      for (k = 0; k < j; k++)
	pchain[i][k].v = tarray[k];
      pchain[i][j].valid = FALSE;


  for (k = 0; k < j; k++)
    {
      if (FP_EQUAL_PAR(pchain[i][k].v.y, 1.0, 1e-3))
	pchain[i][k].v.y = 1.0;
      else if (FP_EQUAL_PAR(pchain[i][k].v.y, 0.0, 1e-3))
	pchain[i][k].v.y = 0.0;

      if (FP_EQUAL_PAR(pchain[i][k].v.x, 1.0, 1e-3))
	pchain[i][k].v.x = 1.0;
      else if (FP_EQUAL_PAR(pchain[i][k].v.x, 0.0, 1e-3))
	pchain[i][k].v.x = 0.0;
    }

#ifdef DEBUG_PRINT_POLY
      printf("Same chain after optimizations\n");
      for (k = 0; k < j; k++)
	printf("%lf %lf\n", pchain[i][k].v.x, pchain[i][k].v.y);
      fflush(stdout);
#endif

#endif

      
#endif
      pchain[i][j].valid = FALSE;

#ifdef DEBUG_PRINT_POLY
      printf("Chain i complete\n");
      fflush(stdout);
#endif

    }

  find_chain_intersections(&nchain);
  npartitions = trace_around_edges(part);
  if (npartitions == -1)
    return -1;

#ifdef DEBUG_PRINT_POLY
  printf("Partitions:\n");
  for (i = 0; i < npartitions; i++)
    {
      printf("\n\npartition %d:\n\t", i);
      for (j = 0; j < part[i].npoints; j++)
	printf("%.13lf, %.13lf\n", part[i].pc[j].x, part[i].pc[j].y);
    }
  printf("\n");
  fflush(stdout);
#endif 


  /* Now put back the data in proper form */
  
  tpatch->NumTrimPortions = npartitions;
  for (i = 0; i < npartitions; i++)
    {
      trimp = &tpatch->TrimPortion[i];

      trimp->PiecePortion.NumOfVertices = part[i].npoints;

      /* Special handling for loops  (Dec 29) .. the code is growing */
      /* like a fat PIG ......*/

      if (FP_EQUAL(part[i].pc[part[i].npoints - 1].x,
		   part[i].pc[0].x) &&
	  FP_EQUAL(part[i].pc[part[i].npoints - 1].y,
		   part[i].pc[0].y))
	{
	  part[i].npoints--;
	  trimp->PiecePortion.NumOfVertices--;
	}

      for (j = 0; j < part[i].npoints; j++)
	{
	  trimp->PiecePortion.poly_vertex[j][0] = part[i].pc[j].x;
	  trimp->PiecePortion.poly_vertex[j][1] = part[i].pc[j].y;
	}
	trimp->PiecePortion.poly_vertex[j][0] = part[i].pc[0].x;
	trimp->PiecePortion.poly_vertex[j][1] = part[i].pc[0].y;
      
      /* Now find point inside each partition */
      
      find_point_inside_polygon(part[i].npoints,
				trimp->PiecePortion.poly_vertex, &p);
      trimp->PiecePortion.ptInsideRegion.u = p.x;
      trimp->PiecePortion.ptInsideRegion.v = p.y;

      free(part[i].pc); 
    }

  for (i = 0; i < nchain; i++)
    free(pchain[i]);

  /* Copy neighbour information */
  for (i = 0; i < npartitions; i++)
    {
      n_info[i].nnb = part[i].nnb;
      memcpy((void *) n_info[i].adj, (void *) part[i].adj, MAX_NEIGHBOURS*sizeof(int));
    }

  return 0;
}



/* The points constituting the polygon are specified in anticlockwise */
/* order. The first is repeated (as vertices[0] and vertices[n]) */
/* Before using this routine, please check-out that you do not */
/* conflict with the global variables defined in basic.h. Return TRUE */
/* if the point lies inside the polygon or even on the boundary */

int is_point_inside_polygon(int n, double vertices[][2], point_typ *v)
{
  register int i;
  int trnum, rseg;
  trap_t *t;
  extern struct global_s global;
  double k1, k2, k3;

#ifndef TRY_STUNT
  for (i = 0; i < n - 1; i++)
    if (FP_EQUAL_PAR(vertices[i][1], vertices[i+1][1], 0.5e-4))
      vertices[i+1][1] = vertices[i][1];
#endif

  memset((void *)seg, 0, sizeof(seg));
  for (i = 1; i <= n; i++)
    {
      seg[i].is_inserted = FALSE;

      seg[i].v0.x = vertices[i][0]; /* x-cood */
      seg[i].v0.y = vertices[i][1]; /* y-cood */
      if (i == 1)
	{
	  seg[n].v1.x = seg[i].v0.x;
	  seg[n].v1.y = seg[i].v0.y;
	}
      else
	{
	  seg[i - 1].v1.x = seg[i].v0.x;
	  seg[i - 1].v1.y = seg[i].v0.y;
	}	
    }

#ifdef CHECK_SIMPLICITY
  if (!is_simple_polygon(PAR_SEG, n))
    {
#ifdef DEBUG_PRINT_POLY
      printf("BAD POLYGON\n");
      for (i=0; i<=n; i++) 
	printf("%10.8lf %10.8lf\n", vertices[i][0], vertices[i][1]);
#endif
      fprintf(stderr, "is_point_inside_polygon: Polygon not simple\n");
      sleep(3);
      return -1;
    }
#endif

  global.nseg = n;
  initialise(n);
  construct_trapezoids(n, seg);

  /* first check if the point lies on any of the edges. If yes, return */
  /* TRUE. */

  for (i = 1; i <= n; i++)
    if (fabs(CROSS(seg[i].v0, seg[i].v1, *v)) <= I_TOLERANCE)
      {
	k1 = SQR_LENGTH(seg[i].v0, seg[i].v1);
	k2 = SQR_LENGTH(seg[i].v0, *v);
	k3 = SQR_LENGTH(seg[i].v1,*v);
	
	if ((k2 <= k1) && (k3 <= k1))
	  return TRUE;
      }
  
  trnum = locate_endpoint(v, v, 1);
  t = &tr[trnum];

  if (t->state == ST_INVALID)
    return 0;

  if ((t->lseg <= 0) || (t->rseg <= 0))
    return 0;
  rseg = t->rseg;
  return _greater_than_equal_to(&seg[rseg].v1, &seg[rseg].v0);
}


/* Retrun TRUE if the curve is in ANTI-clockwise order. False */
/* otherwise */
/* n: # of distinct points. there, i/p = p0, p1, ..... pn (=p0) */

int is_curve_anticlockwise(double curve[][2], int n)
{
  extern struct global_s global;
  int trnum, rseg;
  register int i, j;

  memset((void *)seg, 0, sizeof(seg));
  for (i = 1; i <= n; i++)
    {
      seg[i].is_inserted = FALSE;

      seg[i].v0.x = curve[i][0]; /* x-cood */
      seg[i].v0.y = curve[i][1]; /* y-cood */
      if (i == 1)
	{
	  seg[n].v1.x = seg[i].v0.x;
	  seg[n].v1.y = seg[i].v0.y;
	}
      else
	{
	  seg[i - 1].v1.x = seg[i].v0.x;
	  seg[i - 1].v1.y = seg[i].v0.y;
	}	
    }

#ifdef CHECK_SIMPLICITY
  if (!is_simple_polygon(PAR_SEG, n))
    {
      fprintf(stderr, "is_curve_anticlockwise: Polygon not simple\n");
      return -1;
    }
#endif

  global.nseg = n;
  initialise(n);
  construct_trapezoids(n, seg);

  /* Now search for a trapezoid which lies on the left side of the */
  /* boundary of the polygon */

  for (i = 0; i < TRSIZE; i++)
    {
      if ((tr[i].state == ST_VALID) &&
	  (tr[i].lseg <= 0) && ((rseg = tr[i].rseg) > 0) &&
	  (fabs(seg[rseg].v0.y - seg[rseg].v1.y) > I_TOLERANCE))
	break;
    }
  
  if (i >= TRSIZE)
    printf("is_curve_anticlockwise: bete tera band baja baz gaya!\n");
  else
    {
      rseg = tr[i].rseg;
      if (_greater_than_equal_to(&seg[rseg].v0, &seg[rseg].v1))
	return TRUE;
      else
	return FALSE;
    }

  return TRUE;
}

#endif


/* Return TRUE only if the point lies inside the polygon (i.e. even if */
/* it lies on the edge, we ahould return FALSE) */

int is_point_within_polygon(point_typ *v)
{
  register int i;
  int trnum, rseg, n;
  trap_t *t;
  double k1, k2, k3;

  /* first check if the point lies on any of the edges. If yes, return */
  /* FALSE. */

  n = global.nseg;
  for (i = 1; i <= n; i++)
    if (fabs(CROSS(seg[i].v0, seg[i].v1, *v)) <= I_TOLERANCE)
      {
	k1 = SQR_LENGTH(seg[i].v0, seg[i].v1);
	k2 = SQR_LENGTH(seg[i].v0, *v);
	k3 = SQR_LENGTH(seg[i].v1,*v);
	
	if ((k2 <= k1) && (k3 <= k1))
	  return FALSE;
      }
  
  trnum = locate_endpoint(v, v, 1);
  t = &tr[trnum];

  if (t->state == ST_INVALID)
    return FALSE;

  if ((t->lseg <= 0) || (t->rseg <= 0))
    return FALSE;
  rseg = t->rseg;
  return _greater_than_equal_to(&seg[rseg].v1, &seg[rseg].v0);
}
