#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "linalg.h"

#define NUM_ARB_PATCHES 6

EXTERN_LINALG
extern void allocate_tensor_patch();
extern void XformPatch();

/******************** GetArbControlPoints *************/
/* Get Control Points for the Arb8                    */
/******************************************************/
void GetArbControlPoints(tp, p0,p1,p2,p3,p4,p5,p6,p7)
TENSOR_PATCH *tp;
POINT4D p0,p1,p2,p3,p4,p5,p6,p7;
{
   int i;

   /******* patch 0 *******/

   for(i=0;i<3;i++) {
      tp[0].coeffs[0][0].dCoord[i] = p3.dCoord[i];
      tp[0].coeffs[0][1].dCoord[i] = p0.dCoord[i];
      tp[0].coeffs[1][0].dCoord[i] = p2.dCoord[i];
      tp[0].coeffs[1][1].dCoord[i] = p1.dCoord[i];
   }
   tp[0].coeffs[0][0].dCoord[3] = 1.0;
   tp[0].coeffs[0][1].dCoord[3] = 1.0;
   tp[0].coeffs[1][0].dCoord[3] = 1.0;
   tp[0].coeffs[1][1].dCoord[3] = 1.0;


   /******* patch 1 *******/

   for(i=0;i<3;i++) {
      tp[1].coeffs[0][0].dCoord[i] = p4.dCoord[i];
      tp[1].coeffs[0][1].dCoord[i] = p7.dCoord[i];
      tp[1].coeffs[1][0].dCoord[i] = p5.dCoord[i];
      tp[1].coeffs[1][1].dCoord[i] = p6.dCoord[i];
   }
   tp[1].coeffs[0][0].dCoord[3] = 1.0;
   tp[1].coeffs[0][1].dCoord[3] = 1.0;
   tp[1].coeffs[1][0].dCoord[3] = 1.0;
   tp[1].coeffs[1][1].dCoord[3] = 1.0;

   /******* patch 2 *******/

   for(i=0;i<3;i++) {
      tp[2].coeffs[0][0].dCoord[i] = p0.dCoord[i];
      tp[2].coeffs[0][1].dCoord[i] = p4.dCoord[i];
      tp[2].coeffs[1][0].dCoord[i] = p1.dCoord[i];
      tp[2].coeffs[1][1].dCoord[i] = p5.dCoord[i];
   }
   tp[2].coeffs[0][0].dCoord[3] = 1.0;
   tp[2].coeffs[0][1].dCoord[3] = 1.0;
   tp[2].coeffs[1][0].dCoord[3] = 1.0;
   tp[2].coeffs[1][1].dCoord[3] = 1.0;

   /******* patch 3 *******/

   for(i=0;i<3;i++) {
      tp[3].coeffs[0][0].dCoord[i] = p1.dCoord[i];
      tp[3].coeffs[0][1].dCoord[i] = p5.dCoord[i];
      tp[3].coeffs[1][0].dCoord[i] = p2.dCoord[i];
      tp[3].coeffs[1][1].dCoord[i] = p6.dCoord[i];
   }
   tp[3].coeffs[0][0].dCoord[3] = 1.0;
   tp[3].coeffs[0][1].dCoord[3] = 1.0;
   tp[3].coeffs[1][0].dCoord[3] = 1.0;
   tp[3].coeffs[1][1].dCoord[3] = 1.0;

   /******* patch 4 *******/

   for(i=0;i<3;i++) {
      tp[4].coeffs[0][0].dCoord[i] = p2.dCoord[i];
      tp[4].coeffs[0][1].dCoord[i] = p6.dCoord[i];
      tp[4].coeffs[1][0].dCoord[i] = p3.dCoord[i];
      tp[4].coeffs[1][1].dCoord[i] = p7.dCoord[i];
   }
   tp[4].coeffs[0][0].dCoord[3] = 1.0;
   tp[4].coeffs[0][1].dCoord[3] = 1.0;
   tp[4].coeffs[1][0].dCoord[3] = 1.0;
   tp[4].coeffs[1][1].dCoord[3] = 1.0;

   /******* patch 5 *******/

   for(i=0;i<3;i++) {
      tp[5].coeffs[0][0].dCoord[i] = p3.dCoord[i];
      tp[5].coeffs[0][1].dCoord[i] = p7.dCoord[i];
      tp[5].coeffs[1][0].dCoord[i] = p0.dCoord[i];
      tp[5].coeffs[1][1].dCoord[i] = p4.dCoord[i];
   }
   tp[5].coeffs[0][0].dCoord[3] = 1.0;
   tp[5].coeffs[0][1].dCoord[3] = 1.0;
   tp[5].coeffs[1][0].dCoord[3] = 1.0;
   tp[5].coeffs[1][1].dCoord[3] = 1.0;

}


/******************** gen_brlarb8 ******************************/
/* The "main" routine to create an arb8 (8 vertex polyhedron)  */
/***************************************************************/
int gen_brlarb8(p0,p1,p2,p3,p4,p5,p6,p7,tp)
POINT4D p0,p1,p2,p3,p4,p5,p6,p7;
TENSOR_PATCH **tp;
{
   int i;

   ALLOCN(*tp, TENSOR_PATCH, NUM_ARB_PATCHES);
   for(i=0;i<NUM_ARB_PATCHES;i++) {
      (*tp)[i].iDeg1 = 1;
      (*tp)[i].iDeg0 = 1;
      allocate_tensor_patch(&((*tp)[i]));
   }

   GetArbControlPoints(*tp,p0,p1,p2,p3,p4,p5,p6,p7);

   return NUM_ARB_PATCHES;
}
