#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defs.h"
#include "linalg.h"

#define NUM_TGC_PATCHES 6
#define TGC_LEEWAY_FACTOR 1.1
#define TGC_SDEGREE 2
#define TGC_TDEGREE 1

EXTERN_LINALG
extern void allocate_tensor_patch();
extern void XformPatch();

/******************** getXformMatTgc ******************/
/* Generate transform matrix to shift and rotate      */
/*    the truncated general cone.                     */
/******************************************************/
void getXformMatTgc(center,a,b,c,Xform)
POINT4D center, a, b, c;
double **Xform;
{
   Xform[0][0] = a.dCoord[0];
   Xform[1][0] = b.dCoord[0];
   Xform[2][0] = c.dCoord[0];
   Xform[3][0] = center.dCoord[0];

   Xform[0][1] = a.dCoord[1];
   Xform[1][1] = b.dCoord[1];
   Xform[2][1] = c.dCoord[1];
   Xform[3][1] = center.dCoord[1];

   Xform[0][2] = a.dCoord[2];
   Xform[1][2] = b.dCoord[2];
   Xform[2][2] = c.dCoord[2];
   Xform[3][2] = center.dCoord[2];

   Xform[0][3] = 0.0;
   Xform[1][3] = 0.0;
   Xform[2][3] = 0.0;
   Xform[3][3] = 1.0;
}

/******************** getXformMatTgc2 *****************/
/* Generate transform matrix to shift and rotate      */
/*    the truncated general cone.  Includes rando     */
/*    rotation.                                       */
/******************************************************/
void getXformMatTgc2(center,a,b,c,rotamt,Xform)
POINT4D center, a, b, c;
double **Xform;
double rotamt;
{
   double crot;
   double srot;

   crot = cos(rotamt);
   srot = sin(rotamt);

   Xform[0][0] = a.dCoord[0]*crot+b.dCoord[0]*srot;
   Xform[1][0] = b.dCoord[0]*crot-a.dCoord[0]*srot;
   Xform[2][0] = c.dCoord[0];
   Xform[3][0] = center.dCoord[0];

   Xform[0][1] = a.dCoord[1]*crot+b.dCoord[1]*srot;
   Xform[1][1] = b.dCoord[1]*crot-a.dCoord[1]*srot;
   Xform[2][1] = c.dCoord[1];
   Xform[3][1] = center.dCoord[1];

   Xform[0][2] = a.dCoord[2]*crot+b.dCoord[2]*srot;
   Xform[1][2] = b.dCoord[2]*crot-a.dCoord[2]*srot;
   Xform[2][2] = c.dCoord[2];
   Xform[3][2] = center.dCoord[2];

   Xform[0][3] = 0.0;
   Xform[1][3] = 0.0;
   Xform[2][3] = 0.0;
   Xform[3][3] = 1.0;
}

/******************** GetTgcControlPoints *************/
/* Get Control Points for the TGC                     */
/******************************************************/
void GetTgcControlPoints(tp, height, a, b, c, d, c_vec, d_vec)
TENSOR_PATCH *tp;
POINT4D height;
POINT4D c_vec;
POINT4D d_vec;
double a;
double b;
double c;
double d;

/* note: we need to set upper coords to (c*c_vec + d*d_vec) + height */
{
   double invsq2;

   invsq2 = sqrt(2.0)/2.0;

   /******* patch 0 *******/
   tp[0].coeffs[0][0].dCoord[X] = a;
   tp[0].coeffs[0][0].dCoord[Y] = 0.0;
   tp[0].coeffs[0][0].dCoord[Z] = 0.0;
   tp[0].coeffs[0][0].dCoord[W] = 1.0;

   tp[0].coeffs[0][1].dCoord[X] = c*c_vec.dCoord[0]+height.dCoord[0];
   tp[0].coeffs[0][1].dCoord[Y] = c*c_vec.dCoord[1]+height.dCoord[1];
   tp[0].coeffs[0][1].dCoord[Z] = c*c_vec.dCoord[2]+height.dCoord[2];
   tp[0].coeffs[0][1].dCoord[W] = 1.0;

   tp[0].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[0].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[0].coeffs[1][0].dCoord[Z] = 0.0;
   tp[0].coeffs[1][0].dCoord[W] = invsq2;

   tp[0].coeffs[1][1].dCoord[X] = (c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0])*invsq2;
   tp[0].coeffs[1][1].dCoord[Y] = (c*c_vec.dCoord[1]+d*d_vec.dCoord[1]
                                   +height.dCoord[1])*invsq2;
   tp[0].coeffs[1][1].dCoord[Z] = (c*c_vec.dCoord[2]+d*d_vec.dCoord[2]
                                   +height.dCoord[2])*invsq2;
   tp[0].coeffs[1][1].dCoord[W] = invsq2;

   tp[0].coeffs[2][0].dCoord[X] = 0.0;
   tp[0].coeffs[2][0].dCoord[Y] = b;
   tp[0].coeffs[2][0].dCoord[Z] = 0.0;
   tp[0].coeffs[2][0].dCoord[W] = 1.0;

   tp[0].coeffs[2][1].dCoord[X] = d*d_vec.dCoord[0]+height.dCoord[0];
   tp[0].coeffs[2][1].dCoord[Y] = d*d_vec.dCoord[1]+height.dCoord[1];
   tp[0].coeffs[2][1].dCoord[Z] = d*d_vec.dCoord[2]+height.dCoord[2];
   tp[0].coeffs[2][1].dCoord[W] = 1.0;

   /******* patch 1 *******/
   a=a*(-1.0);
   c=c*(-1.0);
   tp[1].coeffs[0][0].dCoord[X] = 0.0;
   tp[1].coeffs[0][0].dCoord[Y] = b;
   tp[1].coeffs[0][0].dCoord[Z] = 0.0;
   tp[1].coeffs[0][0].dCoord[W] = 1.0;

   tp[1].coeffs[0][1].dCoord[X] = d*d_vec.dCoord[0]+height.dCoord[0];
   tp[1].coeffs[0][1].dCoord[Y] = d*d_vec.dCoord[1]+height.dCoord[1];
   tp[1].coeffs[0][1].dCoord[Z] = d*d_vec.dCoord[2]+height.dCoord[2];
   tp[1].coeffs[0][1].dCoord[W] = 1.0;

   tp[1].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[1].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[1].coeffs[1][0].dCoord[Z] = 0.0;
   tp[1].coeffs[1][0].dCoord[W] = invsq2;

   tp[1].coeffs[1][1].dCoord[X] = (c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0])*invsq2;
   tp[1].coeffs[1][1].dCoord[Y] = (c*c_vec.dCoord[1]+d*d_vec.dCoord[1]
                                   +height.dCoord[1])*invsq2;
   tp[1].coeffs[1][1].dCoord[Z] = (c*c_vec.dCoord[2]+d*d_vec.dCoord[2]
                                   +height.dCoord[2])*invsq2;
   tp[1].coeffs[1][1].dCoord[W] = invsq2;

   tp[1].coeffs[2][0].dCoord[X] = a;
   tp[1].coeffs[2][0].dCoord[Y] = 0.0;
   tp[1].coeffs[2][0].dCoord[Z] = 0.0;
   tp[1].coeffs[2][0].dCoord[W] = 1.0;

   tp[1].coeffs[2][1].dCoord[X] = c*c_vec.dCoord[0]+height.dCoord[0];
   tp[1].coeffs[2][1].dCoord[Y] = c*c_vec.dCoord[1]+height.dCoord[1];
   tp[1].coeffs[2][1].dCoord[Z] = c*c_vec.dCoord[2]+height.dCoord[2];
   tp[1].coeffs[2][1].dCoord[W] = 1.0;

   /******* patch 2 *******/
   b=b*(-1.0);
   d=d*(-1.0);
   tp[2].coeffs[0][0].dCoord[X] = a;
   tp[2].coeffs[0][0].dCoord[Y] = 0.0;
   tp[2].coeffs[0][0].dCoord[Z] = 0.0;
   tp[2].coeffs[0][0].dCoord[W] = 1.0;

   tp[2].coeffs[0][1].dCoord[X] = c*c_vec.dCoord[0]+height.dCoord[0];
   tp[2].coeffs[0][1].dCoord[Y] = c*c_vec.dCoord[1]+height.dCoord[1];
   tp[2].coeffs[0][1].dCoord[Z] = c*c_vec.dCoord[2]+height.dCoord[2];
   tp[2].coeffs[0][1].dCoord[W] = 1.0;

   tp[2].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[2].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[2].coeffs[1][0].dCoord[Z] = 0.0;
   tp[2].coeffs[1][0].dCoord[W] = invsq2;

   tp[2].coeffs[1][1].dCoord[X] = (c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0])*invsq2;
   tp[2].coeffs[1][1].dCoord[Y] = (c*c_vec.dCoord[1]+d*d_vec.dCoord[1]
                                   +height.dCoord[1])*invsq2;
   tp[2].coeffs[1][1].dCoord[Z] = (c*c_vec.dCoord[2]+d*d_vec.dCoord[2]
                                   +height.dCoord[2])*invsq2;
   tp[2].coeffs[1][1].dCoord[W] = invsq2;

   tp[2].coeffs[2][0].dCoord[X] = 0.0;
   tp[2].coeffs[2][0].dCoord[Y] = b;
   tp[2].coeffs[2][0].dCoord[Z] = 0.0;
   tp[2].coeffs[2][0].dCoord[W] = 1.0;

   tp[2].coeffs[2][1].dCoord[X] = d*d_vec.dCoord[0]+height.dCoord[0];
   tp[2].coeffs[2][1].dCoord[Y] = d*d_vec.dCoord[1]+height.dCoord[1];
   tp[2].coeffs[2][1].dCoord[Z] = d*d_vec.dCoord[2]+height.dCoord[2];
   tp[2].coeffs[2][1].dCoord[W] = 1.0;

   /******* patch 3 *******/
   a=a*(-1.0);
   c=c*(-1.0);
   tp[3].coeffs[0][0].dCoord[X] = 0.0;
   tp[3].coeffs[0][0].dCoord[Y] = b;
   tp[3].coeffs[0][0].dCoord[Z] = 0.0;
   tp[3].coeffs[0][0].dCoord[W] = 1.0;

   tp[3].coeffs[0][1].dCoord[X] = d*d_vec.dCoord[0]+height.dCoord[0];
   tp[3].coeffs[0][1].dCoord[Y] = d*d_vec.dCoord[1]+height.dCoord[1];
   tp[3].coeffs[0][1].dCoord[Z] = d*d_vec.dCoord[2]+height.dCoord[2];
   tp[3].coeffs[0][1].dCoord[W] = 1.0;

   tp[3].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[3].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[3].coeffs[1][0].dCoord[Z] = 0.0;
   tp[3].coeffs[1][0].dCoord[W] = invsq2;

   tp[3].coeffs[1][1].dCoord[X] = (c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0])*invsq2;
   tp[3].coeffs[1][1].dCoord[Y] = (c*c_vec.dCoord[1]+d*d_vec.dCoord[1]
                                   +height.dCoord[1])*invsq2;
   tp[3].coeffs[1][1].dCoord[Z] = (c*c_vec.dCoord[2]+d*d_vec.dCoord[2]
                                   +height.dCoord[2])*invsq2;
   tp[3].coeffs[1][1].dCoord[W] = invsq2;

   tp[3].coeffs[2][0].dCoord[X] = a;
   tp[3].coeffs[2][0].dCoord[Y] = 0.0;
   tp[3].coeffs[2][0].dCoord[Z] = 0.0;
   tp[3].coeffs[2][0].dCoord[W] = 1.0;

   tp[3].coeffs[2][1].dCoord[X] = c*c_vec.dCoord[0]+height.dCoord[0];
   tp[3].coeffs[2][1].dCoord[Y] = c*c_vec.dCoord[1]+height.dCoord[1];
   tp[3].coeffs[2][1].dCoord[Z] = c*c_vec.dCoord[2]+height.dCoord[2];
   tp[3].coeffs[2][1].dCoord[W] = 1.0;

   /******* FINISHED WITH SIDES, NOW DEFINE TOP AND BOTTOM *******/

   a = a * TGC_LEEWAY_FACTOR;
   b = b * TGC_LEEWAY_FACTOR;
   c = c * TGC_LEEWAY_FACTOR;
   d = d * TGC_LEEWAY_FACTOR;
   /* reset a,b,c,d to orig. values */
   b = b*(-1.0);
   d = d*(-1.0);

   /******* patch 4 (bottom) *******/
   tp[4].coeffs[0][0].dCoord[X] = a*(-1.0);
   tp[4].coeffs[0][0].dCoord[Y] = b*(-1.0);
   tp[4].coeffs[0][0].dCoord[Z] = 0.0;
   tp[4].coeffs[0][0].dCoord[W] = 1.0;

   tp[4].coeffs[1][0].dCoord[X] = a*(-1.0);
   tp[4].coeffs[1][0].dCoord[Y] = b;
   tp[4].coeffs[1][0].dCoord[Z] = 0.0;
   tp[4].coeffs[1][0].dCoord[W] = 1.0;

   tp[4].coeffs[0][1].dCoord[X] = a;
   tp[4].coeffs[0][1].dCoord[Y] = b*(-1.0);
   tp[4].coeffs[0][1].dCoord[Z] = 0.0;
   tp[4].coeffs[0][1].dCoord[W] = 1.0;

   tp[4].coeffs[1][1].dCoord[X] = a;
   tp[4].coeffs[1][1].dCoord[Y] = b;
   tp[4].coeffs[1][1].dCoord[Z] = 0.0;
   tp[4].coeffs[1][1].dCoord[W] = 1.0;

   /******* patch 5 (top) *******/
   tp[5].coeffs[0][0].dCoord[X] = c*c_vec.dCoord[0]-d*d_vec.dCoord[0]
                                   +height.dCoord[0];
   tp[5].coeffs[0][0].dCoord[Y] = c*c_vec.dCoord[1]-d*d_vec.dCoord[1]
                                   +height.dCoord[1];
   tp[5].coeffs[0][0].dCoord[Z] = c*c_vec.dCoord[2]-d*d_vec.dCoord[2]
                                   +height.dCoord[2];
   tp[5].coeffs[0][0].dCoord[W] = 1.0;

   tp[5].coeffs[1][0].dCoord[X] = c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0];
   tp[5].coeffs[1][0].dCoord[Y] = c*c_vec.dCoord[1]+d*d_vec.dCoord[1]
                                   +height.dCoord[1];
   tp[5].coeffs[1][0].dCoord[Z] = c*c_vec.dCoord[2]+d*d_vec.dCoord[2]
                                   +height.dCoord[2];
   tp[5].coeffs[1][0].dCoord[W] = 1.0;

   tp[5].coeffs[0][1].dCoord[X] = (-1.0)*c*c_vec.dCoord[0]-d*d_vec.dCoord[0]
                                   +height.dCoord[0];
   tp[5].coeffs[0][1].dCoord[Y] = (-1.0)*c*c_vec.dCoord[1]-d*d_vec.dCoord[1]
                                   +height.dCoord[1];
   tp[5].coeffs[0][1].dCoord[Z] = (-1.0)*c*c_vec.dCoord[2]-d*d_vec.dCoord[2]
                                   +height.dCoord[2];
   tp[5].coeffs[0][1].dCoord[W] = 1.0;

   tp[5].coeffs[1][1].dCoord[X] = (-1.0)*c*c_vec.dCoord[0]+d*d_vec.dCoord[0]
                                   +height.dCoord[0];
   tp[5].coeffs[1][1].dCoord[Y] = (-1.0)*c*c_vec.dCoord[1]+d*d_vec.dCoord[1]   
                                   +height.dCoord[1];
   tp[5].coeffs[1][1].dCoord[Z] = (-1.0)*c*c_vec.dCoord[2]+d*d_vec.dCoord[2]   
                                   +height.dCoord[2];
   tp[5].coeffs[1][1].dCoord[W] = 1.0;

}


/******************** gen_brltgc ***************************/
/* The "main" routine to create a truncated general cone.  */
/***********************************************************/
int gen_brltgc(center,height,a_vec,b_vec,c_vec,d_vec,a,b,c,d,tp)
POINT4D center, height, a_vec, b_vec, c_vec, d_vec;
double a, b, c, d;
TENSOR_PATCH **tp;
{
   int i;
   double **XformMatTgc;
   double temp1;
   double temp2;
   double temp3;
   double dotprod;
   double dotprod1;
   double dotprod2;
   double rotamt;
   POINT4D perp_vec;

   ALLOCMAT(XformMatTgc, 4, 4, double);
/* compute perp_vec = a_vec X b_vec */
   perp_vec.dCoord[0] = a_vec.dCoord[1]*b_vec.dCoord[2] -
                        a_vec.dCoord[2]*b_vec.dCoord[1]; 
   perp_vec.dCoord[1] = a_vec.dCoord[2]*b_vec.dCoord[0] -
                        a_vec.dCoord[0]*b_vec.dCoord[2]; 
   perp_vec.dCoord[2] = a_vec.dCoord[0]*b_vec.dCoord[1] -
                        a_vec.dCoord[1]*b_vec.dCoord[0]; 

/*   getXformMatTgc(center,a_vec,b_vec,perp_vec,XformMatTgc);*/

/* decide whether or not to rotate cylinder randomly, then get XformMatTgc */
   dotprod1 = a_vec.dCoord[0]*b_vec.dCoord[0] +
              a_vec.dCoord[1]*b_vec.dCoord[1] +
              a_vec.dCoord[2]*b_vec.dCoord[2];
   dotprod2 = c_vec.dCoord[0]*d_vec.dCoord[0] +
              c_vec.dCoord[1]*d_vec.dCoord[1] +
              c_vec.dCoord[2]*d_vec.dCoord[2];

   if ((FP_EQ_EPS(a,b,(a*0.001))) && (FP_EQ_EPS(c,d,(c*0.001))) &&
       (FP_EQ_EPS(dotprod1,0.0,1e-4)) && (FP_EQ_EPS(dotprod2,0.0,1e-4)) &&
       (EQ_4D_EPS(a_vec,c_vec,1e-4)) && (EQ_4D_EPS(b_vec,d_vec,1e-4))) {
      dotprod = a_vec.dCoord[0]*height.dCoord[0] + 
                a_vec.dCoord[1]*height.dCoord[1] +
                a_vec.dCoord[2]*height.dCoord[2];
      if (FP_EQ_EPS(dotprod,0.0,1e-4)) {
         /* we should rotate the cylinder to help remove degenerate cases */
         if (FP_EQ_EPS(a,c,(a*0.001))) {
            /* this is a nice-shaped cylinder - use Shankar's code */
            return 0;
         } else {
/*            rotamt = (3.1415927/2.0) * ((double)rand())/((double)RAND_MAX);
            getXformMatTgc2(center,a_vec,b_vec,perp_vec,rotamt,XformMatTgc);
*/         getXformMatTgc(center,a_vec,b_vec,perp_vec,XformMatTgc);

         }
      } else {
         getXformMatTgc(center,a_vec,b_vec,perp_vec,XformMatTgc);
      }
   } else {
      getXformMatTgc(center,a_vec,b_vec,perp_vec,XformMatTgc);
   }

/* transform height, c, and d vectors to account for XformMatTgc */
   temp1 = height.dCoord[0]*a_vec.dCoord[0] +
           height.dCoord[1]*a_vec.dCoord[1] +
           height.dCoord[2]*a_vec.dCoord[2];
   temp2 = height.dCoord[0]*b_vec.dCoord[0] +
           height.dCoord[1]*b_vec.dCoord[1] +
           height.dCoord[2]*b_vec.dCoord[2];
   temp3 = height.dCoord[0]*perp_vec.dCoord[0] +
           height.dCoord[1]*perp_vec.dCoord[1] +
           height.dCoord[2]*perp_vec.dCoord[2];
   height.dCoord[0]=temp1;
   height.dCoord[1]=temp2;
   height.dCoord[2]=temp3;
   temp1 = c_vec.dCoord[0]*a_vec.dCoord[0] +
           c_vec.dCoord[1]*a_vec.dCoord[1] +
           c_vec.dCoord[2]*a_vec.dCoord[2];
   temp2 = c_vec.dCoord[0]*b_vec.dCoord[0] +
           c_vec.dCoord[1]*b_vec.dCoord[1] +
           c_vec.dCoord[2]*b_vec.dCoord[2];
   temp3 = c_vec.dCoord[0]*perp_vec.dCoord[0] +
           c_vec.dCoord[1]*perp_vec.dCoord[1] +
           c_vec.dCoord[2]*perp_vec.dCoord[2];
   c_vec.dCoord[0]=temp1;
   c_vec.dCoord[1]=temp2;
   c_vec.dCoord[2]=temp3;
   temp1 = d_vec.dCoord[0]*a_vec.dCoord[0] +
           d_vec.dCoord[1]*a_vec.dCoord[1] +
           d_vec.dCoord[2]*a_vec.dCoord[2];
   temp2 = d_vec.dCoord[0]*b_vec.dCoord[0] +
           d_vec.dCoord[1]*b_vec.dCoord[1] +
           d_vec.dCoord[2]*b_vec.dCoord[2];
   temp3 = d_vec.dCoord[0]*perp_vec.dCoord[0] +
           d_vec.dCoord[1]*perp_vec.dCoord[1] +
           d_vec.dCoord[2]*perp_vec.dCoord[2];
   d_vec.dCoord[0]=temp1;
   d_vec.dCoord[1]=temp2;
   d_vec.dCoord[2]=temp3;

   
   ALLOCN(*tp, TENSOR_PATCH, NUM_TGC_PATCHES);
   for(i=0;i<NUM_TGC_PATCHES-2;i++) {
      (*tp)[i].iDeg0 = TGC_SDEGREE;
      (*tp)[i].iDeg1 = TGC_TDEGREE;
      allocate_tensor_patch(&((*tp)[i]));
   }
   for(i=NUM_TGC_PATCHES-2;i<NUM_TGC_PATCHES;i++) {
      (*tp)[i].iDeg1 = 1;
      (*tp)[i].iDeg0 = 1;
      allocate_tensor_patch(&((*tp)[i]));
   }

   GetTgcControlPoints(*tp,height,a,b,c,d,c_vec,d_vec);

   /* error here - shouldn't transform top and bottom in same way as sides! */
   XformPatch(*tp, NUM_TGC_PATCHES, TGC_SDEGREE, TGC_TDEGREE, XformMatTgc);
   return NUM_TGC_PATCHES;
}
