#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "linalg.h"

#ifdef NEW_CONE_REPRN
#define	NUM_CONE_PATCHES	5
#define CONE_LEEWAY_FACTOR      1.1
#else
#define	NUM_CONE_PATCHES	12
#endif

#define	CONE_SDEGREE		2	
#define	CONE_TDEGREE		1	

EXTERN_LINALG

extern void normalise();
extern void allocate_tensor_patch();
extern void XformPatch();
extern void WriteControlPtsInFile();

void getXformMatCone(center, axis_vec, Xform)
POINT4D center, axis_vec;
double **Xform;
{
  double a, b, c, d, px, py, pz;

  a = axis_vec.dCoord[X];
  b = axis_vec.dCoord[Y];
  c = axis_vec.dCoord[Z];
  d = sqrt(b*b + c*c);
  px = center.dCoord[X];
  py = center.dCoord[Y];
  pz = center.dCoord[Z];

  if (d < 1e-5) {
    Xform[0][0] = 0.0;
    Xform[0][1] = 0.0;
    Xform[0][2] = -1.0;
    Xform[0][3] = 0.0;
 
    Xform[1][0] = 0.0;
    Xform[1][1] = 1.0;
    Xform[1][2] = 0.0;
    Xform[1][3] = 0.0;
 
    Xform[2][0] = 1.0;
    Xform[2][1] = 0.0;
    Xform[2][2] = 0.0;
    Xform[2][3] = 0.0;

    Xform[3][0] = px;
    Xform[3][1] = py;
    Xform[3][2] = pz;
    Xform[3][3] = 1.0;
    return;
  }

  Xform[0][0] = d;
  Xform[0][1] = -a*b/d;
  Xform[0][2] = -a*c/d;
  Xform[0][3] = 0.0;
  
  Xform[1][0] = 0.0;
  Xform[1][1] = c/d;
  Xform[1][2] = -b/d;
  Xform[1][3] = 0.0;
  
  Xform[2][0] = a;
  Xform[2][1] = b;
  Xform[2][2] = c;
  Xform[2][3] = 0.0;
  
  Xform[3][0] = px;
  Xform[3][1] = py;
  Xform[3][2] = pz;
  Xform[3][3] = 1.0;
}


#ifdef NEW_CONE_REPRN		/* new representation!!!! */
void GetNormConeControlPts(tp, cone_length, cone_Lradius, cone_Uradius)
TENSOR_PATCH *tp;
double cone_length, cone_Lradius, cone_Uradius;
{
  double w, w1, rl, ru, l;

  w = sqrt(2.0) / 2.0;  
  w1 = sqrt(3.0) / 2.0;  
  l = cone_length;
  rl = cone_Lradius;
  ru = cone_Uradius;

  tp[0].coeffs[0][0].dCoord[X] = w1*rl;
  tp[0].coeffs[0][0].dCoord[Y] = 0.5*rl;
  tp[0].coeffs[0][0].dCoord[Z] = 0.0;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = w1*ru;
  tp[0].coeffs[1][0].dCoord[Y] = 0.5*ru;
  tp[0].coeffs[1][0].dCoord[Z] = l;
  tp[0].coeffs[1][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = w1*rl;
  tp[0].coeffs[0][1].dCoord[Y] = -0.5*rl;
  tp[0].coeffs[0][1].dCoord[Z] = 0.0;
  tp[0].coeffs[0][1].dCoord[W] = 0.5;

  tp[0].coeffs[1][1].dCoord[X] = w1*ru;
  tp[0].coeffs[1][1].dCoord[Y] = -0.5*ru;
  tp[0].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[0].coeffs[1][1].dCoord[W] = 0.5;

  tp[0].coeffs[0][2].dCoord[X] = 0.0;
  tp[0].coeffs[0][2].dCoord[Y] = -rl;
  tp[0].coeffs[0][2].dCoord[Z] = 0.0;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = 0.0;
  tp[0].coeffs[1][2].dCoord[Y] = -ru;
  tp[0].coeffs[1][2].dCoord[Z] = l;
  tp[0].coeffs[1][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = 0.0;
  tp[1].coeffs[0][0].dCoord[Y] = -rl;
  tp[1].coeffs[0][0].dCoord[Z] = 0.0;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = 0.0;
  tp[1].coeffs[1][0].dCoord[Y] = -ru;
  tp[1].coeffs[1][0].dCoord[Z] = l;
  tp[1].coeffs[1][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = -w1*rl;
  tp[1].coeffs[0][1].dCoord[Y] = -0.5*rl;
  tp[1].coeffs[0][1].dCoord[Z] = 0.0;
  tp[1].coeffs[0][1].dCoord[W] = 0.5;

  tp[1].coeffs[1][1].dCoord[X] = -w1*ru;
  tp[1].coeffs[1][1].dCoord[Y] = -0.5*ru;
  tp[1].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[1].coeffs[1][1].dCoord[W] = 0.5;

  tp[1].coeffs[0][2].dCoord[X] = -w1*rl;
  tp[1].coeffs[0][2].dCoord[Y] = 0.5*rl;
  tp[1].coeffs[0][2].dCoord[Z] = 0.0;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = -w1*ru;
  tp[1].coeffs[1][2].dCoord[Y] = 0.5*ru;
  tp[1].coeffs[1][2].dCoord[Z] = l;
  tp[1].coeffs[1][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = -w1*rl;
  tp[2].coeffs[0][0].dCoord[Y] = 0.5*rl;
  tp[2].coeffs[0][0].dCoord[Z] = 0.0;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = -w1*ru;
  tp[2].coeffs[1][0].dCoord[Y] = 0.5*ru;
  tp[2].coeffs[1][0].dCoord[Z] = l;
  tp[2].coeffs[1][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = 0.0;
  tp[2].coeffs[0][1].dCoord[Y] = rl;
  tp[2].coeffs[0][1].dCoord[Z] = 0.0;
  tp[2].coeffs[0][1].dCoord[W] = 0.5;

  tp[2].coeffs[1][1].dCoord[X] = 0.0;
  tp[2].coeffs[1][1].dCoord[Y] = ru;
  tp[2].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[2].coeffs[1][1].dCoord[W] = 0.5;

  tp[2].coeffs[0][2].dCoord[X] = w1*rl;
  tp[2].coeffs[0][2].dCoord[Y] = 0.5*rl;
  tp[2].coeffs[0][2].dCoord[Z] = 0.0;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = w1*ru;
  tp[2].coeffs[1][2].dCoord[Y] = 0.5*ru;
  tp[2].coeffs[1][2].dCoord[Z] = l;
  tp[2].coeffs[1][2].dCoord[W] = 1.0;

  /* generated the sides!!! Now the lids */

  ru *= CONE_LEEWAY_FACTOR;
  rl *= CONE_LEEWAY_FACTOR;


  tp[3].coeffs[0][0].dCoord[X] = rl;
  tp[3].coeffs[0][0].dCoord[Y] = -rl;
  tp[3].coeffs[0][0].dCoord[Z] = 0.0;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = rl;
  tp[3].coeffs[0][1].dCoord[Y] = rl;
  tp[3].coeffs[0][1].dCoord[Z] = 0.0;
  tp[3].coeffs[0][1].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = -rl;
  tp[3].coeffs[1][0].dCoord[Y] = -rl;
  tp[3].coeffs[1][0].dCoord[Z] = 0.0;
  tp[3].coeffs[1][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][1].dCoord[X] = -rl;
  tp[3].coeffs[1][1].dCoord[Y] = rl;
  tp[3].coeffs[1][1].dCoord[Z] = 0.0;
  tp[3].coeffs[1][1].dCoord[W] = 1.0;

  tp[4].coeffs[0][0].dCoord[X] = -ru;
  tp[4].coeffs[0][0].dCoord[Y] = -ru;
  tp[4].coeffs[0][0].dCoord[Z] = l;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = -ru;
  tp[4].coeffs[0][1].dCoord[Y] = ru;
  tp[4].coeffs[0][1].dCoord[Z] = l;
  tp[4].coeffs[0][1].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = ru;
  tp[4].coeffs[1][0].dCoord[Y] = -ru;
  tp[4].coeffs[1][0].dCoord[Z] = l;
  tp[4].coeffs[1][0].dCoord[W] = 1.0;

  tp[4].coeffs[1][1].dCoord[X] = ru;
  tp[4].coeffs[1][1].dCoord[Y] = ru;
  tp[4].coeffs[1][1].dCoord[Z] = l;
  tp[4].coeffs[1][1].dCoord[W] = 1.0;
}

#else  /* older reprn */
void GetNormConeControlPts(tp, cone_length, cone_Lradius, cone_Uradius)
TENSOR_PATCH *tp;
double cone_length, cone_Lradius, cone_Uradius;
{
  double w, rl, ru, l;

  w = sqrt(2.0) / 2.0;  
  l = cone_length;
  rl = cone_Lradius;
  ru = cone_Uradius;

  tp[0].coeffs[0][0].dCoord[X] = 0.0;
  tp[0].coeffs[0][0].dCoord[Y] = rl;
  tp[0].coeffs[0][0].dCoord[Z] = 0.0;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = 0.0;
  tp[0].coeffs[1][0].dCoord[Y] = ru;
  tp[0].coeffs[1][0].dCoord[Z] = l;
  tp[0].coeffs[1][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = rl*w;
  tp[0].coeffs[0][1].dCoord[Y] = rl*w;
  tp[0].coeffs[0][1].dCoord[Z] = 0.0;
  tp[0].coeffs[0][1].dCoord[W] = w;

  tp[0].coeffs[1][1].dCoord[X] = ru*w;
  tp[0].coeffs[1][1].dCoord[Y] = ru*w;
  tp[0].coeffs[1][1].dCoord[Z] = l*w;
  tp[0].coeffs[1][1].dCoord[W] = w;

  tp[0].coeffs[0][2].dCoord[X] = rl;
  tp[0].coeffs[0][2].dCoord[Y] = 0.0;
  tp[0].coeffs[0][2].dCoord[Z] = 0.0;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = ru;
  tp[0].coeffs[1][2].dCoord[Y] = 0.0;
  tp[0].coeffs[1][2].dCoord[Z] = l;
  tp[0].coeffs[1][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = rl;
  tp[1].coeffs[0][0].dCoord[Y] = 0.0;
  tp[1].coeffs[0][0].dCoord[Z] = 0.0;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = ru;
  tp[1].coeffs[1][0].dCoord[Y] = 0.0;
  tp[1].coeffs[1][0].dCoord[Z] = l;
  tp[1].coeffs[1][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = rl*w;
  tp[1].coeffs[0][1].dCoord[Y] = -rl*w;
  tp[1].coeffs[0][1].dCoord[Z] = 0.0;
  tp[1].coeffs[0][1].dCoord[W] = w;

  tp[1].coeffs[1][1].dCoord[X] = ru*w;
  tp[1].coeffs[1][1].dCoord[Y] = -ru*w;
  tp[1].coeffs[1][1].dCoord[Z] = l*w;
  tp[1].coeffs[1][1].dCoord[W] = w;

  tp[1].coeffs[0][2].dCoord[X] = 0.0;
  tp[1].coeffs[0][2].dCoord[Y] = -rl;
  tp[1].coeffs[0][2].dCoord[Z] = 0.0;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = 0.0;
  tp[1].coeffs[1][2].dCoord[Y] = -ru;
  tp[1].coeffs[1][2].dCoord[Z] = l;
  tp[1].coeffs[1][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = 0.0;
  tp[2].coeffs[0][0].dCoord[Y] = -rl;
  tp[2].coeffs[0][0].dCoord[Z] = 0.0;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = 0.0;
  tp[2].coeffs[1][0].dCoord[Y] = -ru;
  tp[2].coeffs[1][0].dCoord[Z] = l;
  tp[2].coeffs[1][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = -rl*w;
  tp[2].coeffs[0][1].dCoord[Y] = -rl*w;
  tp[2].coeffs[0][1].dCoord[Z] = 0.0;
  tp[2].coeffs[0][1].dCoord[W] = w;

  tp[2].coeffs[1][1].dCoord[X] = -ru*w;
  tp[2].coeffs[1][1].dCoord[Y] = -ru*w;
  tp[2].coeffs[1][1].dCoord[Z] = l*w;
  tp[2].coeffs[1][1].dCoord[W] = w;

  tp[2].coeffs[0][2].dCoord[X] = -rl;
  tp[2].coeffs[0][2].dCoord[Y] = 0.0;
  tp[2].coeffs[0][2].dCoord[Z] = 0.0;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = -ru;
  tp[2].coeffs[1][2].dCoord[Y] = 0.0;
  tp[2].coeffs[1][2].dCoord[Z] = l;
  tp[2].coeffs[1][2].dCoord[W] = 1.0;

  tp[3].coeffs[0][0].dCoord[X] = -rl;
  tp[3].coeffs[0][0].dCoord[Y] = 0.0;
  tp[3].coeffs[0][0].dCoord[Z] = 0.0;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = -ru;
  tp[3].coeffs[1][0].dCoord[Y] = 0.0;
  tp[3].coeffs[1][0].dCoord[Z] = l;
  tp[3].coeffs[1][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = -rl*w;
  tp[3].coeffs[0][1].dCoord[Y] = rl*w;
  tp[3].coeffs[0][1].dCoord[Z] = 0.0;
  tp[3].coeffs[0][1].dCoord[W] = w;

  tp[3].coeffs[1][1].dCoord[X] = -ru*w;
  tp[3].coeffs[1][1].dCoord[Y] = ru*w;
  tp[3].coeffs[1][1].dCoord[Z] = l*w;
  tp[3].coeffs[1][1].dCoord[W] = w;

  tp[3].coeffs[0][2].dCoord[X] = 0.0;
  tp[3].coeffs[0][2].dCoord[Y] = rl;
  tp[3].coeffs[0][2].dCoord[Z] = 0.0;
  tp[3].coeffs[0][2].dCoord[W] = 1.0;

  tp[3].coeffs[1][2].dCoord[X] = 0.0;
  tp[3].coeffs[1][2].dCoord[Y] = ru;
  tp[3].coeffs[1][2].dCoord[Z] = l;
  tp[3].coeffs[1][2].dCoord[W] = 1.0;

  tp[4].coeffs[0][0].dCoord[X] = 0.0;
  tp[4].coeffs[0][0].dCoord[Y] = rl;
  tp[4].coeffs[0][0].dCoord[Z] = 0.0;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = 0.0;
  tp[4].coeffs[0][1].dCoord[Y] = 0.0;
  tp[4].coeffs[0][1].dCoord[Z] = 0.0;
  tp[4].coeffs[0][1].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = rl*w;
  tp[4].coeffs[1][0].dCoord[Y] = rl*w;
  tp[4].coeffs[1][0].dCoord[Z] = 0.0;
  tp[4].coeffs[1][0].dCoord[W] = w;

  tp[4].coeffs[1][1].dCoord[X] = 0.0;
  tp[4].coeffs[1][1].dCoord[Y] = 0.0;
  tp[4].coeffs[1][1].dCoord[Z] = 0.0;
  tp[4].coeffs[1][1].dCoord[W] = w;

  tp[4].coeffs[2][0].dCoord[X] = rl;
  tp[4].coeffs[2][0].dCoord[Y] = 0.0;
  tp[4].coeffs[2][0].dCoord[Z] = 0.0;
  tp[4].coeffs[2][0].dCoord[W] = 1.0;

  tp[4].coeffs[2][1].dCoord[X] = 0.0;
  tp[4].coeffs[2][1].dCoord[Y] = 0.0;
  tp[4].coeffs[2][1].dCoord[Z] = 0.0;
  tp[4].coeffs[2][1].dCoord[W] = 1.0;

  tp[5].coeffs[0][0].dCoord[X] = rl;
  tp[5].coeffs[0][0].dCoord[Y] = 0.0;
  tp[5].coeffs[0][0].dCoord[Z] = 0.0;
  tp[5].coeffs[0][0].dCoord[W] = 1.0;

  tp[5].coeffs[0][1].dCoord[X] = 0.0;
  tp[5].coeffs[0][1].dCoord[Y] = 0.0;
  tp[5].coeffs[0][1].dCoord[Z] = 0.0;
  tp[5].coeffs[0][1].dCoord[W] = 1.0;

  tp[5].coeffs[1][0].dCoord[X] = rl*w;
  tp[5].coeffs[1][0].dCoord[Y] = -rl*w;
  tp[5].coeffs[1][0].dCoord[Z] = 0.0;
  tp[5].coeffs[1][0].dCoord[W] = w;

  tp[5].coeffs[1][1].dCoord[X] = 0.0;
  tp[5].coeffs[1][1].dCoord[Y] = 0.0;
  tp[5].coeffs[1][1].dCoord[Z] = 0.0;
  tp[5].coeffs[1][1].dCoord[W] = w;

  tp[5].coeffs[2][0].dCoord[X] = 0.0;
  tp[5].coeffs[2][0].dCoord[Y] = -rl;
  tp[5].coeffs[2][0].dCoord[Z] = 0.0;
  tp[5].coeffs[2][0].dCoord[W] = 1.0;

  tp[5].coeffs[2][1].dCoord[X] = 0.0;
  tp[5].coeffs[2][1].dCoord[Y] = 0.0;
  tp[5].coeffs[2][1].dCoord[Z] = 0.0;
  tp[5].coeffs[2][1].dCoord[W] = 1.0;

  tp[6].coeffs[0][0].dCoord[X] = 0.0;
  tp[6].coeffs[0][0].dCoord[Y] = -rl;
  tp[6].coeffs[0][0].dCoord[Z] = 0.0;
  tp[6].coeffs[0][0].dCoord[W] = 1.0;

  tp[6].coeffs[0][1].dCoord[X] = 0.0;
  tp[6].coeffs[0][1].dCoord[Y] = 0.0;
  tp[6].coeffs[0][1].dCoord[Z] = 0.0;
  tp[6].coeffs[0][1].dCoord[W] = 1.0;

  tp[6].coeffs[1][0].dCoord[X] = -rl*w;
  tp[6].coeffs[1][0].dCoord[Y] = -rl*w;
  tp[6].coeffs[1][0].dCoord[Z] = 0.0;
  tp[6].coeffs[1][0].dCoord[W] = w;

  tp[6].coeffs[1][1].dCoord[X] = 0.0;
  tp[6].coeffs[1][1].dCoord[Y] = 0.0;
  tp[6].coeffs[1][1].dCoord[Z] = 0.0;
  tp[6].coeffs[1][1].dCoord[W] = w;

  tp[6].coeffs[2][0].dCoord[X] = -rl;
  tp[6].coeffs[2][0].dCoord[Y] = 0.0;
  tp[6].coeffs[2][0].dCoord[Z] = 0.0;
  tp[6].coeffs[2][0].dCoord[W] = 1.0;

  tp[6].coeffs[2][1].dCoord[X] = 0.0;
  tp[6].coeffs[2][1].dCoord[Y] = 0.0;
  tp[6].coeffs[2][1].dCoord[Z] = 0.0;
  tp[6].coeffs[2][1].dCoord[W] = 1.0;

  tp[7].coeffs[0][0].dCoord[X] = -rl;
  tp[7].coeffs[0][0].dCoord[Y] = 0.0;
  tp[7].coeffs[0][0].dCoord[Z] = 0.0;
  tp[7].coeffs[0][0].dCoord[W] = 1.0;

  tp[7].coeffs[0][1].dCoord[X] = 0.0;
  tp[7].coeffs[0][1].dCoord[Y] = 0.0;
  tp[7].coeffs[0][1].dCoord[Z] = 0.0;
  tp[7].coeffs[0][1].dCoord[W] = 1.0;

  tp[7].coeffs[1][0].dCoord[X] = -rl*w;
  tp[7].coeffs[1][0].dCoord[Y] = rl*w;
  tp[7].coeffs[1][0].dCoord[Z] = 0.0;
  tp[7].coeffs[1][0].dCoord[W] = w;

  tp[7].coeffs[1][1].dCoord[X] = 0.0;
  tp[7].coeffs[1][1].dCoord[Y] = 0.0;
  tp[7].coeffs[1][1].dCoord[Z] = 0.0;
  tp[7].coeffs[1][1].dCoord[W] = w;

  tp[7].coeffs[2][0].dCoord[X] = 0.0;
  tp[7].coeffs[2][0].dCoord[Y] = rl;
  tp[7].coeffs[2][0].dCoord[Z] = 0.0;
  tp[7].coeffs[2][0].dCoord[W] = 1.0;

  tp[7].coeffs[2][1].dCoord[X] = 0.0;
  tp[7].coeffs[2][1].dCoord[Y] = 0.0;
  tp[7].coeffs[2][1].dCoord[Z] = 0.0;
  tp[7].coeffs[2][1].dCoord[W] = 1.0;

  tp[8].coeffs[0][0].dCoord[X] = 0.0;
  tp[8].coeffs[0][0].dCoord[Y] = ru;
  tp[8].coeffs[0][0].dCoord[Z] = l;
  tp[8].coeffs[0][0].dCoord[W] = 1.0;

  tp[8].coeffs[1][0].dCoord[X] = 0.0;
  tp[8].coeffs[1][0].dCoord[Y] = 0.0;
  tp[8].coeffs[1][0].dCoord[Z] = l;
  tp[8].coeffs[1][0].dCoord[W] = 1.0;

  tp[8].coeffs[0][1].dCoord[X] = ru*w;
  tp[8].coeffs[0][1].dCoord[Y] = ru*w;
  tp[8].coeffs[0][1].dCoord[Z] = l*w;
  tp[8].coeffs[0][1].dCoord[W] = w;

  tp[8].coeffs[1][1].dCoord[X] = 0.0;
  tp[8].coeffs[1][1].dCoord[Y] = 0.0;
  tp[8].coeffs[1][1].dCoord[Z] = l*w;
  tp[8].coeffs[1][1].dCoord[W] = w;

  tp[8].coeffs[0][2].dCoord[X] = ru;
  tp[8].coeffs[0][2].dCoord[Y] = 0.0;
  tp[8].coeffs[0][2].dCoord[Z] = l;
  tp[8].coeffs[0][2].dCoord[W] = 1.0;

  tp[8].coeffs[1][2].dCoord[X] = 0.0;
  tp[8].coeffs[1][2].dCoord[Y] = 0.0;
  tp[8].coeffs[1][2].dCoord[Z] = l;
  tp[8].coeffs[1][2].dCoord[W] = 1.0;

  tp[9].coeffs[0][0].dCoord[X] = ru;
  tp[9].coeffs[0][0].dCoord[Y] = 0.0;
  tp[9].coeffs[0][0].dCoord[Z] = l;
  tp[9].coeffs[0][0].dCoord[W] = 1.0;

  tp[9].coeffs[1][0].dCoord[X] = 0.0;
  tp[9].coeffs[1][0].dCoord[Y] = 0.0;
  tp[9].coeffs[1][0].dCoord[Z] = l;
  tp[9].coeffs[1][0].dCoord[W] = 1.0;

  tp[9].coeffs[0][1].dCoord[X] = ru*w;
  tp[9].coeffs[0][1].dCoord[Y] = -ru*w;
  tp[9].coeffs[0][1].dCoord[Z] = l*w;
  tp[9].coeffs[0][1].dCoord[W] = w;

  tp[9].coeffs[1][1].dCoord[X] = 0.0;
  tp[9].coeffs[1][1].dCoord[Y] = 0.0;
  tp[9].coeffs[1][1].dCoord[Z] = l*w;
  tp[9].coeffs[1][1].dCoord[W] = w;

  tp[9].coeffs[0][2].dCoord[X] = 0.0;
  tp[9].coeffs[0][2].dCoord[Y] = -ru;
  tp[9].coeffs[0][2].dCoord[Z] = l;
  tp[9].coeffs[0][2].dCoord[W] = 1.0;

  tp[9].coeffs[1][2].dCoord[X] = 0.0;
  tp[9].coeffs[1][2].dCoord[Y] = 0.0;
  tp[9].coeffs[1][2].dCoord[Z] = l;
  tp[9].coeffs[1][2].dCoord[W] = 1.0;

  tp[10].coeffs[0][0].dCoord[X] = 0.0;
  tp[10].coeffs[0][0].dCoord[Y] = -ru;
  tp[10].coeffs[0][0].dCoord[Z] = l;
  tp[10].coeffs[0][0].dCoord[W] = 1.0;

  tp[10].coeffs[1][0].dCoord[X] = 0.0;
  tp[10].coeffs[1][0].dCoord[Y] = 0.0;
  tp[10].coeffs[1][0].dCoord[Z] = l;
  tp[10].coeffs[1][0].dCoord[W] = 1.0;

  tp[10].coeffs[0][1].dCoord[X] = -ru*w;
  tp[10].coeffs[0][1].dCoord[Y] = -ru*w;
  tp[10].coeffs[0][1].dCoord[Z] = l*w;
  tp[10].coeffs[0][1].dCoord[W] = w;

  tp[10].coeffs[1][1].dCoord[X] = 0.0;
  tp[10].coeffs[1][1].dCoord[Y] = 0.0;
  tp[10].coeffs[1][1].dCoord[Z] = l*w;
  tp[10].coeffs[1][1].dCoord[W] = w;

  tp[10].coeffs[0][2].dCoord[X] = -ru;
  tp[10].coeffs[0][2].dCoord[Y] = 0.0;
  tp[10].coeffs[0][2].dCoord[Z] = l;
  tp[10].coeffs[0][2].dCoord[W] = 1.0;

  tp[10].coeffs[1][2].dCoord[X] = 0.0;
  tp[10].coeffs[1][2].dCoord[Y] = 0.0;
  tp[10].coeffs[1][2].dCoord[Z] = l;
  tp[10].coeffs[1][2].dCoord[W] = 1.0;

  tp[11].coeffs[0][0].dCoord[X] = -ru;
  tp[11].coeffs[0][0].dCoord[Y] = 0.0;
  tp[11].coeffs[0][0].dCoord[Z] = l;
  tp[11].coeffs[0][0].dCoord[W] = 1.0;

  tp[11].coeffs[1][0].dCoord[X] = 0.0;
  tp[11].coeffs[1][0].dCoord[Y] = 0.0;
  tp[11].coeffs[1][0].dCoord[Z] = l;
  tp[11].coeffs[1][0].dCoord[W] = 1.0;

  tp[11].coeffs[0][1].dCoord[X] = -ru*w;
  tp[11].coeffs[0][1].dCoord[Y] = ru*w;
  tp[11].coeffs[0][1].dCoord[Z] = l*w;
  tp[11].coeffs[0][1].dCoord[W] = w;

  tp[11].coeffs[1][1].dCoord[X] = 0.0;
  tp[11].coeffs[1][1].dCoord[Y] = 0.0;
  tp[11].coeffs[1][1].dCoord[Z] = l*w;
  tp[11].coeffs[1][1].dCoord[W] = w;

  tp[11].coeffs[0][2].dCoord[X] = 0.0;
  tp[11].coeffs[0][2].dCoord[Y] = ru;
  tp[11].coeffs[0][2].dCoord[Z] = l;
  tp[11].coeffs[0][2].dCoord[W] = 1.0;

  tp[11].coeffs[1][2].dCoord[X] = 0.0;
  tp[11].coeffs[1][2].dCoord[Y] = 0.0;
  tp[11].coeffs[1][2].dCoord[Z] = l;
  tp[11].coeffs[1][2].dCoord[W] = 1.0;
}

#endif



int gen_cone(center, axis_vec, cone_length, cone_Lradius, cone_Uradius, tp)
POINT4D center, axis_vec;
double cone_length, cone_Lradius, cone_Uradius;
TENSOR_PATCH **tp;
{
  int i, j, k;
  double **XformMatCone;

  normalise(&axis_vec);
  ALLOCMAT(XformMatCone, 4, 4, double);
  getXformMatCone(center, axis_vec, XformMatCone);

  ALLOCN(*tp, TENSOR_PATCH, NUM_CONE_PATCHES);

#ifdef NEW_CONE_REPRN  /* which we DO want --culver */

  for (k=0; k<NUM_CONE_PATCHES; k++) {
    if (k > 2)
      (*tp)[k].iDeg0 = (*tp)[k].iDeg1 = 1;
    else 
      {
	(*tp)[k].iDeg1 = CONE_SDEGREE;
	(*tp)[k].iDeg0 = CONE_TDEGREE;
      }
    allocate_tensor_patch(&((*tp)[k]));
  }

#else
  
  for (k=0; k<NUM_CONE_PATCHES; k++) {
    if (k > 3 && k < 8) {
      (*tp)[k].iDeg0 = CONE_SDEGREE;
      (*tp)[k].iDeg1 = CONE_TDEGREE;
    } 
    else 
      {
	(*tp)[k].iDeg1 = CONE_SDEGREE;
	(*tp)[k].iDeg0 = CONE_TDEGREE;
      }
    allocate_tensor_patch(&((*tp)[k]));
  }
#endif


  GetNormConeControlPts(*tp, cone_length, cone_Lradius, cone_Uradius);

/* After getting normalised cylinder's control points ie. cylinder with
   center at origin and axis corresponding to the z axis, all we have
   to do is to transform each control point using XformMat and then 
   writing it into a file */

  XformPatch(*tp, NUM_CONE_PATCHES, CONE_SDEGREE, CONE_TDEGREE, XformMatCone);
  return NUM_CONE_PATCHES;
}

