#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "linalg.h"

#ifdef NEW_CYL_REPRN

/* Use a single trimmed patch to represent the tyop and the bottom of */
/* the cylinder... */

#define	NUM_CYL_PATCHES		5
#define	CYL_SDEGREE		2	
#define	CYL_TDEGREE		1	
#define CYL_LEEWAY_FACTOR       1.1

EXTERN_LINALG

extern void normalise();
extern void allocate_tensor_patch();
extern void XformPatch();

void getXformMatCyl(center, axis_vec, Xform)
POINT4D center, axis_vec;
double **Xform;
{
  double a, b, c, d, px, py, pz;

  a = axis_vec.dCoord[X];
  b = axis_vec.dCoord[Y];
  c = axis_vec.dCoord[Z];
  d = sqrt(b*b + c*c);
  px = center.dCoord[X];
  py = center.dCoord[Y];
  pz = center.dCoord[Z];

  if (d < 1e-5) {
    Xform[0][0] = 0.0;
    Xform[0][1] = 0.0;
    if (a>0) 
       Xform[0][2] = -1.0;
    else
       Xform[0][2] = 1.0;
    Xform[0][3] = 0.0;
  
    Xform[1][0] = 0.0;
    Xform[1][1] = 1.0;
    Xform[1][2] = 0.0;
    Xform[1][3] = 0.0;
 
    if (a>0) 
       Xform[2][0] = 1.0;
    else
       Xform[2][0] = -1.0;
    Xform[2][1] = 0.0;
    Xform[2][2] = 0.0;
    Xform[2][3] = 0.0;
  
    Xform[3][0] = px;
    Xform[3][1] = py;
    Xform[3][2] = pz;
    Xform[3][3] = 1.0;
    return;
  }
  Xform[0][0] = d;
  Xform[0][1] = -a*b/d;
  Xform[0][2] = -a*c/d;
  Xform[0][3] = 0.0;
  
  Xform[1][0] = 0.0;
  Xform[1][1] = c/d;
  Xform[1][2] = -b/d;
  Xform[1][3] = 0.0;
  
  Xform[2][0] = a;
  Xform[2][1] = b;
  Xform[2][2] = c;
  Xform[2][3] = 0.0;
  
  Xform[3][0] = px;
  Xform[3][1] = py;
  Xform[3][2] = pz;
  Xform[3][3] = 1.0;
}
  
void GetNormCylControlPts(tp, cyl_length, cyl_radius)
TENSOR_PATCH *tp;
double cyl_length, cyl_radius;
{
  double w, w1, l, r;

  w = sqrt(2.0) / 2.0;  
  w1 = sqrt(3.0) / 2.0;  
  l = cyl_length;
  r = cyl_radius;

  tp[0].coeffs[0][0].dCoord[X] = w1*r;
  tp[0].coeffs[0][0].dCoord[Y] = 0.5*r;
  tp[0].coeffs[0][0].dCoord[Z] = 0.0;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = w1*r;
  tp[0].coeffs[1][0].dCoord[Y] = 0.5*r;
  tp[0].coeffs[1][0].dCoord[Z] = l;
  tp[0].coeffs[1][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = w1*r;
  tp[0].coeffs[0][1].dCoord[Y] = -0.5*r;
  tp[0].coeffs[0][1].dCoord[Z] = 0.0;
  tp[0].coeffs[0][1].dCoord[W] = 0.5;

  tp[0].coeffs[1][1].dCoord[X] = w1*r;
  tp[0].coeffs[1][1].dCoord[Y] = -0.5*r;
  tp[0].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[0].coeffs[1][1].dCoord[W] = 0.5;

  tp[0].coeffs[0][2].dCoord[X] = 0.0;
  tp[0].coeffs[0][2].dCoord[Y] = -r;
  tp[0].coeffs[0][2].dCoord[Z] = 0.0;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = 0.0;
  tp[0].coeffs[1][2].dCoord[Y] = -r;
  tp[0].coeffs[1][2].dCoord[Z] = l;
  tp[0].coeffs[1][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = 0.0;
  tp[1].coeffs[0][0].dCoord[Y] = -r;
  tp[1].coeffs[0][0].dCoord[Z] = 0.0;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = 0.0;
  tp[1].coeffs[1][0].dCoord[Y] = -r;
  tp[1].coeffs[1][0].dCoord[Z] = l;
  tp[1].coeffs[1][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = -w1*r;
  tp[1].coeffs[0][1].dCoord[Y] = -0.5*r;
  tp[1].coeffs[0][1].dCoord[Z] = 0.0;
  tp[1].coeffs[0][1].dCoord[W] = 0.5;

  tp[1].coeffs[1][1].dCoord[X] = -w1*r;
  tp[1].coeffs[1][1].dCoord[Y] = -0.5*r;
  tp[1].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[1].coeffs[1][1].dCoord[W] = 0.5;

  tp[1].coeffs[0][2].dCoord[X] = -w1*r;
  tp[1].coeffs[0][2].dCoord[Y] = 0.5*r;
  tp[1].coeffs[0][2].dCoord[Z] = 0.0;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = -w1*r;
  tp[1].coeffs[1][2].dCoord[Y] = 0.5*r;
  tp[1].coeffs[1][2].dCoord[Z] = l;
  tp[1].coeffs[1][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = -w1*r;
  tp[2].coeffs[0][0].dCoord[Y] = 0.5*r;
  tp[2].coeffs[0][0].dCoord[Z] = 0.0;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = -w1*r;
  tp[2].coeffs[1][0].dCoord[Y] = 0.5*r;
  tp[2].coeffs[1][0].dCoord[Z] = l;
  tp[2].coeffs[1][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = 0.0;
  tp[2].coeffs[0][1].dCoord[Y] = r;
  tp[2].coeffs[0][1].dCoord[Z] = 0.0;
  tp[2].coeffs[0][1].dCoord[W] = 0.5;

  tp[2].coeffs[1][1].dCoord[X] = 0.0;
  tp[2].coeffs[1][1].dCoord[Y] = r;
  tp[2].coeffs[1][1].dCoord[Z] = 0.5*l;
  tp[2].coeffs[1][1].dCoord[W] = 0.5;

  tp[2].coeffs[0][2].dCoord[X] = w1*r;
  tp[2].coeffs[0][2].dCoord[Y] = 0.5*r;
  tp[2].coeffs[0][2].dCoord[Z] = 0.0;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = w1*r;
  tp[2].coeffs[1][2].dCoord[Y] = 0.5*r;
  tp[2].coeffs[1][2].dCoord[Z] = l;
  tp[2].coeffs[1][2].dCoord[W] = 1.0;

  /* generated the sides!!! Now the lids */
  r *= CYL_LEEWAY_FACTOR;

  tp[3].coeffs[0][0].dCoord[X] = r;
  tp[3].coeffs[0][0].dCoord[Y] = -r;
  tp[3].coeffs[0][0].dCoord[Z] = 0.0;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = r;
  tp[3].coeffs[0][1].dCoord[Y] = r;
  tp[3].coeffs[0][1].dCoord[Z] = 0.0;
  tp[3].coeffs[0][1].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = -r;
  tp[3].coeffs[1][0].dCoord[Y] = -r;
  tp[3].coeffs[1][0].dCoord[Z] = 0.0;
  tp[3].coeffs[1][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][1].dCoord[X] = -r;
  tp[3].coeffs[1][1].dCoord[Y] = r;
  tp[3].coeffs[1][1].dCoord[Z] = 0.0;
  tp[3].coeffs[1][1].dCoord[W] = 1.0;

  tp[4].coeffs[0][0].dCoord[X] = -r;
  tp[4].coeffs[0][0].dCoord[Y] = -r;
  tp[4].coeffs[0][0].dCoord[Z] = l;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = -r;
  tp[4].coeffs[0][1].dCoord[Y] = r;
  tp[4].coeffs[0][1].dCoord[Z] = l;
  tp[4].coeffs[0][1].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = r;
  tp[4].coeffs[1][0].dCoord[Y] = -r;
  tp[4].coeffs[1][0].dCoord[Z] = l;
  tp[4].coeffs[1][0].dCoord[W] = 1.0;

  tp[4].coeffs[1][1].dCoord[X] = r;
  tp[4].coeffs[1][1].dCoord[Y] = r;
  tp[4].coeffs[1][1].dCoord[Z] = l;
  tp[4].coeffs[1][1].dCoord[W] = 1.0;

/*
  tp[0].coeffs[0][0].dCoord[X] = 0.0;
  tp[0].coeffs[0][0].dCoord[Y] = r;
  tp[0].coeffs[0][0].dCoord[Z] = 0.0;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = 0.0;
  tp[0].coeffs[1][0].dCoord[Y] = r;
  tp[0].coeffs[1][0].dCoord[Z] = l;
  tp[0].coeffs[1][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = r*w;
  tp[0].coeffs[0][1].dCoord[Y] = r*w;
  tp[0].coeffs[0][1].dCoord[Z] = 0.0;
  tp[0].coeffs[0][1].dCoord[W] = w;

  tp[0].coeffs[1][1].dCoord[X] = r*w;
  tp[0].coeffs[1][1].dCoord[Y] = r*w;
  tp[0].coeffs[1][1].dCoord[Z] = l*w;
  tp[0].coeffs[1][1].dCoord[W] = w;

  tp[0].coeffs[0][2].dCoord[X] = r;
  tp[0].coeffs[0][2].dCoord[Y] = 0.0;
  tp[0].coeffs[0][2].dCoord[Z] = 0.0;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = r;
  tp[0].coeffs[1][2].dCoord[Y] = 0.0;
  tp[0].coeffs[1][2].dCoord[Z] = l;
  tp[0].coeffs[1][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = r;
  tp[1].coeffs[0][0].dCoord[Y] = 0.0;
  tp[1].coeffs[0][0].dCoord[Z] = 0.0;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = r;
  tp[1].coeffs[1][0].dCoord[Y] = 0.0;
  tp[1].coeffs[1][0].dCoord[Z] = l;
  tp[1].coeffs[1][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = r*w;
  tp[1].coeffs[0][1].dCoord[Y] = -r*w;
  tp[1].coeffs[0][1].dCoord[Z] = 0.0;
  tp[1].coeffs[0][1].dCoord[W] = w;

  tp[1].coeffs[1][1].dCoord[X] = r*w;
  tp[1].coeffs[1][1].dCoord[Y] = -r*w;
  tp[1].coeffs[1][1].dCoord[Z] = l*w;
  tp[1].coeffs[1][1].dCoord[W] = w;

  tp[1].coeffs[0][2].dCoord[X] = 0.0;
  tp[1].coeffs[0][2].dCoord[Y] = -r;
  tp[1].coeffs[0][2].dCoord[Z] = 0.0;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = 0.0;
  tp[1].coeffs[1][2].dCoord[Y] = -r;
  tp[1].coeffs[1][2].dCoord[Z] = l;
  tp[1].coeffs[1][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = 0.0;
  tp[2].coeffs[0][0].dCoord[Y] = -r;
  tp[2].coeffs[0][0].dCoord[Z] = 0.0;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = 0.0;
  tp[2].coeffs[1][0].dCoord[Y] = -r;
  tp[2].coeffs[1][0].dCoord[Z] = l;
  tp[2].coeffs[1][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = -r*w;
  tp[2].coeffs[0][1].dCoord[Y] = -r*w;
  tp[2].coeffs[0][1].dCoord[Z] = 0.0;
  tp[2].coeffs[0][1].dCoord[W] = w;

  tp[2].coeffs[1][1].dCoord[X] = -r*w;
  tp[2].coeffs[1][1].dCoord[Y] = -r*w;
  tp[2].coeffs[1][1].dCoord[Z] = l*w;
  tp[2].coeffs[1][1].dCoord[W] = w;

  tp[2].coeffs[0][2].dCoord[X] = -r;
  tp[2].coeffs[0][2].dCoord[Y] = 0.0;
  tp[2].coeffs[0][2].dCoord[Z] = 0.0;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = -r;
  tp[2].coeffs[1][2].dCoord[Y] = 0.0;
  tp[2].coeffs[1][2].dCoord[Z] = l;
  tp[2].coeffs[1][2].dCoord[W] = 1.0;

  tp[3].coeffs[0][0].dCoord[X] = -r;
  tp[3].coeffs[0][0].dCoord[Y] = 0.0;
  tp[3].coeffs[0][0].dCoord[Z] = 0.0;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = -r;
  tp[3].coeffs[1][0].dCoord[Y] = 0.0;
  tp[3].coeffs[1][0].dCoord[Z] = l;
  tp[3].coeffs[1][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = -r*w;
  tp[3].coeffs[0][1].dCoord[Y] = r*w;
  tp[3].coeffs[0][1].dCoord[Z] = 0.0;
  tp[3].coeffs[0][1].dCoord[W] = w;

  tp[3].coeffs[1][1].dCoord[X] = -r*w;
  tp[3].coeffs[1][1].dCoord[Y] = r*w;
  tp[3].coeffs[1][1].dCoord[Z] = l*w;
  tp[3].coeffs[1][1].dCoord[W] = w;

  tp[3].coeffs[0][2].dCoord[X] = 0.0;
  tp[3].coeffs[0][2].dCoord[Y] = r;
  tp[3].coeffs[0][2].dCoord[Z] = 0.0;
  tp[3].coeffs[0][2].dCoord[W] = 1.0;

  tp[3].coeffs[1][2].dCoord[X] = 0.0;
  tp[3].coeffs[1][2].dCoord[Y] = r;
  tp[3].coeffs[1][2].dCoord[Z] = l;
  tp[3].coeffs[1][2].dCoord[W] = 1.0;
*/

  /* generated the sides ... now generate lids */

/*
  r *= CYL_LEEWAY_FACTOR;

  tp[4].coeffs[0][0].dCoord[X] = r;
  tp[4].coeffs[0][0].dCoord[Y] = -r;
  tp[4].coeffs[0][0].dCoord[Z] = 0.0;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = r;
  tp[4].coeffs[0][1].dCoord[Y] = r;
  tp[4].coeffs[0][1].dCoord[Z] = 0.0;
  tp[4].coeffs[0][1].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = -r;
  tp[4].coeffs[1][0].dCoord[Y] = -r;
  tp[4].coeffs[1][0].dCoord[Z] = 0.0;
  tp[4].coeffs[1][0].dCoord[W] = 1.0;

  tp[4].coeffs[1][1].dCoord[X] = -r;
  tp[4].coeffs[1][1].dCoord[Y] = r;
  tp[4].coeffs[1][1].dCoord[Z] = 0.0;
  tp[4].coeffs[1][1].dCoord[W] = 1.0;

  
  tp[5].coeffs[0][0].dCoord[X] = r;
  tp[5].coeffs[0][0].dCoord[Y] = -r;
  tp[5].coeffs[0][0].dCoord[Z] = l;
  tp[5].coeffs[0][0].dCoord[W] = 1.0;

  tp[5].coeffs[0][1].dCoord[X] = -r;
  tp[5].coeffs[0][1].dCoord[Y] = -r;
  tp[5].coeffs[0][1].dCoord[Z] = l;
  tp[5].coeffs[0][1].dCoord[W] = 1.0;

  tp[5].coeffs[1][0].dCoord[X] = r;
  tp[5].coeffs[1][0].dCoord[Y] = r;
  tp[5].coeffs[1][0].dCoord[Z] = l;
  tp[5].coeffs[1][0].dCoord[W] = 1.0;

  tp[5].coeffs[1][1].dCoord[X] = -r;
  tp[5].coeffs[1][1].dCoord[Y] = r;
  tp[5].coeffs[1][1].dCoord[Z] = l;
  tp[5].coeffs[1][1].dCoord[W] = 1.0;
*/
}

#else  /* old representation with the top and bottom represented as 4 */
       /* bezeir patches each */

#define	NUM_CYL_PATCHES		12
#define	CYL_SDEGREE		2	
#define	CYL_TDEGREE		1	

EXTERN_LINALG

extern void normalise();
extern void allocate_tensor_patch();
extern void XformPatch();

void getXformMatCyl(center, axis_vec, Xform)
POINT4D center, axis_vec;
double **Xform;
{
  double a, b, c, d, px, py, pz;

  a = axis_vec.dCoord[X];
  b = axis_vec.dCoord[Y];
  c = axis_vec.dCoord[Z];
  d = sqrt(b*b + c*c);
  px = center.dCoord[X];
  py = center.dCoord[Y];
  pz = center.dCoord[Z];

  if (d < 1e-5) {
    Xform[0][0] = 0.0;
    Xform[0][1] = 0.0;
    Xform[0][2] = -1.0;
    Xform[0][3] = 0.0;
  
    Xform[1][0] = 0.0;
    Xform[1][1] = 1.0;
    Xform[1][2] = 0.0;
    Xform[1][3] = 0.0;
  
    Xform[2][0] = 1.0;
    Xform[2][1] = 0.0;
    Xform[2][2] = 0.0;
    Xform[2][3] = 0.0;
  
    Xform[3][0] = px;
    Xform[3][1] = py;
    Xform[3][2] = pz;
    Xform[3][3] = 1.0;
    return;
  }
  Xform[0][0] = d;
  Xform[0][1] = -a*b/d;
  Xform[0][2] = -a*c/d;
  Xform[0][3] = 0.0;
  
  Xform[1][0] = 0.0;
  Xform[1][1] = c/d;
  Xform[1][2] = -b/d;
  Xform[1][3] = 0.0;
  
  Xform[2][0] = a;
  Xform[2][1] = b;
  Xform[2][2] = c;
  Xform[2][3] = 0.0;
  
  Xform[3][0] = px;
  Xform[3][1] = py;
  Xform[3][2] = pz;
  Xform[3][3] = 1.0;
}
  
void GetNormCylControlPts(tp, cyl_length, cyl_radius)
TENSOR_PATCH *tp;
double cyl_length, cyl_radius;
{
  double w, l, r;

  w = sqrt(2.0) / 2.0;  
  l = cyl_length;
  r = cyl_radius;

  tp[0].coeffs[0][0].dCoord[X] = 0.0;
  tp[0].coeffs[0][0].dCoord[Y] = r;
  tp[0].coeffs[0][0].dCoord[Z] = 0.0;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = 0.0;
  tp[0].coeffs[1][0].dCoord[Y] = r;
  tp[0].coeffs[1][0].dCoord[Z] = l;
  tp[0].coeffs[1][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = r*w;
  tp[0].coeffs[0][1].dCoord[Y] = r*w;
  tp[0].coeffs[0][1].dCoord[Z] = 0.0;
  tp[0].coeffs[0][1].dCoord[W] = w;

  tp[0].coeffs[1][1].dCoord[X] = r*w;
  tp[0].coeffs[1][1].dCoord[Y] = r*w;
  tp[0].coeffs[1][1].dCoord[Z] = l*w;
  tp[0].coeffs[1][1].dCoord[W] = w;

  tp[0].coeffs[0][2].dCoord[X] = r;
  tp[0].coeffs[0][2].dCoord[Y] = 0.0;
  tp[0].coeffs[0][2].dCoord[Z] = 0.0;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = r;
  tp[0].coeffs[1][2].dCoord[Y] = 0.0;
  tp[0].coeffs[1][2].dCoord[Z] = l;
  tp[0].coeffs[1][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = r;
  tp[1].coeffs[0][0].dCoord[Y] = 0.0;
  tp[1].coeffs[0][0].dCoord[Z] = 0.0;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = r;
  tp[1].coeffs[1][0].dCoord[Y] = 0.0;
  tp[1].coeffs[1][0].dCoord[Z] = l;
  tp[1].coeffs[1][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = r*w;
  tp[1].coeffs[0][1].dCoord[Y] = -r*w;
  tp[1].coeffs[0][1].dCoord[Z] = 0.0;
  tp[1].coeffs[0][1].dCoord[W] = w;

  tp[1].coeffs[1][1].dCoord[X] = r*w;
  tp[1].coeffs[1][1].dCoord[Y] = -r*w;
  tp[1].coeffs[1][1].dCoord[Z] = l*w;
  tp[1].coeffs[1][1].dCoord[W] = w;

  tp[1].coeffs[0][2].dCoord[X] = 0.0;
  tp[1].coeffs[0][2].dCoord[Y] = -r;
  tp[1].coeffs[0][2].dCoord[Z] = 0.0;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = 0.0;
  tp[1].coeffs[1][2].dCoord[Y] = -r;
  tp[1].coeffs[1][2].dCoord[Z] = l;
  tp[1].coeffs[1][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = 0.0;
  tp[2].coeffs[0][0].dCoord[Y] = -r;
  tp[2].coeffs[0][0].dCoord[Z] = 0.0;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = 0.0;
  tp[2].coeffs[1][0].dCoord[Y] = -r;
  tp[2].coeffs[1][0].dCoord[Z] = l;
  tp[2].coeffs[1][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = -r*w;
  tp[2].coeffs[0][1].dCoord[Y] = -r*w;
  tp[2].coeffs[0][1].dCoord[Z] = 0.0;
  tp[2].coeffs[0][1].dCoord[W] = w;

  tp[2].coeffs[1][1].dCoord[X] = -r*w;
  tp[2].coeffs[1][1].dCoord[Y] = -r*w;
  tp[2].coeffs[1][1].dCoord[Z] = l*w;
  tp[2].coeffs[1][1].dCoord[W] = w;

  tp[2].coeffs[0][2].dCoord[X] = -r;
  tp[2].coeffs[0][2].dCoord[Y] = 0.0;
  tp[2].coeffs[0][2].dCoord[Z] = 0.0;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = -r;
  tp[2].coeffs[1][2].dCoord[Y] = 0.0;
  tp[2].coeffs[1][2].dCoord[Z] = l;
  tp[2].coeffs[1][2].dCoord[W] = 1.0;

  tp[3].coeffs[0][0].dCoord[X] = -r;
  tp[3].coeffs[0][0].dCoord[Y] = 0.0;
  tp[3].coeffs[0][0].dCoord[Z] = 0.0;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = -r;
  tp[3].coeffs[1][0].dCoord[Y] = 0.0;
  tp[3].coeffs[1][0].dCoord[Z] = l;
  tp[3].coeffs[1][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = -r*w;
  tp[3].coeffs[0][1].dCoord[Y] = r*w;
  tp[3].coeffs[0][1].dCoord[Z] = 0.0;
  tp[3].coeffs[0][1].dCoord[W] = w;

  tp[3].coeffs[1][1].dCoord[X] = -r*w;
  tp[3].coeffs[1][1].dCoord[Y] = r*w;
  tp[3].coeffs[1][1].dCoord[Z] = l*w;
  tp[3].coeffs[1][1].dCoord[W] = w;

  tp[3].coeffs[0][2].dCoord[X] = 0.0;
  tp[3].coeffs[0][2].dCoord[Y] = r;
  tp[3].coeffs[0][2].dCoord[Z] = 0.0;
  tp[3].coeffs[0][2].dCoord[W] = 1.0;

  tp[3].coeffs[1][2].dCoord[X] = 0.0;
  tp[3].coeffs[1][2].dCoord[Y] = r;
  tp[3].coeffs[1][2].dCoord[Z] = l;
  tp[3].coeffs[1][2].dCoord[W] = 1.0;

  tp[4].coeffs[0][0].dCoord[X] = 0.0;
  tp[4].coeffs[0][0].dCoord[Y] = r;
  tp[4].coeffs[0][0].dCoord[Z] = 0.0;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = 0.0;
  tp[4].coeffs[0][1].dCoord[Y] = 0.0;
  tp[4].coeffs[0][1].dCoord[Z] = 0.0;
  tp[4].coeffs[0][1].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = r*w;
  tp[4].coeffs[1][0].dCoord[Y] = r*w;
  tp[4].coeffs[1][0].dCoord[Z] = 0.0;
  tp[4].coeffs[1][0].dCoord[W] = w;

  tp[4].coeffs[1][1].dCoord[X] = 0.0;
  tp[4].coeffs[1][1].dCoord[Y] = 0.0;
  tp[4].coeffs[1][1].dCoord[Z] = 0.0;
  tp[4].coeffs[1][1].dCoord[W] = w;

  tp[4].coeffs[2][0].dCoord[X] = r;
  tp[4].coeffs[2][0].dCoord[Y] = 0.0;
  tp[4].coeffs[2][0].dCoord[Z] = 0.0;
  tp[4].coeffs[2][0].dCoord[W] = 1.0;

  tp[4].coeffs[2][1].dCoord[X] = 0.0;
  tp[4].coeffs[2][1].dCoord[Y] = 0.0;
  tp[4].coeffs[2][1].dCoord[Z] = 0.0;
  tp[4].coeffs[2][1].dCoord[W] = 1.0;

  tp[5].coeffs[0][0].dCoord[X] = r;
  tp[5].coeffs[0][0].dCoord[Y] = 0.0;
  tp[5].coeffs[0][0].dCoord[Z] = 0.0;
  tp[5].coeffs[0][0].dCoord[W] = 1.0;

  tp[5].coeffs[0][1].dCoord[X] = 0.0;
  tp[5].coeffs[0][1].dCoord[Y] = 0.0;
  tp[5].coeffs[0][1].dCoord[Z] = 0.0;
  tp[5].coeffs[0][1].dCoord[W] = 1.0;

  tp[5].coeffs[1][0].dCoord[X] = r*w;
  tp[5].coeffs[1][0].dCoord[Y] = -r*w;
  tp[5].coeffs[1][0].dCoord[Z] = 0.0;
  tp[5].coeffs[1][0].dCoord[W] = w;

  tp[5].coeffs[1][1].dCoord[X] = 0.0;
  tp[5].coeffs[1][1].dCoord[Y] = 0.0;
  tp[5].coeffs[1][1].dCoord[Z] = 0.0;
  tp[5].coeffs[1][1].dCoord[W] = w;

  tp[5].coeffs[2][0].dCoord[X] = 0.0;
  tp[5].coeffs[2][0].dCoord[Y] = -r;
  tp[5].coeffs[2][0].dCoord[Z] = 0.0;
  tp[5].coeffs[2][0].dCoord[W] = 1.0;

  tp[5].coeffs[2][1].dCoord[X] = 0.0;
  tp[5].coeffs[2][1].dCoord[Y] = 0.0;
  tp[5].coeffs[2][1].dCoord[Z] = 0.0;
  tp[5].coeffs[2][1].dCoord[W] = 1.0;

  tp[6].coeffs[0][0].dCoord[X] = 0.0;
  tp[6].coeffs[0][0].dCoord[Y] = -r;
  tp[6].coeffs[0][0].dCoord[Z] = 0.0;
  tp[6].coeffs[0][0].dCoord[W] = 1.0;

  tp[6].coeffs[0][1].dCoord[X] = 0.0;
  tp[6].coeffs[0][1].dCoord[Y] = 0.0;
  tp[6].coeffs[0][1].dCoord[Z] = 0.0;
  tp[6].coeffs[0][1].dCoord[W] = 1.0;

  tp[6].coeffs[1][0].dCoord[X] = -r*w;
  tp[6].coeffs[1][0].dCoord[Y] = -r*w;
  tp[6].coeffs[1][0].dCoord[Z] = 0.0;
  tp[6].coeffs[1][0].dCoord[W] = w;

  tp[6].coeffs[1][1].dCoord[X] = 0.0;
  tp[6].coeffs[1][1].dCoord[Y] = 0.0;
  tp[6].coeffs[1][1].dCoord[Z] = 0.0;
  tp[6].coeffs[1][1].dCoord[W] = w;

  tp[6].coeffs[2][0].dCoord[X] = -r;
  tp[6].coeffs[2][0].dCoord[Y] = 0.0;
  tp[6].coeffs[2][0].dCoord[Z] = 0.0;
  tp[6].coeffs[2][0].dCoord[W] = 1.0;

  tp[6].coeffs[2][1].dCoord[X] = 0.0;
  tp[6].coeffs[2][1].dCoord[Y] = 0.0;
  tp[6].coeffs[2][1].dCoord[Z] = 0.0;
  tp[6].coeffs[2][1].dCoord[W] = 1.0;

  tp[7].coeffs[0][0].dCoord[X] = -r;
  tp[7].coeffs[0][0].dCoord[Y] = 0.0;
  tp[7].coeffs[0][0].dCoord[Z] = 0.0;
  tp[7].coeffs[0][0].dCoord[W] = 1.0;

  tp[7].coeffs[0][1].dCoord[X] = 0.0;
  tp[7].coeffs[0][1].dCoord[Y] = 0.0;
  tp[7].coeffs[0][1].dCoord[Z] = 0.0;
  tp[7].coeffs[0][1].dCoord[W] = 1.0;

  tp[7].coeffs[1][0].dCoord[X] = -r*w;
  tp[7].coeffs[1][0].dCoord[Y] = r*w;
  tp[7].coeffs[1][0].dCoord[Z] = 0.0;
  tp[7].coeffs[1][0].dCoord[W] = w;

  tp[7].coeffs[1][1].dCoord[X] = 0.0;
  tp[7].coeffs[1][1].dCoord[Y] = 0.0;
  tp[7].coeffs[1][1].dCoord[Z] = 0.0;
  tp[7].coeffs[1][1].dCoord[W] = w;

  tp[7].coeffs[2][0].dCoord[X] = 0.0;
  tp[7].coeffs[2][0].dCoord[Y] = r;
  tp[7].coeffs[2][0].dCoord[Z] = 0.0;
  tp[7].coeffs[2][0].dCoord[W] = 1.0;

  tp[7].coeffs[2][1].dCoord[X] = 0.0;
  tp[7].coeffs[2][1].dCoord[Y] = 0.0;
  tp[7].coeffs[2][1].dCoord[Z] = 0.0;
  tp[7].coeffs[2][1].dCoord[W] = 1.0;

  tp[8].coeffs[0][0].dCoord[X] = 0.0;
  tp[8].coeffs[0][0].dCoord[Y] = r;
  tp[8].coeffs[0][0].dCoord[Z] = l;
  tp[8].coeffs[0][0].dCoord[W] = 1.0;

  tp[8].coeffs[1][0].dCoord[X] = 0.0;
  tp[8].coeffs[1][0].dCoord[Y] = 0.0;
  tp[8].coeffs[1][0].dCoord[Z] = l;
  tp[8].coeffs[1][0].dCoord[W] = 1.0;

  tp[8].coeffs[0][1].dCoord[X] = r*w;
  tp[8].coeffs[0][1].dCoord[Y] = r*w;
  tp[8].coeffs[0][1].dCoord[Z] = l*w;
  tp[8].coeffs[0][1].dCoord[W] = w;

  tp[8].coeffs[1][1].dCoord[X] = 0.0;
  tp[8].coeffs[1][1].dCoord[Y] = 0.0;
  tp[8].coeffs[1][1].dCoord[Z] = l*w;
  tp[8].coeffs[1][1].dCoord[W] = w;

  tp[8].coeffs[0][2].dCoord[X] = r;
  tp[8].coeffs[0][2].dCoord[Y] = 0.0;
  tp[8].coeffs[0][2].dCoord[Z] = l;
  tp[8].coeffs[0][2].dCoord[W] = 1.0;

  tp[8].coeffs[1][2].dCoord[X] = 0.0;
  tp[8].coeffs[1][2].dCoord[Y] = 0.0;
  tp[8].coeffs[1][2].dCoord[Z] = l;
  tp[8].coeffs[1][2].dCoord[W] = 1.0;

  tp[9].coeffs[0][0].dCoord[X] = r;
  tp[9].coeffs[0][0].dCoord[Y] = 0.0;
  tp[9].coeffs[0][0].dCoord[Z] = l;
  tp[9].coeffs[0][0].dCoord[W] = 1.0;

  tp[9].coeffs[1][0].dCoord[X] = 0.0;
  tp[9].coeffs[1][0].dCoord[Y] = 0.0;
  tp[9].coeffs[1][0].dCoord[Z] = l;
  tp[9].coeffs[1][0].dCoord[W] = 1.0;

  tp[9].coeffs[0][1].dCoord[X] = r*w;
  tp[9].coeffs[0][1].dCoord[Y] = -r*w;
  tp[9].coeffs[0][1].dCoord[Z] = l*w;
  tp[9].coeffs[0][1].dCoord[W] = w;

  tp[9].coeffs[1][1].dCoord[X] = 0.0;
  tp[9].coeffs[1][1].dCoord[Y] = 0.0;
  tp[9].coeffs[1][1].dCoord[Z] = l*w;
  tp[9].coeffs[1][1].dCoord[W] = w;

  tp[9].coeffs[0][2].dCoord[X] = 0.0;
  tp[9].coeffs[0][2].dCoord[Y] = -r;
  tp[9].coeffs[0][2].dCoord[Z] = l;
  tp[9].coeffs[0][2].dCoord[W] = 1.0;

  tp[9].coeffs[1][2].dCoord[X] = 0.0;
  tp[9].coeffs[1][2].dCoord[Y] = 0.0;
  tp[9].coeffs[1][2].dCoord[Z] = l;
  tp[9].coeffs[1][2].dCoord[W] = 1.0;

  tp[10].coeffs[0][0].dCoord[X] = 0.0;
  tp[10].coeffs[0][0].dCoord[Y] = -r;
  tp[10].coeffs[0][0].dCoord[Z] = l;
  tp[10].coeffs[0][0].dCoord[W] = 1.0;

  tp[10].coeffs[1][0].dCoord[X] = 0.0;
  tp[10].coeffs[1][0].dCoord[Y] = 0.0;
  tp[10].coeffs[1][0].dCoord[Z] = l;
  tp[10].coeffs[1][0].dCoord[W] = 1.0;

  tp[10].coeffs[0][1].dCoord[X] = -r*w;
  tp[10].coeffs[0][1].dCoord[Y] = -r*w;
  tp[10].coeffs[0][1].dCoord[Z] = l*w;
  tp[10].coeffs[0][1].dCoord[W] = w;

  tp[10].coeffs[1][1].dCoord[X] = 0.0;
  tp[10].coeffs[1][1].dCoord[Y] = 0.0;
  tp[10].coeffs[1][1].dCoord[Z] = l*w;
  tp[10].coeffs[1][1].dCoord[W] = w;

  tp[10].coeffs[0][2].dCoord[X] = -r;
  tp[10].coeffs[0][2].dCoord[Y] = 0.0;
  tp[10].coeffs[0][2].dCoord[Z] = l;
  tp[10].coeffs[0][2].dCoord[W] = 1.0;

  tp[10].coeffs[1][2].dCoord[X] = 0.0;
  tp[10].coeffs[1][2].dCoord[Y] = 0.0;
  tp[10].coeffs[1][2].dCoord[Z] = l;
  tp[10].coeffs[1][2].dCoord[W] = 1.0;

  tp[11].coeffs[0][0].dCoord[X] = -r;
  tp[11].coeffs[0][0].dCoord[Y] = 0.0;
  tp[11].coeffs[0][0].dCoord[Z] = l;
  tp[11].coeffs[0][0].dCoord[W] = 1.0;

  tp[11].coeffs[1][0].dCoord[X] = 0.0;
  tp[11].coeffs[1][0].dCoord[Y] = 0.0;
  tp[11].coeffs[1][0].dCoord[Z] = l;
  tp[11].coeffs[1][0].dCoord[W] = 1.0;

  tp[11].coeffs[0][1].dCoord[X] = -r*w;
  tp[11].coeffs[0][1].dCoord[Y] = r*w;
  tp[11].coeffs[0][1].dCoord[Z] = l*w;
  tp[11].coeffs[0][1].dCoord[W] = w;

  tp[11].coeffs[1][1].dCoord[X] = 0.0;
  tp[11].coeffs[1][1].dCoord[Y] = 0.0;
  tp[11].coeffs[1][1].dCoord[Z] = l*w;
  tp[11].coeffs[1][1].dCoord[W] = w;

  tp[11].coeffs[0][2].dCoord[X] = 0.0;
  tp[11].coeffs[0][2].dCoord[Y] = r;
  tp[11].coeffs[0][2].dCoord[Z] = l;
  tp[11].coeffs[0][2].dCoord[W] = 1.0;

  tp[11].coeffs[1][2].dCoord[X] = 0.0;
  tp[11].coeffs[1][2].dCoord[Y] = 0.0;
  tp[11].coeffs[1][2].dCoord[Z] = l;
  tp[11].coeffs[1][2].dCoord[W] = 1.0;
}

#endif


int gen_cylinder(center, axis_vec, cyl_length, cyl_radius, tp)
POINT4D center, axis_vec;
double cyl_length, cyl_radius;
TENSOR_PATCH **tp;
{
  int i, j, k;
  double **XformMatCyl;

  normalise(&axis_vec);
  ALLOCMAT(XformMatCyl, 4, 4, double);
  getXformMatCyl(center, axis_vec, XformMatCyl);

  ALLOCN(*tp, TENSOR_PATCH, NUM_CYL_PATCHES);

#ifdef NEW_CYL_REPRN		/* new representation */
  
  for (k=0; k<NUM_CYL_PATCHES; k++) {
    if (k > 2)
      (*tp)[k].iDeg0 = (*tp)[k].iDeg1 = 1;
    else 
      {
	(*tp)[k].iDeg1 = CYL_SDEGREE;
	(*tp)[k].iDeg0 = CYL_TDEGREE;
      }
    allocate_tensor_patch(&((*tp)[k]));
  }
  
#else  /* older representation */

  /* Patches 0-3 : sides
     Patches 4-7 : bottom lid
     Patches 8-11 : top lid 
     */

  for (k=0; k<NUM_CYL_PATCHES; k++) {
    if (k > 3 && k < 8) {
      (*tp)[k].iDeg0 = CYL_SDEGREE;
      (*tp)[k].iDeg1 = CYL_TDEGREE;
    } else {
      (*tp)[k].iDeg1 = CYL_SDEGREE;
      (*tp)[k].iDeg0 = CYL_TDEGREE;
    }
    allocate_tensor_patch(&((*tp)[k]));
  }

#endif
  
  GetNormCylControlPts(*tp, cyl_length, cyl_radius);
  
  /* After getting normalised cylinder's control points ie. cylinder with
     center at origin and axis corresponding to the z axis, all we have
     to do is to transform each control point using XformMat and then 
     writing it into a file */
  
  XformPatch(*tp, NUM_CYL_PATCHES, CYL_SDEGREE, CYL_TDEGREE, XformMatCyl);
  return NUM_CYL_PATCHES;  
}

