#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "defs.h"
#include "linalg.h"

#define USESHANKARSPHERE
#define NUM_ELL_PATCHES 8
#define ELL_SDEGREE 2
#define ELL_TDEGREE 2

EXTERN_LINALG
extern void allocate_tensor_patch();
extern void XformPatch();

/******************** getXformMatEll ******************/
/* Generate transform matrix to shift and rotate      */
/*    the ellipsoid.                                  */
/******************************************************/
void getXformMatEll(center,a_vec,b_vec,c_vec,Xform)
POINT4D center, a_vec, b_vec, c_vec;
double **Xform;
{
   Xform[0][0] = a_vec.dCoord[0];
   Xform[1][0] = b_vec.dCoord[0];
   Xform[2][0] = c_vec.dCoord[0];
   Xform[3][0] = center.dCoord[0];

   Xform[0][1] = a_vec.dCoord[1];
   Xform[1][1] = b_vec.dCoord[1];
   Xform[2][1] = c_vec.dCoord[1];
   Xform[3][1] = center.dCoord[1];

   Xform[0][2] = a_vec.dCoord[2];
   Xform[1][2] = b_vec.dCoord[2];
   Xform[2][2] = c_vec.dCoord[2];
   Xform[3][2] = center.dCoord[2];

   Xform[0][3] = 0.0;
   Xform[1][3] = 0.0;
   Xform[2][3] = 0.0;
   Xform[3][3] = 1.0;
}

/******************** getXformMatEll2 *****************/
/* Generate transform matrix to shift and rotate      */
/*    the ellipsoid.  Includes random rotation.       */
/******************************************************/
void getXformMatEll2(center,a_vec,b_vec,c_vec,rotamt1,rotamt2,Xform)
POINT4D center, a_vec, b_vec, c_vec;
double **Xform;
double rotamt1,rotamt2;
{
   double crot1,srot1,crot2,srot2;

   crot1 = cos(rotamt1);
   srot1 = sin(rotamt1);
   crot2 = cos(rotamt2);
   srot2 = sin(rotamt2);

   Xform[0][0] = a_vec.dCoord[0]*crot1 + b_vec.dCoord[0]*srot1;
   Xform[1][0] = b_vec.dCoord[0]*crot2*crot1 + c_vec.dCoord[0]*srot2 -
                 a_vec.dCoord[0]*crot2*srot1;
   Xform[2][0] = c_vec.dCoord[0]*crot2 + a_vec.dCoord[0]*srot2*srot1 -
                 b_vec.dCoord[0]*srot2*crot1;
   Xform[3][0] = center.dCoord[0];

   Xform[0][1] = a_vec.dCoord[1]*crot1+b_vec.dCoord[1]*srot1;
   Xform[1][1] = b_vec.dCoord[1]*crot2*crot1 + c_vec.dCoord[1]*srot2 -
                 a_vec.dCoord[1]*crot2*srot1;
   Xform[2][1] = c_vec.dCoord[1]*crot2 + a_vec.dCoord[1]*srot2*srot1 -
                 b_vec.dCoord[1]*srot2*crot1;
   Xform[3][1] = center.dCoord[1];

   Xform[0][2] = a_vec.dCoord[2]*crot1+b_vec.dCoord[2]*srot1;
   Xform[1][2] = b_vec.dCoord[2]*crot2*crot1 + c_vec.dCoord[2]*srot2 -
                 a_vec.dCoord[2]*crot2*srot1;
   Xform[2][2] = c_vec.dCoord[2]*crot2 + a_vec.dCoord[2]*srot2*srot1 -
                 b_vec.dCoord[2]*srot2*crot1;
   Xform[3][2] = center.dCoord[2];

   Xform[0][3] = 0.0;
   Xform[1][3] = 0.0;
   Xform[2][3] = 0.0;
   Xform[3][3] = 1.0;
}

/******************** GetEllControlPoints *************/
/* Get Control Points for the Ellipsoid               */
/******************************************************/
void GetEllControlPoints(tp, a, b, c)
TENSOR_PATCH *tp;
double a, b, c;
{
   double invsq2;
   double invsq3;

   invsq2 = sqrt(2.0)/2.0;
   invsq3 = 0.5;


   /******* patch 0 *******/
   tp[0].coeffs[0][0].dCoord[X] = 0.0;
   tp[0].coeffs[0][0].dCoord[Y] = 0.0;
   tp[0].coeffs[0][0].dCoord[Z] = c;
   tp[0].coeffs[0][0].dCoord[W] = 1.0;

   tp[0].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[0].coeffs[1][0].dCoord[Y] = 0.0;
   tp[0].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[0].coeffs[1][0].dCoord[W] = invsq2;

   tp[0].coeffs[2][0].dCoord[X] = a;
   tp[0].coeffs[2][0].dCoord[Y] = 0.0;
   tp[0].coeffs[2][0].dCoord[Z] = 0.0;
   tp[0].coeffs[2][0].dCoord[W] = 1.0;

   tp[0].coeffs[0][1].dCoord[X] = 0.0;
   tp[0].coeffs[0][1].dCoord[Y] = 0.0;
   tp[0].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[0].coeffs[0][1].dCoord[W] = invsq2;

   tp[0].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[0].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[0].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[0].coeffs[1][1].dCoord[W] = invsq3;

   tp[0].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[0].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[0].coeffs[2][1].dCoord[Z] = 0.0;
   tp[0].coeffs[2][1].dCoord[W] = invsq2;

   tp[0].coeffs[0][2].dCoord[X] = 0.0;
   tp[0].coeffs[0][2].dCoord[Y] = 0.0;
   tp[0].coeffs[0][2].dCoord[Z] = c;
   tp[0].coeffs[0][2].dCoord[W] = 1.0;

   tp[0].coeffs[1][2].dCoord[X] = 0.0;
   tp[0].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[0].coeffs[1][2].dCoord[Z] = c*invsq2;
   tp[0].coeffs[1][2].dCoord[W] = invsq2;

   tp[0].coeffs[2][2].dCoord[X] = 0.0;
   tp[0].coeffs[2][2].dCoord[Y] = b;
   tp[0].coeffs[2][2].dCoord[Z] = 0.0;
   tp[0].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 1 *******/
   a = a*(-1.0);
   tp[1].coeffs[0][0].dCoord[X] = 0.0;
   tp[1].coeffs[0][0].dCoord[Y] = 0.0;
   tp[1].coeffs[0][0].dCoord[Z] = c;
   tp[1].coeffs[0][0].dCoord[W] = 1.0;

   tp[1].coeffs[1][0].dCoord[X] = 0.0;
   tp[1].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[1].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[1].coeffs[1][0].dCoord[W] = invsq2;

   tp[1].coeffs[2][0].dCoord[X] = 0.0;
   tp[1].coeffs[2][0].dCoord[Y] = b;
   tp[1].coeffs[2][0].dCoord[Z] = 0.0;
   tp[1].coeffs[2][0].dCoord[W] = 1.0;

   tp[1].coeffs[0][1].dCoord[X] = 0.0;
   tp[1].coeffs[0][1].dCoord[Y] = 0.0;
   tp[1].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[1].coeffs[0][1].dCoord[W] = invsq2;

   tp[1].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[1].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[1].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[1].coeffs[1][1].dCoord[W] = invsq3;

   tp[1].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[1].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[1].coeffs[2][1].dCoord[Z] = 0.0;
   tp[1].coeffs[2][1].dCoord[W] = invsq2;

   tp[1].coeffs[0][2].dCoord[X] = 0.0;
   tp[1].coeffs[0][2].dCoord[Y] = 0.0;
   tp[1].coeffs[0][2].dCoord[Z] = c;
   tp[1].coeffs[0][2].dCoord[W] = 1.0;

   tp[1].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[1].coeffs[1][2].dCoord[Y] = 0.0;
   tp[1].coeffs[1][2].dCoord[Z] = c*invsq2;
   tp[1].coeffs[1][2].dCoord[W] = invsq2;

   tp[1].coeffs[2][2].dCoord[X] = a;
   tp[1].coeffs[2][2].dCoord[Y] = 0.0;
   tp[1].coeffs[2][2].dCoord[Z] = 0.0;
   tp[1].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 2 *******/
   b=b*(-1.0);
   tp[2].coeffs[0][0].dCoord[X] = 0.0;
   tp[2].coeffs[0][0].dCoord[Y] = 0.0;
   tp[2].coeffs[0][0].dCoord[Z] = c;
   tp[2].coeffs[0][0].dCoord[W] = 1.0;

   tp[2].coeffs[1][0].dCoord[X] = a*invsq2;
   tp[2].coeffs[1][0].dCoord[Y] = 0.0;
   tp[2].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[2].coeffs[1][0].dCoord[W] = invsq2;

   tp[2].coeffs[2][0].dCoord[X] = a;
   tp[2].coeffs[2][0].dCoord[Y] = 0.0;
   tp[2].coeffs[2][0].dCoord[Z] = 0.0;
   tp[2].coeffs[2][0].dCoord[W] = 1.0;

   tp[2].coeffs[0][1].dCoord[X] = 0.0;
   tp[2].coeffs[0][1].dCoord[Y] = 0.0;
   tp[2].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[2].coeffs[0][1].dCoord[W] = invsq2;

   tp[2].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[2].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[2].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[2].coeffs[1][1].dCoord[W] = invsq3;

   tp[2].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[2].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[2].coeffs[2][1].dCoord[Z] = 0.0;
   tp[2].coeffs[2][1].dCoord[W] = invsq2;

   tp[2].coeffs[0][2].dCoord[X] = 0.0;
   tp[2].coeffs[0][2].dCoord[Y] = 0.0;
   tp[2].coeffs[0][2].dCoord[Z] = c;
   tp[2].coeffs[0][2].dCoord[W] = 1.0;

   tp[2].coeffs[1][2].dCoord[X] = 0.0;
   tp[2].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[2].coeffs[1][2].dCoord[Z] = c*invsq2;
   tp[2].coeffs[1][2].dCoord[W] = invsq2;

   tp[2].coeffs[2][2].dCoord[X] = 0.0;
   tp[2].coeffs[2][2].dCoord[Y] = b;
   tp[2].coeffs[2][2].dCoord[Z] = 0.0;
   tp[2].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 3 *******/
   a=a*(-1.0);
   tp[3].coeffs[0][0].dCoord[X] = 0.0;
   tp[3].coeffs[0][0].dCoord[Y] = 0.0;
   tp[3].coeffs[0][0].dCoord[Z] = c;
   tp[3].coeffs[0][0].dCoord[W] = 1.0;

   tp[3].coeffs[1][0].dCoord[X] = 0.0;
   tp[3].coeffs[1][0].dCoord[Y] = b*invsq2;
   tp[3].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[3].coeffs[1][0].dCoord[W] = invsq2;

   tp[3].coeffs[2][0].dCoord[X] = 0.0;
   tp[3].coeffs[2][0].dCoord[Y] = b;
   tp[3].coeffs[2][0].dCoord[Z] = 0.0;
   tp[3].coeffs[2][0].dCoord[W] = 1.0;

   tp[3].coeffs[0][1].dCoord[X] = 0.0;
   tp[3].coeffs[0][1].dCoord[Y] = 0.0;
   tp[3].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[3].coeffs[0][1].dCoord[W] = invsq2;

   tp[3].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[3].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[3].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[3].coeffs[1][1].dCoord[W] = invsq3;

   tp[3].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[3].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[3].coeffs[2][1].dCoord[Z] = 0.0;
   tp[3].coeffs[2][1].dCoord[W] = invsq2;

   tp[3].coeffs[0][2].dCoord[X] = 0.0;
   tp[3].coeffs[0][2].dCoord[Y] = 0.0;
   tp[3].coeffs[0][2].dCoord[Z] = c;
   tp[3].coeffs[0][2].dCoord[W] = 1.0;

   tp[3].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[3].coeffs[1][2].dCoord[Y] = 0.0;
   tp[3].coeffs[1][2].dCoord[Z] = c*invsq2;
   tp[3].coeffs[1][2].dCoord[W] = invsq2;

   tp[3].coeffs[2][2].dCoord[X] = a;
   tp[3].coeffs[2][2].dCoord[Y] = 0.0;
   tp[3].coeffs[2][2].dCoord[Z] = 0.0;
   tp[3].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 4 *******/
   b=b*(-1.0);
   c=c*(-1.0);
   tp[4].coeffs[0][0].dCoord[X] = 0.0;
   tp[4].coeffs[0][0].dCoord[Y] = 0.0;
   tp[4].coeffs[0][0].dCoord[Z] = c;
   tp[4].coeffs[0][0].dCoord[W] = 1.0;

   tp[4].coeffs[0][1].dCoord[X] = a*invsq2;
   tp[4].coeffs[0][1].dCoord[Y] = 0.0;
   tp[4].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[4].coeffs[0][1].dCoord[W] = invsq2;

   tp[4].coeffs[0][2].dCoord[X] = a;
   tp[4].coeffs[0][2].dCoord[Y] = 0.0;
   tp[4].coeffs[0][2].dCoord[Z] = 0.0;
   tp[4].coeffs[0][2].dCoord[W] = 1.0;

   tp[4].coeffs[1][0].dCoord[X] = 0.0;
   tp[4].coeffs[1][0].dCoord[Y] = 0.0;
   tp[4].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[4].coeffs[1][0].dCoord[W] = invsq2;

   tp[4].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[4].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[4].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[4].coeffs[1][1].dCoord[W] = invsq3;

   tp[4].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[4].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[4].coeffs[1][2].dCoord[Z] = 0.0;
   tp[4].coeffs[1][2].dCoord[W] = invsq2;

   tp[4].coeffs[2][0].dCoord[X] = 0.0;
   tp[4].coeffs[2][0].dCoord[Y] = 0.0;
   tp[4].coeffs[2][0].dCoord[Z] = c;
   tp[4].coeffs[2][0].dCoord[W] = 1.0;

   tp[4].coeffs[2][1].dCoord[X] = 0.0;
   tp[4].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[4].coeffs[2][1].dCoord[Z] = c*invsq2;
   tp[4].coeffs[2][1].dCoord[W] = invsq2;

   tp[4].coeffs[2][2].dCoord[X] = 0.0;
   tp[4].coeffs[2][2].dCoord[Y] = b;
   tp[4].coeffs[2][2].dCoord[Z] = 0.0;
   tp[4].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 5 *******/
   a=a*(-1.0);
   tp[5].coeffs[0][0].dCoord[X] = 0.0;
   tp[5].coeffs[0][0].dCoord[Y] = 0.0;
   tp[5].coeffs[0][0].dCoord[Z] = c;
   tp[5].coeffs[0][0].dCoord[W] = 1.0;

   tp[5].coeffs[0][1].dCoord[X] = 0.0;
   tp[5].coeffs[0][1].dCoord[Y] = b*invsq2;
   tp[5].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[5].coeffs[0][1].dCoord[W] = invsq2;

   tp[5].coeffs[0][2].dCoord[X] = 0.0;
   tp[5].coeffs[0][2].dCoord[Y] = b;
   tp[5].coeffs[0][2].dCoord[Z] = 0.0;
   tp[5].coeffs[0][2].dCoord[W] = 1.0;

   tp[5].coeffs[1][0].dCoord[X] = 0.0;
   tp[5].coeffs[1][0].dCoord[Y] = 0.0;
   tp[5].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[5].coeffs[1][0].dCoord[W] = invsq2;

   tp[5].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[5].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[5].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[5].coeffs[1][1].dCoord[W] = invsq3;

   tp[5].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[5].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[5].coeffs[1][2].dCoord[Z] = 0.0;
   tp[5].coeffs[1][2].dCoord[W] = invsq2;

   tp[5].coeffs[2][0].dCoord[X] = 0.0;
   tp[5].coeffs[2][0].dCoord[Y] = 0.0;
   tp[5].coeffs[2][0].dCoord[Z] = c;
   tp[5].coeffs[2][0].dCoord[W] = 1.0;

   tp[5].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[5].coeffs[2][1].dCoord[Y] = 0.0;
   tp[5].coeffs[2][1].dCoord[Z] = c*invsq2;
   tp[5].coeffs[2][1].dCoord[W] = invsq2;

   tp[5].coeffs[2][2].dCoord[X] = a;
   tp[5].coeffs[2][2].dCoord[Y] = 0.0;
   tp[5].coeffs[2][2].dCoord[Z] = 0.0;
   tp[5].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 6 *******/
   b=b*(-1.0);
   tp[6].coeffs[0][0].dCoord[X] = 0.0;
   tp[6].coeffs[0][0].dCoord[Y] = 0.0;
   tp[6].coeffs[0][0].dCoord[Z] = c;
   tp[6].coeffs[0][0].dCoord[W] = 1.0;

   tp[6].coeffs[0][1].dCoord[X] = a*invsq2;
   tp[6].coeffs[0][1].dCoord[Y] = 0.0;
   tp[6].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[6].coeffs[0][1].dCoord[W] = invsq2;

   tp[6].coeffs[0][2].dCoord[X] = a;
   tp[6].coeffs[0][2].dCoord[Y] = 0.0;
   tp[6].coeffs[0][2].dCoord[Z] = 0.0;
   tp[6].coeffs[0][2].dCoord[W] = 1.0;

   tp[6].coeffs[1][0].dCoord[X] = 0.0;
   tp[6].coeffs[1][0].dCoord[Y] = 0.0;
   tp[6].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[6].coeffs[1][0].dCoord[W] = invsq2;

   tp[6].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[6].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[6].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[6].coeffs[1][1].dCoord[W] = invsq3;

   tp[6].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[6].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[6].coeffs[1][2].dCoord[Z] = 0.0;
   tp[6].coeffs[1][2].dCoord[W] = invsq2;

   tp[6].coeffs[2][0].dCoord[X] = 0.0;
   tp[6].coeffs[2][0].dCoord[Y] = 0.0;
   tp[6].coeffs[2][0].dCoord[Z] = c;
   tp[6].coeffs[2][0].dCoord[W] = 1.0;

   tp[6].coeffs[2][1].dCoord[X] = 0.0;
   tp[6].coeffs[2][1].dCoord[Y] = b*invsq2;
   tp[6].coeffs[2][1].dCoord[Z] = c*invsq2;
   tp[6].coeffs[2][1].dCoord[W] = invsq2;

   tp[6].coeffs[2][2].dCoord[X] = 0.0;
   tp[6].coeffs[2][2].dCoord[Y] = b;
   tp[6].coeffs[2][2].dCoord[Z] = 0.0;
   tp[6].coeffs[2][2].dCoord[W] = 1.0;

   /******* patch 7 *******/
   a=a*(-1.0);
   tp[7].coeffs[0][0].dCoord[X] = 0.0;
   tp[7].coeffs[0][0].dCoord[Y] = 0.0;
   tp[7].coeffs[0][0].dCoord[Z] = c;
   tp[7].coeffs[0][0].dCoord[W] = 1.0;

   tp[7].coeffs[0][1].dCoord[X] = 0.0;
   tp[7].coeffs[0][1].dCoord[Y] = b*invsq2;
   tp[7].coeffs[0][1].dCoord[Z] = c*invsq2;
   tp[7].coeffs[0][1].dCoord[W] = invsq2;

   tp[7].coeffs[0][2].dCoord[X] = 0.0;
   tp[7].coeffs[0][2].dCoord[Y] = b;
   tp[7].coeffs[0][2].dCoord[Z] = 0.0;
   tp[7].coeffs[0][2].dCoord[W] = 1.0;

   tp[7].coeffs[1][0].dCoord[X] = 0.0;
   tp[7].coeffs[1][0].dCoord[Y] = 0.0;
   tp[7].coeffs[1][0].dCoord[Z] = c*invsq2;
   tp[7].coeffs[1][0].dCoord[W] = invsq2;

   tp[7].coeffs[1][1].dCoord[X] = a*invsq3;
   tp[7].coeffs[1][1].dCoord[Y] = b*invsq3;
   tp[7].coeffs[1][1].dCoord[Z] = c*invsq3;
   tp[7].coeffs[1][1].dCoord[W] = invsq3;

   tp[7].coeffs[1][2].dCoord[X] = a*invsq2;
   tp[7].coeffs[1][2].dCoord[Y] = b*invsq2;
   tp[7].coeffs[1][2].dCoord[Z] = 0.0;
   tp[7].coeffs[1][2].dCoord[W] = invsq2;

   tp[7].coeffs[2][0].dCoord[X] = 0.0;
   tp[7].coeffs[2][0].dCoord[Y] = 0.0;
   tp[7].coeffs[2][0].dCoord[Z] = c;
   tp[7].coeffs[2][0].dCoord[W] = 1.0;

   tp[7].coeffs[2][1].dCoord[X] = a*invsq2;
   tp[7].coeffs[2][1].dCoord[Y] = 0.0;
   tp[7].coeffs[2][1].dCoord[Z] = c*invsq2;
   tp[7].coeffs[2][1].dCoord[W] = invsq2;

   tp[7].coeffs[2][2].dCoord[X] = a;
   tp[7].coeffs[2][2].dCoord[Y] = 0.0;
   tp[7].coeffs[2][2].dCoord[Z] = 0.0;
   tp[7].coeffs[2][2].dCoord[W] = 1.0;

}


/******************** gen_brlell *********************/
/* The "main" routine to create an ellipsoid         */
/*****************************************************/
int gen_brlell(center,a_vec,b_vec,c_vec,a,b,c,tp)
POINT4D center, a_vec, b_vec, c_vec;
double a, b, c;
TENSOR_PATCH **tp;
{
   int i;
   double **XformMatEll;
   double dotprod1,dotprod2,dotprod3;
   double rotamt1,rotamt2;

   ALLOCMAT(XformMatEll, 4, 4, double);
/* decide whether or not to rotate cylinder randomly, then get XformMatTgc */
   dotprod1 = a_vec.dCoord[0]*b_vec.dCoord[0] +
              a_vec.dCoord[1]*b_vec.dCoord[1] +
              a_vec.dCoord[2]*b_vec.dCoord[2];
   dotprod2 = a_vec.dCoord[0]*c_vec.dCoord[0] +
              a_vec.dCoord[1]*c_vec.dCoord[1] +
              a_vec.dCoord[2]*c_vec.dCoord[2];
   dotprod3 = b_vec.dCoord[0]*c_vec.dCoord[0] +
              b_vec.dCoord[1]*c_vec.dCoord[1] +
              b_vec.dCoord[2]*c_vec.dCoord[2];

#ifdef USESHANKARSPHERE
   if ((FP_EQ_EPS(a,b,(a*0.001))) && (FP_EQ_EPS(a,c,(a*0.001))) &&
       (FP_EQ_EPS(dotprod1,0.0,1e-4)) && (FP_EQ_EPS(dotprod2,0.0,1e-4)) &&
       (FP_EQ_EPS(dotprod3,0.0,1e-4))) {
      /* This is a sphere - use Shankar's code */
      return 0;
      /* we should rotate the sphere to help remove degenerate cases */
/*      rotamt1 = (3.1415927/2.0) * ((double)rand())/((double)RAND_MAX);
      rotamt2 = (3.1415927/2.0) * ((double)rand())/((double)RAND_MAX);
      getXformMatEll2(center,a_vec,b_vec,c_vec,rotamt1,rotamt2,XformMatEll);
*/
   } else {
#endif
      getXformMatEll(center,a_vec,b_vec,c_vec,XformMatEll);
#ifdef USESHANKARSPHERE
   }
#endif

   ALLOCN(*tp, TENSOR_PATCH, NUM_ELL_PATCHES);
   for(i=0;i<NUM_ELL_PATCHES;i++) {
      (*tp)[i].iDeg0 = ELL_SDEGREE;
      (*tp)[i].iDeg1 = ELL_TDEGREE;
      allocate_tensor_patch(&((*tp)[i]));
   }

   GetEllControlPoints(*tp,a,b,c);

   XformPatch(*tp, NUM_ELL_PATCHES, ELL_SDEGREE, ELL_TDEGREE, XformMatEll);
   return NUM_ELL_PATCHES;
}
