#include <tree.h>
#include <math.h>
#include <macro.h>

#define CONIC_COARSE_FACTOR  3	/* cf for conics */
#define HIDEG_COARSE_FACTOR  2	/* cf for the high (12) degree curves */

#define PRISM_LID_SCALE_FACTOR 1.0 /* check on this!! */

static struct lid_s {
  int is_scanned;
  int iDeg0;
  POINT4D *coeffs;
} lid[MAX_TPNUM];		/* table to hold the curves in */
                                /* order. while thay are being */
                                /* contructed  */

static vector_t corner[4];

static int coarse_factor;	/* how many points to gen. for a curve */

extern TENSOR_PATCH tpatch[];	/* table to hold the generated patches */
extern int tpatch_idx;
extern int top_is_zero;

vector_t topbot_trimcurve[4][MAX_TRIM_CURVE_POINTS];
struct {
  double x, y;
} topbot_inpoint[4];


#ifdef OLD_CODE_FOR_CURVES_SPLITTING

/* Given the inner and outer curves, break them apart in 2 places and */
/* get the 2 trim-curves separately */
static int get_two_curves(vector_t curve[], vector_t curve_inner[],
			  int cidx, int cidx_inner, vector_t  c1[],
			  vector_t c2[], int *c1idx, int *c2idx,
			  int is_clockwise)
{
  register int i, j;
  double cx;			/* x-cood of the centre */
  double s0, s1, t0, t1;
  int itype, opos0, opos1, ipos0, ipos1, tpos;
  struct {
    double x, y;
  } v0, v1, p0, p1, dum0, dum1;

  cidx--;			/* To take care of repeated points */
  cx = curve_inner[0][0];
  p0.x = curve_inner[0][0];
  p0.y = curve_inner[0][1];

  p1.x = 2.0;
  p1.y = p0.y;

  /* Now find which two segments of the other curve intersect this. */
  /* Use segment_intersection routine for this prupose  */
  
  for(i = 0; i <= cidx - 1; i++)
    {
      v0.x = curve[i][0];
      v0.y = curve[i][1];
      v1.x = curve[(i+1)%cidx][0];
      v1.y = curve[(i+1)%cidx][1];
      
      /* delta change */
      if (segments_intersect(&p0, &p1, &v0, &v1, &itype,
			     &s0, &s1, &t0, &t1, &dum0, &dum1))
	break;
    }
  
  if (i > cidx - 1)		/* no intersection */
    {
      fprintf(stderr, "get_two_curves: No intersection found\n");
      return -1;
    }
  else
    opos0 = i;

  /* Similarly, find the other endpoint  */

  p1.x = -2.0;
  p1.y = p0.y;

  /* Now find which two segments of the other curve intersect this. */
  /* Use segment_intersection routine for this prupose  */
  
  for(i = 0; i <= cidx - 1; i++)
    {
      v0.x = curve[i][0];
      v0.y = curve[i][1];
      v1.x = curve[(i+1)%cidx][0];
      v1.y = curve[(i+1)%cidx][1];

      if (segments_intersect(&p0, &p1, &v0, &v1, &itype,
			     &s0, &s1, &p0, &t1, &dum0, &dum1))
	break;
    }
  
  if (i > cidx - 1)		/* no intersection */
    {
      fprintf(stderr, "get_two_curves: No intersection found\n");
      return -1;
    }
  else
    opos1 = i;
  
  ipos0 = 0;
  ipos1 = 3*2*CONIC_COARSE_FACTOR; /* approximate.. each 90 deg arc */
				   /* has 9 points  */

  /* Verify that the two chosen internal positions are correct */
  
  if (curve_inner[ipos0][0] < curve_inner[ipos1][0])
    {
      tpos = ipos0;
      ipos0 = ipos1;
      ipos1 = tpos;
    }

  /* If the X-difference between these positions is lesser than the */
  /* y-difference, choose other points */
  
  if (fabs(curve_inner[ipos0][0] - curve_inner[ipos1][0]) < 
      fabs(curve_inner[ipos0][1] - curve_inner[ipos1][1]))
    {
      ipos0 = 3*CONIC_COARSE_FACTOR;
      ipos1 = ipos0 + 3*2*CONIC_COARSE_FACTOR;

      if (curve_inner[ipos0][0] < curve_inner[ipos1][0])
	{
	  tpos = ipos0;
	  ipos0 = ipos1;
	  ipos1 = tpos;
	}
    }
      
  /* get the two curves (upper and lower) */
  if (is_clockwise)
    {
      /* upper curve */
      j = 0;
      for (i = opos0; i != opos1; i = (i+1)%cidx)
	{
	  c1[j][0] = curve[i][0];
	  c1[j++][1] = curve[i][1];
	}
      c1[j][0] = curve[opos1][0];
      c1[j++][1] = curve[opos1][1];
      for (i = ipos1; i != ipos0; i = (i + cidx_inner - 1)%cidx_inner)
	{
	  c1[j][0] = curve_inner[i][0];
	  c1[j++][1] = curve_inner[i][1];
	}
      c1[j][0] = curve_inner[ipos0][0];
      c1[j++][1] = curve_inner[ipos0][1];
      c1[j][0] = c1[0][0];
      c1[j][1] = c1[0][1];
      *c1idx = j;
      
      /* lower curve */

      j = 0;
      for (i = ipos0; i != ipos1; i = (i + cidx_inner - 1)%cidx_inner)
	{
	  c2[j][0] = curve_inner[i][0];
	  c2[j++][1] = curve_inner[i][1];
	}
      c2[j][0] = curve_inner[ipos1][0];
      c2[j++][1] = curve_inner[ipos1][1];
      for (i = opos1; i != opos0; i = (i + 1)%cidx)
	{
	  c2[j][0] = curve[i][0];
	  c2[j++][1] = curve[i][1];
	}
      c2[j][0] = curve[opos0][0];
      c2[j++][1] = curve[opos0][1];
      c2[j][0] = c2[0][0];
      c2[j][1] = c2[0][1];
      *c2idx = j;
    }
  else				/* outer curve is anticlockwise */
    {
      /* upper curve */
      j = 0;
      for (i = opos0; i != opos1; i = (i + cidx - 1)%cidx)
	{
	  c1[j][0] = curve[i][0];
	  c1[j++][1] = curve[i][1];
	}
      c1[j][0] = curve[opos1][0];
      c1[j++][1] = curve[opos1][1];
      for (i = ipos1; i != ipos0; i = (i + cidx_inner - 1)%cidx_inner)
	{
	  c1[j][0] = curve_inner[i][0];
	  c1[j++][1] = curve_inner[i][1];
	}
      c1[j][0] = curve_inner[ipos0][0];
      c1[j++][1] = curve_inner[ipos0][1];
      c1[j][0] = c1[0][0];
      c1[j][1] = c1[0][1];
      *c1idx = j;
      
      /* lower curve */

      j = 0;
      for (i = ipos0; i != ipos1; i = (i + cidx_inner - 1)%cidx_inner)
	{
	  c2[j][0] = curve_inner[i][0];
	  c2[j++][1] = curve_inner[i][1];
	}
      c2[j][0] = curve_inner[ipos1][0];
      c2[j++][1] = curve_inner[ipos1][1];
      for (i = opos1; i != opos0; i = (i + cidx - 1)%cidx)
	{
	  c2[j][0] = curve[i][0];
	  c2[j++][1] = curve[i][1];
	}
      c2[j][0] = curve[opos0][0];
      c2[j++][1] = curve[opos0][1];
      c2[j][0] = c2[0][0];
      c2[j][1] = c2[0][1];
      *c2idx = j;
    }
  
  /* both the curves have now been generated */
#ifdef DEBUG
  printf("get_two_curves:1\n");
  for (i = 0; i < *c1idx; i++)
    printf("%lf %lf\n", c1[i][0], c1[i][1]);
  printf("get_two_curves:2\n");
  for (i = 0; i < *c2idx; i++)
    printf("%lf %lf\n", c2[i][0], c2[i][1]);
#endif

  return 0;
}

#else  /* A better code for connecting the curve with a hole. (Oct 6) */
       /* Use the code from splitting the s-t domain into 2 regoins */
       /* when we have a trim-curve with a hole. */

int get_two_curves(vector_t opoly[], vector_t cpoly[], 
			  int opoints, int cpoints,
			  vector_t  c1[], vector_t c2[], 
			  int *c1idx, int *c2idx,
			  int is_anticlockwise, int from_planar)
{
  register int i, j;
  int k, l, itype, oposmin, oposmax, cposmax, cposmin;
  double s0, t0, s1, t1, tmin, tmax;
  struct {
    double x, y;
  } v0, v1, p0, p1, q0, q1, q2, q3, cent, q0_sav, q1_sav, q2_sav, q3_sav;
  vector_t vtmp;

  /* If the outer-curve is clockwise, reverse it first so that it is */
  /* anticlockwise */

  if (!from_planar)
    opoints--;

  if (!is_anticlockwise)
    for (k = 0, l = opoints-1; k < l; k++, l--)
      {
	vtmp[0] = opoly[k][0];
	vtmp[1] = opoly[k][1];
	opoly[k][0] = opoly[l][0];
	opoly[k][1] = opoly[l][1];
	opoly[l][0] = vtmp[0];
	opoly[l][1] = vtmp[1];
      }

  /* Also, since by our assumption, the inner curve is always a convex */
  /* polygon. So it is very easy to test for the */
  /* clockwise/anticlockwiseness of the curve. Just consider the */
  /* cross-product of 2 adjacent vectors */

  if (((cpoly[1][0] - cpoly[0][0])*(cpoly[2][1] - cpoly[1][1]) - 
       (cpoly[1][1] - cpoly[0][1])*(cpoly[2][0] - cpoly[1][0])) < 0.0)
    {
      /* inner curve is clockwise */
    for (k = 0, l = cpoints-1; k < l; k++, l--)
      {
	vtmp[0] = cpoly[k][0];
	vtmp[1] = cpoly[k][1];
	cpoly[k][0] = cpoly[l][0];
	cpoly[k][1] = cpoly[l][1];
	cpoly[l][0] = vtmp[0];
	cpoly[l][1] = vtmp[1];
      }
  }

  opoly[opoints][0] = opoly[0][0];
  opoly[opoints][1] = opoly[0][1];

  cpoly[cpoints][0] = cpoly[0][0];
  cpoly[cpoints][1] = cpoly[0][1];

#ifdef DEBUG_PRINT_POLY
  printf("hole_removal: partition 0\n");
  for (i = 0; i < opoints; i++)
    printf("%lf %lf\n", opoly[i][0], opoly[i][1]);

  printf("hole_removal: partition 1\n");
  for (i = 0; i < cpoints; i++)
    printf("%lf %lf\n", cpoly[i][0], cpoly[i][1]);

#endif
  
  /* Find the centroid of the inner chain to get the vertical line to */
  /* use for intersections */

  cent.x = cent.y = 0.0;
  for (i = 0; i < cpoints; i++)
    {
      cent.x += cpoly[i][0];
      cent.y += cpoly[i][1];
    }
  cent.x /= cpoints;
  cent.y /= cpoints;
  
  /* Now find the two positions on the inner chain from where to */
  /* intersect the outer polygon */

#ifdef CENTROID_METHOD
  q0.x = 2.0;
  q2.x = 2.0;
  for (i = 0; i < cpoints; i++)
    if ((cpoly[i][1] > cent.y) && 
	(fabs(cpoly[i][0] - cent.x) < fabs(q0.x - cent.x)))
      {
	q0.x = cpoly[i][0];
	q0.y = cpoly[i][1];
      }
    else if ((cpoly[i][1] < cent.y) && 
	     (fabs(cpoly[i][0] - cent.x) < fabs(q2.x - cent.x)))
      {
	q2.x = cpoly[i][0];
	q2.y = cpoly[i][1];
      }   
  
#else
  q3.x = cent.x;
  q3.y = -1.0;
  q1.x = cent.x;
  q1.y = 2.0;

  tmin = 2.0;
  tmax = -2.0;
  for (i = 0; i < cpoints; i++)
    {
      v0.x = cpoly[i][0];
      v0.y = cpoly[i][1];
      v1.x = cpoly[i+1][0];
      v1.y = cpoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q3, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	{
	  if (t0 > tmax)
	    {
	      tmax = t0;
	      cposmax = i;
	    }
	  if (t0 < tmin)
	    {
	      tmin = t0;
	      cposmin = i;
	    }
	}
    }
  
  /* Get the two actual points on hte inner curve now */
  
  q0.x = (1 - tmax)*q3.x + tmax*q1.x;
  q0.y = (1 - tmax)*q3.y + tmax*q1.y;
  q2.x = (1 - tmin)*q3.x + tmin*q1.x;
  q2.y = (1 - tmin)*q3.y + tmin*q1.y;
  q0_sav = q0;
  q2_sav = q2;

#endif

  /* Now (hopefully), the upper endpoint is q0, and the lower andpoint */
  /* is q2 */
  
  q1.x = q0.x;
  q1.y = 2.0;
      
  tmin = 2.0;
  tmax = -2.0;
  
  /* Detect where this segment q0q1 intersects the outer polygon */
  for (i = 0; i < opoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q0, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmax = i;
	  }
    }

  q1_sav.x = (1 - tmin)*q0.x + tmin*q1.x;
  q1_sav.y = (1 - tmin)*q0.y + tmin*q1.y;


  q3.x = q2.x;
  q3.y = -2.0;
      
  tmin = 2.0;
  tmax = -2.0;

  /* Detect where this segment q2q3 intersects the outer polygon */
  for (i = 0; i < opoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q2, &q3, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmin = i;
	  }
    }

  q3_sav.x = (1 - tmin)*q2.x + tmin*q3.x;
  q3_sav.y = (1 - tmin)*q2.y + tmin*q3.y;

  /* now we have the values oposmax, oposmin, cposmax, cposmin..*/
  /* Generate the two partitions all over again */
  
  k = 0;
  for (i = (oposmin+1)%opoints; i != oposmax; i = (i+1)%opoints)
    {
      c1[k][0] = opoly[i][0];
      c1[k][1] = opoly[i][1];
      k++;
    }
  c1[k][0] = opoly[oposmax][0];
  c1[k][1] = opoly[oposmax][1];  
  k++;

/* Add the new segment q1-q0 in between */

  c1[k][0] = q1_sav.x;
  c1[k][1] = q1_sav.y;
  k++;  
  c1[k][0] = q0_sav.x;
  c1[k][1] = q0_sav.y;
  k++;  
  
  for (i = cposmax; i != cposmin; i = (i + cpoints - 1)%cpoints)
    {
      c1[k][0] = cpoly[i][0];
      c1[k][1] = cpoly[i][1];
      k++;
    }
  c1[k][0] = q2_sav.x;
  c1[k][1] = q2_sav.y;
  k++;  
  c1[k][0] = q3_sav.x;
  c1[k][1] = q3_sav.y;
  k++;  

  if (FP_EQUAL(c1[k-1][0], c1[0][0]) && 
      FP_EQUAL(c1[k-1][1], c1[0][1]))
    k--;
  
  *c1idx = k;
  
 
  /* Now generate partition 2 */
  
  k = 0;
  for (i = (oposmax+1)%opoints; i != oposmin; i = (i+1)%opoints)
    {
      c2[k][0] = opoly[i][0];
      c2[k][1] = opoly[i][1];
      k++;
    }
  c2[k][0] = opoly[oposmin][0];
  c2[k][1] = opoly[oposmin][1];  
  k++;
  c2[k][0] = q3_sav.x;
  c2[k++][1] = q3_sav.y;

  c2[k][0] = q2_sav.x;
  c2[k++][1] = q2_sav.y;

  for (i = cposmin; i != cposmax; i = (i + cpoints - 1)%cpoints)
    {
      c2[k][0] = cpoly[i][0];
      c2[k][1] = cpoly[i][1];
      k++;
    }
  c2[k][0] = q0_sav.x;
  c2[k++][1] = q0_sav.y;

  c2[k][0] = q1_sav.x;
  c2[k++][1] = q1_sav.y;

  if (FP_EQUAL(c2[k-1][0], c2[0][0]) && 
      FP_EQUAL(c2[k-1][1], c2[0][1]))
    k--;

  *c2idx = k;

  /* Remove the possible duplicate points. This cannot be helped */
  for (i = 0; i < *c1idx - 1; i++)
    {
      if (FP_EQUAL(c1[i][0], c1[i+1][0]) && 
	  FP_EQUAL(c1[i][1], c1[i+1][1]))
	{
	  for (j = i+1; j < *c1idx - 1; j++)
	    {
	      c1[j][0] = c1[j+1][0];
	      c1[j][1] = c1[j+1][1];
	    }
	  (*c1idx)--;
	}
    }

  /* Remove the possible duplicate points. This cannot be helped */
  for (i = 0; i < *c2idx - 1; i++)
    {
      if (FP_EQUAL(c2[i][0], c2[i+1][0]) && 
	  FP_EQUAL(c2[i][1], c2[i+1][1]))
	{
	  for (j = i+1; j < *c2idx - 1; j++)
	    {
	      c2[j][0] = c2[j+1][0];
	      c2[j][1] = c2[j+1][1];
	    }
	  (*c2idx)--;
	}
    }

#ifdef DEBUG_PRINT_POLY
  printf("hole_removal: partition 0\n");
  for (i = 0; i < *c1idx; i++)
    printf("%lf %lf\n", c1[i][0], c1[i][1]);

  printf("hole_removal: partition 1\n");
  for (i = 0; i < *c2idx; i++)
    printf("%lf %lf\n", c2[i][0], c2[i][1]);

#endif
  
  return 0;

}

#endif


/* np: pointer to the information about the primitive */
/* vo, v1: reference vectors of one of the cutting planes */
int generate_top_and_bottom_lids_for_prism(node_t *np,
			      vector_t v0, vector_t v1, int ndomains, 
 			      int ndomcrvs[], int *nl1, int *nl2)
{
  register int i, j, k;
  int lastidx, n, l, off, cidx, itype, icount, i_tval, i_ival, cidx_inner;
  TENSOR_PATCH *thisp;
  struct lid_s *lastp;
  matrix_t xmat_rot1, xmat_rot2;
  vector_t v, vv, b[MAX_CURVEDEG], curve[MAX_TRIM_CURVE_POINTS];
  vector_t curve_inner[MAX_TRIM_CURVE_POINTS];
  vector_t ioc1[MAX_TRIM_CURVE_POINTS], ioc2[MAX_TRIM_CURVE_POINTS];
  double parcurve[MAX_TRIM_CURVE_POINTS][2];
  double xmax, xmin, ymax, ymin, xscale, yscale, zmin, zmax;
  double t, step;
  point_typ p0, p1, vi0, vi1, c0, c1;
  double s0, s1, t0, t1;
  int tp0, tp1, c1idx, c2idx;
  TENSOR_PATCH *tpp0, *tpp1;
  int limit, nextfound, one_more_try;
  double xmax_tmp, ymax_tmp, xmin_tmp, ymin_tmp;
  int howmany_pts;

  /* First put all the individual curves in order */

  /* allocate space for the lids to hold the coeffs */
  off = 0;
  for (j = 0; j < tpatch_idx; j++)
    {
      lid[off+j].coeffs = (POINT4D *) new(MAX_CURVEDEG*sizeof(POINT4D));
      lid[off+j].is_scanned = FALSE;
    }
  
  off = 0;
  for (l = 0; l < ndomains; l++)
    {
      for (i = 0; i <= tpatch[0+off].iDeg0; i++)
	lid[0+off].coeffs[i] = tpatch[0+off].coeffs[i][0];
      lid[0+off].is_scanned = TRUE;
      lid[0+off].iDeg0 = tpatch[0+off].iDeg0;
      
      if (ndomains == 1)
	limit = tpatch_idx;
      else
	{
	  if (l == 0)
	    limit = tpatch_idx - 4;
	  else
	    limit = 4;
	}
      for (j = off + 1; j <= limit + off; j++)
	{
	  lastp = &lid[j - 1];
	  lastidx = lastp->iDeg0;
	  
	  nextfound = FALSE;
	  for (i = off + 0; i < (ndomcrvs[l] + off); i++)
	    {
	      if ( lid[i].is_scanned)
		continue;
	      thisp = &(tpatch[i]);
	      if (EQUAL_4D(thisp->coeffs[0][0],
			   lastp->coeffs[lastidx])) 
		{			/* correct order */
		  for (k = 0; k <= thisp->iDeg0; k++)
		    lid[j].coeffs[k] = thisp->coeffs[k][0];
		  lid[j].iDeg0 = thisp->iDeg0;
		  lid[i].is_scanned = TRUE;
		  nextfound = TRUE;
		  break;
		}
	      else if (EQUAL_4D(thisp->coeffs[thisp->iDeg0][0], 
				lastp->coeffs[lastidx]))
		{			/* reverse-order */
		  for (k = 0; k <= thisp->iDeg0; k++)
		    lid[j].coeffs[thisp->iDeg0 - k] =
		      thisp->coeffs[k][0];
		  lid[j].iDeg0 = thisp->iDeg0;
		  lid[i].is_scanned = TRUE;
		  nextfound = TRUE;
		  break;
		}
	    }
	  
	  /* Check whether a succesor patch was found for this j. If */
	  /* not, increase the tolerance and perform the comparison */
	  /* again */
	  
	  if ((j != (limit + off)) && !nextfound) /* next curve not */
						/* found..look again */
	    {
	      one_more_try = FALSE;
	      for (i = off + 0; i < (ndomcrvs[l] + off); i++)
		{
		  if ( lid[i].is_scanned)
		    continue;
		  thisp = &(tpatch[i]);
		  if (EQUAL_4D_PAR(thisp->coeffs[0][0],
				   lastp->coeffs[lastidx], 0.1)) 
		    {			/* correct order */
		      for (k = 0; k <= thisp->iDeg0; k++)
			lid[j].coeffs[k] = thisp->coeffs[k][0];
		      lid[j].iDeg0 = thisp->iDeg0;
		      lid[i].is_scanned = TRUE;
		      one_more_try = TRUE;
		      break;
		    }
		  else if (EQUAL_4D_PAR(thisp->coeffs[thisp->iDeg0][0], 
					lastp->coeffs[lastidx], 0.1))
		    {			/* reverse-order */
		      for (k = 0; k <= thisp->iDeg0; k++)
			lid[j].coeffs[thisp->iDeg0 - k] =
			  thisp->coeffs[k][0];
		      lid[j].iDeg0 = thisp->iDeg0;
		      lid[i].is_scanned = TRUE;
		      one_more_try = TRUE;
		  break;
		    }
		}
	      
	      if (!one_more_try)	/* fata hein!! */
		printf("generate_lids:: cannot complete a closed curve\n");
	    }
	}
      off += ndomcrvs[l];
    }
  
  /* now all the curves have been put in order (either clockwise, or */
  /* anticlockwise). Rotate the axes so that the points are in the X-Y */
  /* plane */
  
  CROSS(v0, v1, v);
  memcpy((void *) &xmat_rot1[0], (void *) v0, sizeof(vector_t)); /* X-axis */
  memcpy((void *) &xmat_rot1[4], (void *) v1, sizeof(vector_t)); /* Y-axis */
  memcpy((void *) &xmat_rot1[8], (void *) v, sizeof(vector_t)); /* Z-axis */
  xmat_rot1[12] = xmat_rot1[13] = xmat_rot1[14] = 0.0;
  xmat_rot1[3] = xmat_rot1[7] = xmat_rot1[11] = 0.0;
  xmat_rot1[15] = 1.0;


  xmax = ymax = -1.0e+10;
  xmin = ymin = 1.0e+10;
  
  if (ndomains > 1)
    limit = tpatch_idx - 4;
  else
    limit = tpatch_idx;

  for (j = 0; j < limit; j++)
    for (i = 0; i <= lid[j].iDeg0; i++)
      {
	yourpt_to_mypt(lid[j].coeffs[i], v);
	matvec_mult(xmat_rot1, v, vv);

/*	mypt_to_yourpt_no_w(vv, lid[j].coeffs[i]);
	lid[j].coeffs[i].dCoord[3] = vv[3]; */

	if (vv[0]/vv[3] > xmax)
	  xmax = vv[0]/vv[3];
	if (vv[0]/vv[3] < xmin)
	  xmin = vv[0]/vv[3];

	if (vv[1]/vv[3] > ymax)
	  ymax = vv[1]/vv[3];
	if (vv[1]/vv[3] < ymin)
	  ymin = vv[1]/vv[3];
      }
  
/*  zmax = zmin = vv[2]; */
  zmax = zmin = vv[2]/vv[3];

#ifdef BAD_UGLY_IDEAS
  /* All the points are in a plane normal to the Z-axis now. Compute */
  /* the bounding box and the scale factor now !! */

  for (j = 0; j < tpatch_idx; j++)
    for (i = 0; i <= lid[j].iDeg0; i++)
      {
	yourpt_to_mypt(lid[j].coeffs[i], v);

	if (v[0]/v[3] > xmax)
	  xmax = v[0]/v[3];
	if (v[0]/v[3] < xmin)
	  xmin = v[0]/v[3];

	if (v[1]/v[3] > ymax)
	  ymax = v[1]/v[3];
	if (v[1]/v[3] < ymin)
	  ymin = v[1]/v[3];
      }
#endif

  /* Extend the bounding-rectangle by a small amount on all the sides */
  /* to take care of the rendering-problems */
  
  xmin_tmp = xmin;
  ymin_tmp = ymin;
  xmax_tmp = xmax;
  ymax_tmp = ymax;

  xmin = (xmin_tmp + xmax_tmp)/2.0 -
    PRISM_LID_SCALE_FACTOR*(xmax_tmp - xmin_tmp)/2.0;
  xmax = (xmin_tmp + xmax_tmp)/2.0 +
    PRISM_LID_SCALE_FACTOR*(xmax_tmp - xmin_tmp)/2.0;
  ymin = (ymin_tmp + ymax_tmp)/2.0 -
    PRISM_LID_SCALE_FACTOR*(ymax_tmp - ymin_tmp)/2.0;
  ymax = (ymin_tmp + ymax_tmp)/2.0 + 
    PRISM_LID_SCALE_FACTOR*(ymax_tmp - ymin_tmp)/2.0;
  
  /* Get the four corners for the rectangular bounding box */
  corner[0][0] = xmin;
  corner[0][1] = ymin;
  corner[0][2] = zmin;
  corner[0][3] = 1.0;

  corner[1][0] = xmax;
  corner[1][1] = ymin;
  corner[1][2] = zmin;
  corner[1][3] = 1.0;

  corner[2][0] = xmax;
  corner[2][1] = ymax;
  corner[2][2] = zmin;
  corner[2][3] = 1.0;

  corner[3][0] = xmin;
  corner[3][1] = ymax;
  corner[3][2] = zmin;
  corner[3][3] = 1.0;
  
  xscale = xmax - xmin;
  yscale = ymax - ymin;

  cidx = 0;
  /* traverse each curve-segment and evaluate the Bezier curve at */
  /* fixed step-size */
  
  for (j = 0; j < limit; j++)
    {
      n = lid[j].iDeg0;
      
      /* If the curve is a straight line, just emit the endpoints */
      if (n == 1)
	{
	  yourpt_to_mypt(lid[j].coeffs[0], v);
	  matvec_mult(xmat_rot1, v, vv);
	  curve[cidx][0] = (vv[0] - xmin)/xscale;
	  curve[cidx][1] = (vv[1] - ymin)/yscale;
	  cidx++;

	  yourpt_to_mypt(lid[j].coeffs[1], v);
	  matvec_mult(xmat_rot1, v, vv);
	  curve[cidx][0] = (vv[0] - xmin)/xscale;
	  curve[cidx][1] = (vv[1] - ymin)/yscale;
	  cidx++;
	  continue;
	}
      else if (n == 2)
	coarse_factor = CONIC_COARSE_FACTOR;
      else 
	coarse_factor = HIDEG_COARSE_FACTOR;

      /* Now evaluate the bezier curve for coarse_factor * n times */

      for (howmany_pts = 0; howmany_pts <= coarse_factor*(n+1); 
	   howmany_pts++)
	{
	  t = howmany_pts*1.0/(coarse_factor*(n+1));
#if 0
      for (t = 0.0; t <= 1.0; t += 1.0/(coarse_factor*(n+1)))
	{
#endif
	  /* Copy the (n+1) points into b-array */
	  
	  for (i = 0; i <= n; i++)  
	    {
	      yourpt_to_mypt(lid[j].coeffs[i], b[i]);
	    }
      
	  for (i = 0; i < n; i++)
	    for (k = 0; k < (n - i); k++)
	      {
		b[k][0] = (1 - t)*b[k][0] + t*b[k + 1][0];
		b[k][1] = (1 - t)*b[k][1] + t*b[k + 1][1];
		b[k][2] = (1 - t)*b[k][2] + t*b[k + 1][2];
		b[k][3] = (1 - t)*b[k][3] + t*b[k + 1][3]; 		
	      }

	  /* The point evaluated is present in b[0] */

	  memcpy((void *) v, (void *) b[0], sizeof(vector_t));
	  matvec_mult(xmat_rot1, v, vv);

	  curve[cidx][0] = (vv[0]/vv[3] - xmin)/xscale;
	  curve[cidx][1] = (vv[1]/vv[3] - ymin)/yscale;
	  cidx++;
	}
	  
    } /* end-for each lid component */
  
#if 0
  if (n == 1)			/* last addn was a straight line! */
    cidx++;
#endif

  /* Perform similar bezier curve generation for the inner curve. Code */
  /* repetation actually not necessary, but easiest way out */

  if (ndomains > 1) {
    cidx_inner = 0;
    /* traverse each curve-segment and evaluate the Bezier curve at */
    /* fixed step-size */
    
    for (j = ndomcrvs[0]; j < tpatch_idx; j++)
      {
	n = lid[j].iDeg0;
	
	/* If the curve is a straight line, just emit the endpoints */
	if (n == 1)
	  {
	    yourpt_to_mypt(lid[j].coeffs[0], v);
	    matvec_mult(xmat_rot1, v, vv);
	    curve_inner[cidx_inner][0] = (vv[0] - xmin)/xscale;
	    curve_inner[cidx_inner][1] = (vv[1] - ymin)/yscale;
	    cidx_inner++;
	    
	    yourpt_to_mypt(lid[j].coeffs[1], v);
	    matvec_mult(xmat_rot1, v, vv);
	    curve_inner[cidx_inner][0] = (vv[0] - xmin)/xscale;
	    curve_inner[cidx_inner][1] = (vv[1] - ymin)/yscale;
	    continue;
	  }
	else if (n == 2)
	  coarse_factor = CONIC_COARSE_FACTOR;
	else 
	  coarse_factor = HIDEG_COARSE_FACTOR;
	
	/* Now evaluate the bezier curve for coarse_factor * n times */
	
	for (t = 0.0; t <= 1.0; t += 1.0/(coarse_factor*(n+1)))
	  {
	    /* Copy the (n+1) points into b-array */
	    
	    for (i = 0; i <= n; i++)  
	      {
		yourpt_to_mypt(lid[j].coeffs[i], b[i]);
	      }
	    
	    for (i = 0; i < n; i++)
	      for (k = 0; k < (n - i); k++)
		{
		  b[k][0] = (1 - t)*b[k][0] + t*b[k + 1][0];
		  b[k][1] = (1 - t)*b[k][1] + t*b[k + 1][1];
		  b[k][2] = (1 - t)*b[k][2] + t*b[k + 1][2];
		  b[k][3] = (1 - t)*b[k][3] + t*b[k + 1][3]; 		
		}
	    
	    /* The point evaluated is present in b[0] */
	    
	    memcpy((void *) v, (void *) b[0], sizeof(vector_t));
	    matvec_mult(xmat_rot1, v, vv);
	    
	    curve_inner[cidx_inner][0] = (vv[0]/vv[3] - xmin)/xscale;
	    curve_inner[cidx_inner][1] = (vv[1]/vv[3] - ymin)/yscale;
	    cidx_inner++;
	  }
	
      } /* end-for each lid component */
    
    if (n == 1)			/* last addn was a straight line! */
      cidx_inner++;
    
  } /* only if ndomains > 1 (for the inner domain) */

  /* No need for lid[] structure now... free all the memory */
  for (i = 0; i < tpatch_idx; i++)
    free(lid[i].coeffs);


  /* Remove the possible duplicate points. This cannot be helped */
  /* Do this only for the outer curve. This is because we are assuming */
  /* that the inner curve can only be a circle, in which case no */
  /* duplicate points will be generated (also, order is */
  /* anti-clockwise) */

  for (i = 0; i < cidx - 1; i++)
    {
      if (FP_EQUAL_TOL(curve[i][0], curve[i+1][0], 1.0e-3) && 
	  FP_EQUAL_TOL(curve[i][1], curve[i+1][1], 1.0e-3))
	{
	  for (j = i+1; j < cidx - 1; j++)
	    {
	      curve[j][0] = curve[j+1][0];
	      curve[j][1] = curve[j+1][1];
	    }
	  cidx--;
	}
    }

  /* Now determine whether the ordering of the curves is correct */

  /* prepare curve for checking its orientation */
  /* copy into other array for future use */
  for (i = 0; i < cidx; i++)
    {
      parcurve[i][0] = curve[i][0];
      parcurve[i][1] = curve[i][1];
    }
  
  parcurve[cidx - 1][0] = parcurve[0][0];
  parcurve[cidx - 1][1] = parcurve[0][1];
  
  /* project the points back to their proper positions */
  
  CROSS(v0, v1, v);
  v0[3] = v1[3] = v[3] = 0.0;

  xmat_rot1[0] = v0[0];
  xmat_rot1[4] = v0[1];
  xmat_rot1[8] = v0[2];		/* X-axis */
  xmat_rot1[12] = v0[3];

  xmat_rot1[1] = v1[0];
  xmat_rot1[5] = v1[1];
  xmat_rot1[9] = v1[2];		/* Y-axis */
  xmat_rot1[13] = v1[3];

  xmat_rot1[2] = v[0];
  xmat_rot1[6] = v[1];
  xmat_rot1[10] = v[2];		/* Z-axis */
  xmat_rot1[14] = v[3];

  xmat_rot1[3] = xmat_rot1[7] = xmat_rot1[11] = 0.0;
  xmat_rot1[15] = 1.0;
  
  for (i = 0; i < 4; i++)
    {
      matvec_mult(xmat_rot1, corner[i], vv);
      for (j = 0; j < 4; j++)
	corner[i][j] = vv[j];
    }

  /* Get the difference vector */
  v[0] = tpatch[0].coeffs[0][1].dCoord[0] -
    tpatch[0].coeffs[0][0].dCoord[0];
  v[1] = tpatch[0].coeffs[0][1].dCoord[1] -
    tpatch[0].coeffs[0][0].dCoord[1];
  v[2] = tpatch[0].coeffs[0][1].dCoord[2] -
    tpatch[0].coeffs[0][0].dCoord[2];

  if (ndomains == 1)  {		/* normal case */
  if (is_curve_anticlockwise(parcurve, cidx - 1))
    {
      if (top_is_zero)
	{
	  for (i = 0; i < cidx; i++)
	    memcpy((void *) topbot_trimcurve[0][i], (void *) curve[i], sizeof(vector_t));
	  for (i = 0; i < cidx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[1][i], (void *) curve[cidx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[1][i][0] = 1.0 - curve[cidx-i-1][0];
	    }
	}
      else
	{
	  for (i = 0; i < cidx; i++)
	    memcpy((void *) topbot_trimcurve[1][i], (void *) curve[i], sizeof(vector_t));
	  for (i = 0; i < cidx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[0][i], (void *) curve[cidx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[0][i][0] = 1.0 - curve[cidx-i-1][0];
	    }
	}
    }
  else				/* clockwise */
    {
      if (top_is_zero)
	{
	  for (i = 0; i < cidx; i++)
	    memcpy((void *) topbot_trimcurve[0][i], (void *) curve[cidx-i-1],
		   sizeof(vector_t)); 
	  for (i = 0; i < cidx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[1][i], (void *) curve[i], sizeof(vector_t));
	      topbot_trimcurve[1][i][0] = 1.0 - curve[i][0];
	    }
	}
      else
	{
	  for (i = 0; i < cidx; i++)
	    memcpy((void *) topbot_trimcurve[1][i], (void *) curve[cidx-i-1],
		   sizeof(vector_t)); 
	  for (i = 0; i < cidx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[0][i], (void *) curve[i], sizeof(vector_t));
	      topbot_trimcurve[0][i][0] = 1.0 - curve[i][0];
	    }
	}
    }

  /* Find the darned point inside each trim region!! */
  for (i = 0; i < cidx; i++)
    {
      parcurve[i][0] = topbot_trimcurve[0][i][0];
      parcurve[i][1] = topbot_trimcurve[0][i][1];
    }
  parcurve[cidx - 1][0] = parcurve[0][0];
  parcurve[cidx - 1][1] = parcurve[0][1];
  find_point_inside_polygon(cidx-1, parcurve, &topbot_inpoint[0]);

  for (i = 0; i < cidx; i++)
    {
      parcurve[i][0] = topbot_trimcurve[1][i][0];
      parcurve[i][1] = topbot_trimcurve[1][i][1];
    }
  parcurve[cidx - 1][0] = parcurve[0][0];
  parcurve[cidx - 1][1] = parcurve[0][1];
  find_point_inside_polygon(cidx-1, parcurve, &topbot_inpoint[1]);

  tp0 = get_new_patch(1, 1);
  tpp0 = &tpatch[tp0];
  tp1 = get_new_patch(1, 1);
  tpp1 = &tpatch[tp1];
  
  mypt_to_yourpt(corner[0], tpp0->coeffs[0][0]);
  mypt_to_yourpt(corner[1], tpp0->coeffs[1][0]);
  mypt_to_yourpt(corner[3], tpp0->coeffs[0][1]);
  mypt_to_yourpt(corner[2], tpp0->coeffs[1][1]);
  
  for (i  = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      corner[i][j] += v[j];

  mypt_to_yourpt(corner[0], tpp1->coeffs[1][0]);
  mypt_to_yourpt(corner[1], tpp1->coeffs[0][0]);
  mypt_to_yourpt(corner[3], tpp1->coeffs[1][1]);
  mypt_to_yourpt(corner[2], tpp1->coeffs[0][1]);

  *nl1 = cidx;

  } /* only one domain */
  else
    {				/* (ndomains > 1) */
      if (is_curve_anticlockwise(parcurve, cidx - 1))
	{
	  
	  /* Now consider the horizontal segment passing the circle. Find */
	  /* where it intersects the outer curve., And construct the two */
	  /* halves of the curves appropriately */
	  get_two_curves(curve, curve_inner, cidx, cidx_inner, ioc1,
			 ioc2, &c1idx, &c2idx, TRUE, FALSE);
	}
      else
	get_two_curves(curve, curve_inner, cidx, cidx_inner, ioc1,
		       ioc2, &c1idx, &c2idx, FALSE, FALSE);	
      
      /* At this stage, both the curves are anti-clockwise */

      if (top_is_zero)
	{
	  for (i = 0; i < c1idx; i++)
	    memcpy((void *) topbot_trimcurve[0][i], (void *) ioc1[i],
		   sizeof(vector_t));
	  for (i = 0; i < c2idx; i++)
	    memcpy((void *) topbot_trimcurve[2][i], (void *) ioc2[i],
		   sizeof(vector_t));
	  
	  for (i = 0; i < c1idx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[1][i], (void *) ioc1[c1idx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[1][i][0] = 1.0 - ioc1[c1idx-i-1][0];
	    }
	  for (i = 0; i < c2idx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[3][i], (void *) ioc2[c2idx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[3][i][0] = 1.0 - ioc2[c2idx-i-1][0];
	    }
	}
      else
	{
	  for (i = 0; i < c1idx; i++)
	    memcpy((void *) topbot_trimcurve[1][i], (void *) ioc1[i],
		   sizeof(vector_t));
	  for (i = 0; i < c2idx; i++)
	    memcpy((void *) topbot_trimcurve[3][i], (void *) ioc2[i],
		   sizeof(vector_t));
	  
	  for (i = 0; i < c1idx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[0][i], (void *) ioc1[c1idx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[0][i][0] = 1.0 - ioc1[c1idx-i-1][0];
	    }
	  for (i = 0; i < c2idx; i++)
	    {
	      memcpy((void *) topbot_trimcurve[2][i], (void *) ioc2[c2idx-i-1],
		     sizeof(vector_t));
	      topbot_trimcurve[2][i][0] = 1.0 - ioc2[c2idx-i-1][0];
	    }
	}

#ifdef DEVP
      else				/* anticlockwise */
	{
	  get_two_curves(curve, curve_inner, cidx, cidx_inner, ioc1,
			 ioc2, &c1idx, &c2idx, FALSE, FALSE);
	  
	  if (top_is_zero)
	    {
	      for (i = 0; i < c1idx; i++)
		memcpy((void *) topbot_trimcurve[0][i], (void *) ioc1[c1idx-i-1],
		       sizeof(vector_t)); 
	      for (i = 0; i < c2idx; i++)
		memcpy((void *) topbot_trimcurve[2][i], (void *) ioc2[c2idx-i-1],
		       sizeof(vector_t)); 

	      for (i = 0; i < c1idx; i++)
		{
		  memcpy((void *) topbot_trimcurve[1][i], (void *) ioc1[i], sizeof(vector_t));
		  topbot_trimcurve[1][i][0] = 1.0 - ioc1[i][0];
		}	
	      for (i = 0; i < c2idx; i++)
		{
		  memcpy((void *) topbot_trimcurve[3][i], (void *) ioc2[i], sizeof(vector_t));
		  topbot_trimcurve[3][i][0] = 1.0 - ioc2[i][0];
		}

	    }
	  else
	    {
 	      for (i = 0; i < c1idx; i++)
		memcpy((void *) topbot_trimcurve[1][i], (void *) ioc1[c1idx-i-1],
		   sizeof(vector_t)); 
 	      for (i = 0; i < c2idx; i++)
		memcpy((void *) topbot_trimcurve[3][i], (void *) ioc2[c2idx-i-1],
		   sizeof(vector_t)); 
	      for (i = 0; i < c1idx; i++)
		{
		  memcpy((void *) topbot_trimcurve[0][i], (void *) ioc1[i], sizeof(vector_t));
		  topbot_trimcurve[0][i][0] = 1.0 - ioc1[i][0];
		}
	      for (i = 0; i < c2idx; i++)
		{
		  memcpy((void *) topbot_trimcurve[2][i], (void *) ioc2[i], sizeof(vector_t));
		  topbot_trimcurve[2][i][0] = 1.0 - ioc2[i][0];
		}
	    }
	}
      
#endif

      /* Find the darned point inside each trim region!! */
      for (i = 0; i < c1idx; i++)
	{
	  parcurve[i][0] = topbot_trimcurve[0][i][0];
	  parcurve[i][1] = topbot_trimcurve[0][i][1];
	}
      parcurve[c1idx][0] = parcurve[0][0];
      parcurve[c1idx][1] = parcurve[0][1];
      find_point_inside_polygon(c1idx, parcurve, &topbot_inpoint[0]);
      
      for (i = 0; i < c1idx; i++)
	{
	  parcurve[i][0] = topbot_trimcurve[1][i][0];
	  parcurve[i][1] = topbot_trimcurve[1][i][1];
	}
      parcurve[c1idx][0] = parcurve[0][0];
      parcurve[c1idx][1] = parcurve[0][1];
      find_point_inside_polygon(c1idx, parcurve, &topbot_inpoint[1]);

      for (i = 0; i < c2idx; i++)
	{
	  parcurve[i][0] = topbot_trimcurve[2][i][0];
	  parcurve[i][1] = topbot_trimcurve[2][i][1];
	}
      parcurve[c2idx][0] = parcurve[0][0];
      parcurve[c2idx][1] = parcurve[0][1];
      find_point_inside_polygon(c2idx, parcurve, &topbot_inpoint[2]);
      
      for (i = 0; i < c2idx; i++)
	{
	  parcurve[i][0] = topbot_trimcurve[3][i][0];
	  parcurve[i][1] = topbot_trimcurve[3][i][1];
	}
      parcurve[c2idx][0] = parcurve[0][0];
      parcurve[c2idx][1] = parcurve[0][1];
      find_point_inside_polygon(c2idx, parcurve, &topbot_inpoint[3]);
      
      tp0 = get_new_patch(1, 1);
      tpp0 = &tpatch[tp0];
      tp1 = get_new_patch(1, 1);
      tpp1 = &tpatch[tp1];
      
      mypt_to_yourpt(corner[0], tpp0->coeffs[0][0]);
      mypt_to_yourpt(corner[1], tpp0->coeffs[1][0]);
      mypt_to_yourpt(corner[3], tpp0->coeffs[0][1]);
      mypt_to_yourpt(corner[2], tpp0->coeffs[1][1]);
      
      for (i  = 0; i < 4; i++)
	for (j = 0; j < 4; j++)
	  corner[i][j] += v[j];
      
      mypt_to_yourpt(corner[0], tpp1->coeffs[1][0]);
      mypt_to_yourpt(corner[1], tpp1->coeffs[0][0]);
      mypt_to_yourpt(corner[3], tpp1->coeffs[1][1]);
      mypt_to_yourpt(corner[2], tpp1->coeffs[0][1]);

      *nl1 = c1idx;
      *nl2 = c2idx;
    }
  
  return cidx;
}
  
