#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include "defs.h"
#include "linalg.h"

#if 0
#define	NUM_SPH_PATCHES		8
#define	SPH_SDEGREE		2	
#define	SPH_TDEGREE		2	
#else
#define	NUM_SPH_PATCHES		6
#define	SPH_SDEGREE		3	
#define	SPH_TDEGREE		3	
#endif

EXTERN_LINALG

extern void normalise();
extern void allocate_tensor_patch();
extern void XformPatch();

void getXformMatSph(center, Xform)
POINT4D center;
double **Xform;
{
  double px, py, pz;

  px = center.dCoord[X];
  py = center.dCoord[Y];
  pz = center.dCoord[Z];

  Xform[0][0] = 1.0;
  Xform[0][1] = 0.0;
  Xform[0][2] = 0.0;
  Xform[0][3] = 0.0;
  
  Xform[1][0] = 0.0;
  Xform[1][1] = 1.0;
  Xform[1][2] = 0.0;
  Xform[1][3] = 0.0;
  
  Xform[2][0] = 0.0;
  Xform[2][1] = 0.0;
  Xform[2][2] = 1.0;
  Xform[2][3] = 0.0;
  
  Xform[3][0] = px;
  Xform[3][1] = py;
  Xform[3][2] = pz;
  Xform[3][3] = 1.0;
}

void getRotationMatSph(Xform, which_axis)
double **Xform;
int which_axis;
{
  struct timeval tval;
  struct timezone tzone;
  double theta, ctheta, stheta;

  gettimeofday(&tval, &tzone);
  srand48(tval.tv_sec);
 
  theta = drand48()*M_PI/2.0;
  ctheta = cos(theta);
  stheta = sin(theta);

  if (which_axis == 1)		/* rotation about X-axis */
    {
      Xform[0][0] = 1.0;
      Xform[0][1] = 0.0;
      Xform[0][2] = 0.0;
      Xform[0][3] = 0.0;
      
      Xform[1][0] = 0.0;
      Xform[1][1] = ctheta;
      Xform[1][2] = -stheta;
      Xform[1][3] = 0.0;
      
      Xform[2][0] = 0.0;
      Xform[2][1] = stheta;
      Xform[2][2] = ctheta;
      Xform[2][3] = 0.0;
      
      Xform[3][0] = 0.0;
      Xform[3][1] = 0.0;
      Xform[3][2] = 0.0;
      Xform[3][3] = 1.0;
    }
  else if (which_axis == 2)	/* Y-axis */
    {
      Xform[0][0] = ctheta;
      Xform[0][1] = 0.0;
      Xform[0][2] = stheta;
      Xform[0][3] = 0.0;
      
      Xform[1][0] = 0.0;
      Xform[1][1] = 1.0;
      Xform[1][2] = 0.0;
      Xform[1][3] = 0.0;
      
      Xform[2][0] = -stheta;
      Xform[2][1] = 0.0;
      Xform[2][2] = ctheta;
      Xform[2][3] = 0.0;
      
      Xform[3][0] = 0.0;
      Xform[3][1] = 0.0;
      Xform[3][2] = 0.0;
      Xform[3][3] = 1.0;
    }
  else				/* rotation about Z-axis */
    {
      Xform[0][0] = ctheta;
      Xform[0][1] = -stheta;
      Xform[0][2] = 0.0;
      Xform[0][3] = 0.0;
      
      Xform[1][0] = stheta;
      Xform[1][1] = ctheta;
      Xform[1][2] = 0.0;
      Xform[1][3] = 0.0;
      
      Xform[2][0] = 0.0;
      Xform[2][1] = 0.0;
      Xform[2][2] = 1.0;
      Xform[2][3] = 0.0;
      
      Xform[3][0] = 0.0;
      Xform[3][1] = 0.0;
      Xform[3][2] = 0.0;
      Xform[3][3] = 1.0;
    }
}

#if 0
void GetNormSphControlPts(tp, sph_radius)
     TENSOR_PATCH *tp;
     double sph_radius;
{
  double weight_factor, sq_wt_factor, r;

  weight_factor = sqrt(2.0) / 2.0;  
  sq_wt_factor = 0.5;
  r = sph_radius;

  tp[0].coeffs[0][0].dCoord[X] = 0.0;
  tp[0].coeffs[0][0].dCoord[Y] = 0.0;
  tp[0].coeffs[0][0].dCoord[Z] = r;
  tp[0].coeffs[0][0].dCoord[W] = 1.0;

  tp[0].coeffs[1][0].dCoord[X] = r*weight_factor;
  tp[0].coeffs[1][0].dCoord[Y] = 0.0;
  tp[0].coeffs[1][0].dCoord[Z] = r*weight_factor;
  tp[0].coeffs[1][0].dCoord[W] = weight_factor;

  tp[0].coeffs[2][0].dCoord[X] = r;
  tp[0].coeffs[2][0].dCoord[Y] = 0.0;
  tp[0].coeffs[2][0].dCoord[Z] = 0.0;
  tp[0].coeffs[2][0].dCoord[W] = 1.0;

  tp[0].coeffs[0][1].dCoord[X] = 0.0;
  tp[0].coeffs[0][1].dCoord[Y] = 0.0;
  tp[0].coeffs[0][1].dCoord[Z] = r*weight_factor;
  tp[0].coeffs[0][1].dCoord[W] = weight_factor;

  tp[0].coeffs[1][1].dCoord[X] = r*sq_wt_factor;
  tp[0].coeffs[1][1].dCoord[Y] = r*sq_wt_factor;
  tp[0].coeffs[1][1].dCoord[Z] = r*sq_wt_factor;
  tp[0].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[0].coeffs[2][1].dCoord[X] = r*weight_factor;
  tp[0].coeffs[2][1].dCoord[Y] = r*weight_factor;
  tp[0].coeffs[2][1].dCoord[Z] = 0.0;
  tp[0].coeffs[2][1].dCoord[W] = weight_factor;

  tp[0].coeffs[0][2].dCoord[X] = 0.0;
  tp[0].coeffs[0][2].dCoord[Y] = 0.0;
  tp[0].coeffs[0][2].dCoord[Z] = r;
  tp[0].coeffs[0][2].dCoord[W] = 1.0;

  tp[0].coeffs[1][2].dCoord[X] = 0.0;
  tp[0].coeffs[1][2].dCoord[Y] = r*weight_factor;
  tp[0].coeffs[1][2].dCoord[Z] = r*weight_factor;
  tp[0].coeffs[1][2].dCoord[W] = weight_factor;

  tp[0].coeffs[2][2].dCoord[X] = 0.0;
  tp[0].coeffs[2][2].dCoord[Y] = r;
  tp[0].coeffs[2][2].dCoord[Z] = 0.0;
  tp[0].coeffs[2][2].dCoord[W] = 1.0;

  tp[1].coeffs[0][0].dCoord[X] = 0.0;
  tp[1].coeffs[0][0].dCoord[Y] = 0.0;
  tp[1].coeffs[0][0].dCoord[Z] = r;
  tp[1].coeffs[0][0].dCoord[W] = 1.0;

  tp[1].coeffs[1][0].dCoord[X] = 0.0;
  tp[1].coeffs[1][0].dCoord[Y] = r*weight_factor;
  tp[1].coeffs[1][0].dCoord[Z] = r*weight_factor;
  tp[1].coeffs[1][0].dCoord[W] = weight_factor;

  tp[1].coeffs[2][0].dCoord[X] = 0.0;
  tp[1].coeffs[2][0].dCoord[Y] = r;
  tp[1].coeffs[2][0].dCoord[Z] = 0.0;
  tp[1].coeffs[2][0].dCoord[W] = 1.0;

  tp[1].coeffs[0][1].dCoord[X] = 0.0;
  tp[1].coeffs[0][1].dCoord[Y] = 0.0;
  tp[1].coeffs[0][1].dCoord[Z] = r*weight_factor;
  tp[1].coeffs[0][1].dCoord[W] = weight_factor;

  tp[1].coeffs[1][1].dCoord[X] = -r*sq_wt_factor;
  tp[1].coeffs[1][1].dCoord[Y] = r*sq_wt_factor;
  tp[1].coeffs[1][1].dCoord[Z] = r*sq_wt_factor;
  tp[1].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[1].coeffs[2][1].dCoord[X] = -r*weight_factor;
  tp[1].coeffs[2][1].dCoord[Y] = r*weight_factor;
  tp[1].coeffs[2][1].dCoord[Z] = 0.0;
  tp[1].coeffs[2][1].dCoord[W] = weight_factor;

  tp[1].coeffs[0][2].dCoord[X] = 0.0;
  tp[1].coeffs[0][2].dCoord[Y] = 0.0;
  tp[1].coeffs[0][2].dCoord[Z] = r;
  tp[1].coeffs[0][2].dCoord[W] = 1.0;

  tp[1].coeffs[1][2].dCoord[X] = -r*weight_factor;
  tp[1].coeffs[1][2].dCoord[Y] = 0.0;
  tp[1].coeffs[1][2].dCoord[Z] = r*weight_factor;
  tp[1].coeffs[1][2].dCoord[W] = weight_factor;

  tp[1].coeffs[2][2].dCoord[X] = -r;
  tp[1].coeffs[2][2].dCoord[Y] = 0.0;
  tp[1].coeffs[2][2].dCoord[Z] = 0.0;
  tp[1].coeffs[2][2].dCoord[W] = 1.0;

  tp[2].coeffs[0][0].dCoord[X] = 0.0;
  tp[2].coeffs[0][0].dCoord[Y] = 0.0;
  tp[2].coeffs[0][0].dCoord[Z] = r;
  tp[2].coeffs[0][0].dCoord[W] = 1.0;

  tp[2].coeffs[1][0].dCoord[X] = -r*weight_factor;
  tp[2].coeffs[1][0].dCoord[Y] = 0.0;
  tp[2].coeffs[1][0].dCoord[Z] = r*weight_factor;
  tp[2].coeffs[1][0].dCoord[W] = weight_factor;

  tp[2].coeffs[2][0].dCoord[X] = -r;
  tp[2].coeffs[2][0].dCoord[Y] = 0.0;
  tp[2].coeffs[2][0].dCoord[Z] = 0.0;
  tp[2].coeffs[2][0].dCoord[W] = 1.0;

  tp[2].coeffs[0][1].dCoord[X] = 0.0;
  tp[2].coeffs[0][1].dCoord[Y] = 0.0;
  tp[2].coeffs[0][1].dCoord[Z] = r*weight_factor;
  tp[2].coeffs[0][1].dCoord[W] = weight_factor;

  tp[2].coeffs[1][1].dCoord[X] = -r*sq_wt_factor;
  tp[2].coeffs[1][1].dCoord[Y] = -r*sq_wt_factor;
  tp[2].coeffs[1][1].dCoord[Z] = r*sq_wt_factor;
  tp[2].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[2].coeffs[2][1].dCoord[X] = -r*weight_factor;
  tp[2].coeffs[2][1].dCoord[Y] = -r*weight_factor;
  tp[2].coeffs[2][1].dCoord[Z] = 0.0;
  tp[2].coeffs[2][1].dCoord[W] = weight_factor;

  tp[2].coeffs[0][2].dCoord[X] = 0.0;
  tp[2].coeffs[0][2].dCoord[Y] = 0.0;
  tp[2].coeffs[0][2].dCoord[Z] = r;
  tp[2].coeffs[0][2].dCoord[W] = 1.0;

  tp[2].coeffs[1][2].dCoord[X] = 0.0;
  tp[2].coeffs[1][2].dCoord[Y] = -r*weight_factor;
  tp[2].coeffs[1][2].dCoord[Z] = r*weight_factor;
  tp[2].coeffs[1][2].dCoord[W] = weight_factor;

  tp[2].coeffs[2][2].dCoord[X] = 0.0;
  tp[2].coeffs[2][2].dCoord[Y] = -r;
  tp[2].coeffs[2][2].dCoord[Z] = 0.0;
  tp[2].coeffs[2][2].dCoord[W] = 1.0;

  tp[3].coeffs[0][0].dCoord[X] = 0.0;
  tp[3].coeffs[0][0].dCoord[Y] = 0.0;
  tp[3].coeffs[0][0].dCoord[Z] = r;
  tp[3].coeffs[0][0].dCoord[W] = 1.0;

  tp[3].coeffs[1][0].dCoord[X] = 0.0;
  tp[3].coeffs[1][0].dCoord[Y] = -r*weight_factor;
  tp[3].coeffs[1][0].dCoord[Z] = r*weight_factor;
  tp[3].coeffs[1][0].dCoord[W] = weight_factor;

  tp[3].coeffs[2][0].dCoord[X] = 0.0;
  tp[3].coeffs[2][0].dCoord[Y] = -r;
  tp[3].coeffs[2][0].dCoord[Z] = 0.0;
  tp[3].coeffs[2][0].dCoord[W] = 1.0;

  tp[3].coeffs[0][1].dCoord[X] = 0.0;
  tp[3].coeffs[0][1].dCoord[Y] = 0.0;
  tp[3].coeffs[0][1].dCoord[Z] = r*weight_factor;
  tp[3].coeffs[0][1].dCoord[W] = weight_factor;

  tp[3].coeffs[1][1].dCoord[X] = r*sq_wt_factor;
  tp[3].coeffs[1][1].dCoord[Y] = -r*sq_wt_factor;
  tp[3].coeffs[1][1].dCoord[Z] = r*sq_wt_factor;
  tp[3].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[3].coeffs[2][1].dCoord[X] = r*weight_factor;
  tp[3].coeffs[2][1].dCoord[Y] = -r*weight_factor;
  tp[3].coeffs[2][1].dCoord[Z] = 0.0;
  tp[3].coeffs[2][1].dCoord[W] = weight_factor;

  tp[3].coeffs[0][2].dCoord[X] = 0.0;
  tp[3].coeffs[0][2].dCoord[Y] = 0.0;
  tp[3].coeffs[0][2].dCoord[Z] = r;
  tp[3].coeffs[0][2].dCoord[W] = 1.0;

  tp[3].coeffs[1][2].dCoord[X] = r*weight_factor;
  tp[3].coeffs[1][2].dCoord[Y] = 0.0;
  tp[3].coeffs[1][2].dCoord[Z] = r*weight_factor;
  tp[3].coeffs[1][2].dCoord[W] = weight_factor;

  tp[3].coeffs[2][2].dCoord[X] = r;
  tp[3].coeffs[2][2].dCoord[Y] = 0.0;
  tp[3].coeffs[2][2].dCoord[Z] = 0.0;
  tp[3].coeffs[2][2].dCoord[W] = 1.0;

  tp[4].coeffs[0][0].dCoord[X] = 0.0;
  tp[4].coeffs[0][0].dCoord[Y] = 0.0;
  tp[4].coeffs[0][0].dCoord[Z] = -r;
  tp[4].coeffs[0][0].dCoord[W] = 1.0;

  tp[4].coeffs[0][1].dCoord[X] = r*weight_factor;
  tp[4].coeffs[0][1].dCoord[Y] = 0.0;
  tp[4].coeffs[0][1].dCoord[Z] = -r*weight_factor;
  tp[4].coeffs[0][1].dCoord[W] = weight_factor;

  tp[4].coeffs[0][2].dCoord[X] = r;
  tp[4].coeffs[0][2].dCoord[Y] = 0.0;
  tp[4].coeffs[0][2].dCoord[Z] = 0.0;
  tp[4].coeffs[0][2].dCoord[W] = 1.0;

  tp[4].coeffs[1][0].dCoord[X] = 0.0;
  tp[4].coeffs[1][0].dCoord[Y] = 0.0;
  tp[4].coeffs[1][0].dCoord[Z] = -r*weight_factor;
  tp[4].coeffs[1][0].dCoord[W] = weight_factor;

  tp[4].coeffs[1][1].dCoord[X] = r*sq_wt_factor;
  tp[4].coeffs[1][1].dCoord[Y] = r*sq_wt_factor;
  tp[4].coeffs[1][1].dCoord[Z] = -r*sq_wt_factor;
  tp[4].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[4].coeffs[1][2].dCoord[X] = r*weight_factor;
  tp[4].coeffs[1][2].dCoord[Y] = r*weight_factor;
  tp[4].coeffs[1][2].dCoord[Z] = 0.0;
  tp[4].coeffs[1][2].dCoord[W] = weight_factor;

  tp[4].coeffs[2][0].dCoord[X] = 0.0;
  tp[4].coeffs[2][0].dCoord[Y] = 0.0;
  tp[4].coeffs[2][0].dCoord[Z] = -r;
  tp[4].coeffs[2][0].dCoord[W] = 1.0;

  tp[4].coeffs[2][1].dCoord[X] = 0.0;
  tp[4].coeffs[2][1].dCoord[Y] = r*weight_factor;
  tp[4].coeffs[2][1].dCoord[Z] = -r*weight_factor;
  tp[4].coeffs[2][1].dCoord[W] = weight_factor;

  tp[4].coeffs[2][2].dCoord[X] = 0.0;
  tp[4].coeffs[2][2].dCoord[Y] = r;
  tp[4].coeffs[2][2].dCoord[Z] = 0.0;
  tp[4].coeffs[2][2].dCoord[W] = 1.0;

  tp[5].coeffs[0][0].dCoord[X] = 0.0;
  tp[5].coeffs[0][0].dCoord[Y] = 0.0;
  tp[5].coeffs[0][0].dCoord[Z] = -r;
  tp[5].coeffs[0][0].dCoord[W] = 1.0;

  tp[5].coeffs[0][1].dCoord[X] = 0.0;
  tp[5].coeffs[0][1].dCoord[Y] = r*weight_factor;
  tp[5].coeffs[0][1].dCoord[Z] = -r*weight_factor;
  tp[5].coeffs[0][1].dCoord[W] = weight_factor;

  tp[5].coeffs[0][2].dCoord[X] = 0.0;
  tp[5].coeffs[0][2].dCoord[Y] = r;
  tp[5].coeffs[0][2].dCoord[Z] = 0.0;
  tp[5].coeffs[0][2].dCoord[W] = 1.0;

  tp[5].coeffs[1][0].dCoord[X] = 0.0;
  tp[5].coeffs[1][0].dCoord[Y] = 0.0;
  tp[5].coeffs[1][0].dCoord[Z] = -r*weight_factor;
  tp[5].coeffs[1][0].dCoord[W] = weight_factor;

  tp[5].coeffs[1][1].dCoord[X] = -r*sq_wt_factor;
  tp[5].coeffs[1][1].dCoord[Y] = r*sq_wt_factor;
  tp[5].coeffs[1][1].dCoord[Z] = -r*sq_wt_factor;
  tp[5].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[5].coeffs[1][2].dCoord[X] = -r*weight_factor;
  tp[5].coeffs[1][2].dCoord[Y] = r*weight_factor;
  tp[5].coeffs[1][2].dCoord[Z] = 0.0;
  tp[5].coeffs[1][2].dCoord[W] = weight_factor;

  tp[5].coeffs[2][0].dCoord[X] = 0.0;
  tp[5].coeffs[2][0].dCoord[Y] = 0.0;
  tp[5].coeffs[2][0].dCoord[Z] = -r;
  tp[5].coeffs[2][0].dCoord[W] = 1.0;

  tp[5].coeffs[2][1].dCoord[X] = -r*weight_factor;
  tp[5].coeffs[2][1].dCoord[Y] = 0.0;
  tp[5].coeffs[2][1].dCoord[Z] = -r*weight_factor;
  tp[5].coeffs[2][1].dCoord[W] = weight_factor;

  tp[5].coeffs[2][2].dCoord[X] = -r;
  tp[5].coeffs[2][2].dCoord[Y] = 0.0;
  tp[5].coeffs[2][2].dCoord[Z] = 0.0;
  tp[5].coeffs[2][2].dCoord[W] = 1.0;

  tp[6].coeffs[0][0].dCoord[X] = 0.0;
  tp[6].coeffs[0][0].dCoord[Y] = 0.0;
  tp[6].coeffs[0][0].dCoord[Z] = -r;
  tp[6].coeffs[0][0].dCoord[W] = 1.0;

  tp[6].coeffs[0][1].dCoord[X] = -r*weight_factor;
  tp[6].coeffs[0][1].dCoord[Y] = 0.0;
  tp[6].coeffs[0][1].dCoord[Z] = -r*weight_factor;
  tp[6].coeffs[0][1].dCoord[W] = weight_factor;

  tp[6].coeffs[0][2].dCoord[X] = -r;
  tp[6].coeffs[0][2].dCoord[Y] = 0.0;
  tp[6].coeffs[0][2].dCoord[Z] = 0.0;
  tp[6].coeffs[0][2].dCoord[W] = 1.0;

  tp[6].coeffs[1][0].dCoord[X] = 0.0;
  tp[6].coeffs[1][0].dCoord[Y] = 0.0;
  tp[6].coeffs[1][0].dCoord[Z] = -r*weight_factor;
  tp[6].coeffs[1][0].dCoord[W] = weight_factor;

  tp[6].coeffs[1][1].dCoord[X] = -r*sq_wt_factor;
  tp[6].coeffs[1][1].dCoord[Y] = -r*sq_wt_factor;
  tp[6].coeffs[1][1].dCoord[Z] = -r*sq_wt_factor;
  tp[6].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[6].coeffs[1][2].dCoord[X] = -r*weight_factor;
  tp[6].coeffs[1][2].dCoord[Y] = -r*weight_factor;
  tp[6].coeffs[1][2].dCoord[Z] = 0.0;
  tp[6].coeffs[1][2].dCoord[W] = weight_factor;

  tp[6].coeffs[2][0].dCoord[X] = 0.0;
  tp[6].coeffs[2][0].dCoord[Y] = 0.0;
  tp[6].coeffs[2][0].dCoord[Z] = -r;
  tp[6].coeffs[2][0].dCoord[W] = 1.0;

  tp[6].coeffs[2][1].dCoord[X] = 0.0;
  tp[6].coeffs[2][1].dCoord[Y] = -r*weight_factor;
  tp[6].coeffs[2][1].dCoord[Z] = -r*weight_factor;
  tp[6].coeffs[2][1].dCoord[W] = weight_factor;

  tp[6].coeffs[2][2].dCoord[X] = 0.0;
  tp[6].coeffs[2][2].dCoord[Y] = -r;
  tp[6].coeffs[2][2].dCoord[Z] = 0.0;
  tp[6].coeffs[2][2].dCoord[W] = 1.0;

  tp[7].coeffs[0][0].dCoord[X] = 0.0;
  tp[7].coeffs[0][0].dCoord[Y] = 0.0;
  tp[7].coeffs[0][0].dCoord[Z] = -r;
  tp[7].coeffs[0][0].dCoord[W] = 1.0;

  tp[7].coeffs[0][1].dCoord[X] = 0.0;
  tp[7].coeffs[0][1].dCoord[Y] = -r*weight_factor;
  tp[7].coeffs[0][1].dCoord[Z] = -r*weight_factor;
  tp[7].coeffs[0][1].dCoord[W] = weight_factor;

  tp[7].coeffs[0][2].dCoord[X] = 0.0;
  tp[7].coeffs[0][2].dCoord[Y] = -r;
  tp[7].coeffs[0][2].dCoord[Z] = 0.0;
  tp[7].coeffs[0][2].dCoord[W] = 1.0;

  tp[7].coeffs[1][0].dCoord[X] = 0.0;
  tp[7].coeffs[1][0].dCoord[Y] = 0.0;
  tp[7].coeffs[1][0].dCoord[Z] = -r*weight_factor;
  tp[7].coeffs[1][0].dCoord[W] = weight_factor;

  tp[7].coeffs[1][1].dCoord[X] = r*sq_wt_factor;
  tp[7].coeffs[1][1].dCoord[Y] = -r*sq_wt_factor;
  tp[7].coeffs[1][1].dCoord[Z] = -r*sq_wt_factor;
  tp[7].coeffs[1][1].dCoord[W] = sq_wt_factor;

  tp[7].coeffs[1][2].dCoord[X] = r*weight_factor;
  tp[7].coeffs[1][2].dCoord[Y] = -r*weight_factor;
  tp[7].coeffs[1][2].dCoord[Z] = 0.0;
  tp[7].coeffs[1][2].dCoord[W] = weight_factor;

  tp[7].coeffs[2][0].dCoord[X] = 0.0;
  tp[7].coeffs[2][0].dCoord[Y] = 0.0;
  tp[7].coeffs[2][0].dCoord[Z] = -r;
  tp[7].coeffs[2][0].dCoord[W] = 1.0;

  tp[7].coeffs[2][1].dCoord[X] = r*weight_factor;
  tp[7].coeffs[2][1].dCoord[Y] = 0.0;
  tp[7].coeffs[2][1].dCoord[Z] = -r*weight_factor;
  tp[7].coeffs[2][1].dCoord[W] = weight_factor;

  tp[7].coeffs[2][2].dCoord[X] = r;
  tp[7].coeffs[2][2].dCoord[Y] = 0.0;
  tp[7].coeffs[2][2].dCoord[Z] = 0.0;
  tp[7].coeffs[2][2].dCoord[W] = 1.0;
}
#else
void GetNormSphControlPts(tp, sph_radius)
TENSOR_PATCH *tp;
double sph_radius;
{
  double weight_factor, sq_wt_factor, r;
  int i, j, k;

  weight_factor = sqrt(2.0) / 2.0;  
  sq_wt_factor = 0.5;
  r = sph_radius/sqrt(3.0);

  for (i=0; i<NUM_SPH_PATCHES; i++)
    for (j=0; j<=SPH_SDEGREE; j++)
      for (k=0; k<=SPH_TDEGREE; k++)
	tp[i].coeffs[j][k].dCoord[W] = 1.0;

/* top */
  tp[0].coeffs[0][3].dCoord[X] = -r;
  tp[0].coeffs[0][3].dCoord[Y] = -r;
  tp[0].coeffs[0][3].dCoord[Z] = r;

  tp[0].coeffs[0][2].dCoord[X] = -0.40068*r;
  tp[0].coeffs[0][2].dCoord[Y] = -1.29966*r;
  tp[0].coeffs[0][2].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[0][1].dCoord[X] = 0.40068*r;
  tp[0].coeffs[0][1].dCoord[Y] = -1.29966*r;
  tp[0].coeffs[0][1].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[0][0].dCoord[X] = r;
  tp[0].coeffs[0][0].dCoord[Y] = -r;
  tp[0].coeffs[0][0].dCoord[Z] = r;

  tp[0].coeffs[1][3].dCoord[X] = -1.29966*r;
  tp[0].coeffs[1][3].dCoord[Y] = -0.40068*r;
  tp[0].coeffs[1][3].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[1][2].dCoord[X] = -0.597556*r; 
  tp[0].coeffs[1][2].dCoord[Y] = -0.597556*r; 
  tp[0].coeffs[1][2].dCoord[Z] = 2.10165*r; 

  tp[0].coeffs[1][1].dCoord[X] = 0.597556*r; 
  tp[0].coeffs[1][1].dCoord[Y] = -0.597556*r; 
  tp[0].coeffs[1][1].dCoord[Z] = 2.10165*r;

  tp[0].coeffs[1][0].dCoord[X] = 1.29966*r; 
  tp[0].coeffs[1][0].dCoord[Y] = -0.40068*r; 
  tp[0].coeffs[1][0].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[2][3].dCoord[X] = -1.29966*r;
  tp[0].coeffs[2][3].dCoord[Y] = 0.40068*r;
  tp[0].coeffs[2][3].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[2][2].dCoord[X] = -0.597556*r;
  tp[0].coeffs[2][2].dCoord[Y] = 0.597556*r;
  tp[0].coeffs[2][2].dCoord[Z] = 2.10165*r;

  tp[0].coeffs[2][1].dCoord[X] = 0.597556*r;
  tp[0].coeffs[2][1].dCoord[Y] = 0.597556*r;
  tp[0].coeffs[2][1].dCoord[Z] = 2.10165*r;

  tp[0].coeffs[2][0].dCoord[X] = 1.29966*r;
  tp[0].coeffs[2][0].dCoord[Y] = 0.40068*r;
  tp[0].coeffs[2][0].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[3][3].dCoord[X] = -r;
  tp[0].coeffs[3][3].dCoord[Y] = r;
  tp[0].coeffs[3][3].dCoord[Z] = r;

  tp[0].coeffs[3][2].dCoord[X] = -0.40068*r;
  tp[0].coeffs[3][2].dCoord[Y] = 1.29966*r;
  tp[0].coeffs[3][2].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[3][1].dCoord[X] = 0.40068*r;
  tp[0].coeffs[3][1].dCoord[Y] = 1.29966*r;
  tp[0].coeffs[3][1].dCoord[Z] = 1.29966*r;

  tp[0].coeffs[3][0].dCoord[X] = r;
  tp[0].coeffs[3][0].dCoord[Y] = r;
  tp[0].coeffs[3][0].dCoord[Z] = r;

/* back */
  tp[1].coeffs[0][3].dCoord[X] = -r;
  tp[1].coeffs[0][3].dCoord[Y] = r;
  tp[1].coeffs[0][3].dCoord[Z] = r;

  tp[1].coeffs[0][2].dCoord[X] = -0.40068*r;
  tp[1].coeffs[0][2].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[0][2].dCoord[Z] = 1.29966*r;

  tp[1].coeffs[0][1].dCoord[X] = 0.40068*r;
  tp[1].coeffs[0][1].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[0][1].dCoord[Z] = 1.29966*r;

  tp[1].coeffs[0][0].dCoord[X] = r;
  tp[1].coeffs[0][0].dCoord[Y] = r;
  tp[1].coeffs[0][0].dCoord[Z] = r;

  tp[1].coeffs[1][3].dCoord[X] = -1.29966*r;
  tp[1].coeffs[1][3].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[1][3].dCoord[Z] = 0.40068*r;

  tp[1].coeffs[1][2].dCoord[X] = -0.597556*r; 
  tp[1].coeffs[1][2].dCoord[Y] = 2.10165*r; 
  tp[1].coeffs[1][2].dCoord[Z] = 0.597556*r; 

  tp[1].coeffs[1][1].dCoord[X] = 0.597556*r; 
  tp[1].coeffs[1][1].dCoord[Y] = 2.10165*r;
  tp[1].coeffs[1][1].dCoord[Z] = 0.597556*r; 

  tp[1].coeffs[1][0].dCoord[X] = 1.29966*r; 
  tp[1].coeffs[1][0].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[1][0].dCoord[Z] = 0.40068*r; 

  tp[1].coeffs[2][3].dCoord[X] = -1.29966*r;
  tp[1].coeffs[2][3].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[2][3].dCoord[Z] = -0.40068*r;

  tp[1].coeffs[2][2].dCoord[X] = -0.597556*r;
  tp[1].coeffs[2][2].dCoord[Y] = 2.10165*r;
  tp[1].coeffs[2][2].dCoord[Z] = -0.597556*r;

  tp[1].coeffs[2][1].dCoord[X] = 0.597556*r;
  tp[1].coeffs[2][1].dCoord[Y] = 2.10165*r;
  tp[1].coeffs[2][1].dCoord[Z] = -0.597556*r;

  tp[1].coeffs[2][0].dCoord[X] = 1.29966*r;
  tp[1].coeffs[2][0].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[2][0].dCoord[Z] = -0.40068*r;

  tp[1].coeffs[3][3].dCoord[X] = -r;
  tp[1].coeffs[3][3].dCoord[Y] = r;
  tp[1].coeffs[3][3].dCoord[Z] = -r;

  tp[1].coeffs[3][2].dCoord[X] = -0.40068*r;
  tp[1].coeffs[3][2].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[3][2].dCoord[Z] = -1.29966*r;

  tp[1].coeffs[3][1].dCoord[X] = 0.40068*r;
  tp[1].coeffs[3][1].dCoord[Y] = 1.29966*r;
  tp[1].coeffs[3][1].dCoord[Z] = -1.29966*r;

  tp[1].coeffs[3][0].dCoord[X] = r;
  tp[1].coeffs[3][0].dCoord[Y] = r;
  tp[1].coeffs[3][0].dCoord[Z] = -r;

/* bottom */
  tp[2].coeffs[0][3].dCoord[X] = -r;
  tp[2].coeffs[0][3].dCoord[Y] = r;
  tp[2].coeffs[0][3].dCoord[Z] = -r;

  tp[2].coeffs[0][2].dCoord[X] = -0.40068*r;
  tp[2].coeffs[0][2].dCoord[Y] = 1.29966*r;
  tp[2].coeffs[0][2].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[0][1].dCoord[X] = 0.40068*r;
  tp[2].coeffs[0][1].dCoord[Y] = 1.29966*r;
  tp[2].coeffs[0][1].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[0][0].dCoord[X] = r;
  tp[2].coeffs[0][0].dCoord[Y] = r;
  tp[2].coeffs[0][0].dCoord[Z] = -r;

  tp[2].coeffs[1][3].dCoord[X] = -1.29966*r;
  tp[2].coeffs[1][3].dCoord[Y] = 0.40068*r;
  tp[2].coeffs[1][3].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[1][2].dCoord[X] = -0.597556*r; 
  tp[2].coeffs[1][2].dCoord[Y] = 0.597556*r; 
  tp[2].coeffs[1][2].dCoord[Z] = -2.10165*r; 

  tp[2].coeffs[1][1].dCoord[X] = 0.597556*r; 
  tp[2].coeffs[1][1].dCoord[Y] = 0.597556*r; 
  tp[2].coeffs[1][1].dCoord[Z] = -2.10165*r;

  tp[2].coeffs[1][0].dCoord[X] = 1.29966*r; 
  tp[2].coeffs[1][0].dCoord[Y] = 0.40068*r; 
  tp[2].coeffs[1][0].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[2][3].dCoord[X] = -1.29966*r;
  tp[2].coeffs[2][3].dCoord[Y] = -0.40068*r;
  tp[2].coeffs[2][3].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[2][2].dCoord[X] = -0.597556*r;
  tp[2].coeffs[2][2].dCoord[Y] = -0.597556*r;
  tp[2].coeffs[2][2].dCoord[Z] = -2.10165*r;

  tp[2].coeffs[2][1].dCoord[X] = 0.597556*r;
  tp[2].coeffs[2][1].dCoord[Y] = -0.597556*r;
  tp[2].coeffs[2][1].dCoord[Z] = -2.10165*r;

  tp[2].coeffs[2][0].dCoord[X] = 1.29966*r;
  tp[2].coeffs[2][0].dCoord[Y] = -0.40068*r;
  tp[2].coeffs[2][0].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[3][3].dCoord[X] = -r;
  tp[2].coeffs[3][3].dCoord[Y] = -r;
  tp[2].coeffs[3][3].dCoord[Z] = -r;

  tp[2].coeffs[3][2].dCoord[X] = -0.40068*r;
  tp[2].coeffs[3][2].dCoord[Y] = -1.29966*r;
  tp[2].coeffs[3][2].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[3][1].dCoord[X] = 0.40068*r;
  tp[2].coeffs[3][1].dCoord[Y] = -1.29966*r;
  tp[2].coeffs[3][1].dCoord[Z] = -1.29966*r;

  tp[2].coeffs[3][0].dCoord[X] = r;
  tp[2].coeffs[3][0].dCoord[Y] = -r;
  tp[2].coeffs[3][0].dCoord[Z] = -r;

/* front */
  tp[3].coeffs[0][3].dCoord[X] = -r;
  tp[3].coeffs[0][3].dCoord[Y] = -r;
  tp[3].coeffs[0][3].dCoord[Z] = -r;

  tp[3].coeffs[0][2].dCoord[X] = -0.40068*r;
  tp[3].coeffs[0][2].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[0][2].dCoord[Z] = -1.29966*r;

  tp[3].coeffs[0][1].dCoord[X] = 0.40068*r;
  tp[3].coeffs[0][1].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[0][1].dCoord[Z] = -1.29966*r;

  tp[3].coeffs[0][0].dCoord[X] = r;
  tp[3].coeffs[0][0].dCoord[Y] = -r;
  tp[3].coeffs[0][0].dCoord[Z] = -r;

  tp[3].coeffs[1][3].dCoord[X] = -1.29966*r;
  tp[3].coeffs[1][3].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[1][3].dCoord[Z] = -0.40068*r;

  tp[3].coeffs[1][2].dCoord[X] = -0.597556*r; 
  tp[3].coeffs[1][2].dCoord[Y] = -2.10165*r; 
  tp[3].coeffs[1][2].dCoord[Z] = -0.597556*r; 

  tp[3].coeffs[1][1].dCoord[X] = 0.597556*r; 
  tp[3].coeffs[1][1].dCoord[Y] = -2.10165*r;
  tp[3].coeffs[1][1].dCoord[Z] = -0.597556*r; 

  tp[3].coeffs[1][0].dCoord[X] = 1.29966*r; 
  tp[3].coeffs[1][0].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[1][0].dCoord[Z] = -0.40068*r; 

  tp[3].coeffs[2][3].dCoord[X] = -1.29966*r;
  tp[3].coeffs[2][3].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[2][3].dCoord[Z] = 0.40068*r;

  tp[3].coeffs[2][2].dCoord[X] = -0.597556*r;
  tp[3].coeffs[2][2].dCoord[Y] = -2.10165*r;
  tp[3].coeffs[2][2].dCoord[Z] = 0.597556*r;

  tp[3].coeffs[2][1].dCoord[X] = 0.597556*r;
  tp[3].coeffs[2][1].dCoord[Y] = -2.10165*r;
  tp[3].coeffs[2][1].dCoord[Z] = 0.597556*r;

  tp[3].coeffs[2][0].dCoord[X] = 1.29966*r;
  tp[3].coeffs[2][0].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[2][0].dCoord[Z] = 0.40068*r;

  tp[3].coeffs[3][3].dCoord[X] = -r;
  tp[3].coeffs[3][3].dCoord[Y] = -r;
  tp[3].coeffs[3][3].dCoord[Z] = r;

  tp[3].coeffs[3][2].dCoord[X] = -0.40068*r;
  tp[3].coeffs[3][2].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[3][2].dCoord[Z] = 1.29966*r;

  tp[3].coeffs[3][1].dCoord[X] = 0.40068*r;
  tp[3].coeffs[3][1].dCoord[Y] = -1.29966*r;
  tp[3].coeffs[3][1].dCoord[Z] = 1.29966*r;

  tp[3].coeffs[3][0].dCoord[X] = r;
  tp[3].coeffs[3][0].dCoord[Y] = -r;
  tp[3].coeffs[3][0].dCoord[Z] = r;

/* right */
  tp[4].coeffs[0][3].dCoord[X] = r;
  tp[4].coeffs[0][3].dCoord[Y] = -r;
  tp[4].coeffs[0][3].dCoord[Z] = r;

  tp[4].coeffs[0][2].dCoord[X] = 1.29966*r;
  tp[4].coeffs[0][2].dCoord[Y] = -1.29966*r;
  tp[4].coeffs[0][2].dCoord[Z] = 0.40068*r;

  tp[4].coeffs[0][1].dCoord[X] = 1.29966*r;
  tp[4].coeffs[0][1].dCoord[Y] = -1.29966*r;
  tp[4].coeffs[0][1].dCoord[Z] = -0.40068*r;

  tp[4].coeffs[0][0].dCoord[X] = r;
  tp[4].coeffs[0][0].dCoord[Y] = -r;
  tp[4].coeffs[0][0].dCoord[Z] = -r;

  tp[4].coeffs[1][3].dCoord[X] = 1.29966*r;
  tp[4].coeffs[1][3].dCoord[Y] = -0.40068*r;
  tp[4].coeffs[1][3].dCoord[Z] = 1.29966*r;

  tp[4].coeffs[1][2].dCoord[X] = 2.10165*r; 
  tp[4].coeffs[1][2].dCoord[Y] = -0.597556*r; 
  tp[4].coeffs[1][2].dCoord[Z] = 0.597556*r; 

  tp[4].coeffs[1][1].dCoord[X] = 2.10165*r;
  tp[4].coeffs[1][1].dCoord[Y] = -0.597556*r; 
  tp[4].coeffs[1][1].dCoord[Z] = -0.597556*r; 

  tp[4].coeffs[1][0].dCoord[X] = 1.29966*r; 
  tp[4].coeffs[1][0].dCoord[Y] = -0.40068*r; 
  tp[4].coeffs[1][0].dCoord[Z] = -1.29966*r;

  tp[4].coeffs[2][3].dCoord[X] = 1.29966*r;
  tp[4].coeffs[2][3].dCoord[Y] = 0.40068*r;
  tp[4].coeffs[2][3].dCoord[Z] = 1.29966*r;

  tp[4].coeffs[2][2].dCoord[X] = 2.10165*r;
  tp[4].coeffs[2][2].dCoord[Y] = 0.597556*r;
  tp[4].coeffs[2][2].dCoord[Z] = 0.597556*r;

  tp[4].coeffs[2][1].dCoord[X] = 2.10165*r;
  tp[4].coeffs[2][1].dCoord[Y] = 0.597556*r;
  tp[4].coeffs[2][1].dCoord[Z] = -0.597556*r;

  tp[4].coeffs[2][0].dCoord[X] = 1.29966*r;
  tp[4].coeffs[2][0].dCoord[Y] = 0.40068*r;
  tp[4].coeffs[2][0].dCoord[Z] = -1.29966*r;

  tp[4].coeffs[3][3].dCoord[X] = r;
  tp[4].coeffs[3][3].dCoord[Y] = r;
  tp[4].coeffs[3][3].dCoord[Z] = r;

  tp[4].coeffs[3][2].dCoord[X] = 1.29966*r;
  tp[4].coeffs[3][2].dCoord[Y] = 1.29966*r;
  tp[4].coeffs[3][2].dCoord[Z] = 0.40068*r;

  tp[4].coeffs[3][1].dCoord[X] = 1.29966*r;
  tp[4].coeffs[3][1].dCoord[Y] = 1.29966*r;
  tp[4].coeffs[3][1].dCoord[Z] = -0.40068*r;

  tp[4].coeffs[3][0].dCoord[X] = r;
  tp[4].coeffs[3][0].dCoord[Y] = r;
  tp[4].coeffs[3][0].dCoord[Z] = -r;

/* left */
  tp[5].coeffs[0][3].dCoord[X] = -r;
  tp[5].coeffs[0][3].dCoord[Y] = -r;
  tp[5].coeffs[0][3].dCoord[Z] = -r;

  tp[5].coeffs[0][2].dCoord[X] = -1.29966*r;
  tp[5].coeffs[0][2].dCoord[Y] = -1.29966*r;
  tp[5].coeffs[0][2].dCoord[Z] = -0.40068*r;

  tp[5].coeffs[0][1].dCoord[X] = -1.29966*r;
  tp[5].coeffs[0][1].dCoord[Y] = -1.29966*r;
  tp[5].coeffs[0][1].dCoord[Z] = 0.40068*r;

  tp[5].coeffs[0][0].dCoord[X] = -r;
  tp[5].coeffs[0][0].dCoord[Y] = -r;
  tp[5].coeffs[0][0].dCoord[Z] = r;

  tp[5].coeffs[1][3].dCoord[X] = -1.29966*r;
  tp[5].coeffs[1][3].dCoord[Y] = -0.40068*r;
  tp[5].coeffs[1][3].dCoord[Z] = -1.29966*r;

  tp[5].coeffs[1][2].dCoord[X] = -2.10165*r; 
  tp[5].coeffs[1][2].dCoord[Y] = -0.597556*r; 
  tp[5].coeffs[1][2].dCoord[Z] = -0.597556*r; 

  tp[5].coeffs[1][1].dCoord[X] = -2.10165*r;
  tp[5].coeffs[1][1].dCoord[Y] = -0.597556*r; 
  tp[5].coeffs[1][1].dCoord[Z] = 0.597556*r; 

  tp[5].coeffs[1][0].dCoord[X] = -1.29966*r; 
  tp[5].coeffs[1][0].dCoord[Y] = -0.40068*r; 
  tp[5].coeffs[1][0].dCoord[Z] = 1.29966*r;

  tp[5].coeffs[2][3].dCoord[X] = -1.29966*r;
  tp[5].coeffs[2][3].dCoord[Y] = 0.40068*r;
  tp[5].coeffs[2][3].dCoord[Z] = -1.29966*r;

  tp[5].coeffs[2][2].dCoord[X] = -2.10165*r;
  tp[5].coeffs[2][2].dCoord[Y] = 0.597556*r;
  tp[5].coeffs[2][2].dCoord[Z] = -0.597556*r;

  tp[5].coeffs[2][1].dCoord[X] = -2.10165*r;
  tp[5].coeffs[2][1].dCoord[Y] = 0.597556*r;
  tp[5].coeffs[2][1].dCoord[Z] = 0.597556*r;

  tp[5].coeffs[2][0].dCoord[X] = -1.29966*r;
  tp[5].coeffs[2][0].dCoord[Y] = 0.40068*r;
  tp[5].coeffs[2][0].dCoord[Z] = 1.29966*r;

  tp[5].coeffs[3][3].dCoord[X] = -r;
  tp[5].coeffs[3][3].dCoord[Y] = r;
  tp[5].coeffs[3][3].dCoord[Z] = -r;

  tp[5].coeffs[3][2].dCoord[X] = -1.29966*r;
  tp[5].coeffs[3][2].dCoord[Y] = 1.29966*r;
  tp[5].coeffs[3][2].dCoord[Z] = -0.40068*r;

  tp[5].coeffs[3][1].dCoord[X] = -1.29966*r;
  tp[5].coeffs[3][1].dCoord[Y] = 1.29966*r;
  tp[5].coeffs[3][1].dCoord[Z] = 0.40068*r;

  tp[5].coeffs[3][0].dCoord[X] = -r;
  tp[5].coeffs[3][0].dCoord[Y] = r;
  tp[5].coeffs[3][0].dCoord[Z] = r;
}
#endif

int gen_sphere(center, sph_radius, tp)
POINT4D center;
double sph_radius;
TENSOR_PATCH **tp;
{
  int i, j, k;
  double **XformMatSph;

  ALLOCMAT(XformMatSph, 4, 4, double);

  ALLOCN(*tp, TENSOR_PATCH, NUM_SPH_PATCHES);
  for (k=0; k<NUM_SPH_PATCHES; k++) {
    (*tp)[k].iDeg0 = SPH_SDEGREE;
    (*tp)[k].iDeg1 = SPH_TDEGREE;
    allocate_tensor_patch(&((*tp)[k]));
  }

  GetNormSphControlPts(*tp, sph_radius);

/* After getting normalized sphere's control points ie. sphere with
   center at origin , all we have
   to do is to transform each control point using XformMat */

#ifdef OLD_CODE_WITHOUT_ROTATION
  getXformMatSph(center, XformMatSph);
  XformPatch(*tp, NUM_SPH_PATCHES, SPH_SDEGREE, SPH_TDEGREE,
	     XformMatSph);
#else
  getRotationMatSph(XformMatSph, 1);
  XformPatch(*tp, NUM_SPH_PATCHES, SPH_SDEGREE, SPH_TDEGREE,
	     XformMatSph);
  getRotationMatSph(XformMatSph, 2);
  XformPatch(*tp, NUM_SPH_PATCHES, SPH_SDEGREE, SPH_TDEGREE,
	     XformMatSph);
  getRotationMatSph(XformMatSph, 3);
  XformPatch(*tp, NUM_SPH_PATCHES, SPH_SDEGREE, SPH_TDEGREE,
	     XformMatSph);
  getXformMatSph(center, XformMatSph);
  XformPatch(*tp, NUM_SPH_PATCHES, SPH_SDEGREE, SPH_TDEGREE,
	     XformMatSph);
#endif

  return NUM_SPH_PATCHES;
}
