#include <tree.h>
#include <macro.h>
#include <math.h>

#define VASE_LEEWAY_FACTOR 1.1

/* External global variables used */

extern TENSOR_PATCH tpatch[MAX_TPNUM];
extern int tpatch_idx;			
int vase_no_lid[2];

/* Local variables */

static struct lid_s {
  int is_scanned;
  int iDeg0;			/* table to hold the curve in */
  POINT4D *coeffs;		/* correct order */
} lid[MAX_TPNUM];

/* Generate the representation for the cone using the side-patches. */
/* Replace each of these curves by (possibly) 4 patches which form the */
/* circle. Also, add in the top and the bottom circles if necessary */

int generate_vase_representation(vase_t *vs, int np)
{
  TENSOR_PATCH pts[MAX_TPNUM];
  register int i, j, k, l;
  int n, tpnum, lidx, d0, d1;
  TENSOR_PATCH *tp0, *tp1, *tp2, *tp3, *tp, *thisp;
  vector_t vx, vdir, vpq, vq, v, vxq, vxqperp;
  vector_t vxq_sav0, vxq_sav1, vxqperp_sav0, vxqperp_sav1, vx_sav0, vx_sav1;
  double len, wcur, r, wcirc, r0, r1;
  POINT4D flp, llp;

  wcirc = 1/sqrt(2.0);
  /* Copy the patch headers in the temporary array */
  for (i = 0; i < np; i++)
    pts[i] = tpatch[i];

  /* The component curves may not be in order. Hence, scan all the */
  /* piecewise Bezier segments and put them in order as done for the */
  /* prism case */

  for (j = 0; j < np; j++)
    {
      lid[j].coeffs = (POINT4D *) new(MAX_CURVEDEG*sizeof(POINT4D));
      lid[j].is_scanned = FALSE;
    }
  
  /* arbitrarily introduce the first segment into the lid */

  for (i = 0; i <= tpatch[0].iDeg0; i++)
    lid[0].coeffs[i] = tpatch[0].coeffs[i][0];
  lid[0].is_scanned = TRUE;
  lid[0].iDeg0 = tpatch[0].iDeg0;

  for (j = 1; j < np; j++)	/* for every lid segment */
    {
      flp = lid[0].coeffs[0];
      llp = lid[j-1].coeffs[lid[j-1].iDeg0];
      
      for (i = 0; i < np; i++)
	{
	  if (lid[i].is_scanned)
	    continue;
	  thisp = &(tpatch[i]);
	  if (EQUAL_4D(llp, thisp->coeffs[0][0]))
	    {		
				/* normal order */
	      for (k = 0; k <= thisp->iDeg0; k++)
		lid[j].coeffs[k] = thisp->coeffs[k][0];
	      lid[j].iDeg0 = thisp->iDeg0;
	      lid[i].is_scanned = TRUE;
	      break;
	    }
	  else if (EQUAL_4D(llp, thisp->coeffs[thisp->iDeg0][0]))
	    {
				/* reverse order */
	      for (k = 0; k <= thisp->iDeg0; k++)
		lid[j].coeffs[thisp->iDeg0 - k] =
		  thisp->coeffs[k][0];
	      lid[j].iDeg0 = thisp->iDeg0;
	      lid[i].is_scanned = TRUE;
	      break;
	    }
	  else if (EQUAL_4D(flp, thisp->coeffs[thisp->iDeg0][0]))
	    {
				/* normal: before lid[0] */
	      for (k = j; k > 0; k--)
		{
		  lid[k].iDeg0 = lid[k-1].iDeg0;
		  for (l = 0; l <= lid[k-1].iDeg0; l++)
		    lid[k].coeffs[l] = lid[k-1].coeffs[l];
		}

	      for (k = 0; k <= thisp->iDeg0; k++)
		lid[0].coeffs[k] = thisp->coeffs[k][0];
	      lid[0].iDeg0 = thisp->iDeg0;
	      lid[i].is_scanned = TRUE;
	      break;
	    }
	  else if (EQUAL_4D(flp, thisp->coeffs[0][0]))
	    {
				/* normal: before lid[0] */
	      for (k = j; k > 0; k--)
		{
		  lid[k].iDeg0 = lid[k-1].iDeg0;
		  for (l = 0; l <= lid[k-1].iDeg0; l++)
		    lid[k].coeffs[l] = lid[k-1].coeffs[l];
		}

	      for (k = 0; k <= thisp->iDeg0; k++)
		lid[0].coeffs[thisp->iDeg0 - k] =
		  thisp->coeffs[k][0];
	      lid[0].iDeg0 = thisp->iDeg0;
	      lid[i].is_scanned = TRUE;
	      break;
	    }  
	} /* end getting succ for each segment */
    } /* end-for all segments */
  
  /* First write only for the special case where the rotation is a */
  /* complete 0-360 revolution. Anyway, the places where there is some */
  /* limited revolution is very small...... */

  /* Normalise the axis vector */
  len = DOT(vs->v, vs->v);
  len = sqrt(len);
  NORMALISE(vs->v, len);

  tpatch_idx = 0;
  for (k = 0; k < np; k++)
    {
      tpnum = get_new_patch(lid[k].iDeg0, 2);
      tp0 = &tpatch[tpnum];
      tpnum = get_new_patch(lid[k].iDeg0, 2);
      tp1 = &tpatch[tpnum];
      tpnum = get_new_patch(lid[k].iDeg0, 2);
      tp2 = &tpatch[tpnum];
      tpnum = get_new_patch(lid[k].iDeg0, 2);
      tp3 = &tpatch[tpnum];

      for (j = 0; j <= lid[k].iDeg0; j++)
	{
	  yourpt_to_mypt(lid[k].coeffs[j], vq);
	  
	  for (i = 0; i < 3; i++)
	    vq[i] /= vq[3];
	  wcur = vq[3];

	  for (i = 0; i < 3; i++)
	    vpq[i] = vq[i] - vs->pt1[i]; 
	  
	  /* Now, we have obtained normalised vs->v and vpq. Get vx */
	  /* from this info */
	  
	  len = DOT(vpq, vs->v);

	  for (i = 0; i < 3; i++)
	    vx[i] = vs->pt1[i] + len*vs->v[i];
	  vx[3] = wcur;
	  
	  /* get radius of the circle */
	  for (i = 0; i < 3; i++)
	    vxq[i] = vq[i] - vx[i];
	  r = DOT(vxq, vxq);
	  r = sqrt(r);
	  NORMALISE(vxq, r);

	  CROSS(vs->v, vxq, vxqperp);
	  len = DOT(vxqperp, vxqperp);
	  len = sqrt(len);
	  NORMALISE(vxqperp, len);

	  /* Save some info for generateing the circles later */

	  if ((k == 0) && (j == 0))
	    {
	      r0 = r;
	      for (i = 0; i < 4; i++)
		vxq_sav0[i] = vxq[i];
	      for (i = 0; i < 4; i++)
		vx_sav0[i] = vx[i];
	      for (i = 0; i < 4; i++)
		vxqperp_sav0[i] = vxqperp[i];
	    }
	  else if ((k == (np - 1)) && (j == lid[k].iDeg0))
	    {
	      r1 = r;
	      for (i = 0; i < 4; i++)
		vxq_sav1[i] = vxq[i];
	      for (i = 0; i < 4; i++)
		vx_sav1[i] = vx[i];
	      for (i = 0; i < 4; i++)
		vxqperp_sav1[i] = vxqperp[i];
	    }
	      

	  /* Now just stick in this point appropriately in the jth */
	  /* column of each of the 4 patches */
	  
	  tp0->coeffs[j][0].dCoord[0] = vq[0]*wcur;
	  tp0->coeffs[j][0].dCoord[1] = vq[1]*wcur;
	  tp0->coeffs[j][0].dCoord[2] = vq[2]*wcur;
	  tp0->coeffs[j][0].dCoord[3] = wcur;
	  
	  tp0->coeffs[j][1].dCoord[0] = (vq[0] + r*vxqperp[0])*wcur*wcirc;
	  tp0->coeffs[j][1].dCoord[1] = (vq[1] + r*vxqperp[1])*wcur*wcirc;
	  tp0->coeffs[j][1].dCoord[2] = (vq[2] + r*vxqperp[2])*wcur*wcirc;
	  tp0->coeffs[j][1].dCoord[3] = wcur*wcirc;
	  
	  tp0->coeffs[j][2].dCoord[0] = (vx[0] + r*vxqperp[0])*wcur;
	  tp0->coeffs[j][2].dCoord[1] = (vx[1] + r*vxqperp[1])*wcur;
	  tp0->coeffs[j][2].dCoord[2] = (vx[2] + r*vxqperp[2])*wcur;
	  tp0->coeffs[j][2].dCoord[3] = wcur;

	  /*********************************************/
	  tp1->coeffs[j][0] = tp0->coeffs[j][2];
	  
	  tp1->coeffs[j][1].dCoord[0] = 
	    (vx[0] + r*vxqperp[0] - r*vxq[0])*wcur*wcirc;
	  tp1->coeffs[j][1].dCoord[1] = 
	    (vx[1] + r*vxqperp[1] - r*vxq[1])*wcur*wcirc;
	  tp1->coeffs[j][1].dCoord[2] = 
	    (vx[2] + r*vxqperp[2] - r*vxq[2])*wcur*wcirc;
	  tp1->coeffs[j][1].dCoord[3] = wcur*wcirc;
	  
	  tp1->coeffs[j][2].dCoord[0] = (vx[0] - r*vxq[0])*wcur;
	  tp1->coeffs[j][2].dCoord[1] = (vx[1] - r*vxq[1])*wcur;
	  tp1->coeffs[j][2].dCoord[2] = (vx[2] - r*vxq[2])*wcur;
	  tp1->coeffs[j][2].dCoord[3] = wcur;

	  /*********************************************/
	  tp2->coeffs[j][0] = tp1->coeffs[j][2];
	  
	  tp2->coeffs[j][1].dCoord[0] = 
	    (vx[0] - r*vxqperp[0] - r*vxq[0])*wcur*wcirc;
	  tp2->coeffs[j][1].dCoord[1] = 
	    (vx[1] - r*vxqperp[1] - r*vxq[1])*wcur*wcirc;
	  tp2->coeffs[j][1].dCoord[2] = 
	    (vx[2] - r*vxqperp[2] - r*vxq[2])*wcur*wcirc;
	  tp2->coeffs[j][1].dCoord[3] = wcur*wcirc;
	  
	  tp2->coeffs[j][2].dCoord[0] = (vx[0] - r*vxqperp[0])*wcur;
	  tp2->coeffs[j][2].dCoord[1] = (vx[1] - r*vxqperp[1])*wcur;
	  tp2->coeffs[j][2].dCoord[2] = (vx[2] - r*vxqperp[2])*wcur;
	  tp2->coeffs[j][2].dCoord[3] = wcur;

	  /*********************************************/
	  tp3->coeffs[j][0] = tp2->coeffs[j][2];
	  
	  tp3->coeffs[j][1].dCoord[0] = 
	    (vx[0] - r*vxqperp[0] + r*vxq[0])*wcur*wcirc;
	  tp3->coeffs[j][1].dCoord[1] = 
	    (vx[1] - r*vxqperp[1] + r*vxq[1])*wcur*wcirc;
	  tp3->coeffs[j][1].dCoord[2] = 
	    (vx[2] - r*vxqperp[2] + r*vxq[2])*wcur*wcirc;
	  tp3->coeffs[j][1].dCoord[3] = wcur*wcirc;
	  
	  tp3->coeffs[j][2] = tp0->coeffs[j][0];
	  
	} /* End-of-j-loop for each curve */
    } /* end-for-all-curves */

  /* Addition: Jan 1 (the new year begins with handling a special case */
  /* NOT mentioned in the manual) ... sigh */
  /* If the curve is closed, do not generate any of the top/bottom */
  /* circles. */
  
  d0 = lid[np - 1].iDeg0;
  if (!EQUAL_4D_PAR(lid[0].coeffs[0], lid[np - 1].coeffs[d0],
		   1.0e-6))
    {
      /* Now we have to generate the top and the bottom circles */
      
      if (!FP_EQUAL_TOL(r0, 0.0, 1.0e-2))
	{
	  tpnum = get_new_patch(1, 1);
	  tp = &tpatch[tpnum];
	  r0 *= VASE_LEEWAY_FACTOR;
	  
	  tp->coeffs[1][0].dCoord[0] = vx_sav0[0] +
	    r0*vxq_sav0[0] - r0*vxqperp_sav0[0];
	  tp->coeffs[1][0].dCoord[1] = vx_sav0[1] +
	    r0*vxq_sav0[1] - r0*vxqperp_sav0[1];
	  tp->coeffs[1][0].dCoord[2] = vx_sav0[2] +
	    r0*vxq_sav0[2] - r0*vxqperp_sav0[2];
	  tp->coeffs[1][0].dCoord[3] = 1.0;
	  
	  tp->coeffs[1][1].dCoord[0] = vx_sav0[0] +
	    r0*vxq_sav0[0] + r0*vxqperp_sav0[0];
	  tp->coeffs[1][1].dCoord[1] = vx_sav0[1] +
	    r0*vxq_sav0[1] + r0*vxqperp_sav0[1];
	  tp->coeffs[1][1].dCoord[2] = vx_sav0[2] +
	    r0*vxq_sav0[2] + r0*vxqperp_sav0[2];
	  tp->coeffs[1][1].dCoord[3] = 1.0;
	  
	  tp->coeffs[0][0].dCoord[0] = vx_sav0[0] -
	    r0*vxq_sav0[0] - r0*vxqperp_sav0[0];
	  tp->coeffs[0][0].dCoord[1] = vx_sav0[1] -
	    r0*vxq_sav0[1] - r0*vxqperp_sav0[1];
	  tp->coeffs[0][0].dCoord[2] = vx_sav0[2] -
	    r0*vxq_sav0[2] - r0*vxqperp_sav0[2];
	  tp->coeffs[0][0].dCoord[3] = 1.0;
	  
	  tp->coeffs[0][1].dCoord[0] = vx_sav0[0] -
	    r0*vxq_sav0[0] + r0*vxqperp_sav0[0];
	  tp->coeffs[0][1].dCoord[1] = vx_sav0[1] -
	    r0*vxq_sav0[1] + r0*vxqperp_sav0[1];
	  tp->coeffs[0][1].dCoord[2] = vx_sav0[2] -
	    r0*vxq_sav0[2] + r0*vxqperp_sav0[2];
	  tp->coeffs[0][1].dCoord[3] = 1.0;
	  vase_no_lid[0] = FALSE;
	}      
      else
	vase_no_lid[0] = TRUE;

      if (!FP_EQUAL_TOL(r1, 0.0, 1.0e-2))
	{
	  tpnum = get_new_patch(1, 1);
	  tp = &tpatch[tpnum];
	  r1 *= VASE_LEEWAY_FACTOR;
	  
	  tp->coeffs[0][0].dCoord[0] = vx_sav1[0] +
	    r1*vxq_sav1[0] - r1*vxqperp_sav1[0];
	  tp->coeffs[0][0].dCoord[1] = vx_sav1[1] +
	    r1*vxq_sav1[1] - r1*vxqperp_sav1[1];
	  tp->coeffs[0][0].dCoord[2] = vx_sav1[2] +
	    r1*vxq_sav1[2] - r1*vxqperp_sav1[2];
	  tp->coeffs[0][0].dCoord[3] = 1.0;
	  
	  tp->coeffs[0][1].dCoord[0] = vx_sav1[0] +
	    r1*vxq_sav1[0] + r1*vxqperp_sav1[0];
	  tp->coeffs[0][1].dCoord[1] = vx_sav1[1] +
	    r1*vxq_sav1[1] + r1*vxqperp_sav1[1];
	  tp->coeffs[0][1].dCoord[2] = vx_sav1[2] +
	    r1*vxq_sav1[2] + r1*vxqperp_sav1[2];
	  tp->coeffs[0][1].dCoord[3] = 1.0;
	  
	  tp->coeffs[1][0].dCoord[0] = vx_sav1[0] -
	    r1*vxq_sav1[0] - r1*vxqperp_sav1[0];
	  tp->coeffs[1][0].dCoord[1] = vx_sav1[1] -
	    r1*vxq_sav1[1] - r1*vxqperp_sav1[1];
	  tp->coeffs[1][0].dCoord[2] = vx_sav1[2] -
	    r1*vxq_sav1[2] - r1*vxqperp_sav1[2];
	  tp->coeffs[1][0].dCoord[3] = 1.0;
	  
	  tp->coeffs[1][1].dCoord[0] = vx_sav1[0] -
	    r1*vxq_sav1[0] + r1*vxqperp_sav1[0];
	  tp->coeffs[1][1].dCoord[1] = vx_sav1[1] -
	    r1*vxq_sav1[1] + r1*vxqperp_sav1[1];
	  tp->coeffs[1][1].dCoord[2] = vx_sav1[2] -
	    r1*vxq_sav1[2] + r1*vxqperp_sav1[2];
	  tp->coeffs[1][1].dCoord[3] = 1.0;

	  vase_no_lid[1] = FALSE;
	}
      else
	vase_no_lid[1] = TRUE;
      
      /* Thus, the last two patches represent the top and bottom */
      /* respectively. Add the proper trimming curve in the */
      /* add_trim_info() routine. Free all the patches in pts.. */
    }
  else
    {				/* closed curve. do nto generate lids */
      vase_no_lid[0] = TRUE;
      vase_no_lid[1] = TRUE;
    }
  
#ifdef _FREE_ALL
  for (k = 0; k < np; i++)
    free(lid[k].coeffs);
#endif

  return tpatch_idx;
}
