#include "defs.h"
#include "linalg.h"

EXTERN_LINALG

extern NDETERMINANT_1D  ndet1D_uv, fdet1D_uv, ndet1D_du, ndet1D_dv;
extern int             MatOrderForMinimFunc;
extern TENSOR_PATCH ntp1, ntp2;
extern TENSOR_PATCH pow_tp0, pow_tp1;

#ifdef _MYMINIMFUNC
double minim_func_for_sing( nDimPt )
double *nDimPt;
{

    int	   iOrder=MatOrderForMinimFunc;
    int    currow, iRow, iCol, irgPivotCol, irgPivotRow, i, j;
    double rec_dPivot, dPivot, dTemp, dMaxPivot, dPivot2, dDet, dDDetu, dDDetv;
    double dGki, dGuki, dGvki, dGuuki, dGuvki, dGvvki;
    double Pu, Nu, Pv, Nv;
    double **rgDet;
    double **rgDDetu;
    double **rgDDetv;

    eval_uv(ndet1D_uv, nDimPt[0], nDimPt[1]);
    eval_uv(ndet1D_du, nDimPt[0], nDimPt[1]);
    eval_uv(ndet1D_dv, nDimPt[0], nDimPt[1]);
    rgDet = ndet1D_uv.eval;
    rgDDetu = ndet1D_du.eval;
    rgDDetv = ndet1D_dv.eval;

    /* start at row one and loop through rows */
    /* until we have completed gaussian elimination portion */
    /* note: we continue to examine rows even after we have found */
    /* a pivot for each location, because one of the later rows */
    /* may render the matrix system inconsistent */
    
    for (currow=0;currow<iOrder-1;currow++) {

	/* find largest remaining value for complete pivoting */
	for (dMaxPivot=0.0,irgPivotRow= -1,irgPivotCol= -1,iRow=currow;iRow<iOrder;iRow++){
	    for (iCol=currow;iCol<iOrder;iCol++) {
		dTemp = rgDet[iRow][iCol];
		if (dTemp<0.0)
		    dTemp = -dTemp;
		
		if (dTemp > dMaxPivot ) {
		    irgPivotRow=iRow;
		    irgPivotCol=iCol;
		    dMaxPivot=dTemp;
		}
	    }
	    
	}

	/* if no pivot was found, there is a row of zeros or linearly 
	   dependent vectors */

	if ((irgPivotRow<0) || (irgPivotCol<0)) {
	    printf("Came here!! That's why SINGULARITY\n");
	    return -1;
	}

	/* otherwise, we have the largest pivot in irgPivotRow and 
	   irgPivotCol */

	if (currow!=irgPivotRow) {
	    /* exchange rows */
	    for (iCol=0;iCol<iOrder;iCol++)  {
		dTemp = rgDet[currow][iCol];
		rgDet[currow][iCol]=	-rgDet[irgPivotRow][iCol];
		rgDet[irgPivotRow][iCol]=dTemp;

		dTemp = rgDDetu[currow][iCol];
		rgDDetu[currow][iCol]=	-rgDDetu[irgPivotRow][iCol];
		rgDDetu[irgPivotRow][iCol]=dTemp;
		dTemp = rgDDetv[currow][iCol];
		rgDDetv[currow][iCol]=	-rgDDetv[irgPivotRow][iCol];
		rgDDetv[irgPivotRow][iCol]=dTemp;
	    }
	}

	if (currow!=irgPivotCol) {
	    /* exchange cols */
	    for (iRow=0;iRow<iOrder;iRow++)  {
		dTemp = rgDet[iRow][currow];
		rgDet[iRow][currow]=	-rgDet[iRow][irgPivotCol];
		rgDet[iRow][irgPivotCol]=dTemp;

		dTemp = rgDDetu[iRow][currow];
		rgDDetu[iRow][currow]=	-rgDDetu[iRow][irgPivotCol];
		rgDDetu[iRow][irgPivotCol]=dTemp;
		dTemp = rgDDetv[iRow][currow];
		rgDDetv[iRow][currow]=	-rgDDetv[iRow][irgPivotCol];
		rgDDetv[iRow][irgPivotCol]=dTemp;
	    }
	}


	/* pivot is now at currow,currow */

	dPivot = rgDet[currow][currow];
	rec_dPivot = 1.0/ dPivot;
	dPivot2 = rec_dPivot * rec_dPivot;

	/* subtract this row from all of the others */
	for (iRow=currow+1;iRow<iOrder;iRow++){
	    dTemp = rgDet[iRow][currow]/dPivot;

	    /* i = currow (row with pivot),
	       k = row row i is being subtracted from */

	    dGki = rgDet[iRow][currow];
	    dGuki = rgDDetu[iRow][currow];
	    dGvki = rgDDetv[iRow][currow];

	    /* no need to zero the entry, but I will do it for
	       easy error checking */

	    rgDet[iRow][currow]=0.0;

	    /* for determinant portion, we must start earlier in
	       row */
	    iCol = 0;

	    for (;iCol<iOrder;iCol++)  {
		if (iCol>currow)
		    rgDet[iRow][iCol] -= dTemp * rgDet[currow][iCol];
		Pu = dGuki*rgDet[currow][iCol] + dGki*rgDDetu[currow][iCol];
		Pv = dGvki*rgDet[currow][iCol] + dGki*rgDDetv[currow][iCol];
		Nu = dPivot2 * (Pu * dPivot 
		     - dGki*rgDet[currow][iCol]*rgDDetu[currow][currow]);
		Nv = dPivot2 * (Pv * dPivot 
		     - dGki*rgDet[currow][iCol]*rgDDetv[currow][currow]);
		rgDDetu[iRow][iCol] -= Nu;
		rgDDetv[iRow][iCol] -= Nv;
	    }
	}	
	
    }

    /* gaussian elimination is complete */
    if (FP_EQ_EPS(rgDet[iOrder-1][iOrder-1], 0, 1e-4))
    { 
      if (FP_EQ_EPS(rgDet[iOrder-2][iOrder-2], 0, 1e-4)) return 0.0;

      for(dDet = 1.0, i = 0; i < iOrder-1; i++) dDet *= rgDet[i][i];
      dDDetu = rgDDetu[iOrder-1][iOrder-1];
      dDDetv = rgDDetv[iOrder-1][iOrder-1];
      dDDetu *= dDet;
      dDDetv *= dDet;
      return dDDetu*dDDetu + dDDetv*dDDetv;
    }
    else
    { 
      for(dDet = 1.0, i = 0; i < iOrder; i++) dDet *= rgDet[i][i];
      for (dDDetu=0.0, i=0; i<iOrder; i++) 
	dDDetu += (rgDDetu[i][i] / rgDet[i][i]);
      for (dDDetv=0.0, i=0; i<iOrder; i++) 
	dDDetv += (rgDDetv[i][i] / rgDet[i][i]);
      dDDetu *= dDet;
      dDDetv *= dDet;
      return dDet*dDet + dDDetu*dDDetu + dDDetv*dDDetv;
    }
}
#else
double minim_func_for_sing( nDimPt )
double *nDimPt;
{
  double t1x, t1y, t1z, t2x, t2y, t2z;
  double nt1x, nt1y, nt1z, nt2x, nt2y, nt2z;
  double retv, cr_n1n2x, cr_n1n2y, cr_n1n2z;
  double Fst_Guv_dist, Nst_Nuv_Cr_mag;

  EvalPointOnPatch(pow_tp0, nDimPt[2], nDimPt[3], &t1x, &t1y, &t1z);
  EvalPointOnPatch(pow_tp1, nDimPt[0], nDimPt[1], &t2x, &t2y, &t2z);
  EvalPointOnPatch(ntp1, nDimPt[2], nDimPt[3], &nt1x, &nt1y, &nt1z);
  EvalPointOnPatch(ntp2, nDimPt[0], nDimPt[1], &nt2x, &nt2y, &nt2z);

  cr_n1n2x = nt1y*nt2z - nt2y*nt1z;
  cr_n1n2y = nt2x*nt1z - nt1x*nt2z;
  cr_n1n2z = nt1x*nt2y - nt2x*nt1y;

  Fst_Guv_dist = (t2x - t1x)*(t2x - t1x) +
                 (t2y - t1y)*(t2y - t1y) +
                 (t2z - t1z)*(t2z - t1z);
  Nst_Nuv_Cr_mag = cr_n1n2x*cr_n1n2x +
                   cr_n1n2y*cr_n1n2y +
                   cr_n1n2z*cr_n1n2z;

  return (Fst_Guv_dist + Nst_Nuv_Cr_mag);
}
#endif


/**************************************************************************
		  	Gaussian Elimination Routines
**************************************************************************/

/* this is the routine which does the gaussian and gauss-jordan methods*/
/* the operation is carefully explained inline, and an algorithmic */
/* explanation is annotated/attached */

int
calc_det( rgDet, rgDDetu, rgDDetv, pdDDetu, pdDDetv, cDim )
double **rgDet;
double **rgDDetu;
double **rgDDetv;
double *pdDDetu, *pdDDetv;
int    cDim;
{

    int	   iOrder=cDim;
    int    currow, iRow, iCol, irgPivotCol, irgPivotRow, i, j;
    double rec_dPivot, dPivot, dTemp, dMaxPivot, dPivot2, dDet, dDDetu, dDDetv;
    double dGki, dGuki, dGvki, dGuuki, dGuvki, dGvvki;
    double Pu, Nu, Pv, Nv;

    /* start at row one and loop through rows */
    /* until we have completed gaussian elimination portion */
    /* note: we continue to examine rows even after we have found */
    /* a pivot for each location, because one of the later rows */
    /* may render the matrix system inconsistent */
    
    for (currow=0;currow<iOrder-1;currow++) {

	/* find largest remaining value for complete pivoting */
	for (dMaxPivot=0.0,irgPivotRow= -1,irgPivotCol= -1,iRow=currow;iRow<iOrder;iRow++){
	    for (iCol=currow;iCol<iOrder;iCol++) {
		dTemp = rgDet[iRow][iCol];
		if (dTemp<0.0)
		    dTemp = -dTemp;
		
		if (dTemp > dMaxPivot ) {
		    irgPivotRow=iRow;
		    irgPivotCol=iCol;
		    dMaxPivot=dTemp;
		}
	    }
	    
	}

	/* if no pivot was found, there is a row of zeros or linearly 
	   dependent vectors */

	if ((irgPivotRow<0) || (irgPivotCol<0)) {
	    printf("Came here!! That's why SINGULARITY\n");
	    *pdDDetu = 0.0;
	    *pdDDetv = 0.0;
	    return -1;
	}

	/* otherwise, we have the largest pivot in irgPivotRow and 
	   irgPivotCol */

	if (currow!=irgPivotRow) {
	    /* exchange rows */
	    for (iCol=0;iCol<iOrder;iCol++)  {
		dTemp = rgDet[currow][iCol];
		rgDet[currow][iCol]=	-rgDet[irgPivotRow][iCol];
		rgDet[irgPivotRow][iCol]=dTemp;

		dTemp = rgDDetu[currow][iCol];
		rgDDetu[currow][iCol]=	-rgDDetu[irgPivotRow][iCol];
		rgDDetu[irgPivotRow][iCol]=dTemp;
		dTemp = rgDDetv[currow][iCol];
		rgDDetv[currow][iCol]=	-rgDDetv[irgPivotRow][iCol];
		rgDDetv[irgPivotRow][iCol]=dTemp;
	    }
	}

	if (currow!=irgPivotCol) {
	    /* exchange cols */
	    for (iRow=0;iRow<iOrder;iRow++)  {
		dTemp = rgDet[iRow][currow];
		rgDet[iRow][currow]=	-rgDet[iRow][irgPivotCol];
		rgDet[iRow][irgPivotCol]=dTemp;

		dTemp = rgDDetu[iRow][currow];
		rgDDetu[iRow][currow]=	-rgDDetu[iRow][irgPivotCol];
		rgDDetu[iRow][irgPivotCol]=dTemp;
		dTemp = rgDDetv[iRow][currow];
		rgDDetv[iRow][currow]=	-rgDDetv[iRow][irgPivotCol];
		rgDDetv[iRow][irgPivotCol]=dTemp;
	    }
	}


	/* pivot is now at currow,currow */

	dPivot = rgDet[currow][currow];
	rec_dPivot = 1.0/ dPivot;
	dPivot2 = rec_dPivot * rec_dPivot;

	/* subtract this row from all of the others */
	for (iRow=currow+1;iRow<iOrder;iRow++){
	    dTemp = rgDet[iRow][currow]/dPivot;

	    /* i = currow (row with pivot),
	       k = row row i is being subtracted from */

	    dGki = rgDet[iRow][currow];
	    dGuki = rgDDetu[iRow][currow];
	    dGvki = rgDDetv[iRow][currow];

	    /* no need to zero the entry, but I will do it for
	       easy error checking */

	    rgDet[iRow][currow]=0.0;

	    /* for determinant portion, we must start earlier in
	       row */
	    iCol = 0;

	    for (;iCol<iOrder;iCol++)  {
		if (iCol>currow)
		    rgDet[iRow][iCol] -= dTemp * rgDet[currow][iCol];
		Pu = dGuki*rgDet[currow][iCol] + dGki*rgDDetu[currow][iCol];
		Pv = dGvki*rgDet[currow][iCol] + dGki*rgDDetv[currow][iCol];
		Nu = dPivot2 * (Pu * dPivot 
		     - dGki*rgDet[currow][iCol]*rgDDetu[currow][currow]);
		Nv = dPivot2 * (Pv * dPivot 
		     - dGki*rgDet[currow][iCol]*rgDDetv[currow][currow]);
		rgDDetu[iRow][iCol] -= Nu;
		rgDDetv[iRow][iCol] -= Nv;
/*
		rgDDetuu[iRow][iCol] += 2*rgDDetu[currow][currow]*Nu*rec_dPivot -
			dPivot2*(dPivot*(rgDet[currow][iCol]*dGuuki +
				         dGki*rgDetuu[currow][iCol] +
					 2*dGuki*rgDetu[currow][iCol]) -
		             rgDetuu[currow][currow]*dGki*rgDet[currow][iCol]);
		rgDDetvv[iRow][iCol] += 2*rgDDetv[currow][currow]*Nv*rec_dPivot -
			dPivot2*(dPivot*(rgDet[currow][iCol]*dGvvki +
				         dGki*rgDetvv[currow][iCol] +
					 2*dGvki*rgDetv[currow][iCol]) -
		             rgDetvv[currow][currow]*dGki*rgDet[currow][iCol]);
		rgDDetuv[iRow][iCol] += 2*rgDDetv[currow][currow]*Nu*rec_dPivot -
			dPivot2*(dPivot*(rgDet[currow][iCol]*dGuvki +
				         dGki*rgDetuv[currow][iCol] +
					 dGuki*rgDetv[currow][iCol] +
					 dGvki*rgDetu[currow][iCol]) +
			     rgDetv[currow][currow]*Pu -
			     rgDetu[currow][currow]*Pv -
		             rgDetuv[currow][currow]*dGki*rgDet[currow][iCol]);
*/
	    }
	}	
	
    }

    /* gaussian elimination is complete */

    /* (Atul: Sep 4).. gave problems with t11. So restored the */
    /* original statements */
   if (FP_EQ_EPS(rgDet[0][0], 0.0, 1e-1) ||
	FP_EQ_EPS(rgDet[iOrder-1][iOrder-1]/rgDet[0][0],0.0,1e-3))
#ifdef SCALE_DOWN_FOR_TRACING_ALONE
    if (FP_EQ_EPS(rgDet[iOrder-1][iOrder-1],0.0,1e-6))
#else
/*
    if (FP_EQ_EPS(rgDet[iOrder-1][iOrder-1],0.0,1e-2))
*/
#endif
    { 
/*
      if (FP_EQ_EPS(rgDet[iOrder-2][iOrder-2], 0.0, 1e-2))
	return SINGULAR_CASE;
*/
       /* for(dDet = 1.0, i = 0; i < iOrder-1; i++) dDet *= rgDet[i][i]; */
      dDDetu = rgDDetu[iOrder-1][iOrder-1];
      dDDetv = rgDDetv[iOrder-1][iOrder-1];
      /* return dDDetu and dDDetv in this form as finally we just need their
	 ratio for tangent computation
      */
      *pdDDetu = dDDetu;
      *pdDDetv = dDDetv;
    }
    else
    { printf("calc_det(): Matrix not sufficiently singular!!\n");
      printf("Det[%d][%d] = %lf\n", iOrder-1, iOrder-1, rgDet[iOrder-1][iOrder-1]);
      exit(-1);
    }
    return 0;
}

/* Solves the system Ax = B  and returns x. It does gauss elimination 
   with complete pivoting. A is a (cDim x cDim) matrix. A and B are 
   left unchanged. If A is singular, it returns 0.
*/
int
gauss_elim(rgrgA, rgX, rgB, cDim)
double **rgrgA;
double *rgX;
double *rgB;
int    cDim;
{

    double **rgrgL, *rgTmpB, *rgTmpX;
    int	   *perm_array, iTemp;
    int	   iOrder=cDim;
    int    currow, iRow, iCol, irgPivotCol, irgPivotRow;
    double dPivot, dTemp, dMaxPivot;

    /* initialize the permutation array */
    ALLOCN(perm_array, int, cDim);
    for(iRow=0; iRow < cDim; iRow++) 
	perm_array[iRow] = iRow;

    ALLOCN(rgTmpX, double, cDim);

    /* copy matrix A into L */
    ALLOCMAT(rgrgL, cDim, cDim, double)
    MATASSIGN(rgrgL, rgrgA, cDim, cDim);

    /* copy matrix B into TmpB */
    ALLOCN(rgTmpB, double, cDim);
    VEC_ASN_OP(rgTmpB, = , rgB, cDim);

    /* start at row one and loop through rows */
    /* until we have completed gaussian elimination portion */
    /* note: we continue to examine rows even after we have found */
    /* a pivot for each location, because one of the later rows */
    /* may render the matrix system inconsistent */
    
    for (currow=0;currow<iOrder-1;currow++) {
	/* find largest remaining value for complete pivoting */
	for (dMaxPivot=0.0,irgPivotRow= -1,irgPivotCol= -1,iRow=currow;iRow<iOrder;iRow++){
	    for (iCol=currow;iCol<iOrder;iCol++) {
		dTemp = rgrgL[iRow][iCol];
		if (dTemp<0.0)
		    dTemp = -dTemp;
		
		if (dTemp > dMaxPivot ) {
		    irgPivotRow=iRow;
		    irgPivotCol=iCol;
		    dMaxPivot=dTemp;
		}
	    }
	    
	}

	/* if no pivot was found, there is a row of zeros or linearly 
	   dependent vectors */

	if ((irgPivotRow<0) || (irgPivotCol<0)) return(0);

	/* otherwise, we have the largest pivot in irgPivotRow and 
	   irgPivotCol */

	if (currow!=irgPivotRow) {
	    /* exchange rows of rgrgL */
	    for (iCol=0;iCol<iOrder;iCol++)  {
		dTemp = rgrgL[currow][iCol];
		rgrgL[currow][iCol]=rgrgL[irgPivotRow][iCol];
		rgrgL[irgPivotRow][iCol]=dTemp;
	    }
	    /* exchange rows of rgTmpB */
	    dTemp = rgTmpB[currow];
	    rgTmpB[currow] = rgTmpB[irgPivotRow];
	    rgTmpB[irgPivotRow] = dTemp;
	}

	if (currow!=irgPivotCol) {
	    /* exchange cols */
	    for (iRow=0;iRow<iOrder;iRow++)  {
		dTemp = rgrgL[iRow][currow];
		rgrgL[iRow][currow]=rgrgL[iRow][irgPivotCol];
		rgrgL[iRow][irgPivotCol]=dTemp;
	    }
	    /* swap the entries of the permutation array */
	    iTemp = perm_array[irgPivotCol];
	    perm_array[irgPivotCol] = perm_array[currow];
	    perm_array[currow] = iTemp;
	}


	/* pivot is now at currow,currow */

	dPivot = rgrgL[currow][currow];

	/* subtract this row from all of the others */
	for (iRow=currow+1;iRow<iOrder;iRow++){
	    dTemp = rgrgL[iRow][currow]/dPivot;

	    /* no need to zero the entry, but I will do it for
	       easy error checking */

	    rgrgL[iRow][currow]=0.0;

	    /* for determinant portion, we must start earlier in
	       row */
	    for (iCol = currow+1;iCol<iOrder;iCol++)  
              rgrgL[iRow][iCol] -= dTemp * rgrgL[currow][iCol];
	    rgTmpB[iRow] -= dTemp*rgTmpB[currow];
	}	
	
    }
   
    /* gaussian elimination is complete, now do back-substitution */

    for (currow=iOrder-1;currow>=0;currow--) 
    { rgTmpX[currow] = rgTmpB[currow];
      for(iRow=currow+1; iRow < iOrder; iRow++)
	rgTmpX[currow] -= rgrgL[currow][iRow]*rgTmpX[iRow];
      rgTmpX[currow] /= rgrgL[currow][currow];
    }

    /* rearrange values of rgTmpX correctly into rgX*/
    for(currow=0; currow < iOrder; currow++)
      rgX[perm_array[currow]] = rgTmpX[currow];

    FREEMAT(rgrgL, cDim, cDim);
    myfree(rgTmpX); myfree(rgTmpB); myfree(perm_array);
    return(1);
}
