#include "fitting.h"
#include "defs.h"

/* The following procedure takes an INT_CURVE_STRUCT and converts it */
/* to ChunFa's reprn to call his fitting routine */

int convert_for_fitting(crv, uv_ptch, st_ptch, uv_array, st_array,
			num_pts)
INT_CURVE_STRUCT *crv;
TENSOR_PATCH *uv_ptch, *st_ptch;
FiveTuple **uv_array, **st_array;
int *num_pts;
{
  int i;
  LIST_POINT4D *tmpptr;

  tmpptr = crv->Curve;
  (*num_pts) = 0;
  while (tmpptr != NULL) {
    (*num_pts)++;
    tmpptr = tmpptr->plp4DNext;
    if (*num_pts >= 100000) {
      printf("convert_for_fitting: Obviously this is going nowhere.\n");
      exit(66);
    }
  }

  tmpptr = crv->Curve;
  ALLOCN((*uv_array), FiveTuple, (*num_pts));
  ALLOCN((*st_array), FiveTuple, (*num_pts));
  
  for (i=0; i<(*num_pts); i++) {
    (*uv_array)[i].s = tmpptr->p4D.dCoord[0];
    (*uv_array)[i].t = tmpptr->p4D.dCoord[1];
    (*st_array)[i].s = tmpptr->p4D.dCoord[2];
    (*st_array)[i].t = tmpptr->p4D.dCoord[3];

    EvalPointOnPatch((*uv_ptch), (*uv_array)[i].s, (*uv_array)[i].t,
		     &((*uv_array)[i].x), &((*uv_array)[i].y),
		     &((*uv_array)[i].z)); 
    EvalPointOnPatch((*st_ptch), (*st_array)[i].s, (*st_array)[i].t,
		     &((*st_array)[i].x), &((*st_array)[i].y),
		     &((*st_array)[i].z)); 
    
    (*uv_array)[i].x = (*st_array)[i].x;
    (*uv_array)[i].y = (*st_array)[i].y;
    (*uv_array)[i].z = (*st_array)[i].z;

    tmpptr = tmpptr->plp4DNext;
  }

  if (tmpptr != NULL) {
    printf("convert_for_fitting: mistake in counting\n");
    exit(-1);
  }
  return 0;
}

/* This routine converts a structure ParamCurve to a UV_SPLINE and a */
/* XYZ_SPLINE. Allocation for the output splines is done inside this */
/* routine. the 'control' array inside i_crv is not freed here */
int convert_spline_reprn(i_crv, uv_ocrv, xyz_ocrv)
ParamCurve *i_crv;
UV_SPLINE_TYPE *uv_ocrv;
XYZ_SPLINE_TYPE *xyz_ocrv;
{
  int i, deg;

  deg = uv_ocrv->Deg = xyz_ocrv->Deg = i_crv->degree;

  ALLOCN(uv_ocrv->CntrlPts, POINT3D, deg+1);
  ALLOCN(xyz_ocrv->CntrlPts, POINT4D, deg+1);

  for (i=0; i<=deg; i++) {
    uv_ocrv->CntrlPts[i].dCoord[2] = 1.0;
    xyz_ocrv->CntrlPts[i].dCoord[3] = 1.0;
  }

  for (i=0; i<=deg; i++) {
    uv_ocrv->CntrlPts[i].dCoord[0] = i_crv->control[i].s;
    uv_ocrv->CntrlPts[i].dCoord[1] = i_crv->control[i].t;
    xyz_ocrv->CntrlPts[i].dCoord[0] = i_crv->control[i].x;
    xyz_ocrv->CntrlPts[i].dCoord[1] = i_crv->control[i].y;
    xyz_ocrv->CntrlPts[i].dCoord[2] = i_crv->control[i].z;
  }
}
