#include "defs.h"
#include <stdio.h>

extern	allocDeterminant_1D();

void
change_to_uv(ndet_xyzw, ndet1D_uv, tpPatch)
NDETERMINANT 		ndet_xyzw;
NDETERMINANT_1D 	ndet1D_uv;
TENSOR_PATCH		tpPatch;
{
  int		i, j, k, l, w;


  for(i = 0; i < ndet1D_uv.cRows; i++)	   /* each element of matrix ndet1_uv*/
    for(j = 0; j < ndet1D_uv.cCols; j++)
    { for(k = 0; k <= tpPatch.iDeg0; k++)  /* each term (k,l) of poly(u,v) at */
	for(l = 0; l <= tpPatch.iDeg1; l++) /* location (i,j) of ndet1_uv */
	{ ndet1D_uv.poly[i][j].coeffs[k][l] = 0.0;

	  /* aX(u,v) + bY(u,v) + cZ(u,v) + dW(u,v) since
	     we substituted the parametrics for the 
	     x,y,z,w */
	  for(w = 0; w < 4; w++) 
	    ndet1D_uv.poly[i][j].coeffs[k][l] += 
	      ndet_xyzw.coeffs[i][j].dCoord[w]*tpPatch.coeffs[k][l].dCoord[w];
         }
      }	  
}

/* Compute the partial derivative of ndet1D_uv and stores it in ndet1D_dx 
   Assumes that : 
   . ndet1D_uv is in the Bernstein basis 
   . members of the struct (ndet1D_dx) are not allocated
   We use:
   dB(n, i, u)/du =  n(B(n-1, i-1, u) - B(n-1, i, u)) 
   and
   dB(n, 0, u)/du = -n(B(n-1, 0, u)), 
   dB(n, n, u)/du = n(B(n-1, n-1, u)), 
*/
void
derivative_uv(ndet1D_uv, ndet1D_dx, x)
NDETERMINANT_1D 	ndet1D_uv;
NDETERMINANT_1D 	*ndet1D_dx;
int			x;
{ 
  int			i, j, k, l, n;
  int			max_pow_u, max_pow_v;

  max_pow_u = ndet1D_uv.poly[0][0].iDeg0;
  max_pow_v = ndet1D_uv.poly[0][0].iDeg1;

  /* allocate members of the struct ndet1D_dx */
  if (x == U) max_pow_u--;
  else max_pow_v--;

  allocDeterminant_1D(ndet1D_dx, ndet1D_uv.cRows, ndet1D_uv.cCols, max_pow_u, 
			max_pow_v);

  for(i = 0; i < ndet1D_uv.cRows; i++)	   /* each element of matrix ndet1_uv*/
    for(j = 0; j < ndet1D_uv.cCols; j++)
    { for(k = 0; k <= max_pow_u; k++)  
	for(l = 0; l <= max_pow_v; l++)
         ndet1D_dx->poly[i][j].coeffs[k][l] = 0.0;
      if (x == V)			   /* compute partial deriv wrt v */
      { for(k = 0; k <= ndet1D_uv.poly[i][j].iDeg0; k++)  
	{ n = ndet1D_uv.poly[i][j].iDeg1;
	  ndet1D_dx->poly[i][j].coeffs[k][0] = 
			-n*ndet1D_uv.poly[i][j].coeffs[k][0];
	  ndet1D_dx->poly[i][j].coeffs[k][n-1]=
			n*ndet1D_uv.poly[i][j].coeffs[k][n];
	  for(l = 1; l < n; l++)
	  { ndet1D_dx->poly[i][j].coeffs[k][l-1] += 
			n*ndet1D_uv.poly[i][j].coeffs[k][l];
	    ndet1D_dx->poly[i][j].coeffs[k][l] -= 
			n*ndet1D_uv.poly[i][j].coeffs[k][l];
	  }
        }
      }
      if (x == U)			   /* compute partial deriv wrt u */
      { for(l = 0; l <= ndet1D_uv.poly[i][j].iDeg1; l++)  
	{ n = ndet1D_uv.poly[i][j].iDeg0;
	  ndet1D_dx->poly[i][j].coeffs[0][l] = 
			-n*ndet1D_uv.poly[i][j].coeffs[0][l];
	  ndet1D_dx->poly[i][j].coeffs[n-1][l]=
			n*ndet1D_uv.poly[i][j].coeffs[n][l];
	  for(k = 1; k < n; k++)
	  { ndet1D_dx->poly[i][j].coeffs[k-1][l] += 
			n*ndet1D_uv.poly[i][j].coeffs[k][l];
	    ndet1D_dx->poly[i][j].coeffs[k][l] -= 
			n*ndet1D_uv.poly[i][j].coeffs[k][l];
	  }
        }
      }
    }
}


void 
eval_uv(ndet1D_uv, u, v)
NDETERMINANT_1D		ndet1D_uv;
double			u, v;
{ 
  double		*ber_u, *ber_v;
  int			m, n; /* m is the max deg u, n is the max deg v */
  int			i, j, k, l;

  /* find the maximum degrees of u and v */
  m = ndet1D_uv.poly[0][0].iDeg0;
  n = ndet1D_uv.poly[0][0].iDeg1;
 
  /* evaluate all the bernstein polynomials of u and v */
  ALLOCN(ber_u, double, (m+1));
  eval_ber(ber_u, u, m);
  ALLOCN(ber_v, double, (n+1));
  eval_ber(ber_v, v, n);

  /* evaluate all the polynomials */
  ndet1D_uv.u = u;
  ndet1D_uv.v = v;
  for(i = 0; i < ndet1D_uv.cRows; i++)	   
    for(j = 0; j < ndet1D_uv.cCols; j++)
    { ndet1D_uv.eval[i][j] = 0.0;
      for(k = 0; k <= ndet1D_uv.poly[i][j].iDeg0; k++)   
	for(l = 0; l <= ndet1D_uv.poly[i][j].iDeg1; l++) 
        { /* each term (k,l) of poly(u,v) at location (i,j) of ndet1_uv */
          ndet1D_uv.eval[i][j] += 
          ndet1D_uv.poly[i][j].coeffs[k][l]*ber_u[k]*ber_v[l];
        }	  
    }

  /* free up all memory */
  myfree(ber_u); 
  myfree(ber_v);
}

/* evaluate all bernstein polynomials for var x */
eval_ber(ber, x, m)
double		*ber;
double		x;
int		m;
{
  int		i, j, k;
  long		*fact; 
  double	*pow_x, *pow_1_x;
  double	one_x;

  /* evaluate all factorials to m */
  ALLOCN(fact, long, m+1);
  fact[0] = 1;	/* just for our purposes */
  for(i = 1; i <= m; i++)
    fact[i] = i*fact[i-1];

  /* evaluate all the powers of x and (1-x) */
  ALLOCN(pow_x, double, (m+1));
  ALLOCN(pow_1_x, double, (m+1));
  pow_x[0] = 1;
  pow_1_x[0] = 1;
  one_x = 1 - x;
  for(i = 1; i <= m; i++)
  { pow_x[i]   = x*pow_x[i-1];
    pow_1_x[i] = one_x*pow_1_x[i-1];
  }

  /* evaluate all the bernstein polynomials of x */
  for(i = 0; i <= m; i++)
    ber[i] = combChoose(m,i)*pow_x[i]*pow_1_x[m-i];

  myfree(fact); myfree(pow_x); myfree(pow_1_x);
}
