#include "defs.h"

COMPLEX plus(a, b)
COMPLEX a, b;
{
  COMPLEX res;

  res.Re = a.Re + b.Re;
  res.Im = a.Im + b.Im;
  return res;
}

COMPLEX minus(a, b)
COMPLEX a, b;
{
  COMPLEX res;

  res.Re = a.Re - b.Re;
  res.Im = a.Im - b.Im;
  return res;
}

COMPLEX mult(a, b)
COMPLEX a, b;
{
  COMPLEX res;
  double p, q, r;

  p = a.Re * ( b.Re + b.Im );
  q = b.Im * ( a.Re + a.Im );
  r = b.Re * ( a.Im - a.Re );
  res.Re = p - q;
  res.Im = p + r;
  return res;
}

COMPLEX recip(a)
COMPLEX a;
{
  COMPLEX res;
  double p;

  p = a.Re * a.Re + a.Im * a.Im;
  res.Re = a.Re / p;
  res.Im = -a.Im / p;
  return res;
}

COMPLEX divid(a,b)
COMPLEX a, b;
{
  COMPLEX rec, res;

  rec = recip(b);
  return mult(a, rec);
}

COMPLEX sqroot(a)
COMPLEX a;
{
  COMPLEX res;
  double p, q;

  p = sqrt(a.Re * a.Re + a.Im * a.Im);

  res.Re = sqrt((p + a.Re) / 2.0);
  res.Im = sqrt((p - a.Re) / 2.0);
  return res;
}
  
COMPLEX  cxpower_pt(z)
COMPLEX z;
{
   COMPLEX one_minus_z;

   one_minus_z.Re = 1.0 - z.Re;
   one_minus_z.Im = -z.Im;

   if (FP_EQ(one_minus_z.Re, 0.0) && FP_EQ(one_minus_z.Im, 0.0)) 
   {
     one_minus_z.Re = HUGE_VAL;
     one_minus_z.Im = 0.0;
     return one_minus_z;
   }
   return divid(z, one_minus_z);
}

COMPLEX  inv_cxpower_pt(z)
COMPLEX z;
{
   COMPLEX one_plus_z;

   if (z.Re == HUGE_VAL) {
     one_plus_z.Re = 1.0;
     one_plus_z.Im = 0.0;
     return one_plus_z;
   }
   one_plus_z.Re = 1.0 + z.Re;
   one_plus_z.Im = z.Im;

   return divid(z, one_plus_z);
}
 
COMPLEX  trans_inv_cxpower_pt(z)
COMPLEX z;
{
   COMPLEX one_plus_z, c1;

   if (z.Re == HUGE_VAL) {
     one_plus_z.Re = 0.5;
     one_plus_z.Im = 0.0;
     return one_plus_z;
   }
   one_plus_z.Re = 1.0 + z.Re;
   one_plus_z.Im = z.Im;
   c1.Re = 1.0 + 2.0*z.Re;
   c1.Im = 2.0*z.Im;

   return divid(one_plus_z, c1);
}
 
