#include "defs.h"
EXTERN_LINALG

long fact( a )
    int a;
{
    long lAcc=1;

    for(;a>0;lAcc*=a, a--);

    return lAcc;
}

long combChoose( a, b )
    int a, b;
{
    int  i;
    long numer, denom;
    double half;
    
    if (b > a) return 0;
    half = a / 2.0;
    if (b > half) b = a - b;
    numer = denom = 1;
    for (i=0; i<b; i++) {
      numer *= a - i;
      denom *= b - i;
    }
    return numer / denom;
}


void bernsteinToPowerBasis_1D( tp_1D )
    TENSOR_PATCH_1D   tp_1D;
{
    int          i, j;


    /* convert from the bernstein to the power basis by the simple
       reparametrization s'=s/(1-s), t'=t/(1-t),

       This converts the coeffs from Xij to Xij' = (m choose i)*
       (n choose j)*Xij (from Manocha 91, a new approach for surface
       intersection) */

	for (i=0;i<=tp_1D.iDeg0;i++)
	    for (j=0;j<=tp_1D.iDeg1;j++) 
		tp_1D.coeffs[i][j] = combChoose( tp_1D.iDeg0, i ) *
		        combChoose( tp_1D.iDeg1, j ) * tp_1D.coeffs[i][j];
}

void bernsteinToPowerBasis( tp )
    TENSOR_PATCH   tp;
{
    int          i, j, iCoord;


    /* convert from the bernstein to the power basis by the simple
       reparametrization s'=s/(1-s), t'=t/(1-t),

       This converts the coeffs from Xij to Xij' = (m choose i)*
       (n choose j)*Xij (from Manocha 91, a new approach for surface
       intersection) */

    for (iCoord=X;iCoord<=W;iCoord++) 
	for (i=0;i<=tp.iDeg0;i++)
	    for (j=0;j<=tp.iDeg1;j++) {
		tp.coeffs[i][j].dCoord[iCoord] *= combChoose( tp.iDeg0, i ) *
		        combChoose( tp.iDeg1, j );
	    }
}

void powerToBernsteinBasis( tp )
    TENSOR_PATCH   tp;
{
    int          i, j, iCoord;


    /* convert from the power to Bernstein basis by the simple
       reparametrization s=s'/(1+s'),
       (see above) */

    for (iCoord=X;iCoord<=W;iCoord++) 
	for (i=0;i<=tp.iDeg0;i++)
	    for (j=0;j<=tp.iDeg1;j++) {
		tp.coeffs[i][j].dCoord[iCoord] /= (combChoose( tp.iDeg0, i ) *
		        combChoose( tp.iDeg1, j ));
	    }
}


void
evaluateBernstein( tp, p2D, pp4D )
TENSOR_PATCH tp;
POINT2D      p2D;
POINT4D      *pp4D;
{
    int    i,j,k;
    double u,v,uprime,vprime;
    double *us, *vs, *uprimes, *vprimes;

    u = p2D.u;
    v = p2D.v;
    uprime = 1.0 - u;
    vprime = 1.0 - v;

    ALLOCN(us, double, tp.iDeg0+1);
    ALLOCN(vs, double, tp.iDeg1+1);
    ALLOCN(uprimes, double, tp.iDeg0+1);
    ALLOCN(vprimes, double, tp.iDeg1+1);
    us[0]=vs[0]=uprimes[0]=vprimes[0]=1.0;

    for (i=1;i<=tp.iDeg0;i++) {
	us[i]=us[i-1]*u;
	uprimes[i]=uprimes[i-1]*uprime;
    }

    for (i=1;i<=tp.iDeg1;i++) {
	vs[i]=vs[i-1]*v;
	vprimes[i]=vprimes[i-1]*vprime;
    }

    pp4D->dCoord[X]=0.0;
    pp4D->dCoord[Y]=0.0;
    pp4D->dCoord[Z]=0.0;
    pp4D->dCoord[W]=0.0;
    for (i=0;i<=tp.iDeg0;i++)
	for (j=0;j<=tp.iDeg1;j++)
	    for (k=X;k<=W;k++)
		pp4D->dCoord[k] += tp.coeffs[i][j].dCoord[k] *
		    combChoose( tp.iDeg0, i ) * 
			us[i] * uprimes[tp.iDeg0-i] *
		    combChoose( tp.iDeg1, j ) *
			vs[j] * vprimes[tp.iDeg1-j];

    myfree(us);
    myfree(vs);
    myfree(uprimes);
    myfree(vprimes);
}

/* evaluate the normal to a Bezier surface at a given point */
void
evaluateNormal( tp, p2D, pp4D )
TENSOR_PATCH tp;
POINT2D      p2D;
POINT4D      *pp4D;
{
   
   static int	     	first = 1;
   static TENSOR_PATCH  tpTemp;
   POINT4D    		UTangent, VTangent;
   int			i, j;

   tpTemp.iDeg0 = tp.iDeg0;
   tpTemp.iDeg1 = tp.iDeg1;

   if (first)
   { first = 0;
     allocate_tensor_patch( &tpTemp);
   }

   /* evaluate UTangent */
   tpTemp.iDeg0--;
   for(i = 0; i <= tpTemp.iDeg0; i++)
     for(j = 0; j <= tpTemp.iDeg1; j++)
     { tpTemp.coeffs[i][j].dCoord[X] = tp.coeffs[i+1][j].dCoord[X] - tp.coeffs[i][j].dCoord[X];
       tpTemp.coeffs[i][j].dCoord[Y] = tp.coeffs[i+1][j].dCoord[Y] - tp.coeffs[i][j].dCoord[Y];
       tpTemp.coeffs[i][j].dCoord[Z] = tp.coeffs[i+1][j].dCoord[Z] - tp.coeffs[i][j].dCoord[Z];
       tpTemp.coeffs[i][j].dCoord[W] = tp.coeffs[i+1][j].dCoord[W] - tp.coeffs[i][j].dCoord[W];
     }
   evaluateBernstein(tpTemp, p2D, &UTangent);
   tpTemp.iDeg0++;
   VEC_SCALAR_OP(UTangent.dCoord, UTangent.dCoord, *, tpTemp.iDeg0, 4);

   /* evaluate VTangent */
   tpTemp.iDeg1--;
   for(i = 0; i <= tpTemp.iDeg0; i++)
     for(j = 0; j <= tpTemp.iDeg1; j++)
     { tpTemp.coeffs[i][j].dCoord[X] = tp.coeffs[i][j+1].dCoord[X] - tp.coeffs[i][j].dCoord[X];
       tpTemp.coeffs[i][j].dCoord[Y] = tp.coeffs[i][j+1].dCoord[Y] - tp.coeffs[i][j].dCoord[Y];
       tpTemp.coeffs[i][j].dCoord[Z] = tp.coeffs[i][j+1].dCoord[Z] - tp.coeffs[i][j].dCoord[Z];
       tpTemp.coeffs[i][j].dCoord[W] = tp.coeffs[i][j+1].dCoord[W] - tp.coeffs[i][j].dCoord[W];
     }
   evaluateBernstein(tpTemp, p2D, &VTangent);
   tpTemp.iDeg1++;
   VEC_SCALAR_OP(VTangent.dCoord, VTangent.dCoord, *, tpTemp.iDeg1, 4);

   NORMALIZE(UTangent.dCoord, 3);
   NORMALIZE(VTangent.dCoord, 3);

   CROSSPROD3(pp4D->dCoord, VTangent.dCoord, UTangent.dCoord);
   NORMALIZE(pp4D->dCoord, 3);
}
