#include "defs.h"

typedef struct {
  double x, y;
} point_typ;

extern point_typ GetPtInsideSplinegon();
/* extern void allocate_tensor_patch(); */

void
allocate_tensor_patch_disabled(ptpPatch)
TENSOR_PATCH    *ptpPatch;
{
  POINT4D       *p4dCoeff_space;
  int           tot_space;
  int           i, j;

  /* allocate memory */
  tot_space = (ptpPatch->iDeg0+1)*(ptpPatch->iDeg1+1);
  ALLOCN(ptpPatch->coeffs, POINT4D *, (ptpPatch->iDeg0+1));
  ALLOCN(p4dCoeff_space, POINT4D, tot_space);

  /* set up pointers */
  for(i = j = 0; i <=  ptpPatch->iDeg0; i++)
  { ptpPatch->coeffs[i] = &p4dCoeff_space[j];
    j+= (ptpPatch->iDeg1 + 1);
  }
}

void copy_ten_p(tps, tpd)
TENSOR_PATCH *tps, *tpd;
{
  int i, j, k;

  tpd->iDeg0 = tps->iDeg0;
  tpd->iDeg1 = tps->iDeg1;
  allocate_tensor_patch(tpd);

  for (i=0; i<=tpd->iDeg0; i++)
  for (j=0; j<=tpd->iDeg1; j++)
    for (k=0; k<4; k++)
      tpd->coeffs[i][j].dCoord[k] = tps->coeffs[i][j].dCoord[k];
}

void copy_flip_ten_p(tps, tpd)
TENSOR_PATCH *tps, *tpd;
{
  int i, j, k;

  tpd->iDeg0 = tps->iDeg0;
  tpd->iDeg1 = tps->iDeg1;
  allocate_tensor_patch(tpd);

  for (i=0; i<=tpd->iDeg0; i++)
  for (j=0; j<=tpd->iDeg1; j++)
    for (k=0; k<4; k++)
      tpd->coeffs[i][j].dCoord[k] = tps->coeffs[i][(tpd->iDeg1)-j].dCoord[k];
}

/* The next two routines copies one spline into another after */
/* allcoating */
void copy_uv_spline(in, out)
UV_SPLINE_TYPE *in, *out;
{
  int i;
  
  out->Deg = in->Deg;
  out->CntrlPts = (POINT3D *) new((out->Deg + 2) * sizeof(POINT3D));
  for (i=0; i<=out->Deg; i++)
    out->CntrlPts[i] = in->CntrlPts[i];
  for (i=0; i<3; i++)
    out->CntrlPts[out->Deg+1].dCoord[i] = -1.0;
}

void copy_flip_uv_spline(in, out)
UV_SPLINE_TYPE *in, *out;
{
  int i;
  
  out->Deg = in->Deg;
  out->CntrlPts = (POINT3D *) new((out->Deg + 2) * sizeof(POINT3D));
  for (i=0; i<=out->Deg; i++) {
    out->CntrlPts[i].dCoord[0] = in->CntrlPts[out->Deg - i].dCoord[0];

    /* Actually v becomes 1-v when flipping, but since it is in */
    /* homogeneous coords., we have to do (1 - v/w)*w = w-v */
    out->CntrlPts[i].dCoord[1] = (in->CntrlPts[out->Deg - i].dCoord[2] -
				  in->CntrlPts[out->Deg - i].dCoord[1]);
    out->CntrlPts[i].dCoord[2] = in->CntrlPts[out->Deg - i].dCoord[2];
  }
  for (i=0; i<3; i++)
    out->CntrlPts[out->Deg+1].dCoord[i] = -1.0;
}

void allocate_uv_spline(spline)
UV_SPLINE_TYPE *spline;
{
  ALLOCN(spline->CntrlPts, POINT3D, spline->Deg+1);
}

void allocate_xyz_spline(spline)
XYZ_SPLINE_TYPE *spline;
{
  ALLOCN(spline->CntrlPts, POINT4D, spline->Deg+1);
}

void copy_xyz_spline(in, out)
XYZ_SPLINE_TYPE *in, *out;
{
  int i;
  
  out->Deg = in->Deg;
  out->CntrlPts = (POINT4D *) new((out->Deg + 2) * sizeof(POINT4D));
  for (i=0; i<=out->Deg; i++)
    out->CntrlPts[i] = in->CntrlPts[i];
  for (i=0; i<4; i++)
    out->CntrlPts[out->Deg+1].dCoord[i] = -1.0;
}

void copy_flip_xyz_spline(in, out)
XYZ_SPLINE_TYPE *in, *out;
{
  int i;
  
  out->Deg = in->Deg;
  out->CntrlPts = (POINT4D *) new((out->Deg + 2) * sizeof(POINT4D));
  for (i=0; i<=out->Deg; i++)
    out->CntrlPts[i] = in->CntrlPts[out->Deg - i];
  for (i=0; i<4; i++)
    out->CntrlPts[out->Deg+1].dCoord[i] = -1.0;
}

/* This copies tpsts into tpstd. If find_inside_point is true, it also */
/* finds an inside point */
void copy_trim_portion_spline_type(tpstd, tpsts, find_inside_point)
TRIM_PORTION_SPLINE_TYPE *tpstd, *tpsts;
short find_inside_point;
{
  point_typ pt;
  int i;

  tpstd->NumOfCurves = tpsts->NumOfCurves;     

/* I have added 8 because I want to be able to partition some curves, */
/* thereby increasing the count */ 
  ALLOCN(tpstd->UV_Splines, UV_SPLINE_TYPE, tpstd->NumOfCurves + 8);
  ALLOCN(tpstd->XYZ_Splines, XYZ_SPLINE_TYPE, tpstd->NumOfCurves + 8);
  ALLOCN(tpstd->AdjInfo, int, tpstd->NumOfCurves + 8);

  for (i=0; i<tpsts->NumOfCurves; i++) {
    copy_uv_spline(&tpsts->UV_Splines[i], &tpstd->UV_Splines[i]);
    copy_xyz_spline(&tpsts->XYZ_Splines[i], &tpstd->XYZ_Splines[i]);
    tpstd->AdjInfo[i] = tpsts->AdjInfo[i];
  }

  if (find_inside_point) {
    pt = GetPtInsideSplinegon(tpstd);
    tpstd->ptInsideRegion.u = pt.x;
    tpstd->ptInsideRegion.v = pt.y;
  } else
    tpstd->ptInsideRegion = tpsts->ptInsideRegion;
}

void copy_flip_trim_portion_spline_type(tpstd, tpsts, find_inside_point)
TRIM_PORTION_SPLINE_TYPE *tpstd, *tpsts;
short find_inside_point;
{
  point_typ pt;
  int i, ncrv;

  ncrv = tpstd->NumOfCurves = tpsts->NumOfCurves;
  ALLOCN(tpstd->UV_Splines, UV_SPLINE_TYPE, tpstd->NumOfCurves);
  ALLOCN(tpstd->XYZ_Splines, XYZ_SPLINE_TYPE, tpstd->NumOfCurves);
  ALLOCN(tpstd->AdjInfo, int, tpstd->NumOfCurves);

  for (i=0; i<tpstd->NumOfCurves; i++) {
    copy_flip_uv_spline(&tpsts->UV_Splines[ncrv-i-1], &tpstd->UV_Splines[i]);
    copy_flip_xyz_spline(&tpsts->XYZ_Splines[ncrv-i-1], &tpstd->XYZ_Splines[i]);
    tpstd->AdjInfo[i] = tpsts->AdjInfo[ncrv-i-1];
  }

  if (find_inside_point) {
    pt = GetPtInsideSplinegon(tpstd);
    tpstd->ptInsideRegion.u = pt.x;
    tpstd->ptInsideRegion.v = pt.y;
  } else {
    tpstd->ptInsideRegion.u = tpsts->ptInsideRegion.u;
    tpstd->ptInsideRegion.v = 1.0 - tpsts->ptInsideRegion.v;
  }
}

void copy_tp_piece_type(tps, tpd)
TRIM_PORTION_PIECEWISE_TYPE *tps, *tpd;
{
  int j;

  tpd->NumOfVertices = tps->NumOfVertices;
  tpd->ptInsideRegion = tps->ptInsideRegion;
  
  tpd->poly_vertex = (double (*)[2])
                           new(MAX_VERTEX_COUNT*2*sizeof(double)); 
  if (tps->AdjInfo != NULL) 
    tpd->AdjInfo = (int *) new(MAX_VERTEX_COUNT*sizeof(int));
  for (j=0; j<=tpd->NumOfVertices; j++) {
    tpd->poly_vertex[j][0] = tps->poly_vertex[j][0];
    tpd->poly_vertex[j][1] = tps->poly_vertex[j][1];
    if (tps->AdjInfo != NULL)
      tpd->AdjInfo[j] = tps->AdjInfo[j];
  }
}

void copy_flip_tp_piece_type(tps, tpd)
TRIM_PORTION_PIECEWISE_TYPE *tps, *tpd;
{
  int j;

  tpd->NumOfVertices = tps->NumOfVertices;
  tpd->ptInsideRegion.u = tps->ptInsideRegion.u;
  tpd->ptInsideRegion.v = 1.0 - tps->ptInsideRegion.v;

  tpd->poly_vertex = (double (*)[2])
                           new(MAX_VERTEX_COUNT*2*sizeof(double)); 
  tpd->AdjInfo = (int *) new(MAX_VERTEX_COUNT*sizeof(int));

  for (j=0; j<=tpd->NumOfVertices; j++) {
    tpd->poly_vertex[j][0] = tps->poly_vertex[tpd->NumOfVertices-j][0];
    tpd->poly_vertex[j][1] = 
      1.0 - tps->poly_vertex[tpd->NumOfVertices-j][1];
    tpd->AdjInfo[j] = tps->AdjInfo[j];
  }
}

void copy_dom_int_curve(dics, dicd)
DOMAIN_INT_CURVE *dics, *dicd;
{
  LIST_POINT2D *ptr1, *ptr2, *ptr3;

  dicd->StEndPoints[0].u = dics->StEndPoints[0].u;
  dicd->StEndPoints[0].v = dics->StEndPoints[0].v;
  dicd->StEndPoints[1].u = dics->StEndPoints[1].u;
  dicd->StEndPoints[1].v = dics->StEndPoints[1].v;

  ptr1 = dics->planeCurve;
  if (ptr1 != NULL) {
    ALLOCN(dicd->planeCurve, LIST_POINT2D, 1);
    ptr2 = dicd->planeCurve;
    ptr2->p2D = ptr1->p2D;
    ptr2->plp2DNext = NULL;
    ptr1 = ptr1->plp2DNext;
  } else {
    dicd->planeCurve = NULL;
    return;
  }
 
  while (ptr1 != NULL) {
    ALLOCN(ptr3, LIST_POINT2D, 1);
    ptr3->p2D = ptr1->p2D;
    ptr3->plp2DNext = NULL;
    ptr2->plp2DNext = ptr3;
    ptr2 = ptr3;
    ptr1 = ptr1->plp2DNext;
  }
}

void copy_trim_patch(intpi, outtpi)
TRIM_PATCH_INFO *intpi, *outtpi;
{
  int i, j;

  copy_ten_p(&(intpi->tp), &(outtpi->tp));
  copy_ten_p(&(intpi->orig_tp), &(outtpi->orig_tp));

  outtpi->umin = intpi->umin;
  outtpi->umax = intpi->umax;
  outtpi->vmin = intpi->vmin;
  outtpi->vmax = intpi->vmax;

  outtpi->primNum = intpi->primNum;
  outtpi->NumTrimPortions = intpi->NumTrimPortions;
  outtpi->SplineReprn = intpi->SplineReprn;
  if (!outtpi->SplineReprn)
    for (i=0; i<outtpi->NumTrimPortions; i++)
      copy_tp_piece_type(&intpi->TrimPortion[i].PiecePortion, 
			 &outtpi->TrimPortion[i].PiecePortion); 
  else
    for (i=0; i<outtpi->NumTrimPortions; i++)
      copy_trim_portion_spline_type(&outtpi->TrimPortion[i].SplinePortion,
				    &intpi->TrimPortion[i].SplinePortion,
				    0);
  outtpi->csg_status = intpi->csg_status;
}

void copy_flip_trim_patch(intpi, outtpi)
TRIM_PATCH_INFO *intpi, *outtpi;
{
  int i, j;

  copy_flip_ten_p(&(intpi->tp), &(outtpi->tp));
  copy_flip_ten_p(&(intpi->orig_tp), &(outtpi->orig_tp));

  outtpi->umin = intpi->umin;
  outtpi->umax = intpi->umax;
  outtpi->vmin = intpi->vmin;
  outtpi->vmax = intpi->vmax;

  outtpi->primNum = intpi->primNum;
  outtpi->NumTrimPortions = intpi->NumTrimPortions;
  outtpi->SplineReprn = intpi->SplineReprn;
  if (!outtpi->SplineReprn)
    for (i=0; i<outtpi->NumTrimPortions; i++)
      copy_flip_tp_piece_type(&intpi->TrimPortion[i].PiecePortion, 
			      &outtpi->TrimPortion[i].PiecePortion); 
  else
    for (i=0; i<outtpi->NumTrimPortions; i++)
      copy_flip_trim_portion_spline_type(&outtpi->TrimPortion[i].SplinePortion,
					 &intpi->TrimPortion[i].SplinePortion,
					 0);
  outtpi->csg_status = intpi->csg_status;
}

void free_ten_p(tp)
TENSOR_PATCH *tp;
{
#if 0
  int i;

  for (i=0; i<=tp->iDeg0; i++)
    myfree(tp->coeffs[i]);

  myfree(tp->coeffs);
#endif
  if (tp->iDeg0 || tp->iDeg1) /* if one of them is nonzero */
    myfree(tp->coeffs[0]);
}

void free_list_point4D(p4Dptr)
LIST_POINT4D *p4Dptr;
{
  LIST_POINT4D *ptr1, *ptr2;

  ptr1 = p4Dptr;
  while (ptr1 != NULL) {
    ptr2 = ptr1->plp4DNext;
    myfree(ptr1);
    ptr1 = ptr2;
  }
  p4Dptr = NULL;
}

void free_list_point4CPX(p4CPXptr)
LIST_POINT4CPX *p4CPXptr;
{
  LIST_POINT4CPX *ptr1, *ptr2;

  ptr1 = p4CPXptr;
  while (ptr1 != NULL) {
    ptr2 = ptr1->plp4CPXNext;
    myfree(ptr1);
    ptr1 = ptr2;
  }
  p4CPXptr = NULL;
}

void free_dom_int_curve(dic)
DOMAIN_INT_CURVE *dic;
{
  LIST_POINT2D *ptr1, *ptr2;

  ptr1 = dic->planeCurve;
  while (ptr1 != NULL) {
    ptr2 = ptr1->plp2DNext;
    myfree(ptr1);
    ptr1 = ptr2;
  }
  dic->planeCurve = NULL;
}

void free_merged_curve(mc)
MERGED_CURVE *mc;
{
  INTERSECTION_CURVE *ptr1, *ptr2;

  ptr1 = mc->int_curve;
  while (ptr1 != NULL) {
    ptr2 = ptr1->next;
    myfree(ptr1);
    ptr1 = ptr2;
  }
  mc->int_curve = NULL;
}

void free_trim_portion_piece_type(tppt)
TRIM_PORTION_PIECEWISE_TYPE *tppt;
{
  myfree(tppt->poly_vertex);
}

void free_trim_portion_spline_type(tpst)
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  register int i, j;
  POINT3D *p3d;
  POINT4D *p4d;

/*
  for (i=0; i<tpst->NumOfCurves; i++) {
    p3d = &tpst->UV_Splines[i].CntrlPts[tpst->UV_Splines[i].Deg+1];
    for (j=0; j<3; j++)
      if (!FP_EQ_EPS(p3d->dCoord[j],-1.0,1e-3))
	printf("free_trim_portion_spline_type: Memory Overwritten %d,1\n", i);
    p4d = &tpst->XYZ_Splines[i].CntrlPts[tpst->XYZ_Splines[i].Deg+1];
    for (j=0; j<4; j++)
      if (!FP_EQ_EPS(p4d->dCoord[j],-1.0,1e-3))
	printf("free_trim_portion_spline_type: Memory Overwritten %d,2\n", i);
  }
*/

  for (i=0; i<tpst->NumOfCurves; i++) {
    myfree((void *) tpst->UV_Splines[i].CntrlPts);
    myfree((void *) tpst->XYZ_Splines[i].CntrlPts);
    tpst->UV_Splines[i].CntrlPts = NULL;
    tpst->XYZ_Splines[i].CntrlPts = NULL;
  }
  myfree(tpst->UV_Splines);
  tpst->UV_Splines = NULL;
  myfree(tpst->XYZ_Splines);
  tpst->XYZ_Splines = NULL;
  myfree(tpst->AdjInfo);
  tpst->AdjInfo = NULL;
}

void free_trim_portion(tp, spl_rep)
TRIM_PORTION_TYPE *tp;
int spl_rep;
{
  TRIM_PORTION_SPLINE_TYPE *tpst;
  TRIM_PORTION_PIECEWISE_TYPE *tppt;

  if (spl_rep) {
    tpst = &tp->SplinePortion;
    free_trim_portion_spline_type(tpst);
  } else {
    tppt = &tp->PiecePortion;
    free_trim_portion_piece_type(tppt);
  }
}

void free_uv_spline(uv_spl)
UV_SPLINE_TYPE *uv_spl;
{
  myfree(uv_spl->CntrlPts);
}

void free_xyz_spline(xyz_spl)
XYZ_SPLINE_TYPE *xyz_spl;
{
  myfree(xyz_spl->CntrlPts);
}

void free_trim_patch(tp)
TRIM_PATCH_INFO *tp;
{
  int i, j;
  LIST_POINT2D *ptr1, *ptr2;

  if (tp == NULL) return;

  free_ten_p(&(tp->tp));
  free_ten_p(&(tp->orig_tp));

  for (i=0; i<tp->NumTrimPortions; i++)
    free_trim_portion(&tp->TrimPortion[i], tp->SplineReprn);
}

void free_int_curve_struct(ics)
INT_CURVE_STRUCT *ics;
{
  LIST_POINT4D *ptr1, *ptr2;

  ptr1 = ics->Curve;
  while (ptr1 != NULL) {
    ptr2 = ptr1->plp4DNext;
    myfree(ptr1);
    ptr1 = ptr2;
  }
  ics->Curve = NULL;
}

