/* TODO : debug!
*/

#include <stdlib.h>
#include <stdio.h>
#include "defs.h"
#include "csurf.h"
#include "primitives.h"

EXTERN_LINALG

#define MAXNUMPATCHPTS 100000
#define free myfree

/*#define MINANG 0.1
#define NEWANG 0.1763
*/
#define MINANG 0.1
#define NEWANG 0.0875

/*#define KEYSERDEBUG*/

INTERSECT_STACK_ELEMENT* newinterstackelement(TENSOR_PATCH* pat,
                                              XYZ_SPLINE_TYPE* cur,
                                              double umin, double umax,
                                              double vmin, double vmax,
                                              double tmin, double tmax,
                                              int oper)
{
   INTERSECT_STACK_ELEMENT* temp;

   temp = (INTERSECT_STACK_ELEMENT*) malloc(sizeof(INTERSECT_STACK_ELEMENT));
   if (temp == NULL) {
      printf("RAN OUT OF MEMORY while allocating INTERSECT_STACK_ELEMENT\n");
   }
   temp->next = NULL;
   temp->patch = pat;
   temp->curve = cur;
   temp->umin = umin;
   temp->umax = umax;
   temp->vmin = vmin;
   temp->vmax = vmax;
   temp->tmin = tmin;
   temp->tmax = tmax;
   temp->oper = oper;

   return temp;
}

free_interstackelement (INTERSECT_STACK_ELEMENT* n)
{
   free(n);
}

HIT_PT_ELEMENT* newhitptelement()
{
   HIT_PT_ELEMENT* temp;

   temp = (HIT_PT_ELEMENT*) malloc(sizeof(HIT_PT_ELEMENT));
   if (temp == NULL) {
      printf("RAN OUT OF MEMORY while allocating HIT_PT_ELEMENT\n");
   }
   temp->next = NULL;

   return temp;
}

free_hitptelement (HIT_PT_ELEMENT* n)
{
   free(n);
}

TENSOR_PATCH* newtensorpatch(int deg0, int deg1)
{
   TENSOR_PATCH* temp;

   temp = (TENSOR_PATCH*) malloc(sizeof(TENSOR_PATCH));
   if (temp == NULL) {
      printf("RAN OUT OF MEMORY while allocating TENSOR_PATCH\n");
   }
   temp->iDeg0 = deg0;
   temp->iDeg1 = deg1;
   allocate_tensor_patch(temp);

   return temp;
}


XYZ_SPLINE_TYPE* newxyzspline()
{
   XYZ_SPLINE_TYPE* temp;

   temp = (XYZ_SPLINE_TYPE*) malloc(sizeof(XYZ_SPLINE_TYPE));
   if (temp == NULL) {
      printf("RAN OUT OF MEMORY while allocating XYZ_SPLINE_TYPE\n");
   }

   return temp;
}

int subdivide_bezier_curve(XYZ_SPLINE_TYPE cur,
                           double t1, double t2,
                           XYZ_SPLINE_TYPE* curout)
/* Subdivide the curve within the region 0 <= t1 < t2 <= 1 */
/* The resulting curve is obtained in curout.  Space must already */
/* be allocated in curout to hold all the control points */
/* This should be close to optimal speed */
{
   int i, j, k;
   POINT4D b[MAX_CURVEDEG][MAX_CURVEDEG];
   double c;

   curout->Deg = cur.Deg;

   for(i=0;i<=cur.Deg;i++) {
      b[0][i].dCoord[0] = cur.CntrlPts[i].dCoord[0];
      b[0][i].dCoord[1] = cur.CntrlPts[i].dCoord[1];
      b[0][i].dCoord[2] = cur.CntrlPts[i].dCoord[2];
      b[0][i].dCoord[3] = cur.CntrlPts[i].dCoord[3];
   }

   for(i=1;i<=cur.Deg;i++) {
      for(k=0; k <= (cur.Deg-i); k++) {
         b[i][k].dCoord[0] = (1.0-t1)*b[i-1][k].dCoord[0] + 
                                   t1*b[i-1][k+1].dCoord[0];
         b[i][k].dCoord[1] = (1.0-t1)*b[i-1][k].dCoord[1] + 
                                   t1*b[i-1][k+1].dCoord[1];
         b[i][k].dCoord[2] = (1.0-t1)*b[i-1][k].dCoord[2] + 
                                   t1*b[i-1][k+1].dCoord[2];
         b[i][k].dCoord[3] = (1.0-t1)*b[i-1][k].dCoord[3] + 
                                   t1*b[i-1][k+1].dCoord[3];
      }
   }
   curout->CntrlPts[0].dCoord[0] = b[cur.Deg][0].dCoord[0];
   curout->CntrlPts[0].dCoord[1] = b[cur.Deg][0].dCoord[1];
   curout->CntrlPts[0].dCoord[2] = b[cur.Deg][0].dCoord[2];
   curout->CntrlPts[0].dCoord[3] = b[cur.Deg][0].dCoord[3];


   for (j=1;j<=cur.Deg;j++) {
      for (i=(cur.Deg-j+1); i<=cur.Deg; i++) {
         for (k=0; k <= (cur.Deg-i); k++) {
            b[i][k].dCoord[0] = (1.0-t2)*b[i-1][k].dCoord[0] + 
                                      t2*b[i-1][k+1].dCoord[0];
            b[i][k].dCoord[1] = (1.0-t2)*b[i-1][k].dCoord[1] + 
                                      t2*b[i-1][k+1].dCoord[1];
            b[i][k].dCoord[2] = (1.0-t2)*b[i-1][k].dCoord[2] + 
                                      t2*b[i-1][k+1].dCoord[2];
            b[i][k].dCoord[3] = (1.0-t2)*b[i-1][k].dCoord[3] + 
                                      t2*b[i-1][k+1].dCoord[3];
         }
      }
      curout->CntrlPts[j].dCoord[0] = b[cur.Deg][0].dCoord[0];
      curout->CntrlPts[j].dCoord[1] = b[cur.Deg][0].dCoord[1];
      curout->CntrlPts[j].dCoord[2] = b[cur.Deg][0].dCoord[2];
      curout->CntrlPts[j].dCoord[3] = b[cur.Deg][0].dCoord[3];
   }
}

clipit(TENSOR_PATCH* pat, XYZ_SPLINE_TYPE* cur, int oper, double* utmin, 
       double* utmax, double* vmin, double* vmax)
/* This routine determines the points at which to Bezier clip - based
   on Sederberg's algorithm */
{
   double a[6];
   double b[6];
   double c[6];
   double d[6];
   POINT3D vec[5];
   double min[3];
   double max[3];
   double size1=0.0;
   double size2=0.0;
   double size3=0.0;
   double size4=0.0;
   double size5=0.0;
   double pt;
   double** st1;
   double** st2;
   double hmax1, hmax2;
   double hmin1, hmin2;
   double cmax1, cmax2;
   double cmin1, cmin2;
   double val1, val2;
   double amt;
   int i,j,k;
   int hullhit1, hullhit2;
   int firstisabove1, firstisabove2;
   int lastisabove1, lastisabove2;


#ifdef KEYSERDEBUG
printf("patch:\n");
for(i=0;i<=pat->iDeg0;i++) {
   for(j=0;j<=pat->iDeg1;j++) {
      printf("%d,%d: %lf %lf %lf %lf\n",i,j,pat->coeffs[i][j].dCoord[0],
                pat->coeffs[i][j].dCoord[1],pat->coeffs[i][j].dCoord[2],
                pat->coeffs[i][j].dCoord[3]);
   }
}
printf("curve:\n");
for(i=0;i<=cur->Deg;i++) {
      printf("%d: %lf %lf %lf %lf\n",i,cur->CntrlPts[i].dCoord[0],
            cur->CntrlPts[i].dCoord[1],cur->CntrlPts[i].dCoord[2],
            cur->CntrlPts[i].dCoord[3]);
}
#endif

   if (oper==0) {
#ifdef KEYSERDEBUG
printf("PATCH CLIP\n");
#endif
      /* clip patch against curve */
      /* first determine three axes of parallelepiped */
      for (i=0;i<3;i++) {
         vec[0].dCoord[i] = (cur->CntrlPts[cur->Deg].dCoord[i] /
                             cur->CntrlPts[cur->Deg].dCoord[3]) -
                            (cur->CntrlPts[0].dCoord[i] /
                             cur->CntrlPts[0].dCoord[3]);
         size1 += vec[0].dCoord[i]*vec[0].dCoord[i];
         vec[1].dCoord[i] = (pat->coeffs[0][0].dCoord[i] /
                             pat->coeffs[0][0].dCoord[3]) -
                            (pat->coeffs[pat->iDeg0][0].dCoord[i] /
                             pat->coeffs[pat->iDeg0][0].dCoord[3]);
         size2 += vec[1].dCoord[i]*vec[1].dCoord[i];
      }

      if(size1 == 0.0) {
         size1=1.0;
         vec[0].dCoord[0] = 1.0;
         vec[0].dCoord[1] = 0.0;
         vec[0].dCoord[2] = 0.0;
      }
      if(size2 == 0.0) {
         size2=1.0;
         vec[1].dCoord[0] = 0.0;
         vec[1].dCoord[1] = 1.0;
         vec[1].dCoord[2] = 0.0;
      }

      size1=sqrt(size1);
      size2=sqrt(size2);
      for (i=0;i<3;i++) {
         vec[0].dCoord[i] = vec[0].dCoord[i]/size1;
         vec[1].dCoord[i] = vec[1].dCoord[i]/size2;
      }

/* Do I need to check to make sure vec0 != vec1 here? */

      amt = vec[0].dCoord[0]*vec[1].dCoord[0] +
            vec[0].dCoord[1]*vec[1].dCoord[1] +
            vec[0].dCoord[2]*vec[1].dCoord[2];
      if ((amt > 0.865 ) || (amt < MINANG)) {
         /* vec0 and vec1 are within 30 degrees of or are
            close to perpendicular - reassign vec1 */
         vec[3].dCoord[0] = vec[0].dCoord[1]*vec[1].dCoord[2] -
                            vec[0].dCoord[2]*vec[1].dCoord[1];
         vec[3].dCoord[1] = vec[0].dCoord[2]*vec[1].dCoord[0] -
                            vec[0].dCoord[0]*vec[1].dCoord[2];
         vec[3].dCoord[2] = vec[0].dCoord[0]*vec[1].dCoord[1] -
                            vec[0].dCoord[1]*vec[1].dCoord[0];
         size4 = vec[3].dCoord[0]*vec[3].dCoord[0] +
                 vec[3].dCoord[1]*vec[3].dCoord[1] +
                 vec[3].dCoord[2]*vec[3].dCoord[2];
         size4 = sqrt(size4);
         vec[3].dCoord[0]=vec[3].dCoord[0]/size4;
         vec[3].dCoord[1]=vec[3].dCoord[1]/size4;
         vec[3].dCoord[2]=vec[3].dCoord[2]/size4;

         vec[4].dCoord[0] = vec[3].dCoord[1]*vec[0].dCoord[2] -
                            vec[3].dCoord[2]*vec[0].dCoord[1];
         vec[4].dCoord[1] = vec[3].dCoord[2]*vec[0].dCoord[0] -
                            vec[3].dCoord[0]*vec[0].dCoord[2];
         vec[4].dCoord[2] = vec[3].dCoord[0]*vec[0].dCoord[1] -
                            vec[3].dCoord[1]*vec[0].dCoord[0];
         size5 = vec[4].dCoord[0]*vec[4].dCoord[0] +
                 vec[4].dCoord[1]*vec[4].dCoord[1] +
                 vec[4].dCoord[2]*vec[4].dCoord[2];
         size5 = sqrt(size5);
         vec[4].dCoord[0]=vec[4].dCoord[0]/size5;
         vec[4].dCoord[1]=vec[4].dCoord[1]/size5;
         vec[4].dCoord[2]=vec[4].dCoord[2]/size5;

         vec[1].dCoord[0] = vec[4].dCoord[0]+NEWANG*vec[0].dCoord[0];
         vec[1].dCoord[1] = vec[4].dCoord[1]+NEWANG*vec[0].dCoord[1];
         vec[1].dCoord[2] = vec[4].dCoord[2]+NEWANG*vec[0].dCoord[2];
         size2 = vec[1].dCoord[0]*vec[1].dCoord[0] +
                 vec[1].dCoord[1]*vec[1].dCoord[1] +
                 vec[1].dCoord[2]*vec[1].dCoord[2];
         size2 = sqrt(size2);
         vec[1].dCoord[0]=vec[1].dCoord[0]/size2;
         vec[1].dCoord[1]=vec[1].dCoord[1]/size2;
         vec[1].dCoord[2]=vec[1].dCoord[2]/size2;

      }

      for (i=0;i<3;i++) {
         vec[2].dCoord[i] = (pat->coeffs[0][0].dCoord[i] /
                             pat->coeffs[0][0].dCoord[3]) -
                            (pat->coeffs[0][pat->iDeg1].dCoord[i] /
                             pat->coeffs[0][pat->iDeg1].dCoord[3]);
         size3 += vec[2].dCoord[i]*vec[2].dCoord[i];
      }
      if(size3 == 0.0) {
         size3=1.0;
         vec[2].dCoord[0] = 0.0;
         vec[2].dCoord[1] = 0.0;
         vec[2].dCoord[2] = 1.0;
      }
      size3=sqrt(size3);
      for (i=0;i<3;i++) {
         vec[2].dCoord[i] = vec[2].dCoord[i]/size3;
      }

      amt = vec[0].dCoord[0]*vec[2].dCoord[0] +
            vec[0].dCoord[1]*vec[2].dCoord[1] +
            vec[0].dCoord[2]*vec[2].dCoord[2];
      if ((amt > 0.865 ) || (amt < MINANG)) {
         /* vec0 and vec2 are within 30 degrees of or are more than
            close to perpendicular - reassign vec2 */
         vec[3].dCoord[0] = vec[0].dCoord[1]*vec[2].dCoord[2] -
                            vec[0].dCoord[2]*vec[2].dCoord[1];
         vec[3].dCoord[1] = vec[0].dCoord[2]*vec[2].dCoord[0] -
                            vec[0].dCoord[0]*vec[2].dCoord[2];
         vec[3].dCoord[2] = vec[0].dCoord[0]*vec[2].dCoord[1] -
                            vec[0].dCoord[1]*vec[2].dCoord[0];
         size4 = vec[3].dCoord[0]*vec[3].dCoord[0] +
                 vec[3].dCoord[1]*vec[3].dCoord[1] +
                 vec[3].dCoord[2]*vec[3].dCoord[2];
         size4 = sqrt(size4);
         vec[3].dCoord[0]=vec[3].dCoord[0]/size4;
         vec[3].dCoord[1]=vec[3].dCoord[1]/size4;
         vec[3].dCoord[2]=vec[3].dCoord[2]/size4;

         vec[4].dCoord[0] = vec[3].dCoord[1]*vec[0].dCoord[2] -
                            vec[3].dCoord[2]*vec[0].dCoord[1];
         vec[4].dCoord[1] = vec[3].dCoord[2]*vec[0].dCoord[0] -
                            vec[3].dCoord[0]*vec[0].dCoord[2];
         vec[4].dCoord[2] = vec[3].dCoord[0]*vec[0].dCoord[1] -
                            vec[3].dCoord[1]*vec[0].dCoord[0];
         size5 = vec[4].dCoord[0]*vec[4].dCoord[0] +
                 vec[4].dCoord[1]*vec[4].dCoord[1] +
                 vec[4].dCoord[2]*vec[4].dCoord[2];
         size5 = sqrt(size5);
         vec[4].dCoord[0]=vec[4].dCoord[0]/size5;
         vec[4].dCoord[1]=vec[4].dCoord[1]/size5;
         vec[4].dCoord[2]=vec[4].dCoord[2]/size5;

         vec[2].dCoord[0] = vec[4].dCoord[0]+NEWANG*vec[0].dCoord[0];
         vec[2].dCoord[1] = vec[4].dCoord[1]+NEWANG*vec[0].dCoord[1];
         vec[2].dCoord[2] = vec[4].dCoord[2]+NEWANG*vec[0].dCoord[2];
         size3 = vec[2].dCoord[0]*vec[2].dCoord[0] +
                 vec[2].dCoord[1]*vec[2].dCoord[1] +
                 vec[2].dCoord[2]*vec[2].dCoord[2];
         size3 = sqrt(size3);
         vec[2].dCoord[0]=vec[2].dCoord[0]/size3;
         vec[2].dCoord[1]=vec[2].dCoord[1]/size3;
         vec[2].dCoord[2]=vec[2].dCoord[2]/size3;

      }

      amt = vec[2].dCoord[0]*vec[1].dCoord[0] +
            vec[2].dCoord[1]*vec[1].dCoord[1] +
            vec[2].dCoord[2]*vec[1].dCoord[2];
      if (amt > 0.865 ) {
         /* vec2 and vec1 are within 30 degrees of each other
            - reassign vec2 */
         vec[3].dCoord[0] = vec[0].dCoord[1]*vec[1].dCoord[2] -
                            vec[0].dCoord[2]*vec[1].dCoord[1];
         vec[3].dCoord[1] = vec[0].dCoord[2]*vec[1].dCoord[0] -
                            vec[0].dCoord[0]*vec[1].dCoord[2];
         vec[3].dCoord[2] = vec[0].dCoord[0]*vec[1].dCoord[1] -
                            vec[0].dCoord[1]*vec[1].dCoord[0];
         size4 = vec[3].dCoord[0]*vec[3].dCoord[0] +
                 vec[3].dCoord[1]*vec[3].dCoord[1] +
                 vec[3].dCoord[2]*vec[3].dCoord[2];
         size4 = sqrt(size4);
         vec[3].dCoord[0]=vec[3].dCoord[0]/size4;
         vec[3].dCoord[1]=vec[3].dCoord[1]/size4;
         vec[3].dCoord[2]=vec[3].dCoord[2]/size4;

         vec[2].dCoord[0] = vec[3].dCoord[0]+NEWANG*vec[0].dCoord[0];
         vec[2].dCoord[1] = vec[3].dCoord[1]+NEWANG*vec[0].dCoord[1];
         vec[2].dCoord[2] = vec[3].dCoord[2]+NEWANG*vec[0].dCoord[2];
         size3 = vec[2].dCoord[0]*vec[2].dCoord[0] +
                 vec[2].dCoord[1]*vec[2].dCoord[1] +
                 vec[2].dCoord[2]*vec[2].dCoord[2];
         size3 = sqrt(size3);
         vec[2].dCoord[0]=vec[2].dCoord[0]/size3;
         vec[2].dCoord[1]=vec[2].dCoord[1]/size3;
         vec[2].dCoord[2]=vec[2].dCoord[2]/size3;

      }

#ifdef KEYSERDEBUG
for(i=0;i<3;i++) {
printf("vec%d: %lf %lf %lf\n",i,vec[i].dCoord[0],vec[i].dCoord[1],vec[i].dCoord[2]);
}
#endif
      /* now determine extents of parallelepiped */
      for (k=0;k<3;k++) {
         /* project control point onto axis, set initial min,max */
         min[k] = (cur->CntrlPts[0].dCoord[0] * vec[k].dCoord[0] +
                   cur->CntrlPts[0].dCoord[1] * vec[k].dCoord[1] +
                   cur->CntrlPts[0].dCoord[2] * vec[k].dCoord[2]) /
                  cur->CntrlPts[0].dCoord[3];
         max[k] = min[k];
      }
/*for(k=0;k<3;k++) {
printf("%d: min,max : %lf %lf\n",k,min[k],max[k]);
}
*/      for (i=1;i<=cur->Deg;i++) {
         for (k=0;k<3;k++) {
            /* project control point onto axis, change min,max if needed */
            pt = (cur->CntrlPts[i].dCoord[0] * vec[k].dCoord[0] +
                  cur->CntrlPts[i].dCoord[1] * vec[k].dCoord[1] +
                  cur->CntrlPts[i].dCoord[2] * vec[k].dCoord[2]) /
                 cur->CntrlPts[i].dCoord[3];
            if (pt < min[k]) min[k] = pt;
            if (pt > max[k]) max[k] = pt;
         }
      }
      /* Add slight safety margin to extents */
/*for(k=0;k<3;k++) {
printf("%d: min,max : %lf %lf\n",k,min[k],max[k]);
}
*/      for (k=0;k<3;k++) {
         if(min[k] != max[k]) {
            amt = 0.005 * (max[k] - min[k]);
         } else {
            val1 = pat->coeffs[0][0].dCoord[0] * vec[k].dCoord[0] +
                   pat->coeffs[0][0].dCoord[1] * vec[k].dCoord[1] +
                   pat->coeffs[0][0].dCoord[2] * vec[k].dCoord[2];
            val2 = pat->coeffs[pat->iDeg0][pat->iDeg1].dCoord[0] * 
                      vec[k].dCoord[0] +
                   pat->coeffs[pat->iDeg0][pat->iDeg1].dCoord[1] * 
                      vec[k].dCoord[1] +
                   pat->coeffs[pat->iDeg0][pat->iDeg1].dCoord[2] * 
                      vec[k].dCoord[2];
            amt = 0.005 * fabs(val1 - val2);
         }
         min[k] = min[k] - amt;
         max[k] = max[k] + amt;
#ifdef KEYSERDEBUG
printf("%d: min,max : %.20lf %.20lf\n",k,min[k],max[k]);
#endif
      }
      /* now determine plane equations from axes and extents */
      for (i=0;i<3;i++) {
         a[2*i] = vec[i].dCoord[0];
         a[2*i+1] = (-1.0)*vec[i].dCoord[0];
         b[2*i] = vec[i].dCoord[1];
         b[2*i+1] = (-1.0)*vec[i].dCoord[1];
         c[2*i] = vec[i].dCoord[2];
         c[2*i+1] = (-1.0)*vec[i].dCoord[2];
         d[2*i] = (-1.0)*max[i];
         d[2*i+1] = min[i];
      }

      /* now clip against all 6 planes */
      cmin1 = 0.0;
      cmax1 = 1.0;
      cmin2 = 0.0;
      cmax2 = 1.0;
      for (i=0;i<6;i++) {
         /* determine convex hull */
         /* st1 is for u-clip, st2 is for v-clip */

         /* Note: one extra location is allocated in each array (it is used by
                  the convex hull code) */
         st1 = (double**) malloc(((pat->iDeg0+1)*(pat->iDeg1+1)+1)*
                                 sizeof(double*));
         for (j=0; j <= ((pat->iDeg0+1)*(pat->iDeg1+1)); j++)
            st1[j]=(double*) malloc(2*sizeof(double));
         st2 = (double**) malloc(((pat->iDeg0+1)*(pat->iDeg1+1)+1)*
                                 sizeof(double*));
         for (j=0; j <= ((pat->iDeg0+1)*(pat->iDeg1+1)); j++)
            st2[j]=(double*) malloc(2*sizeof(double));

         for (j=0; j <= pat->iDeg0; j++) {
            for (k=0; k <= pat->iDeg1; k++) {
               st1[j*(pat->iDeg1+1)+k][0] = ((double) j)/((double) pat->iDeg0);
               st1[j*(pat->iDeg1+1)+k][1] = a[i]*pat->coeffs[j][k].dCoord[0] +
                                            b[i]*pat->coeffs[j][k].dCoord[1] +
                                            c[i]*pat->coeffs[j][k].dCoord[2] +
                                            d[i]*pat->coeffs[j][k].dCoord[3];
               st2[j*(pat->iDeg1+1)+k][0] = ((double) k)/((double) pat->iDeg1);
               st2[j*(pat->iDeg1+1)+k][1] = st1[j*(pat->iDeg1+1)+k][1];
            }
         }
#ifdef KEYSERDEBUG
for (j=0; j <= pat->iDeg0; j++) {
for (k=0; k <= pat->iDeg1; k++) {
printf("st1: %.10lf %.10lf\n",st1[j*(pat->iDeg1+1)+k][0],st1[j*(pat->iDeg1+1)+k][1]);
}
}
for (j=0; j <= pat->iDeg0; j++) {
for (k=0; k <= pat->iDeg1; k++) {
printf("st2: %.10lf %.10lf\n",st2[j*(pat->iDeg1+1)+k][0],st2[j*(pat->iDeg1+1)+k][1]);
}
}
#endif

         if (st1[0][1] > 0.0) firstisabove1 = 1;
         else firstisabove1 = 0;
         if (st2[0][1] > 0.0) firstisabove2 = 1;
         else firstisabove2 = 0;
         if (st1[((pat->iDeg0+1)*(pat->iDeg1+1)-1)][1] > 0.0) lastisabove1 = 1;
         else lastisabove1 = 0;
         if (st2[((pat->iDeg0+1)*(pat->iDeg1+1)-1)][1] > 0.0) lastisabove2 = 1;
         else lastisabove2 = 0;
         hullhit1 = ch2d(st1,((pat->iDeg0+1)*(pat->iDeg1+1)));
         hullhit2 = ch2d(st2,((pat->iDeg0+1)*(pat->iDeg1+1)));
#ifdef KEYSERDEBUG
for (j=0; j <= hullhit1; j++) {
printf("CH1: %.20lf %.20lf\n",st1[j][0],st1[j][1]);
}
for (j=0; j <= hullhit2; j++) {
printf("CH2: %.20lf %.20lf\n",st2[j][0],st2[j][1]);
}
#endif
         /* find points where hull intersects 0 line */
         hmax1 = -1.0;
         hmin1 = 2.0;
         hmax2 = -1.0;
         hmin2 = 2.0;
         for (j=0;j<hullhit1;j++) {
            if (st1[j][1] == 0.0) {
               if (st1[j][0] > hmax1) hmax1 = st1[j][0];
               if (st1[j][0] < hmin1) hmin1 = st1[j][0];
            }
            if (((st1[j][1] < 0.0) && (st1[j+1][1] > 0.0)) ||
                ((st1[j][1] > 0.0) && (st1[j+1][1] < 0.0))) {
               pt = st1[j][0] + (st1[j+1][0] - st1[j][0]) *
                    (st1[j][1]/(st1[j][1]-st1[j+1][1]));
               if (pt > hmax1) hmax1 = pt;
               if (pt < hmin1) hmin1 = pt;
            }
         }
         for (j=0;j<hullhit2;j++) {
            if (st2[j][1] == 0.0) {
               if (st2[j][0] > hmax2) hmax2 = st2[j][0];
               if (st2[j][0] < hmin2) hmin2 = st2[j][0];
            }
            if (((st2[j][1] < 0.0) && (st2[j+1][1] > 0.0)) ||
                ((st2[j][1] > 0.0) && (st2[j+1][1] < 0.0))) {
               pt = st2[j][0] + (st2[j+1][0] - st2[j][0]) *
                    (st2[j][1]/(st2[j][1]-st2[j+1][1]));
               if (pt > hmax2) hmax2 = pt;
               if (pt < hmin2) hmin2 = pt;
            }
         }
#ifdef KEYSERDEBUG
printf("hmin1, hmax1: %.20lf %.20lf\n",hmin1,hmax1);
printf("hmin2, hmax2: %.20lf %.20lf\n",hmin2,hmax2);
#endif
         if (hmax1 < 0.0) {
            /* no intersections - the hull is either above or below */
            if (st1[0][1] > 0.0) {
               /* The entire hull is above the line - no intersection */
/*printf("Reject - allabove 1\n");
*/               cmin1 = 1.0;
               cmax1 = 0.0;
            }
         } else {
            if (firstisabove1) {
               /* throw out stuff below min */
               if (hmin1 > cmin1) cmin1 = hmin1;
            }
            if (lastisabove1) {
               /* throw out stuff above max */
               if (hmax1 < cmax1) cmax1 = hmax1;
            }
         }
#ifdef KEYSERDEBUG
printf("cmin1, cmax1: %.20lf %.20lf\n",cmin1,cmax1);
#endif
         if (hmax2 < 0.0) {
            /* no intersections - the hull is either above or below */
            if (st2[0][1] > 0.0) {
               /* The entire hull is above the line - no intersection */
/*printf("Reject - allabove 1\n");
*/               cmin2 = 1.0;
               cmax2 = 0.0;
            }
         } else {
            if (firstisabove2) {
               /* throw out stuff below min */
               if (hmin2 > cmin2) cmin2 = hmin2;
            }
            if (lastisabove2) {
               /* throw out stuff above max */
               if (hmax2 < cmax2) cmax2 = hmax2;
            }
         }
#ifdef KEYSERDEBUG
printf("cmin2, cmax2: %.20lf %.20lf\n",cmin2,cmax2);
#endif
         /* now free up st1, st2 */
         for (j=0; j <= ((pat->iDeg0+1)*(pat->iDeg1+1)); j++) {
            free(st1[j]);
            free(st2[j]);
         }
         free(st1);
         free(st2);
      }
      /* Add slight safety margin to min, max as suggested by Sederberg */
      cmin1 = 0.99*cmin1;
      cmax1 = 0.99*cmax1 + 0.01;
      cmin2 = 0.99*cmin2;
      cmax2 = 0.99*cmax2 + 0.01;
#ifdef KEYSERDEBUG
printf("return %lf %lf, %lf %lf\n",cmin1,cmax1,cmin2,cmax2);
#endif
      *utmin = cmin1;
      *utmax = cmax1;
      *vmin = cmin2;
      *vmax = cmax2;
   } else {
      /* clip curve against patch */
      /* first determine three axes of parallelepiped */
#ifdef KEYSERDEBUG
printf("CURVE CLIP\n");
#endif
      for (i=0;i<3;i++) {
         vec[0].dCoord[i] = (pat->coeffs[pat->iDeg0][0].dCoord[i] /
                             pat->coeffs[pat->iDeg0][0].dCoord[3]) -
                            (pat->coeffs[0][0].dCoord[i] /
                             pat->coeffs[0][0].dCoord[3]);
         size1 += vec[0].dCoord[i]*vec[0].dCoord[i];
         vec[1].dCoord[i] = (pat->coeffs[0][pat->iDeg1].dCoord[i] /
                             pat->coeffs[0][pat->iDeg1].dCoord[3]) -
                            (pat->coeffs[0][0].dCoord[i] /
                             pat->coeffs[0][0].dCoord[3]);
         size2 += vec[1].dCoord[i]*vec[1].dCoord[i];
      }
      if(size1 == 0.0) {
         size1=1.0;
         vec[0].dCoord[0] = 1.0;
         vec[0].dCoord[1] = 0.0;
         vec[0].dCoord[2] = 0.0;
      }
      if(size2 == 0.0) {
         size2=1.0;
         vec[1].dCoord[0] = 0.0;
         vec[1].dCoord[1] = 1.0;
         vec[1].dCoord[2] = 0.0;
      }
      size1=sqrt(size1);
      size2=sqrt(size2);
      for (i=0;i<3;i++) {
         vec[0].dCoord[i] = vec[0].dCoord[i]/size1;
         vec[1].dCoord[i] = vec[1].dCoord[i]/size2; 
      }
      if (EQUAL3D_EPS(vec[0],vec[1],1e-4)) {
         /* vec0 and vec1 are the same, set vec2 perp. to vec1 */
         if (!(FP_EQ_EPS(vec[1].dCoord[0],1.0,1e-4))) {
            vec[1].dCoord[0] = 1.0;
            vec[1].dCoord[1] = 0.0;
            vec[1].dCoord[2] = 0.0;
         } else {
            vec[1].dCoord[0] = 0.0;
            vec[1].dCoord[1] = 1.0;
            vec[1].dCoord[2] = 0.0;
         }
      }

      if ((vec[0].dCoord[0]*vec[1].dCoord[0] +
           vec[0].dCoord[1]*vec[1].dCoord[1] +
           vec[0].dCoord[2]*vec[1].dCoord[2]) > 0.865 ) {
         /* vec0 and vec1 are within 30 degrees of each other
            reassign vec1 */
         vec[2].dCoord[0] = vec[0].dCoord[1]*vec[1].dCoord[2] -
                            vec[0].dCoord[2]*vec[1].dCoord[1];
         vec[2].dCoord[1] = vec[0].dCoord[2]*vec[1].dCoord[0] -
                            vec[0].dCoord[0]*vec[1].dCoord[2];
         vec[2].dCoord[2] = vec[0].dCoord[0]*vec[1].dCoord[1] -
                            vec[0].dCoord[1]*vec[1].dCoord[0];
         size3 = vec[2].dCoord[0]*vec[2].dCoord[0] +
                 vec[2].dCoord[1]*vec[2].dCoord[1] +
                 vec[2].dCoord[2]*vec[2].dCoord[2];
         size3 = sqrt(size3);
         vec[2].dCoord[0]=vec[2].dCoord[0]/size3;
         vec[2].dCoord[1]=vec[2].dCoord[1]/size3;
         vec[2].dCoord[2]=vec[2].dCoord[2]/size3;

         vec[1].dCoord[0] = vec[2].dCoord[1]*vec[0].dCoord[2] -
                            vec[2].dCoord[2]*vec[0].dCoord[1];
         vec[1].dCoord[1] = vec[2].dCoord[2]*vec[0].dCoord[0] -
                            vec[2].dCoord[0]*vec[0].dCoord[2];
         vec[1].dCoord[2] = vec[2].dCoord[0]*vec[0].dCoord[1] -
                            vec[2].dCoord[1]*vec[0].dCoord[0];
         size2 = vec[1].dCoord[0]*vec[1].dCoord[0] +
                 vec[1].dCoord[1]*vec[1].dCoord[1] +
                 vec[1].dCoord[2]*vec[1].dCoord[2];
         size2 = sqrt(size2);
         vec[1].dCoord[0]=vec[1].dCoord[0]/size2;
         vec[1].dCoord[1]=vec[1].dCoord[1]/size2;
         vec[1].dCoord[2]=vec[1].dCoord[2]/size2;
      }

      vec[3].dCoord[0] = vec[0].dCoord[1]*vec[1].dCoord[2] -
                         vec[0].dCoord[2]*vec[1].dCoord[1];
      vec[3].dCoord[1] = vec[0].dCoord[2]*vec[1].dCoord[0] -
                         vec[0].dCoord[0]*vec[1].dCoord[2];
      vec[3].dCoord[2] = vec[0].dCoord[0]*vec[1].dCoord[1] -
                         vec[0].dCoord[1]*vec[1].dCoord[0];
      size4 = vec[3].dCoord[0]*vec[3].dCoord[0] +
              vec[3].dCoord[1]*vec[3].dCoord[1] +
              vec[3].dCoord[2]*vec[3].dCoord[2];
      size4 = sqrt(size4);
      vec[3].dCoord[0]=vec[3].dCoord[0]/size4;
      vec[3].dCoord[1]=vec[3].dCoord[1]/size4;
      vec[3].dCoord[2]=vec[3].dCoord[2]/size4;

      vec[4].dCoord[0] = vec[0].dCoord[0]+vec[1].dCoord[0];
      vec[4].dCoord[1] = vec[0].dCoord[1]+vec[1].dCoord[1];
      vec[4].dCoord[2] = vec[0].dCoord[2]+vec[1].dCoord[2];
      size5 = vec[4].dCoord[0]*vec[4].dCoord[0] +
              vec[4].dCoord[1]*vec[4].dCoord[1] +
              vec[4].dCoord[2]*vec[4].dCoord[2];
      size5 = sqrt(size5);
      vec[4].dCoord[0]=vec[4].dCoord[0]/size5;
      vec[4].dCoord[1]=vec[4].dCoord[1]/size5;
      vec[4].dCoord[2]=vec[4].dCoord[2]/size5;

      vec[2].dCoord[0] = vec[3].dCoord[0]+NEWANG*vec[4].dCoord[0];
      vec[2].dCoord[1] = vec[3].dCoord[1]+NEWANG*vec[4].dCoord[1];
      vec[2].dCoord[2] = vec[3].dCoord[2]+NEWANG*vec[4].dCoord[2];
      size3 = vec[2].dCoord[0]*vec[2].dCoord[0] +
              vec[2].dCoord[1]*vec[2].dCoord[1] +
              vec[2].dCoord[2]*vec[2].dCoord[2];
      size3 = sqrt(size3);
      vec[2].dCoord[0]=vec[2].dCoord[0]/size3;
      vec[2].dCoord[1]=vec[2].dCoord[1]/size3;
      vec[2].dCoord[2]=vec[2].dCoord[2]/size3;

#ifdef KEYSERDEBUG
for(i=0;i<3;i++) {
printf("vec%d: %lf %lf %lf\n",i,vec[i].dCoord[0],vec[i].dCoord[1],vec[i].dCoord[2]);
}
#endif
      /* now determine extents of parallelepiped */
      for (k=0;k<3;k++) {
         /* project control point onto axis, change min,max if needed */
         min[k] = (pat->coeffs[0][0].dCoord[0] * vec[k].dCoord[0] +
                   pat->coeffs[0][0].dCoord[1] * vec[k].dCoord[1] +
                   pat->coeffs[0][0].dCoord[2] * vec[k].dCoord[2]) /
                  pat->coeffs[0][0].dCoord[3];
         max[k] = min[k];
      }
      for (i=0;i<=pat->iDeg0;i++) {
         for (j=0;j<=pat->iDeg1;j++) {
            for (k=0;k<3;k++) {
               /* project control point onto axis, change min,max if needed */
               pt = (pat->coeffs[i][j].dCoord[0] * vec[k].dCoord[0] +
                     pat->coeffs[i][j].dCoord[1] * vec[k].dCoord[1] +
                     pat->coeffs[i][j].dCoord[2] * vec[k].dCoord[2]) /
                    pat->coeffs[i][j].dCoord[3];
               if (pt < min[k]) min[k] = pt;
               if (pt > max[k]) max[k] = pt;
            }
         }
      }
/*for (k=0;k<3;k++) {
printf("%d: min,max : %.20lf %.20lf\n",k,min[k],max[k]);
}
*/
      /* Add slight safety margin to extents */
      for (k=0;k<3;k++) {
         if (min[k] != max[k]) {
            amt = 0.005 * (max[k] - min[k]);
         } else {
            val1 = cur->CntrlPts[0].dCoord[0] * vec[k].dCoord[0] +
                   cur->CntrlPts[0].dCoord[1] * vec[k].dCoord[1] +
                   cur->CntrlPts[0].dCoord[2] * vec[k].dCoord[2];
            val2 = cur->CntrlPts[cur->Deg].dCoord[0] * vec[k].dCoord[0] +
                   cur->CntrlPts[cur->Deg].dCoord[1] * vec[k].dCoord[1] +
                   cur->CntrlPts[cur->Deg].dCoord[2] * vec[k].dCoord[2];
            amt = 0.005 * fabs(val1 - val2);
         }
         min[k] = min[k] - amt;
         max[k] = max[k] + amt;
#ifdef KEYSERDEBUG
printf("%d: min,max : %.20lf %.20lf\n",k,min[k],max[k]);
#endif
      }

      /* now determine plane equations from axes and extents */
      for (i=0;i<3;i++) {
         a[2*i] = vec[i].dCoord[0];
         a[2*i+1] = (-1.0)*vec[i].dCoord[0];
         b[2*i] = vec[i].dCoord[1];
         b[2*i+1] = (-1.0)*vec[i].dCoord[1];
         c[2*i] = vec[i].dCoord[2];
         c[2*i+1] = (-1.0)*vec[i].dCoord[2];
         d[2*i] = (-1.0)*max[i];
         d[2*i+1] = min[i];
      }

      /* now clip against all 6 planes */
      cmin1 = 0.0;
      cmax1 = 1.0;
      for (i=0;i<6;i++) {
         /* determine convex hull */
         /* Note: one extra position allocated to be used by chull code */
         st1 = (double**) malloc((cur->Deg+2)*sizeof(double*));
         for (j=0;j<=(cur->Deg+1);j++)
            st1[j]=(double*) malloc(2*sizeof(double));
         for (j=0;j<=cur->Deg;j++) {
            st1[j][0] = ((double) j) / ((double) cur->Deg);
            st1[j][1] = a[i]*cur->CntrlPts[j].dCoord[0] +
                        b[i]*cur->CntrlPts[j].dCoord[1] +
                        c[i]*cur->CntrlPts[j].dCoord[2] +
                        d[i]*cur->CntrlPts[j].dCoord[3];

         }
#ifdef KEYSERDEBUG
for (j=0;j<=cur->Deg;j++) {
printf("st1: %.10lf %.10lf\n",st1[j][0],st1[j][1]);
}
#endif
         if (st1[0][1] > 0.0) firstisabove1 = 1;
         else firstisabove1 = 0;
         if (st1[cur->Deg][1] > 0.0) lastisabove1 = 1;
         else lastisabove1 = 0;
         hullhit1 = ch2d(st1,(cur->Deg+1));
#ifdef KEYSERDEBUG
for (j=0; j <= hullhit1; j++) {
printf("CH1: %.20lf %.20lf\n",st1[j][0],st1[j][1]);
}
#endif
         /* find points where hull intersects 0 line */
         hmax1 = -1.0;
         hmin1 = 2.0;
         for (j=0;j<hullhit1;j++) {
            if (st1[j][1] == 0.0) {
               if (st1[j][0] > hmax1) hmax1 = st1[j][0];
               if (st1[j][0] < hmin1) hmin1 = st1[j][0];
            }
            if (((st1[j][1] < 0.0) && (st1[j+1][1] > 0.0)) ||
                ((st1[j][1] > 0.0) && (st1[j+1][1] < 0.0))) {
               pt = st1[j][0] + (st1[j+1][0] - st1[j][0]) *
                    (st1[j][1]/(st1[j][1]-st1[j+1][1]));
               if (pt > hmax1) hmax1 = pt;
               if (pt < hmin1) hmin1 = pt;
            }
         }
#ifdef KEYSERDEBUG
printf("hmin1, hmax1: %.20lf %.20lf\n",hmin1,hmax1);
#endif
         if (hmax1 < 0.0) {
            /* no intersections - the hull is either above or below */
            if (st1[0][1] > 0.0) {
               /* The entire hull is above the line - no intersection */
               cmin1 = 1.0;
               cmax1 = 0.0;
            }
         } else {
            if (firstisabove1) {
               /* throw out stuff below min */
               if (hmin1 > cmin1) cmin1 = hmin1;
            }
            if (lastisabove1) {
               /* throw out stuff above max */
               if (hmax1 < cmax1) cmax1 = hmax1;
            }
         }
#ifdef KEYSERDEBUG
printf("cmin1, cmax1: %.20lf %.20lf\n",cmin1,cmax1);
#endif
         /* now free up st1 */
         for (j=0;j<=(cur->Deg+1);j++)
            free(st1[j]);
         free(st1);
      }
      /* Add slight safety margin to min, max as suggested by Sederberg */
      cmin1 = 0.99*cmin1;
      cmax1 = 0.99*cmax1 + 0.01;
      *utmin = cmin1;
      *utmax = cmax1;
#ifdef KEYSERDEBUG
printf("return %lf %lf\n",cmin1,cmax1);
#endif
   }
}

int curve_surface_intersect(TENSOR_PATCH* pat, XYZ_SPLINE_TYPE* cur, 
                            double tolc, double tolp, int* numhits, 
                            double* thits, double* uhits, double* vhits)
/* This routine implements Sederberg's Bezier clipping algorithm to find
   the intersection between a curve and a surface.  The surface is given
   by pat, the curve by cur, and the desired tolerance (in terms of t, u,
   and v) by tol.  Returned are the number of intersections (numinter),
   and arrays containing the t (for curve) and u and v (for surface) values
   of the intersection points */
{
   INTERSECT_STACK_ELEMENT* stk = NULL;
   INTERSECT_STACK_ELEMENT* elt = NULL;
   TENSOR_PATCH* patch = NULL;
   TENSOR_PATCH* patch2 = NULL;
   TENSOR_PATCH* pp1 = NULL;
   TENSOR_PATCH* pp2 = NULL;
   TENSOR_PATCH* pp3 = NULL;
   TENSOR_PATCH* pp4 = NULL;
   XYZ_SPLINE_TYPE* curve = NULL;
   XYZ_SPLINE_TYPE* curve2 = NULL;
   XYZ_SPLINE_TYPE* curve3 = NULL;
   double min1, max1, min2, max2, min3, max3;
   double umin,umax,vmin,vmax,tmin,tmax;
   HIT_PT_ELEMENT* hitstk = NULL;
   HIT_PT_ELEMENT* temphit = NULL;
   int i;
   int ok_to_add;
   int numinter;

   /* add pat,cur to stack */
   patch = newtensorpatch(pat->iDeg0,pat->iDeg1);
   curve = newxyzspline();
   copy_ten_p(pat,patch);
   copy_xyz_spline(cur,curve);
   stk = newinterstackelement(patch, curve, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 1);
   hitstk = newhitptelement();
   numinter = 0;
/*printf("called\n");
*/
   /* while stack not empty */
   while (stk != NULL) {
      if( (((stk->umax) - (stk->umin)) <= tolp) && 
          (((stk->vmax) - (stk->vmin)) <= tolp) &&
          (((stk->tmax) - (stk->tmin)) <= tolc)) {

         /* if there is an intersection on this patch, then add     */
         /* this point to the list of hit points, otherwise discard */
#ifdef KEYSERDEBUG
printf("testing1\n");
#endif
         clipit(stk->patch, stk->curve, 1, &min3, &max3, &min2, &max2);
#ifdef KEYSERDEBUG
printf("testing2\n");
#endif
         clipit(stk->patch, stk->curve, 0, &min1, &max1, &min2, &max2);
         if ((max1 >= min1) && (max2 >= min2) && (max3 >= min3)) {
#ifdef KEYSERDEBUG
printf("adding\n");
#endif
            /* there is an intersection - add this to hit list */

            ok_to_add = 1;
            for(temphit=hitstk;((temphit != NULL) && (ok_to_add));
                temphit = temphit->next) {
               if ((FP_EQ_EPS(temphit->t,(stk->tmin+stk->tmax)/2.0,tolc)) &&
                   (FP_EQ_EPS(temphit->u,(stk->umin+stk->umax)/2.0,tolp)) &&
                   (FP_EQ_EPS(temphit->v,(stk->vmin+stk->vmax)/2.0,tolp))) {
                  ok_to_add = 0;
               }

            }

            if (ok_to_add) {
               temphit = newhitptelement();
               temphit->t = (stk->tmax + stk->tmin) / 2.0;
               temphit->u = (stk->umax + stk->umin) / 2.0;
               temphit->v = (stk->vmax + stk->vmin) / 2.0;
               temphit->next = hitstk;
               hitstk = temphit;
               numinter++;
            }
         }
         /* remove top stack element */
         elt = stk;
         stk = stk->next;
/*
         free_ten_p(elt->patch);
*/
         free_xyz_spline(elt->curve);
         free_interstackelement(elt);
      } else {
         /* call clipit for that patch and curve */
/*         if ( ((stk->oper == 0) && (((stk->umax) - (stk->umin)) <= tolp) &&
              (((stk->vmax) - (stk->vmin)) <= tolp)) || 
              ((stk->oper == 1) && (((stk->tmax) - (stk->tmin)) <= tolc))) {
            stk->oper = 1 - stk->oper;
         }
*/

/*printf("udiff: %lf, vdiff: %lf, tdiff: %lf\n",stk->umax-stk->umin,
                         stk->vmax-stk->vmin, stk->tmax-stk->tmin);
*/
         clipit(stk->patch, stk->curve, stk->oper, &min1, &max1, &min2, &max2);

         if ((max1 < min1) || ((stk->oper == 0) && (max2 < min2))) {
            /* The entire patch is clipped away - no intersecions! */
            elt = stk;
            stk = stk->next;
/*
            free_ten_p(elt->patch);
*/
            free_xyz_spline(elt->curve);
            free_interstackelement(elt);
         } else {
            if (stk->oper == 0) {
               /* clipping patch against curve */
               elt = stk->next;
               patch = stk->patch;
               curve = stk->curve;
               umin = stk->umin;
               umax = stk->umax;
               vmin = stk->vmin;
               vmax = stk->vmax;
               tmin = stk->tmin;
               tmax = stk->tmax;
               free_interstackelement(stk);
               stk=elt;

               if ((max1-min1) > .8) {
                  /* Not good enough - divide in half in u */
                  if ((max2-min2) > .8) {
                     /* Not good enough - divide in half in v */
                     /* need to push four patches */
                     /* first allocate space for new patches */
                     pp1 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     pp2 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     pp3 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     pp4 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     /* now get patches */
                     subdivide_bezier_patch(*patch, 0.0, 0.5, 0.0, 0.5, pp1);
                     subdivide_bezier_patch(*patch, 0.0, 0.5, 0.5, 1.0, pp2);
                     subdivide_bezier_patch(*patch, 0.5, 1.0, 0.0, 0.5, pp3);
                     subdivide_bezier_patch(*patch, 0.5, 1.0, 0.5, 1.0, pp4);
                     /* now push patches */
                     curve2 = newxyzspline();
                     copy_xyz_spline(curve,curve2);
                     elt = newinterstackelement(pp1, curve2,
                                                umin, 0.5*(umax-umin) + umin,
                                                vmin, 0.5*(vmax-vmin) + vmin, 
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;

                     curve2 = newxyzspline();
                     copy_xyz_spline(curve,curve2);
                     elt = newinterstackelement(pp2, curve2,
                                                umin, 0.5*(umax-umin) + umin,
                                                0.5*(vmax-vmin) + vmin, vmax,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;

                     curve2 = newxyzspline();
                     copy_xyz_spline(curve,curve2);
                     elt = newinterstackelement(pp3, curve2,
                                                0.5*(umax-umin) + umin, umax,
                                                vmin, 0.5*(vmax-vmin) + vmin,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;

                     elt = newinterstackelement(pp4, curve,
                                                0.5*(umax-umin) + umin, umax,
                                                0.5*(vmax-vmin) + vmin, vmax, 
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;
#ifdef KEYSERDEBUG
printf("1->4\n");
#endif
                     
/*
                     free_ten_p(patch);
*/
                  } else {
                     /* Good - will subdivide at min2, max2 in v */
                     /* need to push two patches */
                     /* first allocate space for new patches */
                     pp1 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     pp2 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     /* now get patches */
                     subdivide_bezier_patch(*patch, 0.0, 0.5, min2, max2, pp1);
                     subdivide_bezier_patch(*patch, 0.5, 1.0, min2, max2, pp2);
                     /* now push patches */
                     curve2 = newxyzspline();
                     copy_xyz_spline(curve,curve2);
                     elt = newinterstackelement(pp1, curve2,
                                                umin, 0.5*(umax-umin) + umin,
                                                min2*(vmax-vmin) + vmin, 
                                                max2*(vmax-vmin) + vmin,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;

                     elt = newinterstackelement(pp2, curve,
                                                0.5*(umax-umin) + umin, umax,
                                                min2*(vmax-vmin) + vmin, 
                                                max2*(vmax-vmin) + vmin,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;
#ifdef KEYSERDEBUG
printf("1->2\n");
#endif
/*
                     free_ten_p(patch);
*/
                  }
               } else {
                  /* Good - will subdivide at min1, max1 in u */
                  if ((max2-min2) > .8) {               
                     /* Not good enough - divide in half in v */
                     /* need to push two patches */
                     /* first allocate space for new patches */
                     pp1 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     pp2 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     /* now get patches */
                     subdivide_bezier_patch(*patch, min1, max1, 0.0, 0.5, pp1);
                     subdivide_bezier_patch(*patch, min1, max1, 0.5, 1.0, pp2);
                     /* now push patches */
                     curve2 = newxyzspline();
                     copy_xyz_spline(curve,curve2);
                     elt = newinterstackelement(pp1, curve2, 
                                                min1*(umax-umin) + umin,
                                                max1*(umax-umin) + umin,
                                                vmin, 0.5*(vmax-vmin) + vmin,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;

                     elt = newinterstackelement(pp2, curve, 
                                                min1*(umax-umin) + umin,
                                                max1*(umax-umin) + umin,
                                                0.5*(vmax-vmin) + vmin, vmax,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;
#ifdef KEYSERDEBUG
printf("1->2\n");
#endif
/*
                     free_ten_p(patch);
*/
                  } else {
                     /* Good - will subdivide at min2, max2 in v */
                     /* first allocate space for new patch */
                     pp1 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                     /* need to push one patch */
                     subdivide_bezier_patch(*patch,min1,max1,min2,max2,pp1);
                     elt = newinterstackelement(pp1, curve, 
                                                min1*(umax-umin) + umin,
                                                max1*(umax-umin) + umin,
                                                min2*(vmax-vmin) + vmin,
                                                max2*(vmax-vmin) + vmin,
                                                tmin, tmax, 1);
                     elt->next = stk;
                     stk = elt;
/*
                     free_ten_p(patch);
*/
                  }
               }
            } else {
               /* clipping curve against patch */
               elt = stk->next;
               patch = stk->patch;
               curve = stk->curve;
               umin = stk->umin;
               umax = stk->umax;
               vmin = stk->vmin;
               vmax = stk->vmax;
               tmin = stk->tmin;
               tmax = stk->tmax;
               free_interstackelement(stk);
               stk=elt;

               if ((max1-min1) > .8) {
                  /* Not good enough, divide curve in half and push both */
                  /* first allocate space for new curves */
                  curve2 = newxyzspline();
                  curve3 = newxyzspline();
                  curve2->Deg = curve->Deg;
                  curve2->CntrlPts = (POINT4D *) new((curve->Deg + 1) * 
                                                     sizeof(POINT4D));
                  curve3->Deg = curve->Deg;
                  curve3->CntrlPts = (POINT4D *) new((curve->Deg + 1) * 
                                                  sizeof(POINT4D));             
                  /* now get curves */
                  subdivide_bezier_curve(*curve, 0.0, 0.5, curve2);
                  subdivide_bezier_curve(*curve, 0.5, 1.0, curve3);
                  /* now push curves */
                  /* allocate space for new patch */
                  pp1 = newtensorpatch(patch->iDeg0,patch->iDeg1);
                  copy_ten_p(patch,pp1);
                  elt = newinterstackelement(pp1, curve2,
                                             umin, umax,
                                             vmin, vmax,
                                             tmin, 
                                             0.5*(tmax-tmin) + tmin, 0);
                  elt->next = stk;
                  stk = elt;

                  elt = newinterstackelement(patch, curve3,
                                             umin, umax,
                                             vmin, vmax,
                                             0.5*(tmax-tmin) + tmin,
                                             tmax, 0);
                  elt->next = stk;
                  stk = elt;
                  free_xyz_spline(curve);
#ifdef KEYSERDEBUG
printf("C:1->2\n");
#endif
               } else {
                  /* Subdivide curve at min1, max1; push center curve */
                  curve2 = newxyzspline();
                  curve2->Deg = curve->Deg;
                  curve2->CntrlPts = (POINT4D *) new((curve->Deg + 1) * 
                                                  sizeof(POINT4D));
                  subdivide_bezier_curve(*curve, min1, max1, curve2);
                  elt = newinterstackelement(patch, curve2,
                                             umin, umax,
                                             vmin, vmax,
                                             min1*(tmax-tmin) + tmin,
                                             max1*(tmax-tmin) + tmin, 0);
                  elt->next = stk;
                  stk = elt;
                  free_xyz_spline(curve);
               }
            }
         }
      }
   }
   /* pull off hit points and return */
#define MAXNUMHITS 50

   if (numinter>MAXNUMHITS) {
      printf("ERROR - too many curve-surface intersections - %d\n",numinter);
      printf("        Will only report first %d intersections\n",MAXNUMHITS);
      numinter=MAXNUMHITS;
   }
   /* now fill hit arrays */
   for (i=0;i<numinter;i++) {
      thits[i] = hitstk->t;
      uhits[i] = hitstk->u;
      vhits[i] = hitstk->v;
      temphit = hitstk;
      hitstk = hitstk->next;
      free_hitptelement(temphit);      
   }
   *numhits = numinter;
}




/* Following is the convex hull routine */

/*
 * Ken Clarkson wrote this.  Copyright (c) 1996 by AT&T..
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 */


/*
 * two-dimensional convex hull
 * read points from stdin,
 *      one point per line, as two numbers separated by whitespace
 * on stdout, points on convex hull in order around hull, given
 *      by their numbers in input order
 * the results should be "robust", and not return a wildly wrong hull,
 *	despite using floating point
 * works in O(n log n); I think a bit faster than Graham scan;
 * 	somewhat like Procedure 8.2 in Edelsbrunner's "Algorithms in Combinatorial
 *	Geometry".
 */


int ccw(double **P, int i, int j, int k) {
   double a = P[i][0] - P[j][0],
          b = P[i][1] - P[j][1],
          c = P[k][0] - P[j][0],
          d = P[k][1] - P[j][1];
   return a*d - b*c <= 0.0;	   /* true if points i, j, k counterclockwise */
}


#define CMPM(c,A,B) \
	v = (*(double**)A)[c] - (*(double**)B)[c];\
	if (v>0) return 1;\
	if (v<0) return -1;

int cmpl(const void *a, const void *b) {
   double v; 
   CMPM(0,a,b);
   CMPM(1,b,a);
   return 0;
}
int cmph(const void *a, const void *b) {return cmpl(b,a);}


int make_chain(double** verts, int n, int (*cmp)(const void*, const void*)) {
   int i, j, s = 1;
   double* t;

   qsort(verts, n, sizeof(double*), cmp);
   for (i=2; i<n; i++) {
      for (j=s; j>=1 && ccw(verts, i, j, j-1); j--){}
      s = j+1;
      t = verts[s]; verts[s] = verts[i]; verts[i] = t;
   }
   return s;
}

int ch2d(double** P, int n)  {
   int u;
   u = make_chain(P, n, cmpl);			/* make lower hull */
   if (!n) return 0;
   P[n] = P[0];
   return u+make_chain(P+u, n-u+1, cmph);	/* make upper hull */
}
