#ifndef _LOOP
#include "defs.h"
#include "complex.h"

#define MAX_ITER_LIM	100

EXTERN_LINALG

void
cx_generalized_power_iterate( rgrgB, rgrgD, rgEVec, iDim, iDegree, dGuess, pdEVal, m, it_val )
double      **rgrgB, **rgrgD;
COMPLEX     *rgEVec;
int         iDim, iDegree, m;
COMPLEX     dGuess, *pdEVal;
int	    *it_val;
{

    COMPLEX **rgrgcB, **rgrgcD, **rgrgA, *rgdWork;
    COMPLEX *rgQ;
    COMPLEX **LastMat; 
    int    i,j,k,l;
    double p, q;
    COMPLEX dMagQVec, tmp;
    COMPLEX dAlpha1, dAlpha2;
    COMPLEX dNewEst, dS, dT;
    COMPLEX dOldEst, dOldS, dOldT;
    COMPLEX dNumer, dDenom;
    int dim, base;
    FILE *df, *bf;

    base = iDim - iDegree;
    dim = iDim / iDegree;
    p = dGuess.Re;
    q = dGuess.Im;

    /* allocate/initialize the matrices */
    ALLOCMAT(rgrgcB, iDim, iDim, COMPLEX);
    ALLOCMAT(rgrgcD, iDim, iDim, COMPLEX);
    ALLOCMAT(rgrgA, iDim, iDim, COMPLEX);
    ALLOCMAT(LastMat, iDegree, iDegree, COMPLEX);
    ALLOCN(rgdWork, COMPLEX, iDim);
    ALLOCN(rgQ, COMPLEX, iDim);

    /* perform inverse power iteartions as per
       Golub and Van Loan, pg 383 */

    for (i=0; i<iDim; i++)
    for (j=0; j<iDim; j++)
    {
      rgrgcB[i][j].Re = rgrgB[i][j];
      rgrgcB[i][j].Im = 0.0;
      rgrgcD[i][j].Re = rgrgD[i][j];
      rgrgcD[i][j].Im = 0.0;
    }


    for (i=0;i<iDim;i++) {
	rgQ[i].Re = rgEVec[i].Re;
	rgQ[i].Im = rgEVec[i].Im;
    }

    /* normalize and use as next estimate */
    dMagQVec.Re = 0.0;
    dMagQVec.Im = 0.0;
    for(i=0;i<iDim;i++)
	dMagQVec = plus(dMagQVec, mult(rgQ[i],rgQ[i]));
    
    dMagQVec = sqroot(dMagQVec);
    
    for (i=0; i<iDim; i++)
       rgQ[i] = divid(rgQ[i], dMagQVec);

    /* calc (D*lambda - B) */
    for (j = 0; j < dim; j++)
    {
        l = j*iDegree;
        for ( i = 0; i < iDegree; i++)
            for ( k = 0; k < iDegree; k++)
        {
            rgrgA[base + i][l + k] = minus(mult(rgrgcD[base + i][l+k],dGuess), 
                                           rgrgcB[base + i][l + k]);
       }
    }

    dAlpha1 = minus(mult(rgrgcD[0][0],dGuess), rgrgcB[0][0]);
    dAlpha2 = minus(mult(rgrgcD[0][iDegree],dGuess), rgrgcB[0][iDegree]);

    CDecompose(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2);

    dNewEst = dGuess;
    dS = divid(rgQ[1],rgQ[0]); 
    dT = divid(rgQ[m],rgQ[0]); 
    *it_val = 0;

    do {
        dOldEst = dNewEst;
 	 /* printf("CxPower: Est = %lf + i%lf\n",dNewEst.Re,dNewEst.Im); */
        dOldS = dS;
        dOldT = dT;
        
	for (i=0; i<iDim; i++) {
	  rgdWork[i].Re = 0.0;
	  rgdWork[i].Im = 0.0;
	  for (j=0; j<iDim; j++) {
	    rgdWork[i] = plus(rgdWork[i], mult(rgrgcD[i][j],rgQ[j]));
	  }
	}
	for (i=0; i<iDim; i++)
	  rgQ[i] = rgdWork[i];

	CSolve(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2, rgQ);

	/* normalize and use as next estimate */
        dMagQVec.Re = 0.0;
        dMagQVec.Im = 0.0;
        for(i=0;i<iDim;i++)
            dMagQVec = plus(dMagQVec, mult(rgQ[i],rgQ[i]));

        dMagQVec = sqroot(dMagQVec);

	if (FP_EQ_EPS(0.0, dMagQVec.Re, 1e-10) && 
            FP_EQ_EPS(0.0, dMagQVec.Im, 1e-10)) {
	    fprintf(stderr, "\ncx_power_iterate(): zero magnitude e-vector.\n");
	    exit(-3);
	}

        for (i=0; i<iDim; i++)
           rgQ[i] = divid(rgQ[i], dMagQVec);

	for (i=0; i<iDim; i++) {
          rgdWork[i].Re = 0.0;
          rgdWork[i].Im = 0.0;
          for (j=0; j<iDim; j++) {
            rgdWork[i] = plus(rgdWork[i], mult(rgrgcB[i][j],rgQ[j]));
          }
        }
	dNumer.Re = 0.0;
	dNumer.Im = 0.0;
	for (i=0; i<iDim; i++)
	  dNumer = plus(dNumer, mult(rgdWork[i], rgQ[i]));

	for (i=0; i<iDim; i++) {
          rgdWork[i].Re = 0.0;
          rgdWork[i].Im = 0.0;
          for (j=0; j<iDim; j++) {
            rgdWork[i] = plus(rgdWork[i], mult(rgrgcD[i][j],rgQ[j]));
          }
        }
	dDenom.Re = 0.0;
	dDenom.Im = 0.0;
	for (i=0; i<iDim; i++)
	  dDenom = plus(dDenom, mult(rgdWork[i], rgQ[i]));

	dNewEst = divid(dNumer, dDenom);
	dS = divid(rgQ[1],rgQ[0]);
	dT = divid(rgQ[m],rgQ[0]);
	*it_val = (*it_val) + 1;

    } while (((!FP_EQ_EPS( dOldEst.Re, dNewEst.Re, 1e-3)) || 
	      (!FP_EQ_EPS( dOldEst.Im, dNewEst.Im, 1e-3)) ||
	      (!(FP_EQ_EPS( dOldS.Re, dS.Re, 1e-3))) || 
	      (!(FP_EQ_EPS( dOldS.Im, dS.Im, 1e-3))) || 
	      (!(FP_EQ_EPS( dOldT.Re, dT.Re, 1e-3))) || 
	      (!(FP_EQ_EPS( dOldT.Im, dT.Im, 1e-3)))) &&
	      (*it_val < MAX_ITER_LIM)); 

    pdEVal->Re = dNewEst.Re;
    pdEVal->Im = dNewEst.Im;
    for (i=0; i<iDim; i++) {
      rgEVec[i].Re = rgQ[i].Re;
      rgEVec[i].Im = rgQ[i].Im;
    }
    myfree(rgQ); myfree(rgdWork); 
    myfree(LastMat);
    FREEMAT(rgrgA, iDim, iDim);
    FREEMAT(rgrgcB, iDim, iDim);
    FREEMAT(rgrgcD, iDim, iDim);
}

Matrix_Vector_Solve(A, vec, dim)
double **A;
double *vec;
int dim;
{
   int i,j,k;
   double *LM, *v, *ipvt;
   int lda, lda1, job;
   double rcond;

   ALLOCN(LM, DOUBLE, dim*dim);
   ALLOCN(v, DOUBLE, dim);
   ALLOCN(ipvt, DOUBLE, dim);

   for (i = j = 0; j < dim; j++)
    for (k=0; k<dim; k++)
      LM[i++] = A[j][k];

   lda = lda1 = dim;

#ifdef __SGI
   dgeco_(LM,&lda,&lda1,ipvt,&rcond,v);
#else
   dgeco(LM,&lda,&lda1,ipvt,&rcond,v);
#endif

   job = 1;
   lda = lda1 = dim;

#ifdef __SGI
   dgesl_(LM,&lda,&dim,ipvt,vec,&job);
#else
   dgesl(LM,&lda,&dim,ipvt,vec,&job);
#endif

   myfree(LM); myfree(v); myfree(ipvt);
}

CMatrix_Vector_Solve(LM, vec, dim)
COMPLEX **LM, *vec;
int dim;
{
  double **R, **M;
  double *M_inv, *R_x, *R_y, *M_x, *M_y, *v, *work;
  double **LeftM, **TmpM, **chckMat, **R_Minv;
  double det[2];
  double rcond;
  int i, j, k, *ipvt;
  int lda, lda1, job;

  ALLOCMAT(R, dim, dim, double);
  ALLOCMAT(M, dim, dim, double);
  ALLOCMAT(LeftM, dim, dim, double);
  ALLOCMAT(TmpM, dim, dim, double);
  ALLOCMAT(R_Minv, dim, dim, double);
  ALLOCMAT(chckMat, dim, dim, double);
  ALLOCN(M_inv, DOUBLE, dim*dim);
  ALLOCN(R_x, DOUBLE, dim);
  ALLOCN(R_y, DOUBLE, dim);
  ALLOCN(M_x, DOUBLE, dim);
  ALLOCN(M_y, DOUBLE, dim);
  ALLOCN(v, DOUBLE, dim);
  ALLOCN(work, DOUBLE, dim);
  ALLOCN(ipvt, int, dim);

  for (k=i=0; i<dim; i++) {
    for (j=0; j<dim; j++)
    {
      R[i][j] = LM[i][j].Re;
      M[i][j] = LM[i][j].Im;
      M_inv[k++] = M[i][j];
    }
    R_x[i] = vec[i].Re;
    M_x[i] = vec[i].Im;
  }

  lda1 = lda = dim;

#ifdef __SGI
  dgeco_(M_inv,&lda,&lda1,ipvt,&rcond,v);
#else
  dgeco(M_inv,&lda,&lda1,ipvt,&rcond,v);
#endif

  if (FP_EQ(rcond,0.0)) {
    printf("ILL_CONDITIONED MATRIX: QUITTING ... \n");
    exit(-1);
  }
  job = 01;
  dgedi_(M_inv,&lda,&lda1,ipvt,det,work,&job);

  for (k=i=0; i<dim; i++)
  for (j=0; j<dim; j++)
    TmpM[i][j] = M_inv[k++];

/* chckMat is just to check if it is an identity matrix */
  MATMULT(chckMat, TmpM, M, dim, dim, dim);

  MATMULT(R_Minv, R, TmpM, dim, dim, dim);
  MATVECMULT(R_y, R_Minv, R_x, dim, dim);
  for (i=0; i<dim; i++)
    R_y[i] += M_x[i];

  MATMULT(LeftM, R_Minv, R, dim, dim, dim);
  for (i=0; i<dim; i++)
  for (j=0; j<dim; j++)
    LeftM[i][j] += M[i][j];

  Matrix_Vector_Solve(LeftM, R_y, dim);

  MATMULT(LeftM, TmpM, R, dim, dim, dim);
  MATVECMULT(M_y, LeftM, R_y, dim, dim);
  MATVECMULT(M_x, TmpM, R_x, dim, dim);
  for (i=0; i<dim; i++)
    M_y[i] -= M_x[i];

  for (i=0; i<dim; i++) {
    vec[i].Re = R_y[i];
    vec[i].Im = M_y[i];
  }

  myfree(M_inv);
  myfree(R_x); myfree(R_y); myfree(M_x); myfree(M_y); myfree(work);
  FREEMAT(R, dim, dim);
  FREEMAT(M, dim, dim);
  FREEMAT(LeftM, dim, dim);
  FREEMAT(R_Minv, dim, dim);
  FREEMAT(chckMat, dim, dim);
  FREEMAT(TmpM, dim, dim);
}

/*
  Uses the structure of Block Companion Matrices to perform LU decomposition
  The matrix has been defined in the global variable C[DIM1][DIM1].
*/
  CDecompose(A,LM,order,degree,alpha1,alpha2)
    COMPLEX **A, **LM;
    COMPLEX alpha1, alpha2;
    int order,degree;
{
    int i,j,k,dim,base,lda,lda1;
    COMPLEX factor, tmp_comp;

    dim = order/degree; /* dim corresponds to the degree of the matrix
polynomial */

    base = order - degree;
    factor = divid(alpha2,alpha1);
    for (i = 1; i < dim; i++)
    {
         for (j = 0; j < degree; j++)
             for (k = 0; k < degree; k++)
             {
		 tmp_comp = mult(factor, A[base+j][(i-1)*degree + k]);
                 A[base+j][i*degree + k] = 
	 	 	     minus(A[base+j][i*degree + k], tmp_comp);
             }
    }

/* After the previous loop is completed, C basically contains a lower
   triangular matrix except for the bottom right degree X degree matrix.
   And that is precisely LM, which is computed below                     */

    for (j = 0; j < degree; j++)
       for (k = 0; k < degree; k++)
       {
            LM[j][k] = A[base + j][base + k];
       }
}

CSolve(A,LM,order,degree,alpha1,alpha2,vec)
COMPLEX **A, **LM;
int order, degree;
COMPLEX alpha1, alpha2;
COMPLEX *vec;
{
  int i,j,k,dim,base,lda,job;
  COMPLEX factor, *vec1;

  ALLOCN(vec1, COMPLEX, degree);

  base = order - degree;
  factor = divid(alpha2, alpha1);

  for (i = 0; i < base; i++)
  {
     vec[i] = divid(vec[i], alpha1);
  }  

  for (i = base; i < order; i++)
  {
    for (j = 0; j < base; j++)
      vec[i] = minus(vec[i], mult(A[i][j], vec[j]));
    vec1[i-base] = vec[i];
  }

  CMatrix_Vector_Solve(LM, vec1, degree);

  for (i=0; i<degree; i++)
    vec[i+base] = vec1[i];

  for (i=base-1; i>=0; i--)
    vec[i] = minus(vec[i], mult(factor, vec[i+degree]));

  myfree(vec1);
}
#endif
