#ifndef _LOOP
#include "defs.h"
#include "complex.h"

#define FRACTION	8

extern double Step_Size_Out;
extern LIST_POINT4D *gplp4DEndPoints;

extern  void  form_power_companion_matrix();
extern  void  add_list_point();
extern  void  matrix_singular();
extern	void  call_svd();
extern  int   calc_det();
EXTERN_LINALG

#define SQ(a) ((a)*(a))
#define		PLAY_SAFE 	     0
#define		MAX_ITER	     100
#define         MIN_FIVE_IMAG_VAL    0.05
#define         MIN_ONE_IMAG_VAL     0.01
#define         MAX_IMAG_PATHS       5    /* This constant gives the */
				          /* maximum number of positive */
				          /* complex paths whose imaginary */
				          /* values are less than */
				          /* MIN_IMAG_VAL */

COMPLEX max_cxval;

#ifdef _NEW_CTRACE
/* This CTrace code was written on 5/9/95. This is written to take
   care of reaching all the loop turning points including those from
   complex paths which did not originate at u = 0 */
void CTrace( ndet1D_uv, fdet1D_uv, tp0, tp1, pplp4DOutput, base_case, row_p )
NDETERMINANT_1D  	ndet1D_uv, fdet1D_uv;
TENSOR_PATCH            *tp0, *tp1;
LIST_POINT4D     	**pplp4DOutput;
int                     base_case, *row_p;
{
  int i, j, k, m, n;
  NDETERMINANT_1D       ndet1D_Tmp, ndet1D_v;
  NDETERMINANT_1D       fdet1D_Tmp, fdet1D_v;
  int			iDeg0, iDeg1, iter;
  double                curr_uval, step_uval;
  double                a, b, c, d;
  int                   iSizeResultant, iDeg_Curve, iSize_C;
  double                **rgrgB, *dEReal, *dEImag, **rgrgEVectors;
  COMPLEX               tmp_comp, tmp_comp2;
  int                   num_positive_imag_paths;
  int                   min_five_idx[MAX_IMAG_PATHS], num_five_min_imag_paths;
                        /* imag values less than 0.05 */
  int                   min_one_idx[MAX_IMAG_PATHS], num_one_min_imag_paths;
                        /* imag values less than 0.01 */
  POINT4D               pt4D;
  
  iDeg0 = ndet1D_uv.poly[0][0].iDeg0;
  iDeg1 = ndet1D_uv.poly[0][0].iDeg1;
  allocDeterminant_1D(&ndet1D_Tmp,ndet1D_uv.cRows,ndet1D_uv.cCols,iDeg0,iDeg1);
  allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols, 0, iDeg1);
  if (base_case) {
    allocDeterminant_1D(&fdet1D_Tmp,fdet1D_uv.cRows,fdet1D_uv.cCols,
			tp1->iDeg0, tp1->iDeg1);
    allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols,
			0, tp1->iDeg1);
  }

  /* copy ndet1D_uv to ndet1D_Tmp */
  copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);
  
  /* convert bernstein basis to power basis for each entry of Tmp */
  for(i = 0; i < ndet1D_Tmp.cRows; i++)
    for(j = 0; j < ndet1D_Tmp.cCols; j++)
      bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);
  
  if (base_case) {
    /* copy fdet1D_uv to fdet1D_Tmp */
    copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);
    
    /* convert bernstein basis to power basis for each entry of Tmp */
    for(i = 0; i < fdet1D_Tmp.cRows; i++)
      for(j = 0; j < fdet1D_Tmp.cCols; j++)
	bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
  }

  iDeg_Curve = ndet1D_v.poly[0][0].iDeg1;
  iSizeResultant = ndet1D_v.cRows;

  iSize_C = iSizeResultant*iDeg_Curve;

  ALLOCMAT(rgrgB, iSize_C, iSize_C, double);
  ALLOCN(dEReal, double, iSize_C);
  ALLOCN(dEImag, double, iSize_C);
  ALLOCMAT(rgrgEVectors, iSize_C, iSize_C, double);
  curr_uval = 0.0;
  while (curr_uval < (1.0+EPSILON)) {
    compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, curr_uval);

    a = b = c = d = 0.0;
    while (FP_EQ_EPS(a, 0.0, 1e-4)) 
      a = 0.7 + 0.3*drand48();
    while (FP_EQ_EPS(b, 0.0, 1e-4)) 
      b = 0.7 + 0.3*drand48();
    while (FP_EQ_EPS(c, 0.0, 1e-4))
      c = 0.7 + 0.3*drand48();
    while (FP_EQ_EPS(d, 0.0, 1e-4))
      d = 0.7 + 0.3*drand48();

    if (form_companion_matrix(ndet1D_v,rgrgB,iSize_C,a,b,c,d) == -1) {
      printf("New CTrace: form_companion_matrix : Leading Matrix SINGULAR\n");
      exit(-1);
    }
    call_rg(rgrgB, iSize_C, iSize_C, dEReal, dEImag, rgrgEVectors, 1); 
    num_five_min_imag_paths = num_one_min_imag_paths = 0;
    num_positive_imag_paths = 0;

    for (i=0; i<iSize_C; i++) {
      if (FP_EQ(dEImag[i], 0.0)) 
	 /* dEReal[i] = trans_inverse_power_pt(dEReal[i], a, b, c, d); */
	dEReal[i] = (a*dEReal[i] + b) / ((a+c)*dEReal[i] + b + d);
      else {
	tmp_comp.Re = a*dEReal[i] + b;
	tmp_comp.Im = a*dEImag[i];
	tmp_comp2.Re = (a+c)*dEReal[i] + b + d;
	tmp_comp2.Im = (a+c)*dEImag[i];
	 /* tmp_comp = trans_inv_cxpower_pt(tmp_comp, a, b, c, d); */
 	tmp_comp = divid(tmp_comp, tmp_comp2);
	dEReal[i] = tmp_comp.Re;
	dEImag[i] = tmp_comp.Im;
	if ((dEReal[i] < (1.0+EPSILON)) && (dEReal[i] > -EPSILON) && 
	    (dEImag[i] > 0.0)) {
	  num_positive_imag_paths++;
	  if (dEImag[i] < MIN_ONE_IMAG_VAL) 
	    min_one_idx[num_one_min_imag_paths++] = i;
	  else {
	    if (dEImag[i] < MIN_FIVE_IMAG_VAL) 
	      min_five_idx[num_five_min_imag_paths++] = i;
	  }
	}
      }
    }
    if (num_one_min_imag_paths > 0) {
      step_uval = 0.01;
      for (j=0; j<num_one_min_imag_paths; j++) {
	pt4D.dCoord[U] = curr_uval + 0.01;
	pt4D.dCoord[V] = dEReal[min_one_idx[j]];
	find_next_point(ndet1D_v, tp0->iDeg0, tp0->iDeg1, &pt4D, V,
			pt4D.dCoord[V] + 0.005, *tp0, *tp1, fdet1D_v,
			row_p, 1, &iter);
	if (iter < MAX_ITER) {
	  for (k=0; k<4; k++)
	    pt4D.dCoord[k] = power_pt(pt4D.dCoord[k]);
	  addPointToList(pt4D, pplp4DOutput);
	}
      }
    } else {
      if (num_five_min_imag_paths > 0) 
	step_uval = 0.02;
      else step_uval = 0.05;
    }
    curr_uval += step_uval;
  }
}
#else
void 
CTrace( ndet1D_uv,  p4DStart, plp4DStart, pplp4DOutput, m, n, tp0, tp1, dr )
NDETERMINANT_1D  	ndet1D_uv;
POINT4CPX          	p4DStart;
LIST_POINT4D     	**plp4DStart;
LIST_POINT4D     	**pplp4DOutput;
int			m, n, dr;
TENSOR_PATCH		tp0, tp1;
{
  POINT4CPX	  	p4DNext, p4DCurrent;
  POINT4D		ppp4D, ppm4D, ppav4D;
  double		ges;
  int		  	i, j, k, iter_val;
  int			iDeg0, iDeg1;
  NDETERMINANT_1D       ndet1D_Tmp, ndet1D_u, fdet1D_u, ndet1D_v;
  double		exp_step_size = 0.05;
  int			once_inc, once_dec, cnt, *row_p;
  double		chck_slope, diff_in_imag;

  max_cxval.Re = HUGE_VAL;
  max_cxval.Im = 0.0;
  /* allocate various vectors */
  iDeg0  = ndet1D_uv.poly[0][0].iDeg0;
  iDeg1  = ndet1D_uv.poly[0][0].iDeg1;
  allocDeterminant_1D(&ndet1D_Tmp,ndet1D_uv.cRows,ndet1D_uv.cCols,iDeg0,iDeg1);
  allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols, 1, iDeg0);
  allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols, 1, iDeg1);

    p4DCurrent = p4DStart;

    if (p4DCurrent.cpxCoord[V].Im < 0.5) exp_step_size = 0.01;
    if (p4DCurrent.cpxCoord[V].Im < 0.05) exp_step_size = 0.001;

    once_inc = 0;
    once_dec = 0;

    while ((p4DCurrent.cpxCoord[U].Re >= -EPSILON) && (p4DCurrent.cpxCoord[U].Re <= (1+EPSILON)) &&
           (p4DCurrent.cpxCoord[V].Im >= EPSILON) && (p4DCurrent.cpxCoord[V].Im <= (5+EPSILON)) )
/* &&
           (p4DCurrent.cpxCoord[S].Im >= EPSILON) && (p4DCurrent.cpxCoord[S].Im <= (5+EPSILON)) &&
           (p4DCurrent.cpxCoord[T].Im >= EPSILON) && (p4DCurrent.cpxCoord[T].Im <= (5+EPSILON)))
*/
    { 
      p4DNext.cpxCoord[U].Re = p4DCurrent.cpxCoord[U].Re + exp_step_size;
      p4DNext.cpxCoord[U].Im = p4DCurrent.cpxCoord[U].Im;
      p4DNext.cpxCoord[V].Re = p4DCurrent.cpxCoord[V].Re;
      p4DNext.cpxCoord[V].Im = p4DCurrent.cpxCoord[V].Im;
      p4DNext.cpxCoord[S].Re = p4DCurrent.cpxCoord[S].Re;
      p4DNext.cpxCoord[S].Im = p4DCurrent.cpxCoord[S].Im;
      p4DNext.cpxCoord[T].Re = p4DCurrent.cpxCoord[T].Re;
      p4DNext.cpxCoord[T].Im = p4DCurrent.cpxCoord[T].Im;

      if (!FP_EQ(p4DNext.cpxCoord[U].Im, 0.0)) {
	printf("Something WRONG!!! U is COMPLEX: %lf\n", p4DNext.cpxCoord[U].Im);
	exit(-1);
      }

      if ((p4DNext.cpxCoord[U].Re >= -EPSILON) && (p4DNext.cpxCoord[U].Re <= (1+EPSILON)))
      {
        copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

        for(i = 0; i < ndet1D_Tmp.cRows; i++)
          for(j = 0; j < ndet1D_Tmp.cCols; j++)
            bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);
  
	compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, p4DNext.cpxCoord[U].Re);

	if (find_next_cxpoint(ndet1D_v, m, n, &p4DNext, V, 
			p4DNext.cpxCoord[V], &iter_val) == -1) {
	  return;
	}

	if (iter_val < MAX_ITER) {
	  diff_in_imag = p4DCurrent.cpxCoord[V].Im - p4DNext.cpxCoord[V].Im;
	  if (p4DNext.cpxCoord[V].Im > p4DCurrent.cpxCoord[V].Im) {
	    exp_step_size *= 2.0;
	    if (exp_step_size > 0.05) exp_step_size = 0.05;
	  }
/*
	  if (p4DNext.cpxCoord[V].Im < p4DCurrent.cpxCoord[V].Im) {
	    chck_slope = (exp_step_size * p4DCurrent.cpxCoord[V].Im) /
	  	         (1.5 * diff_in_imag) + p4DCurrent.cpxCoord[U].Re;
	    if (chck_slope > 1+EPSILON) {
	      printf("Imaginary value not decreasing sufficiently ...\n");
	      return;
	    }
	    if ((p4DNext.cpxCoord[V].Re > p4DCurrent.cpxCoord[V].Re) &&
	        (p4DCurrent.cpxCoord[V].Re > 1+EPSILON)) {
	      printf("Looks like real part is outside the domain because %lf -> %lf\n", p4DCurrent.cpxCoord[V].Re, p4DNext.cpxCoord[V].Re);
	      return;
	    }
	    printf("Imaginary part decreasing with value %lf + i%lf\n",
		  p4DNext.cpxCoord[V].Re, p4DNext.cpxCoord[V].Im);
	  }
*/
	
 	  if (diff_in_imag > 0.0) {
	    exp_step_size = (p4DCurrent.cpxCoord[V].Im * exp_step_size) /
			    (FRACTION * diff_in_imag);
	    if (exp_step_size > 0.05) exp_step_size = 0.05;
	  }
	}
/*
	if (p4DNext.cpxCoord[V].Im < 0.5) exp_step_size = 0.01;
	if (p4DNext.cpxCoord[V].Im < 0.05) exp_step_size = 0.001;

*/
	if ((FP_EQ_EPS(p4DNext.cpxCoord[V].Im, 0.0, 1e-3)) &&
	    (FP_EQ_EPS(p4DNext.cpxCoord[S].Im, 0.0, 1e-3)) &&
	    (FP_EQ_EPS(p4DNext.cpxCoord[T].Im, 0.0, 1e-3)) &&
	    (p4DNext.cpxCoord[U].Re > -EPSILON) && 
	    (p4DNext.cpxCoord[U].Re < 1+EPSILON) &&
	    (p4DNext.cpxCoord[V].Re > -EPSILON) && 
	    (p4DNext.cpxCoord[V].Re < 1+EPSILON) &&
	    (p4DNext.cpxCoord[S].Re > -EPSILON) && 
	    (p4DNext.cpxCoord[S].Re < 1+EPSILON) &&
	    (p4DNext.cpxCoord[T].Re > -EPSILON) && 
	    (p4DNext.cpxCoord[T].Re < 1+EPSILON) ) 
 	{
	  ppm4D.dCoord[U] = p4DNext.cpxCoord[U].Re ;
	  ppm4D.dCoord[V] = p4DNext.cpxCoord[V].Re - 0.01;
	  ppm4D.dCoord[S] = p4DNext.cpxCoord[S].Re;
	  ppm4D.dCoord[T] = p4DNext.cpxCoord[T].Re;
	  ppp4D.dCoord[U] = p4DNext.cpxCoord[U].Re ;
	  ppp4D.dCoord[V] = p4DNext.cpxCoord[V].Re + 0.01;
	  ppp4D.dCoord[S] = p4DNext.cpxCoord[S].Re;
	  ppp4D.dCoord[T] = p4DNext.cpxCoord[T].Re;
 	  find_next_point(ndet1D_v, m, n, &ppm4D, V, ppm4D.dCoord[V], 
			  tp0, tp1, fdet1D_u, row_p, 0, &cnt);
	  find_next_point(ndet1D_v, m, n, &ppp4D, V, ppp4D.dCoord[V], 
			  tp0, tp1, fdet1D_u, row_p, 0, &cnt);
	  for (i=0; i<4; i++) {
	    ppav4D.dCoord[i] = (ppm4D.dCoord[i] + ppp4D.dCoord[i])/2.0;
	  }

	  if ((ppav4D.dCoord[U] > -EPSILON) && (ppav4D.dCoord[U] < 1+EPSILON) &&
	      (ppav4D.dCoord[V] > -EPSILON) && (ppav4D.dCoord[V] < 1+EPSILON) &&
	      (ppav4D.dCoord[S] > -EPSILON) && (ppav4D.dCoord[S] < 1+EPSILON) &&
	      (ppav4D.dCoord[T] > -EPSILON) && (ppav4D.dCoord[T] < 1+EPSILON))
	  {
	    add_list_point(pplp4DOutput, ppav4D, tp1, 0, 1, dr);

	    for (i=0; i<4; i++) {
	      ppm4D.dCoord[i] = power_pt(ppm4D.dCoord[i]);
	      ppp4D.dCoord[i] = power_pt(ppp4D.dCoord[i]);
	    }

	    addPointToList(ppm4D, plp4DStart);
	    addPointToList(ppp4D, plp4DStart);
	    return;
	  }
	}

	if (iter_val >= MAX_ITER) {
	  printf("ITER_VAL EXCEEDED LIMIT\n");
	  return;
	}
#ifdef __GL
        if (!FP_EQ(p4DCurrent.cpxCoord[V].Im, 0.0)) {
          dr_uvi_point(p4DCurrent.cpxCoord[U].Re, 
		       p4DCurrent.cpxCoord[V].Re, 
		       p4DCurrent.cpxCoord[V].Im);
          dr_uvi_point(p4DCurrent.cpxCoord[U].Re, 
		       p4DCurrent.cpxCoord[V].Re, 
		       -p4DCurrent.cpxCoord[V].Im);
        }
#endif
	p4DCurrent = p4DNext;
      }
      else break;
    }

/*
    printf("Off patch at (uv) %lf %lf %lf (bernstein).\n\n",
    p4DNext.cpxCoord[U].Re, p4DNext.cpxCoord[V].Re, p4DNext.cpxCoord[V].Im);
*/

}
#endif

/* expects ndet1D_u.poly in power basis, p4Duvst in bernstein,
   m = deg of patch in u, n = degree of patch in v,
   dRootGuess is in bernstein_basis, and returns
   the new point in p4Duvst in bernstein  basis */

int find_next_cxpoint(ndet1D_u, m, n, p4Duvst, iPolyIn, dRootGuess, it_val )
NDETERMINANT_1D		ndet1D_u;
int			iPolyIn, m, n;
POINT4CPX		*p4Duvst;
COMPLEX                 dRootGuess;
int			*it_val;
{ 
  int			i, iSize;
  double		**rgrgB, **rgrgA_plus_B;
  double		**rgrgU, **rgrgV, *rgSigma;
  COMPLEX		*rgEVec;
  COMPLEX		iPolyIn_val;

  iSize = ndet1D_u.cRows*ndet1D_u.poly[0][0].iDeg1;

  ALLOCMAT(rgrgB, iSize, iSize, double);
  ALLOCMAT(rgrgU, iSize, iSize, double);
  ALLOCMAT(rgrgV, iSize, iSize, double);
  ALLOCN(rgSigma, DOUBLE, iSize);
  ALLOCMAT(rgrgA_plus_B, iSize, iSize, double);
  ALLOCN(rgEVec, COMPLEX, iSize);

  /* convert all values into power-basis space */
  for(i = 0; i < 4; i++) {
    p4Duvst->cpxCoord[i] = cxpower_pt(p4Duvst->cpxCoord[i]);
    if ((p4Duvst->cpxCoord[i]).Re == HUGE_VAL) return -1;
  }

  iPolyIn_val = p4Duvst->cpxCoord[iPolyIn];

  form_power_companion_matrix(ndet1D_u, rgrgB, rgrgA_plus_B, iSize);
/*
  form_companion_matrix(ndet1D_u, rgrgB, iSize);
  MATASSIGN(rgrgA_plus_B, rgrgB, iSize, iSize);
  for (i=0; i<iSize; i++)
    rgrgA_plus_B[i][i] += 1.0;

  call_svd(rgrgA_plus_B, iSize, iSize, rgrgU, rgSigma, rgrgV);
  for (i=0; i<iSize; i++)
    printf("Sigma[%d] = %18.15lf\n", i, rgSigma[i]);
*/

  construct_kernel_cxvec(ndet1D_u.poly[0][0].iDeg1, m, n, 
		       cxpower_pt(dRootGuess), p4Duvst->cpxCoord[S], 
		       p4Duvst->cpxCoord[T], rgEVec);

  /* get the new values of s, t, u, v */
  cx_generalized_power_iterate(rgrgB, rgrgA_plus_B, rgEVec, iSize, 
		ndet1D_u.cRows, dRootGuess, &iPolyIn_val, m, it_val); 

  p4Duvst->cpxCoord[iPolyIn]= cxpower_pt(iPolyIn_val);
  p4Duvst->cpxCoord[S]= divid(rgEVec[1],rgEVec[0]);
  p4Duvst->cpxCoord[T]= divid(rgEVec[m],rgEVec[0]);
/*
  This code was commented because even though rgEVec[0] might be
  very small rgEVec[1 or m] might be smaller and S and T values
  might be accurate
  p4Duvst->cpxCoord[S]= (FP_EQ(rgEVec[0].Re,0.0) && FP_EQ(rgEVec[0].Im,0.0)) ? 
		       max_cxval : divid(rgEVec[1],rgEVec[0]);
  p4Duvst->cpxCoord[T]= (FP_EQ(rgEVec[0].Re,0.0) && FP_EQ(rgEVec[0].Im,0.0)) ?
		       max_cxval : divid(rgEVec[m],rgEVec[0]);
*/

  /* convert all values back into Bernstein-basis space */
  for(i = 0; i < 4; i++)
    p4Duvst->cpxCoord[i] = inv_cxpower_pt(p4Duvst->cpxCoord[i]);

  FREEMAT(rgrgB, iSize, iSize);
  FREEMAT(rgrgU, iSize, iSize);
  FREEMAT(rgrgV, iSize, iSize);
  FREEMAT(rgrgA_plus_B, iSize, iSize);
  myfree(rgEVec); myfree(rgSigma);
  return 0;
}


/* constructed as (1, s, s^2 .... s^m, st, s^2t, .....s^mt^2n) */
/* refer determinant.c for further details, also, dLamda is how far
from the current eigenvalue you expect the new one to be (sign and
direction) (in power basis).
*/
construct_kernel_cxvec(cDegCurve, m, n, dLambda, s, t, rgEVec)
int			cDegCurve, m, n;
COMPLEX			dLambda, s, t;
COMPLEX			*rgEVec;
{ 
  int			i, j, k, l;
  COMPLEX		*pow_s, *pow_t, *pow_lambda;

  ALLOCN(pow_s, COMPLEX, m);
  ALLOCN(pow_t, COMPLEX, 2*n);
  ALLOCN(pow_lambda, COMPLEX, cDegCurve);

  pow_s[0].Re = pow_t[0].Re = pow_lambda[0].Re = 1.0;
  pow_s[0].Im = pow_t[0].Im = pow_lambda[0].Im = 0.0;

  for(i = 1; i < m; i++) 
    pow_s[i] = mult(s,pow_s[i-1]);
  for(i = 1; i < 2*n; i++)
    pow_t[i] = mult(t,pow_t[i-1]);
  for(i = 1; i < cDegCurve; i++)
    pow_lambda[i] = mult(dLambda,pow_lambda[i-1]);
  
  for (l = k = 0; l < cDegCurve; l++)
	 for(i = 0; i < 2*n; i++)
	  for(j = 0; j < m; j++)
		 rgEVec[k++] = mult(mult(pow_lambda[l],pow_s[j]),pow_t[i]);
  
  myfree(pow_s); myfree(pow_t); myfree(pow_lambda);
}
#endif
