#ifndef _defs_h
#define _defs_h

#include <stdio.h>
#include <math.h>
#include <errno.h>
#include "linalg.h"
#include "fitting.h"

#ifdef _DMALLOC
#include "dmalloc.h"
#else
#include <stdlib.h>
#endif

#define DOUBLE double

/* fuzzy compare */

#define FP_EQ_EPS( a, b, c )  ((((a) - (b)) < (c)) && (((a) - (b)) > -(c)))

#define EPSILON  1.0e-9
#define ODD( a )     ( (a) % 2 == 1 )

#define FP_EQ( a, b )  ((((a) - (b)) < EPSILON) && (((a) - (b)) > -EPSILON))

#define EQ_4D( a, b ) ( (FP_EQ((a).dCoord[0], (b).dCoord[0])) && \
			(FP_EQ((a).dCoord[1], (b).dCoord[1])) && \
			(FP_EQ((a).dCoord[2], (b).dCoord[2])) && \
			(FP_EQ((a).dCoord[3], (b).dCoord[3])) )

#define EQ_4D_EPS(a,b,c) ((FP_EQ_EPS((a).dCoord[0],(b).dCoord[0],(c))) && \
			  (FP_EQ_EPS((a).dCoord[1],(b).dCoord[1],(c))) && \
			  (FP_EQ_EPS((a).dCoord[2],(b).dCoord[2],(c))) && \
			  (FP_EQ_EPS((a).dCoord[3],(b).dCoord[3],(c))) )

#define NEW_EQ_4D_EPS(a,b,c)  ((FP_EQ_EPS((((a).dCoord[0])/((a).dCoord[3])),(((b).dCoord[0])/((b).dCoord[3])),(c)))&& \
 (FP_EQ_EPS((((a).dCoord[1])/((a).dCoord[3])),(((b).dCoord[1])/((b).dCoord[3])),(c))) && \
 (FP_EQ_EPS((((a).dCoord[2])/((a).dCoord[3])),(((b).dCoord[2])/((b).dCoord[3])),(c)))) 

#define EQ_2D( a, m, n, b ) ( ( (((a).dCoord[m]) -  ((b).u)) * \
			        (((a).dCoord[m]) -  ((b).u)) + \
			        (((a).dCoord[n]) -  ((b).v)) * \
			        (((a).dCoord[n]) -  ((b).v)) ) < 1.0e-5 ) /* was 1e-3 */

#define EQUAL2D( a, b ) ( ( (((a).u) -  ((b).u)) * \
                            (((a).u) -  ((b).u)) + \
                            (((a).v) -  ((b).v)) * \
                            (((a).v) -  ((b).v)) ) < 1.0e-5 ) /* was 1e-3 */

#define EQUAL2D_EPS( a, b, c ) ( ( (((a).u) -  ((b).u)) * \
                            (((a).u) -  ((b).u)) + \
                            (((a).v) -  ((b).v)) * \
                            (((a).v) -  ((b).v)) ) < (c) )

#define EQUAL3D_ABS( a, b, c ) (((((a).dCoord[0]) - ((b).dCoord[0])) * \
				 (((a).dCoord[0]) - ((b).dCoord[0])) + \
				 (((a).dCoord[1]) - ((b).dCoord[1])) * \
                                 (((a).dCoord[1]) - ((b).dCoord[1])) + \
				 (((a).dCoord[2]) - ((b).dCoord[2])) * \
                                 (((a).dCoord[2]) - ((b).dCoord[2]))) \
				 < (c) )

#define EQUAL3D_EPS( a, b, c ) ((((((a).dCoord[0]) - ((b).dCoord[0])) * \
				  (((a).dCoord[0]) - ((b).dCoord[0])) + \
				  (((a).dCoord[1]) - ((b).dCoord[1])) * \
                                  (((a).dCoord[1]) - ((b).dCoord[1])) + \
				  (((a).dCoord[2]) - ((b).dCoord[2])) * \
                                  (((a).dCoord[2]) - ((b).dCoord[2]))) / \
				 ((((a).dCoord[0]) * ((a).dCoord[0])) + \
				  (((a).dCoord[1]) * ((a).dCoord[1])) + \
				  (((a).dCoord[2]) * ((a).dCoord[2])))) \
				 < (c) )

/* boolean consts */
#define		FALSE		0
#define		TRUE		1

#define		ERROR_VAL	-10.0
#define		MAX_RAND_VAL	2147483647.0

/* the order and values of these should be contiguous */
/* added H to comply with BRLCAD */
#ifdef INBRLCAD
#define         X               0
#define         Y               1
#define         Z               2
#define		H		3
#define		W		H
#else
#define         X               0
#define         Y               1
#define         Z               2
#define		W		3
#define         U               0
#define         V               1
#define         S               2
#define         T               3
#endif

/* Constants for CSG_STATUS */
#define IN_CSG		1
#define OUT_CSG		-1
#define PART_CSG	0

/* Constants for partition_plygon call */
#define MAX_NEIGHBOURS  10

/* Constants for CSG_OPERATION */
#define DOUNION		1
#define INTERSECTION	2
#define DIFFERENCE	3
#define PLANE_CUT	4

/* Constants for ORIENTATION */
#define	LEFT		0
#define RIGHT		1

/* flags for allocDeterminant call, saying how the size should be 
   calculated */
#define FINAL       1
#define UNFACTORED  2

/* flags for outputPolyLines call, specifying the stage of writing points */
#define START       1
#define CONTINUE    2
#define FINISH      3
#define UVST_SIZE  256

/* The constant TINFO_PT_LIST_SIZE defines the number of elements
   of the array ptList in the TRACE_INFO structure */
#define TINFO_PT_LIST_SIZE 8 
#define MAX_COMP_PER_PATCH 15 
#define MAX_PATCHES_PER_SOLID  2000
#define MAX_PARTNS_PER_PATCH  15
#define MAX_TRIM_CURVES_PER_PATCH  10
#define MAX_SPLINES_PER_TRIM_REGION 100
#define MAX_SPLIT_ARRAY_LENGTH 20
#define MAX_LOOPS_PER_PATCH 4
#define MAX_CURVE_DEG 5
#define MAX_POINTS_PER_TRIM_REGION 400
#define MAX_VERTEX_COUNT  300
#define MAX_NUM_PORTIONS  15
#define MAX_SHADY_POINTS  10
#define POW_TOL        	  1e-4

#define W_ST 1
#define W_UV 2

/* Constants for loop and singular cases */
#define NO_LOOP_CASE 0
#define LOOP_CASE 1
#define SINGULAR_CASE	2

/* macros to access the Dixon matrix */
#define ROW_INDEX( a, b, maxDegB ) ((a)*((maxDegB)+1)+(b))
#define COL_INDEX( s, t, maxDegS ) ((t)*((maxDegS)+1)+(s))
#define  ALLOCN(PTR,TYPE,N)                                                \
         { if (N<=0) PTR = (TYPE *) NULL;                                    \
	 else if ((PTR = (TYPE *) malloc((unsigned) (N)*sizeof(TYPE)))==NULL) { \
	     printf("Tried to allocate %d, %d bytes\n", N,sizeof(TYPE));	\
	     printf("ERRONO is %d\n", errno);				\
	     perror("malloc failed");                                    \
	     exit(-1);                                                     \
	 } }

typedef struct _POINT2D {
    double  u,v;
} POINT2D;

typedef struct { 
    double 	dCoord[3];
} POINT3D;

typedef struct { 
    double 	dCoord[4];
} POINT4D;

typedef struct {
    double low;
    double high;
} INTERVAL;

typedef struct {
    INTERVAL i3Dbox[3];
} BOUNDINGBOX;

typedef struct {
    INTERVAL i2Dbox[2];
} RECT_BOX;

typedef struct {
    int Npts;
    POINT4D *ptList;
    RECT_BOX domain_def;
} TRACE_INFO;

typedef struct {
    int patchNum0;
    int patchNum1;
} OVERLAPBOXES;

typedef struct {
    OVERLAPBOXES ptches;
    int          same_side;
} OVR_BOXES;

typedef struct {
    double 	Re;
    double 	Im;
} COMPLEX;

typedef struct {
    COMPLEX	cpxCoord[4];
} POINT4CPX;

typedef struct _LIST_POINT4D {
    POINT4D               p4D;
    struct _LIST_POINT4D  *plp4DNext;
} LIST_POINT4D;

typedef struct _LIST_POINT2D {
    POINT2D               p2D;
    struct _LIST_POINT2D  *plp2DNext;
} LIST_POINT2D;

typedef struct {
    LIST_POINT4D	*Curve;
    POINT4D		endPoints[2];
} INT_CURVE_STRUCT;

typedef struct {
    LIST_POINT2D	*planeCurve;
    POINT2D		StEndPoints[2];
} DOMAIN_INT_CURVE;

typedef struct _LIST_POINT4CPX {
    POINT4CPX               p4CPX;
    struct _LIST_POINT4CPX  *plp4CPXNext;
} LIST_POINT4CPX;

/* iDeg0 is degree in s, call it m, iDeg1 is degree in t, call it n */
typedef struct 
{ int		iDeg0, iDeg1;	/* degrees of the tensor product patch */
  POINT4D	**coeffs;	/* coefficients of the tensor product patch */
} TENSOR_PATCH;

typedef struct {
  int Deg;
  POINT3D *CntrlPts;
} UV_SPLINE_TYPE;

typedef struct {
  int Deg;
  POINT4D *CntrlPts;
} XYZ_SPLINE_TYPE;

typedef struct _INTERSECTION_CURVE {
    UV_SPLINE_TYPE		uv_spline;
    XYZ_SPLINE_TYPE		xyz_spline;
    int				neighbor;
    struct _INTERSECTION_CURVE	*other_solid_ptr;
    struct _INTERSECTION_CURVE	*next; 
} INTERSECTION_CURVE;

typedef struct {
    short   which_solid; /* is 1 or 2 depending on first or second */
                         /* solid */
    short   same_solid;
    short   patch_num;
    double  *param;
    short   param_length; /* gives length of the param array */
    POINT4D *st_pt, *end_pt;
    short   do_twice;
    short   other_patch_num, other_region_num;
} SPLIT_TPST_STRUCT;

typedef struct {
    INTERSECTION_CURVE	*int_curve;
} MERGED_CURVE;

typedef struct {
    int entering_from, going_into;
} CURVE_ADJCY;

/* iDeg0 is degree in u, call it m, iDeg1 is degree in v, call it n */
typedef struct 
{ int		iDeg0, iDeg1;	/* degrees of the tensor product patch */
  double	**coeffs;	/* coefficients of the tensor product patch */
} TENSOR_PATCH_1D;

typedef struct {
  int NumOfCurves; /* Number of curves defining the trim curve */
  UV_SPLINE_TYPE *UV_Splines; /* definition of each spline both in domain and */
                           /* space */
  XYZ_SPLINE_TYPE *XYZ_Splines; 
  int *AdjInfo; /* Adjacency info for each spline */
  POINT2D ptInsideRegion; /* one point inside the trim curve */
} TRIM_PORTION_SPLINE_TYPE;

typedef struct {
  int NumOfVertices;
  double (*poly_vertex)[2];
  int *AdjInfo;   
  POINT2D ptInsideRegion;
} TRIM_PORTION_PIECEWISE_TYPE;

typedef union {
    TRIM_PORTION_SPLINE_TYPE	SplinePortion;
    TRIM_PORTION_PIECEWISE_TYPE PiecePortion;
} TRIM_PORTION_TYPE;

typedef struct {
    TENSOR_PATCH	tp;	/* tensor patch */
    TENSOR_PATCH	orig_tp; /* original, undivided tpatch */
    int			primNum;
    int			SplineReprn;
    int			NumTrimPortions;
    TRIM_PORTION_TYPE	TrimPortion[MAX_NUM_PORTIONS];
    int			csg_status;
    double		umin, umax, vmin, vmax;
} TRIM_PATCH_INFO;

typedef struct {
    POINT4D             spt, ept;
    int                 patchNo;
    int                 adj_pnum;
} NEIGHBORING_CURVE;

typedef struct {
    POINT3D		pt3d;
    short		patchNo;
    short		s_or_e; /* start or end point in the curve */
				/* 0 - start, 1 - end */
    short		marked; /* -1 - unmarked, otherwise */
				/* adjacent patch number */
    short		curve_num; /* Xn. curve number of that patch */
    short               my_curve_no; /* trimming curve number where pt
					lies */
    short               adj_patch_no; /* corresponding patch number */
    short               adj_curve_no; /* corresponding adj trimming
					 curve no. */
} CURVE_MATCHING_TYPE;

typedef struct {
    double              param;
    int                 bndry_crv_num;
} ARBIT_STRUCT;

typedef struct {
    POINT4D             p4d;
    int                 bndry_crv_num;
} ANOTHER_ARBIT_STRUCT;

typedef struct {
/* spl_c_num - spline curve num; ending - 0 or 1; 
   merge_c_num - merged curve num; param - parameter of spl_c */

    short		spl_c_num, ending, merge_c_num;
    double		param;
    POINT4D             p4d;
} PART_SPLINE_TYPE;

typedef struct {
  TRIM_PATCH_INFO *PatchPtr;
  int             NumPatches;
} SOLID;

typedef struct {
  double x1, y1, x2, y2;
  double Cx, Cy, Cc;
} LINE;

/* Here, we are using a,b,s,t as in sederberg's thesis, so
    (assuming, b from 0 to n, a from 0 to 2m, s from 0 to m, t from 
     0 to 2n):
        cRows = (2m + 1)(n + 1)
	cCols = (2n + 1)(m + 1) */

typedef struct _NDETERMINANT 
{
    int        cRows, cCols;
    POINT4D    **coeffs;
} NDETERMINANT;

typedef struct {
  int curve_num;
  double param_arr[20];
  int param_length;
  int initial;
  int adj_patch;
} SAME_SIDE_OVR_STRUCT;

typedef struct {
  int adj[MAX_NEIGHBOURS];	/* neighbouring partitions */
  int nnb;			/* #neighbours */
} neighbour_t;

typedef struct _NDETERMINANT_1D 
{
    int        		   cRows, cCols;
    double		   u, v;
    double		   **eval;
    TENSOR_PATCH_1D        **poly;
} NDETERMINANT_1D;

extern double   power_pt();
extern double inverse_power_pt();
extern form_companion_matrix();


#endif /* _defs_h */
