#include "defs.h"
#include "complex.h"
#include <sys/time.h>
#include "lp.h"
#ifdef __GL
#include <gl.h>
#endif

extern TENSOR_PATCH pow_tp0;
extern TENSOR_PATCH pow_tp1;
extern double Step_Size_Out;
extern  int     compar();
extern  int     comparU();
extern  int     comparV();
extern  double  dist2D();

extern  int    	imp_time, cs_time, sp, trace_time;

EXTERN_LINALG

#define NEW_EPS		5.0*1e-5

int getNumUList(uvstList, NumPoints, endUVal, domain)
POINT4D *uvstList;
int NumPoints;
double **endUVal;
RECT_BOX *domain;
{
  int i, j, k;
  double previous_max;
  int numUIntervals;

  if (NumPoints == 0) return 0;

/* I am mallocing NumPoints+2 because I think we can prove that there
   cannot be more than NumPoints+2 v-intervals */
  ALLOCN(*endUVal, double, NumPoints+2);
   previous_max = (*endUVal)[0] = domain->i2Dbox[0].low;
   /* previous_max = (*endUVal)[0] = uvstList[0].dCoord[U]; */

  for (k=i=0; i<NumPoints; i++) 
    if (uvstList[i].dCoord[U] > previous_max) {
      k++;
      previous_max = (*endUVal)[k] = uvstList[i].dCoord[U];
    }
  if (previous_max < domain->i2Dbox[0].high-EPSILON)
    (*endUVal)[++k] = domain->i2Dbox[0].high;

  numUIntervals = k;
  return numUIntervals;
}
 
int getNumVList(uvstList, NumPoints, endVVal, domain)
POINT4D *uvstList;
int NumPoints;
double **endVVal;
RECT_BOX *domain;
{
  int i, j, k;
  double previous_max;
  int numVIntervals;

  if (NumPoints == 0) return 0;

/* I am mallocing NumPoints+2 because I think we can prove that there
   cannot be more than NumPoints+2 v-intervals */
  ALLOCN(*endVVal, double, NumPoints+2);
   previous_max = (*endVVal)[0] = domain->i2Dbox[1].low;
   /* previous_max = (*endVVal)[0] = uvstList[0].dCoord[V]; */

  for (k=i=0; i<NumPoints; i++) 
    if (uvstList[i].dCoord[V] > previous_max) {
      k++;
      previous_max = (*endVVal)[k] = uvstList[i].dCoord[V];
    }
  if (previous_max < domain->i2Dbox[1].high-EPSILON)
    (*endVVal)[++k] = domain->i2Dbox[1].high;

  numVIntervals = k;
  return numVIntervals;
}
 
doDomainDecomp(uvstList, NumPoints, domain, s_dom, fdet1D_uv,
               ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
               tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
               fdet1D_Tmp, fdet1D_u, fdet1D_v,
               row_perm, base_case, cComp, compNo, first_time)
POINT4D *uvstList;
int NumPoints;
RECT_BOX domain, *s_dom;
NDETERMINANT_1D fdet1D_uv, ndet1D_uv;
NDETERMINANT_1D ndet1D_du, ndet1D_dv;
TENSOR_PATCH tpPatch0, tpPatch1;
INT_CURVE_STRUCT **cComp;
NDETERMINANT_1D      ndet1D_Tmp, ndet1D_u, ndet1D_v;
NDETERMINANT_1D      fdet1D_Tmp, fdet1D_u, fdet1D_v;
int *row_perm;
int base_case, *compNo, first_time;
{

  POINT4D p4DEnd;
  LIST_POINT4D *pplp4DOutput;
  double min_dom_width, max_dom_width, udiff, vdiff;
  double s_dom_domain_dist, ul_diff, uh_diff, vl_diff, vh_diff;
  double trace_step_size;
  int numVList, numUList;
  TRACE_INFO *t_info;
  double *endVVal, *endUVal, RetVal;
  double *nDPt, **nnDirVec;
  double ftol, bisect_pt, d2D_distance;
  int ord, NumIter;
  int                  num_points, num_cpoints;
  POINT4D              *pdPoints, *newEndPt;
  POINT2D              tpt2_1, tpt2_2;
  POINT4CPX            *pcpxPoints;
  struct timeval   tp1, tp2;
  struct timezone  tzp1, tzp2;
  float vert[2];
  int i, j, k, m, n, cIndex, ptIndex;
  int AnyOtherIndex, retCode;
  int tmpCompNo, traceRet;

  udiff = domain.i2Dbox[0].high - domain.i2Dbox[0].low;
  vdiff = domain.i2Dbox[1].high - domain.i2Dbox[1].low;
  min_dom_width = (udiff < vdiff) ? udiff : vdiff;
  max_dom_width = (udiff > vdiff) ? udiff : vdiff;

  m = tpPatch0.iDeg0;
  n = tpPatch0.iDeg1;
  i = (tpPatch1.iDeg0 > tpPatch1.iDeg1)? tpPatch1.iDeg0: tpPatch1.iDeg1;
  ALLOCN(pdPoints, POINT4D, i*ndet1D_uv.cRows);

/* The following code just removes some startpoints which are very
   close to some others. This causes problems in domain decomposition.
*/
  if (NumPoints > 2) {
    j=0; 
    while (j<NumPoints) {
      k = j+1;
      while (k < NumPoints) {
	if (EQ_4D_EPS(uvstList[j], uvstList[k], 1e-2)) {
	  for (i=k+1; i<NumPoints; i++) 
	    uvstList[i-1] = uvstList[i];
	  NumPoints--;
	} else k++;
      }
      j++;
    }
  }

#ifdef __GL
  winset(giduv);
  color(WHITE);
  bgnline();
  for (i=0; i<=udiff*0.85*UVST_SIZE; i++) {
    vert[0] = i + (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=vdiff*0.85*UVST_SIZE; i++) {
    vert[0] = (0.1+domain.i2Dbox[0].high*0.85)*UVST_SIZE;
    vert[1] = i + (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=udiff*0.85*UVST_SIZE; i++) {
    vert[0] = i + (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = (0.1+domain.i2Dbox[1].high*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=vdiff*0.85*UVST_SIZE; i++) {
    vert[0] = (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = i + (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  winset(gidxyz);
#endif

/* Actually FIRST_TIME is a wrong name. If it is 1, it actually
   meant that it was called inside doDomain.. itself */
  if (NumPoints == 0) return;
  if (NumPoints < 2 && first_time) return;
  if (NumPoints == 2) {
#ifdef TRIVTRACE
    if (max_dom_width <= 0.06) {
      if (gettimeofday(&tp1, &tzp1)) {
        printf("Error getting time\n");
        exit(-1);
      }
      TrivialTrace(&pplp4DOutput, uvstList[0], uvstList[1], tpPatch1);
      if (gettimeofday(&tp2, &tzp2)) {
        printf("Error getting time\n");
        exit(-1);
      }
      trace_time += (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);
      return;
    }
#endif
    if (FP_EQ_EPS(min_dom_width, 0.0, 1e-3)) return;
    if (min_dom_width < 0.002) trace_step_size = min_dom_width/2.0;
    else if (max_dom_width <= 0.2) trace_step_size = min_dom_width/4.0;
     /* else if (max_dom_width <= 0.2) trace_step_size = 0.008; */
    else trace_step_size = Step_Size_Out;

    ALLOCN(newEndPt, POINT4D, 2);
    cIndex = ptIndex = -1;
    tmpCompNo = 0;
    if (*compNo == 0) {
      cIndex = 0;
      if (uvstList[0].dCoord[V] < uvstList[1].dCoord[V])
        ptIndex = 0;
      else if (FP_EQ_EPS(uvstList[1].dCoord[V],uvstList[0].dCoord[V],1e-6)){
        if (uvstList[0].dCoord[U] < uvstList[1].dCoord[U])
          ptIndex = 0;
        else ptIndex = 1;
      }
      else ptIndex = 1;
      tmpCompNo = 1;
      AnyOtherIndex = -1;
      (*cComp)[cIndex].endPoints[0] = uvstList[ptIndex];
      (*cComp)[cIndex].endPoints[1] = uvstList[1-ptIndex];
    }
    else {
      setUpTraceOrder(cComp, uvstList, *compNo, &cIndex,
                      &ptIndex, &AnyOtherIndex, &retCode, &newEndPt);
      if ((cIndex == -1) && (ptIndex == -1)) {
        cIndex = *compNo;
        if (uvstList[0].dCoord[V] < uvstList[1].dCoord[V])
          ptIndex = 0;
        else if (uvstList[1].dCoord[V] == uvstList[0].dCoord[V]) {
          if (uvstList[0].dCoord[U] < uvstList[1].dCoord[U])
            ptIndex = 0;
          else ptIndex = 1;
        }
        else ptIndex = 1;
        tmpCompNo = 1;
        AnyOtherIndex = -1;
        if (cIndex == MAX_COMP_PER_PATCH) {
          printf("Too many components. Halting Intersection.\n");
          return;
        }
        (*cComp)[cIndex].endPoints[0] = uvstList[ptIndex];
        (*cComp)[cIndex].endPoints[1] = uvstList[1-ptIndex];
      }
      else tmpCompNo = 2;
    }
     
    if (gettimeofday(&tp1, &tzp1)) {
      printf("Error getting time\n");
      exit(-1);
    }

/* The following few lines of code is added to take care of
   tracing when the endpoints are closer than the step size */
    tpt2_1.u = uvstList[0].dCoord[0];
    tpt2_1.v = uvstList[0].dCoord[1];
    tpt2_2.u = uvstList[1].dCoord[0];
    tpt2_2.v = uvstList[1].dCoord[1];
    d2D_distance = dist2D(&tpt2_1, &tpt2_2);
    d2D_distance = sqrt(d2D_distance);
    if (d2D_distance < trace_step_size)
      trace_step_size = 0.7 * d2D_distance; /* 0.7 was chosen because
					      sqrt(2) is minimum */
    
    if (!(traceRet = Trace(ndet1D_uv, ndet1D_du, ndet1D_dv,
                           uvstList[ptIndex], domain, &p4DEnd,
          &((*cComp)[cIndex].Curve), tpPatch0, tpPatch1, 0, 0, fdet1D_uv,
          row_perm, base_case, 1, trace_step_size, tmpCompNo, compNo)) ) {
      if (!EQ_4D(p4DEnd, uvstList[1-ptIndex]))
        add_list_point(&((*cComp)[cIndex].Curve), uvstList[1-ptIndex],
                       tpPatch1, 0, 0, 1);
      if (tmpCompNo == 2) {
        (*cComp)[cIndex].endPoints[0] = newEndPt[0];
        (*cComp)[cIndex].endPoints[1] = newEndPt[1];
      }
    }
    myfree(newEndPt);
    if (gettimeofday(&tp2, &tzp2)) {
      printf("Error getting time\n");
      exit(-1);
    }
    trace_time += (tp2.tv_sec - tp1.tv_sec)*1e6 +
                  (tp2.tv_usec - tp1.tv_usec);

    if (AnyOtherIndex != -1 && !traceRet)
      MergeCurveLists(cComp, cIndex, AnyOtherIndex, retCode, compNo);
    return;
  }
  if (NumPoints == 4) {
   if (max_dom_width <= 0.06) {
    printf("Looks like we are either approaching a singularity\n");
    printf("or have two very close components\n");
    /* TRACE SINGULARITY PROPERLY WITH VERY SMALL STEP SIZE*/
    pow_tp0 = tpPatch0;
    pow_tp1 = tpPatch1;
    ord = 4;
    ftol = 0.01;
    ALLOCN(nDPt, double, 4);
    ALLOCMAT(nnDirVec, 4, 4, DOUBLE);
    RetVal = minim_func_for_sing(nDPt);
    nDPt[0] = uvstList[0].dCoord[0];
    nDPt[1] = uvstList[0].dCoord[1];
    nDPt[2] = uvstList[0].dCoord[2];
    nDPt[3] = uvstList[0].dCoord[3];
/*
    nDPt[0] = (domain.i2Dbox[0].low + domain.i2Dbox[0].high) / 2.0;
    nDPt[1] = (domain.i2Dbox[1].low + domain.i2Dbox[1].high) / 2.0;
*/
    for (i=0; i<4; i++)
    for (j=0; j<4; j++)
      nnDirVec[i][j] = (i == j) ? 1.0 : 0.0;
    powell(nDPt, nnDirVec, ord, ftol, &NumIter, &RetVal, minim_func_for_sing);
   printf("Values returned from POWELL are %g  %g  %g\n", nDPt[0], nDPt[1], RetVal);
    if (FP_EQ_EPS(RetVal, 0.0, 1.0e-10) &&
        nDPt[0] > (domain.i2Dbox[0].low + 1e-3) &&
        nDPt[0] < (domain.i2Dbox[0].high - 1e-3) &&
        nDPt[1] > (domain.i2Dbox[1].low + 1e-3) &&
        nDPt[1] < (domain.i2Dbox[1].high - 1e-3)) {
      printf("Actually Located a SINGULARITY at U = %g and V = %g\n", nDPt[0], nDPt[1]);
      ALLOCN(t_info, TRACE_INFO, 4); /* 4 is because we create four
                                        new subdomains around the
                                        singular point */
      for (i=0; i<4; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }
      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[0].high = nDPt[0];
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[0].domain_def.i2Dbox[1].high = nDPt[1];
      t_info[1].domain_def.i2Dbox[0].low = nDPt[0];
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[1].high = nDPt[1];
      t_info[2].domain_def.i2Dbox[0].low = nDPt[0];
      t_info[2].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[2].domain_def.i2Dbox[1].low = nDPt[1];
      t_info[2].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
      t_info[3].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[3].domain_def.i2Dbox[0].high = nDPt[0];
      t_info[3].domain_def.i2Dbox[1].low = nDPt[1];
      t_info[3].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
      for (i=0; i<NumPoints; i++)
        if (uvstList[i].dCoord[0] < nDPt[0]) {
          if (uvstList[i].dCoord[1] < nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else if (uvstList[i].dCoord[1] > nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 3, &t_info, domain, 0);
        } else if (uvstList[i].dCoord[0] > nDPt[0]) {
          if (uvstList[i].dCoord[1] < nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
          else if (uvstList[i].dCoord[1] > nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 2, &t_info, domain, 0);
        }
/* Now must do curve-surface Xsection with isolines passing thru sing. point
*/
      if (gettimeofday(&tp1, &tzp1)) {
        printf("Error getting time\n");
        exit(-1);
      }
      compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, nDPt[1]);
      if (base_case)
        compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, nDPt[1]);
      find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,U,power_pt(nDPt[1]),
                        fdet1D_u, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      if (gettimeofday(&tp2, &tzp2)) {
        printf("Error getting time\n");
        exit(-1);
      }
      cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                  (tp2.tv_usec - tp1.tv_usec);

      for (i=0; i < num_points; i++)
        for (j=0; j<4; j++)
          addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                            t_info[j].domain_def, 1);

      if (gettimeofday(&tp1, &tzp1)) {
        printf("Error getting time\n");
        exit(-1);
      }
      compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, nDPt[0]);
      if (base_case)
        compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, nDPt[0]);
      find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,V,power_pt(nDPt[0]),
                        fdet1D_v, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      if (gettimeofday(&tp2, &tzp2)) {
        printf("Error getting time\n");
        exit(-1);
      }
      cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                  (tp2.tv_usec - tp1.tv_usec);
      for (i = 0; i < num_points; i++) {
        if (!FP_EQ_EPS(inverse_power_pt(pdPoints[i].dCoord[1]), nDPt[1],
1e-5))
          for (j=0; j<4; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
      }

      s_dom->i2Dbox[0].low = s_dom->i2Dbox[0].high = 0.0;
      s_dom->i2Dbox[1].low = s_dom->i2Dbox[1].high = 0.0;
      for (i=0; i<4; i++)
        if (t_info[i].Npts > 1)
          doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                         t_info[i].domain_def, s_dom, fdet1D_uv,
                         ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                         tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                         fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm,
                         base_case, cComp, compNo, 1);

#ifdef __FREE_ALL
     for (i=0; i<4; i++)
       myfree(t_info[i].ptList);
     myfree(t_info);
     myfree(nDPt);
     FREEMAT(nnDirVec, 2, 2);
     myfree(pdPoints);
#endif
     return;
    }
#ifdef __FREE_ALL
    myfree(nDPt);
    FREEMAT(nnDirVec, 2, 2);
#endif
   } else {
    ul_diff = (domain.i2Dbox[0].low - s_dom->i2Dbox[0].low);
    uh_diff = (domain.i2Dbox[0].high - s_dom->i2Dbox[0].high);
    vl_diff = (domain.i2Dbox[1].low - s_dom->i2Dbox[1].low);
    vh_diff = (domain.i2Dbox[1].high - s_dom->i2Dbox[1].high);
    s_dom_domain_dist = ul_diff*ul_diff + uh_diff*uh_diff +
                        vl_diff*vl_diff + vh_diff*vh_diff;
    if (FP_EQ_EPS(s_dom_domain_dist, 0.0, 0.005)) {
      ALLOCN(t_info, TRACE_INFO, 2); /* 2 is because we create two
                                        new subdomains by
                                        bisection */
      for (i=0; i<2; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;

      if (udiff > vdiff) {
        bisect_pt = (domain.i2Dbox[0].low +
                     domain.i2Dbox[0].high) / 2.0;
        t_info[0].domain_def.i2Dbox[0].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
        t_info[1].domain_def.i2Dbox[0].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[0], bisect_pt)) continue;
          if (uvstList[i].dCoord[0] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        if (gettimeofday(&tp1, &tzp1)) {
          printf("Error getting time\n");
          exit(-1);
        }
        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, bisect_pt);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(bisect_pt),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (gettimeofday(&tp2, &tzp2)) {
          printf("Error getting time\n");
          exit(-1);
        }
        cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      else {
        bisect_pt = (domain.i2Dbox[1].low +
                     domain.i2Dbox[1].high) / 2.0;
        t_info[0].domain_def.i2Dbox[1].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
        t_info[1].domain_def.i2Dbox[1].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[1], bisect_pt)) continue;
          if (uvstList[i].dCoord[1] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        if (gettimeofday(&tp1, &tzp1)) {
          printf("Error getting time\n");
          exit(-1);
        }
        compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, bisect_pt);
        find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,U,power_pt(bisect_pt),
                          fdet1D_u, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (gettimeofday(&tp2, &tzp2)) {
          printf("Error getting time\n");
          exit(-1);
        }
        cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      for (i=0; i<2; i++) {
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
      }

      for (i=0; i<2; i++)
        myfree(t_info[i].ptList);
      myfree(t_info);
    } else (*s_dom) = domain;
   }
  }

  qsort(uvstList, NumPoints, sizeof(POINT4D), comparV);
  numVList = getNumVList(uvstList, NumPoints, &endVVal, &domain);
  if (numVList <= 0) {
    printf("Something Wrong in Domain Decomp.\n");
    return;
  }
  if (numVList > 1) {
    ALLOCN(t_info, TRACE_INFO, numVList);
    for (i=0; i<numVList; i++) {
      t_info[i].Npts = 0;
      ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
    }

    if (gettimeofday(&tp1, &tzp1)) {
      printf("Error getting time\n");
      exit(-1);
    }

    compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, endVVal[0]);
    if (base_case)
      compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, endVVal[0]);
    find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                      pcpxPoints,U,power_pt(endVVal[0]),
                      fdet1D_u, row_perm, m, n, base_case,
                      tpPatch0, tpPatch1, 0);

    if (gettimeofday(&tp2, &tzp2)) {
      printf("Error getting time\n");
      exit(-1);
    }
    cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                (tp2.tv_usec - tp1.tv_usec);

    for (i = 0; i < num_points; i++)
      addPointsToTinfoArray(pdPoints[i], -1, 0, &t_info, domain, 1);
    for (i=0; i<numVList; i++) {
      t_info[i].domain_def.i2Dbox[U].low = domain.i2Dbox[U].low;
      t_info[i].domain_def.i2Dbox[U].high = domain.i2Dbox[U].high;
      t_info[i].domain_def.i2Dbox[V].low = endVVal[i];
      t_info[i].domain_def.i2Dbox[V].high = endVVal[i+1];

      if (gettimeofday(&tp1, &tzp1)) {
        printf("Error getting time\n");
        exit(-1);
      }
      compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, endVVal[i+1]);
      if (base_case)
        compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, endVVal[i+1]);
      find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,U,power_pt(endVVal[i+1]),
                        fdet1D_u, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      if (gettimeofday(&tp2, &tzp2)) {
        printf("Error getting time\n");
        exit(-1);
      }
      cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                  (tp2.tv_usec - tp1.tv_usec);

      if (i == numVList-1) {
        for (j = 0; j < num_points; j++)
          addPointsToTinfoArray(pdPoints[j], -1, i, &t_info, domain, 1);
      } else {
        for (j = 0; j < num_points; j++)
          addPointsToTinfoArray(pdPoints[j], 0, i, &t_info, domain, 1);
      }
      if (t_info[i].Npts >= 2)
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
    }
  } else { /* Case when NumVList = 1 and we are dividing w.r.t U */
    qsort(uvstList, NumPoints, sizeof(POINT4D), comparU);
    numUList = getNumUList(uvstList, NumPoints, &endUVal, &domain);
    if (numUList <= 0) {
      printf("Something Wrong in Domain Decomp.\n");
      return;
    }
    if (numUList > 1) {
      ALLOCN(t_info, TRACE_INFO, numUList);
      for (i=0; i<numUList; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      if (gettimeofday(&tp1, &tzp1)) {
        printf("Error getting time\n");
        exit(-1);
      }
      compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, endUVal[0]);
      if (base_case)
        compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, endUVal[0]);
      find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,V,power_pt(endUVal[0]),
                        fdet1D_v, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      if (gettimeofday(&tp2, &tzp2)) {
        printf("Error getting time\n");
        exit(-1);
      }
      cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                  (tp2.tv_usec - tp1.tv_usec);

      for (i = 0; i < num_points; i++)
        addPointsToTinfoArray(pdPoints[i], -1, 0, &t_info, domain, 1);
      for (i=0; i<numUList; i++) {
        t_info[i].domain_def.i2Dbox[U].low = endUVal[i];
        t_info[i].domain_def.i2Dbox[U].high = endUVal[i+1];
        t_info[i].domain_def.i2Dbox[V].low = domain.i2Dbox[V].low;
        t_info[i].domain_def.i2Dbox[V].high = domain.i2Dbox[V].high;

        if (gettimeofday(&tp1, &tzp1)) {
          printf("Error getting time\n");
          exit(-1);
        }
        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, endUVal[i+1]);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, endUVal[i+1]);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(endUVal[i+1]),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (gettimeofday(&tp2, &tzp2)) {
          printf("Error getting time\n");
          exit(-1);
        }
        cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);

        if (i == numUList-1) {
          for (j = 0; j < num_points; j++)
            addPointsToTinfoArray(pdPoints[j], -1, i, &t_info, domain, 1);
        } else {
          for (j = 0; j < num_points; j++)
            addPointsToTinfoArray(pdPoints[j], 0, i, &t_info, domain, 1);
        }
        if (t_info[i].Npts >= 2)
          doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                         t_info[i].domain_def, s_dom, fdet1D_uv,
                         ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                         tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                         fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm,
                         base_case,cComp, compNo, 1);
      }
    }
    else { /* Case when numUList and numVList are 1 => Cannot divide
              the domain further */
#ifndef DEBUG_PRINT_POLY
      printf("numUList and numVList are 1: So doing bisection\n");
#endif
      ALLOCN(t_info, TRACE_INFO, 2); /* 2 is because we create two
                                        new subdomains by
                                        bisection */
      for (i=0; i<2; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;

      if (udiff > vdiff) {
        bisect_pt = (domain.i2Dbox[0].low +
                     domain.i2Dbox[0].high) / 2.0;
        t_info[0].domain_def.i2Dbox[0].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
        t_info[1].domain_def.i2Dbox[0].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[0], bisect_pt)) continue;
          if (uvstList[i].dCoord[0] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        if (gettimeofday(&tp1, &tzp1)) {
          printf("Error getting time\n");
          exit(-1);
        }
        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, bisect_pt);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(bisect_pt),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (gettimeofday(&tp2, &tzp2)) {
          printf("Error getting time\n");
          exit(-1);
        }
        cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      else {
        bisect_pt = (domain.i2Dbox[1].low +
                     domain.i2Dbox[1].high) / 2.0;
        t_info[0].domain_def.i2Dbox[1].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
        t_info[1].domain_def.i2Dbox[1].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[1], bisect_pt)) continue;
          if (uvstList[i].dCoord[1] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        if (gettimeofday(&tp1, &tzp1)) {
          printf("Error getting time\n");
          exit(-1);
        }
        compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, bisect_pt);
        find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,U,power_pt(bisect_pt),
                          fdet1D_u, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (gettimeofday(&tp2, &tzp2)) {
          printf("Error getting time\n");
          exit(-1);
        }
        cs_time +=  (tp2.tv_sec - tp1.tv_sec)*1e6 +
                    (tp2.tv_usec - tp1.tv_usec);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      for (i=0; i<2; i++) {
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
      }

      for (i=0; i<2; i++)
        myfree(t_info[i].ptList);
      myfree(t_info);
    }
#ifdef __FREE_ALL
  myfree(endUVal);
#endif
  } 
#ifdef __FREE_ALL
  myfree(endVVal);
  myfree(pdPoints);
#endif
}

addPointsToTinfoArray(point4D_val, flag, index, ppArray, domain, inv)
POINT4D point4D_val;
int flag, index, inv;
TRACE_INFO **ppArray;
RECT_BOX domain;
{
  int i;
  double ulow, uhigh, vlow, vhigh;
 
  if ((flag != 0) && (flag != -1)) {
    printf("Incorrect Value of FLAG\n");
    return;
  }

  ulow = domain.i2Dbox[U].low;
  uhigh = domain.i2Dbox[U].high;
  vlow = domain.i2Dbox[V].low;
  vhigh = domain.i2Dbox[V].high;

  if (inv)
    for(i=0; i < 4; i++)
      point4D_val.dCoord[i] = inverse_power_pt(point4D_val.dCoord[i]);

  if ((point4D_val.dCoord[U] >= (ulow-NEW_EPS)) && (point4D_val.dCoord[U] <= (uhigh+NEW_EPS)) &&
      (point4D_val.dCoord[V] >= (vlow-NEW_EPS)) && (point4D_val.dCoord[V] <= (vhigh+NEW_EPS)) &&
      (point4D_val.dCoord[S] >= (-NEW_EPS)) && (point4D_val.dCoord[S] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[T] >= (-NEW_EPS)) && (point4D_val.dCoord[T] <= (1.0+NEW_EPS)))
  {
    (*ppArray)[index].ptList[(*ppArray)[index].Npts] = point4D_val;
    (*ppArray)[index].Npts += 1;
    if ((*ppArray)[index].Npts == 8) 
      printf("OverFlow in TINFO array: Seg Fault Imminent\n");
    if (flag == 0) {
      (*ppArray)[index+1].ptList[(*ppArray)[index+1].Npts] = point4D_val;
      (*ppArray)[index+1].Npts += 1;
      if ((*ppArray)[index+1].Npts == 8) 
        printf("OverFlow in TINFO array: Seg Fault Imminent\n");
    }
  }
}

