void copy_trim_patch_info(tpi1, tpi2)
TRIM_PATCH_INFO tpi1, *tpi2;
{
  int i, j;

  tpi2->tp = tpi1.tp;
  tpi2->orig_tp = tpi1.orig_tp;
  tpi2->NumTrimCurves = tpi1.NumTrimCurves;
  tpi2->NumTrimPortions = tpi1.NumTrimPortions;
  tpi2->csg_status = tpi1.csg_status;
  for (i=0; i<tpi2->NumTrimCurves; i++) {
    tpi2->Curves[i].StEndPoints[0] = tpi1.Curves[i].StEndPoints[0];
    tpi2->Curves[i].StEndPoints[1] = tpi1.Curves[i].StEndPoints[1];
    tpi2->Curves[i].planeCurve = tpi1.Curves[i].planeCurve;
  }

  for (i=0; i<tpi2->NumTrimPortions; i++) {
    tpi2->TrimPortion[i].NumOfVertices = 
				tpi1.TrimPortion[i].NumOfVertices;
    tpi2->TrimPortion[i].ptInsideRegion = 
				tpi1.TrimPortion[i].ptInsideRegion;
    memcpy(tpi2->TrimPortion[i].poly_vertex,
	   tpi1.TrimPortion[i].poly_vertex, 
	   sizeof(tpi2->TrimPortion[i].poly_vertex));
  }
}

POINT2D GiveNextClockCorner(point)
POINT2D point;
{
  double ret_u, ret_v, pu, pv;
  POINT2D rv;

  pu = point.u;
  pv = point.v;

  if (pu < -EPSILON || pu > 1.0+EPSILON || 
      pv < -EPSILON || pv > 1.0+EPSILON ) {
    printf("Incorrect parameter value: POINT\n");
    rv.u = rv.v = -1.0;
    return rv;
  }
  
  if (FP_EQ(pu, 0.0)) {
    ret_u = pu;
    if (FP_EQ(pv, 0.0))
      ret_v = pv;
    else 
      ret_v = 1.0;
  } else if (FP_EQ(pu, 1.0)) {
    ret_u = pu;
    if (FP_EQ(pv, 1.0)) ret_v = pv;
    else ret_v = 0.0;
  } else if (FP_EQ(pv, 0.0)) {
    ret_v = pv;
    if (FP_EQ(pu, 1.0)) ret_u = pu;
    else ret_u = 0.0;
  } else if (FP_EQ(pv, 1.0)) {
    ret_v = pv;
    if (FP_EQ(pu, 0.0)) ret_u = pu;
    else ret_u = 1.0;
  }
      
  rv.u = ret_u;
  rv.v = ret_v;
  return rv;
}

POINT2D GiveNextAntiClockCorner(point)
POINT2D point;
{
  double ret_u, ret_v, pu, pv;
  POINT2D rv;

  pu = point.u;
  pv = point.v;

  if (pu < -EPSILON || pu > 1.0+EPSILON || 
      pv < -EPSILON || pv > 1.0+EPSILON ) {
    printf("Incorrect parameter value: POINT\n");
    rv.u = rv.v = -1.0;
    return rv;
  }
  
  if (FP_EQ(pu, 0.0)) {
    ret_u = pu;
    if (FP_EQ(pv, 1.0))
      ret_v = pv;
    else 
      ret_v = 0.0;
  } else if (FP_EQ(pu, 1.0)) {
    ret_u = pu;
    if (FP_EQ(pv, 0.0)) ret_v = pv;
    else ret_v = 1.0;
  } else if (FP_EQ(pv, 0.0)) {
    ret_v = pv;
    if (FP_EQ(pu, 0.0)) ret_u = pu;
    else ret_u = 1.0;
  } else if (FP_EQ(pv, 1.0)) {
    ret_v = pv;
    if (FP_EQ(pu, 1.0)) ret_u = pu;
    else ret_u = 0.0;
  }
      
  rv.u = ret_u;
  rv.v = ret_v;
  return rv;
}

int CompleteTheLoop(t_curve, vert_array)
DOMAIN_INT_CURVE t_curve;
double vert_array[][2];
{
  double st_u, st_v, end_u, end_v;
  int i, j, vert_cnt;
  LIST_POINT2D *tmp_ptr;
  POINT2D last_corner, tmp_corner;

  vert_cnt = 0;
  st_u = t_curve.StEndPoints[0].u;
  st_v = t_curve.StEndPoints[0].v;
  end_u = t_curve.StEndPoints[1].u;
  end_v = t_curve.StEndPoints[1].v;
  
  tmp_ptr = t_curve.planeCurve;
  while (tmp_ptr != NULL) {
    if (vert_cnt == MAX_VERTEX_COUNT) {
      printf("Too many points in the polygon\n");
      return -1;
    }
    vert_array[vert_cnt][0] = tmp_ptr->p2D.u;
    vert_array[vert_cnt++][1] = tmp_ptr->p2D.v;
    tmp_ptr = tmp_ptr->plp2DNext;
  }

  if (FP_EQ(st_u, end_u) || FP_EQ(st_v, end_v)) {
    if (vert_cnt == MAX_VERTEX_COUNT) {
      printf("Too many points in the polygon\n");
      return -1;
    }
    vert_array[vert_cnt][0] = st_u;
    vert_array[vert_cnt][1] = st_v;
    return vert_cnt;
  }

  last_corner = GiveNextClockCorner(t_curve.StEndPoints[0]);
  tmp_corner = GiveNextAntiClockCorner(t_curve.StEndPoints[1]);
  while (!EQUAL2D(tmp_corner, last_corner)) {
    if (vert_cnt == MAX_VERTEX_COUNT) {
      printf("Too many points in the polygon\n");
      return -1;
    }
    vert_array[vert_cnt][0] = tmp_corner.u;
    vert_array[vert_cnt++][1] = tmp_corner.v;
    tmp_corner = GiveNextAntiClockCorner(tmp_corner);
  }
  if (vert_cnt == MAX_VERTEX_COUNT) {
    printf("Too many points in the polygon\n");
    return -1;
  }
  vert_array[vert_cnt][0] = last_corner.u;
  vert_array[vert_cnt][1] = last_corner.v;
  return vert_cnt;
}

