#include "defs.h"
#include "complex.h"
#include "lp.h"
#ifdef __GL
#include <gl.h>
#endif

/* This routine is used mostly when arr is patchNumber and brr is */
/* first_or_second array. This returns the original patch number in */
/* the solid. Used to get final solid's graph */
int ReturnIndexInArray(arr, num, value, brr, val)
int *arr, *brr, num, value, val;
{
  int i;

  for (i=0; i<num; i++)
    if ((arr[i] == value) && (brr[i] == val))
      return i;
  return -1;
}

void myfree(ptr)
void *ptr;
{
  if (ptr == NULL) return;
  free(ptr);
}

void EvalPointOnPatch(tp, s, t, xv, yv, zv)
TENSOR_PATCH    tp;
double          s, t;
double          *xv, *yv, *zv;
{
  int           m, n;
  int           i, j;
  int		*ppp;
  int           polynomial;
  double        *ber1, *ber2;
  double        x_st, y_st, z_st, w_st;

  m = tp.iDeg0;
  n = tp.iDeg1;

  ALLOCN(ber1, double, m+1);
  ALLOCN(ber2, double, n+1);

  eval_ber(ber1, s, m);
  eval_ber(ber2, t, n);

  x_st = y_st = z_st = w_st = 0.0;
  polynomial = 1;

  for (i=0; i<=m; i++) {
    for (j=0; j<=n; j++) {
      if ((tp.coeffs[i][j].dCoord[W] > 0.999) && 
	  (tp.coeffs[i][j].dCoord[W] < 1.001)) {
	polynomial = 0;
	break;
      }
    }
    if (polynomial == 0) break;
  }

  for (i=0; i<=m; i++)
  for (j=0; j<=n; j++)
  {
    x_st += tp.coeffs[i][j].dCoord[X] * ber1[i] * ber2[j] ;
    y_st += tp.coeffs[i][j].dCoord[Y] * ber1[i] * ber2[j] ;
    z_st += tp.coeffs[i][j].dCoord[Z] * ber1[i] * ber2[j] ;
    w_st += tp.coeffs[i][j].dCoord[W] * ber1[i] * ber2[j] ;
  }

  if ( FP_EQ_EPS(w_st, 0.0, 1e-6) && !polynomial)
  {
    printf("W_ST IS ZERO\n");
    exit(-1);
  }

   /* w_st /= Scale; */
  if (polynomial) {
    *xv = x_st;
    *yv = y_st;
    *zv = z_st;
  } else {
    *xv = x_st / w_st ;
    *yv = y_st / w_st ;
    *zv = z_st / w_st ;
  }

#ifdef __FREE_ALL
  myfree(ber1);
  myfree(ber2);
#endif
}

int compar( v1, v2 )
int *v1, *v2;
{
  return ( (*v1) - (*v2) );
}

int comparU(p1, p2)
POINT4D *p1, *p2;
{
  if (p1->dCoord[U] < p2->dCoord[U]) return -1;
  else if (p1->dCoord[U] > p2->dCoord[U]) return 1;
  return 0;
}

int comparV(p1, p2)
POINT4D *p1, *p2;
{
  if (p1->dCoord[V] < p2->dCoord[V]) return -1;
  else if (p1->dCoord[V] > p2->dCoord[V]) return 1;
  return 0;
}

void Const2DFrom4D(ret_curve, orig_curve, m, n)
LIST_POINT2D **ret_curve;
LIST_POINT4D *orig_curve;
int m, n;
{
  LIST_POINT2D *tmp1, *tmp2;
  LIST_POINT4D *tmp3;

  *ret_curve = NULL;
  if (orig_curve == NULL) return;
  tmp3 = orig_curve;
  ALLOCN(tmp2, LIST_POINT2D, 1);
  tmp2->p2D.u = tmp3->p4D.dCoord[m];
  tmp2->p2D.v = tmp3->p4D.dCoord[n];
  tmp2->plp2DNext = NULL;
  *ret_curve = tmp1 = tmp2;
  tmp3 = tmp3->plp4DNext;

  for (;tmp3;) {
    ALLOCN(tmp2, LIST_POINT2D, 1);
    tmp2->p2D.u = tmp3->p4D.dCoord[m];
    tmp2->p2D.v = tmp3->p4D.dCoord[n];
    tmp2->plp2DNext = NULL;
    tmp1->plp2DNext = tmp2;
    tmp1 = tmp2;
    tmp3 = tmp3->plp4DNext;
  }
}
  
void Const2DFrom4DReverse(ret_curve, orig_curve, m, n)
LIST_POINT2D **ret_curve;
LIST_POINT4D *orig_curve;
int m, n;
{
  LIST_POINT2D *tmp2;
  LIST_POINT4D *tmp3;

  *ret_curve = NULL;
  if (orig_curve == NULL) return;
  tmp3 = orig_curve;
  for (;tmp3;) {
    ALLOCN(tmp2, LIST_POINT2D, 1);
    tmp2->p2D.u = tmp3->p4D.dCoord[m];
    tmp2->p2D.v = tmp3->p4D.dCoord[n];
    tmp2->plp2DNext = *ret_curve;
    *ret_curve = tmp2;
    tmp3 = tmp3->plp4DNext;
  }
}
  
void Reverse2DList(ret_curve)
DOMAIN_INT_CURVE *ret_curve;
{
  LIST_POINT2D *tmp, *tmp2;
  POINT2D pt2D;

  pt2D = ret_curve->StEndPoints[0];
  ret_curve->StEndPoints[0] = ret_curve->StEndPoints[1];
  ret_curve->StEndPoints[1] = pt2D;
  tmp = ret_curve->planeCurve;

  ret_curve->planeCurve = NULL;
  for (;tmp;) {
    tmp2 = tmp;
    tmp = tmp->plp2DNext;
    tmp2->plp2DNext = ret_curve->planeCurve;
    ret_curve->planeCurve = tmp2;
  }
}
  
double dist2D(pt1, pt2)
POINT2D *pt1, *pt2;
{
  double ret, uv, vv;

  uv = pt1->u - pt2->u;
  vv = pt1->v - pt2->v;
  ret = uv * uv + vv * vv;

  return ret;
}

int closer_pt(pt, pt1, pt2, dist)
POINT2D *pt, *pt1, *pt2;
double *dist;
{
  double dist0, dist1;

  dist0 = dist2D(pt, pt1);
  dist1 = dist2D(pt, pt2);
  if (dist0 < dist1) { *dist = dist0; return 0; }
  else { *dist = dist1; return 1; }
}

Merge2DLists(dom_curve, num_curv, comp, flag)
DOMAIN_INT_CURVE **dom_curve;
int *num_curv, flag;
INT_CURVE_STRUCT *comp;
{
  int p, q;
  int i, j, k;
  LIST_POINT2D *tmp_curve;
  POINT2D comp_ep0, comp_ep1;
  double dist0, dist1;
  int close0, close1;
  int degenerate_case; /* = 1 when the two endpts of comp are
			  within 2e-5 dist. apart */

  if (flag != 1 && flag != 2) {
    printf("Merge2DLists: Error in flag input\n");
    return;
  }
  if (flag == 1) {
    p = 2;
    q = 3;
  } else {
    p = 0;
    q = 1;
  }

  comp_ep0.u = comp->endPoints[0].dCoord[p];
  comp_ep0.v = comp->endPoints[0].dCoord[q];
  comp_ep1.u = comp->endPoints[1].dCoord[p];
  comp_ep1.v = comp->endPoints[1].dCoord[q];
  if (EQUAL2D_EPS(comp_ep0, comp_ep1, 1e-8))
    degenerate_case = 1;
  else degenerate_case = 0;

  if (degenerate_case) 
    printf("Merge2DLists: AKJHDKUDKJGF\n");

  for (i=0; i<(*num_curv); i++) {
    close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
		   &comp_ep0, &comp_ep1, &dist0);

    close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
		   &comp_ep0, &comp_ep1, &dist1);

    if (EQ_2D(comp->endPoints[0], p, q, 
	      (*dom_curve)[i].StEndPoints[0]) && close0 == 0
	&& dist0 <= dist1) { 
      tmp_curve = NULL;
      Const2DFrom4D(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&tmp_curve, (*dom_curve)[i].planeCurve);
      (*dom_curve)[i].planeCurve = tmp_curve;
      (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[1].dCoord[p];
      (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[1].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[0]) && close0
	    == 0 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[1]) && close0
	    == 1 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[0], p, q, 
	      (*dom_curve)[i].StEndPoints[1]) && close1 == 0
	       && dist1 <= dist0) {
      tmp_curve = NULL;
      Const2DFrom4DReverse(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&((*dom_curve)[i].planeCurve), tmp_curve);
      (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[1].dCoord[p];
      (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[1].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[0]) && close1
	    == 0 && dist1 <= dist0) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[1]) && close1
	    == 1 && dist1 <= dist0) { 
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[1], p, q, 
	      (*dom_curve)[i].StEndPoints[0]) && close0 == 1
	       && dist0 <= dist1) {  
      tmp_curve = NULL;
      Const2DFrom4DReverse(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&tmp_curve, (*dom_curve)[i].planeCurve);
      (*dom_curve)[i].planeCurve = tmp_curve;
      (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[0].dCoord[p];
      (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[0].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[0]) && close0 == 0&& dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[1]) && close0
	    == 1 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[1], p, q, 
	       (*dom_curve)[i].StEndPoints[1]) && close1 == 1
	       && dist1 <= dist0) {
      tmp_curve = NULL;
      Const2DFrom4D(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&((*dom_curve)[i].planeCurve), tmp_curve);
      (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[0].dCoord[p];
      (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[0].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[0]) && close1
	    == 0 && dist1 <= dist0) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[1]) && close1
	    == 1 && dist1 <= dist0) { 
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    }
  }

  if (i == (*num_curv)) {
    Const2DFrom4DReverse(&((*dom_curve)[i].planeCurve),comp->Curve,p,q);
    (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[0].dCoord[p];
    (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[0].dCoord[q];
    (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[1].dCoord[p];
    (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[1].dCoord[q];
    (*num_curv)++;
  } 
}

int getNumofPoints(pList)
LIST_POINT4D *pList;
{
  int i;
  LIST_POINT4D *tmp_ptr;

  tmp_ptr = pList;
  i = 0;
  if (tmp_ptr == NULL) return 0;
  if (tmp_ptr->plp4DNext == NULL) return 1;

  while (tmp_ptr != NULL) {
    i++;
    tmp_ptr = tmp_ptr->plp4DNext;
  }
  return i;
}

#ifdef __MYCODE
int dump_one_patch(outfile, trim, patchnum)
FILE *outfile;
TRIM_PATCH_INFO *trim;
int patchnum;
{
  register int i, j, k;
  int l;
 
  k = patchnum;
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);

      for (i = 0; i <= trim[k].tp.iDeg0; i++)
        for (j = 0; j <= trim[k].tp.iDeg1; j++)
          {
            for (l = 0; l < 4; l++)
              fprintf(outfile, "%lf ",
                      trim[k].tp.coeffs[i][j].dCoord[l]);
            fprintf(outfile, "\n");
          }

  fflush(outfile);
  return 0;
}

int dump_trim_patches(outfile, trim, npatches)
FILE *outfile;
TRIM_PATCH_INFO *trim;
int npatches;
{
  register int i, j, k;
  int l;
 
  fprintf(outfile, "\n\n Printing trim_patch_info dump\n");
  for (k = 0; k < npatches; k++)
    {
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);

      for (i = 0; i <= trim[k].tp.iDeg0; i++)
        for (j = 0; j <= trim[k].tp.iDeg1; j++)
          {
            for (l = 0; l < 4; l++)
              fprintf(outfile, "%lf ",
                      trim[k].tp.coeffs[i][j].dCoord[l]);
            fprintf(outfile, "\n");
          }
    }

  fflush(outfile);
  return 0;
}
#endif

int FindNumTrimSegments(ptr)
LIST_POINT2D *ptr;
{
  int num=0;
  LIST_POINT2D *tmp;
  POINT2D tmp2D;

  if (ptr == NULL) return 0;
  num = 0;
  tmp2D = ptr->p2D;
  if (!num && !FP_EQ_EPS(tmp2D.u,0.0,1e-4) && 
      !FP_EQ_EPS(tmp2D.u,1.0,1e-4) &&
      !FP_EQ_EPS(tmp2D.v,0.0,1e-4) && 
      !FP_EQ_EPS(tmp2D.v,1.0,1e-4)) num++;
  tmp = ptr->plp2DNext;
  if (tmp == NULL) return num;
  while (tmp != NULL) {
    tmp2D = tmp->p2D;
    num++;
    tmp = tmp->plp2DNext;
  }
  if (!FP_EQ_EPS(tmp2D.u, 0.0,1e-4) && !FP_EQ_EPS(tmp2D.u, 1.0,1e-4) &&
      !FP_EQ_EPS(tmp2D.v, 0.0,1e-4) && !FP_EQ_EPS(tmp2D.v, 1.0,1e-4))
    num++; 
  return num;
}

int FindNumElmtsIn(curve, curveArray)
LIST_POINT2D *curve;
double curveArray[][2];
{
  LIST_POINT2D *ptr;
  POINT2D elmt1, elmt2;
  int count;

  count = 0;
  ptr = curve;

  if (ptr == NULL) {
    printf("Warning: No element in the list\n");
    return count;
  }
  
  elmt1 = ptr->p2D;
  curveArray[count][0] = elmt1.u;
  curveArray[count++][1] = elmt1.v;

  if (ptr->plp2DNext == NULL) {
    printf("Warning: Only one element in a loop\n");
    return count;
  }

  ptr = ptr->plp2DNext;

  while (ptr->plp2DNext != NULL) {
    curveArray[count][0] = ptr->p2D.u;
    curveArray[count++][1] = ptr->p2D.v;

    ptr = ptr->plp2DNext;
  }

  if (ptr->plp2DNext != NULL) {
    printf("Something wrong with WHILE loop\n");
    exit(-1);
  }
  elmt2 = ptr->p2D;

  if (!EQUAL2D_EPS(elmt1, elmt2, 1e-6)) {
    printf("First and Last element of loop not the same\n");
  }

  curveArray[count][0] = ptr->p2D.u;
  curveArray[count][1] = ptr->p2D.v;

  return count;
}

