#include "defs.h"
#include <math.h>
#include "basic.h"
#ifdef BEZ_CLIP
#include "csurf.h"
#endif

extern void bernsteinToPowerBasis();
extern void allocDeterminant();
extern void implicitize();
extern void powerToBernsteinBasis();
extern void moveFactoredToFinalDeterminant();
extern int check_for_base_points();
extern void allocDeterminant_1D();
extern void change_to_uv();
extern int find_start_points();
extern double inverse_power_pt();
extern int is_point_inside_polygon();
extern point_typ GetPtInsideSplinegon();

extern int NumOvrPatches;
extern OVR_BOXES overlap_pairs[];

extern short first_complete_ovrlap[], second_complete_ovrlap[];
extern double Scale;

EXTERN_LINALG

static int markArray[MAX_NUM_PORTIONS];

int IsPatchFacingOutside(solid, numPatches, patchNo)
TRIM_PATCH_INFO *solid;
int numPatches, patchNo;
{
  TENSOR_PATCH u_tp, v_tp;
  int m, n;
  int i, j, k, l;
  double uval, vval;
  double uxval, uyval, uzval;
  double vxval, vyval, vzval;
  double nxval, nyval, nzval;
  double xval, yval, zval;
  double nmod;

  m = solid[patchNo].tp.iDeg0;
  n = solid[patchNo].tp.iDeg1;  

  u_tp.iDeg0 = m-1;
  u_tp.iDeg1 = n;
  v_tp.iDeg0 = m;
  v_tp.iDeg1 = n-1;
  allocate_tensor_patch(&u_tp);
  allocate_tensor_patch(&v_tp);

  for (i=0; i<=u_tp.iDeg0; i++)
    for (j=0; j<=u_tp.iDeg1; j++) {
      for (k=0; k<4; k++)
	u_tp.coeffs[i][j].dCoord[k] = m*
	  (solid[patchNo].tp.coeffs[i+1][j].dCoord[k] - 
	   solid[patchNo].tp.coeffs[i][j].dCoord[k]);
    }
  for (i=0; i<=v_tp.iDeg0; i++)
    for (j=0; j<=v_tp.iDeg1; j++) {
      for (k=0; k<4; k++)
	v_tp.coeffs[i][j].dCoord[k] = n*
	  (solid[patchNo].tp.coeffs[i][j+1].dCoord[k] - 
	   solid[patchNo].tp.coeffs[i][j].dCoord[k]);
    }

  if (solid[patchNo].NumTrimPortions != 1) {
    printf("IsPatchFacingOutside: Too many trim portions\n");
    exit(-1);
  }

  if (solid[patchNo].SplineReprn) {
    uval = solid[patchNo].TrimPortion[0].SplinePortion.ptInsideRegion.u;
    vval = solid[patchNo].TrimPortion[0].SplinePortion.ptInsideRegion.v;
  } else {
    uval = solid[patchNo].TrimPortion[0].PiecePortion.ptInsideRegion.u;
    vval = solid[patchNo].TrimPortion[0].PiecePortion.ptInsideRegion.v;
  }
  EvalPointOnPatch(u_tp, uval, vval, &uxval, &uyval, &uzval);
  EvalPointOnPatch(v_tp, uval, vval, &vxval, &vyval, &vzval);
  EvalPointOnPatch(solid[patchNo].tp, uval, vval, &xval, &yval, &zval);

  nxval = uyval*vzval - uzval*vyval;
  nyval = uzval*vxval - uxval*vzval;
  nzval = uxval*vyval - uyval*vxval;
  nmod = sqrt(nxval*nxval + nyval*nyval + nzval*nzval);

  if (FP_EQ_EPS(nmod, 0.0, 1e-4)) {
    printf("IsPatchFacingOutside: nmod too small\n");
    exit(-1);
  }

  nxval /= nmod;
  nyval /= nmod;
  nzval /= nmod;

  xval += nxval * 0.01;
  yval += nyval * 0.01;
  zval += nzval * 0.01;
  if (RobustInOutTest(xval, yval, zval, solid, numPatches) == 0)
    return 1;
  else return 0;
}

#ifdef BEZ_CLIP
int FindNumCurveSurfInt(xv, yv, zv, xv2, yv2, zv2, solid, numPatches)
double xv, yv, zv;
double xv2, yv2, zv2;
TRIM_PATCH_INFO *solid;
int numPatches;
{
  int              i, j, k, numXsections, is_inside;
  int              p, q, r, s, num_points;
  XYZ_SPLINE_TYPE  spl;
  TENSOR_PATCH     tp0;
  double           vval, sval, tval;
  double           xval1, yval1, zval1;
  point_typ	   pt;
  int		   NumShadyPoints; /* maintains count of boundary
				      intersections */
  POINT4D	   ShadyPtList[MAX_SHADY_POINTS];
  double           vhits[20], shits[20], thits[20];
  int		   ind, shady_and_ignore;

  spl.Deg = 1;
  allocate_xyz_spline(&spl);
  spl.CntrlPts[0].dCoord[0] = xv;
  spl.CntrlPts[0].dCoord[1] = yv;
  spl.CntrlPts[0].dCoord[2] = zv;
  spl.CntrlPts[0].dCoord[3] = 1.0;
  spl.CntrlPts[1].dCoord[0] = xv+xv2;
  spl.CntrlPts[1].dCoord[1] = yv+yv2;
  spl.CntrlPts[1].dCoord[2] = zv+zv2;
  spl.CntrlPts[1].dCoord[3] = 1.0;

  numXsections = 0;
  NumShadyPoints = 0;
  shady_and_ignore = 0;
  for (k=0; k<numPatches; k++) {
    copy_ten_p(&solid[k].tp, &tp0);

    curve_surface_intersect(&tp0, &spl, CLIP_TOL, CLIP_TOL, &num_points,
			    vhits, shits, thits);
    for (i=0; i<num_points; i++) {
      vval = vhits[i];
      sval = shits[i];
      tval = thits[i];
      EvalPointOnPatch(tp0, sval, tval, &xval1, &yval1, &zval1);
      if (sval >= -EPSILON && sval <= 1.0+EPSILON &&
          tval >= -EPSILON && tval <= 1.0+EPSILON && vval >= -EPSILON) {
	if (FP_EQ(vval,0.0) || FP_EQ(vval,1.0)) 
	  printf("Warning: vval ( %lf ) is very close to boundary\n");
	for (ind=0; ind<NumShadyPoints; ind++) 
	  if (FP_EQ_EPS(xval1,ShadyPtList[ind].dCoord[0],1e-3) &&
	      FP_EQ_EPS(yval1,ShadyPtList[ind].dCoord[1],1e-3) &&
	      FP_EQ_EPS(zval1,ShadyPtList[ind].dCoord[2],1e-3)) {
	    shady_and_ignore = 1;
	    break;
	  } 

	if (shady_and_ignore)
	  shady_and_ignore = 0;
	else {
	  is_inside = 0;
	  for (j=0; j<solid[k].NumTrimPortions; j++) {
	    pt.x = sval;
	    pt.y = tval;
	    if (solid[k].SplineReprn)
	      is_inside = IsPtInsideSplinegon(
				  &solid[k].TrimPortion[j].SplinePortion,
				  &pt);
	    else
	      is_inside = is_point_inside_polygon(
		               solid[k].TrimPortion[j].PiecePortion.NumOfVertices,
			       solid[k].TrimPortion[j].PiecePortion.poly_vertex,
			       &pt);
	    if (is_inside) {
	      ShadyPtList[ind].dCoord[0] = xval1;
	      ShadyPtList[ind].dCoord[1] = yval1;
	      ShadyPtList[ind].dCoord[2] = zval1;
	      NumShadyPoints++;
	      break;
	    }
	  }
          if (is_inside) numXsections++;
	}
      }
    }
    free_ten_p(&tp0);
  }
  return numXsections;
}
#else
#ifdef _CALC_IMPLICIT
int FindNumCurveSurfInt(xv, yv, zv, xv2, yv2, zv2, solid, numPatches,
			ndetImplicit, imp_scale)
double xv, yv, zv;
double xv2, yv2, zv2;
TRIM_PATCH_INFO *solid;
int numPatches;
NDETERMINANT *ndetImplicit;
double imp_scale;
{
  NDETERMINANT     fdetImplicit;
  NDETERMINANT_1D  fdet1D_Tmp, ndet1D_Tmp;
  int              Nrows, Ncols;
  int              *row_perm, *col_perm;
  int              base_case, rank_row_val;
  int              num_points, num_cpoints;
  POINT4D          *pdPoints;
  POINT4CPX        *pcpxPoints = NULL;
  int              i, j, k, numXsections, is_inside;
  int              p, q, r, s;
  TENSOR_PATCH     tpPatch0, tpPatch1, tp1;
  double           vval, sval, tval;
  double           xval1, yval1, zval1;
  double           xval2, yval2, zval2;
  double           maxscale = 0;
  double           mxsc = 0;
  point_typ	   pt;
  int		   NumShadyPoints; /* maintains count of boundary
				      intersections */
  POINT4D	   ShadyPtList[MAX_SHADY_POINTS];
  int		   ind, shady_and_ignore;

  Scale = 1.0;
  tp1.iDeg0 = tpPatch1.iDeg0 = 0;
  tp1.iDeg1 = tpPatch1.iDeg1 = 1;
  allocate_tensor_patch(&tpPatch1);
  allocate_tensor_patch(&tp1);
  tp1.coeffs[0][0].dCoord[X] = xv*Scale;
  tp1.coeffs[0][0].dCoord[Y] = yv*Scale;
  tp1.coeffs[0][0].dCoord[Z] = zv*Scale;
  tp1.coeffs[0][0].dCoord[W] = 1.0;
  tp1.coeffs[0][1].dCoord[X] = (xv+xv2)*Scale;
  tp1.coeffs[0][1].dCoord[Y] = (yv+yv2)*Scale;
  tp1.coeffs[0][1].dCoord[Z] = (zv+zv2)*Scale;
  tp1.coeffs[0][1].dCoord[W] = 1.0;
#ifdef _SCALE
  for (p=0; p<=tp1.iDeg0; p++)
  for (q=0; q<=tp1.iDeg1; q++)
    for (r=0; r<=Z; r++) {
      if (maxscale < fabs(tp1.coeffs[p][q].dCoord[r]))
        maxscale = fabs(tp1.coeffs[p][q].dCoord[r]);
    }
#endif

  numXsections = 0;
  NumShadyPoints = 0;
  shady_and_ignore = 0;
  for (k=0; k<numPatches; k++) {
    for (p=0; p<=tpPatch1.iDeg0; p++)
      for (q=0; q<=tpPatch1.iDeg1; q++) {
	for (r=0; r<=Z; r++)
	  tpPatch1.coeffs[p][q].dCoord[r] = tp1.coeffs[p][q].dCoord[r] / imp_scale;
	tpPatch1.coeffs[p][q].dCoord[W] = tp1.coeffs[p][q].dCoord[W];
      }

#ifdef _SCALE
    mxsc = maxscale;
#else
    mxsc = 1.0;
#endif
    tpPatch0.iDeg0 = solid[k].tp.iDeg0;
    tpPatch0.iDeg1 = solid[k].tp.iDeg1;
    allocate_tensor_patch(&tpPatch0);
    for (p=0; p<=tpPatch0.iDeg0; p++)
    for (q=0; q<=tpPatch0.iDeg1; q++) {
      for (r=0; r<=Z; r++) {
        tpPatch0.coeffs[p][q].dCoord[r] =
                        solid[k].tp.coeffs[p][q].dCoord[r];
#ifdef _SCALE
        if (mxsc < fabs(tpPatch0.coeffs[p][q].dCoord[r]))
          mxsc = fabs(tpPatch0.coeffs[p][q].dCoord[r]);
#else
	mxsc = 1.0;
#endif
      }
      tpPatch0.coeffs[p][q].dCoord[W] = solid[k].tp.coeffs[p][q].dCoord[W];
    }

    Scale = imp_scale;
    scale_tensor_patch(&tpPatch0, 1);
    Scale = 1.0;

#ifdef _SCALE
    for (p=0; p<=tpPatch0.iDeg0; p++)
    for (q=0; q<=tpPatch0.iDeg1; q++)
      for (r=0; r<=Z; r++)
        tpPatch0.coeffs[p][q].dCoord[r] /= mxsc;

    for (p=0; p<=tpPatch1.iDeg0; p++)
    for (q=0; q<=tpPatch1.iDeg1; q++)
      for (r=0; r<=Z; r++)
        tpPatch1.coeffs[p][q].dCoord[r] /= mxsc;
#endif

    Nrows = ndetImplicit[k].cRows;
    Ncols = ndetImplicit[k].cCols;
    base_case = check_for_base_points(ndetImplicit[k], Nrows, Ncols,
                                      &rank_row_val, &row_perm, &col_perm);
    if (base_case) {
      fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
      ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
      for (i=0; i<fdetImplicit.cCols; i++)
        ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

      for (i=0; i<rank_row_val; i++)
      {
        for (j=0; j<rank_row_val; j++)
        {
          fdetImplicit.coeffs[i][j].dCoord[X] = ndetImplicit[k].coeffs[row_perm[i]][col_perm[j]].dCoord[X];
          fdetImplicit.coeffs[i][j].dCoord[Y] = ndetImplicit[k].coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
          fdetImplicit.coeffs[i][j].dCoord[Z] = ndetImplicit[k].coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
          fdetImplicit.coeffs[i][j].dCoord[W] = ndetImplicit[k].coeffs[row_perm[i]][col_perm[j]].dCoord[W];
        }
      }
    }

    if (base_case) {
      allocDeterminant_1D( &fdet1D_Tmp, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
         2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);
      allocDeterminant_1D(&ndet1D_Tmp, rank_row_val, rank_row_val, 
			  tpPatch1.iDeg0, tpPatch1.iDeg1);
      change_to_uv( ndetImplicit[k], fdet1D_Tmp, tpPatch1 );
      change_to_uv( fdetImplicit, ndet1D_Tmp, tpPatch1 );
    }
    else {
      allocDeterminant_1D( &ndet1D_Tmp, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
         2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);
      change_to_uv( ndetImplicit[k], ndet1D_Tmp, tpPatch1 );
    }
    ALLOCN(pdPoints, POINT4D, ndet1D_Tmp.cRows);
    if (find_start_points(ndet1D_Tmp, &num_points,pdPoints,&num_cpoints,
                          pcpxPoints, V, power_pt(0.0),
                          fdet1D_Tmp, row_perm, tpPatch0.iDeg0, tpPatch0.iDeg1,
                          base_case, tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i=0; i<num_points; i++) {
      vval = inverse_power_pt(pdPoints[i].dCoord[V]);
      sval = inverse_power_pt(pdPoints[i].dCoord[S]);
      tval = inverse_power_pt(pdPoints[i].dCoord[T]);
      EvalPointOnPatch(tpPatch0, sval, tval, &xval1, &yval1, &zval1);
      EvalPointOnPatch(tpPatch1, 0.0, vval, &xval2, &yval2, &zval2);
/*
      if (!(FP_EQ_EPS(xval1, xval2, 1e-3) &&
	    FP_EQ_EPS(yval1, yval2, 1e-3) &&
	    FP_EQ_EPS(zval1, zval2, 1e-3))) {
	printf("The intersection points ARE (%lf,%lf,%lf) and (%lf,%lf,%lf)\n", xval1,yval1,zval1,xval2,yval2,zval2);
	exit(-1); 
      }
*/
      if (sval >= -EPSILON && sval <= 1.0+EPSILON &&
          tval >= -EPSILON && tval <= 1.0+EPSILON && vval >= -EPSILON) {
	if (FP_EQ(vval,0.0) || FP_EQ(vval,1.0)) 
	  printf("Warning: vval ( %lf ) is very close to boundary\n");
/*
	if (FP_EQ(sval,0.0) || FP_EQ(sval,1.0) || 
	    FP_EQ(tval,0.0) || FP_EQ(tval,1.0)) {
*/
	for (ind=0; ind<NumShadyPoints; ind++) 
	  if (FP_EQ_EPS(xval1,ShadyPtList[ind].dCoord[0],1e-3) &&
	      FP_EQ_EPS(yval1,ShadyPtList[ind].dCoord[1],1e-3) &&
	      FP_EQ_EPS(zval1,ShadyPtList[ind].dCoord[2],1e-3)) {
	    shady_and_ignore = 1;
	    break;
	  } 

/*
	}
*/
	if (shady_and_ignore)
	  shady_and_ignore = 0;
	else {
	  is_inside = 0;
	  for (j=0; j<solid[k].NumTrimPortions; j++) {
	    pt.x = sval;
	    pt.y = tval;
	    if (solid[k].SplineReprn)
	      is_inside = IsPtInsideSplinegon(
				  &solid[k].TrimPortion[j].SplinePortion,
				  &pt);
	    else
	      is_inside = is_point_inside_polygon(
		               solid[k].TrimPortion[j].PiecePortion.NumOfVertices,
			       solid[k].TrimPortion[j].PiecePortion.poly_vertex,
			       &pt);
	    if (is_inside) {
	      ShadyPtList[ind].dCoord[0] = xval1;
	      ShadyPtList[ind].dCoord[1] = yval1;
	      ShadyPtList[ind].dCoord[2] = zval1;
	      NumShadyPoints++;
	      break;
	    }
	  }
          if (is_inside) numXsections++;
	}
      }
    }
    if (base_case) {
      for (i=0; i<fdetImplicit.cCols; i++) {
        myfree(fdetImplicit.coeffs[i]);
	fdetImplicit.coeffs[i] = NULL;
      }
      myfree(fdetImplicit.coeffs);
      fdetImplicit.coeffs = NULL;
      
      free_alloc_det_1D(&fdet1D_Tmp);
    }
    free_alloc_det_1D(&ndet1D_Tmp);
    free_ten_p(&tpPatch0);
  }
  free_ten_p(&tpPatch1);
  free_ten_p(&tp1);
  
  return numXsections;
}
#else
int FindNumCurveSurfInt(xv, yv, zv, xv2, yv2, zv2, solid, numPatches)
double xv, yv, zv;
double xv2, yv2, zv2;
TRIM_PATCH_INFO *solid;
int numPatches;
{
  NDETERMINANT     ndetImplicit, fdetImplicit, ndet;
  NDETERMINANT_1D  fdet1D_Tmp, ndet1D_Tmp;
  int              Nrows, Ncols;
  int              *row_perm, *col_perm;
  int              base_case, rank_row_val;
  int              num_points, num_cpoints;
  POINT4D          *pdPoints;
  POINT4CPX        *pcpxPoints;
  int              i, j, k, numXsections, is_inside;
  int              p, q, r, s;
  TENSOR_PATCH     tpPatch0, tpPatch1, tp1;
  double           vval, sval, tval;
  double           xval1, yval1, zval1;
  double           xval2, yval2, zval2;
  double           maxscale = 0;
  double           mxsc = 0;
  point_typ	   pt;
  int		   NumShadyPoints; /* maintains count of boundary
				      intersections */
  POINT4D	   ShadyPtList[MAX_SHADY_POINTS];
  int		   ind, shady_and_ignore;

  tp1.iDeg0 = tpPatch1.iDeg0 = 0;
  tp1.iDeg1 = tpPatch1.iDeg1 = 1;
  allocate_tensor_patch(&tpPatch1);
  allocate_tensor_patch(&tp1);
  tp1.coeffs[0][0].dCoord[X] = xv*Scale;
  tp1.coeffs[0][0].dCoord[Y] = yv*Scale;
  tp1.coeffs[0][0].dCoord[Z] = zv*Scale;
  tp1.coeffs[0][0].dCoord[W] = 1.0;
  tp1.coeffs[0][1].dCoord[X] = (xv+xv2)*Scale;
  tp1.coeffs[0][1].dCoord[Y] = (yv+yv2)*Scale;
  tp1.coeffs[0][1].dCoord[Z] = (zv+zv2)*Scale;
  tp1.coeffs[0][1].dCoord[W] = 1.0;
#ifdef _SCALE
  for (p=0; p<=tp1.iDeg0; p++)
  for (q=0; q<=tp1.iDeg1; q++)
    for (r=0; r<=Z; r++) {
      if (maxscale < fabs(tp1.coeffs[p][q].dCoord[r]))
        maxscale = fabs(tp1.coeffs[p][q].dCoord[r]);
    }
#endif

  numXsections = 0;
  NumShadyPoints = 0;
  shady_and_ignore = 0;
  for (k=0; k<numPatches; k++) {
    for (p=0; p<=tpPatch1.iDeg0; p++)
    for (q=0; q<=tpPatch1.iDeg1; q++)
      for (r=0; r<=W; r++)
        tpPatch1.coeffs[p][q].dCoord[r] = tp1.coeffs[p][q].dCoord[r];

#ifdef _SCALE
    mxsc = maxscale;
#else
    mxsc = 1.0;
#endif
    tpPatch0.iDeg0 = solid[k].tp.iDeg0;
    tpPatch0.iDeg1 = solid[k].tp.iDeg1;
    allocate_tensor_patch(&tpPatch0);
    for (p=0; p<=tpPatch0.iDeg0; p++)
    for (q=0; q<=tpPatch0.iDeg1; q++) {
      for (r=0; r<=Z; r++) {
        tpPatch0.coeffs[p][q].dCoord[r] =
                        Scale*solid[k].tp.coeffs[p][q].dCoord[r];
#ifdef _SCALE
        if (mxsc < fabs(tpPatch0.coeffs[p][q].dCoord[r]))
          mxsc = fabs(tpPatch0.coeffs[p][q].dCoord[r]);
#else
	mxsc = 1.0;
#endif
      }
      tpPatch0.coeffs[p][q].dCoord[W] = solid[k].tp.coeffs[p][q].dCoord[W];
    }

#ifdef _SCALE
    for (p=0; p<=tpPatch0.iDeg0; p++)
    for (q=0; q<=tpPatch0.iDeg1; q++)
      for (r=0; r<=Z; r++)
        tpPatch0.coeffs[p][q].dCoord[r] /= mxsc;

    for (p=0; p<=tpPatch1.iDeg0; p++)
    for (q=0; q<=tpPatch1.iDeg1; q++)
      for (r=0; r<=Z; r++)
        tpPatch1.coeffs[p][q].dCoord[r] /= mxsc;
#endif

    bernsteinToPowerBasis( tpPatch0 );
    allocDeterminant( &ndet, tpPatch0.iDeg0, tpPatch0.iDeg1, UNFACTORED);
    implicitize(tpPatch0, ndet);
    powerToBernsteinBasis(tpPatch0);
    allocDeterminant( &ndetImplicit, tpPatch0.iDeg0, tpPatch0.iDeg1,FINAL);
    moveFactoredToFinalDeterminant( ndetImplicit, ndet,
                                    tpPatch0.iDeg0, tpPatch0.iDeg1 );
    Nrows = ndetImplicit.cRows;
    Ncols = ndetImplicit.cCols;
    base_case = check_for_base_points(ndetImplicit, Nrows, Ncols,
                                      &rank_row_val, &row_perm, &col_perm);
    if (base_case) {
      fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
      ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
      for (i=0; i<fdetImplicit.cCols; i++)
        ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

      for (i=0; i<rank_row_val; i++)
      {
        for (j=0; j<rank_row_val; j++)
        {
          fdetImplicit.coeffs[i][j].dCoord[X] = ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[X];
          fdetImplicit.coeffs[i][j].dCoord[Y] = ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
          fdetImplicit.coeffs[i][j].dCoord[Z] = ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
          fdetImplicit.coeffs[i][j].dCoord[W] = ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[W];
        }
      }
    }

    if (base_case) {
      allocDeterminant_1D( &fdet1D_Tmp, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
         2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);
      allocDeterminant_1D(&ndet1D_Tmp, rank_row_val, rank_row_val, 
			  tpPatch1.iDeg0, tpPatch1.iDeg1);
      change_to_uv( ndetImplicit, fdet1D_Tmp, tpPatch1 );
      change_to_uv( fdetImplicit, ndet1D_Tmp, tpPatch1 );
    }
    else {
      allocDeterminant_1D( &ndet1D_Tmp, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
         2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);
      change_to_uv( ndetImplicit, ndet1D_Tmp, tpPatch1 );
    }
    ALLOCN(pdPoints, POINT4D, ndet1D_Tmp.cRows);
    if (find_start_points(ndet1D_Tmp, &num_points,pdPoints,&num_cpoints,
                          pcpxPoints, V, power_pt(0.0),
                          fdet1D_Tmp, row_perm, tpPatch0.iDeg0, tpPatch0.iDeg1,
                          base_case, tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i=0; i<num_points; i++) {
      vval = inverse_power_pt(pdPoints[i].dCoord[V]);
      sval = inverse_power_pt(pdPoints[i].dCoord[S]);
      tval = inverse_power_pt(pdPoints[i].dCoord[T]);
      EvalPointOnPatch(tpPatch0, sval, tval, &xval1, &yval1, &zval1);
      EvalPointOnPatch(tpPatch1, 0.0, vval, &xval2, &yval2, &zval2);
/*
      if (!(FP_EQ_EPS(xval1, xval2, 1e-3) &&
	    FP_EQ_EPS(yval1, yval2, 1e-3) &&
	    FP_EQ_EPS(zval1, zval2, 1e-3))) {
	printf("The intersection points ARE (%lf,%lf,%lf) and (%lf,%lf,%lf)\n", xval1,yval1,zval1,xval2,yval2,zval2);
	exit(-1); 
      }
*/
      if (sval >= -EPSILON && sval <= 1.0+EPSILON &&
          tval >= -EPSILON && tval <= 1.0+EPSILON && vval >= -EPSILON) {
	if (FP_EQ(vval,0.0) || FP_EQ(vval,1.0)) 
	  printf("Warning: vval ( %lf ) is very close to boundary\n");
/*
	if (FP_EQ(sval,0.0) || FP_EQ(sval,1.0) || 
	    FP_EQ(tval,0.0) || FP_EQ(tval,1.0)) {
*/
	for (ind=0; ind<NumShadyPoints; ind++) 
	  if (FP_EQ_EPS(xval1,ShadyPtList[ind].dCoord[0],1e-3) &&
	      FP_EQ_EPS(yval1,ShadyPtList[ind].dCoord[1],1e-3) &&
	      FP_EQ_EPS(zval1,ShadyPtList[ind].dCoord[2],1e-3)) {
	    shady_and_ignore = 1;
	    break;
	  } 

/*
	}
*/
	if (shady_and_ignore)
	  shady_and_ignore = 0;
	else {
	  is_inside = 0;
	  for (j=0; j<solid[k].NumTrimPortions; j++) {
	    pt.x = sval;
	    pt.y = tval;
	    if (solid[k].SplineReprn)
	      is_inside = IsPtInsideSplinegon(
				  &solid[k].TrimPortion[j].SplinePortion,
				  &pt);
	    else
	      is_inside = is_point_inside_polygon(
		               solid[k].TrimPortion[j].PiecePortion.NumOfVertices,
			       solid[k].TrimPortion[j].PiecePortion.poly_vertex,
			       &pt);
	    if (is_inside) {
	      ShadyPtList[ind].dCoord[0] = xval1;
	      ShadyPtList[ind].dCoord[1] = yval1;
	      ShadyPtList[ind].dCoord[2] = zval1;
	      NumShadyPoints++;
	      break;
	    }
	  }
          if (is_inside) numXsections++;
	}
      }
    }
  }
  return numXsections;
}
#endif /* CLAC_IMPLICIT */
#endif /* BEZ_CLIP */

#ifdef MANY_TESTS
int RobustInOutTest(xv, yv, zv, solid, numPatches)
double xv, yv, zv;
TRIM_PATCH_INFO *solid;
int numPatches;
{
  int odd_cnt=0, even_cnt=0;
  int i, j, k, ret_val;
  double nxv[7], nyv[7], nzv[7];

  for (i=0; i<7; i++) {
    nxv[i] = 0.0;
    nyv[i] = 0.0;
    nzv[i] = 0.0;
  }
  nxv[0] = nxv[6] = 50.0;  
  nxv[1] = -50.0;  
  nyv[2] = nyv[6] = 50.0;  
  nyv[3] = -50.0;  
  nzv[4] = nzv[6] = 50.0;  
  nzv[5] = -50.0;  

  for (i=0; i<6; i++) {
    ret_val = FindNumCurveSurfInt(xv,yv,zv,nxv[i],nyv[i],nzv[i],
				  solid, numPatches);
    if (ret_val%2 == 0) even_cnt++;
    else odd_cnt++;
  }

  if (odd_cnt > even_cnt) return 1;
  else if (even_cnt > odd_cnt) return 0;
  else {
    printf("COULD NOT MAKE UP AS DECISION! TIE-BREAKER\n");
    return FindNumCurveSurfInt(xv, yv, zv, nxv[6], nyv[6], nzv[6],
				            solid, numPatches);
  }
}
#else
#ifdef _CALC_IMPLICIT
int RobustInOutTest(xv, yv, zv, solid, numPatches, sol_imp, imp_scale)
double xv, yv, zv;
TRIM_PATCH_INFO *solid;
int numPatches;
NDETERMINANT *sol_imp;
double imp_scale;
{
  int odd_cnt=0, even_cnt=0;
  int i, j, k, ret_val;
  BOUNDINGBOX *boxes, bigBB;
  double nxv[7], nyv[7], nzv[7];

  ALLOCN(boxes, BOUNDINGBOX, numPatches);

  bigBB.i3Dbox[0].low = 100000;
  bigBB.i3Dbox[0].high = -100000;
  bigBB.i3Dbox[1].low = 100000;
  bigBB.i3Dbox[1].high = -100000;
  bigBB.i3Dbox[2].low = 100000;
  bigBB.i3Dbox[2].high = -100000;

  for (i=0; i<numPatches; i++)
    compute_bounding_box(solid[i].tp, &(boxes[i]));

  for (i=0; i<numPatches; i++) {
    for (j=0; j<3; j++) {
      if (boxes[i].i3Dbox[j].low < bigBB.i3Dbox[j].low)
	bigBB.i3Dbox[j].low = boxes[i].i3Dbox[j].low;
      if (boxes[i].i3Dbox[j].high > bigBB.i3Dbox[j].high)
	bigBB.i3Dbox[j].high = boxes[i].i3Dbox[j].high;
    }
  }

  nxv[0] = bigBB.i3Dbox[0].low - 5.0;
  nxv[1] = bigBB.i3Dbox[0].high + 5.0;
  nxv[2] = bigBB.i3Dbox[0].low - 5.0;
  nxv[3] = bigBB.i3Dbox[0].high + 5.0;
  nyv[0] = bigBB.i3Dbox[1].low - 5.0;
  nyv[1] = bigBB.i3Dbox[1].high + 5.0;
  nyv[2] = bigBB.i3Dbox[1].high + 5.0;
  nyv[3] = bigBB.i3Dbox[1].low - 5.0;
  nzv[0] = bigBB.i3Dbox[2].low - 5.0;
  nzv[1] = bigBB.i3Dbox[2].high + 5.0;
  nzv[2] = bigBB.i3Dbox[2].high + 5.0;
  nzv[3] = bigBB.i3Dbox[2].low - 5.0;
  nxv[4] = bigBB.i3Dbox[0].high + 5.0;
  nyv[4] = (bigBB.i3Dbox[1].low + bigBB.i3Dbox[1].high)/2.0;
  nzv[4] = (bigBB.i3Dbox[2].low + bigBB.i3Dbox[2].high)/2.0;
  nxv[5] = (bigBB.i3Dbox[0].low + bigBB.i3Dbox[0].high)/2.0;
  nyv[5] = bigBB.i3Dbox[1].high + 5.0;
  nzv[5] = (bigBB.i3Dbox[2].low + bigBB.i3Dbox[2].high)/2.0;
  nxv[6] = (bigBB.i3Dbox[0].low + bigBB.i3Dbox[0].high)/2.0;
  nyv[6] = (bigBB.i3Dbox[1].low + bigBB.i3Dbox[1].high)/2.0;
  nzv[6] = bigBB.i3Dbox[2].high + 5.0;

  for (i=0; i<4; i++) {
    ret_val = FindNumCurveSurfInt(xv,yv,zv,nxv[i]-xv,nyv[i]-yv,nzv[i]-zv,
				  solid, numPatches, sol_imp, imp_scale);
    if (ret_val%2 == 0) even_cnt++;
    else odd_cnt++;
  }

  if (odd_cnt > even_cnt) return 1;
  else if (even_cnt > odd_cnt) return 0;
  else {
    /* printf("COULD NOT MAKE UP A DECISION! TIE-BREAKER\n"); */
    return FindNumCurveSurfInt(xv, yv, zv, nxv[4]-xv, nyv[4]-yv, nzv[4]-zv,
			       solid, numPatches,
			       sol_imp, imp_scale);
  }
  myfree(boxes);
}
#else
int RobustInOutTest(xv, yv, zv, solid, numPatches)
double xv, yv, zv;
TRIM_PATCH_INFO *solid;
int numPatches;
{
  int odd_cnt=0, even_cnt=0;
  int i, j, k, ret_val;
  BOUNDINGBOX *boxes, bigBB;
  double nxv[7], nyv[7], nzv[7];

  ALLOCN(boxes, BOUNDINGBOX, numPatches);

  bigBB.i3Dbox[0].low = 100000;
  bigBB.i3Dbox[0].high = -100000;
  bigBB.i3Dbox[1].low = 100000;
  bigBB.i3Dbox[1].high = -100000;
  bigBB.i3Dbox[2].low = 100000;
  bigBB.i3Dbox[2].high = -100000;

  for (i=0; i<numPatches; i++)
    compute_bounding_box(solid[i].tp, &(boxes[i]));

  for (i=0; i<numPatches; i++) {
    for (j=0; j<3; j++) {
      if (boxes[i].i3Dbox[j].low < bigBB.i3Dbox[j].low)
	bigBB.i3Dbox[j].low = boxes[i].i3Dbox[j].low;
      if (boxes[i].i3Dbox[j].high > bigBB.i3Dbox[j].high)
	bigBB.i3Dbox[j].high = boxes[i].i3Dbox[j].high;
    }
  }

  nxv[0] = bigBB.i3Dbox[0].low - 5.0;
  nxv[1] = bigBB.i3Dbox[0].high + 5.0;
  nxv[2] = bigBB.i3Dbox[0].low - 5.0;
  nxv[3] = bigBB.i3Dbox[0].high + 5.0;
  nyv[0] = bigBB.i3Dbox[1].low - 5.0;
  nyv[1] = bigBB.i3Dbox[1].high + 5.0;
  nyv[2] = bigBB.i3Dbox[1].high + 5.0;
  nyv[3] = bigBB.i3Dbox[1].low - 5.0;
  nzv[0] = bigBB.i3Dbox[2].low - 5.0;
  nzv[1] = bigBB.i3Dbox[2].high + 5.0;
  nzv[2] = bigBB.i3Dbox[2].high + 5.0;
  nzv[3] = bigBB.i3Dbox[2].low - 5.0;
  nxv[4] = bigBB.i3Dbox[0].high + 5.0;
  nyv[4] = (bigBB.i3Dbox[1].low + bigBB.i3Dbox[1].high)/2.0;
  nzv[4] = (bigBB.i3Dbox[2].low + bigBB.i3Dbox[2].high)/2.0;
  nxv[5] = (bigBB.i3Dbox[0].low + bigBB.i3Dbox[0].high)/2.0;
  nyv[5] = bigBB.i3Dbox[1].high + 5.0;
  nzv[5] = (bigBB.i3Dbox[2].low + bigBB.i3Dbox[2].high)/2.0;
  nxv[6] = (bigBB.i3Dbox[0].low + bigBB.i3Dbox[0].high)/2.0;
  nyv[6] = (bigBB.i3Dbox[1].low + bigBB.i3Dbox[1].high)/2.0;
  nzv[6] = bigBB.i3Dbox[2].high + 5.0;

  for (i=0; i<6; i++) {
    ret_val = FindNumCurveSurfInt(xv,yv,zv,nxv[i]-xv,nyv[i]-yv,nzv[i]-zv,
				  solid, numPatches);
    if (ret_val%2 == 0) even_cnt++;
    else odd_cnt++;
  }

  if (odd_cnt > even_cnt) return 1;
  else if (even_cnt > odd_cnt) return 0;
  else {
    /* printf("COULD NOT MAKE UP A DECISION! TIE-BREAKER\n"); */
    return FindNumCurveSurfInt(xv, yv, zv, nxv[6]-xv, nyv[6]-yv, nzv[6]-zv,
				            solid, numPatches);
  }
  myfree(boxes);
}
#endif /* CALC_IMPLICIT */
#endif /* MANY_TESTS */

void DeleteRedundantTrimCurvePoints(dic)
DOMAIN_INT_CURVE *dic;
{
  int i, j, k;
  double udiff, vdiff;
  double mod1, mod2;
  int going_back;
  double u1, v1, u2, v2, u3, v3;
  double uvec1, vvec1, uvec2, vvec2, dotprod;
  LIST_POINT2D *tmpptr1, *tmpptr2, *tmpptr3;

  tmpptr1 = dic->planeCurve;
  if (tmpptr1 == NULL) {
    printf("DeleteRedundantTrimCurvePoints: Null list input\n");
    return;
  }
  tmpptr2 = tmpptr1->plp2DNext;
  if (tmpptr2 == NULL) {
    printf("DeleteRedundantTrimCurvePoints: Only one point in curve\n");
    return;
  }
  tmpptr3 = tmpptr2->plp2DNext;
  going_back = 0;
  while (tmpptr3 != NULL) {
    u1 = tmpptr1->p2D.u;
    v1 = tmpptr1->p2D.v;
    u2 = tmpptr2->p2D.u;
    v2 = tmpptr2->p2D.v;
    u3 = tmpptr3->p2D.u;
    v3 = tmpptr3->p2D.v;
    uvec1 = u2 - u1;
    vvec1 = v2 - v1;
    uvec2 = u3 - u2;
    vvec2 = v3 - v2;
    mod1 = sqrt(uvec1*uvec1 + vvec1*vvec1);
    mod2 = sqrt(uvec2*uvec2 + vvec2*vvec2);
    dotprod = (uvec1*uvec2 + vvec1*vvec2)/(mod1*mod2);

    if ((dotprod >= 0.997 /* cos(4.5) */ && dotprod <= 1.0001) ||
	mod1 <= 0.005 || mod2 <= 0.005) {
      tmpptr1->plp2DNext = tmpptr3;

      tmpptr2 = tmpptr3;
      tmpptr3 = tmpptr3->plp2DNext;    
      going_back = 0;
    } else if (dotprod >= -1.0001 && dotprod <= -0.997) {
      tmpptr2->plp2DNext = tmpptr3->plp2DNext;

      tmpptr3 = tmpptr3->plp2DNext;
      going_back = 1;
    } else {
      tmpptr1 = tmpptr2;
      tmpptr2 = tmpptr3;
      tmpptr3 = tmpptr3->plp2DNext;
      going_back = 0;
    }
  }
  if (tmpptr3 != NULL) {
    printf("DeleteRedundantTrimCurvePoints: while loop quit incorrectly\n");
    exit(-1);
  }
  if (going_back) {
    tmpptr2->plp2DNext = NULL;
    dic->StEndPoints[0] = tmpptr2->p2D;
  }
}

void perf_dfs(node, value, n_info, NumInt)
int node, value, *NumInt;
neighbour_t *n_info;
{
  int i, j;

  if (markArray[node])
    {				/* perform check */
      for (i=0; i<n_info[node].nnb; i++) {
	j = n_info[node].adj[i];
	if (markArray[j] && (NumInt[j] == NumInt[node])) {
	  printf("perf_dfs: Error in Adjacency List\n");
	  exit(77);
	}
      }
    }
  else				/* unmarked */
    {
      markArray[node] = TRUE;
      NumInt[node] = value;
      for (i=0; i<n_info[node].nnb; i++) {
	j = n_info[node].adj[i];
	perf_dfs(j, 1-value, n_info, NumInt);
      }
    }
}

int GetLineEq(u1, v1, u2, v2, line)
double u1, v1, u2, v2;
LINE *line;
{
  if (FP_EQ_EPS(u1, u2, 1e-3) && FP_EQ_EPS(v1, v2, 1e-3)) {
    /*   printf("Illegal line (%lf,%lf) to (%lf,%lf). Both endpoints are same\n",
       u1,v1,u2,v2); */
    return 0;
  }

  line->x1 = u1;
  line->y1 = v1;
  line->x2 = u2;
  line->y2 = v2;

  line->Cx = v2 - v1;
  line->Cy = u1 - u2;
  line->Cc = (line->Cx)*u1 + (line->Cy)*v1;

  return 1;
}

int SegInLineTest(line, u31, v31, u32, v32)
double u31, v31, u32, v32;
LINE *line;
{
  double u1, v1, u2, v2;
  double act_val1, act_val2;
  int online, TryU;

  online = 0;

  u1 = line->x1;
  v1 = line->y1;
  u2 = line->x2;
  v2 = line->y2;

  act_val1 = (line->Cx)*u31 + (line->Cy)*v31;
  act_val2 = (line->Cx)*u32 + (line->Cy)*v32;

  if (FP_EQ_EPS(act_val1, line->Cc, 7.0e-4) &&
      FP_EQ_EPS(act_val2, line->Cc, 7.0e-4))
    online = 1;

  if (online) {
    if ((((u2-u31)*(u31-u1) > -1e-4) && ((v2-v31)*(v31-v1) > -1e-4)) ||
	(((u2-u32)*(u32-u1) > -1e-4) && ((v2-v32)*(v32-v1) > -1e-4)))
      return 1;
  }
  return 0;
}

int SegInPolygonBoundary(lines, NumLines, pt1, pt2)
LINE *lines;
int NumLines;
POINT2D *pt1, *pt2;
{
  int i;
  double u31, v31, u32, v32;

  u31 = pt1->u;
  v31 = pt1->v;
  u32 = pt2->u;
  v32 = pt2->v;

  for (i=0; i<NumLines; i++)
    if (SegInLineTest(&(lines[i]), u31, v31, u32, v32))
      return 1;

  return 0;
}

void MergeIfNecessary(CurveList, n)
DOMAIN_INT_CURVE **CurveList;
int *n;
{
  int i, j, k;
  LIST_POINT2D *last, *lastj;

  for (i=0; i<(*n); i++) {
    last = (*CurveList)[i].planeCurve;
    if (last == NULL) {
      printf("Something wrong in the lists\n");
      exit(-1);
    }
    while (last->plp2DNext != NULL) 
      last = last->plp2DNext;
    for (j=i+1; j<(*n); j++) {
      if (EQUAL2D_EPS((*CurveList)[i].StEndPoints[1],
		      (*CurveList)[j].StEndPoints[0], 2.5e-6)) {
	(*CurveList)[i].StEndPoints[1] =
	  (*CurveList)[j].StEndPoints[1];
	last->plp2DNext = (*CurveList)[j].planeCurve;
	while (last->plp2DNext != NULL) 
	  last = last->plp2DNext;
	for (k=j; k<((*n)-1); k++) {
	  (*CurveList)[k].StEndPoints[0] = (*CurveList)[k+1].StEndPoints[0];
	  (*CurveList)[k].StEndPoints[1] = (*CurveList)[k+1].StEndPoints[1];
	  (*CurveList)[k].planeCurve = (*CurveList)[k+1].planeCurve;
	}
	(*n)--;
      } else {
	if (EQUAL2D_EPS((*CurveList)[i].StEndPoints[0],
		      (*CurveList)[j].StEndPoints[1], 2.5e-6)) {
	  lastj = (*CurveList)[j].planeCurve;
	  if (lastj == NULL) {
	    printf("Something wrong in the lists\n");
	    exit(-1);
	  }
	  while (lastj->plp2DNext != NULL) 
	    lastj = lastj->plp2DNext;
	  lastj->plp2DNext = (*CurveList)[i].planeCurve;
	  (*CurveList)[i].StEndPoints[0] =
	    (*CurveList)[j].StEndPoints[0];
	  (*CurveList)[i].planeCurve = (*CurveList)[j].planeCurve;
	  while (lastj->plp2DNext != NULL) 
	    lastj = lastj->plp2DNext;
	  last = lastj;
	  for (k=j; k<((*n)-1); k++) {
	    (*CurveList)[k].StEndPoints[0] = (*CurveList)[k+1].StEndPoints[0];
	    (*CurveList)[k].StEndPoints[1] = (*CurveList)[k+1].StEndPoints[1];
	    (*CurveList)[k].planeCurve = (*CurveList)[k+1].planeCurve;
	  }
	  (*n)--;
	}
      }
    }
  }
  
}

int GetNonDegenerateCurveSegs(Incurve, lines, NumLines, Outcurve, index)
DOMAIN_INT_CURVE *Incurve, **Outcurve;
LINE *lines;
int NumLines, *index;
{
  LIST_POINT2D *ptr1, *ptr2;
  LIST_POINT2D *lastptr, *tmpptr;
  POINT2D StPt, EndPt;
  int i, counter;

  ptr1 = Incurve->planeCurve;
  ptr2 = ptr1->plp2DNext;

  if ((ptr1 == NULL) || (ptr2 == NULL))
    return 0;
  if (ptr2->plp2DNext == NULL) 
    return 0;

  while (ptr2->plp2DNext != NULL) {
    lastptr = (*Outcurve)[(*index)].planeCurve = NULL;
    while (SegInPolygonBoundary(lines, NumLines, 
				&(ptr1->p2D), &(ptr2->p2D)) &&
	   (ptr2->plp2DNext != NULL)) {
      ptr1 = ptr2;
      ptr2 = ptr2->plp2DNext;
    }
    if (ptr2->plp2DNext != NULL) {
      StPt = ptr1->p2D;
      while ((!SegInPolygonBoundary(lines, NumLines, 
				    &(ptr1->p2D), &(ptr2->p2D))) &&
	     (ptr2->plp2DNext != NULL)) {
	ALLOCN(tmpptr, LIST_POINT2D, 1);
	tmpptr->plp2DNext = NULL;
	tmpptr->p2D = ptr1->p2D;
	if (lastptr == NULL)
	  lastptr = (*Outcurve)[(*index)].planeCurve = tmpptr;
	else {
	  lastptr->plp2DNext = tmpptr;
	  lastptr = tmpptr;
	}
	ptr1 = ptr2;
	ptr2 = ptr2->plp2DNext;
      }
      if (ptr2->plp2DNext != NULL)
	EndPt = ptr1->p2D;
      else
	EndPt = ptr2->p2D;

      ALLOCN(tmpptr, LIST_POINT2D, 1);
      tmpptr->plp2DNext = NULL;
      tmpptr->p2D = EndPt;
      if (lastptr == NULL)
	lastptr = (*Outcurve)[(*index)].planeCurve = tmpptr;
      else {
	lastptr->plp2DNext = tmpptr;
	lastptr = tmpptr;
      }
      (*Outcurve)[(*index)].StEndPoints[0] = StPt;
      (*Outcurve)[(*index)++].StEndPoints[1] = EndPt;
    }
    MergeIfNecessary(Outcurve, index);
  }
}

/* This routine determines the component classification of each */
/* component of a solid */
#ifdef _CALC_IMPLICIT
void DetermineXsectionsforSolid(tpst, tp, solid, numsolidpatches,
				sol_imp, imp_scale, graph,
				num_comp, NumInt, comp_num)
TRIM_PORTION_SPLINE_TYPE *tpst;
TENSOR_PATCH *tp;
double imp_scale;
TRIM_PATCH_INFO *solid;
NDETERMINANT *sol_imp; /* All implicit patches of solid */
int numsolidpatches, num_comp, *NumInt, comp_num;
neighbour_t *graph;
{
  int n0, i;
  point_typ pt;
  double xv, yv, zv;

  pt = GetPtInsideSplinegon(tpst);
  EvalPointOnPatch((*tp), pt.x, pt.y, &xv, &yv, &zv);
  NumInt[comp_num] = (RobustInOutTest(xv, yv, zv, solid,
				      numsolidpatches, sol_imp, imp_scale)) % 2;
  for (i=0; i<MAX_NUM_PORTIONS; i++)
    markArray[i] = 0;
  perf_dfs(comp_num, NumInt[comp_num], graph, NumInt);
}
#else
void DetermineXsectionsforSolid(tpst, tp, solid, numsolidpatches, graph,
				num_comp, NumInt, comp_num)
TRIM_PORTION_SPLINE_TYPE *tpst;
TENSOR_PATCH *tp;
TRIM_PATCH_INFO *solid;
int numsolidpatches, num_comp, *NumInt, comp_num;
neighbour_t *graph;
{
  int n0, i;
  point_typ pt;
  double xv, yv, zv;

  pt = GetPtInsideSplinegon(tpst);
  EvalPointOnPatch((*tp), pt.x, pt.y, &xv, &yv, &zv);
  NumInt[comp_num] = (RobustInOutTest(xv, yv, zv, solid, numsolidpatches)) % 2;
  for (i=0; i<MAX_NUM_PORTIONS; i++)
    markArray[i] = 0;
  perf_dfs(comp_num, NumInt[comp_num], graph, NumInt);
}
#endif

void ObtainFinalPiecesofSolid(solid, tpst, numpatches, num_partns,
			      sol_comp, NumInt, orig_region_num, oper,
			      first)
TRIM_PATCH_INFO **solid;
TRIM_PORTION_SPLINE_TYPE **tpst;
int numpatches, *num_partns, *sol_comp, *NumInt, oper;
short orig_region_num[][MAX_PARTNS_PER_PATCH], first;
{
  register int i, j, k;
  int special_case, spec_case2, spec_case3;

  for (i=0; i<numpatches; i++) {
    special_case = spec_case2 = spec_case3 = 0;
    if (!first) {
      for (j=0; j<NumOvrPatches; j++) {
	if (overlap_pairs[j].ptches.patchNum1 == i) 
	  if (overlap_pairs[j].same_side)
	    special_case = 1;
	  else
	    if (first_complete_ovrlap[j] && !second_complete_ovrlap[j])
	      spec_case2 = 1;
	break;
      }
    } else {
      for (j=0; j<NumOvrPatches; j++) {
	if (overlap_pairs[j].ptches.patchNum0 == i) 
	  if (!overlap_pairs[j].same_side) {
	    if (!first_complete_ovrlap[j] && second_complete_ovrlap[j])
	      spec_case2 = 1;
	  } else {
	    if (first_complete_ovrlap[j] && !second_complete_ovrlap[j])
	      spec_case3 = 1;
	  }
	break;
      }
    }

    if (special_case && (num_partns[i] > 2)) {
      printf("ObtainFinalPiecesofSolid: Might be making a mistake\n");
    }
    (*solid)[i].NumTrimPortions = 0;
    for (j=0; j<num_partns[i]; j++) {
      k = sol_comp[i*MAX_PARTNS_PER_PATCH+j];
      if (oper == PLANE_CUT) {
	printf("ObtainFinalPiecesofSolid: Have not implemented PLANE_CUT yet\n");
	exit(-1);
      } else {
	switch (oper) {
	case INTERSECTION: 
	  if ((NumInt[k] % 2 == 1) && !spec_case3 && !special_case) {
	    /* I have to also obtain pointinsideregion and fill in */
	    /* orig_region_num */
	    copy_trim_portion_spline_type
	      (&((*solid)[i].TrimPortion[(*solid)[i].NumTrimPortions++].
		 SplinePortion), &tpst[i][j]);
	    orig_region_num[i][(*solid)[i].NumTrimPortions-1] = j;
	  }
	  break;
	case DOUNION: 
	  if (((NumInt[k] % 2 == 0) && !spec_case2) || special_case) {
	    copy_trim_portion_spline_type
              (&((*solid)[i].TrimPortion[(*solid)[i].NumTrimPortions++].
		 SplinePortion), &tpst[i][j], 1);
	    orig_region_num[i][(*solid)[i].NumTrimPortions-1] = j;
	  }
	  break;
	case DIFFERENCE: 
	  if (NumInt[k] % 2 == 0) {
	    if (first) {
	      copy_trim_portion_spline_type
		(&((*solid)[i].TrimPortion[(*solid)[i].NumTrimPortions++].
		   SplinePortion), &tpst[i][j], 1);
	      orig_region_num[i][(*solid)[i].NumTrimPortions-1] = j;
	    }
	  } else if (NumInt[k] % 2 == 1) {
	    if (!first && !special_case) {
	      copy_trim_portion_spline_type
		(&((*solid)[i].TrimPortion[(*solid)[i].NumTrimPortions++].
		   SplinePortion), &tpst[i][j], 1);
	      orig_region_num[i][(*solid)[i].NumTrimPortions-1] = j;
	    }
	  }
	  else {
	    printf("HAHAHA!! WHAT A BIG JOKE\n");
	     exit(-1);
	  }
	  break;
	}
      }
    }
    if ((*solid)[i].NumTrimPortions == 0) 
      (*solid)[i].csg_status = OUT_CSG;
    else
      (*solid)[i].csg_status = IN_CSG;
  }
}

void SplitIntoNewRegions(trimpatch, newnum, oldnum, solid, 
			 numsolidpatches, curves, oper, to_keep,
			 cutting_plane, if_first)
TRIM_PATCH_INFO *trimpatch, *solid;
DOMAIN_INT_CURVE *curves;
int oldnum, newnum, oper, numsolidpatches, if_first, to_keep;
POINT4D *cutting_plane;
/* the parameter if_first is to identify if the trimpatch belongs
   to the first solid or the second solid in the CSG operation */
{
  TRIM_PATCH_INFO *tmpstruct;
  int num_portions, i, j, k, r, try_new_one;
  int numpolysegs, NumInt[MAX_NUM_PORTIONS];

  DOMAIN_INT_CURVE *TempIntCurve;
  double xval, yval, zval;
  LINE line_eq[200];
  double upar, vpar;
  neighbour_t n_info[MAX_NUM_PORTIONS][MAX_NUM_PORTIONS];

  num_portions = trimpatch->NumTrimPortions;
  if (num_portions > 1) {
    printf("SplitIntoNewRegions: Too many trim portions\n");
    exit(-1);
  }
  ALLOCN(tmpstruct, TRIM_PATCH_INFO, num_portions);
  for (i=0; i<num_portions; i++) 
    for (j=0; j<MAX_NUM_PORTIONS; j++) {
      tmpstruct[i].TrimPortion[j].PiecePortion.poly_vertex = 
	(double (*)[2]) new(MAX_VERTEX_COUNT*2*sizeof(double));
      tmpstruct[i].TrimPortion[j].PiecePortion.AdjInfo = NULL;
      printf("SplitIntoNewRegions: No neighbors for piecewise reprn\n");
    }

  numpolysegs = newnum - oldnum;

/* CHANGED ON 6/25/95 to make sure none of the segments overlapped
   with the trim polygon */

  try_new_one = 0;
  if (numpolysegs > 0)
    for (i=0; i<num_portions; i++)
      if (partition_polygon(&(trimpatch->TrimPortion[i]), 
			    &(curves[oldnum]),
			    numpolysegs, &(tmpstruct[i]), n_info[i]) == -1) {
	try_new_one = 1;
	break;
      }

  if (try_new_one) {
    ALLOCN(TempIntCurve, DOMAIN_INT_CURVE, 10);
    for (i=0; i<10; i++)
      TempIntCurve[i].planeCurve = NULL;
    
    for (j=0; j<num_portions; j++) {
      if (trimpatch->TrimPortion[j].PiecePortion.NumOfVertices > 200) {
	printf("Too many points in trim curve\n");
	exit(-1);
      }
      for (k=0; k<trimpatch->TrimPortion[j].PiecePortion.NumOfVertices; k++) 
	if (!GetLineEq(trimpatch->TrimPortion[j].PiecePortion.poly_vertex[k][0],
		       trimpatch->TrimPortion[j].PiecePortion.poly_vertex[k][1],
		       trimpatch->TrimPortion[j].PiecePortion.poly_vertex[k+1][0],
		       trimpatch->TrimPortion[j].PiecePortion.poly_vertex[k+1][1],
		       &(line_eq[k]))) {
	  printf("Line segment too small to get accurate equation\n");
	  exit(-1);
	}
      r = 0;
      for (i=oldnum; i<newnum; i++)
	GetNonDegenerateCurveSegs(
		      &(curves[i]), line_eq,
		      trimpatch->TrimPortion[j].PiecePortion.NumOfVertices,
		      &TempIntCurve, &r);
    }

    if (r != numpolysegs)
#ifdef DEBUG_PRINT_POLY
      printf("Number of segments changed\n");
#endif
    if (r > 0)
      for (i=0; i<num_portions; i++) 
	if (partition_polygon(&(trimpatch->TrimPortion[i]), 
			      &(TempIntCurve[0]),
			      r, &(tmpstruct[i]), n_info[i]) == -1) {
	  printf("Stack not empty problems continue\n");
	  sleep(3);
	  exit(-1);
	}
  }

  trimpatch->NumTrimPortions = 0;
  for (i=0; i<num_portions; i++) {
   for (k=0; k<MAX_NUM_PORTIONS; k++)
     markArray[k] = 0;

   if (numpolysegs > 0) {
     j = 0;
     upar = tmpstruct[i].TrimPortion[j].PiecePortion.ptInsideRegion.u;
     vpar = tmpstruct[i].TrimPortion[j].PiecePortion.ptInsideRegion.v;
     EvalPointOnPatch(trimpatch->tp, upar, vpar, &xval, &yval, &zval);
     if (oper == PLANE_CUT) {
       if (to_keep != 1 && to_keep != -1) {
	 printf("Invalid TO_KEEP value: %d\n", to_keep);
	 exit(-1);
       }
       if ((cutting_plane->dCoord[X]*xval + cutting_plane->dCoord[Y]*yval +
	    cutting_plane->dCoord[Z]*zval + cutting_plane->dCoord[W]) * to_keep >= 0.0) 
	 NumInt[j] = TRUE;
       else
	 NumInt[j] = FALSE;
     } else 
       NumInt[j] = (RobustInOutTest(xval, yval, zval, solid,
				   numsolidpatches)) % 2;

     /* Doing a DFS of n_info to set NumInt */
     /* and reduce in_out tests */
     perf_dfs(0, NumInt[0], n_info[i], NumInt);
     

     for (j=0; j<tmpstruct[i].NumTrimPortions; j++) { 
       if (oper == PLANE_CUT) {
	 if (NumInt[j] == TRUE)
	   trimpatch->TrimPortion[trimpatch->NumTrimPortions++] =
	     tmpstruct[i].TrimPortion[j];
       } else {
	 switch (oper) {
	 case INTERSECTION: 
	   if (NumInt[j] % 2 == 1)
	     trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	       tmpstruct[i].TrimPortion[j];
	   break;
	 case DOUNION: 
	   if (NumInt[j] % 2 == 0)
	     trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	       tmpstruct[i].TrimPortion[j];
	   break;
	 case DIFFERENCE: 
	   if (NumInt[j] % 2 == 0) {
	     if (if_first)
	       trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
		 tmpstruct[i].TrimPortion[j];
	   } else if (NumInt[j] % 2 == 1) {
	     if (!if_first)
	       trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
		 tmpstruct[i].TrimPortion[j];
	   }
	   else {
	     printf("HAHAHA!! WHAT A BIG JOKE\n");
	     exit(-1);
	   }
	   break;
	 }
       }
     } 
   } else { /* numpolysegs == 0 */
     upar = trimpatch->TrimPortion[i].PiecePortion.ptInsideRegion.u;
     vpar = trimpatch->TrimPortion[i].PiecePortion.ptInsideRegion.v;
     EvalPointOnPatch(trimpatch->tp, upar, vpar, &xval, &yval, &zval);
     if (oper == PLANE_CUT) {
       if (to_keep != 1 && to_keep != -1) {
	 printf("Invalid TO_KEEP value: %d\n", to_keep);
	 exit(-1);
       }
       if ((cutting_plane->dCoord[X]*xval + cutting_plane->dCoord[Y]*yval +
	    cutting_plane->dCoord[Z]*zval + cutting_plane->dCoord[W]) * to_keep >= 0.0) 
	 trimpatch->TrimPortion[trimpatch->NumTrimPortions++] =
	   trimpatch->TrimPortion[i];
     } else {
       NumInt[0] = RobustInOutTest(xval, yval, zval, solid, numsolidpatches);
       switch (oper) {
       case INTERSECTION: 
	 if (NumInt[0] % 2 == 1)
	   trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	     trimpatch->TrimPortion[i];
	 break;
       case DOUNION: 
	 if (NumInt[0] % 2 == 0)
	   trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	     trimpatch->TrimPortion[i];
	 break;
       case DIFFERENCE: 
	 if (NumInt[0] % 2 == 0) {
	   if (if_first)
	     trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	       trimpatch->TrimPortion[i];
	 } else if (NumInt[0] % 2 == 1) {
	   if (!if_first)
	     trimpatch->TrimPortion[trimpatch->NumTrimPortions++] = 
	       trimpatch->TrimPortion[i];
	 }
	 else {
	   printf("HAHAHA!! WHAT A BIG JOKE\n");
	   exit(-1);
	 }
	 break;
       }
     }
   }
 }
  if (trimpatch->NumTrimPortions == 0) trimpatch->csg_status = OUT_CSG;
  else  trimpatch->csg_status = IN_CSG;

#ifdef __FREE_ALL
  for (i=0; i<num_portions; i++)
    if (tmpstruct[i].NumTrimPortions > 0)
      free_trim_patch(&(tmpstruct[i]));
  myfree(tmpstruct);
#endif
}
