
/* Macros and stuff for linear algebra operations with vectors and matrices. */

/* Initial declarations necessary for linear algebra macros */
#define LINALGINIT		double LAmag,LAsum; int LAi,LAj,LAk;
#define EXTERN_LINALG		extern double LAmag,LAsum; extern int LAi,LAj,LAk;

#define CROSSPROD3(a,b,c)       {a[0]=b[1]*c[2]-b[2]*c[1]; \
                                 a[1]=b[2]*c[0]-b[0]*c[2]; \
                                 a[2]=b[0]*c[1]-b[1]*c[0];}

/* Vector macros */
#define VECTORZERO(a,m)	         {for(LAi=0; LAi<m; a[LAi++] = 0); }
#define VECTORNEG(a,b,m)        {for(LAi=0; LAi<m; LAi++) a[LAi] = -b[LAi]; }

/* binary vector operations */
#define VEC_BIN_OP(a,b,op,c,m){for(LAi=0; LAi<m; LAi++)   \
				    a[LAi] = b[LAi] op c[LAi]; }
#define VEC_SCALAR_OP(a,b,op,c,m){for(LAi=0; LAi<m; LAi++) \
				    a[LAi] = b[LAi] op c; }

/* vector assignment operations */
#define VEC_ASN_OP(a,op,b,m){for(LAi=0; LAi<m; LAi++) a[LAi] op b[LAi]; }

#define NORMALIZE(a,m)		{ for(LAi=0,LAmag=0.;LAi<m;LAi++) \
				    LAmag += a[LAi]*a[LAi];	  \
				  LAmag=1./sqrt(LAmag);		  \
				  for(LAi=0; LAi<m; a[LAi++] *= LAmag); }
#define PRINT_VEC(a, m, string)	{ printf("%s : (", string);	  \
				  for(LAi=0; LAi<m; LAi++)	  \
					printf("%g ",a[LAi]);	  \
				  printf(")\n"); }
/*================= Matrix macros ===================*/

#define ALLOCMAT(a, m, n, type) {ALLOCN(a, type*, m);	  \
				 for(LAi=0; LAi<m; LAi++) 	  \
				   ALLOCN(a[LAi], type, n);}
#define MALLOCMAT(a, m, n, type) {MALLOCN(a, type*, m);	  \
				 for(LAi=0; LAi<m; LAi++) 	  \
				   MALLOCN(a[LAi], type, n);}
#define FREEMAT(a, m, n)        {for(LAi=0; LAi<m; \
				     LAi++){myfree(a[LAi]);a[LAi]=NULL;} \
				 myfree(a); a = NULL;}

/* a = identity */
#define IDENTMAT(a,m)		{for(LAi=0; LAi<m; LAi++) \
                                    for(LAj=0; LAj<m; LAj++) \
                                       a[LAi][LAj] = (LAi==LAj) ? 1. : 0.;}
/* vec a = mat b * vec c; mat b is m x n matrix */
#define MATVECMULT(a,b,c,m,n)	{for(LAi=0; LAi<m; LAi++) {\
				       for(LAj=0,LAsum=0.; LAj<n; LAj++) \
				          LAsum+=b[LAi][LAj]*c[LAj]; \
                                       a[LAi] = LAsum; \
                                    }}
/* a = b * c  (b is m x n and c is n x p) */
#define MATMULT(a,b,c,m,n,p)	{for(LAi=0; LAi<m; LAi++) \
				    for(LAj=0; LAj<p; LAj++) { \
				       for(LAk=0,LAsum=0.; LAk<n; LAk++) \
				          LAsum+=b[LAi][LAk]*c[LAk][LAj]; \
                                       a[LAi][LAj] = LAsum; \
                                    }}
/* a = b (m x n matrices)*/
#define MATASSIGN(a,b,m,n)	{for(LAi=0; LAi<m; LAi++) \
				    for(LAj=0; LAj<n; LAj++) { \
                                       a[LAi][LAj] = b[LAi][LAj]; \
                                    }}
/*      t		 */
/* a = b ( b is m x n )*/
#define MATTRANSPOSE(a,b,m,n)	{for(LAi=0; LAi<m; LAi++) \
				    for(LAj=0; LAj<n; LAj++) { \
                                       a[LAj][LAi] = b[LAi][LAj]; \
                                    }}

/* b[][a] = c, length of column vector is m */
#define SETCOL(a,b,c,m)         {for(LAi=0; LAi<m; LAi++) \
				    b[LAi][a] = c[LAi]; \
				}
/* c = b[][a], length of column vector is m */
#define GETCOL(a,b,c,m)         {for(LAi=0; LAi<m; LAi++) \
				    c[LAi] = b[LAi][a]; \
				}
/* b[a][] = c, length of row vector is n */
#define SETROW(a,b,c,n)         {for(LAi=0; LAi<n; LAi++) \
				    b[a][LAi] = c[LAi]; \
				}
/* c = b[a][], length of row vector is n */
#define GETROW(a,b,c,n)         {for(LAi=0; LAi<n; LAi++) \
				    c[LAi] = b[a][LAi]; \
				}
#define PRINT_MAT(a, p, q, string){ printf("%s :\n", string);	  \
				  printf("{");		          \
				  for(LAi=0; LAi<p; LAi++)	  \
				  { printf("{");		  \
				    for(LAj=0; LAj<q; LAj++)	  \
				    { printf("%lg ",a[LAi][LAj]);\
				      if (LAj<q-1) printf(",");   \
				    }				  \
				    if (LAi<p-1) printf("},\n"); else printf("}\n"); \
				  }				  \
				  printf("}\n");	          \
				  }
