#include "defs.h"
#include "lp.h"

extern void subdivide_bezier_patch();
TENSOR_PATCH ntp1, ntp2;

EXTERN_LINALG

int doLoopDetection(tp1, tp2)
TENSOR_PATCH tp1, tp2;
{
  int i, j, k;
  int status1, status2;
  TENSOR_PATCH p_norm_tp1, p_norm_tp2, p_norm_tp3;
  double tx, ty, tz;
  double s, t;
  
  DetGaussMap(tp1, &p_norm_tp1);
  ntp1 = p_norm_tp1;
/*
  This piece was just written to find the normal vector at some point
  so that I could rotate and create a singularity.
  s = 0.5;
  t = 0.5;
*/
  EvalPointOnPatch(tp1, s, t, &tx, &ty, &tz);
  EvalPointOnPatch(p_norm_tp1, s, t, &tx, &ty, &tz);
  DetGaussMap(tp2, &p_norm_tp2);
  ntp2 = p_norm_tp2;
  EvalPointOnPatch(tp2, s, t, &tx, &ty, &tz);
  EvalPointOnPatch(p_norm_tp2, s, t, &tx, &ty, &tz);
/*
*/

  p_norm_tp3.iDeg0 = p_norm_tp2.iDeg0;
  p_norm_tp3.iDeg1 = p_norm_tp2.iDeg1;
  allocate_tensor_patch(&p_norm_tp3);
  for (i=0; i<=p_norm_tp3.iDeg0; i++)
  for (j=0; j<=p_norm_tp3.iDeg1; j++) {
    for (k=0; k<3; k++) 
      p_norm_tp3.coeffs[i][j].dCoord[k] = -p_norm_tp2.coeffs[i][j].dCoord[k];
    p_norm_tp3.coeffs[i][j].dCoord[3] = p_norm_tp2.coeffs[i][j].dCoord[3];
  }

  status1 = do_lin_prog(p_norm_tp1, p_norm_tp2, -1);
  status2 = do_lin_prog(p_norm_tp1, p_norm_tp3, -1);

  free_ten_p(&p_norm_tp1);
  free_ten_p(&p_norm_tp2);
  free_ten_p(&p_norm_tp3);

  if (status1 == INFEASIBLE || status2 == INFEASIBLE) return LOOP_CASE;
  else return NO_LOOP_CASE;
}

void SplitCurveatC(tp, param, constVal, tp1, c, rev)
TENSOR_PATCH tp, *tp1;
int constVal, param, rev;
double c;
{
  POINT4D **ptList;
  int i, j, k, l;
  int deg;

  if (param == U) deg = tp.iDeg0;
  else deg = tp.iDeg1;

  ALLOCMAT(ptList, deg+1, deg+1, POINT4D);
  for (i=0; i<=deg; i++) {
    if (param == U) {
      if (!rev) 
	for (j=0; j<4; j++)
	  ptList[0][i].dCoord[j] = tp.coeffs[i][constVal].dCoord[j];
      else 
	for (j=0; j<4; j++)
	  ptList[0][i].dCoord[j] = tp.coeffs[deg-i][constVal].dCoord[j];
    } else { /* param == V */
      if (!rev) 
	for (j=0; j<4; j++)
	  ptList[0][i].dCoord[j] = tp.coeffs[constVal][i].dCoord[j];
      else 
	for (j=0; j<4; j++)
	  ptList[0][i].dCoord[j] = tp.coeffs[constVal][deg - i].dCoord[j];
    }
  }

  for (i=1; i<=deg; i++)
  for (j=i; j<=deg; j++)
    for (k=0; k<4; k++)
      ptList[i][j].dCoord[k] = (1-c)*ptList[i-1][j-1].dCoord[k] +
			       c*ptList[i-1][j].dCoord[k];

  for (i=0; i<=deg; i++) {
    if (param == U) {
      if (!rev)
	for (j=0; j<4; j++)
	  tp1->coeffs[i][constVal].dCoord[j] = ptList[i][i].dCoord[j];
      else 
	for (j=0; j<4; j++)
	 tp1->coeffs[i][constVal].dCoord[j]=ptList[deg-i][deg-i].dCoord[j];
    } else { /* param == V */
      if (!rev)
	for (j=0; j<4; j++)
	  tp1->coeffs[constVal][i].dCoord[j] = ptList[i][i].dCoord[j];
      else 
	for (j=0; j<4; j++)
	 tp1->coeffs[constVal][i].dCoord[j]=ptList[deg-i][deg-i].dCoord[j];
    }
  }
  FREEMAT(ptList, deg+1, deg+1);
}

void DividePatchatC(tp, tp1, tp2, c, param)
TENSOR_PATCH tp, *tp1, *tp2;
double c;
int param;
{
  int i, j, k;
  int udeg, vdeg, deg;

  udeg = tp.iDeg0;
  vdeg = tp.iDeg1;

  tp1->iDeg0 = tp2->iDeg0 = udeg;
  tp1->iDeg1 = tp2->iDeg1 = vdeg;
  allocate_tensor_patch(tp1);
  allocate_tensor_patch(tp2);

  if (param == U) deg = vdeg;
  else deg = udeg;

  for (j=0; j<=deg; j++) {
    SplitCurveatC(tp, param, j, tp1, c, 0);
    SplitCurveatC(tp, param, j, tp2, 1-c, 1);
  }
}

#ifdef DONT_USE_THIS
void HandleLoopCase(solid1, ns1, solid2, ns2, oper, step_size)
TRIM_PATCH_INFO **solid1, **solid2;
int ns1, ns2, oper;
double step_size;
{
  TRIM_PATCH_INFO *tmpsolid1, *tmpsolid2;
  INT_CURVE_STRUCT **ccomp;
  double c1, c2;
  int i, j, k, tot_curve_num, param;
  LIST_POINT2D *tmp2D;
  LIST_POINT4D *tmp4D;
  int p, q, f_index, sol_index;
  double mult_fac_u, add_fac_u, mult_fac_v, add_fac_v;
  double u_min, u_max, v_min, v_max;
  static int parityvar = 0;

  ALLOCN(tmpsolid1, TRIM_PATCH_INFO, 2);
  ALLOCN(tmpsolid2, TRIM_PATCH_INFO, 2);

  tmpsolid1[0].tp.iDeg0 = tmpsolid1[1].tp.iDeg0 = 
				(*solid1)[ns1].tp.iDeg0;
  tmpsolid1[0].tp.iDeg1 = tmpsolid1[1].tp.iDeg1 = 
				(*solid1)[ns1].tp.iDeg1;
  tmpsolid2[0].tp.iDeg0 = tmpsolid2[1].tp.iDeg0 = 
				(*solid2)[ns2].tp.iDeg0;
  tmpsolid2[0].tp.iDeg1 = tmpsolid2[1].tp.iDeg1 = 
				(*solid2)[ns2].tp.iDeg1;

  for (i=0; i<2; i++) {
    allocate_tensor_patch(&(tmpsolid1[i]));
    allocate_tensor_patch(&(tmpsolid2[i]));
  }

  if (!(parityvar%2)) {
    param = U;
  }
  else {
    param = V;
  }
  parityvar++;

  c1 = 0.4 + drand48() * 0.2;
  c2 = 0.4 + drand48() * 0.2;

  if (param == U) {
    subdivide_bezier_patch((*solid1)[ns1].tp, 0.0, c1, 0.0, 1.0, 
			   &(tmpsolid1[0].tp));
    subdivide_bezier_patch((*solid1)[ns1].tp, c1, 1.0, 0.0, 1.0, 
			   &(tmpsolid1[1].tp));
    subdivide_bezier_patch((*solid2)[ns2].tp, 0.0, c2, 0.0, 1.0, 
			   &(tmpsolid2[0].tp));
    subdivide_bezier_patch((*solid2)[ns2].tp, c2, 1.0, 0.0, 1.0, 
			   &(tmpsolid2[1].tp));
  } else { /* param == V */
    subdivide_bezier_patch((*solid1)[ns1].tp, 0.0, 1.0, 0.0, c1,
			   &(tmpsolid1[0].tp));
    subdivide_bezier_patch((*solid1)[ns1].tp, 0.0, 1.0, c1, 1.0,
			   &(tmpsolid1[1].tp));
    subdivide_bezier_patch((*solid2)[ns2].tp, 0.0, 1.0, 0.0, c2,
			   &(tmpsolid2[0].tp));
    subdivide_bezier_patch((*solid2)[ns2].tp, 0.0, 1.0, c2, 1.0,
			   &(tmpsolid2[1].tp));
  }

/*
  DividePatchatC((*solid1)[ns1].tp, &(tmpsolid1[0].tp), &(tmpsolid1[1].tp),
		 c1, param);
  DividePatchatC((*solid2)[ns2].tp, &(tmpsolid2[0].tp), &(tmpsolid2[1].tp),
		 c2, param);
*/

  
  for (i=0; i<2; i++) {
    tmpsolid1[i].NumTrimPortions = 1;
    tmpsolid1[i].SplineReprn = 0;
    printf("HandleLoopCase: Patch has SplineReprn 0\n");
    tmpsolid1[i].TrimPortion[0].PiecePortion.NumOfVertices = 4;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[0][0] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[0][1] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[1][1] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[3][0] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[4][0] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.poly_vertex[4][1] = 0.0;
    tmpsolid1[i].TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
    tmpsolid1[i].TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;

    tmpsolid2[i].NumTrimPortions = 1;
    tmpsolid2[i].TrimPortion[0].PiecePortion.NumOfVertices = 4;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[0][0] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[0][1] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[1][1] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[3][0] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[4][0] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.poly_vertex[4][1] = 0.0;
    tmpsolid2[i].TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
    tmpsolid2[i].TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;
  }

  ALLOCN(ccomp, INT_CURVE_STRUCT *, 4);
  for (i=0; i<4; i++)
    ALLOCN(ccomp[i], INT_CURVE_STRUCT, MAX_COMP_PER_PATCH);

  findXsectionCurve(&tmpsolid1, 2, &tmpsolid2, 2, 0, oper, step_size);

  if (param == U) {
    mult_fac_v = 1.0;
    add_fac_v = 0.0;
  } else {
    mult_fac_u = 1.0;
    add_fac_u = 0.0;
  }
/* First Solid; First Patch */
  tot_curve_num = tmpsolid1[0].NumTrimCurves;
  p = 2;
  q = 3;
  f_index = 0;
  sol_index = 0;
  if (sol_index == 0) {
    if (param == U) {
      mult_fac_u = c1;
      add_fac_u = 0.0;
    } else {
      mult_fac_v = c1;
      add_fac_v = 0.0;
    }
  } else {
    if (param == U) {
      mult_fac_u = 1-c1;
      add_fac_u = c1;
    } else {
      mult_fac_v = 1-c1;
      add_fac_v = c1;
    }
  }
  if (tot_curve_num >= MAX_COMP_PER_PATCH) {
    printf("Too many components. Cannot Handle. Exiting.. \n");
    exit(-1);
  }
  for (i=0; i<tot_curve_num; i++) {
    ccomp[f_index][i].endPoints[0].dCoord[p] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[0].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[0].dCoord[q] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[0].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].endPoints[1].dCoord[p] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[1].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[1].dCoord[q] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[1].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].Curve = NULL;
    tmp2D = tmpsolid1[0].Curves[i].planeCurve;
    while (tmp2D != NULL) {
      ALLOCN(tmp4D, LIST_POINT4D, 1);
      tmp4D->p4D.dCoord[p] = tmp2D->p2D.u*mult_fac_u+add_fac_u;
      tmp4D->p4D.dCoord[q] = tmp2D->p2D.v*mult_fac_v+add_fac_v;
      tmp4D->plp4DNext = ccomp[f_index][i].Curve;
      ccomp[f_index][i].Curve = tmp4D;
      tmp2D = tmp2D->plp2DNext;
    }
  }
  for (j=0; j<tot_curve_num; j++) 
    Merge2DLists(solid1, ns1, ccomp[f_index][j], 1);

/* First Solid; Second Patch */
  tot_curve_num = tmpsolid1[1].NumTrimCurves;
  p = 2;
  q = 3;
  f_index = 1;
  sol_index = 1;
  if (sol_index == 0) {
    if (param == U) {
      mult_fac_u = c1;
      add_fac_u = 0.0;
    } else {
      mult_fac_v = c1;
      add_fac_v = 0.0;
    }
  } else {
    if (param == U) {
      mult_fac_u = 1-c1;
      add_fac_u = c1;
    } else {
      mult_fac_v = 1-c1;
      add_fac_v = c1;
    }
  }
  if (tot_curve_num >= MAX_COMP_PER_PATCH) {
    printf("Too many components. Cannot Handle. Exiting.. \n");
    exit(-1);
  }
  for (i=0; i<tot_curve_num; i++) {
    ccomp[f_index][i].endPoints[0].dCoord[p] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[0].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[0].dCoord[q] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[0].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].endPoints[1].dCoord[p] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[1].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[1].dCoord[q] =
	tmpsolid1[sol_index].Curves[i].StEndPoints[1].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].Curve = NULL;
    tmp2D = tmpsolid1[1].Curves[i].planeCurve;
    while (tmp2D != NULL) {
      ALLOCN(tmp4D, LIST_POINT4D, 1);
      tmp4D->p4D.dCoord[p] = tmp2D->p2D.u*mult_fac_u+add_fac_u;
      tmp4D->p4D.dCoord[q] = tmp2D->p2D.v*mult_fac_v+add_fac_v;
      tmp4D->plp4DNext = ccomp[f_index][i].Curve;
      ccomp[f_index][i].Curve = tmp4D;
      tmp2D = tmp2D->plp2DNext;
    }
  }
  for (j=0; j<tot_curve_num; j++) 
    Merge2DLists(solid1, ns1, ccomp[f_index][j], 1);

/* Second Solid; First Patch */
  tot_curve_num = tmpsolid2[0].NumTrimCurves;
  p = 0;
  q = 1;
  f_index = 2;
  sol_index = 0;
  if (sol_index == 0) {
    if (param == U) {
      mult_fac_u = c2;
      add_fac_u = 0.0;
    } else {
      mult_fac_v = c2;
      add_fac_v = 0.0;
    }
  } else {
    if (param == U) {
      mult_fac_u = 1-c2;
      add_fac_u = c2;
    } else {
      mult_fac_v = 1-c2;
      add_fac_v = c2;
    }
  }
  if (tot_curve_num >= MAX_COMP_PER_PATCH) {
    printf("Too many components. Cannot Handle. Exiting.. \n");
    exit(-1);
  }
  for (i=0; i<tot_curve_num; i++) {
    ccomp[f_index][i].endPoints[0].dCoord[p] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[0].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[0].dCoord[q] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[0].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].endPoints[1].dCoord[p] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[1].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[1].dCoord[q] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[1].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].Curve = NULL;
    tmp2D = tmpsolid2[0].Curves[i].planeCurve;
    while (tmp2D != NULL) {
      ALLOCN(tmp4D, LIST_POINT4D, 1);
      tmp4D->p4D.dCoord[p] = tmp2D->p2D.u*mult_fac_u+add_fac_u;
      tmp4D->p4D.dCoord[q] = tmp2D->p2D.v*mult_fac_v+add_fac_v;
      tmp4D->plp4DNext = ccomp[f_index][i].Curve;
      ccomp[f_index][i].Curve = tmp4D;
      tmp2D = tmp2D->plp2DNext;
    }
  }
  for (j=0; j<tot_curve_num; j++) 
    Merge2DLists(solid2, ns2, ccomp[f_index][j], 2);

/* Second Solid; Second Patch */
  tot_curve_num = tmpsolid2[1].NumTrimCurves;
  p = 0;
  q = 1;
  f_index = 3;
  sol_index = 1;
  if (sol_index == 0) {
    if (param == U) {
      mult_fac_u = c2;
      add_fac_u = 0.0;
    } else {
      mult_fac_v = c2;
      add_fac_v = 0.0;
    }
  } else {
    if (param == U) {
      mult_fac_u = 1-c2;
      add_fac_u = c2;
    } else {
      mult_fac_v = 1-c2;
      add_fac_v = c2;
    }
  }
  if (tot_curve_num >= MAX_COMP_PER_PATCH) {
    printf("Too many components. Cannot Handle. Exiting.. \n");
    exit(-1);
  }
  for (i=0; i<tot_curve_num; i++) {
    ccomp[f_index][i].endPoints[0].dCoord[p] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[0].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[0].dCoord[q] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[0].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].endPoints[1].dCoord[p] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[1].u*mult_fac_u+add_fac_u;
    ccomp[f_index][i].endPoints[1].dCoord[q] =
	tmpsolid2[sol_index].Curves[i].StEndPoints[1].v*mult_fac_v+add_fac_v;
    ccomp[f_index][i].Curve = NULL;
    tmp2D = tmpsolid2[1].Curves[i].planeCurve;
    while (tmp2D != NULL) {
      ALLOCN(tmp4D, LIST_POINT4D, 1);
      tmp4D->p4D.dCoord[p] = tmp2D->p2D.u*mult_fac_u+add_fac_u;
      tmp4D->p4D.dCoord[q] = tmp2D->p2D.v*mult_fac_v+add_fac_v;
      tmp4D->plp4DNext = ccomp[f_index][i].Curve;
      ccomp[f_index][i].Curve = tmp4D;
      tmp2D = tmp2D->plp2DNext;
    }
  }
  for (j=0; j<tot_curve_num; j++) 
    Merge2DLists(solid2, ns2, ccomp[f_index][j], 2);

  free_trim_patch(&(tmpsolid1[0]));
  free_trim_patch(&(tmpsolid1[1]));
  free_trim_patch(&(tmpsolid2[0]));
  free_trim_patch(&(tmpsolid2[1]));
}
#endif /* DONT_USE_THIS */

