#define GOLD 	1.618034
#define GLIMIT	100.0
#define TINY	1.0e-20
#define MAX(a,b) 	((a) > (b) ? (a) : (b))
#define fabs(a) 	((a) < 0.0 ? -a : a)
#define SIGN(a,b)	((b) > 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d)	(a)=(b); (b)=(c); (c)=(d);
#define ITER_MAX	100
#define ITMAX	200
#define CGOLD	0.3819660
#define ZEPS	1.0e-10
#define TOL	2.0e-4

int ncom = 0;
double *pcom;
double *xicom;
double (*nrfunc)();
static double sqrarg;
#define SQR(a) 	(sqrarg = (a), sqrarg*sqrarg)

void mnbrak(ax, bx, cx, fa, fb, fc, func)
double *ax, *bx, *cx, *fa, *fb, *fc;
double (*func)();
{
  double ulim, u, r, q, fu, dum;

  *fa = (*func)(*ax);
  *fb = (*func)(*bx);
  if (*fb > *fa) {
    SHFT(dum,*ax,*bx,dum);
    SHFT(dum,*fb,*fa,dum);
  }

  *cx = (*bx) + GOLD*(*bx - *ax);
  *fc = (*func)(*cx);

  while (*fb > *fc) {
    r = (*bx - *ax) * (*fb - *fc);
    q = (*bx - *cx) * (*fb - *fa);
    u = (*bx) - ((*bx - *cx)*q - (*bx - *ax)*r) /
		(2.0*SIGN(MAX(fabs(q-r), TINY), q-r));
    ulim = (*bx) + GLIMIT*(*cx - *bx);
    if ((*bx - u)*(u - *cx) > 0.0) {
      fu = (*func)(u);
      if (fu < *fc) {
	*ax = (*bx);
	*bx = u;
	*fa = (*fb);
	*fb = fu;
	return;
      } 
      else if (fu > *fb) {
	*cx = u;
	*fc = fu;
	return;
      }
      u = (*cx) + GOLD*(*cx - *bx);
      fu = (*func)(u);
    }
    else if ((*cx - u)*(u - ulim) > 0.0) {
      fu = (*func)(u);
      if (fu < *fc) {
	SHFT(*bx,*cx,u,*cx+GOLD*(*cx - *bx));
	SHFT(*fb,*fc,fu,(*func)(u));
      }
    }
    else if ((u - ulim)*(ulim - *cx) >= 0.0) {
      u = ulim;
      fu = (*func)(u);
    }
    else {
      u = (*cx) + GOLD*(*cx - *bx);
      fu = (*func)(u);
    }

    SHFT(*ax,*bx,*cx, u);
    SHFT(*fa, *fb, *fc, fu);
  }
}

double brent(ax, bx, cx, f, tol, xmin)
double ax, bx, cx, tol, *xmin;
double (*f)();
{
  int iter;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e = 0.0;

  a = ((ax < cx) ? ax : cx);
  b = ((ax > cx) ? ax : cx);
  x = w = v = bx;
  fw = fv = fx = (*f)(x);

  for (iter=1; iter <= ITER_MAX; iter++) {
    xm = 0.5*(a+b);
    tol2 = 2.0*(tol1=tol*fabs(x)+ZEPS);
    if (fabs(x-xm) <= (tol2 - 0.5*(b-a))) {
      *xmin = x;
      return fx;
    }

    if (fabs(e) > tol1) {
      r = (x-w) * (fx - fv);
      q = (x-v) * (fx - fw);
      p = (x-v)*q - (x-w)*r;
      q = 2.0*(q-r);
      if (q > 0.0) p = -p;
      q = fabs(q);
      etemp = e;
      e = d;
      if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p > q*(b-x))
	d = CGOLD*(e=(x > xm ? a-x : b-x));
      else {
	d = p/q;
	u = x + d;
	if (u-a < tol2 || b-u < tol2)
	  d = SIGN(tol1, xm-x);
      }
    }
    else {
      d = CGOLD*(e=(x > xm ? a-x : b-x));
    }

    u = (fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
    fu = (*f)(u);
    if (fu <= fx) {
      if (u >= x) a = x;
      else b = x;
      SHFT(v,w,x,u);
      SHFT(fv,fw,fx,fu);
    } else {
      if (u < x) a = u;
      else b = u;
      if (fu <= fw || w == x) {
	v = w;
	w = u;
	fv = fw;
	fw = fu;
      }
      else if (fu <= fv || v == x || v == w) {
	v = u;
	fv = fu;
      }
    }
  }
   /* printf("Too many iterations in BRENT\n"); */
  *xmin = x;
  return fx;
}

double f1dim(x)
double x;
{
  int j;
  double f, *xt;

  xt = (double *) malloc(ncom * sizeof(double));
  for (j=0; j<ncom; j++)
    xt[j] = pcom[j] + x*xicom[j];

  f = (*nrfunc)(xt);
  myfree(xt);
  return f;
}

void linmin(p,xi,n,fret,func)
double *p, *xi, *fret, (*func)();
int n;
{
  int j;
  double xx, xmin, fx, fb, fa, bx, ax;

  ncom = n;
  pcom = (double *) malloc(n * sizeof(double));
  xicom = (double *) malloc(n * sizeof(double));
  nrfunc = func;

  for (j=0; j<n; j++) {
    pcom[j] = p[j];
    xicom[j] = xi[j];
  }

  ax = 0.0;
  xx = 1.0;
  bx = 2.0;
  mnbrak(&ax,&xx,&bx,&fa,&fx,&fb,f1dim);
  *fret = brent(ax,xx,bx,f1dim,0.1,&xmin);

  for (j=0; j<n; j++) {
    xi[j] *= xmin;
    p[j] += xi[j];
  }

  myfree(xicom); myfree(pcom);
}

void powell(p,xi,n,ftol,iter,fret,func)
double *p, **xi, ftol, *fret, (*func)();
int n, *iter;
{
  int i, ibig, j;
  double t,fptt,fp,del;
  double *pt,*ptt,*xit;
  double tmp1, tmp2;

  pt = (double *) malloc(n * sizeof(double));
  ptt = (double *) malloc(n * sizeof(double));
  xit = (double *) malloc(n * sizeof(double));
  *fret = (*func)(p);

  for (j=0; j<n; j++) pt[j] = p[j];
  for (*iter=1;;(*iter)++) {
    fp = (*fret);
    ibig = 0;
    del = 0.0;
    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) xit[j] = xi[j][i];
      fptt = (*fret);
      linmin(p,xit,n,fret,func);
      if (fabs(fptt - (*fret)) > del) {
	del = fabs(fptt - (*fret));
	ibig = i;
      }
    }
    if (2.0*fabs(fp - (*fret)) <= ftol*(fabs(fp) + fabs(*fret))) {
      myfree(xit); myfree(ptt); myfree(pt); return;
    }
    if (*iter == ITMAX) printf("Too many iterations in POWELL\n");
    for (j=0; j<n; j++) {
      ptt[j] = 2.0*p[j] - pt[j];
      xit[j] = p[j] - pt[j];
      pt[j] = p[j];
    }
    fptt = (*func)(ptt);
    if (fptt < fp) {
      t = 2.0*(fp-2.0*(*fret)+fptt)*SQR(fp-(*fret)-del)-del*SQR(fp-fptt);
      if (t < 0.0) {
	linmin(p,xit,n,fret,func);
	for (j=0; j<n; j++) xi[j][ibig] = xit[j];
      }
    }
  }
}

