#include "defs.h"

#define ITER_LIMIT	100

EXTERN_LINALG

void
power_iterate( rgrgB, rgEVec, iDim, iDegree, dGuess, pdEVal, m, iteration_count, base_point_case )
double      **rgrgB;
double      *rgEVec;
int         iDim, iDegree, m;
double      dGuess, *pdEVal;
int	    *iteration_count;
int base_point_case;
{

    double *rgQ, *rgdWork, *ipvt;
    double **rgrgA, *LastMat;
    double **rgrgTmp, **rgrgU, **rgrgV, *rgK;
    int    i,j,k,l;
    double dOldEst, dNewEst, dRCond, dMagQVec;
    double dS, dT, dOldS, dOldT;
    double dNumer, dDenom, dAlpha1, dAlpha2;
    int dim, base;

    base = iDim - iDegree;
    dim = iDim / iDegree;

    /* allocate/initialize the matrices */
    ALLOCMAT(rgrgA, iDim, iDim, double);
    ALLOCN(LastMat, DOUBLE, iDegree * iDegree);
    ALLOCMAT(rgrgTmp, iDim, iDim, double);
    ALLOCMAT(rgrgU, iDim, iDim, double);
    ALLOCMAT(rgrgV, iDim, iDim, double);
    ALLOCN(rgK, DOUBLE, iDim);
    ALLOCN(rgdWork, DOUBLE, iDim);
    ALLOCN(rgQ, DOUBLE, iDim);
    ALLOCN(ipvt, DOUBLE, iDegree);

    /* perform inverse power iteartions as per
       Golub and Van Loan, pg 383 */

    for (i=0;i<iDim;i++)
	rgQ[i]=rgEVec[i];

    /* normalize and use as next estimate */
    for(i=0, dMagQVec=0.0;i<iDim;i++)
	dMagQVec += rgQ[i]*rgQ[i];
    
    dMagQVec = sqrt(dMagQVec);
    
    if (FP_EQ_EPS(0.0, dMagQVec, 1e-10)) {
	fprintf(stderr, "\npower_iterate(): zero magnitude e-vector.\n");
	exit(-3);
    }
    
    VEC_SCALAR_OP( rgQ, rgQ, /, dMagQVec, iDim );

    /* calc (D*lambda - B) and check if it singular to proceed */
    for (i = 0; i < iDim; i++)
    for (j = 0; j < iDim; j++)
      rgrgTmp[i][j] = (i == j) ? (dGuess - rgrgB[i][j]) : (-rgrgB[i][j]);
    call_svd(rgrgTmp, iDim, iDim, rgrgU, rgK, rgrgV);
    if (FP_EQ_EPS(rgK[iDim-1], 0.0, 1e-10)) {
      *iteration_count = 0;
      *pdEVal = dGuess;
      VEC_ASN_OP(rgEVec, =, rgQ, iDim );
      return;
    }

    /* calc (D*lambda - B) */
    for (j = 0; j < dim; j++)
    {
        l = j*iDegree;
        for ( i = 0; i < iDegree; i++)
            for ( k = 0; k < iDegree; k++)
        {
            rgrgA[base + i][l + k] = ((base + i) == (l+k)) ? 
				     (dGuess - rgrgB[base + i][l + k]) :
				     (-rgrgB[base + i][l + k]) ;
       }
    }

    dAlpha1 = dGuess - rgrgB[0][0];
    dAlpha2 = (iDegree == 0) ? (dAlpha1) : (- rgrgB[0][iDegree]);

    Decompose(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2, ipvt);

/* #endif */

    *iteration_count = 0;
    dNewEst = dGuess;
    dS = rgQ[1]/rgQ[0]; 
    dT = rgQ[m]/rgQ[0]; /* How did they get 4. In general it should
			   be m, for bicubic it should be 3 */

    do {
	dOldEst = dNewEst;
        dOldS = dS;
	dOldT = dT;
        
/* do rgQ = rgrgD*rgQ 
        MATVECMULT(rgdWork, rgrgD, rgQ, iDim, iDim);
	VEC_ASN_OP(rgQ, =, rgdWork, iDim);
*/

	/* solve for rgQ */
        Solve(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2, rgQ, ipvt);

	/* normalize and use as next estimate */
	for(i=0, dMagQVec=0.0;i<iDim;i++)
	    dMagQVec += rgQ[i]*rgQ[i];

	dMagQVec = sqrt(dMagQVec);

	if (FP_EQ_EPS(0.0, dMagQVec, 1e-10)) {
	    fprintf(stderr, "\npower_iterate(): zero magnitude e-vector.\n");
	    exit(-3);
	}

	VEC_SCALAR_OP( rgQ, rgQ, /, dMagQVec, iDim );

	MATVECMULT( rgdWork, rgrgB, rgQ, iDim, iDim );
	/* calc dot prod of rgQ and rgTemp to get new lambda */
	for (dNumer=0.0, i=0; i<iDim; i++)
	    dNumer += rgdWork[i]*rgQ[i];

	 /* MATVECMULT( rgdWork, rgrgD, rgQ, iDim, iDim ); */
	/* calc dot prod of rgQ and rgTemp to get new lambda */
	for (dDenom=0.0, i=0; i<iDim; i++)
	    dDenom += rgQ[i]*rgQ[i];
  
	dNewEst = dNumer/dDenom;

	if (!base_point_case) {
	  dS = rgQ[1]/rgQ[0]; 
	  dT = rgQ[m]/rgQ[0]; /* Again 4 used ??? */
	}
	*iteration_count = (*iteration_count) + 1;

         /* printf("Lambda %lf, S %lf, T %lf\n", dNewEst, inverse_power_pt(dS), inverse_power_pt(dT)); */
	    
/*
    } while ((!FP_EQ_EPS( dOldEst, dNewEst, 1e-3)) && 
	     (*iteration_count < ITER_LIMIT));
*/
    } while (((!FP_EQ_EPS( dOldEst, dNewEst, 1e-5)) || 
	     (!(FP_EQ_EPS( dOldS, dS, 1e-3))) || 
	     (!(FP_EQ_EPS( dOldT, dT, 1e-3)))) &&
	     (*iteration_count < ITER_LIMIT));

    *pdEVal = dNewEst;
    VEC_ASN_OP( rgEVec, =, rgQ, iDim );

    myfree(rgQ); myfree(rgdWork);
    myfree(LastMat); myfree(ipvt);
    myfree(rgK);
    FREEMAT(rgrgA, iDim, iDim);
    FREEMAT(rgrgTmp, iDim, iDim);
    FREEMAT(rgrgU, iDim, iDim);
    FREEMAT(rgrgV, iDim, iDim);
}

void
generalized_power_iterate( rgrgB, rgrgD, rgEVec, iDim, iDegree, dGuess, pdEVal, m, iteration_count, base_point_case )
double      **rgrgB, **rgrgD;
double      *rgEVec;
int         iDim, iDegree, m;
double      dGuess, *pdEVal;
int	    *iteration_count;
int base_point_case;
{

    double *rgQ, *rgdWork, *ipvt;
    double **rgrgA, *LastMat;
    double **rgrgTmp, **rgrgU, **rgrgV, *rgK;
    int    i,j,k,l;
    double dOldEst, dNewEst, dRCond, dMagQVec;
    double dS, dT, dOldS, dOldT;
    double dNumer, dDenom, dAlpha1, dAlpha2;
    int dim, base;

    base = iDim - iDegree;
    dim = iDim / iDegree;

    /* allocate/initialize the matrices */
    ALLOCMAT(rgrgA, iDim, iDim, double);
    ALLOCN(LastMat, DOUBLE, iDegree * iDegree);
    ALLOCMAT(rgrgTmp, iDim, iDim, double);
    ALLOCMAT(rgrgU, iDim, iDim, double);
    ALLOCMAT(rgrgV, iDim, iDim, double);
    ALLOCN(rgK, DOUBLE, iDim);
    ALLOCN(rgdWork, DOUBLE, iDim);
    ALLOCN(rgQ, DOUBLE, iDim);
    ALLOCN(ipvt, DOUBLE, iDegree);

    /* perform inverse power iteartions as per
       Golub and Van Loan, pg 383 */

    for (i=0;i<iDim;i++)
	rgQ[i]=rgEVec[i];

    /* normalize and use as next estimate */
    for(i=0, dMagQVec=0.0;i<iDim;i++)
	dMagQVec += rgQ[i]*rgQ[i];
    
    dMagQVec = sqrt(dMagQVec);
    
    if (FP_EQ_EPS(0.0, dMagQVec, 1e-10)) {
	fprintf(stderr, "\npower_iterate(): zero magnitude e-vector.\n");
	exit(-3);
    }
    
    VEC_SCALAR_OP( rgQ, rgQ, /, dMagQVec, iDim );

    /* calc (D*lambda - B) and check if it singular to proceed */
    for (i = 0; i < iDim; i++)
    for (j = 0; j < iDim; j++)
      rgrgTmp[i][j] = rgrgD[i][j]*dGuess - rgrgB[i][j];
    call_svd(rgrgTmp, iDim, iDim, rgrgU, rgK, rgrgV);
    if (FP_EQ_EPS(rgK[iDim-1], 0.0, 1e-10)) {
      *iteration_count = 0;
      *pdEVal = dGuess;
      VEC_ASN_OP(rgEVec, =, rgQ, iDim );
      return;
    }

    /* calc (D*lambda - B) */
    for (j = 0; j < dim; j++)
    {
        l = j*iDegree;
        for ( i = 0; i < iDegree; i++)
            for ( k = 0; k < iDegree; k++)
        {
            rgrgA[base + i][l + k] = rgrgD[base + i][l+k]*dGuess -  
			     	     rgrgB[base + i][l + k];
       }
    }

    if (dim > 1) {
      dAlpha1 = rgrgD[0][0]*dGuess - rgrgB[0][0];
      dAlpha2 = rgrgD[0][iDegree]*dGuess - rgrgB[0][iDegree];
    } else 
      dAlpha1 = dAlpha2 = 0.0;

    Decompose(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2, ipvt);

/* #endif */

    *iteration_count = 0;
    dNewEst = dGuess;
    if (!base_point_case) {
      dS = rgQ[1]/rgQ[0]; 
      dT = rgQ[m]/rgQ[0]; /*	 How did they get 4. In general it should
				 be m,	 for bicubic it should be 3 */
    } else {
      dS = dT = 1.0; /* default value */
    }
    
    do {
	dOldEst = dNewEst;
        dOldS = dS;
	dOldT = dT;
        
	/* do rgQ = rgrgD*rgQ */
        MATVECMULT(rgdWork, rgrgD, rgQ, iDim, iDim);
	VEC_ASN_OP(rgQ, =, rgdWork, iDim);


	/* solve for rgQ */
        Solve(rgrgA, LastMat, iDim, iDegree, dAlpha1, dAlpha2, rgQ, ipvt);

	/* normalize and use as next estimate */
	for(i=0, dMagQVec=0.0;i<iDim;i++)
	    dMagQVec += rgQ[i]*rgQ[i];

	dMagQVec = sqrt(dMagQVec);

	if (FP_EQ_EPS(0.0, dMagQVec, 1e-10)) {
	    fprintf(stderr, "\npower_iterate(): zero magnitude e-vector.\n");
	    exit(-3);
	}

	VEC_SCALAR_OP( rgQ, rgQ, /, dMagQVec, iDim );

	MATVECMULT( rgdWork, rgrgB, rgQ, iDim, iDim );
	/* calc dot prod of rgQ and rgTemp to get new lambda */
	for (dNumer=0.0, i=0; i<iDim; i++)
	    dNumer += rgdWork[i]*rgQ[i];

	MATVECMULT( rgdWork, rgrgD, rgQ, iDim, iDim );
	/* calc dot prod of rgQ and rgTemp to get new lambda */
	for (dDenom=0.0, i=0; i<iDim; i++)
	    dDenom += rgdWork[i]*rgQ[i];
  
	dNewEst = dNumer/dDenom;

	if (!base_point_case) {
	  dS = rgQ[1]/rgQ[0]; 
	  dT = rgQ[m]/rgQ[0]; /* Again 4 used ??? */
	}
	*iteration_count = (*iteration_count) + 1;

         /* printf("Lambda %lf, S %lf, T %lf\n", dNewEst, inverse_power_pt(dS), inverse_power_pt(dT)); */
	    
/*
    } while ((!FP_EQ_EPS( dOldEst, dNewEst, 1e-3)) && 
	     (*iteration_count < ITER_LIMIT));
*/
    } while (((!FP_EQ_EPS( dOldEst, dNewEst, 1e-5)) || 
	     (!(FP_EQ_EPS( dOldS, dS, 1e-3))) || 
	     (!(FP_EQ_EPS( dOldT, dT, 1e-3)))) &&
	     (*iteration_count < ITER_LIMIT));

    *pdEVal = dNewEst;
    VEC_ASN_OP( rgEVec, =, rgQ, iDim );

    myfree(rgQ); myfree(rgdWork);
    myfree(LastMat); myfree(ipvt);
    myfree(rgK);
    FREEMAT(rgrgA, iDim, iDim);
    FREEMAT(rgrgTmp, iDim, iDim);
    FREEMAT(rgrgU, iDim, iDim);
    FREEMAT(rgrgV, iDim, iDim);
}

/*
  Uses the structure of Block Companion Matrices to perform LU decomposition
  The matrix has been defined in the global variable C[DIM1][DIM1].
*/
  Decompose(A,LM,order,degree,alpha1,alpha2,ipvt)
    double **A, *LM;
    int order,degree,*ipvt;
    double alpha1,alpha2;

{
    int i,j,k,dim,base,lda,lda1;
    double factor, rcond, *vec;

    ALLOCN(vec, DOUBLE, degree);
    dim = order/degree; /* dim corresponds to the degree of the matrix polynomial */

    base = order - degree;
    factor = alpha2/alpha1;
    for (i = 1; i < dim; i++)
    {
         for (j = 0; j < degree; j++)
             for (k = 0; k < degree; k++)
             {
                 A[base+j][i*degree + k] -= factor * A[base+j][(i-1)*degree +
k];
             }
    }
/* After the previous loop is completed, C basically contains a lower
   triangular matrix except for the bottom right degree X degree matrix.
   And that is precisely LM, which is computed below                     */

    for (i = j = 0; j < degree; j++)
       for (k = 0; k < degree; k++)
       {
            LM[i++] = A[base + j][base + k];
       }
    lda = degree;
    lda1 = degree;

/* Perform decomposition on LM - efficiently done by the Fortran code */
#ifdef __SGI
    dgeco_(LM,&lda,&lda1,ipvt,&rcond,vec);
#else
    dgeco(LM,&lda,&lda1,ipvt,&rcond,vec);
#endif

    myfree(vec);
}

/*  Given the LU decomposition, solve for the variables. The RHS vector is in
   vec[]
*/

Solve(A,LM,order,degree,alpha1,alpha2,vec,ipvt)
    double **A, *LM;
    int order,degree,*ipvt;
    double alpha1,alpha2;
    double *vec;

{
    int i,j,k,dim,base,lda,job;
    double factor, *vec1;

    ALLOCN(vec1, DOUBLE, degree);

    base = order - degree;
    factor = alpha2/alpha1;

/* While solving for L U x = b, we first solve for L z = b and then U x = z.
   We are now trying to compute z. The following loop computes first base
   entries easily. */
   for (i = 0; i < base; i++)
   {
        vec[i] = vec[i] / alpha1;
   }

/* Now formulate the problem of LM * last degree elements of z = last
   degree elements of b */
   for (i = base; i < order; i++)
   {
       for (j = 0; j < base; j++)
           vec[i] = vec[i] - A[i][j] * vec[j];
       vec1[i-base] = vec[i];
   }

   job = 1;
   lda = degree;

/* Solve for last degree elements of z using Fortran code */
#ifdef __SGI
   dgesl_(LM,&lda,&degree,ipvt,vec1,&job);
#else
   dgesl(LM,&lda,&degree,ipvt,vec1,&job);
#endif

   for (i = 0; i < degree; i++)
       vec[i+base] = vec1[i];

/* Finally solve now for U x = z. U is very simple. x found directly */
   for (i = base - 1; i >=0; i--)
   {
       vec[i] = vec[i] -  factor * vec[i+degree];
   }
   myfree(vec1);
}

