#ifdef __GL
#include <gl/gl.h>
#include <gl/device.h>
#include <gl/sphere.h>
#include "color.h"

long vert[4][2] = {{200,200}, {200,400},{400,400},{400,200}};

/*
#define XYZ_HEIGHT 1024
#define XYZ_WIDTH  1280
#define XOFF       0
#define YOFF       0
*/

#define XYZ_HEIGHT 600
#define XYZ_WIDTH  1000
#define XOFF       50
#define YOFF       50

#define HSPACING   20
#define VSPACING   50
#define PIX_RADIUS 1
#define UVST_SIZE  256

#define RGB_RED    0xff0000

void setupWindows();
long openwindow( int, int, int, int, Colorindex, char *, int );
void draw_uv_start_point(double, double);
void draw_st_start_point(double, double);
void draw_uv_point(double, double);
void draw_st_point(double, double);

double cpointsS1[4][4][3];
double cpointsS2[4][4][3];

/* window ids */
long giduv, gidst, gidxyz;

double ctlpoints[4][4][3];
double knots[8]={0.0,0.0,0.0,0.0,1.0,1.0,1.0,1.0};

init_surface()
{
    int u, v;
    for(u=0;u<4;u++) {
	for(v=0;v<4;v++) {
	    ctlpoints[u][v][0]=2.0*((double)u-1.5);
	    ctlpoints[u][v][1]=2.0*((double)v-1.5);

	    if ((u==1||u==2) &&(v==1||v==2))
		ctlpoints[u][v][2]=3.0;
	    else
		ctlpoints[u][v][2]=-3.0;
	}
    }
}

/*
long blackvec[3]={0,0,0};
long purplevec[3]={150,0,150};
*/
long bluevec[3]={0,0,200};
long backgrndvec[3]={64,64,64};

static float purplePlastic[] = {
    AMBIENT, .1,.1,.1,
    DIFFUSE, .8, .12, .75,
    SPECULAR, .8, .8, .8,
    SHININESS, 30,
    LMNULL
    };

static float greenPlastic[] = {
    AMBIENT, .1,.1,.1,
    DIFFUSE, .08, .82, .15,
    SPECULAR, .8, .8, .8,
    SHININESS, 30,
    LMNULL
    };

static float bluePlastic[] = {
    AMBIENT, .1,.1,.1,
    DIFFUSE, .08, .12, .85,
    SPECULAR, .8, .8, .8,
    SHININESS, 30,
    LMNULL
    };

static float myModel[] = {
    AMBIENT, .4, .4, .4,
    LOCALVIEWER, 1.0,
    LMNULL
    };

static float myLight[] = {
    AMBIENT, .0, .0, .0,
    LCOLOR, 1.0, 1.0, 1.0,
    POSITION, 0.0, -1.0, 0.0, 0.0,
    SPOTLIGHT, 0.0, 180.0,
    SPOTDIRECTION, 0.0, 0.0, -1.0,
    LMNULL
    };

static float myLight2[] = {
    AMBIENT, .0, .0, .0,
    LCOLOR, 1.0, 1.0, 1.0,
    POSITION, -1.0, 0.0, 0.0, 0.0,
    SPOTLIGHT, 0.0, 180.0,
    SPOTDIRECTION, 0.0, 0.0, -1.0,
    LMNULL
    };

int incpitch=0, incyaw=0;
Coord zChange=0, yChange=0, xChange=0;

long init_ctrace_win(x0,y0,x1,y1,bkgrnd,win_name)
int x0, x1, y0, y1;
Colorindex bkgrnd;
char *win_name;
{
POINT tmp1, tmp2;
    long xorig, yorig, xsize, ysize;
    long gid;
    float vert0[3], vert1[3];
    float vv[3];
    int i, rot_deg;
    int sc_rot_trans;
    int sc, rot, trans;
    float planexy[4][3] = {
      {0.0, 0.0, 0.0},
      {1.4, 0.0, 0.0},
      {1.4, 1.4, 0.0},
      {0.0, 1.4, 0.0}
    };

    float planeyz[4][3] = {
      {0.0, 0.0, 0.0},
      {0.0, 1.4, 0.0},
      {0.0, 1.4, 2.0},
      {0.0, 0.0, 2.0}
    };

    float planezx[4][3] = {
      {0.0, 0.0, 0.0},
      {1.4, 0.0, 0.0},
      {1.4, 0.0, 2.0},
      {0.0, 0.0, 2.0}
    };

    /* set up graphics environment */
    prefposition(x0, x1, y0, y1);
    FOREGROUND;
    gid = winopen(win_name);
    color(bkgrnd);
    getorigin(&xorig, &yorig);
    getsize(&xsize, &ysize);
    singlebuffer();
    RGBmode();
    gconfig();
    lsetdepth(0, 0x7FFFFF);
    zbuffer(TRUE);
    mmode(MPROJECTION);
    perspective(900, xsize/(float)ysize, 0.01, 10000.0);
    mmode(MVIEWING);
    def_light_calc();
    use_light_calc();
    shademodel(GOURAUD);
    c3f(purplevec);       /* clear the frame */
    clear();
    zclear();
/*
    printf("input eye coords: ");
    scanf("%lf", &vv[0]);
    scanf("%lf", &vv[1]);
    scanf("%lf", &vv[2]);
*/
      /* lookat(2.5625, -2.075, 0.1125, 0.0, 0.0, 0.0, 0); */
     lookat(0.7, -3.0, 0.0, 0.5, 0.0, 0.0, 0);
      /* lookat(vv[0], vv[1], vv[2], 0.0, 0.0, 0.0, 0); */

    lmbind(MATERIAL, 9);
    vert0[0] = 0.0;
    vert0[1] = 0.0;
    vert0[2] = 0.0;
    vert1[0] = 1.4;
    vert1[1] = 0.0;
    vert1[2] = 0.0;
    vv[0] = 0.0;
    vv[1] = 1.0;
    vv[2] = 0.0;
    n3f(vv);
    scale(2.4, 2.4, 2.4);
    translate(-0.4, -0.4, 0.0);
     rotate(-30, 'x');
    rotate(400, 'y');
/*
    scale(3.0, 4.5, 3.0);
    rotate(450,'y');
    translate(-1.0, -0.8, 0.0);
    rotate(-420, 'x');
    translate(0.5, 0.4, 0.0);
    scale(1.2, 1.5, 0.5);
*/
    bgnline();
    v3f(vert0);
    v3f(vert1);
    endline();
    vert1[0] = 0.0;
    vert1[1] = 2.9;
    vert1[2] = 0.0;
    bgnline();
    v3f(vert0);
    v3f(vert1);
    endline();
    vert1[0] = 0.0;
    vert1[1] = 0.0;
    vert1[2] = 2.0;
    bgnline();
    v3f(vert0);
    v3f(vert1);
    endline();
    vert0[0] = 1.0;
    vert0[1] = 0.0;
    vert0[2] = 0.0;
    vert1[0] = 1.0;
    vert1[1] = 1.0;
    vert1[2] = 0.0;
    bgnline();
    v3f(vert0);
    v3f(vert1);
    endline();
    vert0[0] = 0.0;
    vert0[1] = 1.0;
    vert0[2] = 0.0;
    bgnline();
    v3f(vert0);
    v3f(vert1);
    endline();

    cmov(0.5, -0.05, 0.0);
    charstr("u");
    cmov(-0.57, 2.7, 0.0);
    charstr("Rv");
    cmov(-0.09, -0.05, 1.5);
    charstr("Iv");
    cmov(0.05, -0.05, 0.0);
    charstr("0");
    cmov(1.0, -0.05, 0.0);
    charstr("1");
    cmov(-0.2, 1.0, 0.0);
    charstr("1");
    cmov(-0.05, -0.05, 1.0);
    charstr("1");

     /* lmbind(MATERIAL, 12); */
/*
    bgnpolygon();
    for (i=0; i<4; i++)
       v3f(planexy[i]);
    endpolygon();
    bgnpolygon();
    for (i=0; i<4; i++)
       v3f(planeyz[i]);
    endpolygon();
    bgnpolygon();
    for (i=0; i<4; i++)
       v3f(planezx[i]);
    endpolygon();
*/
    return gid;
}

long openwindow( int x0, int y0, int x1, int y1, Colorindex bkgrnd, 
		char *pchName, int uv )
{
    long gid;
    float vert[2];
    int i;

    prefposition(x0,x1,y0,y1);
    gid = winopen(pchName);
/*
    RGBmode();
    gconfig();
    c3f(purplevec);
*/
    color(BLACK);
    clear();
    color(WHITE);
    bgnline();
    for (i=0; i<UVST_SIZE; i++) {
      vert[0] = i;
      vert[1] = 0.1 * UVST_SIZE;
      v2f(vert);
    }
    endline();
    bgnline();
    for (i=0; i<UVST_SIZE; i++) {
      vert[0] = 0.1 * UVST_SIZE;
      vert[1] = i;
      v2f(vert);
    }
    endline();
    bgnline();
    for (i=0; i<0.85*UVST_SIZE; i++) {
      vert[0] = i + 0.1*UVST_SIZE;
      vert[1] = 0.95 * UVST_SIZE;
      v2f(vert);
    }
    endline();
    bgnline();
    for (i=0; i<0.85*UVST_SIZE; i++) {
      vert[0] = 0.95 * UVST_SIZE;
      vert[1] = i + 0.1*UVST_SIZE;
      v2f(vert);
    }
    endline();

    cmov2(0.13*UVST_SIZE, 0.05*UVST_SIZE);
    charstr("0");
    cmov2(0.93*UVST_SIZE, 0.05*UVST_SIZE);
    charstr("1");
    cmov2(0.05*UVST_SIZE, 0.94*UVST_SIZE);
    charstr("0");
    cmov2(0.97*UVST_SIZE, 0.94*UVST_SIZE);
    charstr("1");

    cmov2(0.525*UVST_SIZE, 0.05*UVST_SIZE);
    if (uv) charstr("u");
    else charstr("s");
    cmov2(0.06*UVST_SIZE, 0.525*UVST_SIZE);
    if (uv) charstr("v");
    else charstr("t");
    return gid;
}

short lastpointuv[2]={0,0};
short lastpointst[2]={0,0};

void draw_uv_point(double u, double v)
{
    short newpoint[2];

    newpoint[0] = 0.1*UVST_SIZE + u*0.85*UVST_SIZE;
    newpoint[1] = 0.1*UVST_SIZE + v*0.85*UVST_SIZE;

    winset(giduv);
    color(BLUE);
    circfs(newpoint[0], newpoint[1], PIX_RADIUS);
/*
    bgnpoint();
    v2s(newpoint);
    endpoint();
    printf("\nuv Drew from %d %d to %d %d.\n",lastpointuv[0],lastpointuv[1],
    newpoint[0],newpoint[1]);
    lastpointuv[0]=newpoint[0];
    lastpointuv[1]=newpoint[1];
*/
    gflush();
    winset(gidxyz);
}

void setupWindows( c_trace )
int c_trace;
{
    foreground();
    
    gidxyz=init_gl(XOFF,YOFF,
		     XOFF+XYZ_WIDTH,YOFF+XYZ_HEIGHT,
		     WHITE,"xyz Space");
    winset(gidxyz);
}

float lst_point[3];

void dr_uvi_point(double u, double vr, double vi)
{
    float sph[4];
    static int i=0;
    float cood[3];

    winset(giduv);
    sph[0] = u;
    sph[1] = vr;
    sph[2] = vi;
    if (vr < 0.5) sph[3] = 0.008;
    else if (vr < 1.0) sph[3] = 0.013;
    else if (vr < 2.0) sph[3] = 0.025;
    else sph[3] = 0.03;
    if (vi != 0.0) {
      lmbind(MATERIAL, 13);
      if (i != 0) {
	i++;
	cood[0] = u;
	cood[1] = vr;
	cood[2] = vi;
	bgnline();
	v3f(lst_point);
	v3f(cood);
	endline();
      }
    }
    else lmbind(MATERIAL, 2);

    sphdraw(sph);
    lst_point[0] = u;
    lst_point[1] = vr;
    lst_point[2] = vi;
    winset(gidxyz);
}

void draw_uv_start_point(double u, double v)
{
    lastpointuv[0] = u * UVST_SIZE;
    lastpointuv[1] = v * UVST_SIZE;
}

void draw_st_point(double s, double t)
{
    short newpoint[2];

    newpoint[0] = 0.1*UVST_SIZE + s*0.85*UVST_SIZE;
    newpoint[1] = 0.1*UVST_SIZE + t*0.85*UVST_SIZE;

    winset(gidst);
    color(RED);
    circfs(newpoint[0], newpoint[1], PIX_RADIUS);
/*
    bgnpoint();
    v2s(newpoint);
    endpoint();
    printf("\nst Drew from %d %d to %d %d.\n",lastpointst[0],lastpointst[1],
    newpoint[0],newpoint[1]);
    lastpointst[0]=newpoint[0];
    lastpointst[1]=newpoint[1];
*/
    gflush();
    winset(gidxyz);
}

void draw_st_start_point(double s, double t)
{
    lastpointst[0] = s * UVST_SIZE;
    lastpointst[1] = t * UVST_SIZE;
}

/*
void sgiLoop( void )
{

    Device dev;
    short val;
    int fContinue=1, fRedraw=1;

    qdevice(QKEY);
    qdevice(ESCKEY);
    qdevice(UPARROWKEY);
    qdevice(DOWNARROWKEY);
    qdevice(LEFTARROWKEY);
    qdevice(RIGHTARROWKEY);
    qdevice(JKEY);
    qdevice(KKEY);
    qdevice(HKEY);
    qdevice(LKEY);
    qdevice(IKEY);
    qdevice(MKEY);
    qreset();

    while (fContinue) {
	switch (dev = qread(&val)) {
	  case QKEY:
	  case ESCKEY:
	    fContinue=0;
	    break;
	  case UPARROWKEY:
	    incpitch -=10;
	    fRedraw=1;
	    break;
	  case DOWNARROWKEY:
	    incpitch +=10;
	    fRedraw=1;
	    break;
	  case RIGHTARROWKEY:
	    incyaw +=10;
	    fRedraw=1;
	    break;
	  case LEFTARROWKEY:
	    incyaw -=10;
	    fRedraw=1;
	    break;
	  case JKEY:
	    zChange +=10;
	    fRedraw=1;
	    break;
	  case KKEY:
	    zChange -=10;
	    fRedraw=1;
	    break;
	  case IKEY:
	    yChange +=10;
	    fRedraw=1;
	    break;
	  case MKEY:
	    yChange -=10;
	    fRedraw=1;
	    break;
	  case LKEY:
	    xChange +=10;
	    fRedraw=1;
	    break;
	  case HKEY:
	    xChange -=10;
	    fRedraw=1;
	    break;
	  case REDRAW:
	    if (val=gidxyz) {
		reshapeviewport();
		fRedraw=1;
	    }
	}

	if (fRedraw) {
	    redrawXYZ();
	    fRedraw=0;
	}
    }
}
*/
#endif
