#include "defs.h"
#include "complex.h"
#include "lp.h"

#define  ORD 	54
#define  NEW_EPS 5.0*1e-5

extern double	power_pt();
extern double   inverse_power_pt();
extern double   trans_inverse_power_pt();
extern COMPLEX  trans_inv_cxpower_pt();
extern double	RefinePoint();
extern double	RefinePoint1D();
extern TENSOR_PATCH pow_tp0, pow_tp1;

extern int unchanged;
extern double unchanged_val;

int solindex;
double SOLUTION[ORD][3];

int special_plane_case();

EXTERN_LINALG

/* get all start-points on the open & closed components */
int
getComponents( pplp4DStartingPoints, plp4CPXStartingPoints, fdet1D_uv,
	       ndet1D_uv, tpPatch0, tpPatch1, 
               box, cbox, row_perm, base_case, ctrace )
LIST_POINT4D     **pplp4DStartingPoints;
LIST_POINT4CPX   **plp4CPXStartingPoints;
NDETERMINANT_1D  ndet1D_uv, fdet1D_uv;
TENSOR_PATCH     tpPatch0, tpPatch1;
BOUNDINGBOX 	 box, *cbox;
int		 *row_perm;
int		 base_case, ctrace;
{
    *pplp4DStartingPoints = NULL; 
    *plp4CPXStartingPoints = NULL;
    if (getStartPoints(pplp4DStartingPoints, 
		       plp4CPXStartingPoints, fdet1D_uv, ndet1D_uv, 
		       tpPatch0, tpPatch1, box, cbox, 
                       row_perm, base_case, ctrace) == -1)
      return -1;
     /* printList(*pplp4DStartingPoints); */
    return 0;
}


/* right now this only finds all the start points on one patch. 
   should eventually  be done for both patches
*/
getStartPoints( pplp4DStartingPoints, pplp4CPXStartingPoints, fdet1D_uv,
		ndet1D_uv, tpPatch0, tpPatch1, 
                box, cbox, row_perm, to_use_f, ctrace)
LIST_POINT4D     	**pplp4DStartingPoints;
LIST_POINT4CPX		**pplp4CPXStartingPoints;
NDETERMINANT_1D  	ndet1D_uv, fdet1D_uv;
TENSOR_PATCH     	tpPatch0;
TENSOR_PATCH     	tpPatch1;
BOUNDINGBOX		box, *cbox;
int			*row_perm;
int			to_use_f, ctrace;
{ 
   int		 	i, j;
   NDETERMINANT_1D	ndet1D_Tmp, ndet1D_u, ndet1D_v;
   NDETERMINANT_1D	fdet1D_Tmp, fdet1D_u, fdet1D_v;
   int			num_points, num_cpoints;
   POINT4D		*pdPoints;
   POINT4CPX		*pcpxPoints;
   double		parametrized_value;
   int			m, n, stts;

   m = tpPatch0.iDeg0;
   n = tpPatch0.iDeg1;

   i = (tpPatch1.iDeg0 > tpPatch1.iDeg1)? tpPatch1.iDeg0: tpPatch1.iDeg1;

   ALLOCN(pdPoints, POINT4D, i*ndet1D_uv.cRows);

   ALLOCN(pcpxPoints, POINT4CPX, i*ndet1D_uv.cRows);

   allocDeterminant_1D(&ndet1D_Tmp, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			tpPatch1.iDeg0, tpPatch1.iDeg1);

   allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			0, tpPatch1.iDeg0);

   allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			0, tpPatch1.iDeg1);

   if (to_use_f) {
     allocDeterminant_1D(&fdet1D_Tmp, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        tpPatch1.iDeg0, tpPatch1.iDeg1);

     allocDeterminant_1D(&fdet1D_u, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg0);

     allocDeterminant_1D(&fdet1D_v, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg1);
   }

   /* copy ndet1D_uv to ndet1D_Tmp */
   copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

   /* convert bernstein basis to power basis for each entry of Tmp */
   for(i = 0; i < ndet1D_Tmp.cRows; i++)
     for(j = 0; j < ndet1D_Tmp.cCols; j++)
       bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);

   if (to_use_f) {
     /* copy fdet1D_uv to fdet1D_Tmp */
     copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);

     /* convert bernstein basis to power basis for each entry of Tmp */
     for(i = 0; i < fdet1D_Tmp.cRows; i++)
       for(j = 0; j < fdet1D_Tmp.cCols; j++)
         bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
   }

#ifdef DO_LINPROG_CS
     if (boxes_overlap(box, cbox[2])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 2);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
#endif
       /* find all open component points for (u, 0) */
        /* printf("At curve-surf intersection V = 0\n"); */
       compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, 0.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, 0.0); 
       if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,U,power_pt(0.0), 
			     fdet1D_u, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addPointToList(pdPoints[i], pplp4DStartingPoints);
#ifdef DO_LINPROG_CS
       }
     }
   
     if (boxes_overlap(box, cbox[3])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 3);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
#endif
       /* find all open component points for (u, 1) */
        /* printf("At curve-surf intersection V = 1\n"); */
       compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, 1.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, 1.0); 
       if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,U,power_pt(1.0), 
			     fdet1D_u, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addPointToList(pdPoints[i], pplp4DStartingPoints);
#ifdef DO_LINPROG_CS
       }
     }
#endif
   
   if (!ctrace) {
#ifdef DO_LINPROG_CS
       if (boxes_overlap(box, cbox[0])) {
         stts = do_lin_prog(tpPatch0, tpPatch1, 0);
         if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
#endif
         /* find all open component points for (0, v) */
          /* printf("At curve-surf intersection U = 0\n"); */
         compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, 0.0); 
         if (to_use_f)
           compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, 0.0);
         if (find_start_points(ndet1D_v,&num_points,pdPoints,
			       &num_cpoints, pcpxPoints,V,power_pt(0.0), 
			       fdet1D_v, row_perm, m, n, to_use_f,
			       tpPatch0, tpPatch1, 0) == -1)
           return -1;
         for(i = 0; i < num_points; i++)
           addPointToList(pdPoints[i], pplp4DStartingPoints);
#ifdef DO_LINPROG_CS
         }
       }
#endif
   }
   else {
     compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, 0.0); 
     if (to_use_f)
       compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, 0.0);
     if (find_start_points(ndet1D_v,&num_points,pdPoints,
			   &num_cpoints, pcpxPoints,V,power_pt(0.0), 
			   fdet1D_v, row_perm, m, n, to_use_f,
			   tpPatch0, tpPatch1, 1) == -1)
     return -1;
     for(i = 0; i < num_points; i++)
       addPointToList(pdPoints[i], pplp4DStartingPoints);
     for(i = 0; i < num_cpoints; i++)
       addCPointToList(pcpxPoints[i], pplp4CPXStartingPoints);
   }
   
#ifdef DO_LINPROG_CS
     if (boxes_overlap(box, cbox[1])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 1);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
#endif
       /* find all open component points for (1, v) */
        /* printf("At curve-surf intersection U = 1\n"); */
       compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, 1.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, 1.0); 
       if (find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,V,power_pt(1.0), 
			     fdet1D_v, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addPointToList(pdPoints[i], pplp4DStartingPoints);
#ifdef DO_LINPROG_CS
       }
     }
#endif
   
#ifdef __FREE_ALL
   free_alloc_det_1D(&ndet1D_Tmp);
   free_alloc_det_1D(&ndet1D_u);
   free_alloc_det_1D(&ndet1D_v);
   if (to_use_f) {
     free_alloc_det_1D(&fdet1D_Tmp);
     free_alloc_det_1D(&fdet1D_u);
     free_alloc_det_1D(&fdet1D_v);
   }
   myfree(pdPoints);
   if (ctrace)
     myfree(pcpxPoints);
#endif
   return 0;
}

/* This finds only the starting points of open curves in uv domain
   that do not share the boundary
*/
int
getXtraComponents( pplp4DStartingPoints, plp4CPXStartingPoints, fdet1D_uv,
	       ndet1D_uv, tpPatch0, tpPatch1, 
               box, cbox, row_perm, base_case)
LIST_POINT4D     **pplp4DStartingPoints;
LIST_POINT4CPX   **plp4CPXStartingPoints;
NDETERMINANT_1D  ndet1D_uv, fdet1D_uv;
TENSOR_PATCH     tpPatch0, tpPatch1;
BOUNDINGBOX 	 box, *cbox;
int		 *row_perm;
int		 base_case;
{
    *pplp4DStartingPoints = NULL; 
    *plp4CPXStartingPoints = NULL;
    if (getXtraStartPoints(pplp4DStartingPoints, 
		       plp4CPXStartingPoints, fdet1D_uv, ndet1D_uv, 
		       tpPatch0, tpPatch1, box, cbox, 
                       row_perm, base_case) == -1)
      return -1;
     /* printList(*pplp4DStartingPoints); */
    return 0;
}


/* This finds only the starting points of open curves in uv domain
   that do not share the boundary
*/
getXtraStartPoints( pplp4DStartingPoints, pplp4CPXStartingPoints, fdet1D_uv,
		ndet1D_uv, tpPatch0, tpPatch1, 
                box, cbox, row_perm, to_use_f)
LIST_POINT4D     	**pplp4DStartingPoints;
LIST_POINT4CPX		**pplp4CPXStartingPoints;
NDETERMINANT_1D  	ndet1D_uv, fdet1D_uv;
TENSOR_PATCH     	tpPatch0;
TENSOR_PATCH     	tpPatch1;
BOUNDINGBOX		box, *cbox;
int			*row_perm;
int			to_use_f;
{ 
   int		 	i, j;
   NDETERMINANT_1D	ndet1D_Tmp, ndet1D_u, ndet1D_v;
   NDETERMINANT_1D	fdet1D_Tmp, fdet1D_u, fdet1D_v;
   int			num_points, num_cpoints;
   POINT4D		*pdPoints;
   POINT4CPX		*pcpxPoints;
   double		parametrized_value;
   int			m, n, stts;

   m = tpPatch0.iDeg0;
   n = tpPatch0.iDeg1;

   i = (tpPatch1.iDeg0 > tpPatch1.iDeg1)? tpPatch1.iDeg0: tpPatch1.iDeg1;

   ALLOCN(pdPoints, POINT4D, i*ndet1D_uv.cRows);

   ALLOCN(pcpxPoints, POINT4CPX, i*ndet1D_uv.cRows);

   allocDeterminant_1D(&ndet1D_Tmp, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			tpPatch1.iDeg0, tpPatch1.iDeg1);

   allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			0, tpPatch1.iDeg0);

   allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols, 
			0, tpPatch1.iDeg1);

   if (to_use_f) {
     allocDeterminant_1D(&fdet1D_Tmp, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        tpPatch1.iDeg0, tpPatch1.iDeg1);

     allocDeterminant_1D(&fdet1D_u, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg0);

     allocDeterminant_1D(&fdet1D_v, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg1);
   }

   /* copy ndet1D_uv to ndet1D_Tmp */
   copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

   /* convert bernstein basis to power basis for each entry of Tmp */
   for(i = 0; i < ndet1D_Tmp.cRows; i++)
     for(j = 0; j < ndet1D_Tmp.cCols; j++)
       bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);

   if (to_use_f) {
     /* copy fdet1D_uv to fdet1D_Tmp */
     copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);

     /* convert bernstein basis to power basis for each entry of Tmp */
     for(i = 0; i < fdet1D_Tmp.cRows; i++)
       for(j = 0; j < fdet1D_Tmp.cCols; j++)
         bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
   }

     if (boxes_overlap(box, cbox[2])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 2);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
       /* find all open component points for (u, 0) */
        /* printf("At curve-surf intersection V = 0\n"); */
       compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, 0.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, 0.0); 
       if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,U,power_pt(0.0), 
			     fdet1D_u, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addXtraPointToList(pdPoints[i], pplp4DStartingPoints);
       }
     }
   
     if (boxes_overlap(box, cbox[3])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 3);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
       /* find all open component points for (u, 1) */
        /* printf("At curve-surf intersection V = 1\n"); */
       compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, 1.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, 1.0); 
       if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,U,power_pt(1.0), 
			     fdet1D_u, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addXtraPointToList(pdPoints[i], pplp4DStartingPoints);
       }
     }
   
     if (boxes_overlap(box, cbox[0])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 0);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
         /* find all open component points for (0, v) */
          /* printf("At curve-surf intersection U = 0\n"); */
         compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, 0.0); 
         if (to_use_f)
           compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, 0.0);
         if (find_start_points(ndet1D_v,&num_points,pdPoints,
			       &num_cpoints, pcpxPoints,V,power_pt(0.0), 
			       fdet1D_v, row_perm, m, n, to_use_f,
			       tpPatch0, tpPatch1, 0) == -1)
           return -1;
         for(i = 0; i < num_points; i++)
           addXtraPointToList(pdPoints[i], pplp4DStartingPoints);
       }
     }
   
     if (boxes_overlap(box, cbox[1])) {
       stts = do_lin_prog(tpPatch0, tpPatch1, 1);
       if ((stts == INFEASIBLE) || (stts == AMBIGUOUS)) {
       /* find all open component points for (1, v) */
        /* printf("At curve-surf intersection U = 1\n"); */
       compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, 1.0); 
       if (to_use_f)
         compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, 1.0); 
       if (find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
			     pcpxPoints,V,power_pt(1.0), 
			     fdet1D_v, row_perm, m, n, to_use_f,
			     tpPatch0, tpPatch1, 0) == -1)
         return -1;
       for(i = 0; i < num_points; i++)
         addXtraPointToList(pdPoints[i], pplp4DStartingPoints);
       }
     }
   
#ifdef __FREE_ALL
   free_alloc_det_1D(&ndet1D_Tmp);
   free_alloc_det_1D(&ndet1D_u);
   free_alloc_det_1D(&ndet1D_v);
   if (to_use_f) {
     free_alloc_det_1D(&fdet1D_Tmp);
     free_alloc_det_1D(&fdet1D_u);
     free_alloc_det_1D(&fdet1D_v);
   }
   myfree(pdPoints);
#endif
   return 0;
}

/* this will evaluate the matrix ndet1D_uv_power for a curve */
/* variable is one of u or v whose curve has to be found.
   dVal is the value of the other variable in Bernstein Basis, ie:
   if variable == U, value of v is dVal and
   if variable == V, value of u is dVal
*/
compute_curve_1D(ndet1D_curve, ndet1D_uv_power, variable, dVal)
NDETERMINANT_1D		ndet1D_curve, ndet1D_uv_power;
int			variable;
double			dVal;
{
  int			i, j, k, l, n;
  double		*pow_val;

  if (variable == U)
  { n = ndet1D_uv_power.poly[0][0].iDeg1; 
    ALLOCN(pow_val, double, n+1);
    eval_ber(pow_val, dVal, n);
    for(i = 0; i <= n; i++)
      pow_val[i] /= combChoose(n, i);
  }
  else 
  { n = ndet1D_uv_power.poly[0][0].iDeg0; 
    ALLOCN(pow_val, double, n+1);
    eval_ber(pow_val, dVal, n);
    for(i = 0; i <= n; i++)
      pow_val[i] /= combChoose(n, i);
  }

  for(i = 0; i < ndet1D_uv_power.cRows; i++)
    for(j = 0; j < ndet1D_uv_power.cCols; j++)
    { if (variable == U)
      { for(k = 0; k <= ndet1D_uv_power.poly[i][j].iDeg0; k++)
	{ ndet1D_curve.poly[i][j].coeffs[0][k] = 0.0;
	  for(l = 0; l <= ndet1D_uv_power.poly[i][j].iDeg1; l++)
	    ndet1D_curve.poly[i][j].coeffs[0][k] += 
	    ndet1D_uv_power.poly[i][j].coeffs[k][l]*pow_val[l];
        }
      }
      else /* if (variable == V) */
      { for(k = 0; k <= ndet1D_uv_power.poly[i][j].iDeg1; k++)
	{ ndet1D_curve.poly[i][j].coeffs[0][k] = 0.0;
	  for(l = 0; l <= ndet1D_uv_power.poly[i][j].iDeg0; l++)
	    ndet1D_curve.poly[i][j].coeffs[0][k] += 
	    ndet1D_uv_power.poly[i][j].coeffs[l][k]*pow_val[l];
        }
      }
    }

  myfree(pow_val);
}

/* assumes that ndet1D_u has the equation of the curve in the 
   power basis in coeffs[0][d], for coeff of u^d.
   array pdPoints is assumed allocated outside this routine.
   start points output in power_basis.
   dConstParam input in power_basis.
*/
find_start_points(ndet1D_u, num_points, pdPoints, num_cpoints, 
		  pcpxPoints, iPolyIn, dConstParam, fdet1D_u, 
		  row_p, iDeg0Patch0, iDeg1Patch0, base, tp0, tp1, ct)
NDETERMINANT_1D		ndet1D_u, fdet1D_u;	
int			*num_points, *num_cpoints;
POINT4D			*pdPoints;
POINT4CPX		*pcpxPoints;
int			iPolyIn, iDeg0Patch0, iDeg1Patch0, base;
double			dConstParam;
TENSOR_PATCH		tp0, tp1;
int			*row_p, ct;
{ 
  int			i, j, k, l, iSizeResultant;
  int			p, q, r, prev_check;
  int			iConstIn = (iPolyIn == U)? V: U;
  double		**rgrgA_plus_B, **rgrgB;
  double		*dEReal, *dEImag, *dEBeta, **rgrgEVectors;
  double		Sval, Tval;
  double		fst, scnd, tp0X, tp0Y, tp0Z, tp1X, tp1Y, tp1Z;
  int			iDeg_Curve;
 int			iSize_C;
  int			chck_val;
  COMPLEX		tmp_comp, tmp_comp2, CdC;
  double		*nDCoord, **dirM;
  int			nIter;
  double		funcRet, err_dist, big_dist;
  double		a, b, c, d;
  double		discr; /* this contains c^2+d^2 if complex
				  number is c + id */

  iDeg_Curve = ndet1D_u.poly[0][0].iDeg1;
  iSizeResultant = ndet1D_u.cRows;

  iSize_C = iSizeResultant*iDeg_Curve;

  ALLOCMAT(rgrgB, iSize_C, iSize_C, double);
  ALLOCN(dEReal, double, iSize_C);
  ALLOCN(dEImag, double, iSize_C);
  ALLOCMAT(rgrgEVectors, iSize_C, iSize_C, double);
  ALLOCN(nDCoord, double, 3);
  ALLOCMAT(dirM, 3, 3, double);

  a = b = c = d = 0.0;
  while (FP_EQ_EPS(a, 0.0, 1e-4)) 
    a = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(b, 0.0, 1e-4)) 
    b = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(c, 0.0, 1e-4))
    c = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(d, 0.0, 1e-4))
    d = 0.7 + 0.3*drand48();

/* new code added */
#ifdef ALG_PRUNE
  *num_cpoints = 0;
  *num_points = 0;
  prune(ndet1D_u, iDeg0Patch0, iDeg1Patch0, iPolyIn, dConstParam, tp0,
	tp1, fdet1D_u, row_p, base, 0.0, 1.0, num_points, pdPoints);
  
#else
  if (1) /* (form_companion_matrix(ndet1D_u,rgrgB,iSize_C,
				      a, b, c, d) == -1) */ {
    ALLOCMAT(rgrgA_plus_B, iSize_C, iSize_C, double);

    if (form_power_companion_matrix(ndet1D_u,rgrgB,rgrgA_plus_B,iSize_C) == -1)
      return -1;

    /* find the eigenvalues of C */
    ALLOCN(dEBeta, double, iSize_C);
    call_rgg(rgrgB, rgrgA_plus_B, iSize_C, iSize_C, dEReal, 
	     dEImag, dEBeta, rgrgEVectors, 1);
    for (i=0; i<iSize_C; i++) 
      if (!FP_EQ(dEBeta[i], 0.0)) {
        dEReal[i] /= dEBeta[i];
        dEImag[i] /= dEBeta[i];
      }
      else dEReal[i] = dEImag[i] = ERROR_VAL;
    FREEMAT(rgrgA_plus_B, iSize_C, iSize_C);
    myfree(dEBeta);
  } else {
    call_rg(rgrgB, iSize_C, iSize_C, dEReal, dEImag, rgrgEVectors, 1); 
    for (i=0; i<iSize_C; i++)
      if (FP_EQ(dEImag[i], 0.0)) 
	 /* dEReal[i] = trans_inverse_power_pt(dEReal[i], a, b, c, d); */
	dEReal[i] = (a*dEReal[i] + b) / ((a+c)*dEReal[i] + b + d);
      else {
	tmp_comp.Re = a*dEReal[i] + b;
	tmp_comp.Im = a*dEImag[i];
	tmp_comp2.Re = (a+c)*dEReal[i] + b + d;
	tmp_comp2.Im = (a+c)*dEImag[i];
	 /* tmp_comp = trans_inv_cxpower_pt(tmp_comp, a, b, c, d); */
 	tmp_comp = divid(tmp_comp, tmp_comp2);
	dEReal[i] = tmp_comp.Re;
	dEImag[i] = tmp_comp.Im;
      }
  }

  /* return the real eigenvalues as start points (u, v) and find 
    from their corresponding eigenvectors the points (s, t) on
     the other patch.
  */

  chck_val = -2; /* this has been added to skip one value of i
                    if a complex eigenval is found */
  for(i = j = k = 0; i < iSize_C; i++) {
      
      if (FP_EQ(dEImag[i], 0.0) && (dEReal[i] >= -NEW_EPS) &&
	  (dEReal[i] <= 1+NEW_EPS)) {
	prev_check = 0;
	for (l=0; l<j; l++) {
	  if (FP_EQ_EPS(dEReal[i], 
	      inverse_power_pt(pdPoints[l].dCoord[iPolyIn]),1e-3))
	    prev_check = 1;
	}
	if (prev_check) continue;

	pdPoints[j].dCoord[iPolyIn] = power_pt(dEReal[i]);
	pdPoints[j].dCoord[iConstIn] = dConstParam;
	if (tp0.iDeg0 == 1 && tp0.iDeg1 == 1) {
	  fst = (iPolyIn == U) ? dEReal[i] :
				  inverse_power_pt(dConstParam);
	  scnd = (iPolyIn == V) ? dEReal[i] :
				  inverse_power_pt(dConstParam);
	  EvalPointOnPatch(tp1, fst, scnd, &tp1X, &tp1Y, &tp1Z);
	  if (special_plane_case(&tp0, tp1X, tp1Y, tp1Z, &Sval, &Tval)
	      == -1)
	    continue;
	  pdPoints[j].dCoord[S] = power_pt(Sval);
	  pdPoints[j].dCoord[T] = power_pt(Tval);
	} else {
	  if (base || FP_EQ_EPS(rgrgEVectors[0][i], 0.0, 1e-2)) {
	    extract_st_val(tp0, tp1, dEReal[i], iPolyIn, 
		  inverse_power_pt(dConstParam), &Sval, &Tval);
	    pdPoints[j].dCoord[S] = power_pt(Sval);
	    pdPoints[j].dCoord[T] = power_pt(Tval);
	  }
	  else {
	    if (FP_EQ(rgrgEVectors[0][i], 0.0)) { 
	        pdPoints[j].dCoord[S] = HUGE_VAL;
	        pdPoints[j].dCoord[T] = HUGE_VAL;
	    }
	    else { 
	        pdPoints[j].dCoord[S] = 
				rgrgEVectors[1][i]/rgrgEVectors[0][i];
	        pdPoints[j].dCoord[T] = (iSize_C <= iDeg0Patch0)? 0.0:
	        rgrgEVectors[iDeg0Patch0][i]/rgrgEVectors[0][i];
	    }
	  }
	}
	if (iPolyIn == U) {
	  unchanged_val = scnd = inverse_power_pt(dConstParam);
	  unchanged = V;
	  fst = nDCoord[0] = dEReal[i];
	} else {
	  fst = unchanged_val = inverse_power_pt(dConstParam);
	  unchanged = U;
	  scnd = nDCoord[0] = dEReal[i];
	}
/*
	nDCoord[1] = inverse_power_pt(pdPoints[j].dCoord[S]);
	nDCoord[2] = inverse_power_pt(pdPoints[j].dCoord[T]);
	pow_tp0 = tp0;
	pow_tp1 = tp1;
	for (p=0; p<3; p++)
	for (q=0; q<3; q++) 
	  if (p == q) dirM[p][q] = 1.0;
	  else dirM[p][q] = 0.0;
	powell(nDCoord, dirM, 3, 0.01, &nIter, &funcRet, RefinePoint1D);
	if (funcRet < 1e-6) {
	  if (iPolyIn == U) {
	    pdPoints[j].dCoord[0] = power_pt(nDCoord[0]);
	    pdPoints[j].dCoord[1] = dConstParam;
	    pdPoints[j].dCoord[2] = power_pt(nDCoord[1]);
	    pdPoints[j].dCoord[3] = power_pt(nDCoord[2]);
	  } else {
	    pdPoints[j].dCoord[0] = dConstParam;
	    pdPoints[j].dCoord[1] = power_pt(nDCoord[0]);
	    pdPoints[j].dCoord[2] = power_pt(nDCoord[1]);
	    pdPoints[j].dCoord[3] = power_pt(nDCoord[2]);
	  }
	  j++;
	}
*/
	EvalPointOnPatch(tp0, inverse_power_pt(pdPoints[j].dCoord[S]),
			 inverse_power_pt(pdPoints[j].dCoord[T]),
			 &tp0X, &tp0Y, &tp0Z);
	EvalPointOnPatch(tp1, fst, scnd, &tp1X, &tp1Y, &tp1Z);
	big_dist = tp1X*tp1X + tp1Y*tp1Y + tp1Z*tp1Z;
	err_dist = (tp1X-tp0X)*(tp1X-tp0X) +
		   (tp1Y-tp0Y)*(tp1Y-tp0Y) +
	   	   (tp1Z-tp0Z)*(tp1Z-tp0Z);
	err_dist /= big_dist;
	if (FP_EQ_EPS(err_dist, 0.0, 1e-5)) j++;
/*
	if (FP_EQ_EPS(tp0X,tp1X,1e-3) && FP_EQ_EPS(tp0Y,tp1Y,1e-3) &&
	    FP_EQ_EPS(tp0Z,tp1Z,1e-3))
	  j++;
*/
      }
      else if (ct && !FP_EQ(dEImag[i],0.0)) { /* Imaginary part is non-zero */
	  /* if (i == chck_val+1) this is the conjugate eigenval */
	  if (dEImag[i] < 0.0) /* this is the conjugate eigenval */
	     continue;
	  else {
	      /* chck_val = i; */
	     tmp_comp.Re  = dEReal[i];
	     tmp_comp.Im  = dEImag[i];
	     prev_check = 0;
	     for (l=0; l<k; l++) {
	       if (FP_EQ_EPS(dEReal[i], 
	           	     pcpxPoints[l].cpxCoord[iPolyIn].Re,
			     1e-1) &&
	           FP_EQ_EPS(dEImag[i], 
	           	     pcpxPoints[l].cpxCoord[iPolyIn].Im,
			     1e-1)) 
	         prev_check = 1;
	     }
	     if (prev_check) continue;

	     pcpxPoints[k].cpxCoord[iPolyIn] = tmp_comp;
	     pcpxPoints[k].cpxCoord[iConstIn].Re = 
	       inverse_power_pt(dConstParam);
	     pcpxPoints[k++].cpxCoord[iConstIn].Im = 0.0;
/*
	     if (base) {
	       CdC.Re = dConstParam;
	       CdC.Im = 0.0;
	       printf("Should not have come here\n");
	     }
	     else {
	       if (FP_EQ(rgrgEVectors[0][i], 0.0) &&
		 	               FP_EQ(rgrgEVectors[0][i+1], 0.0)) { 
	           pcpxPoints[k].cpxCoord[S].Re = HUGE_VAL;
	           pcpxPoints[k].cpxCoord[S].Im = 0.0;
	           pcpxPoints[k].cpxCoord[T].Re = HUGE_VAL;
	           pcpxPoints[k].cpxCoord[T].Im = 0.0;
	       }
	       else { 
	  	   discr = rgrgEVectors[0][i] * rgrgEVectors[0][i] +
		  	   rgrgEVectors[0][i+1] * rgrgEVectors[0][i+1];
		   pcpxPoints[k].cpxCoord[S].Re = 
			   ( rgrgEVectors[1][i] * rgrgEVectors[0][i] +
			     rgrgEVectors[1][i+1] * rgrgEVectors[0][i+1] ) /
			   discr;
		   pcpxPoints[k].cpxCoord[S].Im =
			   ( rgrgEVectors[1][i+1] * rgrgEVectors[0][i] - 
		 	     rgrgEVectors[1][i] * rgrgEVectors[0][i+1] ) /
			   discr;
		   if (iSize_C <= iDeg0Patch0) {
		      pcpxPoints[k].cpxCoord[T].Re = 0.0;
		      pcpxPoints[k].cpxCoord[T].Im = 0.0;
		   }
		   else {
		      pcpxPoints[k].cpxCoord[T].Re =
		      ( rgrgEVectors[iDeg0Patch0][i] * rgrgEVectors[0][i] +
		      rgrgEVectors[iDeg0Patch0][i+1] * rgrgEVectors[0][i+1] ) /
		      discr; 
		      pcpxPoints[k].cpxCoord[T].Im =
		      ( rgrgEVectors[iDeg0Patch0][i+1] * rgrgEVectors[0][i] -
		        rgrgEVectors[iDeg0Patch0][i] * rgrgEVectors[0][i+1] ) /
		      discr;
		   }
	       }
	     }
*/
	  }
      }
  }
  *num_points = j; 
  *num_cpoints = k;
#endif

  myfree(dEReal); myfree(dEImag); 
  myfree(nDCoord);
  FREEMAT(rgrgEVectors, iSize_C, iSize_C);
  FREEMAT(rgrgB, iSize_C, iSize_C);
  FREEMAT(dirM, 3, 3);
  return 0;
}

/* assumes that ndet1D_u has the equation of the curve in the 
   power basis in coeffs[0][d], for coeff of u^d.
   array pdPoints is assumed allocated outside this routine.
   start points output in power_basis.
   dConstParam input in power_basis.
*/
find_st_points_use_prune(ndet1D_u, num_points, pdPoints, num_cpoints, 
		  pcpxPoints, iPolyIn, dConstParam, fdet1D_u, 
		  row_p, iDeg0Patch0, iDeg1Patch0, base, tp0, tp1)
NDETERMINANT_1D		ndet1D_u, fdet1D_u;	
int			*num_points, *num_cpoints;
POINT4D			*pdPoints;
POINT4CPX		*pcpxPoints;
int			iPolyIn, iDeg0Patch0, iDeg1Patch0, base;
double			dConstParam;
TENSOR_PATCH		tp0, tp1;
int			*row_p;
{ 
  int			i, j, k, iSizeResultant;
  int			p, q, r;
  int			iConstIn = (iPolyIn == U)? V: U;
  double		**rgrgA_plus_B, **rgrgB;
  double		*dEReal, *dEImag, *dEBeta, **rgrgEVectors;
  double		Sval, Tval;
  int			iDeg_Curve;
  int			iSize_C;
  int			num_sol = 0;
  int			chck_val;
  COMPLEX		tmp_comp, CdC;
  double		discr; /* this contains c^2+d^2 if complex
				  number is c + id */

  iDeg_Curve = ndet1D_u.poly[0][0].iDeg1;
  iSizeResultant = ndet1D_u.cRows;

  iSize_C = iSizeResultant*iDeg_Curve;

  if (iSize_C > ORD) {
    printf("Not enough memory in the SOLUTION array\n");
    return -1;
  }

  ALLOCMAT(rgrgA_plus_B, iSize_C, iSize_C, double);
  ALLOCMAT(rgrgB, iSize_C, iSize_C, double);

  if (form_power_companion_matrix(ndet1D_u,rgrgB,rgrgA_plus_B,iSize_C) == -1)
    return -1;

  algebraic_prune(rgrgA_plus_B, rgrgB, iDeg0Patch0, iSize_C, 
                  iSizeResultant, &num_sol, base);

  *num_cpoints = 0;
  for (j=0; j<num_sol; j++) 
    printf("SOLUTION #%d is %lf\n", j, SOLUTION[j][0]);
  
  for (j=0; j<num_sol; j++) {
    pdPoints[j].dCoord[iPolyIn] = power_pt(SOLUTION[j][0]);
    pdPoints[j].dCoord[iConstIn] = dConstParam;
    if (base) {
      extract_st_val(tp0, tp1, SOLUTION[j][0], iPolyIn,
	    inverse_power_pt(dConstParam), &Sval, &Tval);
      pdPoints[j].dCoord[S] = power_pt(Sval);
      pdPoints[j].dCoord[T] = power_pt(Tval);
    }
    else {
      pdPoints[j].dCoord[S] = power_pt(SOLUTION[j][1]);
      pdPoints[j].dCoord[T] = power_pt(SOLUTION[j][2]);
    }
  }

  FREEMAT(rgrgA_plus_B, iSize_C, iSize_C);
  FREEMAT(rgrgB, iSize_C, iSize_C);

  *num_points = num_sol;
  return 0;
}

/* This function gets a matrix with 1 variable ndet1D and then
   it evaluates the whole matrix given the value of the variable.
   Now we have an 18 X 18 matrix.
   Then it computes a new matrix in one variable 's' which is
   of size 6 X 6 (for bicubic case) and evaluates s and t. This
   function is called when the patch has base points.
*/

getMatrixforS_tmp(ndet1D_s, fdet1D_u, u_val, row_p)
NDETERMINANT_1D  ndet1D_s, fdet1D_u;
double           u_val;
int              *row_p;
{
  double        **num_mat;
  double        *pow_val;
  int           power_plus_one, max_power, i, j, k;

  max_power = fdet1D_u.poly[0][0].iDeg1;
  ALLOCN(pow_val, double, max_power+1);
  eval_ber(pow_val, u_val, max_power);
  for(i = 0; i <= max_power; i++)
    pow_val[i] /= combChoose(max_power, i);

  ALLOCMAT(num_mat, fdet1D_u.cRows, fdet1D_u.cCols, double);
  for (i=0; i<fdet1D_u.cRows; i++)
  for (j=0; j<fdet1D_u.cCols; j++)
  {
    num_mat[i][j] = 0.0;
    for (k=0; k<=max_power; k++)
      num_mat[i][j] += fdet1D_u.poly[i][j].coeffs[0][k]*pow_val[k];
  }

  for (i=0; i<ndet1D_s.cRows; i++)
  for (j=0; j<ndet1D_s.cCols; j++)
  {
    power_plus_one = ndet1D_s.poly[i][j].iDeg1 + 1;
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[i][j].coeffs[0][k] = num_mat[row_p[i]][j*power_plus_one + k];
  }
  myfree(pow_val);
  FREEMAT(num_mat, fdet1D_u.cRows, fdet1D_u.cCols);
}

getMatrixforT_tmp(ndet1D_t, fdet1D_u, u_val, row_p)
NDETERMINANT_1D  ndet1D_t, fdet1D_u;
double           u_val;
int              *row_p;
{
  double        **num_mat;
  double        *pow_val;
  int           power_plus_one, max_power, i, j, k;

  max_power = fdet1D_u.poly[0][0].iDeg1;
  ALLOCN(pow_val, double, max_power+1);
  eval_ber(pow_val, u_val, max_power);
  for(i = 0; i <= max_power; i++)
    pow_val[i] /= combChoose(max_power, i);

  ALLOCMAT(num_mat, fdet1D_u.cRows, fdet1D_u.cCols, double);
  for (i=0; i<fdet1D_u.cRows; i++)
  for (j=0; j<fdet1D_u.cCols; j++)
  {
    num_mat[i][j] = 0.0;
    for (k=0; k<=max_power; k++)
      num_mat[i][j] += fdet1D_u.poly[i][j].coeffs[0][k]*pow_val[k];
  }

  for (i=0; i<ndet1D_t.cRows; i++)
  for (j=0; j<ndet1D_t.cCols; j++)
  {
    power_plus_one = ndet1D_t.poly[i][j].iDeg1 + 1;
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[i][j].coeffs[0][k] = num_mat[row_p[i]][k*ndet1D_t.cCols+j];
  }
  myfree(pow_val);
  FREEMAT(num_mat, fdet1D_u.cRows, fdet1D_u.cCols);
}

extract_st_val_tmp(fdu, tp0, tp1, eig, eig_var, curve_val, rp, sval, tval)
NDETERMINANT_1D		fdu;
TENSOR_PATCH		tp0, tp1;
double			eig, curve_val, *sval, *tval;
int			eig_var, *rp;
{
  double	Xval, Yval, Zval;
  double  	Xtest, Ytest, Ztest;
  double 	**AA, **BBB;
  double 	**PP, **QQQ;
  int		sSize_C, tSize_C;
  int		nlda, ndim, nmatz, nierr;
  int		tlda, tdim, tmatz, tierr;
  NDETERMINANT_1D ndet1D_s, ndet1D_t;
  double	s_val[20], t_val[20];
  int		s_cnt, t_cnt;
  int		j, k;
  int		iConstIn = (eig_var == U)? V: U;
  double	first_val, second_val;
  int		mnum, nnum;
  double	*nwr, *nwi, *nbeta, **nzz;
  double	*twr, *twi, *tbeta, **tzz;

  if (eig_var < iConstIn) {
    first_val = eig;
    second_val = curve_val;
  }
  else {
    first_val = curve_val;
    second_val = eig;
  }
  mnum = tp0.iDeg0;
  nnum = tp0.iDeg1;
  EvalPointOnPatch(tp1, first_val, second_val, &Xval, &Yval, &Zval);
  allocDeterminant_1D(&ndet1D_s, 2*nnum, 2*nnum, 0, mnum-1);
  getMatrixforS_tmp(ndet1D_s, fdu, eig, rp);
  sSize_C = ndet1D_s.poly[0][0].iDeg1 * ndet1D_s.cRows;
  ALLOCMAT(AA, sSize_C, sSize_C, double);
  ALLOCMAT(BBB, sSize_C, sSize_C, double);
  form_power_companion_matrix(ndet1D_s, BBB, AA, sSize_C);
  ALLOCN(nwr, double, sSize_C);
  ALLOCN(nwi, double, sSize_C);
  ALLOCN(nbeta, double, sSize_C);
  ALLOCMAT(nzz, sSize_C, sSize_C, double);
  call_rgg(BBB, AA, sSize_C, sSize_C, nwr, nwi, nbeta, nzz, 1);
  s_cnt = 0;
  for (j=0; j<sSize_C; j++)
    if ( /* (!FP_EQ(nbeta[j], 0.0)) && */ (FP_EQ(nwi[j], 0.0))) {
      s_val[s_cnt] = nwr[j] / nbeta[j];
      if ((s_val[s_cnt] >= -NEW_EPS) && (s_val[s_cnt] <= 1.0+NEW_EPS)) s_cnt++;
    }

  allocDeterminant_1D(&ndet1D_t, mnum, mnum, 0, 2*nnum - 1);
  getMatrixforT_tmp(ndet1D_t, fdu, eig, rp);
  tSize_C = ndet1D_t.poly[0][0].iDeg1 * ndet1D_t.cRows;
  ALLOCMAT(PP, tSize_C, tSize_C, double);
  ALLOCMAT(QQQ, tSize_C, tSize_C, double);
  form_power_companion_matrix(ndet1D_t, QQQ, PP, tSize_C);
  ALLOCN(twr, double, tSize_C);
  ALLOCN(twi, double, tSize_C);
  ALLOCN(tbeta, double, tSize_C);
  ALLOCMAT(tzz, tSize_C, tSize_C, double);
  call_rgg(QQQ, PP, tSize_C, tSize_C, twr, twi, tbeta, tzz, 1);
  t_cnt = 0;
  for (j=0; j<tSize_C; j++)
    if ( /* (!FP_EQ(tbeta[j], 0.0)) && */ (FP_EQ(twi[j], 0.0))) {
      t_val[t_cnt] = twr[j] / tbeta[j];
      if ((t_val[t_cnt] >= -NEW_EPS) && (t_val[t_cnt] <= 1.0+NEW_EPS)) t_cnt++;
    }

  *sval = *tval = -1.0;
  for (j=0; j<s_cnt; j++)
  for (k=0; k<t_cnt; k++)
  {
    EvalPointOnPatch(tp0, s_val[j], t_val[k], &Xtest, &Ytest, &Ztest);
    if (FP_EQ_EPS(Xval, Xtest, 1e-2) && FP_EQ_EPS(Yval, Ytest, 1e-2) && 
	FP_EQ_EPS(Zval, Ztest, 1e-2))
    {
      *sval = s_val[j];
      *tval = t_val[k];
    }
  }

  myfree(nwr); myfree(nwi); myfree(nbeta);
  myfree(twr); myfree(twi); myfree(tbeta);
  FREEMAT(AA, sSize_C, sSize_C);
  FREEMAT(BBB, sSize_C, sSize_C);
  FREEMAT(nzz, sSize_C, sSize_C);
  FREEMAT(PP, tSize_C, tSize_C);
  FREEMAT(QQQ, tSize_C, tSize_C);
  FREEMAT(tzz, tSize_C, tSize_C)
#ifdef __FREE_ALL
  free_alloc_det_1D(&ndet1D_s);
  free_alloc_det_1D(&ndet1D_t);
#endif
}

getMatrixforST(ndet1D_s, ndet1D_t, tp, xv, yv, zv, lc1, lc2)
NDETERMINANT_1D  ndet1D_s, ndet1D_t;
TENSOR_PATCH	 tp;
double           xv, yv, zv, lc1, lc2;
{
  double 	**x_array, **y_array, **z_array, **f1_array, **f2_array;
  int           power_plus_one, i, j, k;
  int		degS, degT;
  long		combProd;

  degS = tp.iDeg0;
  degT = tp.iDeg1;
  ALLOCMAT(x_array, degS+1, degT+1, double);
  ALLOCMAT(y_array, degS+1, degT+1, double);
  ALLOCMAT(z_array, degS+1, degT+1, double);
  ALLOCMAT(f1_array, degS+1, degT+1, double);
  ALLOCMAT(f2_array, degS+1, degT+1, double);
  for (i=0; i<=degS; i++)
  for (j=0; j<=degT; j++)
  {
    combProd = combChoose(degS,i)*combChoose(degT,j);
    x_array[i][j] = (tp.coeffs[i][j].dCoord[X] - xv*tp.coeffs[i][j].dCoord[W])*
		    combProd;
    y_array[i][j] = (tp.coeffs[i][j].dCoord[Y] - yv*tp.coeffs[i][j].dCoord[W])*
		    combProd;
    z_array[i][j] = (tp.coeffs[i][j].dCoord[Z] - zv*tp.coeffs[i][j].dCoord[W])*
		    combProd;
    f1_array[i][j] = lc1*x_array[i][j] + (1.0 - lc1)*y_array[i][j];
    f2_array[i][j] = lc2*z_array[i][j] + (1.0 - lc2)*y_array[i][j];
  }
  
  if ((ndet1D_s.cRows != 2*degT) || (ndet1D_t.cRows != 2*degS)) {
    printf("Dimensions of ndet1D_s or ndet1D_t are WRONG!!\n");
    exit(-1);
  }

/* 
   We are actually applying Sylvester's resultant to two equations in
   one unknown (t). The eqns. are
		X(s,t) - xv*W(s,t) = 0
		Y(s,t) - yv*W(s,t) = 0
		Z(s,t) - zv*W(s,t) = 0
   We take a linear combination to get 2 eqns. and eliminate t to
   obtain value for s.
*/

  power_plus_one = ndet1D_s.poly[0][0].iDeg1 + 1;

  for (i=0; i<degT; i++) { /* For each of the first degT rows */
    for (j=0; j<i; j++) /* Fill first i columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[i][j].coeffs[0][k] = 0.0;

    for (j=i; j<i+degT+1; j++) /* the next degT+1 columns with proper coeffs */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[i][j].coeffs[0][k] = f1_array[k][i+degT-j];

    for (j=i+degT+1; j<2*degT; j++) /* Fill rest of columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[i][j].coeffs[0][k] = 0.0;
  }
  for (i=0; i<degT; i++) { /* For each of the first degT rows */
    for (j=0; j<i; j++) /* Fill first i columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[2*degT-1-i][j].coeffs[0][k] = 0.0;

    for (j=i; j<i+degT+1; j++) /* the next degT+1 columns with proper coeffs */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[2*degT-1-i][j].coeffs[0][k] = f2_array[k][i+degT-j];

    for (j=i+degT+1; j<2*degT; j++) /* Fill rest of columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_s.poly[2*degT-1-i][j].coeffs[0][k] = 0.0;
  }

/* 
   We are actually applying Sylvester's resultant to two equations in
   one unknown (s). The eqns. are
		X(s,t) - xv*W(s,t) = 0
		Y(s,t) - yv*W(s,t) = 0
		Z(s,t) - zv*W(s,t) = 0
   We take a linear combination to get 2 eqns. and eliminate s to
   obtain value for t.
*/

  power_plus_one = ndet1D_t.poly[0][0].iDeg1 + 1;

  for (i=0; i<degS; i++) { /* For each of the first degS rows */
    for (j=0; j<i; j++) /* Fill first i columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[i][j].coeffs[0][k] = 0.0;

    for (j=i; j<i+degS+1; j++) /* the next degS+1 columns with proper coeffs */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[i][j].coeffs[0][k] = f1_array[i+degS-j][k];

    for (j=i+degS+1; j<2*degS; j++) /* Fill rest of columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[i][j].coeffs[0][k] = 0.0;
  }
  for (i=0; i<degS; i++) { /* For each of the first degS rows */
    for (j=0; j<i; j++) /* Fill first i columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[2*degS-1-i][j].coeffs[0][k] = 0.0;

    for (j=i; j<i+degS+1; j++) /* the next degS+1 columns with proper coeffs */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[2*degS-1-i][j].coeffs[0][k] = f2_array[i+degS-j][k];

    for (j=i+degS+1; j<2*degS; j++) /* Fill rest of columns with 0 */
    for (k=0; k<power_plus_one; k++)
      ndet1D_t.poly[2*degS-1-i][j].coeffs[0][k] = 0.0;
  }

#ifdef __FREE_ALL
  FREEMAT(x_array, degS+1, degT+1);
  FREEMAT(y_array, degS+1, degT+1);
  FREEMAT(z_array, degS+1, degT+1);
  FREEMAT(f1_array, degS+1, degT+1);
  FREEMAT(f2_array, degS+1, degT+1);
#endif
}

extract_st_val(tp0, tp1, eig, eig_var, curve_val, sval, tval)
TENSOR_PATCH		tp0, tp1;
double			eig, curve_val, *sval, *tval;
int			eig_var;
{
  double	Xval, Yval, Zval;
  double  	Xtest, Ytest, Ztest;
  double 	**AA, **BBB;
  double 	**PP, **QQQ;
  int		sSize_C, tSize_C;
  int		nlda, ndim, nmatz, nierr;
  int		tlda, tdim, tmatz, tierr;
  NDETERMINANT_1D ndet1D_s, ndet1D_t;
  double	s_val[20], t_val[20];
  int		s_cnt, t_cnt;
  int		j, k;
  int		iConstIn = (eig_var == U)? V: U;
  double	first_val, second_val;
  double	lc1,lc2, min_dist, tmp_dist, tot_dist;
  int		mnum, nnum, mins_ind, mint_ind;
  double	*nwr, *nwi, *nbeta, **nzz;
  double	*twr, *twi, *tbeta, **tzz;
  double 	a, b, c, d;

  if (eig_var < iConstIn) {
    first_val = eig;
    second_val = curve_val;
  }
  else {
    first_val = curve_val;
    second_val = eig;
  }
  mnum = tp0.iDeg0;
  nnum = tp0.iDeg1;
  EvalPointOnPatch(tp1, first_val, second_val, &Xval, &Yval, &Zval);
  tot_dist = Xval*Xval + Yval*Yval + Zval*Zval;

  a = b = c = d = 0.0;
  while (FP_EQ_EPS(a, 0.0, 1e-4)) 
    a = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(b, 0.0, 1e-4)) 
    b = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(c, 0.0, 1e-4))
    c = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(d, 0.0, 1e-4))
    d = 0.7 + 0.3*drand48();

  lc1 = 0.7;
  lc2 = 0.6;
  allocDeterminant_1D(&ndet1D_s, 2*nnum, 2*nnum, 0, mnum);
  allocDeterminant_1D(&ndet1D_t, 2*mnum, 2*mnum, 0, nnum);
  getMatrixforST(ndet1D_s, ndet1D_t, tp0, Xval, Yval, Zval, lc1, lc2);
  sSize_C = ndet1D_s.poly[0][0].iDeg1 * ndet1D_s.cRows;
  ALLOCMAT(AA, sSize_C, sSize_C, double);
  ALLOCMAT(BBB, sSize_C, sSize_C, double);
  ALLOCN(nwr, double, sSize_C);
  ALLOCN(nwi, double, sSize_C);
  ALLOCN(nbeta, double, sSize_C);
  ALLOCMAT(nzz, sSize_C, sSize_C, double);
  s_cnt = 0;
  if (1) /* (form_companion_matrix(ndet1D_s, BBB, sSize_C, a, b, c, d) */
	 /* == -1) */ {
#ifdef DEBUG1
    printf("RGG\n");
#endif
    form_power_companion_matrix(ndet1D_s, BBB, AA, sSize_C);
    call_rgg(BBB, AA, sSize_C, sSize_C, nwr, nwi, nbeta, nzz, 0);
    for (j=0; j<sSize_C; j++)
      if ( /* (!FP_EQ(nbeta[j], 0.0)) && */ (FP_EQ(nwi[j], 0.0))) {
        s_val[s_cnt] = nwr[j] / nbeta[j];
        if ((s_val[s_cnt] >= -NEW_EPS) && (s_val[s_cnt] <= 1.0+NEW_EPS)) 
	  s_cnt++;
      }
  } else {
    call_rg(BBB, sSize_C, sSize_C, nwr, nwi, nzz, 0); 
    for (j=0; j<sSize_C; j++)
      if (FP_EQ(nwi[j], 0.0)) {
	 /* s_val[s_cnt] = trans_inverse_power_pt(nwr[j], a, b, c, d); */
	s_val[s_cnt] = (a*nwr[j] + b) / ((a+c)*nwr[j] + b + d);
        if ((s_val[s_cnt] >= -NEW_EPS) && (s_val[s_cnt] <= 1.0+NEW_EPS)) 
	  s_cnt++;
      }
  }

  tSize_C = ndet1D_t.poly[0][0].iDeg1 * ndet1D_t.cRows;
  ALLOCMAT(PP, tSize_C, tSize_C, double);
  ALLOCMAT(QQQ, tSize_C, tSize_C, double);
  ALLOCN(twr, double, tSize_C);
  ALLOCN(twi, double, tSize_C);
  ALLOCN(tbeta, double, tSize_C);
  ALLOCMAT(tzz, tSize_C, tSize_C, double);
  t_cnt = 0;
  if (1) /* (form_companion_matrix(ndet1D_t, QQQ, tSize_C, a, b, c, d) */
	 /* == -1) */ {
#ifdef DEBUG1
    printf("RGG\n");
#endif
    form_power_companion_matrix(ndet1D_t, QQQ, PP, tSize_C);
    call_rgg(QQQ, PP, tSize_C, tSize_C, twr, twi, tbeta, tzz, 0);
    for (j=0; j<tSize_C; j++)
      if ( /* (!FP_EQ(tbeta[j], 0.0)) && */ (FP_EQ(twi[j], 0.0))) {
        t_val[t_cnt] = twr[j] / tbeta[j];
        if ((t_val[t_cnt] >= -NEW_EPS) && (t_val[t_cnt] <= 1.0+NEW_EPS)) 
	  t_cnt++;
      }
  } else {
    call_rg(QQQ, tSize_C, tSize_C, twr, twi, tzz, 0); 
    for (j=0; j<tSize_C; j++)
      if (FP_EQ(twi[j], 0.0)) {
	 /* t_val[t_cnt] = trans_inverse_power_pt(twr[j], a, b, c, d); */
	t_val[t_cnt] = (a*twr[j] + b) / ((a+c)*twr[j] + b + d);
        if ((t_val[t_cnt] >= -NEW_EPS) && (t_val[t_cnt] <= 1.0+NEW_EPS)) 
	  t_cnt++;
      }
  }

  *sval = *tval = -1.0;
  min_dist = HUGE_VAL;
  for (j=0; j<s_cnt; j++)
  for (k=0; k<t_cnt; k++)
  {
    EvalPointOnPatch(tp0, s_val[j], t_val[k], &Xtest, &Ytest, &Ztest);
    tmp_dist = (Xtest-Xval)*(Xtest-Xval) +
	       (Ytest-Yval)*(Ytest-Yval) +
	       (Ztest-Zval)*(Ztest-Zval);
    if (tmp_dist < min_dist) {
      min_dist = tmp_dist;
      mins_ind = j;
      mint_ind = k;
    }
  }
/*
  if (FP_EQ_EPS(Xval, Xtest, 1e-3) && FP_EQ_EPS(Yval, Ytest, 1e-3) && 
      FP_EQ_EPS(Zval, Ztest, 1e-3))
  {
    *sval = s_val[j];
    *tval = t_val[k];
  }
*/

  myfree(nwr); myfree(nwi); myfree(nbeta);
  myfree(twr); myfree(twi); myfree(tbeta);
  FREEMAT(AA, sSize_C, sSize_C);
  FREEMAT(BBB, sSize_C, sSize_C);
  FREEMAT(nzz, sSize_C, sSize_C);
  FREEMAT(PP, tSize_C, tSize_C);
  FREEMAT(QQQ, tSize_C, tSize_C);
  FREEMAT(tzz, tSize_C, tSize_C)
#ifdef __FREE_ALL
  free_alloc_det_1D(&ndet1D_s);
  free_alloc_det_1D(&ndet1D_t);
#endif

#ifdef SCALE_DOWN_FOR_TRACING_ALONE
  if ((min_dist/tot_dist) > 1.0e-9) return;
#else
  if ((min_dist/tot_dist) > 1.0e-6) return;
#endif

  *sval = s_val[mins_ind];
  *tval = t_val[mint_ind];
}

/* Trying to solve a1*s + b1*t + c1  = 0 and a2*s + b2*t + c2 = 0 */
int solve_simult_lin_eq(a1, b1, c1, a2, b2, c2, s, t)
double a1, b1, c1, a2, b2, c2;
double *s, *t;
{
  double sdenom, tdenom;

  sdenom = a1*b2 - a2*b1;
  tdenom = -sdenom;
  if (FP_EQ_EPS(sdenom, 0.0, 1e-4)) {
    printf("solve_simult_lin_eq: Equations are not independent\n");
    return -1;
  }
  
  *s = (b1*c2 - b2*c1) / sdenom;
  *t = (a1*c2 - a2*c1) / tdenom;
  return 0;
}

#ifdef OLD_PLANE_CASE
int special_plane_case(tp, xval, yval, zval, sval, tval)
TENSOR_PATCH *tp;
double xval, yval, zval, *sval, *tval;
{
  double aff00[3];
  double p00[4], p01[4], p10[4], p11[4], pt[4];
  double v1[3], v2[3], norm[3];
  double v1mag, normmag;
  double x01, y01, x10, y10, x11, y11, xpt, ypt;
  double x_at_t0, x_at_t1;
  int i, j, k;

  for (i=0; i<3; i++) {
    aff00[i] = tp->coeffs[0][0].dCoord[i] / tp->coeffs[0][0].dCoord[W];
    p00[i] = 0.0;
  }
  p00[3] = tp->coeffs[0][0].dCoord[W];
  
  for (i=0; i<3; i++) {
    p01[i] = ((tp->coeffs[0][1].dCoord[i]/tp->coeffs[0][1].dCoord[W]) -
	      aff00[i]) ;
    p10[i] = ((tp->coeffs[1][0].dCoord[i]/tp->coeffs[1][0].dCoord[W]) -
	      aff00[i]) ;
    p11[i] = ((tp->coeffs[1][1].dCoord[i]/tp->coeffs[1][1].dCoord[W]) -
	      aff00[i]) ;
  }
  p01[3] = tp->coeffs[0][1].dCoord[W];
  p10[3] = tp->coeffs[1][0].dCoord[W];
  p11[3] = tp->coeffs[1][1].dCoord[W];
  
  pt[0] = xval - aff00[0];
  pt[1] = yval - aff00[1];
  pt[2] = zval - aff00[2];
  pt[3] = 1.0;

  for (i=0; i<3; i++) {
    v1[i] = p01[i] ;
    v2[i] = p10[i] ;
  }

/* norm is cross-product of v1 and v2 */
  norm[0] = v1[1]*v2[2] - v2[1]*v1[2];
  norm[1] = v1[2]*v2[0] - v2[2]*v1[0];
  norm[2] = v1[0]*v2[1] - v2[0]*v1[1];

  v1mag = normmag = 0.0;
  for (i=0; i<3; i++) {
    v1mag += v1[i]*v1[i];
    normmag += norm[i]*norm[i];
  }
  v1mag = sqrt(v1mag);
  normmag = sqrt(normmag);
  for (i=0; i<3; i++) {
    v1[i] /= v1mag;
    norm[i] /= normmag;
  }

/* v2 is the cross-product of norm and v1 */
  v2[0] = norm[1]*v1[2] - v1[1]*norm[2];
  v2[1] = norm[2]*v1[0] - v1[2]*norm[0];
  v2[2] = norm[0]*v1[1] - v1[0]*norm[1];
  
  x01 = y01 = x10 = y10 = x11 = y11 = xpt = ypt = 0.0;
  for (i=0; i<3; i++) {
    x01 += v1[i]*p01[i];
    y01 += v2[i]*p01[i];
    x10 += v1[i]*p10[i];
    y10 += v2[i]*p10[i];
    x11 += v1[i]*p11[i];
    y11 += v2[i]*p11[i];
    xpt += v1[i]*pt[i];
    ypt += v2[i]*pt[i];
  }

  if (!FP_EQ_EPS(y01, 0.0, 1e-2) || !FP_EQ_EPS(y10, y11, 1e-2)) {
    printf("special_plane_case:TRANSFORMATION NOT CORRECT\n");
    return;
  }

  *sval = ypt / y11;
  x_at_t0 = (*sval) * x10;
  x_at_t1 = x01 + (*sval) * (x11 - x01);
  *tval = (xpt - x_at_t0) / (x_at_t1 - x_at_t0);
}
#else
int special_plane_case(tp, xval, yval, zval, sval, tval)
TENSOR_PATCH *tp;
double xval, yval, zval, *sval, *tval;
{
  double cx_cnst, cx_s, cx_t, cx_st;
  double cy_cnst, cy_s, cy_t, cy_st;
  double cz_cnst, cz_s, cz_t, cz_st;
  double rnd1, rnd2;
  double a1, a2, a3, a4, b1, b2, b3, b4;
  double c13, c14, c23, c24, c1423;
  double s1, s2, t1, t2;
  double discr, denr, tmp;

  cx_cnst = tp->coeffs[0][0].dCoord[0] - xval*tp->coeffs[0][0].dCoord[3];
  cx_s = (tp->coeffs[1][0].dCoord[0] - tp->coeffs[0][0].dCoord[0]) -
          xval*(tp->coeffs[1][0].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cx_t = (tp->coeffs[0][1].dCoord[0] - tp->coeffs[0][0].dCoord[0]) -
          xval*(tp->coeffs[0][1].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cx_st = ((tp->coeffs[0][0].dCoord[0] + tp->coeffs[1][1].dCoord[0]) - 
	   (tp->coeffs[1][0].dCoord[0] + tp->coeffs[0][1].dCoord[0])) -
     xval*((tp->coeffs[0][0].dCoord[3] + tp->coeffs[1][1].dCoord[3]) - 
	   (tp->coeffs[1][0].dCoord[3] + tp->coeffs[0][1].dCoord[3]));

  cy_cnst = tp->coeffs[0][0].dCoord[1] - yval*tp->coeffs[0][0].dCoord[3];
  cy_s = (tp->coeffs[1][0].dCoord[1] - tp->coeffs[0][0].dCoord[1]) -
          yval*(tp->coeffs[1][0].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cy_t = (tp->coeffs[0][1].dCoord[1] - tp->coeffs[0][0].dCoord[1]) -
          yval*(tp->coeffs[0][1].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cy_st = ((tp->coeffs[0][0].dCoord[1] + tp->coeffs[1][1].dCoord[1]) - 
	   (tp->coeffs[1][0].dCoord[1] + tp->coeffs[0][1].dCoord[1])) -
     yval*((tp->coeffs[0][0].dCoord[3] + tp->coeffs[1][1].dCoord[3]) - 
	   (tp->coeffs[1][0].dCoord[3] + tp->coeffs[0][1].dCoord[3]));

  cz_cnst = tp->coeffs[0][0].dCoord[2] - zval*tp->coeffs[0][0].dCoord[3];
  cz_s = (tp->coeffs[1][0].dCoord[2] - tp->coeffs[0][0].dCoord[2]) -
          zval*(tp->coeffs[1][0].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cz_t = (tp->coeffs[0][1].dCoord[2] - tp->coeffs[0][0].dCoord[2]) -
          zval*(tp->coeffs[0][1].dCoord[3] - tp->coeffs[0][0].dCoord[3]);
  cz_st = ((tp->coeffs[0][0].dCoord[2] + tp->coeffs[1][1].dCoord[2]) - 
	   (tp->coeffs[1][0].dCoord[2] + tp->coeffs[0][1].dCoord[2])) -
     zval*((tp->coeffs[0][0].dCoord[3] + tp->coeffs[1][1].dCoord[3]) - 
	   (tp->coeffs[1][0].dCoord[3] + tp->coeffs[0][1].dCoord[3]));

  rnd1 = 0.5 + 0.3*drand48();
  rnd2 = 0.4 + 0.5*drand48();
  a1 = rnd1*cx_cnst + rnd2*cz_cnst;
  a2 = rnd1*cx_s + rnd2*cz_s;
  a3 = rnd1*cx_t + rnd2*cz_t;
  a4 = rnd1*cx_st + rnd2*cz_st;

  rnd1 = 0.5 + 0.3*drand48();
  rnd2 = 0.4 + 0.5*drand48();
  b1 = rnd1*cy_cnst + rnd2*cz_cnst;
  b2 = rnd1*cy_s + rnd2*cz_s;
  b3 = rnd1*cy_t + rnd2*cz_t;
  b4 = rnd1*cy_st + rnd2*cz_st;

  /* coeffs of quadratic in s are c13 + (c14+c23)s + c24 s^2 = 0 */
  c13 = a1*b3 - a3*b1;
  c14 = a1*b4 - a4*b1;
  c23 = a2*b3 - a3*b2;
  c24 = a2*b4 - a4*b2;
  c1423 = c14 + c23;

  if (FP_EQ_EPS(c24, 0.0, 1e-9)) {
    /* printf("special_plane_case: quadratic term vanishes\n"); */
    if (FP_EQ_EPS(c1423, 0.0, 1e-9)) {
      printf("special_plane_case: very degenerate patch\n");
      exit(-1);
    }
    *sval = -c13 / c1423;
    tmp = a3 + (*sval)*a4;
    if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
      tmp = b3 + (*sval)*b4;
      if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
	printf("special_plane_case: both denrs for obtaining tval are zero\n");
	exit(-1);
      }
      *tval = -(b1 + (*sval)*b2) / tmp;
    } else 
      *tval = -(a1 + (*sval)*a2) / tmp;
    return 0;
  }
  
/*
  if (!FP_EQ_EPS(c24, 0.0, 1e-5))
    printf("special_plane_case: proper quadratic equation. Could be error\n");
*/  
  discr = (c1423*c1423 - 4*c13*c24);
  if (discr < 0.0) {
    printf("special_plane_case: discriminant is negative\n");
    exit(-1);
  }
  discr = sqrt(discr);
  denr = 1.0 / (2.0*c24);
  s1 = (-c1423 + discr) * denr;
  s2 = (-c1423 - discr) * denr;
  if ((s1 > -EPSILON) && (s1 < (1.0+EPSILON))) {
    if ((s2 > -EPSILON) && (s2 < (1.0+EPSILON))) {
      printf("special_plane_case: s1 and s2 lie within the domain. Surrender\n");
      exit(-1);
    }
    *sval = s1;
    tmp = a3 + (*sval)*a4;
    if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
      tmp = b3 + (*sval)*b4;
      if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
	printf("special_plane_case: both denrs for obtaining tval are zero\n");
	exit(-1);
      }
      *tval = -(b1 + (*sval)*b2) / tmp;
    } else 
      *tval = -(a1 + (*sval)*a2) / tmp;
    return 0;
  }
  if ((s2 > -EPSILON) && (s2 < (1.0+EPSILON))) {
    *sval = s2;
    tmp = a3 + (*sval)*a4;
    if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
      tmp = b3 + (*sval)*b4;
      if (FP_EQ_EPS(tmp, 0.0, 1e-9)) {
	printf("special_plane_case: both denrs for obtaining tval are zero\n");
	exit(-1);
      }
      *tval = -(b1 + (*sval)*b2) / tmp;
    } else 
      *tval = -(a1 + (*sval)*a2) / tmp;
    return 0;
  }
  /* s1 and s2 outside the domain */
  *sval = -1.0;
  *tval = -1.0;
  return -1;
}
#endif

GeneralInverseOp(tp0, tp1, eig, eig_var, curve_val, sval, tval)
TENSOR_PATCH		tp0, tp1;
double			eig, curve_val, *sval, *tval;
int			eig_var;
{
  double	Xval, Yval, Zval;
  int		iConstIn = (eig_var == U)? V: U;
  double	first_val, second_val;
  short         plane_case = 0;

  if ((tp0.iDeg0 == 1) && (tp0.iDeg1 == 1)) 
    plane_case = 1;

  if (plane_case) {
    if (eig_var < iConstIn) {
      first_val = eig;
      second_val = curve_val;
    }
    else {
      first_val = curve_val;
      second_val = eig;
    }
    EvalPointOnPatch(tp1, first_val, second_val, &Xval, &Yval, &Zval);
    special_plane_case(&tp0, Xval, Yval, Zval, sval, tval);
    return;
  }

  extract_st_val(tp0, tp1, eig, eig_var, curve_val, sval, tval);
}

/* add point to the global list of points */
addPointToList(point4D_val, ppList)
POINT4D 		point4D_val;
LIST_POINT4D     	**ppList;
{ 
  LIST_POINT4D		*new_point;
  LIST_POINT4D		*tmp_ptr;
  int			i, same_point;
  
  /* undo the parametrization u' = u/1-u etc. */
  for(i=0; i < 4; i++)
    point4D_val.dCoord[i] = inverse_power_pt(point4D_val.dCoord[i]);
  
  if ((point4D_val.dCoord[U] >= -NEW_EPS) && (point4D_val.dCoord[U] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[V] >= -NEW_EPS) && (point4D_val.dCoord[V] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[S] >= -NEW_EPS) && (point4D_val.dCoord[S] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[T] >= -NEW_EPS) && (point4D_val.dCoord[T] <= (1.0+NEW_EPS)))
  { 
    tmp_ptr = (*ppList);
    same_point = 0;
    while (tmp_ptr != NULL) {
      if (EQ_4D_EPS(point4D_val, tmp_ptr->p4D, 1e-3)) {
	same_point = 1;
	break;
      }
      tmp_ptr = tmp_ptr->plp4DNext;
    }

    if (!same_point) {
      ALLOCN(new_point, LIST_POINT4D, 1);
      new_point->plp4DNext = *ppList; 
      new_point->p4D = point4D_val;
      *ppList = new_point;
    }
  }
}

addXtraPointToList(point4D_val, ppList)
POINT4D 		point4D_val;
LIST_POINT4D     	**ppList;
{ 
  LIST_POINT4D		*new_point;
  int			i;
  double		tmp_var;
  LIST_POINT4D		*tmp_ptr;
  int			same_point;
  
  /* undo the parametrization u' = u/1-u etc. */
  for(i=0; i < 4; i++)
    point4D_val.dCoord[i] = inverse_power_pt(point4D_val.dCoord[i]);

/* I am swapping U with S and V with T because we are dealing with
   the second patch */
  tmp_var = point4D_val.dCoord[U];
  point4D_val.dCoord[U] = point4D_val.dCoord[S];
  point4D_val.dCoord[S] = tmp_var;
  tmp_var = point4D_val.dCoord[V];
  point4D_val.dCoord[V] = point4D_val.dCoord[T];
  point4D_val.dCoord[T] = tmp_var;
  
  if ((point4D_val.dCoord[U] >= -NEW_EPS) && (point4D_val.dCoord[U] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[V] >= -NEW_EPS) && (point4D_val.dCoord[V] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[S] >= -NEW_EPS) && (point4D_val.dCoord[S] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[T] >= -NEW_EPS) && (point4D_val.dCoord[T] <= (1.0+NEW_EPS)))
  { 
    if (!(FP_EQ_EPS(point4D_val.dCoord[U], 0.0, NEW_EPS) ||
	  FP_EQ_EPS(point4D_val.dCoord[U], 1.0, NEW_EPS) ||
	  FP_EQ_EPS(point4D_val.dCoord[V], 0.0, NEW_EPS) ||
	  FP_EQ_EPS(point4D_val.dCoord[V], 1.0, NEW_EPS) )) {
      tmp_ptr = (*ppList);
      same_point = 0;
      while (tmp_ptr != NULL) {
        if (EQ_4D_EPS(point4D_val, tmp_ptr->p4D, 1e-3)) {
	  same_point = 1;
	  break;
        }
	tmp_ptr = tmp_ptr->plp4DNext;
      }

      if (!same_point) {
        ALLOCN(new_point, LIST_POINT4D, 1);
        new_point->plp4DNext = *ppList; 
        new_point->p4D = point4D_val;
        *ppList = new_point;
      }
    }
  }
}

/* delete point from the global list of points */
deletePointFromList(point4D_val, ppList)
POINT4D 		point4D_val;
LIST_POINT4D     	**ppList;
{ 
  LIST_POINT4D		*new_point, *next_point;
  POINT4D		Dpt, Npt;
  int			i;
  
  new_point = *ppList;
  if (!new_point) return;
  Dpt = new_point->p4D;
  if (FP_EQ_EPS(Dpt.dCoord[U], point4D_val.dCoord[U], 9e-2) &&
      FP_EQ_EPS(Dpt.dCoord[V], point4D_val.dCoord[V], 9e-2)) {
    *ppList = new_point->plp4DNext;
    return;
  }

  for (new_point = *ppList; new_point; new_point=new_point->plp4DNext){
    next_point = new_point->plp4DNext;
    if (!next_point) return;
    Npt = next_point->p4D;
    if (FP_EQ_EPS(Npt.dCoord[U], point4D_val.dCoord[U], 1e-2) &&
        FP_EQ_EPS(Npt.dCoord[V], point4D_val.dCoord[V], 1e-2)) {
      new_point->plp4DNext = next_point->plp4DNext;
      myfree(next_point);
      return;
    }
  }
}

/* add complex point to the global list of points */
addCPointToList(point4CPX_val, ppList)
POINT4CPX 		point4CPX_val;
LIST_POINT4CPX     	**ppList;
{ 
  LIST_POINT4CPX	*new_point;
  int			i;
  
  /* undo the parametrization u' = u/1-u etc. */
/*
  for(i=0; i < 2; i++)
   point4CPX_val.cpxCoord[i] = inv_cxpower_pt(point4CPX_val.cpxCoord[i]);
*/
  
  if ( (point4CPX_val.cpxCoord[V].Re <= 3.001) &&
       (point4CPX_val.cpxCoord[V].Re >= -2.001) &&
       (point4CPX_val.cpxCoord[V].Im <= 2.001) )
  {
    ALLOCN(new_point, LIST_POINT4CPX, 1);
    new_point->plp4CPXNext = *ppList; 
    new_point->p4CPX = point4CPX_val;
    *ppList = new_point;
  }
}

/* print all the points in the list */
printList(pList)
LIST_POINT4D     	*pList;
{ 
  LIST_POINT4D		*tmp_point;

  tmp_point = pList;

  while(tmp_point)
  { 
/*
    printf("Start Point (%lf %lf)\n", tmp_point->p4D.dCoord[U], 
    tmp_point->p4D.dCoord[V], tmp_point->p4D.dCoord[S], 
    tmp_point->p4D.dCoord[T]);
*/
    tmp_point = tmp_point->plp4DNext;
  }
}

/* take a point in the Bernstein-basis space and return it in
   the Power-basis space, assuming that the parametrization is
   s' = s/(1-s)
   where s' is in Power-basis space and s is in Bernstein-basis 
   space.
*/
double
power_pt(dVal)
double dVal;
{ 
  double rVal;
  rVal = (FP_EQ(dVal, 1.0))? HUGE_VAL : dVal/(1.0 - dVal);
  return(rVal);
}

/* take a point in the Power-basis space and return it in
   the Bernstein-basis space, assuming that the parametrization 
   is:
   s' = s/(1-s) ie s = s'/(1+s')
   where s' is in Power-basis space and s is in Bernstein-basis 
   space.
*/

double
inverse_power_pt(dVal)
double	dVal;
{ 
  double rVal;
  rVal = (dVal == HUGE_VAL)? 1.0 : dVal/(1.0 + dVal);
  return(rVal);
}

double
trans_inverse_power_pt(dVal, a, b, c, d)
double	dVal, a, b, c, d;
{ 
  double rVal;
  rVal = (dVal == HUGE_VAL)? a/(a+c) : (b + a*dVal)/((b+d) + (a+c)*dVal);
  return(rVal);
}

