#include "defs.h"

extern double Step_Size_Out;
extern double Sphradius;
extern LIST_POINT4D *gplp4DEndPoints;

extern	void  call_svd();
extern  int   calc_det();
extern void  powell();
EXTERN_LINALG

#define SQ(a) ((a)*(a))
#define		PLAY_SAFE 	0
#define		ITER_LIM	100
#define		MIN_STEP_SIZE	0.00005
#define		NEW_EPS		5.0*1e-5

TENSOR_PATCH pow_tp0;
TENSOR_PATCH pow_tp1;
int unchanged;
double unchanged_val;

int IsPtInDomain(p4D, dom, eps)
POINT4D *p4D;
RECT_BOX *dom;
double eps;
{
  if ((p4D->dCoord[U] >= dom->i2Dbox[0].low-eps) && 
      (p4D->dCoord[U] <= (dom->i2Dbox[0].high+eps)) &&
      (p4D->dCoord[V] >= dom->i2Dbox[1].low-eps) && 
      (p4D->dCoord[V] <= (dom->i2Dbox[1].high+eps)) &&
      (p4D->dCoord[S] >= -eps) && 
      (p4D->dCoord[S] <= (1+eps)) &&
      (p4D->dCoord[T] >= -eps) && 
      (p4D->dCoord[T] <= (1+eps)))
    return 1;
  return 0;
}

double RefinePoint1D(coord)
double *coord;
{
  double uval,vval,sval,tval;
  double x0val,y0val,z0val;
  double x1val,y1val,z1val, ret;

  if (unchanged != U && unchanged != V) {
    printf("Improper input to RefinePoint1D:: %d\n", unchanged);
    exit(-1);
  }
  if (unchanged == U) {
    uval = unchanged_val;
    vval = coord[0];
  } else {
    vval = unchanged_val;
    uval = coord[0];
  } 
  sval = coord[1];
  tval = coord[2];

  EvalPointOnPatch(pow_tp0, sval, tval, &x0val, &y0val, &z0val);
  EvalPointOnPatch(pow_tp1, uval, vval, &x1val, &y1val, &z1val);

  ret = (x0val - x1val) * (x0val - x1val) +
	(y0val - y1val) * (y0val - y1val) +
	(z0val - z1val) * (z0val - z1val);
  return ret;
}
  
double RefinePoint(coord)
double *coord;
{
  double uval,vval,sval,tval;
  double x0val,y0val,z0val;
  double x1val,y1val,z1val, ret;

  uval = coord[0];
  vval = coord[1];
  sval = coord[2];
  tval = coord[3];

  EvalPointOnPatch(pow_tp0, sval, tval, &x0val, &y0val, &z0val);
  EvalPointOnPatch(pow_tp1, uval, vval, &x1val, &y1val, &z1val);

  ret = (x0val - x1val) * (x0val - x1val) +
	(y0val - y1val) * (y0val - y1val) +
	(z0val - z1val) * (z0val - z1val);
  return ret;
}
  
void InvertPointList(PointList)
LIST_POINT4D **PointList;
{
  LIST_POINT4D *tmpListPtr1, *tmpListPtr2;

  tmpListPtr1 = *PointList;
  *PointList = NULL;

  for (;tmpListPtr1;) {
    tmpListPtr2 = tmpListPtr1;
    tmpListPtr1 = tmpListPtr1->plp4DNext;
    tmpListPtr2->plp4DNext = *PointList;
    *PointList = tmpListPtr2;
  }
}

void Invert2DPointList(PointList)
LIST_POINT2D **PointList;
{
  LIST_POINT2D *tmpListPtr1, *tmpListPtr2;

  tmpListPtr1 = *PointList;
  *PointList = NULL;

  for (;tmpListPtr1;) {
    tmpListPtr2 = tmpListPtr1;
    tmpListPtr1 = tmpListPtr1->plp2DNext;
    tmpListPtr2->plp2DNext = *PointList;
    *PointList = tmpListPtr2;
  }
}

void ConCat2D(PtLst1, PtLst2)
LIST_POINT2D **PtLst1, *PtLst2;
{
  LIST_POINT2D *tmpListPtr1, *tmpListPtr2;

  tmpListPtr1 = *PtLst1;
  tmpListPtr2 = PtLst2;
  if (*PtLst1 == NULL) {
    *PtLst1 = PtLst2;
    return;
  }
  for (;tmpListPtr1;) {
    if (tmpListPtr1->plp2DNext == NULL) {
      while (tmpListPtr2 != NULL &&
	     FP_EQ_EPS(tmpListPtr1->p2D.u, tmpListPtr2->p2D.u, 1e-3) &&
	     FP_EQ_EPS(tmpListPtr1->p2D.v, tmpListPtr2->p2D.v, 1e-3))
	tmpListPtr2 = tmpListPtr2->plp2DNext;
      tmpListPtr1->plp2DNext = tmpListPtr2;
      return;
    }
    tmpListPtr1 = tmpListPtr1->plp2DNext;
  }
}

void ConCat(PtLst1, PtLst2)
LIST_POINT4D **PtLst1, *PtLst2;
{
  LIST_POINT4D *tmpListPtr1, *tmpListPtr2;

  tmpListPtr1 = *PtLst1;
  tmpListPtr2 = PtLst2;
  if (*PtLst1 == NULL) {
    *PtLst1 = PtLst2;
    return;
  }
  for (;tmpListPtr1;) {
    if (tmpListPtr1->plp4DNext == NULL) {
      while (tmpListPtr2 != NULL && EQ_4D(tmpListPtr1->p4D, tmpListPtr2->p4D))
	tmpListPtr2 = tmpListPtr2->plp4DNext;
      tmpListPtr1->plp4DNext = tmpListPtr2;
      return;
    }
    tmpListPtr1 = tmpListPtr1->plp4DNext;
  }
}

void setUpTraceOrder(cComp, uvstList, TotCompNo, retIndex, StIndex,
		     OtherIndex, Code, retEndPt)
INT_CURVE_STRUCT **cComp;
POINT4D *uvstList, **retEndPt;
int TotCompNo, *retIndex, *StIndex, *OtherIndex, *Code;
{
  int i, j, k, firstCompNo, SecCompNo;
  POINT4D t4D;

  *retIndex = *StIndex = *OtherIndex = -1;

  for (i=0; i<TotCompNo; i++) {
    if (EQ_4D(uvstList[0], (*cComp)[i].endPoints[0])) {
      InvertPointList(&((*cComp)[i].Curve));
      (*retEndPt)[0] = (*cComp)[i].endPoints[1];
      (*retEndPt)[1] = uvstList[1];
      *retIndex = i;
      *StIndex = 0;
      break;
    }
    else if (EQ_4D(uvstList[0], (*cComp)[i].endPoints[1])) {
      (*retEndPt)[0] = (*cComp)[i].endPoints[0];
      (*retEndPt)[1] = uvstList[1];
      *retIndex = i;
      *StIndex = 0;
      break;
    }
  }

  if ((*retIndex == -1) && (*StIndex == -1))
    for (i=0; i<TotCompNo; i++) {
      if (EQ_4D(uvstList[1], (*cComp)[i].endPoints[0])) {
        InvertPointList(&((*cComp)[i].Curve));
        (*retEndPt)[0] = (*cComp)[i].endPoints[1];
        (*retEndPt)[1] = uvstList[0];
        *retIndex = i;
        *StIndex = 1;
        break;
      }
      else if (EQ_4D(uvstList[1], (*cComp)[i].endPoints[1])) {
	(*retEndPt)[0] = (*cComp)[i].endPoints[0];
        (*retEndPt)[1] = uvstList[0];
        *retIndex = i;
        *StIndex = 1;
        break;
      }
    }
  else 
    for (j=0; j<TotCompNo; j++) {
      if (EQ_4D(uvstList[1], (*cComp)[j].endPoints[0])) {
	*OtherIndex = j;
	*Code = 2;
	break;
      } else if (EQ_4D(uvstList[1], (*cComp)[j].endPoints[1])) {
	*OtherIndex = j;
	*Code = 3;
	break;
      }
    } 
}

void MergeCurveLists(cComp, fIndex, sIndex, Code, TotCompNo)
INT_CURVE_STRUCT **cComp;
int fIndex, sIndex, Code, *TotCompNo;
{
  int i, j, k, tmp, swap;

  if (fIndex == sIndex) return;
  if (fIndex > sIndex) {
    tmp = fIndex;
    fIndex = sIndex;
    sIndex = tmp;
    swap = 1;
  }  
  else swap = 0;

    switch (Code) {
      case 0: InvertPointList(&((*cComp)[sIndex].Curve));
	      ConCat(&((*cComp)[fIndex].Curve), (*cComp)[sIndex].Curve);
	      (*cComp)[fIndex].endPoints[0] =
				      (*cComp)[sIndex].endPoints[1];
	      for (i=sIndex; i<(*TotCompNo)-1; i++) {
		(*cComp)[i].Curve = (*cComp)[i+1].Curve;
		(*cComp)[i].endPoints[0] = (*cComp)[i+1].endPoints[0];
		(*cComp)[i].endPoints[1] = (*cComp)[i+1].endPoints[1];
	      }
	      break;
      case 1: if (!swap) {
		ConCat(&((*cComp)[fIndex].Curve), (*cComp)[sIndex].Curve);
	        (*cComp)[fIndex].endPoints[0] =
				      (*cComp)[sIndex].endPoints[0];
	      } else {
		ConCat(&((*cComp)[sIndex].Curve), (*cComp)[fIndex].Curve);
		(*cComp)[fIndex].Curve = (*cComp)[sIndex].Curve;
		(*cComp)[fIndex].endPoints[1] = 
				      (*cComp)[sIndex].endPoints[1];
	      }
	      for (i=sIndex; i<(*TotCompNo)-1; i++) {
		(*cComp)[i].Curve = (*cComp)[i+1].Curve;
		(*cComp)[i].endPoints[0] = (*cComp)[i+1].endPoints[0];
		(*cComp)[i].endPoints[1] = (*cComp)[i+1].endPoints[1];
	      }
	      break;
      case 2: if (!swap) {
		ConCat(&((*cComp)[sIndex].Curve), (*cComp)[fIndex].Curve);
	        (*cComp)[fIndex].Curve = (*cComp)[sIndex].Curve;
	        (*cComp)[fIndex].endPoints[1] =
				      (*cComp)[sIndex].endPoints[1];
	      } else {
		ConCat(&((*cComp)[fIndex].Curve), (*cComp)[sIndex].Curve);
	        (*cComp)[fIndex].endPoints[0] =
				      (*cComp)[sIndex].endPoints[0];
	      }
	      for (i=sIndex; i<(*TotCompNo)-1; i++) {
		(*cComp)[i].Curve = (*cComp)[i+1].Curve;
		(*cComp)[i].endPoints[0] = (*cComp)[i+1].endPoints[0];
		(*cComp)[i].endPoints[1] = (*cComp)[i+1].endPoints[1];
	      }
	      break;
      case 3: InvertPointList(&((*cComp)[sIndex].Curve));
	      ConCat(&((*cComp)[sIndex].Curve), (*cComp)[fIndex].Curve);
	      (*cComp)[fIndex].Curve = (*cComp)[sIndex].Curve;
	      (*cComp)[fIndex].endPoints[1] =
				      (*cComp)[sIndex].endPoints[0];
	      for (i=sIndex; i<(*TotCompNo)-1; i++) {
		(*cComp)[i].Curve = (*cComp)[i+1].Curve;
		(*cComp)[i].endPoints[0] = (*cComp)[i+1].endPoints[0];
		(*cComp)[i].endPoints[1] = (*cComp)[i+1].endPoints[1];
	      }
	      break;
    }
    (*TotCompNo) -= 1;
}

void TrivialTrace(pplp4DOutput, p4DStart, p4DEnd, tp1)
LIST_POINT4D            **pplp4DOutput;
POINT4D                 p4DStart, p4DEnd;
TENSOR_PATCH		tp1;
{
  int i, j, k;
  double uvlength, stlength;
  double udiff, vdiff, sdiff, tdiff;
  POINT4D interm_point;
  POINT2D uvVec, stVec;

  udiff = (p4DEnd.dCoord[U] - p4DStart.dCoord[U])/4.0;
  vdiff = (p4DEnd.dCoord[V] - p4DStart.dCoord[V])/4.0;
  sdiff = (p4DEnd.dCoord[S] - p4DStart.dCoord[S])/4.0;
  tdiff = (p4DEnd.dCoord[T] - p4DStart.dCoord[T])/4.0;
  uvlength = sqrt(udiff*udiff + vdiff*vdiff);
  stlength = sqrt(sdiff*sdiff + tdiff*tdiff);

  for (i=0; i<4; i++) {
    interm_point.dCoord[U] = p4DStart.dCoord[U]+i*udiff;
    interm_point.dCoord[V] = p4DStart.dCoord[V]+i*vdiff;
    interm_point.dCoord[S] = p4DStart.dCoord[S]+i*sdiff;
    interm_point.dCoord[T] = p4DStart.dCoord[T]+i*tdiff;
    add_list_point(pplp4DOutput, interm_point, tp1, 0, 0, 1);                   
  }
  add_list_point(pplp4DOutput, p4DEnd, tp1, 0, 0, 1);
}
  

int 
Trace(ndet1D_uv, ndet1D_du, ndet1D_dv, p4DStart, domain, p4DEnd, pplp4DOutput,
      tp0, tp1, inv, trace, fdet1D_uv, row_p, base, dr, ssize, tCNo, cNo, Disp)  
NDETERMINANT_1D  	ndet1D_uv, ndet1D_du, ndet1D_dv, fdet1D_uv;
POINT4D          	p4DStart, *p4DEnd;
RECT_BOX 		domain;
LIST_POINT4D     	**pplp4DOutput;
POINT2D                 *Disp;
TENSOR_PATCH		tp0, tp1;
int			inv, trace, base, *row_p, dr;
double			ssize;
int			tCNo, *cNo;
{
  POINT4D	  	p4DNext, p4DCurrent, p4DSing, p4DPrev;
  POINT2D	  	p2DNormal, p2DTangent, p2DTP;
  double	  	dTemp, chosen_step_size, minim_step_size;
  double	  	*rgKernel, udiff, vdiff, mindiff, st_separation;
  int		  	i, j, k, fFirst;
  int			m, n;
  int			iDeg0, iDeg1;
  int			jDeg0, jDeg1;
  int			iDir;
  NDETERMINANT_1D       ndet1D_Tmp, ndet1D_u, ndet1D_v;
  NDETERMINANT_1D       fdet1D_Tmp, fdet1D_u, fdet1D_v;
  int			it_cnt, retval;
  int			found_next_point, out_of_bounds;
  int			found_ppoint;

  udiff = domain.i2Dbox[0].high - domain.i2Dbox[0].low;
  vdiff = domain.i2Dbox[1].high - domain.i2Dbox[1].low;
  mindiff = (udiff < vdiff) ? udiff : vdiff;
  /*
  if (ssize < 3*MIN_STEP_SIZE)
    minim_step_size = ssize / 3.0;
  else minim_step_size = MIN_STEP_SIZE;
  */
  minim_step_size = MIN_STEP_SIZE;

  m = tp0.iDeg0;
  n = tp0.iDeg1;
  retval = -1;
  chosen_step_size = ssize;
  /* allocate various vectors */
  ALLOCN(rgKernel, double, ndet1D_uv.cCols);
  iDeg0  = ndet1D_uv.poly[0][0].iDeg0;
  iDeg1  = ndet1D_uv.poly[0][0].iDeg1;
  allocDeterminant_1D(&ndet1D_Tmp,ndet1D_uv.cRows,ndet1D_uv.cCols,iDeg0,iDeg1);
  allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols, 1, iDeg0);
  allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols, 1, iDeg1);

  if (base) {
    jDeg0  = fdet1D_uv.poly[0][0].iDeg0;
    jDeg1  = fdet1D_uv.poly[0][0].iDeg1;
    allocDeterminant_1D(&fdet1D_Tmp,fdet1D_uv.cRows,fdet1D_uv.cCols,jDeg0,jDeg1);
    allocDeterminant_1D(&fdet1D_u, fdet1D_uv.cRows, fdet1D_uv.cCols, 1, jDeg0);
    allocDeterminant_1D(&fdet1D_v, fdet1D_uv.cRows, fdet1D_uv.cCols, 1, jDeg1);
  }

  /* trace in both the directions from each start point */
  for(iDir = 0; iDir < 2; iDir++)
  { 
     /* printf("VALUE OF IDIR IS %d\n", iDir); */
    eval_uv(ndet1D_uv, p4DStart.dCoord[U], p4DStart.dCoord[V]);
    p4DCurrent = p4DStart;
    fFirst = 1;

    found_ppoint = 0;
    /* keep tracing as long as we are inside the patch */
    /* add conditions for loops here later */
    while ((p4DCurrent.dCoord[U] >= (domain.i2Dbox[0].low-NEW_EPS)) && 
	   (p4DCurrent.dCoord[U] <= (domain.i2Dbox[0].high+NEW_EPS)) &&
           (p4DCurrent.dCoord[V] >= (domain.i2Dbox[1].low-NEW_EPS)) && 
	   (p4DCurrent.dCoord[V] <= (domain.i2Dbox[1].high+NEW_EPS)) &&
           (p4DCurrent.dCoord[S] >= -NEW_EPS) && (p4DCurrent.dCoord[S] <= (1+NEW_EPS)) &&
           (p4DCurrent.dCoord[T] >= -NEW_EPS) && (p4DCurrent.dCoord[T] <= (1+NEW_EPS)))
    { /* add point to the list of points on the component */
      found_next_point = 0;
      out_of_bounds = 0;
     while (!found_next_point && chosen_step_size > minim_step_size) {

      /* compute partial derivative matrices */
      eval_uv(ndet1D_du, p4DCurrent.dCoord[U], p4DCurrent.dCoord[V]);
      eval_uv(ndet1D_dv, p4DCurrent.dCoord[U], p4DCurrent.dCoord[V]);

      /* find the direction in which to go next */
      calc_det(ndet1D_uv.eval, ndet1D_du.eval, ndet1D_dv.eval,
               &p4DNext.dCoord[U], &p4DNext.dCoord[V], ndet1D_uv.cRows, 0);
/*
       printf("Printing partial values: %18.12lf %18.12lf %18.12lf %18.12lf\n", 
               p4DCurrent.dCoord[U], p4DCurrent.dCoord[V],
               p4DNext.dCoord[U], p4DNext.dCoord[V]);
*/

#ifdef TRACE_SINGULAR_CHECK
      COMMENT check to make sure that new point is not singular 
      if ((FP_EQ_EPS(p4DNext.dCoord[U], 0.0, 1e-9)) && 
          (FP_EQ_EPS(p4DNext.dCoord[V],0.0, 1e-9)))
      { 
        printf("Trace(): curve singularity; both derivatives zero\n");
	if (fFirst) { p2DTP.u = p2DTP.v = 0.0; }
	if (HandleSingularCase(p4DCurrent, p2DTP, ndet1D_uv, &p4DSing, tp0,
			   tp1, fdet1D_uv, row_p, base) == -1) {
	  printf("Could not handle singularities properly\n");
          return;
	}
	p4DCurrent = p4DSing;
	eval_uv(ndet1D_uv, p4DCurrent.dCoord[U], p4DCurrent.dCoord[V]);
	add_list_point(pplp4DOutput, p4DCurrent, tp1, inv, trace, dr);
	eval_uv(ndet1D_du, p4DCurrent.dCoord[U], p4DCurrent.dCoord[V]);
	eval_uv(ndet1D_dv, p4DCurrent.dCoord[U], p4DCurrent.dCoord[V]);
	calc_det(ndet1D_uv.eval, ndet1D_du.eval, ndet1D_dv.eval,
		 &p4DNext.dCoord[U], &p4DNext.dCoord[V], ndet1D_uv.cRows, 0);
	if ((FP_EQ(p4DNext.dCoord[U], 0.0)) &&
	    (FP_EQ(p4DNext.dCoord[V],0.0))) {
	  printf("Tracing becoming Numerically Unstable. Returning ...\n");
	  return;
	}
      }
#endif

      /* normalize p4DNext.dCoord[U] and p4DNext.dCoord[V] before using */
      dTemp = 1.0/sqrt(p4DNext.dCoord[U]*p4DNext.dCoord[U] + 
		       p4DNext.dCoord[V]*p4DNext.dCoord[V]);

      /* we must make sure that the normal is continuous -- there are two
         "normals" to every curve -- we don't care which one we get, we just
         want it to be continuous, so check dot product !<0  (first time through
         will randomly choose the orientation) */

      p4DNext.dCoord[U]*=dTemp;
      p4DNext.dCoord[V]*=dTemp;
      if (!fFirst) {
	  dTemp = p2DNormal.u*p4DNext.dCoord[U] + p2DNormal.v*p4DNext.dCoord[V];
	  if (dTemp < 0.0) {
	      p4DNext.dCoord[U]= -p4DNext.dCoord[U];
	      p4DNext.dCoord[V]= -p4DNext.dCoord[V];
	  }
      }
      else 
	fFirst=0;

      p2DNormal.u = p4DNext.dCoord[U];
      p2DNormal.v = p4DNext.dCoord[V];

      /* choose the direction of the tangent here */
      p2DTangent.u = (iDir)? p2DNormal.v : -p2DNormal.v;
      p2DTangent.v = (iDir)? -p2DNormal.u : p2DNormal.u;
      p2DTP = p2DTangent;

      if (fabs(p2DTangent.u) > fabs(p2DTangent.v))
      { 
/*
	if ((chosen_step_size > Disp->u) && (Disp->u > 0.01))
	  chosen_step_size = 0.3*Disp->u;
*/
	p4DNext.dCoord[U] = p4DCurrent.dCoord[U] + chosen_step_size * p2DTangent.u;
        p4DNext.dCoord[V] = p4DCurrent.dCoord[V];
      }
      else
      { 
/*
	if ((chosen_step_size > Disp->v) && (Disp->v > 0.01))
	  chosen_step_size = 0.3*Disp->v;
*/
	p4DNext.dCoord[U] = p4DCurrent.dCoord[U];
        p4DNext.dCoord[V] = p4DCurrent.dCoord[V] + chosen_step_size * p2DTangent.v;
      }
      p4DNext.dCoord[S] = p4DCurrent.dCoord[S];
      p4DNext.dCoord[T] = p4DCurrent.dCoord[T];
  
      if ((p4DNext.dCoord[U] >= (domain.i2Dbox[0].low-NEW_EPS)) && 
	  (p4DNext.dCoord[U] <= (domain.i2Dbox[0].high+NEW_EPS)) &&
          (p4DNext.dCoord[V] >= (domain.i2Dbox[1].low-NEW_EPS)) && 
	  (p4DNext.dCoord[V] <= (domain.i2Dbox[1].high+NEW_EPS)) &&
          (p4DNext.dCoord[S] >= -NEW_EPS) && (p4DNext.dCoord[S] <= (1+NEW_EPS)) &&
          (p4DNext.dCoord[T] >= -NEW_EPS) && (p4DNext.dCoord[T] <= (1+NEW_EPS)))
      { 

        eval_uv(ndet1D_uv, p4DNext.dCoord[U], p4DNext.dCoord[V]);
/************** a bit iffy, but am taking it out for now ********
        if (!matrix_singular(ndet1D_uv,rgKernel))
*****************************************************************/
        copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

        for(i = 0; i < ndet1D_Tmp.cRows; i++)
	  for(j = 0; j < ndet1D_Tmp.cCols; j++)
	    bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);
	      
	if (base) {
          copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);

          for(i = 0; i < fdet1D_Tmp.cRows; i++)
	    for(j = 0; j < fdet1D_Tmp.cCols; j++)
	      bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
 	}
	      
        if (p4DNext.dCoord[V] == p4DCurrent.dCoord[V])
        { 
	  compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, p4DNext.dCoord[U]);
	  if (base)
	    compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, p4DNext.dCoord[U]);
  
  	  /* guess for location of next point is p4DNext.dCoord[U],
	     p4DNext.V + stepsize*normalized tangent in v dir */
	  find_next_point(ndet1D_v, m, n, &p4DNext, V,
	                  p4DNext.dCoord[V] + chosen_step_size * p2DTangent.v, 
	                  tp0, tp1, fdet1D_v, row_p, base, &it_cnt);
/* Added on 7/20/95 by SK to make sure that two adjacent points are */
/* within distance chosen_step_size on both the patches */
/*
	  st_separation = (p4DNext.dCoord[S] - p4DCurrent.dCoord[S])*
	                  (p4DNext.dCoord[S] - p4DCurrent.dCoord[S]) +
			  (p4DNext.dCoord[T] - p4DCurrent.dCoord[T])*
	                  (p4DNext.dCoord[T] - p4DCurrent.dCoord[T]);
			   
	  if ((it_cnt == ITER_LIM) || 
	      (st_separation > chosen_step_size*chosen_step_size)) {
*/
	  if (it_cnt == ITER_LIM) {
	    printf("Reducing step size ...\n");
	    chosen_step_size /= 2.0;
	    if (found_ppoint) {
	      p4DCurrent = p4DPrev;
	      (*pplp4DOutput) = (*pplp4DOutput)->plp4DNext;
	      found_ppoint = 0;
	    }
	    eval_uv(ndet1D_uv,p4DCurrent.dCoord[U],p4DCurrent.dCoord[V]);
	  } else { found_next_point = 1; found_ppoint = 1; }
        }
        else
        { 
	  compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, p4DNext.dCoord[V]);
	  if (base)
	    compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, p4DNext.dCoord[V]);

	  find_next_point(ndet1D_u, m, n, &p4DNext, U,
	                  p4DNext.dCoord[U] + chosen_step_size * p2DTangent.u, 
	                  tp0, tp1, fdet1D_u, row_p, base, &it_cnt);
/* Added on 7/20/95 by SK to make sure that two adjacent points are */
/* within distance chosen_step_size on both the patches */
/*
	  st_separation = (p4DNext.dCoord[S] - p4DCurrent.dCoord[S])*
	                  (p4DNext.dCoord[S] - p4DCurrent.dCoord[S]) +
			  (p4DNext.dCoord[T] - p4DCurrent.dCoord[T])*
	                  (p4DNext.dCoord[T] - p4DCurrent.dCoord[T]);
			   
	  if ((it_cnt == ITER_LIM) || (IsPtInDomain(&p4DNext, &domain,
						    NEW_EPS) &&
	      (st_separation > chosen_step_size*chosen_step_size))) {
*/
	  if (it_cnt == ITER_LIM) {
	    printf("Reducing step size ...\n");
	    chosen_step_size /= 2.0;
	    if (found_ppoint) {
	      p4DCurrent = p4DPrev;
	      (*pplp4DOutput) = (*pplp4DOutput)->plp4DNext;
	      found_ppoint = 0;
	    }
	    eval_uv(ndet1D_uv,p4DCurrent.dCoord[U],p4DCurrent.dCoord[V]);
	  } else { found_next_point = 1; found_ppoint = 1; }
        }
	if (found_next_point) {
          eval_uv(ndet1D_uv, p4DNext.dCoord[U], p4DNext.dCoord[V]);

	  if (trace && (p4DNext.dCoord[U]<p4DCurrent.dCoord[U])) {
	    break;
	  }
          if (IsPtInDomain(&p4DNext, &domain, NEW_EPS)) {
	    if (!EQ_4D_EPS(p4DCurrent, p4DNext, 1e-4)) {
              add_list_point(pplp4DOutput, p4DCurrent, tp1, inv, trace, dr);
	      *p4DEnd = p4DCurrent;
  	      if (tCNo == 1) { (*cNo) += 1; tCNo = 0; }
	      retval = 0;
	    }
	  } else out_of_bounds = 1;
	  p4DPrev = p4DCurrent;
          p4DCurrent = p4DNext;
	}
      }  
      else { out_of_bounds = 1; break; }
     }
     if (out_of_bounds) break;
     if (!found_next_point && !out_of_bounds) {
       printf("Illconditioned Problem inspite of reducing stepsize\n");
       return;
     }
    }
  }

  /* free up all allocated mem */
  if (base) {
    free_alloc_det_1D(&fdet1D_Tmp);
    free_alloc_det_1D(&fdet1D_u);
    free_alloc_det_1D(&fdet1D_v);
  }
  free_alloc_det_1D(&ndet1D_Tmp);
  free_alloc_det_1D(&ndet1D_u);
  free_alloc_det_1D(&ndet1D_v);
  myfree(rgKernel);
  return retval;
}

#ifdef TRACE_SINGULAR_CHECK
HandleSingularCase(p4DC, p2DTP, ndet1D_uv, p4DNew, tpPatch0, 
                   tpPatch1, fdet1D_uv, rp, base)
POINT4D  p4DC, *p4DNew;
POINT2D  p2DTP;
NDETERMINANT_1D ndet1D_uv, fdet1D_uv;
TENSOR_PATCH   tpPatch0, tpPatch1;
int *rp, base;
{
  NDETERMINANT_1D ndet1D_Tmp, ndet1D_u, ndet1D_v;
  NDETERMINANT_1D fdet1D_Tmp, fdet1D_u, fdet1D_v;
  LIST_POINT4D *SingList, *TmpElmt;
  int i, j;
  int iD0, iD1;
  int                  num_points, num_cpoints;
  POINT4D              *pdPoints;
  POINT4CPX            *pcpxPoints;
  double u_val, v_val, tgt_val, tgt_est;
  double diff_u, diff_v;

  iD0 = tpPatch0.iDeg0;
  iD1 = tpPatch0.iDeg1;

   i = (tpPatch1.iDeg0 > tpPatch1.iDeg1)? tpPatch1.iDeg0: tpPatch1.iDeg1;

   ALLOCN(pdPoints, POINT4D, i*ndet1D_uv.cRows);

   ALLOCN(pcpxPoints, POINT4CPX, i*ndet1D_uv.cRows);

  allocDeterminant_1D(&ndet1D_Tmp, ndet1D_uv.cRows, ndet1D_uv.cCols,
                       tpPatch1.iDeg0, tpPatch1.iDeg1);

  allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols,
                       0, tpPatch1.iDeg0);

  allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols,
                       0, tpPatch1.iDeg1);

  if (base) {
    allocDeterminant_1D(&fdet1D_Tmp, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        tpPatch1.iDeg0, tpPatch1.iDeg1);

    allocDeterminant_1D(&fdet1D_u, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg0);

    allocDeterminant_1D(&fdet1D_v, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg1);
  }

   /* copy ndet1D_uv to ndet1D_Tmp */
   copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

   /* convert bernstein basis to power basis for each entry of Tmp */
   for(i = 0; i < ndet1D_Tmp.cRows; i++)
     for(j = 0; j < ndet1D_Tmp.cCols; j++)
       bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);

   if (base) {
     /* copy fdet1D_uv to fdet1D_Tmp */
     copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);

     /* convert bernstein basis to power basis for each entry of Tmp */
     for(i = 0; i < fdet1D_Tmp.cRows; i++)
       for(j = 0; j < fdet1D_Tmp.cCols; j++)
         bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
   }
  
  u_val = p4DC.dCoord[U];
  v_val = p4DC.dCoord[V];
  SingList = NULL;

   /* if (p2DTP.v < 0.0) { */
    compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, v_val-Step_Size_Out);
    if (base)
      compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, v_val-Step_Size_Out);
    if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			pcpxPoints,U,power_pt(v_val-Step_Size_Out),
			fdet1D_u, rp, iD0, iD1, base,
			tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i = 0; i < num_points; i++) {
     diff_u = inverse_power_pt(pdPoints[i].dCoord[U]) - u_val;
     diff_v = inverse_power_pt(pdPoints[i].dCoord[V]) - v_val;
     if ((fabs(diff_u) <= 2*Step_Size_Out) &&
         (fabs(diff_v) <= 2*Step_Size_Out))
      addPointToList(pdPoints[i], &SingList); 
    }
   /* } else { */
    compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, v_val+Step_Size_Out);
    if (base)
      compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, v_val+Step_Size_Out);
    if (find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
			pcpxPoints,U,power_pt(v_val+Step_Size_Out),
			fdet1D_u, rp, iD0, iD1, base,
			tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i = 0; i < num_points; i++) {
     diff_u = inverse_power_pt(pdPoints[i].dCoord[U]) - u_val;
     diff_v = inverse_power_pt(pdPoints[i].dCoord[V]) - v_val;
     if ((fabs(diff_u) <= 2*Step_Size_Out) &&
         (fabs(diff_v) <= 2*Step_Size_Out))
      addPointToList(pdPoints[i], &SingList); 
    }
   /* } */

   /* if (p2DTP.u < 0.0) { */
    compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, u_val-Step_Size_Out);
    if (base)
      compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, u_val-Step_Size_Out);
    if (find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
			pcpxPoints,V,power_pt(u_val-Step_Size_Out),
			fdet1D_v, rp, iD0, iD1, base,
			tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i = 0; i < num_points; i++) {
     diff_u = inverse_power_pt(pdPoints[i].dCoord[U]) - u_val;
     diff_v = inverse_power_pt(pdPoints[i].dCoord[V]) - v_val;
     if ((fabs(diff_u) <= 2*Step_Size_Out) &&
         (fabs(diff_v) <= 2*Step_Size_Out))
      addPointToList(pdPoints[i], &SingList); 
    }

   /* } else { */
    compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, u_val+Step_Size_Out);
    if (base)
      compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, u_val+Step_Size_Out);
    if (find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
  			pcpxPoints,V,power_pt(u_val+Step_Size_Out),
			fdet1D_v, rp, iD0, iD1, base,
			tpPatch0, tpPatch1, 0) == -1)
      return -1;
    for (i = 0; i < num_points; i++) {
     diff_u = inverse_power_pt(pdPoints[i].dCoord[U]) - u_val;
     diff_v = inverse_power_pt(pdPoints[i].dCoord[V]) - v_val;
     if ((fabs(diff_u) <= 2*Step_Size_Out) &&
         (fabs(diff_v) <= 2*Step_Size_Out))
      addPointToList(pdPoints[i], &SingList); 
    }
   /* } */

  if (SingList == (LIST_POINT4D *)0) {
    printf("Something Wrong!! Could not find neighborhood point\n");
    return -1;
  }
  if (SingList->plp4DNext == (LIST_POINT4D *)0) {
    *p4DNew = SingList->p4D;
    return 0;
  }
  else {
    if ((p2DTP.u == 0.0) && (p2DTP.v == 0.0)) {
      *p4DNew = SingList->p4D;
      return 0;
    }
     /* tgt_val = p2DTP.v / p2DTP.u; */
    for (TmpElmt = SingList; TmpElmt; TmpElmt = SingList) {
      SingList = SingList->plp4DNext;
      diff_u = (TmpElmt->p4D).dCoord[U] - u_val;
      diff_v = (TmpElmt->p4D).dCoord[V] - v_val;
      tgt_val = sqrt(diff_u*diff_u + diff_v*diff_v);
      diff_u /= tgt_val;
      diff_v /= tgt_val;
       /* tgt_est = diff_v / diff_u; */
      if ((FP_EQ_EPS(p2DTP.u, diff_u, 9e-2)) &&
          (FP_EQ_EPS(p2DTP.v, diff_v, 9e-2))) {
        *p4DNew = TmpElmt->p4D;
	return 0;
      }
    }
  }

  return -1;
}
#endif
         
/* expects ndet1D_u.poly in power basis, p4Duvst in bernstein,
   m = deg of patch in u, n = degree of patch in v, 
   dRootGuess in bernstein_basis, and returns
   the new point in p4Duvst in bernstein  basis */

#ifdef NAIVE_POWER_METHOD
find_next_pointAllEval(ndet1D_u, m, n, p4Duvst, iPolyIn, dRootGuess )
NDETERMINANT_1D		ndet1D_u;
int			iPolyIn, m, n;
POINT4D			*p4Duvst;
double                  dRootGuess;
{
   int			num_points;
   POINT4D		*pdPoints;
   int			i, iSize, iClosestRoot;
   int			iConstIn = (iPolyIn == U)? V: U;
   double               dPolyOrig, dMinDiff, dRoot;
   double		dir_of_guess, dtemp;

   iSize = ndet1D_u.cRows*ndet1D_u.poly[0][0].iDeg1;

   dPolyOrig = p4Duvst->dCoord[iPolyIn]; 
  
   dir_of_guess = dRootGuess - dPolyOrig;

   ALLOCN(pdPoints, POINT4D, iSize);

   /* find_start_points just takes a polynomial, makes companion
      matrix, solves for all eigenvalues, and then returns an the
      real (not complex) roots (eigenvalues) in pdPoints 
      (intersection of line || to iPolyIn axis at iConstIn = 
      power_pt(p4Duvst[iConstIn]) and the curve in ndet1D_u
      (ndet1d_u already has curve filled in by compute_curve1D) */

/*
   printf("Just before calling FIND_START_POINTS from find_next_pointAllEval\n");
*/
   find_start_points( ndet1D_u, &num_points, pdPoints, iPolyIn, 
		     power_pt(p4Duvst->dCoord[iConstIn]), m );

   /* now we assume that the closest to the current value is the correct
      one, or closest to the guessed value */

   /* should actually only look in the direction of the guess */
   for (i=0,iClosestRoot=-1, dMinDiff=HUGE_VAL;i<num_points;i++) {
       dRoot = inverse_power_pt(pdPoints[i].dCoord[iPolyIn]); 
 /* 
   printf("uvst bernstein (%lf %lf %lf %lf)\n", 
inverse_power_pt(pdPoints[i].dCoord[U]),inverse_power_pt(pdPoints[i].dCoord[V]),
inverse_power_pt(pdPoints[i].dCoord[S]),inverse_power_pt(pdPoints[i].dCoord[T]));
*/
       /*
       if (((dir_of_guess > 0) && (dRoot > dPolyOrig)) ||
	   ((dir_of_guess < 0) && (dRoot < dPolyOrig)))
       */
       { dtemp  = SQ(p4Duvst->dCoord[0]-inverse_power_pt(pdPoints[i].dCoord[0]))
		 +SQ(p4Duvst->dCoord[1]-inverse_power_pt(pdPoints[i].dCoord[1]))
		 +SQ(p4Duvst->dCoord[2]-inverse_power_pt(pdPoints[i].dCoord[2]))
		 +SQ(p4Duvst->dCoord[3]-inverse_power_pt(pdPoints[i].dCoord[3]));
	 if (dtemp < dMinDiff) 
	 { dMinDiff = dtemp;
	   iClosestRoot = i;
         }
      }
   }
 
/*
   printf("\nOriginal value is %lf, guessed value is %lf, closest root is %lf.\n",
	 dPolyOrig, dRootGuess, inverse_power_pt(pdPoints[iClosestRoot].dCoord[iPolyIn])); 
*/
   
   *p4Duvst=pdPoints[iClosestRoot];

   /* convert all values back into Bernstein-basis space */
   for(i = 0; i < 4; i++)
       p4Duvst->dCoord[i] = inverse_power_pt(p4Duvst->dCoord[i]);

   myfree(pdPoints);
}
#endif

#ifdef ALG_PRUNE
prune(ndet1D_u, m, n, iPolyIn, other_const, tp0, tp1, fdet1D_u, row_p,
      base, low, high, num_points, pdPoints)
NDETERMINANT_1D		ndet1D_u, fdet1D_u;
int			iPolyIn, m, n;
int			base, *row_p;
TENSOR_PATCH		tp0, tp1;
int                     *num_points;
POINT4D                 *pdPoints;
double                  low, high, other_const;
{ 
  int i, j, k;
  int it_cnt, iConstIn;
  POINT4D ret_point;
  double guess, dist;

  iConstIn = (iPolyIn == U) ? V : U;
  
  if (low > high - 1e-4) {
    printf("Interval [low,high] is incorrect\n");
    exit(-1);
  }

  guess = 0.5*(low + high);
  ret_point.dCoord[iConstIn] = inverse_power_pt(other_const);
  ret_point.dCoord[iPolyIn] = guess;
  ret_point.dCoord[S] = ret_point.dCoord[T] = 0.0;

  find_closest_point(ndet1D_u, m, n, &ret_point, iPolyIn, guess, tp0,
		     tp1, fdet1D_u, row_p, base, &it_cnt);

  if (it_cnt >= ITER_LIM) {
    printf("Inverse Iteration could not converge. Trying with different guess value\n");
    guess = 0.6*(low + high);
    ret_point.dCoord[iConstIn] = other_const;
    ret_point.dCoord[iPolyIn] = guess;
    ret_point.dCoord[S] = ret_point.dCoord[T] = 0.0;
    
    find_closest_point(ndet1D_u, m, n, &ret_point, iPolyIn, guess, tp0,
		       tp1, fdet1D_u, row_p, base, &it_cnt);
    if (it_cnt >= ITER_LIM) {
      printf("Could not converge second try also.\n");
      return;
    }
  }

  if (ret_point.dCoord[iPolyIn] > (low - 1e-4) &&
      ret_point.dCoord[iPolyIn] < (high + 1e-4)) {
    for (i=0; i<4; i++)
      pdPoints[(*num_points)++].dCoord[i] = power_pt(ret_point.dCoord[i]);

    dist = fabs(guess - ret_point.dCoord[iPolyIn]);

    if ((guess - dist) > (low + 1e-4)) 
      prune(ndet1D_u, m, n, iPolyIn, other_const, tp0, tp1, fdet1D_u,
	    row_p, base, low, guess - dist - 1e-2, num_points, pdPoints);

    if ((guess + dist) < (high - 1e-4)) 
      prune(ndet1D_u, m, n, iPolyIn, other_const, tp0, tp1, fdet1D_u,
	    row_p, base, guess + dist + 1e-2, high, num_points, pdPoints);
  }
  return;
}

/* expects ndet1D_u.poly in power basis, p4Duvst in bernstein,
   m = deg of patch in u, n = degree of patch in v,
   dRootGuess is in bernstein_basis, and returns
   the new point in p4Duvst in bernstein  basis */

find_closest_point(ndet1D_u, m, n, p4Duvst, iPolyIn, dRootGuess, 
		tp0, tp1, fdet1D_u, row_p, base, iter_cnt )
NDETERMINANT_1D		ndet1D_u, fdet1D_u;
int			iPolyIn, m, n;
POINT4D			*p4Duvst;
double                  dRootGuess;
int			*iter_cnt, base, *row_p;
TENSOR_PATCH		tp0, tp1;
{ 
  int			i, j, iSize;
  double		**rgrgB, **rgrgA_plus_B;
  double		*rgEVec;
  double		iPolyIn_val, iConstVal;
  int			iConstIn = (iPolyIn == U)? V: U;
  double		sval, tval, fret, fval;
  double		xval1, yval1, zval1;
  double		xval2, yval2, zval2;
  double 		*pow_coord, **dirMat;
  int			nIter;
  double		a, b, c, d;

  iSize = ndet1D_u.cRows*ndet1D_u.poly[0][0].iDeg1;

  ALLOCMAT(rgrgB, iSize, iSize, double);
  ALLOCMAT(rgrgA_plus_B, iSize, iSize, double);
  ALLOCN(rgEVec, double, iSize);

  /* convert all values into power-basis space */
  for(i = 0; i < 4; i++)
    p4Duvst->dCoord[i] = power_pt(p4Duvst->dCoord[i]);

  iPolyIn_val = p4Duvst->dCoord[iPolyIn];

  const_rand_kernel_vec(rgEVec, iSize);

  form_power_companion_matrix(ndet1D_u, rgrgB, rgrgA_plus_B, iSize);

  /* get the new values of s, t, u, v */
  generalized_power_iterate(rgrgB, rgrgA_plus_B, rgEVec, iSize,
			    ndet1D_u.cRows, dRootGuess, &iPolyIn_val,
			    m, iter_cnt, base);  
  if (base || FP_EQ_EPS(rgEVec[0], 0.0, 1e-4) ||
      (*iter_cnt == 0)) {
    iConstVal = inverse_power_pt(p4Duvst->dCoord[iConstIn]);
    extract_st_val(tp0, tp1, iPolyIn_val, iPolyIn, iConstVal, 
		   &sval, &tval);
    p4Duvst->dCoord[iConstIn] = iConstVal;
    p4Duvst->dCoord[iPolyIn] = iPolyIn_val;
    p4Duvst->dCoord[S] = sval;
    p4Duvst->dCoord[T] = tval;
  }
  else {
    p4Duvst->dCoord[iPolyIn]= power_pt(iPolyIn_val);
    p4Duvst->dCoord[S]= rgEVec[1]/rgEVec[0];
    p4Duvst->dCoord[T]= rgEVec[m]/rgEVec[0];

  /* convert all values back into Bernstein-basis space */
    for(i = 0; i < 4; i++)
      p4Duvst->dCoord[i] = inverse_power_pt(p4Duvst->dCoord[i]);
  }

  if (base && sval == -1 && tval == -1) {
     /* printf("FIND_NEXT_POINT: extract_st_val could not return good vals\n"); */
    return;
  }

  EvalPointOnPatch(tp0, p4Duvst->dCoord[S], p4Duvst->dCoord[T],
		   &xval1, &yval1, &zval1);
  EvalPointOnPatch(tp1, p4Duvst->dCoord[U], p4Duvst->dCoord[V],
		   &xval2, &yval2, &zval2);
  fval = (xval2 - xval1) * (xval2 - xval1) +
	 (yval2 - yval1) * (yval2 - yval1) +
	 (zval2 - zval1) * (zval2 - zval1);
  if (fval > 1.0e-6) {
    pow_tp0 = tp0;
    pow_tp1 = tp1;
    ALLOCN(pow_coord, double, 3); /* 3 is the dim. of minim. problem */
    ALLOCMAT(dirMat, 3, 3, double);
    if (iPolyIn == U) {
      unchanged = V;
      unchanged_val = p4Duvst->dCoord[V];
      pow_coord[0] = p4Duvst->dCoord[0];
      pow_coord[1] = p4Duvst->dCoord[2];
      pow_coord[2] = p4Duvst->dCoord[3];
    } else {
      unchanged = U;
      unchanged_val = p4Duvst->dCoord[U];
      pow_coord[0] = p4Duvst->dCoord[1];
      pow_coord[1] = p4Duvst->dCoord[2];
      pow_coord[2] = p4Duvst->dCoord[3];
    }
    for (i=0; i<3; i++) 
      for (j=0; j<3; j++)
        if (i == j) dirMat[i][j] = 1.0;
        else dirMat[i][j] = 0.0;
    powell(pow_coord, dirMat, 3, POW_TOL, &nIter, &fret, RefinePoint1D);
    if (fret > 1.0e-6) {
      p4Duvst->dCoord[S] = p4Duvst->dCoord[T] = -1;
    } else {
      if (iPolyIn == U) {
        p4Duvst->dCoord[0] = pow_coord[0]; 
        p4Duvst->dCoord[1] = unchanged_val; 
        p4Duvst->dCoord[2] = pow_coord[1]; 
        p4Duvst->dCoord[3] = pow_coord[2]; 
      } else {
        p4Duvst->dCoord[0] = unchanged_val; 
        p4Duvst->dCoord[1] = pow_coord[0]; 
        p4Duvst->dCoord[2] = pow_coord[1]; 
        p4Duvst->dCoord[3] = pow_coord[2]; 
      }
    }
    myfree(pow_coord);
    FREEMAT(dirMat, 3, 3);
  }

  FREEMAT(rgrgB, iSize, iSize);
  FREEMAT(rgrgA_plus_B, iSize, iSize);
  myfree(rgEVec);
}
#endif

/* expects ndet1D_u.poly in power basis, p4Duvst in bernstein,
   m = deg of patch in u, n = degree of patch in v,
   dRootGuess is in bernstein_basis, and returns
   the new point in p4Duvst in bernstein  basis */

find_next_point(ndet1D_u, m, n, p4Duvst, iPolyIn, dRootGuess, 
		tp0, tp1, fdet1D_u, row_p, base, iter_cnt )
NDETERMINANT_1D		ndet1D_u, fdet1D_u;
int			iPolyIn, m, n;
POINT4D			*p4Duvst;
double                  dRootGuess;
int			*iter_cnt, base, *row_p;
TENSOR_PATCH		tp0, tp1;
{ 
  int			i, j, iSize;
  double		**rgrgB, **rgrgA_plus_B;
  double		*rgEVec;
  double		iPolyIn_val, iConstVal;
  int			iConstIn = (iPolyIn == U)? V: U;
  double		sval, tval, fret, fval;
  double		xval1, yval1, zval1;
  double		xval2, yval2, zval2;
  double 		*pow_coord, **dirMat;
  int			nIter;
  double		a, b, c, d;

  iSize = ndet1D_u.cRows*ndet1D_u.poly[0][0].iDeg1;

  ALLOCMAT(rgrgB, iSize, iSize, double);
  ALLOCMAT(rgrgA_plus_B, iSize, iSize, double);
  ALLOCN(rgEVec, double, iSize);

  a = b = c = d = 0.0;
  while (FP_EQ_EPS(a, 0.0, 1e-4))
    a = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(b, 0.0, 1e-4))
    b = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(c, 0.0, 1e-4))
    c = 0.7 + 0.3*drand48();
  while (FP_EQ_EPS(d, 0.0, 1e-4))
    d = 0.7 + 0.3*drand48();

  /* convert all values into power-basis space */
  for(i = 0; i < 4; i++)
    p4Duvst->dCoord[i] = power_pt(p4Duvst->dCoord[i]);

  iPolyIn_val = p4Duvst->dCoord[iPolyIn];

  /* if curve is currently perfectly horiz or vert, just guess a little
     in one direction */

/*  if (FP_EQ(dRootGuess-iPolyIn_val,0.0))
      dRootGuess += 0.001;
*/

  if (base || FP_EQ(dRootGuess,1.0) || (p4Duvst->dCoord[S] == HUGE_VAL) 
           || (p4Duvst->dCoord[T] == HUGE_VAL))
    const_rand_kernel_vec(rgEVec, iSize);
  else
    construct_kernel_vec(ndet1D_u.poly[0][0].iDeg1, m, n, 
		       dRootGuess/(1 - dRootGuess), p4Duvst->dCoord[S], 
		       p4Duvst->dCoord[T], rgEVec);

  if (1) /* (form_companion_matrix(ndet1D_u, rgrgB, iSize, a, b, c, d) == -1) */ {
    form_power_companion_matrix(ndet1D_u, rgrgB, rgrgA_plus_B, iSize);

    /* get the new values of s, t, u, v */
    generalized_power_iterate(rgrgB, rgrgA_plus_B, rgEVec, iSize, ndet1D_u.cRows, 
			      dRootGuess, &iPolyIn_val, m, iter_cnt, base); 
  } else {
    dRootGuess = power_pt(dRootGuess);
    dRootGuess = (b - d*dRootGuess) / (c*dRootGuess - a);
    power_iterate(rgrgB, rgEVec, iSize, ndet1D_u.cRows, dRootGuess,
		  &iPolyIn_val, m, iter_cnt, base);
     /* iPolyIn_val = trans_inverse_power_pt(iPolyIn_val, a, b, c, d);
*/
    iPolyIn_val = (a*iPolyIn_val+b)/((a+c)*iPolyIn_val+(b+d));
  }

  if (base || FP_EQ_EPS(rgEVec[0], 0.0, 1e-4) ||
      (*iter_cnt == 0)) {
    iConstVal = inverse_power_pt(p4Duvst->dCoord[iConstIn]);
    extract_st_val(tp0, tp1, iPolyIn_val, iPolyIn, iConstVal, 
		   &sval, &tval);
    p4Duvst->dCoord[iConstIn] = iConstVal;
    p4Duvst->dCoord[iPolyIn] = iPolyIn_val;
    p4Duvst->dCoord[S] = sval;
    p4Duvst->dCoord[T] = tval;
  }
  else {
    p4Duvst->dCoord[iPolyIn]= power_pt(iPolyIn_val);
    p4Duvst->dCoord[S]= rgEVec[1]/rgEVec[0];
    p4Duvst->dCoord[T]= rgEVec[m]/rgEVec[0];
/*
    if ((p4Duvst->dCoord[S] <= -EPSILON) || (p4Duvst->dCoord[S] >= 1+EPSILON))
      p4Duvst->dCoord[S] = HUGE_VAL;
    if ((p4Duvst->dCoord[T] <= -EPSILON) || (p4Duvst->dCoord[T] >= 1+EPSILON))
      p4Duvst->dCoord[T] = HUGE_VAL;
*/

  /* convert all values back into Bernstein-basis space */
    for(i = 0; i < 4; i++)
      p4Duvst->dCoord[i] = inverse_power_pt(p4Duvst->dCoord[i]);
  }

  if (base && sval == -1 && tval == -1) {
     /* printf("FIND_NEXT_POINT: extract_st_val could not return good vals\n"); */
    return;
  }

  EvalPointOnPatch(tp0, p4Duvst->dCoord[S], p4Duvst->dCoord[T],
		   &xval1, &yval1, &zval1);
  EvalPointOnPatch(tp1, p4Duvst->dCoord[U], p4Duvst->dCoord[V],
		   &xval2, &yval2, &zval2);
  fval = (xval2 - xval1) * (xval2 - xval1) +
	 (yval2 - yval1) * (yval2 - yval1) +
	 (zval2 - zval1) * (zval2 - zval1);
  if (fval > 1.0e-6) {
    pow_tp0 = tp0;
    pow_tp1 = tp1;
    ALLOCN(pow_coord, double, 3); /* 3 is the dim. of minim. problem */
    ALLOCMAT(dirMat, 3, 3, double);
    if (iPolyIn == U) {
      unchanged = V;
      unchanged_val = p4Duvst->dCoord[V];
      pow_coord[0] = p4Duvst->dCoord[0];
      pow_coord[1] = p4Duvst->dCoord[2];
      pow_coord[2] = p4Duvst->dCoord[3];
    } else {
      unchanged = U;
      unchanged_val = p4Duvst->dCoord[U];
      pow_coord[0] = p4Duvst->dCoord[1];
      pow_coord[1] = p4Duvst->dCoord[2];
      pow_coord[2] = p4Duvst->dCoord[3];
    }
    for (i=0; i<3; i++) 
      for (j=0; j<3; j++)
        if (i == j) dirMat[i][j] = 1.0;
        else dirMat[i][j] = 0.0;
    powell(pow_coord, dirMat, 3, POW_TOL, &nIter, &fret, RefinePoint1D);
    if (fret > 1.0e-6) {
      p4Duvst->dCoord[S] = p4Duvst->dCoord[T] = -1;
    } else {
      if (iPolyIn == U) {
        p4Duvst->dCoord[0] = pow_coord[0]; 
        p4Duvst->dCoord[1] = unchanged_val; 
        p4Duvst->dCoord[2] = pow_coord[1]; 
        p4Duvst->dCoord[3] = pow_coord[2]; 
      } else {
        p4Duvst->dCoord[0] = unchanged_val; 
        p4Duvst->dCoord[1] = pow_coord[0]; 
        p4Duvst->dCoord[2] = pow_coord[1]; 
        p4Duvst->dCoord[3] = pow_coord[2]; 
      }
    }
    myfree(pow_coord);
    FREEMAT(dirMat, 3, 3);
  }

    FREEMAT(rgrgB, iSize, iSize);
    FREEMAT(rgrgA_plus_B, iSize, iSize);
    myfree(rgEVec);
}


/* constructed as (1, s, s^2 .... s^m, st, s^2t, .....s^mt^2n) */
/* refer determinant.c for further details, also, dLamda is how far
from the current eigenvalue you expect the new one to be (sign and
direction) (in power basis).
*/
construct_kernel_vec(cDegCurve, m, n, dLambda, s, t, rgEVec)
int			cDegCurve, m, n;
double			dLambda, s, t;
double			*rgEVec;
{ 
  int			i, j, k, l;
  double		*pow_s, *pow_t, *pow_lambda;

  ALLOCN(pow_s, double, m);
  ALLOCN(pow_t, double, 2*n);
  ALLOCN(pow_lambda, double, cDegCurve);

  for(i = 1, pow_s[0] = 1; i < m; i++) 
 pow_s[i] = s*pow_s[i-1];
  for(i = 1, pow_t[0] = 1; i < 2*n; i++)
 pow_t[i] = t*pow_t[i-1];
  for(i = 1, pow_lambda[0] = 1; i < cDegCurve; i++)
 pow_lambda[i] = dLambda*pow_lambda[i-1];
  
  for (l = k = 0; l < cDegCurve; l++)
	 for(i = 0; i < 2*n; i++)
	  for(j = 0; j < m; j++)
		 rgEVec[k++] = pow_lambda[l]*pow_s[j]*pow_t[i];
  
  myfree(pow_s); myfree(pow_t); myfree(pow_lambda);
}

const_rand_kernel_vec(rgEVec, size)
double		*rgEVec;
int		size;
{
  double rnd_val;
  double pow_val, mag;
  int i;

  mag = 0.0;
  pow_val = MAX_RAND_VAL;
  while (1) {
    mag = 0.0;
    for (i=0; i<size; i++) {
      rnd_val = (double) drand48();
      rgEVec[i] = rnd_val;
      mag += rgEVec[i] * rgEVec[i];
    }
    mag = sqrt(mag);
    if (!FP_EQ_EPS(mag,0.0,1e-10)) break;
  }
}

/* assumes that rgrgC has been allocated outside */
int form_companion_matrix(ndet1D_u, rgrgC, iSize_C, a, b, c, d)
NDETERMINANT_1D		ndet1D_u;	
double			**rgrgC;
int			iSize_C;
double			a, b, c, d;
{ 
  int			i, j, k, l, n;
  int			p, q, r;
  double		***rgrgM_i;
  double		***rgrgM_j;
  double		**rgrgMd, **rgrgMd_inv;
  double		**rgrgU,**rgrgUTrans;
  double		**rgrgV, **rgrgVTrans;
  double		*rgSigma, **rgrgSigma_inv;
  double		*zvec, rcond;
  int			iDeg_Curve;
  double		det[2];
  int 			job;
  double		**gen_trans_mat;
  double		*a_power, *b_power, *c_power, *d_power;
  double 		*tmp_div, *tmp_quot;

  if (ndet1D_u.cRows != ndet1D_u.cCols)
  { printf("form_companion_matrix():input matrix is not square\n");
 exit(-1);
  }
  iDeg_Curve = ndet1D_u.poly[0][0].iDeg1;
  n = ndet1D_u.cRows;

  ALLOCMAT(rgrgMd, n, n, double); ALLOCMAT(rgrgMd_inv, n, n, double);
  ALLOCMAT(rgrgU, n, n, double); ALLOCMAT(rgrgUTrans, n, n, double);
  ALLOCMAT(rgrgV, n, n, double); ALLOCMAT(rgrgVTrans, n, n, double);
  ALLOCN(rgSigma, double, n); ALLOCMAT(rgrgSigma_inv, n, n, double);
  ALLOCN(zvec, double, n); 
  ALLOCN(a_power, double, iDeg_Curve+1);
  ALLOCN(b_power, double, iDeg_Curve+1);
  ALLOCN(c_power, double, iDeg_Curve+1);
  ALLOCN(d_power, double, iDeg_Curve+1);
  ALLOCN(tmp_div, double, iDeg_Curve+1);
  ALLOCN(tmp_quot, double, iDeg_Curve+1);

  a_power[0] = b_power[0] = c_power[0] = d_power[0] = 1.0;
  for (i=1; i<=iDeg_Curve; i++) {
    a_power[i] = a*a_power[i-1];
    b_power[i] = b*b_power[i-1];
    c_power[i] = c*c_power[i-1];
    d_power[i] = d*d_power[i-1];
  }

  ALLOCN(rgrgM_i, double**, iDeg_Curve);
  for(i = 0; i < iDeg_Curve; i++)
    ALLOCMAT(rgrgM_i[i], n, n, double);

  ALLOCN(rgrgM_j, double**, iDeg_Curve+1);
  for(i = 0; i <= iDeg_Curve; i++)
    ALLOCMAT(rgrgM_j[i], n, n, double);

  ALLOCMAT(gen_trans_mat, iDeg_Curve+1, iDeg_Curve+1, double);

  for (i=0; i<=iDeg_Curve; i++)
  for (j=0; j<=iDeg_Curve; j++)
    gen_trans_mat[i][j] = 0.0;

  /* get the identity matrices in place */
  for(i = 0; i < iSize_C; i++)
  for(j = 0; j < iSize_C; j++)
    rgrgC[i][j] = (j == i + n)? 1: 0;

  /* fill in the Mi's */
  for(i = 0; i < n; i++)
  for(j = 0; j < n; j++) {
    for(k = 0; k <= iDeg_Curve; k++)
      rgrgM_j[k][i][j] = ndet1D_u.poly[i][j].coeffs[0][k];
  }

  for (i=iDeg_Curve; i>=0; i--)
    gen_trans_mat[0][i] = combChoose(iDeg_Curve,i)*c_power[i]*
			  d_power[iDeg_Curve-i];

  for (i=1; i<=iDeg_Curve; i++) {
    for (j=0; j<=iDeg_Curve; j++)
      tmp_div[j] = gen_trans_mat[i-1][j];
    tmp_quot[0] = tmp_div[0] / d;
    tmp_quot[iDeg_Curve] = 0.0;
    for (j=1; j<iDeg_Curve; j++)
      tmp_quot[j] = (tmp_div[j] - c*tmp_quot[j-1]) / d;
    gen_trans_mat[i][0] = tmp_quot[0]*b;
    gen_trans_mat[i][iDeg_Curve] = tmp_quot[iDeg_Curve-1]*a;
    for (j=1; j<iDeg_Curve; j++)
      gen_trans_mat[i][j] = tmp_quot[j-1]*a + tmp_quot[j]*b;
  }

/*
  for (i=0; i<=iDeg_Curve; i++)
  for (j=0; j<=i; j++) 
    gen_trans_mat[i][iDeg_Curve-i+j] = combChoose(i, j);
*/

  for (k=0; k<iDeg_Curve; k++) {
    for (i=0; i<n; i++)
    for (j=0; j<n; j++) {
      rgrgM_i[k][i][j] = 0.0;
      for (l=0; l<=iDeg_Curve; l++) 
	rgrgM_i[k][i][j] += rgrgM_j[l][i][j]*gen_trans_mat[l][k];
    }
  }
  /* store Md */
  for (i=0; i<n; i++)
  for (j=0; j<n; j++) {
    rgrgMd[i][j] = 0.0;
    for (l=0; l<=iDeg_Curve; l++) {
      if (gen_trans_mat[l][iDeg_Curve] == 0) continue;
      rgrgMd[i][j] += rgrgM_j[l][i][j]*gen_trans_mat[l][iDeg_Curve];
    }
  }

  /* compute its SVD */
#ifdef _SVD_METH

  call_svd(rgrgMd, n, n, rgrgU, rgSigma, rgrgV);

  if (FP_EQ(rgSigma[n-1], 0.0))
  { 
    return(-1);
  }
  /* Check to see if the matrix rgrgMd is singular using dgeco */
  /* construct Md_inverse */
  MATTRANSPOSE(rgrgUTrans, rgrgU, n, n);
  MATTRANSPOSE(rgrgVTrans, rgrgV, n, n);
  IDENTMAT(rgrgSigma_inv, n);
  for(i = 0; i < n; i++) rgrgSigma_inv[i][i] /= rgSigma[i];
  MATMULT(rgrgU, rgrgV, rgrgSigma_inv, n, n, n); /* rgrgU is tmp array */
  MATMULT(rgrgMd_inv, rgrgU, rgrgUTrans, n, n, n);

#else

  MATASSIGN(rgrgMd_inv, rgrgMd, n, n);
#ifdef __SGI
  dgeco_(rgrgMd_inv, &n, &n, rgSigma, &rcond, zvec);
#else
  dgeco(rgrgMd_inv, &n, &n, rgSigma, &rcond, zvec);
#endif
  printf("The value of RCOND is %.10lf\n", rcond);
  if (FP_EQ_EPS(rcond, 0.0, 1e-5)) { 
    return(-1);
  }
  job = 01;
#ifdef __SGI
  dgedi_(rgrgMd_inv, &n, &n, rgSigma, det, zvec, &job);
#else
  dgedi(rgrgMd_inv, &n, &n, rgSigma, det, zvec, &job);
#endif
  
#endif

  /* multiply all other matrices M_i in C with Md_inverse */
  for(i = 0; i < iDeg_Curve; i++)
  { MATMULT(rgrgU, rgrgMd_inv, rgrgM_i[i], n, n, n); /* use rgrgU as tmp array*/
    MATASSIGN(rgrgM_i[i], rgrgU, n, n);
  }

  /* store Mi's in C */
  for(i = 0; i < n; i++)
  { p = i + iSize_C - n;
    for(k = 0; k < iDeg_Curve; k++)
    { r = k*n;
      for(j = 0; j < n; j++)
      { q = r + j;
        rgrgC[p][q] = -rgrgM_i[k][i][j];
      }	
    }
  }

  FREEMAT(rgrgMd, n, n); FREEMAT(rgrgMd_inv, n, n); 
  FREEMAT(rgrgU, n, n); FREEMAT(rgrgUTrans, n, n); 
  FREEMAT(rgrgV, n, n); FREEMAT(rgrgVTrans, n, n); 
  FREEMAT(rgrgSigma_inv, n, n); myfree(rgSigma);
  myfree(zvec);
  for(i = 0; i < iDeg_Curve; i++)
    FREEMAT(rgrgM_i[i], n, n);
  myfree(rgrgM_i);
  for(i = 0; i <= iDeg_Curve; i++)
    FREEMAT(rgrgM_j[i], n, n);
  myfree(rgrgM_j);
  FREEMAT(gen_trans_mat, iDeg_Curve+1, iDeg_Curve+1);
  myfree(a_power); myfree(b_power); myfree(c_power); myfree(d_power);
  myfree(tmp_div); myfree(tmp_quot);
  return 0;
}

/* assumes that rgrgB and rgrgA_plus_B have been allocated outside */
form_power_companion_matrix(ndet1D_u, rgrgB, rgrgA_plus_B, iSize_C)
NDETERMINANT_1D		ndet1D_u;	
double			**rgrgB;
double			**rgrgA_plus_B;
int			iSize_C;
{ 
  int			i, j, k, n;
  int			p, q, r;
  double		***rgrgM_i;
  double		**rgrgA;
  double		**rgrgMd;
  double		**rgrgU,**rgrgUTrans;
  double		**rgrgV;
  double		*rgSigma, **rgrgSigma_inv;
  int			iDeg_Curve;

  if (ndet1D_u.cRows != ndet1D_u.cCols)
  { printf("form_power_companion_matrix():input matrix is not square\n");
    exit(-1);
  }
  iDeg_Curve = ndet1D_u.poly[0][0].iDeg1;
  n = ndet1D_u.cRows;

  ALLOCMAT(rgrgA, iSize_C, iSize_C, double); 
  ALLOCMAT(rgrgU, iSize_C, iSize_C, double); 
  ALLOCMAT(rgrgUTrans, iSize_C, iSize_C, double);
  ALLOCMAT(rgrgV, iSize_C, iSize_C, double); 
  ALLOCN(rgSigma, double, iSize_C); 
  ALLOCMAT(rgrgSigma_inv, iSize_C, iSize_C, double);
  ALLOCN(rgrgM_i, double**, iDeg_Curve);
  for(i = 0; i < iDeg_Curve; i++)
   ALLOCMAT(rgrgM_i[i], n, n, double);
  ALLOCMAT(rgrgMd, n, n, double);

  /* get the identity matrices in place */
  for(i = 0; i < iSize_C; i++)
   for(j = 0; j < iSize_C; j++)
   { rgrgA[i][j] = (j == i) ? 1: 0;
     rgrgB[i][j] = (j == i+n) ? 1: 0;
   }

  /* fill in the Mi's */
  for(i = 0; i < n; i++)
  { for(j = 0; j < n; j++)
	 for(k = 0; k < iDeg_Curve; k++)
  	     rgrgM_i[k][i][j] = ndet1D_u.poly[i][j].coeffs[0][k];
  }

  /* store Md */
  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      rgrgMd[i][j] = ndet1D_u.poly[i][j].coeffs[0][iDeg_Curve];

  /* store Mi's in B */
  for(i = 0; i < n; i++)
  { p = i + iSize_C - n;
    for(k = 0; k < iDeg_Curve; k++)
    { r = k*n;
      for(j = 0; j < n; j++)
      { q = r + j;
	rgrgB[p][q] = -rgrgM_i[k][i][j];
      }	
    }
  }

  /* store Md in A */
  for(i = 0; i < n; i++)
  { p = i + iSize_C - n;
    for(j = 0; j < n; j++)
    { q = j + iSize_C - n;
      rgrgA[p][q] = rgrgMd[i][j];
    }	
  }

  /* compute A + B */
  for(i = 0; i < iSize_C; i++)
    for(j = 0; j < iSize_C; j++)
      rgrgA_plus_B[i][j] = rgrgA[i][j] + rgrgB[i][j];

  FREEMAT(rgrgMd, n, n); 
  FREEMAT(rgrgA, iSize_C, iSize_C); 
  FREEMAT(rgrgU, iSize_C, iSize_C); 
  FREEMAT(rgrgUTrans, iSize_C, iSize_C); 
  FREEMAT(rgrgV, iSize_C, iSize_C); 
  FREEMAT(rgrgSigma_inv, iSize_C, iSize_C); myfree(rgSigma);
  for(i = 0; i < iDeg_Curve; i++)
    FREEMAT(rgrgM_i[i], n, n);
  myfree(rgrgM_i);
}



add_list_point(pplp4Dlist, p4Dpt, tp1, invert, trace, dr)
LIST_POINT4D	**pplp4Dlist;
POINT4D		p4Dpt;
TENSOR_PATCH	tp1;
int 		invert, trace, dr;
{ 
  LIST_POINT4D	*plp4Dtemp;
  POINT2D       p2D;
  POINT4D       p4D, chck_pt4D;
extern double Scale;

  if ((*pplp4Dlist) != NULL) {
    chck_pt4D = (*pplp4Dlist)->p4D;
    if (EQ_4D(chck_pt4D, p4Dpt)) return;
  }

  ALLOCN(plp4Dtemp, LIST_POINT4D, 1);
  plp4Dtemp->plp4DNext = *pplp4Dlist;
  plp4Dtemp->p4D = p4Dpt;
  *pplp4Dlist = plp4Dtemp;

#ifdef __GL
   /* if (trace) { */
#ifdef ONE_PATCH
  if (1) {
    if (!invert) {
       /* draw_st_point( p4Dpt.dCoord[S], p4Dpt.dCoord[T] ); */
      if (trace)
        dr_uvi_point(p4Dpt.dCoord[U], p4Dpt.dCoord[V], 0.0);
      else draw_uv_point( p4Dpt.dCoord[U], p4Dpt.dCoord[V] );
    }
    else {
      if (trace)
        dr_uvi_point(p4Dpt.dCoord[S], p4Dpt.dCoord[T], 0.0);
      else draw_uv_point( p4Dpt.dCoord[S], p4Dpt.dCoord[T] );
       /* draw_st_point( p4Dpt.dCoord[U], p4Dpt.dCoord[V] ); */
    }
  }
#endif

  p2D.u = p4Dpt.dCoord[U];
  p2D.v = p4Dpt.dCoord[V];
  evaluateBernstein( tp1, p2D, &p4D );
  p4D.dCoord[W]/=Scale;
  draw_xyz_point(p4D.dCoord[X]/p4D.dCoord[W],
                 p4D.dCoord[Y]/p4D.dCoord[W],
                 p4D.dCoord[Z]/p4D.dCoord[W], dr );
#endif
}

/* determines if the matrix is singular or not.
   if it is, this returns a vector in the kernel of the matrix
*/

int
matrix_singular(ndet1D_uv, rgKernel)
NDETERMINANT_1D		ndet1D_uv;
double			*rgKernel;
{

  int			dim, singular = FALSE;
  double		**rgrgU, *rgSigma, **rgrgV;
  double		max_val, min_val;

  dim = ndet1D_uv.cRows;
  ALLOCN(rgSigma, double, dim);
  ALLOCMAT(rgrgU, dim, dim, double);
  ALLOCMAT(rgrgV, dim, dim, double);
  call_svd(ndet1D_uv.eval, dim, dim, rgrgU, rgSigma, rgrgV);

   /* printf(" (%4.9f) \n", rgSigma[dim-1]);
*/

  max_val = rgSigma[0];
  min_val = rgSigma[dim-1];

  singular = FP_EQ_EPS(min_val, 0.0, 1e-3);
  /* singular = FP_EQ_EPS(min_val, 0.0, 1e-7); */

  if (singular)
  { /* get the vector v in the kernel */
    /* the last column of v?? */
    GETCOL(dim-1, rgrgV, rgKernel, dim);
  }
  
  myfree(rgrgU); 
  myfree(rgSigma); 
  myfree(rgrgV);
 
  return(singular);
}

