#ifdef __GL

#include <basic_i.h>
#include <sys/time.h>
int is_planar_overlap = FALSE;

extern struct global_s global;
extern node_t qs[];		/* Query structure */
extern trap_t tr[];		/* Trapezoid structure */
extern segment_t seg[];		/* Segment table */

static int initialise(nseg)
     int nseg;
{
  register int i;

  for (i = 1; i <= nseg; i++)
    seg[i].is_inserted = FALSE;
  generate_random_ordering(nseg);
  
  return 0;
}

/*#undef STANDALONE*/
#ifdef STANDALONE

int main(argc, argv)
     int argc;
     char *argv[];
{
  int n, nmonpoly;
  FILE *infile;
  int op[SEGSIZE][3], i;

  if (argc < 2)
    {
      fprintf(stderr, "usage: triangulate <filename>\n");
      exit(1);
    }
  else
    if ((infile = fopen(argv[1], "r")) == NULL)
      {
	perror(argv[1]);
	exit(1);
      }
  
#ifdef CLOCK
  clock();
#endif
  
  n = read_segments(infile);
  if (1)
    if (!is_polygon_simple(n))
      {
	printf("triangulate polygon: polygon not simple\n");
	return -1;
      }

  for (i = 0; i < 1; i++)
    {
      initialise(n);
      construct_trapezoids(n, seg);
      nmonpoly = monotonate_trapezoids(n);
      triangulate_monotone_polygons(nmonpoly, op);
    }
  
#ifdef CLOCK
  printf("CPU time used: %ld microseconds\n", clock());
#endif
  
  return 0;
}

#else  /* Not standalone. Use this as an interface routine */

/* The points constituting the polygon are specified in anticlockwise
 * order. If there are n points in the polygon, i/p would be the
 * points p0, p1....p(n) (where p0 and pn are the same point). The
 * output is contained in the array "triangles".
 * Every triangle is output in anticlockwise order and the 3
 * integers are the indices of the points. Thus, the triangle (i, j, k)
 * refers to the triangle formed by the points (pi, pj, pk). Before
 * using this routine, please check-out that you do not conflict with
 * the global variables  defined in basic.h.
 *
 * n:         number of points in polygon (p0 = pn)
 * vertices:  the vertices p0, p1..., p(n) of the polygon
 * triangles: output array containing the triangles 
 *
 * Return value: less than zero if the polygon is non-simple and the
 *               triangulation was unsuccessful. Else >= 0 
 */

int triangulate_polygon(n, vertices, triangles, check_simplicity)
     int n;
     double vertices[][2];
     int triangles[][3];
     int check_simplicity;
{
  register int i;
  int nmonpoly;

  memset((void *)seg, 0, SEGSIZE*sizeof(segment_t));
  for (i = 1; i <= n; i++)
    {
      seg[i].is_inserted = FALSE;

      seg[i].v0.x = vertices[i][0]; /* x-cood */
      seg[i].v0.y = vertices[i][1]; /* y-cood */
      if (i == 1)
	{
	  seg[n].v1.x = seg[i].v0.x;
	  seg[n].v1.y = seg[i].v0.y;
	}
      else
	{
	  seg[i - 1].v1.x = seg[i].v0.x;
	  seg[i - 1].v1.y = seg[i].v0.y;
	}	
    }
  
  global.nseg = n;

#ifdef DEBUG_PRINT_POLY_PRINT
  printf("TRiangulate_input\n");
  for (i = 1; i <= n; i++)
	printf("%.8lf %.8lf\n", seg[i].v0.x, seg[i].v0.y);
fflush(stdout);
#endif

  if (check_simplicity)
    if (!is_polygon_simple(n))
      {
	/* printf("triangulate polygon: polygon not simple\n"); */
#ifndef NOPRINT_VERSION
	  for (i = 1; i <= n; i++)
		printf("%.8lf %.8lf\n", seg[i].v0.x, seg[i].v0.y);
#endif
	  fflush(stdout);
	return -1;
      }
  
  initialise(n);
  construct_trapezoids(n, seg);
  nmonpoly = monotonate_trapezoids(n);
  triangulate_monotone_polygons(nmonpoly, triangles);
  
  return 0;
}

#endif

int is_polygon_simple(n)
     int n;
{
  register int i, j;
  point_typ p0, p1;
  double s0, s1, t0, t1;
  int itype;

  for (i = 1; i < n-1; i++)
    for (j = i+2; j <= n; j++)
      if (i == 1 && j == n)
	continue;
      else 
	if (segments_intersect(&seg[i].v0, &seg[i].v1,
			       &seg[j].v0, &seg[j].v1, &itype,
			       &s0, &s1, &t0, &t1, &p0, &p1) > 0)
	  return FALSE;
  
  return TRUE;
}

#endif
