#include <intersect.h>
#include <math.h>

static sinfo_t ch[MAX_INTERSECTIONS_PER_SEGMENT]; /* used to store all */
						  /* the intersections */
						  /* of one segment in */
						  /* sorted order */

/* normalise vector: used for angle calculations */
static int normalise(vector_t *v)
{
  double denom;

  denom = sqrt(v->x*v->x + v->y*v->y);
  v->x /= denom;
  v->y /= denom;

  return 0;
}

/* Add one more intersection to the edge. Maintain this list in sorted */
/* order. */
static int insert_iinfo(einfo_t *e, int nch, int segno, int inum,
			double s, double t)
{
  register int i;
  int j;
  
  for (i = 0; i < MAX_INTERSECTIONS_PER_EDGE; i++)
    if (e->intr[i].s > s)
      break;
  
  for (j = e->nintr; j > i; j--)
    e->intr[j] = e->intr[j - 1];

  e->intr[i].nch = nch;
  e->intr[i].nseg = segno;
  e->intr[i].inum = inum;
  e->intr[i].s = s;
  e->intr[i].t = t;
      
  e->nintr++;
  return i;
}


/* Add one more intersection to the list of all intersections of a */
/* particular segment. */
static int insert_sinfo(sinfo_t ch[], int nch, point_typ *p,
			int nedge, double s, double t, int h)
{
  register int i;
  int j;
  
  for (i = 0; i < nch; i++)
    if ((ch[i].t > t) || (FP_EQUAL(ch[i].t, t) && (ch[i].h > h)))
      break;
  
  /* The extra h-information added to take care of the case when a */
  /* segment just grazes the commonpoint of 2 edges and lies inside. */
  /* In this case, we consider it to be two intersections */

  for (j = nch; j > i; j--)
    ch[j] = ch[j - 1];

  ch[i].s = s;
  ch[i].t = t;
  ch[i].h = h;
  ch[i].nedge = nedge;
  ch[i].v = *p;
      
  return 0;
}

/* return TRUE if p2 is to the left of (p0 -> p1) */
static int intr_is_left_of(point_typ *p0, point_typ *p1,
		    point_typ *p2)
{
  double area;
  area = CROSS(*p0, *p1, *p2);
  return (area > 0.0);
}


/* return the angle formed by the points (u0, u1, u2) in that order. */
/* Basically, it is a numerical value between 1 and -3. Used for */
/* comparison only */
/* smaller angle -> larger return value */
static double intr_angle_between(point_typ *u0, point_typ *u1,
		    point_typ *u2)
{
  vector_t v0, v1;
  
  v0.x = u1->x - u0->x;
  v0.y = u1->y - u0->y;
  v1.x = u2->x - u1->x;
  v1.y = u2->y - u1->y;
  
  normalise(&v0);
  normalise(&v1);
  if (CROSS_VEC(v0, v1) >= 0.0)
    return  -DOT(v0, v1); /* 0: 1.0    180: -1.0 */
  else
    return (-2.0 + DOT(v0, v1));
}


/* Find the intersections of all the chaons with all theedges of the */
/* polygon */
int find_intersections_of_segment_with_polygon(segment_t *edge, 
         struct global_s global,					      
	 point_typ *v0, point_typ *v1, double *ss, double *tt, int *first_indx, 
	 double *ss1, double *tt1, int *sec_indx, int was_in, int is_in)
{
  register int i, k;
  point_typ pp0, pp1, p0, p1, vv0, vv1;
  int nch, ich, n, p, q, gnch, bflag, itype, nintr;
  double s0, s1, t0, t1;
  int pos, prev_edge, nedge, delta, hackflag;
  pchain_t *pc;
  int N = global.nseg;

  
  ich = 0;			/* index for #intersections */
  bflag = hackflag = FALSE;
  for (i = 0; i < global.nseg; i++)
    {
      if ((n = shady_segments_intersect(&edge[i].v0, &edge[i].v1,
				  v0, v1, &itype, 
				  &s0, &s1, &t0, &t1,
				  &p0, &p1, I_NONEPT)) == 0)
	{
	  /* segments do not intersect */
	  continue;
	}
      else if (n == 1)
	{
	  int cond;
	  /* insert information about this intrn. in the edgeinfo */
	  /* structure, and the ch-table */

#ifdef OLD_VERSION	  
	  cond = FP_EQUAL(s0, 0.0) &&
	    (FP_EQUAL(t0, 0.0) || FP_EQUAL(t0, 1.0));
	  
	  if ((FP_EQUAL(s0, 0.0) && cond)
	      /* take care in next edge */
	      || 
	      (FP_EQUAL(s0, 1.0) && /* coincident with */
	       /* next edge */
	       intr_are_vectors_coincident(v0, v1, 
					   &edge[(i+1)%N].v0,
					   &edge[(i+1)%N].v1))
	      ||
	      ((s0 < 1.0) && !(cond) && FP_EQUAL(t0, 1.0) &&
	       (!intr_is_left_of(&edge[i].v0, &edge[i].v1,
				 v0)))
	      ||
	      (FP_EQUAL(t0, 0.0) && (s0 < 1.0) &&
	       (!cond)  &&
	       (!intr_is_left_of(&edge[i].v0, 
				 &edge[i].v1,
				 v1)))
	      ||
	      (FP_EQUAL(t0, 0.0) && FP_EQUAL(s0, 1.0) &&
	       !(intr_angle_between(&edge[i].v0, 
				    &edge[i].v1,
				    v1) > 
		 intr_angle_between(&edge[i].v0, 
				    &edge[i].v1,
				    &edge[(i+1)%N].v1)))) 
	    ;
	  else
	    {
	      if ((FP_EQUAL(s0, 0.0) || FP_EQUAL(s0, 1.0))
		  && !(FP_EQUAL(t0, 0.0) || 
		       FP_EQUAL(t0, 1.0)))
		{
		  int tp, h;
		  point_typ p_tmp;
		  
		  /* Check if the (possibly) grazing edge is inside or */
		  /* outside */
		  if (FP_EQUAL(s0, 1.0))
		    {
		      tp = (i+1)%N;
		      vv0.x = edge[i].v1.x - edge[i].v0.x;
		      vv0.y = edge[i].v1.y - edge[i].v0.y;
		      
		      vv1.x = edge[tp].v1.x - edge[tp].v0.x;
		      vv1.y = edge[tp].v1.y - edge[tp].v0.y;
			      
		      if (CROSS_VEC(vv0, vv1) > 0.0)
			{
			  if (intr_is_left_of(&edge[i].v0,
					      &edge[i].v1,
					      v0))
			    p_tmp = *v0;
			  else
			    p_tmp = *v1;
			  
			  /* check if the segment intersects the */
			  /* polygon at the common point of two edges. */
			  /* If yes, insert the intrn. info, else it */
			  /* is grazing condn. ignore it */

			  if (intr_angle_between(&edge[i].v0, 
						 &edge[i].v1,
						 &p_tmp) > 
			      intr_angle_between(&edge[i].v0, 
						 &edge[i].v1,
						 &edge[(i+1)%N].v1))
			    {
			      insert_sinfo(ch, ich, &p0, i, s0, t0,
					   -1);
			      ich++;
			      continue;
			    }
			  else
			    continue;
			}
		      else	/* cross product is < 0.0 */
			{
			  if (intr_is_left_of(&edge[i].v0,
					      &edge[i].v1,
					      v0))
			    p_tmp = *v1;
			  else
			    p_tmp = *v0;
			  
			  /* check if the segment intersects the */
			  /* polygon at the common point of two edges. */
			  /* If yes, insert the intrn. info, else it */
			  /* is grazing condn. ignore it */

			  if (intr_angle_between(&edge[i].v0, 
						 &edge[i].v1,
						 &p_tmp) < 
			      intr_angle_between(&edge[i].v0, 
						 &edge[i].v1,
						 &edge[(i+1)%N].v1))
			    {
			      insert_sinfo(ch, ich, &p0, i, s0, t0,
					   -1);
			      ich++;
			      continue;
			    }
			  else			  
			    continue;
			}
		    }
		  else		/* s0 = 0.0. Even if the segment is */
				/* grazing/proper intersection, do not */
				/* care abput it (since s = 1.0 condn. */
				/* will incorporate it. So ignore the */
				/* intersection  */
		    continue;
		}
	      
	      /* NORMAL CASE (single, well-behaved intersection) */
	      insert_sinfo(ch, ich, &p0, i, s0, t0, -1);
	      ich++;
	    }
#else
	  /* Newer code */
	  insert_sinfo(ch, ich, &p0, i, s0, t0, -1);
	  ich++;	  
#endif

	}
      else			/* two intersection (coincident lines) */
	{
	  if (itype == I_PAR)
	    {
	      insert_sinfo(ch, ich, &p0, i, s0, t0, -1);
	      ich++;
	      insert_sinfo(ch, ich, &p1, i, s1, t1, -1);
	      ich++;
	    }
	  else
	    {
	      insert_sinfo(ch, ich, &p0, i, s1, t0, -1);
	      ich++;
	      insert_sinfo(ch, ich, &p1, i, s0, t1, -1);
	      ich++;
	    }
	}
    } /* end-forloop */
  
  /* Now, we have all the intersections for the segment into */
  /* the table ch[]. Return in intersection information of the last */
  /* intersection with the polygon */
  
  if (ich > 0)			/* there ARE intersections */
    {
      nintr = ich--;
      
      if (was_in && !is_in)
	{
	  *ss = ch[ich].s;
	  *tt = ch[ich].t;
	  *first_indx = ch[ich].nedge;
	  *ss1 = *tt1 = -1.0;
	  *sec_indx = -1;
	}
      else if (is_in && !was_in)
	{
	  *ss = ch[0].s;
	  *tt = ch[0].t;
	  *first_indx = ch[0].nedge;
	  *ss1 = *tt1 = -1.0;
	  *sec_indx = -1;
	}
      else 
	{
	  *ss = ch[0].s;
	  *tt = ch[0].t;
	  *first_indx = ch[0].nedge;
	  if (!FP_EQUAL_PAR(ch[0].t, ch[ich].t, 1e-3)) {
	    *ss1 = ch[ich].s;
	    *tt1 = ch[ich].t;
	    *sec_indx = ch[ich].nedge;
	  } else nintr = 1;
	}
      return nintr;
    }
  
  return 0;
} 




