#include <stdio.h>
#include <malloc.h>
/* #include <fmclient.h> */

#ifdef __PERFORM_TIME
#include "timesgi.h"
#endif

#include "defs.h"
#include "complex.h"
#include "lp.h"
#include "primitives.h"
#include "tree.h"
#ifdef BEZ_CLIP
#include "csurf.h"
#endif

#ifdef __GL
#include <gl.h>
#endif

typedef struct {
  double x, y;
} mypoint_t;

typedef struct {
  mypoint_t v0, v1;
  int is_inserted;
  int root0, root1;
} segment_t;

#define SEGSIZE 400		/* max# of segments */
#define XYZ_HEIGHT 600
#define XYZ_WIDTH  1000
#define	MAX_PATCH_NUM	1200
#define MAX_OVERLAP_NUM	100
#define MAX_PATCH_DEGREE 7
#define NEW_EPS		5.0*1e-5
#define SMALL_DISPLACEMENT 0.01 /* This constant is to find if solids meeting
				  at a plane are on the same side or not */

extern  int    Trace(); 
extern  int     getComponents();
extern	void 	read_input();
extern	void 	print_tensor_patch();
extern  void    bernsteinToPowerBasis();
extern  void    allocDeterminant();
extern  void    allocDeterminant_1D();
extern  void    moveFactoredToFinalDeterminant();
extern  void    printDetMatrix();
extern  void    outputPolyLines();
extern  void    output_patch_mesh();
extern  void    powell();
extern  void    MergeCurveLists();
extern  void    setUpTraceOrder();
extern  void	SpiltIntoNewRegions();
extern  double  RefinePoint();
extern  void	copy_trim_patch();
extern  int     compar();
extern  int     comparU();
extern  int     comparV();
extern void     copy_ten_p();
extern INTERSECTION_CURVE *MergeSplines();
extern INTERSECTION_CURVE *MergeSplinesFurther();

extern TENSOR_PATCH pow_tp0;
extern TENSOR_PATCH pow_tp1;

extern int tpatch_idx;
extern TENSOR_PATCH tpatch[];

#ifdef __GL
extern  void    draw_the_patches();

extern  int curve_or_full;
extern Matrix   Mp;
extern	long	giduv, gidxyz, gidst;
#endif

double		Scale=1.0;
double          Sphradius;
NDETERMINANT_1D  ndet1D_uv, fdet1D_uv, ndet1D_du, ndet1D_dv;
int		MatOrderForMinimFunc;
double		Step_Size_Out;
int    		imp_time, cs_time, sp, trace_time;
LIST_POINT4D    *gplp4DEndPoints;

/* if you change any of the next two variables in any way, make */
/* corresponding changes in spline.c */
short first_same_solid[MAX_PATCHES_PER_SOLID][1][MAX_PARTNS_PER_PATCH][MAX_SPLINES_PER_TRIM_REGION];
                                       /* this is the global */
                                       /* variable that holds */
				       /* the boolean for */
				       /* [patchno][trimportion */
				       /* no][regionno][curveno] */

short first_complete_ovrlap[20], second_complete_ovrlap[20];

short second_same_solid[MAX_PATCHES_PER_SOLID][1][MAX_PARTNS_PER_PATCH][MAX_SPLINES_PER_TRIM_REGION];

short NumCurvesBeforeProjection[20];

short special_crvs_in_first_sol[MAX_OVERLAP_NUM][MAX_SPLINES_PER_TRIM_REGION],
      special_crvs_in_sec_sol[MAX_OVERLAP_NUM][MAX_SPLINES_PER_TRIM_REGION];

extern short fXsc[100][5][2][3], sXsc[100][5][2][3]; 

NEIGHBORING_CURVE ngh_crv[MAX_OVERLAP_NUM];
int NumNghCrvs;
OVR_BOXES overlap_pairs[MAX_OVERLAP_NUM];
SAME_SIDE_OVR_STRUCT ssos[MAX_OVERLAP_NUM][MAX_SPLINES_PER_TRIM_REGION];
int NumOvrPatches = 0;

static int    planar_ovrlap_flag1[MAX_PATCH_NUM], planar_ovrlap_flag2[MAX_PATCH_NUM];
static segment_t seg2[SEGSIZE];

#ifdef __PERFORM_TIME
  static double XsectTime = 0.0;
  static double LPTime = 0.0;
  static double ImplicitTime = 0.0;
  static double PartitionTime = 0.0;
  static double ClassifyTime = 0.0;
  static double FittingTime = 0.0;
  static TimerType Time1, Time2;
#endif

static float  whitevec[3] = {1.0, 1.0, 1.0};
static float  blackvec[3] = {0, 0, 0};
static float  grayvec[3] = {0.8, 0.8, 0.8};
static float  cyanvec[3] = {.3, .9, .7};
static float  yelvec[3] = {1, 1, 0};
static float  purplevec[3] = {0.2, 0.2, 0.2};

static float  navyvec[3] = {0.0, 0.0, 1.0};

/* the following static variables are used as trim curves either in */
/* piecewise case or in spline case */
static DOMAIN_INT_CURVE **fs_pw_curves, **ss_pw_curves;
static int *fs_nc, *ss_nc; /* indicates number of curves */
static MERGED_CURVE **fs_spl_curves, **ss_spl_curves;

LINALGINIT

/* This routine is used mostly when arr is patchNumber and brr is */
/* first_or_second array. This returns the original patch number in */
/* the solid. Used to get final solid's graph */
int ReturnIndexInArray(arr, num, value, brr, val)
int *arr, *brr, num, value, val;
{
  int i;

  for (i=0; i<num; i++)
    if ((arr[i] == value) && (brr[i] == val))
      return i;
  return -1;
}

void myfree(ptr)
void *ptr;
{
  if (ptr == NULL) return;
  /* ptr = realloc(ptr, 0); */
  free(ptr); 
  ptr = NULL;
}

void EvalPointOnPatch(tp, s, t, xv, yv, zv)
TENSOR_PATCH    tp;
double          s, t;
double          *xv, *yv, *zv;
{
  int           m, n;
  int           i, j;
  int		*ppp;
  int           polynomial;
  double        *ber1, *ber2;
  double        x_st, y_st, z_st, w_st;

  m = tp.iDeg0;
  n = tp.iDeg1;

  ALLOCN(ber1, double, m+1);
  ALLOCN(ber2, double, n+1);

  eval_ber(ber1, s, m);
  eval_ber(ber2, t, n);

  x_st = y_st = z_st = w_st = 0.0;
  polynomial = 1;

  for (i=0; i<=m; i++) {
    for (j=0; j<=n; j++) {
      if (!FP_EQ_EPS(tp.coeffs[i][j].dCoord[W], 1.0, 1e-3)) {
/*
      if ((tp.coeffs[i][j].dCoord[W] > 0.999) && 
	  (tp.coeffs[i][j].dCoord[W] < 1.001)) {
*/
	polynomial = 0;
	break;
      }
    }
    if (polynomial == 0) break;
  }

  for (i=0; i<=m; i++)
  for (j=0; j<=n; j++)
  {
    x_st += tp.coeffs[i][j].dCoord[X] * ber1[i] * ber2[j] ;
    y_st += tp.coeffs[i][j].dCoord[Y] * ber1[i] * ber2[j] ;
    z_st += tp.coeffs[i][j].dCoord[Z] * ber1[i] * ber2[j] ;
    w_st += tp.coeffs[i][j].dCoord[W] * ber1[i] * ber2[j] ;
  }

  if ( FP_EQ_EPS(w_st, 0.0, 1e-6) && !polynomial)
  {
    printf("W_ST IS ZERO\n");
    exit(-1);
  }

   /* w_st /= Scale; */
  if (polynomial) {
    *xv = x_st;
    *yv = y_st;
    *zv = z_st;
  } else {
    *xv = x_st / w_st ;
    *yv = y_st / w_st ;
    *zv = z_st / w_st ;
  }

#ifdef __FREE_ALL
  myfree(ber1);
  myfree(ber2);
#endif
}

int compar( v1, v2 )
int *v1, *v2;
{
  return ( (*v1) - (*v2) );
}

int comparU(p1, p2)
POINT4D *p1, *p2;
{
  if (p1->dCoord[U] < p2->dCoord[U]) return -1;
  else if (p1->dCoord[U] > p2->dCoord[U]) return 1;
  return 0;
}

int comparV(p1, p2)
POINT4D *p1, *p2;
{
  if (p1->dCoord[V] < p2->dCoord[V]) return -1;
  else if (p1->dCoord[V] > p2->dCoord[V]) return 1;
  return 0;
}

void Const2DFrom4D(ret_curve, orig_curve, m, n)
LIST_POINT2D **ret_curve;
LIST_POINT4D *orig_curve;
int m, n;
{
  LIST_POINT2D *tmp1, *tmp2;
  LIST_POINT4D *tmp3;

  *ret_curve = NULL;
  if (orig_curve == NULL) return;
  tmp3 = orig_curve;
  ALLOCN(tmp2, LIST_POINT2D, 1);
  tmp2->p2D.u = tmp3->p4D.dCoord[m];
  tmp2->p2D.v = tmp3->p4D.dCoord[n];
  tmp2->plp2DNext = NULL;
  *ret_curve = tmp1 = tmp2;
  tmp3 = tmp3->plp4DNext;

  for (;tmp3;) {
    ALLOCN(tmp2, LIST_POINT2D, 1);
    tmp2->p2D.u = tmp3->p4D.dCoord[m];
    tmp2->p2D.v = tmp3->p4D.dCoord[n];
    tmp2->plp2DNext = NULL;
    tmp1->plp2DNext = tmp2;
    tmp1 = tmp2;
    tmp3 = tmp3->plp4DNext;
  }
}
  
void Const2DFrom4DReverse(ret_curve, orig_curve, m, n)
LIST_POINT2D **ret_curve;
LIST_POINT4D *orig_curve;
int m, n;
{
  LIST_POINT2D *tmp2;
  LIST_POINT4D *tmp3;

  *ret_curve = NULL;
  if (orig_curve == NULL) return;
  tmp3 = orig_curve;
  for (;tmp3;) {
    ALLOCN(tmp2, LIST_POINT2D, 1);
    tmp2->p2D.u = tmp3->p4D.dCoord[m];
    tmp2->p2D.v = tmp3->p4D.dCoord[n];
    tmp2->plp2DNext = *ret_curve;
    *ret_curve = tmp2;
    tmp3 = tmp3->plp4DNext;
  }
}
  
void Reverse2DList(ret_curve)
DOMAIN_INT_CURVE *ret_curve;
{
  LIST_POINT2D *tmp, *tmp2;
  POINT2D pt2D;

  pt2D = ret_curve->StEndPoints[0];
  ret_curve->StEndPoints[0] = ret_curve->StEndPoints[1];
  ret_curve->StEndPoints[1] = pt2D;
  tmp = ret_curve->planeCurve;

  ret_curve->planeCurve = NULL;
  for (;tmp;) {
    tmp2 = tmp;
    tmp = tmp->plp2DNext;
    tmp2->plp2DNext = ret_curve->planeCurve;
    ret_curve->planeCurve = tmp2;
  }
}
  
double dist3D(pt1, pt2)
POINT3D *pt1, *pt2;
{
  double ret, xv, yv, zv;

  xv = pt1->dCoord[0] - pt2->dCoord[0];
  yv = pt1->dCoord[1] - pt2->dCoord[1];
  zv = pt1->dCoord[2] - pt2->dCoord[2];

  ret = xv * xv + yv * yv + zv * zv;

  return ret;
}

double dist2D(pt1, pt2)
POINT2D *pt1, *pt2;
{
  double ret, uv, vv;

  uv = pt1->u - pt2->u;
  vv = pt1->v - pt2->v;
  ret = uv * uv + vv * vv;

  return ret;
}

int closer_pt(pt, pt1, pt2, dist)
POINT2D *pt, *pt1, *pt2;
double *dist;
{
  double dist0, dist1;

  dist0 = dist2D(pt, pt1);
  dist1 = dist2D(pt, pt2);
  if (dist0 < dist1) { *dist = dist0; return 0; }
  else { *dist = dist1; return 1; }
}

Merge2DLists(dom_curve, num_curv, comp, flag)
DOMAIN_INT_CURVE **dom_curve;
int *num_curv, flag;
INT_CURVE_STRUCT *comp;
{
  int p, q;
  int i, j, k;
  LIST_POINT2D *tmp_curve;
  POINT2D comp_ep0, comp_ep1;
  double dist0, dist1;
  int close0, close1;
  int degenerate_case; /* = 1 when the two endpts of comp are
			  within 2e-5 dist. apart */

  if (flag != 1 && flag != 2) {
    printf("Merge2DLists: Error in flag input\n");
    return;
  }
  if (flag == 1) {
    p = 2;
    q = 3;
  } else {
    p = 0;
    q = 1;
  }

  comp_ep0.u = comp->endPoints[0].dCoord[p];
  comp_ep0.v = comp->endPoints[0].dCoord[q];
  comp_ep1.u = comp->endPoints[1].dCoord[p];
  comp_ep1.v = comp->endPoints[1].dCoord[q];
  if (EQUAL2D_EPS(comp_ep0, comp_ep1, 1e-8))
    degenerate_case = 1;
  else degenerate_case = 0;

  if (degenerate_case) 
    printf("Merge2DLists: AKJHDKUDKJGF\n");

  for (i=0; i<(*num_curv); i++) {
    close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
		   &comp_ep0, &comp_ep1, &dist0);

    close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
		   &comp_ep0, &comp_ep1, &dist1);

    if (EQ_2D(comp->endPoints[0], p, q, 
	      (*dom_curve)[i].StEndPoints[0]) && close0 == 0
	&& dist0 <= dist1) { 
      tmp_curve = NULL;
      Const2DFrom4D(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&tmp_curve, (*dom_curve)[i].planeCurve);
      (*dom_curve)[i].planeCurve = tmp_curve;
      (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[1].dCoord[p];
      (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[1].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[0]) && close0
	    == 0 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[1]) && close0
	    == 1 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[0], p, q, 
	      (*dom_curve)[i].StEndPoints[1]) && close1 == 0
	       && dist1 <= dist0) {
      tmp_curve = NULL;
      Const2DFrom4DReverse(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&((*dom_curve)[i].planeCurve), tmp_curve);
      (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[1].dCoord[p];
      (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[1].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[0]) && close1
	    == 0 && dist1 <= dist0) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[1]) && close1
	    == 1 && dist1 <= dist0) { 
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[1], p, q, 
	      (*dom_curve)[i].StEndPoints[0]) && close0 == 1
	       && dist0 <= dist1) {  
      tmp_curve = NULL;
      Const2DFrom4DReverse(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&tmp_curve, (*dom_curve)[i].planeCurve);
      (*dom_curve)[i].planeCurve = tmp_curve;
      (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[0].dCoord[p];
      (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[0].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[0]) && close0 == 0&& dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[0],
		    (*dom_curve)[j].StEndPoints[1]) && close0
	    == 1 && dist0 <= dist1) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[0].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[0].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  ConCat2D(&((*dom_curve)[j].planeCurve),
		   (*dom_curve)[i].planeCurve);
	  (*dom_curve)[i].planeCurve = 
		   (*dom_curve)[j].planeCurve;
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQ_2D(comp->endPoints[1], p, q, 
	       (*dom_curve)[i].StEndPoints[1]) && close1 == 1
	       && dist1 <= dist0) {
      tmp_curve = NULL;
      Const2DFrom4D(&tmp_curve, comp->Curve, p, q);
      ConCat2D(&((*dom_curve)[i].planeCurve), tmp_curve);
      (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[0].dCoord[p];
      (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[0].dCoord[q];
      for (j=i+1; j<(*num_curv); j++) {
	close0 = closer_pt(&((*dom_curve)[i].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist0);
	close1 = closer_pt(&((*dom_curve)[i].StEndPoints[1]),
			   &((*dom_curve)[j].StEndPoints[0]),
			   &((*dom_curve)[j].StEndPoints[1]),
&dist1);
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[0]) && close1
	    == 0 && dist1 <= dist0) {
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[1].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[1].v;
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D((*dom_curve)[i].StEndPoints[1],
		    (*dom_curve)[j].StEndPoints[1]) && close1
	    == 1 && dist1 <= dist0) { 
#ifndef DEBUG_PRINT_POLY
	  printf("2nd Merge: \n");
#endif
	  (*dom_curve)[i].StEndPoints[1].u = 
		(*dom_curve)[j].StEndPoints[0].u;
	  (*dom_curve)[i].StEndPoints[1].v = 
		(*dom_curve)[j].StEndPoints[0].v;
	  Invert2DPointList(&((*dom_curve)[j].planeCurve));
	  ConCat2D(&((*dom_curve)[i].planeCurve),
		   (*dom_curve)[j].planeCurve);
	  for (k=j+1; k<(*num_curv); k++) {
	    (*dom_curve)[k-1].StEndPoints[0] = 
	    		(*dom_curve)[k].StEndPoints[0]; 
	    (*dom_curve)[k-1].StEndPoints[1] = 
	    		(*dom_curve)[k].StEndPoints[1]; 
	    (*dom_curve)[k-1].planeCurve =
	    		(*dom_curve)[k].planeCurve;
	  }
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    }
  }

  if (i == (*num_curv)) {
    Const2DFrom4DReverse(&((*dom_curve)[i].planeCurve),comp->Curve,p,q);
    (*dom_curve)[i].StEndPoints[0].u = comp->endPoints[0].dCoord[p];
    (*dom_curve)[i].StEndPoints[0].v = comp->endPoints[0].dCoord[q];
    (*dom_curve)[i].StEndPoints[1].u = comp->endPoints[1].dCoord[p];
    (*dom_curve)[i].StEndPoints[1].v = comp->endPoints[1].dCoord[q];
    (*num_curv)++;
  } 
}

int getNumofPoints(pList)
LIST_POINT4D *pList;
{
  int i;
  LIST_POINT4D *tmp_ptr;

  tmp_ptr = pList;
  i = 0;
  if (tmp_ptr == NULL) return 0;
  if (tmp_ptr->plp4DNext == NULL) return 1;

  while (tmp_ptr != NULL) {
    i++;
    tmp_ptr = tmp_ptr->plp4DNext;
  }
  return i;
}

#ifdef __MYCODE
int dump_one_patch(outfile, trim, patchnum)
FILE *outfile;
TRIM_PATCH_INFO *trim;
int patchnum;
{
  register int i, j, k;
  int l;
 
  k = patchnum;
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);

      for (i = 0; i <= trim[k].tp.iDeg0; i++)
        for (j = 0; j <= trim[k].tp.iDeg1; j++)
          {
            for (l = 0; l < 4; l++)
              fprintf(outfile, "%lf ",
                      trim[k].tp.coeffs[i][j].dCoord[l]);
            fprintf(outfile, "\n");
          }

  fflush(outfile);
  return 0;
}

int dump_trim_patches(outfile, trim, npatches)
FILE *outfile;
TRIM_PATCH_INFO *trim;
int npatches;
{
  register int i, j, k;
  int l;
 
  fprintf(outfile, "\n\n Printing trim_patch_info dump\n");
  for (k = 0; k < npatches; k++)
    {
      fprintf(outfile, "\n%d %d\n", trim[k].tp.iDeg0, trim[k].tp.iDeg1);

      for (i = 0; i <= trim[k].tp.iDeg0; i++)
        for (j = 0; j <= trim[k].tp.iDeg1; j++)
          {
            for (l = 0; l < 4; l++)
              fprintf(outfile, "%lf ",
                      trim[k].tp.coeffs[i][j].dCoord[l]);
            fprintf(outfile, "\n");
          }
    }

  fflush(outfile);
  return 0;
}
#endif

int FindNumTrimSegments(ptr)
LIST_POINT2D *ptr;
{
  int num=0;
  LIST_POINT2D *tmp;
  POINT2D tmp2D;

  if (ptr == NULL) return 0;
  num = 0;
  tmp2D = ptr->p2D;
  if (!num && !FP_EQ_EPS(tmp2D.u,0.0,1e-4) && 
      !FP_EQ_EPS(tmp2D.u,1.0,1e-4) &&
      !FP_EQ_EPS(tmp2D.v,0.0,1e-4) && 
      !FP_EQ_EPS(tmp2D.v,1.0,1e-4)) num++;
  tmp = ptr->plp2DNext;
  if (tmp == NULL) return num;
  while (tmp != NULL) {
    tmp2D = tmp->p2D;
    num++;
    tmp = tmp->plp2DNext;
  }
  if (!FP_EQ_EPS(tmp2D.u, 0.0,1e-4) && !FP_EQ_EPS(tmp2D.u, 1.0,1e-4) &&
      !FP_EQ_EPS(tmp2D.v, 0.0,1e-4) && !FP_EQ_EPS(tmp2D.v, 1.0,1e-4))
    num++; 
  return num;
}

int FindNumElmtsIn(curve, curveArray)
LIST_POINT2D *curve;
double curveArray[][2];
{
  LIST_POINT2D *ptr;
  POINT2D elmt1, elmt2;
  int count;

  count = 0;
  ptr = curve;

  if (ptr == NULL) {
    printf("Warning: No element in the list\n");
    return count;
  }
  
  elmt1 = ptr->p2D;
  curveArray[count][0] = elmt1.u;
  curveArray[count++][1] = elmt1.v;

  if (ptr->plp2DNext == NULL) {
    printf("Warning: Only one element in a loop\n");
    return count;
  }

  ptr = ptr->plp2DNext;

  while (ptr->plp2DNext != NULL) {
    curveArray[count][0] = ptr->p2D.u;
    curveArray[count++][1] = ptr->p2D.v;

    ptr = ptr->plp2DNext;
  }

  if (ptr->plp2DNext != NULL) {
    printf("Something wrong with WHILE loop\n");
    exit(-1);
  }
  elmt2 = ptr->p2D;

  if (!EQUAL2D_EPS(elmt1, elmt2, 1e-6)) {
    printf("First and Last element of loop not the same\n");
  }

  curveArray[count][0] = ptr->p2D.u;
  curveArray[count][1] = ptr->p2D.v;

  return count;
}

GetUVPair(pt2D, intpt2D, curve, orientation)
POINT2D *pt2D, *intpt2D;
LIST_POINT2D *curve;
int *orientation;
{
  POINT2D tmp1, tmp2, tmp3, vec;
  int index, i, NumPoints = 0;
  LIST_POINT2D *tmpPtr;
  double norm, crossp;

  tmpPtr = curve;
  for (;tmpPtr;) {
    NumPoints++;
    tmpPtr = tmpPtr->plp2DNext;
  } 
  if (NumPoints < 5) {
    printf("Too few points in the curve\n");
    pt2D->u = -1.0;
    pt2D->v = -1.0;
  }
  index = NumPoints/2;
  tmpPtr = curve;
  for (i=0; i<index; i++)
    tmpPtr = tmpPtr->plp2DNext;
  *intpt2D = tmp1 = tmpPtr->p2D;
  tmp2 = tmpPtr->plp2DNext->p2D;
  tmp3.u = (tmp1.u+tmp2.u)/2.0;
  tmp3.v = (tmp1.v+tmp2.v)/2.0;
  vec.u = 0.5 - tmp3.u;
  vec.v = 0.5 - tmp3.v;
  if (FP_EQ(vec.u,0.0) && FP_EQ(vec.v,0.0)) { 
    printf("Vector Value is almost zero\n");
    vec.u = -tmp3.u;
    vec.v = -tmp3.v;
  }
  crossp = (tmp2.u - tmp1.u)*vec.v - (tmp2.v - tmp1.v)*vec.u;
  if (FP_EQ(crossp,0.0)) {
    printf("Cross Product is almost zero\n");
    vec.u = 0.5-tmp3.u;
    vec.v = 1.0-tmp3.v;
    crossp = (tmp2.u - tmp1.u)*vec.v - (tmp2.v - tmp1.v)*vec.u;
  }
  norm = sqrt(vec.u*vec.u + vec.v*vec.v);
  vec.u /= norm;
  vec.v /= norm;
  pt2D->u = tmp3.u + 0.1*vec.u;
  pt2D->v = tmp3.v + 0.1*vec.v;
  if (crossp > 0.0) *orientation = LEFT;
  else *orientation = RIGHT;
}  

#ifdef PRIM_TEST
int plane_cut_op(first_solid, first_set_patches, cutting_plane, v0, v1, to_keep, ssize, parent_solid)
TRIM_PATCH_INFO **first_solid, **parent_solid;
int first_set_patches, to_keep;
POINT4D *v0, *v1, *cutting_plane;
double ssize;
{
  int i, j, k;
  double xval, yval, zval, uval, vval;
  double minx, maxx, miny, maxy;
  TRIM_PATCH_INFO *second_solid;
  BOUNDINGBOX *boxes, bigBB;
  double v0_mag, v1_mag, cutP_mag;
  double Xmat[4][4];
  POINT4D in_pts[8], out_pts[8], fin_in_pts[4], fin_out_pts[4];
  int parent_patches, second_set_patches;
  int frep, srep;

  v0_mag = v0->dCoord[X] * v0->dCoord[X] +
	   v0->dCoord[Y] * v0->dCoord[Y] +
	   v0->dCoord[Z] * v0->dCoord[Z];
  v1_mag = v1->dCoord[X] * v1->dCoord[X] +
	   v1->dCoord[Y] * v1->dCoord[Y] +
	   v1->dCoord[Z] * v1->dCoord[Z];
  cutP_mag = cutting_plane->dCoord[X] * cutting_plane->dCoord[X] +
	   cutting_plane->dCoord[Y] * cutting_plane->dCoord[Y] +
	   cutting_plane->dCoord[Z] * cutting_plane->dCoord[Z];
  if (!(FP_EQ(v0_mag, 1.0) && FP_EQ(v1_mag, 1.0) && FP_EQ(cutP_mag, 1.0))) {
    printf("Plane vectors are not normalised: %lf %lf %lf\n", 
	   v0_mag, v1_mag, cutP_mag);
    exit(-1);
  }

  second_set_patches = 1;
  ALLOCN(second_solid, TRIM_PATCH_INFO, second_set_patches); 
						/* patch repr. cutting plane */
  for (i=0; i<second_set_patches; i++) {
    second_solid[i].tp.iDeg0 = second_solid[i].tp.iDeg1 = 1;
    allocate_tensor_patch(&(second_solid[i].tp));
    second_solid[i].NumTrimPortions = 1;
    second_solid[i].SplineReprn = 0;
    second_solid[i].TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
    second_solid[i].TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;
    second_solid[i].TrimPortion[0].PiecePortion.NumOfVertices = 4;
    
    /* Additional code to initialise u, v, values:: Aug 5 (Atul) */
    second_solid[i].umax = second_solid[i].vmax = 1.0;
    second_solid[i].umin = second_solid[i].vmin = 0.0;

    for (j=0; j<=4; j++) {
      second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[j][0] = 0.0;
      second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[j][1] = 0.0;
    }
    second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
    second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
    second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
    second_solid[i].TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
  }
  
  ALLOCN(boxes, BOUNDINGBOX, first_set_patches);
  for (i=0; i<3; i++) {
    bigBB.i3Dbox[i].low = 100000;
    bigBB.i3Dbox[i].high = -100000;
  }
  for (i=0; i<first_set_patches; i++)
    compute_bounding_box((*first_solid)[i].tp, &(boxes[i]));

  for (i=0; i<first_set_patches; i++) {
    for (j=0; j<3; j++) {
      if (boxes[i].i3Dbox[j].low < bigBB.i3Dbox[j].low)
        bigBB.i3Dbox[j].low = boxes[i].i3Dbox[j].low;
      if (boxes[i].i3Dbox[j].high > bigBB.i3Dbox[j].high)
        bigBB.i3Dbox[j].high = boxes[i].i3Dbox[j].high;
    }
  }
  for (i=0; i<3; i++) {
    bigBB.i3Dbox[i].low -= 5.0;
    bigBB.i3Dbox[i].high += 5.0;
  }

  in_pts[0].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[0].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[0].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[1].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[1].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[1].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[2].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[2].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[2].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[3].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[3].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[3].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[4].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[4].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[4].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[5].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[5].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[5].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[6].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[6].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[6].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[7].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[7].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[7].dCoord[Z] = bigBB.i3Dbox[2].high;
  for (i=0; i<8; i++) {
    in_pts[i].dCoord[W] = 1.0;
    for (j=0; j<2; j++)
      out_pts[i].dCoord[j] = 0.0;
    out_pts[i].dCoord[2] = -cutting_plane->dCoord[W];
    out_pts[i].dCoord[W] = 1.0;
  }

  for (i=0; i<4; i++) {
    for (j=0; j<3; j++)
      fin_out_pts[i].dCoord[j] = 0.0;
    fin_in_pts[i].dCoord[Z] = -cutting_plane->dCoord[W];
    fin_in_pts[i].dCoord[W] = 1.0;
    fin_out_pts[i].dCoord[W] = 1.0;
  }

  for (i=0; i<4; i++)
  for (j=0; j<4; j++)
    Xmat[i][j] = 0.0;
  Xmat[0][0] = v0->dCoord[X];
  Xmat[0][1] = v0->dCoord[Y];
  Xmat[0][2] = v0->dCoord[Z];
  Xmat[1][0] = v1->dCoord[X];
  Xmat[1][1] = v1->dCoord[Y];
  Xmat[1][2] = v1->dCoord[Z];
  Xmat[2][0] = cutting_plane->dCoord[X];
  Xmat[2][1] = cutting_plane->dCoord[Y];
  Xmat[2][2] = cutting_plane->dCoord[Z];
  Xmat[3][3] = 1.0;
 
  for (k=0; k<8; k++) {
    for (i=0; i<2; i++) 
    for (j=0; j<3; j++) 
      out_pts[k].dCoord[i] += Xmat[i][j]*in_pts[k].dCoord[j];
  }

  minx = maxx = out_pts[0].dCoord[X];
  miny = maxy = out_pts[0].dCoord[Y];
  for (i=1; i<8; i++) {
    if (out_pts[i].dCoord[X] < minx) minx = out_pts[i].dCoord[X]; 
    if (out_pts[i].dCoord[X] > maxx) maxx = out_pts[i].dCoord[X]; 
    if (out_pts[i].dCoord[Y] < miny) miny = out_pts[i].dCoord[Y]; 
    if (out_pts[i].dCoord[Y] > maxy) maxy = out_pts[i].dCoord[Y]; 
  }

  fin_in_pts[0].dCoord[X] = minx;
  fin_in_pts[0].dCoord[Y] = miny;
  fin_in_pts[1].dCoord[X] = minx;
  fin_in_pts[1].dCoord[Y] = maxy;
  fin_in_pts[2].dCoord[X] = maxx;
  fin_in_pts[2].dCoord[Y] = miny;
  fin_in_pts[3].dCoord[X] = maxx;
  fin_in_pts[3].dCoord[Y] = maxy;

  for (k=0; k<4; k++) {
    for (i=0; i<3; i++) 
    for (j=0; j<3; j++) 
      fin_out_pts[k].dCoord[i] += Xmat[j][i]*fin_in_pts[k].dCoord[j];
  }

  for (i=0; i<=1; i++)
  for (j=0; j<=1; j++)
    second_solid[0].tp.coeffs[i][j] = fin_out_pts[2*i + j];

  copy_ten_p(&(second_solid[0].tp), &(second_solid[0].orig_tp));

  frep = srep = 1;
  /* changing call to perform_CSG to accomodate splines */
  parent_patches = perform_CSG(first_solid, first_set_patches, &second_solid,
			       second_set_patches, frep, srep,
			       PLANE_CUT, to_keep, cutting_plane,
			       ssize, parent_solid); 

  free_trim_patch(second_solid);
  myfree(boxes);
  return parent_patches;
}
#else
int plane_cut_op(first_solid, first_set_patches, cutting_plane, v0, v1, to_keep, ssize, parent_solid)
TRIM_PATCH_INFO **first_solid, **parent_solid;
int first_set_patches, to_keep;
POINT4D *v0, *v1, *cutting_plane;
double ssize;
{
  int i, j, k;
  double xval, yval, zval, uval, vval;
  double minx, maxx, miny, maxy, minz, zdist;
  TRIM_PATCH_INFO *second_solid;
  BOUNDINGBOX *boxes, bigBB;
  double v0_mag, v1_mag, cutP_mag;
  double Xmat[4][4];
  POINT4D in_pts[8], out_pts[8], fin_in_pts[4], fin_out_pts[4];
  int parent_patches, second_set_patches;
  int frep, srep, sense, totpatches;
  ppped_t pp; /* I am going to construct a ppped instead of the plane */
	      /* and use the difference operation */
  node_t np; /* using this structure so that I can use some of the */
	     /* routines again */

  v0_mag = v0->dCoord[X] * v0->dCoord[X] +
	   v0->dCoord[Y] * v0->dCoord[Y] +
	   v0->dCoord[Z] * v0->dCoord[Z];
  v1_mag = v1->dCoord[X] * v1->dCoord[X] +
	   v1->dCoord[Y] * v1->dCoord[Y] +
	   v1->dCoord[Z] * v1->dCoord[Z];
  cutP_mag = cutting_plane->dCoord[X] * cutting_plane->dCoord[X] +
	   cutting_plane->dCoord[Y] * cutting_plane->dCoord[Y] +
	   cutting_plane->dCoord[Z] * cutting_plane->dCoord[Z];
  if (!(FP_EQ(v0_mag, 1.0) && FP_EQ(v1_mag, 1.0) && FP_EQ(cutP_mag, 1.0))) {
    printf("Plane vectors are not normalised: %lf %lf %lf\n", 
	   v0_mag, v1_mag, cutP_mag);
    exit(-1);
  }

  ALLOCN(boxes, BOUNDINGBOX, first_set_patches);
  for (i=0; i<3; i++) {
    bigBB.i3Dbox[i].low = 100000;
    bigBB.i3Dbox[i].high = -100000;
  }
  for (i=0; i<first_set_patches; i++)
    compute_bounding_box((*first_solid)[i].tp, &(boxes[i]));

  for (i=0; i<first_set_patches; i++) {
    for (j=0; j<3; j++) {
      if (boxes[i].i3Dbox[j].low < bigBB.i3Dbox[j].low)
        bigBB.i3Dbox[j].low = boxes[i].i3Dbox[j].low;
      if (boxes[i].i3Dbox[j].high > bigBB.i3Dbox[j].high)
        bigBB.i3Dbox[j].high = boxes[i].i3Dbox[j].high;
    }
  }
  for (i=0; i<3; i++) {
    bigBB.i3Dbox[i].low -= 5.0;
    bigBB.i3Dbox[i].high += 5.0;
  }

  in_pts[0].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[0].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[0].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[1].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[1].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[1].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[2].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[2].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[2].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[3].dCoord[X] = bigBB.i3Dbox[0].low;
  in_pts[3].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[3].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[4].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[4].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[4].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[5].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[5].dCoord[Y] = bigBB.i3Dbox[1].low;
  in_pts[5].dCoord[Z] = bigBB.i3Dbox[2].high;
  in_pts[6].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[6].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[6].dCoord[Z] = bigBB.i3Dbox[2].low;
  in_pts[7].dCoord[X] = bigBB.i3Dbox[0].high;
  in_pts[7].dCoord[Y] = bigBB.i3Dbox[1].high;
  in_pts[7].dCoord[Z] = bigBB.i3Dbox[2].high;

  for (i=0; i<8; i++) {
    in_pts[i].dCoord[W] = 1.0;
    for (j=0; j<2; j++)
      out_pts[i].dCoord[j] = 0.0;
    out_pts[i].dCoord[2] = -cutting_plane->dCoord[W];
    out_pts[i].dCoord[W] = 1.0;
  }

  minz = HUGE_VAL;
  for (i=0; i<8; i++) {
    zdist = 0.0;
    for (j=0; j<4; j++) 
      zdist += in_pts[i].dCoord[j]*cutting_plane->dCoord[j];
    if (zdist < minz)
      minz = zdist;
  }
  if (minz > -EPSILON) {
    printf("plane_cut_op: minz is positive\n");
    printf("So I just copy first_solid and return in parent_solid\n");
    parent_patches = first_set_patches;
    ALLOCN((*parent_solid), TRIM_PATCH_INFO, parent_patches);
    k = 0;
    for (i=0; i<first_set_patches; i++) {
      copy_trim_patch((*first_solid)+i, (*parent_solid)+k);
      k++;
    }
    return parent_patches;
  } else {
    zdist = -minz + 10.0; /* Just making sure that the back of the */
			  /* ppped does not cut the first_solid */
  }

  for (j=0; j<3; j++)
    fin_out_pts[0].dCoord[j] = 0.0;
  fin_in_pts[0].dCoord[Z] = -cutting_plane->dCoord[W];
  fin_in_pts[0].dCoord[W] = 1.0;
  fin_out_pts[0].dCoord[W] = 1.0;

  for (i=0; i<4; i++)
  for (j=0; j<4; j++)
    Xmat[i][j] = 0.0;
  Xmat[0][0] = v0->dCoord[X];
  Xmat[0][1] = v0->dCoord[Y];
  Xmat[0][2] = v0->dCoord[Z];
  Xmat[1][0] = v1->dCoord[X];
  Xmat[1][1] = v1->dCoord[Y];
  Xmat[1][2] = v1->dCoord[Z];
  Xmat[2][0] = cutting_plane->dCoord[X];
  Xmat[2][1] = cutting_plane->dCoord[Y];
  Xmat[2][2] = cutting_plane->dCoord[Z];
  Xmat[3][3] = 1.0;
 
  for (k=0; k<8; k++) {
    for (i=0; i<2; i++) 
      for (j=0; j<3; j++) 
	out_pts[k].dCoord[i] += Xmat[i][j]*in_pts[k].dCoord[j];
  }

  minx = maxx = out_pts[0].dCoord[X];
  miny = maxy = out_pts[0].dCoord[Y];
  for (i=1; i<8; i++) {
    if (out_pts[i].dCoord[X] < minx) minx = out_pts[i].dCoord[X]; 
    if (out_pts[i].dCoord[X] > maxx) maxx = out_pts[i].dCoord[X]; 
    if (out_pts[i].dCoord[Y] < miny) miny = out_pts[i].dCoord[Y]; 
    if (out_pts[i].dCoord[Y] > maxy) maxy = out_pts[i].dCoord[Y]; 
  }

  for (i=0; i<3; i++) {
    pp.v0[i] = v0->dCoord[i];
    pp.v1[i] = v1->dCoord[i];
    pp.v2[i] = -cutting_plane->dCoord[i];
  }
  pp.v0[3] = pp.v1[3] = pp.v2[3] = 1.0;

  pp.len0 = maxx - minx;
  pp.len1 = maxy - miny;
  pp.len2 = zdist;

  fin_in_pts[0].dCoord[X] = minx;
  fin_in_pts[0].dCoord[Y] = miny;

  for (i=0; i<3; i++) 
    for (j=0; j<3; j++) 
      fin_out_pts[0].dCoord[i] += Xmat[j][i]*fin_in_pts[0].dCoord[j];

  pp.c0.x = fin_out_pts[0].dCoord[0];
  pp.c0.y = fin_out_pts[0].dCoord[1];
  pp.c0.z = fin_out_pts[0].dCoord[2];

  tpatch_idx = 0;
  sense = generate_patches_for_ppped(&pp);
  totpatches = 6;
  
  np.prim = (primitive_t *) new(sizeof(primitive_t));
  np.left = np.right = NULL;
  np.nodetype = T_LEAF;
  np.prim->ptype = P_PPPED;
  np.patch = (TENSOR_PATCH *) new(totpatches *
				   sizeof(TENSOR_PATCH));
  memcpy((void *) np.patch, (void *) tpatch, 
	 totpatches * sizeof(TENSOR_PATCH));
  np.npatches = totpatches;
  add_trim_patch_info(&np, P_PPPED, sense, 0, 0, 0);
  
  frep = srep = 1;
  /* changing call to perform_CSG to accomodate splines */
  parent_patches = perform_CSG(first_solid, first_set_patches, &np.trim,
			       np.npatches, frep, srep,
			       DIFFERENCE, to_keep, NULL,
			       ssize, parent_solid); 

  free_trim_patch(second_solid);
  myfree(boxes);
  return parent_patches;
}
#endif

int resolveNoXsection(first_solid, first_set_patches, second_solid, second_set_patches, parent_solid, operation)
TRIM_PATCH_INFO **first_solid, **second_solid, **parent_solid;
int first_set_patches, second_set_patches;
int operation;
{
  int i, j, k, p, q, r;
  int parent_patches;
  double utmp, vtmp;
  FILE *fp;
  double normX, normY, normZ;
  int is_facing_outside;
  double xval, yval, zval, uval, vval;
  TENSOR_PATCH norm_tp;

  switch (operation) {
    case DOUNION: 
		parent_patches = first_set_patches + second_set_patches;
		ALLOCN((*parent_solid), TRIM_PATCH_INFO, parent_patches);
		k = 0;
		for (i=0; i<first_set_patches; i++) {
		  copy_trim_patch((*first_solid)+i, (*parent_solid)+k);
		  k++;
		}
		for (i=0; i<second_set_patches; i++) {
		  copy_trim_patch((*second_solid)+i,
				  (*parent_solid)+k);
		  if ((*parent_solid)[k].SplineReprn) {
		    for (j=0; j<(*parent_solid)[k].NumTrimPortions; j++)
		      for (p=0;
			   p<(*parent_solid)[k].TrimPortion[j].SplinePortion.
			   NumOfCurves; p++)
			(*parent_solid)[k].TrimPortion[j].SplinePortion.AdjInfo[p] = 
			  (*parent_solid)[k].TrimPortion[j].SplinePortion.AdjInfo[p] +
			  first_set_patches;
		  }
		  k++;
		}
		break;
    case DIFFERENCE: 
    case PLANE_CUT: 
		parent_patches = first_set_patches;
		ALLOCN((*parent_solid), TRIM_PATCH_INFO, parent_patches);
		k = 0;
		for (i=0; i<first_set_patches; i++) {
		  copy_trim_patch((*first_solid)+i, (*parent_solid)+k);
		  k++;
		}
		break;
    case INTERSECTION: 
		parent_patches = 0;
		break;
  }

  return parent_patches;
}

#ifdef _CALC_IMPLICIT
void CalculateImplicit(tp0, ndetImplicit, imp_scale)
TENSOR_PATCH *tp0;
NDETERMINANT *ndetImplicit;
double imp_scale;
{
  int i, j, k;
  TENSOR_PATCH tpPatch0;
  NDETERMINANT ndet;

  tpPatch0.iDeg0 = tp0->iDeg0;
  tpPatch0.iDeg1 = tp0->iDeg1;
  allocate_tensor_patch(&tpPatch0);

  /* Now calculating the exact Scale factor */
  Scale = imp_scale;
  for (i=0; i<=tp0->iDeg0; i++)
    for (j=0; j<=tp0->iDeg1; j++)
      for (k=X; k<=W; k++)
	tpPatch0.coeffs[i][j].dCoord[k] = tp0->coeffs[i][j].dCoord[k];

  /* Scale the tensor patch */
  scale_tensor_patch(&tpPatch0, 1);

  bernsteinToPowerBasis( tpPatch0 );

  /* tpPatch0 is now in the power basis under the reparametrization 
     of s=s/1-s, t=t/1-t, but tpPatch1 is still a bezier tensor prod
     patch */

  allocDeterminant( &ndet, tpPatch0.iDeg0, tpPatch0.iDeg1, UNFACTORED );

  implicitize( tpPatch0, ndet );

  /*  powerToBernsteinBasis( tpPatch0 ); */

  allocDeterminant( ndetImplicit, tpPatch0.iDeg0, tpPatch0.iDeg1, FINAL );
  moveFactoredToFinalDeterminant( (*ndetImplicit), ndet, 
				 tpPatch0.iDeg0, tpPatch0.iDeg1 );

  Scale = 1.0;
  free_ten_p(&tpPatch0);
  free_alloc_determ(&ndet);
}
#endif

/* @@@ */
int perform_CSG(first_solid, first_set_patches, second_solid,
		second_set_patches, frep, srep, operation, to_keep,
		cutting_plane, ssize, parent_solid) 
TRIM_PATCH_INFO **first_solid, **second_solid, **parent_solid;
int first_set_patches, second_set_patches; /* numbers of patches */
int operation; /* one of DOUNION, INTERSECTION, DIFFERENCE */
int to_keep, frep, srep; /* frep == 1 means spline, 0 means */
				    /* piecewise for first solid, */
				    /* similar for second solid */ 
POINT4D *cutting_plane;
double ssize;
{
  int num_pts;
  double ss;
  UV_SPLINE_TYPE *ic;
  int *init_fsolid_curves, *init_ssolid_curves;
  int i, j, k, l, m, ncurves, orientation, NumInt;
  int NumInt1, NumInt2, NumInt3;
  double xval, yval, zval, uval, vval;
  double intxval, intyval, intzval;
  POINT2D PrevPt, NextPt;
  POINT4D fpt, ept;
  POINT2D uvPair, intuvPair, tmp2D;
  TRIM_PATCH_INFO *tmps1, *tmps2, *tmp_psolid, *tp;
  int parent_patches, tmp_ppatches;
  int *patchNumber, *first_or_second;
  FILE *fp;
  int p, q, r, numTrimSegs;
  int npts, half_npts, retFXSC /* return value from findXsectionCurve */ ;
  point_typ inside_pt;
  LIST_POINT2D *tmp_ptr1, *tmp_ptr2;
  double utmp, vtmp;
  double normX, normY, normZ;
  int is_facing_outside;
  TENSOR_PATCH norm_tp;
  CURVE_MATCHING_TYPE fcmt[100], scmt[100]; /* assuming 100 is enough */
  TRIM_PORTION_SPLINE_TYPE **fop_tpst, **sop_tpst, *tmp_tpst;
  neighbour_t fn_info[MAX_PATCHES_PER_SOLID][MAX_PARTNS_PER_PATCH], 
              sn_info[MAX_PATCHES_PER_SOLID][MAX_PARTNS_PER_PATCH];
  neighbour_t *global_graph_f, *global_graph_s;
  int num_partn_first[MAX_PATCHES_PER_SOLID], 
      num_partn_second[MAX_PATCHES_PER_SOLID];
  int fcmt_num, scmt_num, f_num_comp, s_num_comp, adj_patch, deg;
  int fNumInt[5], sNumInt[5]; /* 5 should be enough for number of */
			      /* components per solid */

  SPLIT_TPST_STRUCT sts_array[MAX_SPLIT_ARRAY_LENGTH];
  int sts_arr_len;
  int first_sol_comp[MAX_PATCHES_PER_SOLID*MAX_PARTNS_PER_PATCH],
      second_sol_comp[MAX_PATCHES_PER_SOLID*MAX_PARTNS_PER_PATCH]; 
  short first_orig_region_num[MAX_PATCHES_PER_SOLID][MAX_PARTNS_PER_PATCH], 
        sec_orig_region_num[MAX_PATCHES_PER_SOLID][MAX_PARTNS_PER_PATCH];
        /* These two variables contain the original region number of */
	/* the current partition number in a patch. This region number */
	/* is used to access first_same_solid and second_same_solid */
	/* when obtaining graph of the resulting solid */
  short final_patch_number[MAX_PATCHES_PER_SOLID][MAX_PARTNS_PER_PATCH];
        /* This table stores the number of this patch in the final */
	/* solid */
  short **spec_crvs_for_ovr_loops_first, **spec_crvs_for_ovr_loops_second;
  short partn_num, prev_dir, dir, curve_num; /* This is to split some curves just */
			                     /* because their adjacent patches had to */
			                     /* be split because of loops */
  double tmp_doub;
  short *same_solid;
  short checkXnCrvs;

#ifdef _CALC_IMPLICIT
  NDETERMINANT *first_sol_implicit, *sec_sol_implicit;
  double imp_scale = -1.0;

  for (i=0; i<first_set_patches; i++)
    for (j=0; j<=(*first_solid)[i].tp.iDeg0; j++)
      for (k=0; k<=(*first_solid)[i].tp.iDeg1; k++)
	for (l=0; l<3; l++)
	  if (imp_scale <
	      fabs((*first_solid)[i].tp.coeffs[j][k].dCoord[l]))
	    imp_scale =
	      fabs((*first_solid)[i].tp.coeffs[j][k].dCoord[l]);
  for (i=0; i<second_set_patches; i++)
    for (j=0; j<=(*second_solid)[i].tp.iDeg0; j++)
      for (k=0; k<=(*second_solid)[i].tp.iDeg1; k++)
	for (l=0; l<3; l++)
	  if (imp_scale <
	      fabs((*second_solid)[i].tp.coeffs[j][k].dCoord[l]))
	    imp_scale =
	      fabs((*second_solid)[i].tp.coeffs[j][k].dCoord[l]);

  ALLOCN(first_sol_implicit, NDETERMINANT, first_set_patches);
  ALLOCN(sec_sol_implicit, NDETERMINANT, second_set_patches);

  imp_scale /= 10.0;
#ifdef __PERFORM_TIME
  Time1 = GetTimer;
#endif
  for (i=0; i<first_set_patches; i++)
    CalculateImplicit(&(*first_solid)[i].tp, &first_sol_implicit[i], imp_scale);
  for (i=0; i<second_set_patches; i++)
    CalculateImplicit(&(*second_solid)[i].tp, &sec_sol_implicit[i], imp_scale);
#ifdef __PERFORM_TIME
  Time2 = GetTimer;
  ImplicitTime += TIMEDIFF(Time2, Time1);
#endif
  
#endif

  ALLOCMAT(spec_crvs_for_ovr_loops_first, MAX_PATCHES_PER_SOLID,
	   MAX_SPLINES_PER_TRIM_REGION, short);
  ALLOCMAT(spec_crvs_for_ovr_loops_second, MAX_PATCHES_PER_SOLID,
	   MAX_SPLINES_PER_TRIM_REGION, short);
  for (i=0; i<MAX_PATCHES_PER_SOLID; i++)
    for (j=0; j<MAX_SPLINES_PER_TRIM_REGION; j++)
      spec_crvs_for_ovr_loops_first[i][j] =
	spec_crvs_for_ovr_loops_second[i][j] = -1;

  for (i=0; i<MAX_PATCHES_PER_SOLID; i++)
    for (j=0; j<MAX_PARTNS_PER_PATCH; j++)
      fn_info[i][j].nnb = sn_info[i][j].nnb = 0;

  ALLOCN(patchNumber, int, first_set_patches+second_set_patches);
  ALLOCN(first_or_second, int, first_set_patches+second_set_patches);

  /* Depending on the values of frep and srep allocate the static */
  /* globals */
  if (frep != srep) {
    printf("perform_CSG: frep and srep are not the same\n");
    exit(-1);
  }

  /* this is common to both reprns - spline and piecewise */
  ALLOCN(fs_nc, int, first_set_patches);
  ALLOCN(ss_nc, int, second_set_patches);
  for (i=0; i<first_set_patches; i++)
    fs_nc[i] = 0;
  for (i=0; i<second_set_patches; i++)
    ss_nc[i] = 0;

  if (!frep) {
    ALLOCN(tmps1, TRIM_PATCH_INFO, first_set_patches);
    ALLOCN(tmps2, TRIM_PATCH_INFO, second_set_patches);
    ALLOCN(init_fsolid_curves, int, first_set_patches);
    ALLOCN(init_ssolid_curves, int, second_set_patches);

    /* allocate domain_curves */
    ALLOCN(fs_pw_curves, DOMAIN_INT_CURVE *, first_set_patches);
    ALLOCN(ss_pw_curves, DOMAIN_INT_CURVE *, second_set_patches);
    for (i=0; i<first_set_patches; i++) 
      ALLOCN(fs_pw_curves[i], DOMAIN_INT_CURVE, MAX_TRIM_CURVES_PER_PATCH);
    for (i=0; i<second_set_patches; i++) 
      ALLOCN(ss_pw_curves[i], DOMAIN_INT_CURVE, MAX_TRIM_CURVES_PER_PATCH);
    /* initialize the structures */
    for (i=0; i<first_set_patches; i++)
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	fs_pw_curves[i][j].planeCurve = NULL;
    for (i=0; i<second_set_patches; i++)
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	ss_pw_curves[i][j].planeCurve = NULL;

    for (i=0; i<first_set_patches; i++) {
      /* copy_trim_patch_info((*first_solid)[i], &(tmps1[i])); */
      copy_trim_patch(&((*first_solid)[i]), &(tmps1[i]));
      init_fsolid_curves[i] = 0;
    }
    for (i=0; i<second_set_patches; i++) {
      /* copy_trim_patch_info((*second_solid)[i], &(tmps2[i])); */
      copy_trim_patch(&((*second_solid)[i]), &(tmps2[i]));
      init_ssolid_curves[i] = 0;
    }
  } else {
    /* allocate splines */
    ALLOCN(fs_spl_curves, MERGED_CURVE *, first_set_patches);
    ALLOCN(ss_spl_curves, MERGED_CURVE *, second_set_patches);
    for (i=0; i<first_set_patches; i++) 
      ALLOCN(fs_spl_curves[i], MERGED_CURVE, MAX_TRIM_CURVES_PER_PATCH);
    for (i=0; i<second_set_patches; i++) 
      ALLOCN(ss_spl_curves[i], MERGED_CURVE, MAX_TRIM_CURVES_PER_PATCH);
    /* initialize the structures */
    for (i=0; i<first_set_patches; i++)
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	fs_spl_curves[i][j].int_curve = NULL;
    for (i=0; i<second_set_patches; i++)
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	ss_spl_curves[i][j].int_curve = NULL;

    for (i=0; i<100; i++)
      for (j=0; j<5; j++)
	for (k=0; k<2; k++)
	  for (l=0; l<3; l++) 
	    fXsc[i][j][k][l] = sXsc[i][j][k][l] = -1;
  }

  for (i=0; i<20; i++)
    first_complete_ovrlap[i] = second_complete_ovrlap[i] = 0;

#ifdef _CALC_IMPLICIT
  retFXSC = findXsectionCurve(first_solid, first_set_patches,
			      first_sol_implicit, second_solid, 
		    	      second_set_patches, sec_sol_implicit, imp_scale,
			      frep, operation, ssize,
			      spec_crvs_for_ovr_loops_first,
			      spec_crvs_for_ovr_loops_second); 
#else
  retFXSC = findXsectionCurve(first_solid, first_set_patches, second_solid, 
		    	      second_set_patches, frep, operation,
			      ssize,
			      spec_crvs_for_ovr_loops_first,
			      spec_crvs_for_ovr_loops_second); 
#endif

  for (i=0; i<first_set_patches; i++) {
    int new_num_crvs, ind1, ind2;
    INTERSECTION_CURVE *lf_ptr, *ll_ptr;
    TRIM_PORTION_SPLINE_TYPE *loc_tpst;
    MERGED_CURVE *temp_mc;
    int temp_mc_ind;
    POINT2D st, end;
    double param;

    if (fs_nc[i] > 0) {
      temp_mc_ind = 0;
      ind1 = 0;
      ALLOCN(temp_mc, MERGED_CURVE, fs_nc[i]);
      while (ind1<fs_nc[i]) {
	return_st_end_in_p2d(&(fs_spl_curves[i][ind1]), &st, &end, &lf_ptr, &ll_ptr);
	if (EQUAL2D_EPS(st, end, 1e-6)) {
	  ind1++;
	  continue;
	}
	loc_tpst = &(*first_solid)[i].TrimPortion[0].SplinePortion;
	ind2 = 0;
	while ((ind2<loc_tpst->NumOfCurves) && !EvaluateParam(&loc_tpst->UV_Splines[ind2], &st, &param))
	  ind2++;
	if (ind2 == loc_tpst->NumOfCurves) {
	  temp_mc[temp_mc_ind++].int_curve = fs_spl_curves[i][ind1].int_curve;
	  for (ind2=ind1+1; ind2<fs_nc[i]; ind2++)
	    fs_spl_curves[i][ind2-1].int_curve = fs_spl_curves[i][ind2].int_curve;
	  fs_spl_curves[i][fs_nc[i]].int_curve = NULL;
	  fs_nc[i]--;
	  continue;
	}
	ind2 = 0;
	while ((ind2<loc_tpst->NumOfCurves) && !EvaluateParam(&loc_tpst->UV_Splines[ind2], &end, &param))
	  ind2++;
	if (ind2 == loc_tpst->NumOfCurves) {
	  temp_mc[temp_mc_ind++].int_curve = fs_spl_curves[i][ind1].int_curve;
	  for (ind2=ind1+1; ind2<fs_nc[i]; ind2++)
	    fs_spl_curves[i][ind2-1].int_curve = fs_spl_curves[i][ind2].int_curve;
	  fs_spl_curves[i][fs_nc[i]].int_curve = NULL;
	  fs_nc[i]--;
	  continue;
	}
	ind1++;
      }
	
      if (temp_mc_ind > 0) {
	new_num_crvs = 1;
	for (j=1; j<temp_mc_ind; j++)
	  MergeSplinesFurther(&temp_mc, &new_num_crvs, &temp_mc[j], 1);
	for (j=0; j<new_num_crvs; j++)
	  fs_spl_curves[i][(fs_nc[i])++].int_curve = temp_mc[j].int_curve;
      }
      myfree(temp_mc);
    }
  }

  for (i=0; i<second_set_patches; i++) {
    int new_num_crvs, ind1, ind2;
    INTERSECTION_CURVE *lf_ptr, *ll_ptr;
    TRIM_PORTION_SPLINE_TYPE *loc_tpst;
    MERGED_CURVE *temp_mc;
    int temp_mc_ind;
    POINT2D st, end;
    double param;

    if (ss_nc[i] > 0) {
      temp_mc_ind = 0;
      ind1 = 0;
      ALLOCN(temp_mc, MERGED_CURVE, ss_nc[i]);
      while (ind1<ss_nc[i]) {
	return_st_end_in_p2d(&(ss_spl_curves[i][ind1]), &st, &end, &lf_ptr, &ll_ptr);
	if (EQUAL2D_EPS(st, end, 1e-6)) {
	  ind1++;
	  continue;
	}
	loc_tpst = &(*second_solid)[i].TrimPortion[0].SplinePortion;
	ind2 = 0;
	while ((ind2<loc_tpst->NumOfCurves) && !EvaluateParam(&loc_tpst->UV_Splines[ind2], &st, &param))
	  ind2++;
	if (ind2 == loc_tpst->NumOfCurves) {
	  temp_mc[temp_mc_ind++].int_curve = ss_spl_curves[i][ind1].int_curve;
	  for (ind2=ind1+1; ind2<ss_nc[i]; ind2++)
	    ss_spl_curves[i][ind2-1].int_curve = ss_spl_curves[i][ind2].int_curve;
	  ss_spl_curves[i][ss_nc[i]].int_curve = NULL;
	  ss_nc[i]--;
	  continue;
	}
	ind2 = 0;
	while ((ind2<loc_tpst->NumOfCurves) && !EvaluateParam(&loc_tpst->UV_Splines[ind2], &end, &param))
	  ind2++;
	if (ind2 == loc_tpst->NumOfCurves) {
	  temp_mc[temp_mc_ind++].int_curve = ss_spl_curves[i][ind1].int_curve;
	  for (ind2=ind1+1; ind2<ss_nc[i]; ind2++)
	    ss_spl_curves[i][ind2-1].int_curve = ss_spl_curves[i][ind2].int_curve;
	  ss_spl_curves[i][ss_nc[i]].int_curve = NULL;
	  ss_nc[i]--;
	  continue;
	}
	ind1++;
      }
	
      if (temp_mc_ind > 0) {
	new_num_crvs = 1;
	for (j=1; j<temp_mc_ind; j++)
	  MergeSplinesFurther(&temp_mc, &new_num_crvs, &temp_mc[j], 0);
	for (j=0; j<new_num_crvs; j++)
	  ss_spl_curves[i][(ss_nc[i])++].int_curve = temp_mc[j].int_curve;
      }
      myfree(temp_mc);
    }
  }

  /* I am disabling the checkXnCrvs flag if two patches overlap in the */
  /* opposite direction because the curves as I compute it will not be */
  /* closed in space */
  checkXnCrvs = 1;
  for (i=0; i<NumOvrPatches; i++) {
    if (!overlap_pairs[i].same_side)
      break;
  }
  if (i != NumOvrPatches)
    checkXnCrvs = 0;

#ifdef _LOTS_OF_HOLES
  (*parent_solid) = (*first_solid);
  return first_set_patches;
#else
  if (!retFXSC) 
    return resolveNoXsection(first_solid, first_set_patches, second_solid,
		             second_set_patches, parent_solid, operation);

  /* check if the curves in each solid form a closed loop on the */
  /* surface */
  if (frep && 
      !CheckIntersectionCurves(first_solid, fs_spl_curves, 
			       first_set_patches, fs_nc,
			       &fcmt[0], &fcmt_num,
			       second_solid, ss_spl_curves, 
			       second_set_patches, ss_nc, 
			       &scmt[0], &scmt_num, !checkXnCrvs)) {
    printf("Intersection Curves do not close properly\n");
    exit(-1);
  }

  /* For same_side overlap, I revert all the intersection curves of */
  /* the second patch which overlaps to the number before projection */
  for (i=0; i<NumOvrPatches; i++)
    if (overlap_pairs[i].same_side)
      ss_nc[overlap_pairs[i].ptches.patchNum1] = NumCurvesBeforeProjection[i];

  /* This code is used only if we have piecewise reprns */
  if (!frep) 
    for (i=0; i<first_set_patches; i++) 
      for (j=0; j<(*first_solid)[i].NumTrimPortions; j++) {
	k = (*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices - 1;
	PrevPt.u = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][0];
	PrevPt.v = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][1];
	NextPt.u = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][0];
	NextPt.v = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][1];
	if (EQUAL2D_EPS(PrevPt, NextPt, 2e-6)) {
	  for (l=k; l<(*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices; l++) {
	    (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][0] = 
	      (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][0];
	    (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][1] = 
	      (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][1];
	  }
	  (*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices--;
	}
	k = 1;
	while (k < (*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices) {
	  PrevPt.u = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][0];
	  PrevPt.v = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][1];
	  NextPt.u = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][0];
	  NextPt.v = (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][1];
	  if (EQUAL2D_EPS(PrevPt, NextPt, 2e-6)) {
	    for (l=k+1; l<(*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices; l++) {
	      (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][0] = 
		(*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][0];
	      (*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][1] = 
		(*first_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][1];
	    }
	    (*first_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices--;
	  } else {
	    k++;
	  }
	}
      }

  if (!srep)
    for (i=0; i<second_set_patches; i++) 
      for (j=0; j<(*second_solid)[i].NumTrimPortions; j++) {
	k = (*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices - 1;
	PrevPt.u = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][0];
	PrevPt.v = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][1];
	NextPt.u = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][0];
	NextPt.v = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][1];
	if (EQUAL2D_EPS(PrevPt, NextPt, 2e-6)) {
	  for (l=k; l<(*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices; l++) {
	    (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][0] = 
	      (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][0];
	    (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][1] = 
	      (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][1];
	  }
	  (*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices--;
	}
	k = 1;
	while (k < (*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices) {
	  PrevPt.u = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][0];
	  PrevPt.v = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k][1];
	  NextPt.u = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][0];
	  NextPt.v = (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[k+1][1];
	  if (EQUAL2D_EPS(PrevPt, NextPt, 2e-6)) {
	    for (l=k+1; 
		 l<(*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices; l++) {
	      (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][0] = 
		(*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][0];
	      (*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l][1] = 
		(*second_solid)[i].TrimPortion[j].PiecePortion.poly_vertex[l+1][1];
	    }
	    (*second_solid)[i].TrimPortion[j].PiecePortion.NumOfVertices--;
	  } else {
	    k++;
	  }
	}
      }

#ifdef _PRINT_DOTS
  printf(". ");
  fflush(stdout);
#endif

  if (!frep) {
    tmp_ppatches = 0;
    for (i=0; i<first_set_patches; i++) {
#ifdef _PRINT_DOTS
      printf(". ");
      fflush(stdout);
#endif
      if (planar_ovrlap_flag1[i] == 1) {
	if ((*first_solid)[i].NumTrimPortions == 0)
	  (*first_solid)[i].csg_status = OUT_CSG;
      } else {
	if (init_fsolid_curves[i] > fs_nc[i]) {
	  printf("Something wrong in findXsectionCurve\n");
	  return;
	}
	SplitIntoNewRegions(&((*first_solid)[i]), fs_nc[i], 
			    init_fsolid_curves[i], tmps2,
			    second_set_patches, to_keep, cutting_plane, 1);
      }
      if ((*first_solid)[i].csg_status != OUT_CSG) {
	patchNumber[tmp_ppatches] = i;
	first_or_second[tmp_ppatches] = 1;
	tmp_ppatches++;
      }
    }
  } else {
    tmp_ppatches = 0;
    if ((first_set_patches >= MAX_PATCHES_PER_SOLID) || 
	(second_set_patches >= MAX_PATCHES_PER_SOLID)) {
      printf("perform_CSG: one of the solids has more than 100 patches\n");
      sleep(3);
    }
    sts_arr_len = 0;
    ALLOCMAT(fop_tpst, first_set_patches, MAX_PARTNS_PER_PATCH, 
	     TRIM_PORTION_SPLINE_TYPE);
    NumNghCrvs = 0;
#ifdef __PERFORM_TIME
    Time1 = GetTimer;
#endif
    for (i=0; i<first_set_patches; i++) {
      if ((*first_solid)[i].NumTrimPortions != 1) {
	printf("perform_CSG: a patch has more than one portion\n");
	exit(-1);
      }
      for (j=0; j<(*first_solid)[i].NumTrimPortions; j++) {
	num_partn_first[i] = 
	  PartitionSplinegon(&(*first_solid)[i].TrimPortion[j].SplinePortion, 
			     1, i, j, &fs_spl_curves[i][0], fs_nc[i], 
			     &fop_tpst[i], fn_info[i], first_solid,
			     sts_array, &sts_arr_len,
			     &spec_crvs_for_ovr_loops_first[i][0],
			     *second_solid);  
	if (num_partn_first[i] > MAX_PARTNS_PER_PATCH) {
	  printf("perform_CSG: too many partitions returned\n");
	  exit(-1);
	}
      }
    }
    /* Now in PartitionSplinegon I fill in ngh_crv array. I go through the array */
    /* and assign the appropriate values for AdjInfo and same_solid */
    for (i=0; i<NumNghCrvs; i++) {
      partn_num = GetAdjacentPartn(fop_tpst, ngh_crv[i].patchNo,
				   num_partn_first[ngh_crv[i].patchNo],
				   &ngh_crv[i].spt, &ngh_crv[i].ept, -1,
				   &curve_num, &dir, -1);
      fop_tpst[ngh_crv[i].patchNo][partn_num].AdjInfo[curve_num] =
	ngh_crv[i].adj_pnum; 
      first_same_solid[ngh_crv[i].patchNo][0][partn_num][curve_num] = 0;
    }
#ifdef __PERFORM_TIME
    Time2 = GetTimer;
    PartitionTime += TIMEDIFF(Time2, Time1);
#endif

    /* I am going to do the same thing for the second solid inside */
    /* frep itself because if frep is true, so is srep */
    ALLOCMAT(sop_tpst, second_set_patches, MAX_PARTNS_PER_PATCH, 
	     TRIM_PORTION_SPLINE_TYPE);
    NumNghCrvs = 0;
#ifdef __PERFORM_TIME
    Time1 = GetTimer;
#endif
    for (i=0; i<second_set_patches; i++) {
      if ((*second_solid)[i].NumTrimPortions != 1) {
	printf("perform_CSG: a patch has more than one portion\n");
	exit(-1);
      }
      for (j=0; j<(*second_solid)[i].NumTrimPortions; j++) {
	num_partn_second[i] = 
	  PartitionSplinegon(&(*second_solid)[i].TrimPortion[j].SplinePortion, 
			     0, i, j, &ss_spl_curves[i][0], ss_nc[i], 
			     &sop_tpst[i], sn_info[i], second_solid,
			     sts_array, &sts_arr_len,
			     &spec_crvs_for_ovr_loops_second[i][0],
			     *first_solid);  
	/* This is just to make sure that I am not getting more than */
	/* one trim portion from the same_side overlapping patch */
	/* because I am making some simplifying assumptions based on */
	/* that later */
	for (k=0; k<NumOvrPatches; k++)
	  if ((i == overlap_pairs[k].ptches.patchNum1) &&
	      (num_partn_second[i] > 1) && overlap_pairs[k].same_side) {
	    printf("Overlapping second patch has more than one trim portion\n");
	    break;
	  }
	if (num_partn_second[i] > MAX_PARTNS_PER_PATCH) {
	  printf("perform_CSG: too many partitions returned\n");
	  exit(-1);
	}
      }
    }
#ifdef __PERFORM_TIME
    Time2 = GetTimer;
    PartitionTime += TIMEDIFF(Time2, Time1);
#endif

    /* for every patch of the second_solid that overlaps with the */
    /* first on the same_side, assign portions of the boundary that */
    /* actually lie inside the first solid the same_solid value 0 */
    for (i=0; i<NumOvrPatches; i++) {
      k = sop_tpst[overlap_pairs[i].ptches.patchNum1][0].NumOfCurves-1;
      if (overlap_pairs[i].same_side) {
	for (; k>=0; k--) {
	  for (j=sop_tpst[overlap_pairs[i].ptches.patchNum1][0].NumOfCurves-1;
	       j>ssos[i][k].curve_num; j--)
	    second_same_solid[overlap_pairs[i].ptches.patchNum1][0]
	                     [0][j+ssos[i][k].param_length] = 
	    second_same_solid[overlap_pairs[i].ptches.patchNum1][0][0][j];
	  for (j=ssos[i][k].curve_num+1;
	       j<=ssos[i][k].curve_num+ssos[i][k].param_length; j++)
	    second_same_solid[overlap_pairs[i].ptches.patchNum1][0][0][j] = 
	    second_same_solid[overlap_pairs[i].ptches.patchNum1][0]
                             [0][ssos[i][k].curve_num];
	  AdjustTrimCurves(&sop_tpst[overlap_pairs[i].ptches.patchNum1][0], 
			   ssos[i][k].curve_num, &ssos[i][k].param_arr[0],
			   ssos[i][k].param_length); 
	  j = ssos[i][k].curve_num + 1 - ssos[i][k].initial;
	  while (j <= (ssos[i][k].curve_num + ssos[i][k].param_length)) {
	    adj_patch = sop_tpst[overlap_pairs[i].ptches.patchNum1][0].AdjInfo[j];
	    deg = sop_tpst[overlap_pairs[i].ptches.patchNum1][0].XYZ_Splines[j].Deg;
	    partn_num = GetAdjacentPartn(sop_tpst, adj_patch,
					 num_partn_second[adj_patch],
					 &sop_tpst[overlap_pairs[i].ptches.patchNum1][0].XYZ_Splines[j].CntrlPts[0], 
					 &sop_tpst[overlap_pairs[i].ptches.patchNum1][0].XYZ_Splines[j].CntrlPts[deg],
					 -1, &curve_num, &dir,
					 overlap_pairs[i].ptches.patchNum1);
	    sop_tpst[adj_patch][partn_num].AdjInfo[curve_num] = ssos[i][k].adj_patch;
	    second_same_solid[adj_patch][0][partn_num][curve_num] = 0;
	    sop_tpst[overlap_pairs[i].ptches.patchNum1][0].AdjInfo[j] = ssos[i][k].adj_patch;
	    second_same_solid[overlap_pairs[i].ptches.patchNum1][0][0][j] = 0;
	    j += 2;
	  }
	}
      }
    }

    /* Now in PartitionSplinegon I fill in ngh_crv array. I go through the array */
    /* and assign the appropriate values for AdjInfo and same_solid */
    for (i=0; i<NumNghCrvs; i++) {
      partn_num = GetAdjacentPartn(sop_tpst, ngh_crv[i].patchNo,
				   num_partn_second[ngh_crv[i].patchNo],
				   &ngh_crv[i].spt, &ngh_crv[i].ept, -1,
				   &curve_num, &dir, -1);
      sop_tpst[ngh_crv[i].patchNo][partn_num].AdjInfo[curve_num] =
	ngh_crv[i].adj_pnum; 
      second_same_solid[ngh_crv[i].patchNo][0][partn_num][curve_num] = 0;
    }

    for (i=0; i<sts_arr_len; i++) {
      if (sts_array[i].which_solid == 1) {
	partn_num = GetAdjacentPartn(fop_tpst, sts_array[i].patch_num,
				     num_partn_first[sts_array[i].patch_num],
				     sts_array[i].st_pt, sts_array[i].end_pt,
				     -1, &curve_num, &prev_dir,
				     sts_array[i].other_patch_num);
	if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	    printf("Its happening here: 1\n");
	  }

	same_solid = &first_same_solid[sts_array[i].patch_num][0][partn_num][0];
	if (fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves >=
	    MAX_SPLINES_PER_TRIM_REGION-sts_array[i].param_length) {
	  printf("Too many curves in one region. Quitting ...\n");
	  exit(-1);
	}
	for (j=fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves-1; 
	     j>=curve_num; j--) {
	 copy_uv_spline(&fop_tpst[sts_array[i].patch_num][partn_num].UV_Splines[j], 
			&fop_tpst[sts_array[i].patch_num][partn_num].
			UV_Splines[j+sts_array[i].param_length]);
	 if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	   printf("Its happening here: 21\n");
	 }
	 copy_xyz_spline(&fop_tpst[sts_array[i].patch_num][partn_num].XYZ_Splines[j], 
			 &fop_tpst[sts_array[i].patch_num][partn_num].
			 XYZ_Splines[j+sts_array[i].param_length]);
	 if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	   printf("Its happening here: 22\n");
	 }
	 fop_tpst[sts_array[i].patch_num][partn_num].
	   AdjInfo[j+sts_array[i].param_length] = 
	     fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j];
	 same_solid[j+sts_array[i].param_length] = same_solid[j];
        }

	if (prev_dir) {
	  k = 0;
	  while (k <= (sts_array[i].param_length-k-1)) {
	    if (k == sts_array[i].param_length-k-1)
	      sts_array[i].param[k] = 1.0 - sts_array[i].param[k];
	    else {
	      tmp_doub = sts_array[i].param[k];
	      sts_array[i].param[k] = 1.0 -
		sts_array[i].param[sts_array[i].param_length-k-1];
	      sts_array[i].param[sts_array[i].param_length-k-1] = 1.0 -
		tmp_doub; 
	    }
	    k++;
	  }
	}
	SubdivideTrimCurve(&fop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[curve_num],
			   &fop_tpst[sts_array[i].patch_num][partn_num].
			   XYZ_Splines[curve_num],
			   sts_array[i].param_length, sts_array[i].param,
			   &fop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[curve_num],
			   &fop_tpst[sts_array[i].patch_num][partn_num].
			   XYZ_Splines[curve_num]);
	if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	    printf("Its happening here: 3\n");
	  }

	fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves +=
	  sts_array[i].param_length;
	for (j=curve_num+1; j<=curve_num+sts_array[i].param_length;
	     j++) {
	  fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j] =
	    fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[curve_num];
	  same_solid[j] = same_solid[curve_num];
	}

	/* Now I have to split some trim curves if the same_solid flag */
	/* is 0. This is because the curve being cut on other side */
	/* belongs to the other solid, so the curve must be an */
	/* intersection curve. If I split an intersection curve, I */
	/* must do it on both sides te get proper adjacencies. */
/*
	for (j=0; j<NumOvrPatches; j++)
	  if (overlap_pairs[j].ptches.patchNum1 == sts_array[i].other_patch_num)
	    break;
*/
	if (!sts_array[i].same_solid && sts_array[i].do_twice) {
	  partn_num = GetAdjacentPartn(fop_tpst, sts_array[i].patch_num,
				       num_partn_first[sts_array[i].patch_num],
				       sts_array[i].st_pt, sts_array[i].end_pt,
				       partn_num, &curve_num, &dir,
				       sts_array[i].other_patch_num);
	  if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	    printf("Its happening here: 4\n");
	  }

	  same_solid = &first_same_solid[sts_array[i].patch_num][0][partn_num][0];
	  if (fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves >=
	      MAX_SPLINES_PER_TRIM_REGION-sts_array[i].param_length) {
	    printf("Too many curves in one region. Quitting ...\n");
	    exit(-1);
	  }
	  for (j=fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves-1; 
	       j>=curve_num; j--) {
	    copy_uv_spline(&fop_tpst[sts_array[i].patch_num][partn_num].UV_Splines[j], 
			   &fop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[j+sts_array[i].param_length]);
	    copy_xyz_spline(&fop_tpst[sts_array[i].patch_num][partn_num].XYZ_Splines[j], 
			    &fop_tpst[sts_array[i].patch_num][partn_num].
			    XYZ_Splines[j+sts_array[i].param_length]);
	    fop_tpst[sts_array[i].patch_num][partn_num].
	      AdjInfo[j+sts_array[i].param_length] = 
		fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j];
	    same_solid[j+sts_array[i].param_length] = same_solid[j];
	  }
	  if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	    printf("Its happening here: 5\n");
	  }

	  if ((dir && !prev_dir) || (prev_dir && !dir)) {
	    k = 0;
	    while (k <= (sts_array[i].param_length-k-1)) {
	      if (k == sts_array[i].param_length-k-1)
		sts_array[i].param[k] = 1.0 - sts_array[i].param[k];
	      else {
		tmp_doub = sts_array[i].param[k];
		sts_array[i].param[k] = 1.0 -
		  sts_array[i].param[sts_array[i].param_length-k-1];
		sts_array[i].param[sts_array[i].param_length-k-1] = 1.0 -
		  tmp_doub; 
	      }
	      k++;
	    }
	  }
	  SubdivideTrimCurve(&fop_tpst[sts_array[i].patch_num][partn_num].
			     UV_Splines[curve_num],
			     &fop_tpst[sts_array[i].patch_num][partn_num].
			     XYZ_Splines[curve_num],
			     sts_array[i].param_length, sts_array[i].param,
			     &fop_tpst[sts_array[i].patch_num][partn_num].
			     UV_Splines[curve_num],
			     &fop_tpst[sts_array[i].patch_num][partn_num].
			     XYZ_Splines[curve_num]);
	  if (FP_EQ_EPS(fop_tpst[sts_array[i].patch_num][0].XYZ_Splines[0].CntrlPts[0].dCoord[3], 0.0, 1e-3)) {
	    printf("Its happening here: 6\n");
	  }

	  fop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves +=
	    sts_array[i].param_length;
	  for (j=curve_num+1; j<=curve_num+sts_array[i].param_length;
	       j++) {
	    fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j] =
	      fop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[curve_num];
	    same_solid[j] = same_solid[curve_num];
	  }
	}
      } else {
	partn_num = GetAdjacentPartn(sop_tpst, sts_array[i].patch_num,
				     num_partn_second[sts_array[i].patch_num],
				     sts_array[i].st_pt, sts_array[i].end_pt,
				     -1, &curve_num, &prev_dir,
				     sts_array[i].other_patch_num);
	same_solid = &second_same_solid[sts_array[i].patch_num][0][partn_num][0];
	if (sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves >=
	    MAX_SPLINES_PER_TRIM_REGION-sts_array[i].param_length) {
	  printf("Too many curves in one region. Quitting ...\n");
	  exit(-1);
	}
	for (j=sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves-1; 
	     j>=curve_num; j--) {
	 copy_uv_spline(&sop_tpst[sts_array[i].patch_num][partn_num].UV_Splines[j], 
			&sop_tpst[sts_array[i].patch_num][partn_num].
			UV_Splines[j+sts_array[i].param_length]);
	 copy_xyz_spline(&sop_tpst[sts_array[i].patch_num][partn_num].XYZ_Splines[j], 
			 &sop_tpst[sts_array[i].patch_num][partn_num].
			 XYZ_Splines[j+sts_array[i].param_length]);
	 sop_tpst[sts_array[i].patch_num][partn_num].
	   AdjInfo[j+sts_array[i].param_length] = 
	     sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j];
	 same_solid[j+sts_array[i].param_length] = same_solid[j];
        }
	if (prev_dir) {
	  k = 0;
	  while (k <= (sts_array[i].param_length-k-1)) {
	    if (k == sts_array[i].param_length-k-1)
	      sts_array[i].param[k] = 1.0 - sts_array[i].param[k];
	    else {
	      tmp_doub = sts_array[i].param[k];
	      sts_array[i].param[k] = 1.0 -
		sts_array[i].param[sts_array[i].param_length-k-1];
	      sts_array[i].param[sts_array[i].param_length-k-1] = 1.0 -
		tmp_doub; 
	    }
	    k++;
	  }
	}
	SubdivideTrimCurve(&sop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[curve_num],
			   &sop_tpst[sts_array[i].patch_num][partn_num].
			   XYZ_Splines[curve_num],
			   sts_array[i].param_length, sts_array[i].param,
			   &sop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[curve_num],
			   &sop_tpst[sts_array[i].patch_num][partn_num].
			   XYZ_Splines[curve_num]);
	sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves +=
	  sts_array[i].param_length;
	for (j=curve_num+1; j<=curve_num+sts_array[i].param_length;
	     j++) {
	  sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j] =
	    sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[curve_num];
	  same_solid[j] = same_solid[curve_num];
	}

	/* Now I have to split some trim curves if the same_solid flag */
	/* is 0. This is because the curve being cut on other side */
	/* belongs to the other solid, so the curve must be an */
	/* intersection curve. If I split an intersection curve, I */
	/* must do it on both sides te get proper adjacencies. */
/*
	for (j=0; j<NumOvrPatches; j++)
	  if (overlap_pairs[j].ptches.patchNum0 == sts_array[i].other_patch_num)
	    break;
*/
	if (!sts_array[i].same_solid && sts_array[i].do_twice) {
	  partn_num = GetAdjacentPartn(sop_tpst, sts_array[i].patch_num,
				       num_partn_second[sts_array[i].patch_num],
				       sts_array[i].st_pt, sts_array[i].end_pt,
				       partn_num, &curve_num, &dir,
				       sts_array[i].other_patch_num);
	  same_solid = &second_same_solid[sts_array[i].patch_num][0][partn_num][0];
	  if (sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves >=
	      MAX_SPLINES_PER_TRIM_REGION-sts_array[i].param_length) {
	    printf("Too many curves in one region. Quitting ...\n");
	    exit(-1);
	  }
	  for (j=sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves-1; 
	       j>=curve_num; j--) {
	    copy_uv_spline(&sop_tpst[sts_array[i].patch_num][partn_num].UV_Splines[j], 
			   &sop_tpst[sts_array[i].patch_num][partn_num].
			   UV_Splines[j+sts_array[i].param_length]);
	    copy_xyz_spline(&sop_tpst[sts_array[i].patch_num][partn_num].XYZ_Splines[j], 
			    &sop_tpst[sts_array[i].patch_num][partn_num].
			    XYZ_Splines[j+sts_array[i].param_length]);
	    sop_tpst[sts_array[i].patch_num][partn_num].
	      AdjInfo[j+sts_array[i].param_length] = 
		sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j];
	    same_solid[j+sts_array[i].param_length] = same_solid[j];
	  }
	  if ((dir && !prev_dir) || (prev_dir && !dir)) {
	    k = 0;
	    while (k <= (sts_array[i].param_length-k-1)) {
	      if (k == sts_array[i].param_length-k-1)
		sts_array[i].param[k] = 1.0 - sts_array[i].param[k];
	      else {
		tmp_doub = sts_array[i].param[k];
		sts_array[i].param[k] = 1.0 -
		  sts_array[i].param[sts_array[i].param_length-k-1];
		sts_array[i].param[sts_array[i].param_length-k-1] = 1.0 -
		  tmp_doub; 
	      }
	      k++;
	    }
	  }
	  SubdivideTrimCurve(&sop_tpst[sts_array[i].patch_num][partn_num].
			     UV_Splines[curve_num],
			     &sop_tpst[sts_array[i].patch_num][partn_num].
			     XYZ_Splines[curve_num],
			     sts_array[i].param_length, sts_array[i].param,
			     &sop_tpst[sts_array[i].patch_num][partn_num].
			     UV_Splines[curve_num],
			     &sop_tpst[sts_array[i].patch_num][partn_num].
			     XYZ_Splines[curve_num]);
	  sop_tpst[sts_array[i].patch_num][partn_num].NumOfCurves +=
	    sts_array[i].param_length;
	  for (j=curve_num+1; j<=curve_num+sts_array[i].param_length;
	       j++) {
	    sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[j] =
	      sop_tpst[sts_array[i].patch_num][partn_num].AdjInfo[curve_num];
	    same_solid[j] = same_solid[curve_num];
	  }
	}
      }
    }

#ifdef __PERFORM_TIME
    Time1 = GetTimer;
#endif
    ObtainNewGlobalGraph(fop_tpst, first_set_patches, num_partn_first,
			 fn_info, 1, first_sol_comp, &f_num_comp, 
			 &global_graph_f);

    /* I am finding out (patch,partn) that has component num 0 */
    for (i=0; i<first_set_patches; i++) {
      for (k=0; k<NumOvrPatches; k++)
	if (overlap_pairs[k].ptches.patchNum0 == i) break;
      if (k != NumOvrPatches) continue;
      else break;
    }
/*
      for (j=0; j<num_partn_first[i]; j++)
	if (first_sol_comp[i*MAX_PARTNS_PER_PATCH+j] == 0)
	  break;
      if (j != num_partn_first[i]) break;
    }
    DetermineXsectionsforSolid(&fop_tpst[i][j], &(*first_solid)[i].tp, 
			       *second_solid, second_set_patches,
			       global_graph_f, num_comp, fNumInt); 
*/
#ifdef _CALC_IMPLICIT
    DetermineXsectionsforSolid(&fop_tpst[i][0], &(*first_solid)[i].tp, 
			       *second_solid, second_set_patches,
			       sec_sol_implicit, imp_scale, 
			       global_graph_f, f_num_comp, fNumInt,
			       first_sol_comp[i*MAX_PARTNS_PER_PATCH+0]); 
#else
    DetermineXsectionsforSolid(&fop_tpst[i][0], &(*first_solid)[i].tp, 
			       *second_solid, second_set_patches,
			       global_graph_f, f_num_comp, fNumInt,
			       first_sol_comp[i*MAX_PARTNS_PER_PATCH+0]); 
#endif

    /* This call goes through sop_tpst and gives a unique global */
    /* component number (ranging from 0 .. num_comp-1) for each new */
    /* region */
    /* ALLOCN(global_graph_s, neighbour_t, num_comp); */
    ObtainNewGlobalGraph(sop_tpst, second_set_patches, num_partn_second,
			 sn_info, 0, second_sol_comp, &s_num_comp, 
			 &global_graph_s);

    /* I am finding out (patch,partn) that has component num 0 */
    for (i=0; i<second_set_patches; i++) {
      for (k=0; k<NumOvrPatches; k++)
	if (overlap_pairs[k].ptches.patchNum1 == i) break;
      if (k != NumOvrPatches) continue;
      else break;
    }
/*
      for (j=0; j<num_partn_second[i]; j++)
	if (second_sol_comp[i*MAX_PARTNS_PER_PATCH+j] == 0)
	  break;
      if (j != num_partn_first[i]) break;
    }
    DetermineXsectionsforSolid(&sop_tpst[i][j], &(*second_solid)[i].tp, 
			       *first_solid, first_set_patches,
			       global_graph_s, num_comp, sNumInt); 
*/
#ifdef _CALC_IMPLICIT
    DetermineXsectionsforSolid(&sop_tpst[i][0], &(*second_solid)[i].tp, 
			       *first_solid, first_set_patches,
			       first_sol_implicit, imp_scale, 
			       global_graph_s, s_num_comp, sNumInt,
			       second_sol_comp[i*MAX_PARTNS_PER_PATCH+0]); 
#else
    DetermineXsectionsforSolid(&sop_tpst[i][0], &(*second_solid)[i].tp, 
			       *first_solid, first_set_patches,
			       global_graph_s, s_num_comp, sNumInt,
			       second_sol_comp[i*MAX_PARTNS_PER_PATCH+0]); 
#endif

    ObtainFinalPiecesofSolid(first_solid, fop_tpst, first_set_patches,
			     num_partn_first, first_sol_comp, fNumInt,
			     first_orig_region_num, operation, 1);

    for (i=0; i<first_set_patches; i++) 
      if ((*first_solid)[i].csg_status != OUT_CSG) {
	patchNumber[tmp_ppatches] = i;
	first_or_second[tmp_ppatches] = 1;
	tmp_ppatches++;
      }

    ObtainFinalPiecesofSolid(second_solid, sop_tpst, second_set_patches,
			     num_partn_second, second_sol_comp, sNumInt,
			     sec_orig_region_num, operation, 0);

    for (i=0; i<second_set_patches; i++) 
      if ((*second_solid)[i].csg_status != OUT_CSG) {
	patchNumber[tmp_ppatches] = i;
	first_or_second[tmp_ppatches] = 2;
	tmp_ppatches++;
      }
  }
#ifdef __PERFORM_TIME
  Time2 = GetTimer;
  ClassifyTime += TIMEDIFF(Time2, Time1);
#endif

  if (!srep) {
    if (operation != PLANE_CUT) {
      for (i=0; i<second_set_patches; i++) {
#ifdef _PRINT_DOTS
	printf(". ");
	fflush(stdout);
#endif
	if (planar_ovrlap_flag2[i] == 1) {
	  if ((*second_solid)[i].NumTrimPortions == 0 || operation != DOUNION)
	    (*second_solid)[i].csg_status = OUT_CSG;
	} else {
	  if (init_ssolid_curves[i] > ss_nc[i]) {
	    printf("Something wrong in findXsectionCurve\n");
	    return;
	  }
	  SplitIntoNewRegions(&((*second_solid)[i]), ss_nc[i],
			      init_ssolid_curves[i], tmps1, 
			      first_set_patches, ss_pw_curves[i], 
			      operation, to_keep, cutting_plane, 0);
	}
	if ((*second_solid)[i].csg_status != OUT_CSG) {
	  patchNumber[tmp_ppatches] = i;
	  first_or_second[tmp_ppatches] = 2;
	  tmp_ppatches++;
	}
      }
    } else {
      (*second_solid)[0].NumTrimPortions = ss_nc[0];
      (*second_solid)[0].csg_status = IN_CSG;
      for (i=0; i<(*second_solid)[0].NumTrimPortions; i++) {
	npts = 0;
	if (!(FP_EQ(ss_pw_curves[0][i].StEndPoints[0].u,
		    ss_pw_curves[0][i].StEndPoints[1].u) &&
	      FP_EQ(ss_pw_curves[0][i].StEndPoints[0].v,
		    ss_pw_curves[0][i].StEndPoints[1].v))) {
	  printf("The two endpoints are (%lf,%lf) and (%lf,%lf)\n",
		 ss_pw_curves[0][i].StEndPoints[0].u,
		 ss_pw_curves[0][i].StEndPoints[0].v,
		 ss_pw_curves[0][i].StEndPoints[1].u,
		 ss_pw_curves[0][i].StEndPoints[1].v);
	  printf("PLANE_CUT_OP: No loop in plane patch\n");
	  /* exit(-1); changed on 1/4/95 */
	}
	tmp_ptr1 = ss_pw_curves[0][i].planeCurve;
	while (tmp_ptr1 != NULL) {
	  (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts][0] =
	    tmp_ptr1->p2D.u;
	  (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts++][1] =
	    tmp_ptr1->p2D.v;
	  tmp_ptr1 = tmp_ptr1->plp2DNext;
	}
	(*second_solid)[0].TrimPortion[i].PiecePortion.NumOfVertices = npts - 1;
	
	/* is_curve_clockwise actually SHOULD be is_curve_anticlockwise */
	if (!is_curve_anticlockwise(
	 (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex, npts-1)) {
	  if (ODD(npts))
	    half_npts = (npts - 1)/2;
	  else half_npts = npts/2;
	  for (j=0; j<half_npts; j++) {
	    tmp2D.u = (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[j][0];
	    tmp2D.v = (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[j][1];
	    (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[j][0] = 
	      (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts-j-1][0];
	    (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[j][1] = 
	      (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts-j-1][1];
	    (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts-j-1][0] =
	      tmp2D.u;
	    (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex[npts-j-1][1] =
	      tmp2D.v;
	  }
	}
	find_point_inside_polygon(npts-1,
	       (*second_solid)[0].TrimPortion[i].PiecePortion.poly_vertex,
				  &inside_pt);
	(*second_solid)[0].TrimPortion[i].PiecePortion.ptInsideRegion.u = inside_pt.x;
	(*second_solid)[0].TrimPortion[i].PiecePortion.ptInsideRegion.v = inside_pt.y;
      }
      patchNumber[tmp_ppatches] = 0;
      first_or_second[tmp_ppatches] = 2;
      tmp_ppatches++;
    }
  }
    
  ALLOCN(tmp_psolid, TRIM_PATCH_INFO, tmp_ppatches);
  for (i=0; i<tmp_ppatches; i++) {
    if (first_or_second[i] == 1) {
      copy_trim_patch((*first_solid)+patchNumber[i], tmp_psolid + i); 
      if (tmp_psolid[i].SplineReprn)
	for (j=0; j<tmp_psolid[i].NumTrimPortions; j++) {
	  tmp_tpst = &tmp_psolid[i].TrimPortion[j].SplinePortion;
	  for (q=1,k=0; k<tmp_tpst->NumOfCurves; k++) {
	    POINT2D pt;

	    ic = &tmp_tpst->UV_Splines[k];
	    num_pts = NumPointsOnCurve(ic, 0.03);
	    ss = 1.0 / num_pts;
	    for (r=0; r<num_pts; r++) {
	      Eval2DCurve(ic, r*ss, &pt);
	      seg2[q].v0.x = pt.u;
	      seg2[q].v0.y = pt.v;
	      if (q != 1) {
		seg2[q-1].v1.x = pt.u;
		seg2[q-1].v1.y = pt.v;
	      }
	      seg2[q].is_inserted = FALSE;
	      q++;
	    }
	  }
	  seg2[q-1].v1.x = seg2[1].v0.x;
	  seg2[q-1].v1.y = seg2[1].v0.y;
	  if (!is_simple_polygon2(seg2, q-1))
	    fprintf(stderr, "perf_csg: polygon not simple\n"); 

	  for (k=0; k<tmp_tpst->NumOfCurves; k++) {
	    p = first_same_solid[patchNumber[i]][0]
	      [first_orig_region_num[patchNumber[i]][j]][k];
	    if (p) {
	      if ((tmp_tpst->AdjInfo[k] = ReturnIndexInArray(patchNumber,
		 tmp_ppatches,tmp_tpst->AdjInfo[k],first_or_second,1))
		  == -1) {
		printf("perform_CSG: Could not find proper final adjacency\n");
		exit(-1);
	      }
	    } else {
	      if ((tmp_tpst->AdjInfo[k] = ReturnIndexInArray(patchNumber,
		 tmp_ppatches,tmp_tpst->AdjInfo[k],first_or_second,2))
		  == -1) {
		printf("perform_CSG: Could not find proper final adjacency\n");
		exit(-1);
	      }
	    }
	  }
	}
    }
    else if (first_or_second[i] == 2) {
      if (operation == DIFFERENCE) {
	copy_flip_trim_patch((*second_solid)+patchNumber[i],
			     tmp_psolid + i); 
	if (tmp_psolid[i].SplineReprn)
	  for (j=0; j<tmp_psolid[i].NumTrimPortions; j++) {
	    tmp_tpst = &tmp_psolid[i].TrimPortion[j].SplinePortion;
	    k = 0;
	    while (k < tmp_tpst->NumOfCurves-k-1) {
	      p = second_same_solid[patchNumber[i]][0]
		                   [sec_orig_region_num[patchNumber[i]][j]][k];
	      second_same_solid[patchNumber[i]][0]
		               [sec_orig_region_num[patchNumber[i]][j]][k] = 
	      second_same_solid[patchNumber[i]][0]
		               [sec_orig_region_num[patchNumber[i]][j]]
			       [tmp_tpst->NumOfCurves-k-1];
	      second_same_solid[patchNumber[i]][0]
		               [sec_orig_region_num[patchNumber[i]][j]]
			       [tmp_tpst->NumOfCurves-k-1] = p;
	      k++;
	    }
	  }
      } else
	copy_trim_patch((*second_solid)+patchNumber[i], tmp_psolid + i); 
      if (tmp_psolid[i].SplineReprn)
	for (j=0; j<tmp_psolid[i].NumTrimPortions; j++) {
	  tmp_tpst = &tmp_psolid[i].TrimPortion[j].SplinePortion;
	  for (q=1,k=0; k<tmp_tpst->NumOfCurves; k++) {
	    POINT2D pt;

	    ic = &tmp_tpst->UV_Splines[k];
	    num_pts = NumPointsOnCurve(ic, 0.03);
	    ss = 1.0 / num_pts;
	    for (r=0; r<num_pts; r++) {
	      Eval2DCurve(ic, r*ss, &pt);
	      seg2[q].v0.x = pt.u;
	      seg2[q].v0.y = pt.v;
	      if (q != 1) {
		seg2[q-1].v1.x = pt.u;
		seg2[q-1].v1.y = pt.v;
	      }
	      seg2[q].is_inserted = FALSE;
	      q++;
	    }
	  }
	  seg2[q-1].v1.x = seg2[1].v0.x;
	  seg2[q-1].v1.y = seg2[1].v0.y;
	  if (!is_simple_polygon2(seg2, q-1))
	    fprintf(stderr, "perf_csg: polygon not simple\n"); 

	  for (k=0; k<tmp_tpst->NumOfCurves; k++) {
	    p = second_same_solid[patchNumber[i]][0]
	      [sec_orig_region_num[patchNumber[i]][j]][k];
	    if (p) {
	      if ((tmp_tpst->AdjInfo[k] = ReturnIndexInArray(patchNumber,
		       tmp_ppatches,tmp_tpst->AdjInfo[k],first_or_second,2))
		  == -1) {
		printf("perform_CSG: Could not find proper final adjacency\n");
		exit(-1);
	      }
	    } else {
	      if ((tmp_tpst->AdjInfo[k] = ReturnIndexInArray(patchNumber,
		       tmp_ppatches,tmp_tpst->AdjInfo[k],first_or_second,1))
		  == -1) {
		printf("perform_CSG: Could not find proper final adjacency\n");
		exit(-1);
	      }
	    }
	  }
	}
    }
    else {
      printf("Uh-oh! first_or_second isn't 1 or 2!\n");
    }
  }

  parent_patches = 0;
  for (i=0; i<tmp_ppatches; i++)
    for (j=0; j<tmp_psolid[i].NumTrimPortions; j++)
      final_patch_number[i][j] = parent_patches++;
  
  ALLOCN((*parent_solid), TRIM_PATCH_INFO, parent_patches);
  r = 0;
  for (p=0; p<tmp_ppatches; p++) {
    /* I am commenting out the if statement because otherwise */
    /* adjacency info is hard to get */
    /* if (tmp_psolid[p].NumTrimPortions > 1) { */
    for (q=0; q<tmp_psolid[p].NumTrimPortions; q++) {
      subdivide_bezier_patch_using_trim_curve(tmp_psolid+p, q,
					      (*parent_solid) + r);
      for (i=0;
	   i<tmp_psolid[p].TrimPortion[q].SplinePortion.NumOfCurves;
	   i++) {
	m = tmp_psolid[p].TrimPortion[q].SplinePortion.XYZ_Splines[i].Deg;
	fpt = tmp_psolid[p].TrimPortion[q].SplinePortion.
	  XYZ_Splines[i].CntrlPts[0];
	ept = tmp_psolid[p].TrimPortion[q].SplinePortion.
	  XYZ_Splines[i].CntrlPts[m];
	j = tmp_psolid[p].TrimPortion[q].SplinePortion.AdjInfo[i];
	if (j != p)
	  k = GetAdjacentTrimPortion(&tmp_psolid[j], &fpt, &ept, &l, -1, &dir);
	else
	  k = GetAdjacentTrimPortion(&tmp_psolid[j], &fpt, &ept, &l, q, &dir);
	if ((*parent_solid)[r].TrimPortion[0].SplinePortion.AdjInfo[i]
	    == -1) {
	  (*parent_solid)[r].TrimPortion[0].SplinePortion.AdjInfo[i]
	    = final_patch_number[j][k];
/*
	  (*parent_solid)[final_patch_number[j][k]].TrimPortion[0].
	    SplinePortion.AdjInfo[l] = r;
*/
	}
      }
      r++;
    }
/*
   } else {
     copy_trim_patch(tmp_psolid+p, (*parent_solid) + r);
     r++;
   }
*/
  }

  /* START HERE: I am almost done. Just go through the rest of the */
  /* routine for correct use of structures */
#ifdef __GL
  draw_the_patches("", *parent_solid, parent_patches, NULL, 0, 0, 0, "Result");
#endif

  /* freeing the static global pointers */
  myfree(fs_nc);
  myfree(ss_nc);
  FREEMAT(spec_crvs_for_ovr_loops_first, MAX_PATCHES_PER_SOLID,
	  MAX_SPLINES_PER_TRIM_REGION);
  FREEMAT(spec_crvs_for_ovr_loops_second, MAX_PATCHES_PER_SOLID,
	  MAX_SPLINES_PER_TRIM_REGION); 
  if (!frep) {
    for (i=0; i<first_set_patches; i++) {
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	free_dom_int_curve(&(fs_pw_curves[i][j]));
      myfree(&fs_pw_curves[i][0]);
    }
    for (i=0; i<second_set_patches; i++) {
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	free_dom_int_curve(&(ss_pw_curves[i][j]));
      myfree(&ss_pw_curves[i][0]);
    }
    myfree(fs_pw_curves);
    myfree(ss_pw_curves);
    myfree(init_fsolid_curves);
    myfree(init_ssolid_curves);
    for (i=0; i<first_set_patches; i++)
      free_trim_patch(&(tmps1[i]));
    for (i=0; i<second_set_patches; i++)
      free_trim_patch(&(tmps2[i]));
    myfree(tmps1);
    myfree(tmps2);
  } else {
    for (i=0; i<first_set_patches; i++) {
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	free_merged_curve(&(fs_spl_curves[i][j]));
      myfree(fs_spl_curves[i]);
      fs_spl_curves[i] = NULL;
    }
    for (i=0; i<second_set_patches; i++) {
      for (j=0; j<MAX_TRIM_CURVES_PER_PATCH; j++)
	free_merged_curve(&(ss_spl_curves[i][j]));
      myfree(ss_spl_curves[i]);
      ss_spl_curves[i] = NULL;
    }
    myfree(fs_spl_curves);
    myfree(ss_spl_curves);
    FREEMAT(fop_tpst, first_set_patches, MAX_PARTNS_PER_PATCH);
    FREEMAT(sop_tpst, second_set_patches, MAX_PARTNS_PER_PATCH);
    myfree(global_graph_f); myfree(global_graph_s);
  }

  for (i=0; i<tmp_ppatches; i++)
    free_trim_patch(&(tmp_psolid[i]));
  myfree(patchNumber);
  myfree(first_or_second);
  myfree(tmp_psolid);
  return parent_patches;

#endif /* _LOTS_OF_HOLES */
}

void RefineEndPointsFurther(tp0, tp1, ccmp)
TENSOR_PATCH tp0, tp1;
INT_CURVE_STRUCT *ccmp;
{
  double *pow_coord, **dirMat;
  double *pow_coord2, **dirMat2;
  int nIter, i, j;
  double fret;
  LIST_POINT4D *tmp_ptr;

  pow_tp0 = tp0;
  pow_tp1 = tp1;
  ALLOCN(pow_coord, double, 4);
  ALLOCN(pow_coord2, double, 4);
  ALLOCMAT(dirMat, 4, 4, double);
  ALLOCMAT(dirMat2, 4, 4, double);
  for (i=0; i<4; i++) {
    pow_coord[i] = (*ccmp).endPoints[0].dCoord[i];
    pow_coord2[i] = (*ccmp).endPoints[1].dCoord[i];
    for (j=0; j<4; j++) {
      if (i == j) {
	dirMat[i][j] = 1.0;
	dirMat2[i][j] = 1.0;
      } else {
	dirMat[i][j] = 0.0;
	dirMat2[i][j] = 0.0;
      }
    }
  }
  powell(pow_coord, dirMat, 4, POW_TOL, &nIter, &fret, RefinePoint);
  if (fret > 1e-3) 
    printf("Intersection point computed in prop_int is incorrect:: %lf\n", fret);
  powell(pow_coord2, dirMat2, 4, POW_TOL, &nIter, &fret, RefinePoint);
  if (fret > 1e-3) 
    printf("Intersection point computed in prop_int is incorrect:: %lf\n", fret);
  for (i=0; i<4; i++) {
    (*ccmp).endPoints[0].dCoord[i] = pow_coord[i];
    (*ccmp).endPoints[1].dCoord[i] = pow_coord2[i];
  }

  tmp_ptr = (*ccmp).Curve;
  tmp_ptr->p4D = (*ccmp).endPoints[1];
  if (tmp_ptr->plp4DNext == NULL)
    printf("RefineEndPointsFurther:Got a chain with only 1 point\n");
  while (tmp_ptr->plp4DNext != NULL)
    tmp_ptr = tmp_ptr->plp4DNext;
  tmp_ptr->p4D = (*ccmp).endPoints[0];

  myfree(pow_coord); myfree(pow_coord2);
  FREEMAT(dirMat, 4, 4);
  FREEMAT(dirMat2, 4, 4);
}

usefont(fontname,fontsize)
char fontname[];
float fontsize;
/*
 * GET READY TO USE A FONT
 * Dinesh Manocha  2-20-89
 */
{
long  *font1pt,*fsized;
/*-----------------------------------------------------------------*
font1pt = (long *)fmfindfont(fontname);
if(!font1pt) 
  { printf("Can't Find Font '%s'; Substituting Times-Roman\n",fontname);
    font1pt = (long *)fmfindfont("Times-Roman");
    if(!font1pt) 
      { printf("Can't Find Times-Roman Font. System Problem.\n");
        exit(-1);
      }
  }
fsized = (long *)fmscalefont(font1pt,fontsize);
fmsetfont(fsized);
*-----------------------------------------------------------------*/
}

void FillXscArray(fptr, lptr, pnum, cnum, first_sol, bndry_info, bi_size)
INTERSECTION_CURVE *fptr, *lptr;
int pnum, cnum, first_sol, bi_size;
ANOTHER_ARBIT_STRUCT *bndry_info;
{
  POINT3D first, last, curr_pt;
  int deg, i;
  double wcrd;
  int first_done, last_done;

  first_done = last_done = 0;
  wcrd = 1.0 / fptr->xyz_spline.CntrlPts[0].dCoord[3];
  first.dCoord[0] = fptr->xyz_spline.CntrlPts[0].dCoord[0] * wcrd;
  first.dCoord[1] = fptr->xyz_spline.CntrlPts[0].dCoord[1] * wcrd;
  first.dCoord[2] = fptr->xyz_spline.CntrlPts[0].dCoord[2] * wcrd; 

  deg = lptr->xyz_spline.Deg;
  wcrd = 1.0 / lptr->xyz_spline.CntrlPts[deg].dCoord[3];
  last.dCoord[0] = lptr->xyz_spline.CntrlPts[deg].dCoord[0] * wcrd;
  last.dCoord[1] = lptr->xyz_spline.CntrlPts[deg].dCoord[1] * wcrd;
  last.dCoord[2] = lptr->xyz_spline.CntrlPts[deg].dCoord[2] * wcrd;

  for (i=0; i<bi_size; i++) {
    if (first_done && last_done) break;

    wcrd = 1.0 / bndry_info[i].p4d.dCoord[3];
    curr_pt.dCoord[0] = bndry_info[i].p4d.dCoord[0] * wcrd;
    curr_pt.dCoord[1] = bndry_info[i].p4d.dCoord[1] * wcrd;
    curr_pt.dCoord[2] = bndry_info[i].p4d.dCoord[2] * wcrd;
    
    if (!first_done && EQUAL3D_ABS(first, curr_pt, 0.003)) {
      first_done = 1;
      if (first_sol)
	fXsc[pnum][cnum][0][0] = bndry_info[i].bndry_crv_num;
      else
	sXsc[pnum][cnum][0][0] = bndry_info[i].bndry_crv_num;
    }
    if (!last_done && EQUAL3D_ABS(last, curr_pt, 0.003)) {
      last_done = 1;
      if (first_sol)
	fXsc[pnum][cnum][1][0] = bndry_info[i].bndry_crv_num;
      else
	sXsc[pnum][cnum][1][0] = bndry_info[i].bndry_crv_num;
    }
  }
}

#ifdef _CALC_IMPLICIT
int findXsectionCurve(first_solid, first_set_patches, first_imp, second_solid,
		      second_set_patches, second_imp, imp_scale, frep, op, step_size,
		      ovr_crv_loops_f, ovr_crv_loops_s)
TRIM_PATCH_INFO **first_solid, **second_solid;
double imp_scale;
NDETERMINANT *first_imp, *second_imp; /* These two are arrays of */
				      /* implicit reprns of all */
				      /* patches in both solids */
int first_set_patches, second_set_patches, op, frep; /* frep is for */
						     /* representation */
						     /* 0 - piecewise */
						     /* 1 - splines */
double step_size;
short **ovr_crv_loops_f, **ovr_crv_loops_s;
{ 

  TENSOR_PATCH	   tpPatch0[MAX_PATCH_NUM], tpPatch1[MAX_PATCH_NUM];
  TENSOR_PATCH     ptch0, ptch1;
  double	   local_scale, temporary_store, dot_prod;
  char             str[15];
  BOUNDINGBOX	   boxP0[MAX_PATCH_NUM], boxP1[MAX_PATCH_NUM];
  BOUNDINGBOX	   box0, box1;
  BOUNDINGBOX	   cboxP0[MAX_PATCH_NUM][4], cboxP1[MAX_PATCH_NUM][4];
  BOUNDINGBOX	   cbox0[4], cbox1[4];
  OVERLAPBOXES     obox_store[MAX_OVERLAP_NUM],
                   ov_patch[MAX_OVERLAP_NUM], reject_pairs[MAX_OVERLAP_NUM]; 
  LIST_POINT4D	   *p4DSp;
  unsigned int	   i, j, k, l, x, y, z;
  int		   p, q, r, s;
  FILE             *pfOutputFile, *fp;
  int              cFine, Num_intersecting_patches, patch_cnt;
  int              NumRejectPairs;
  int		   do_complex, inp_ret, answer, dr, rf;
#ifdef __PERFORM_TIME
  time_t	   time1, time2;
#endif
  INT_CURVE_STRUCT **curveComp, *trimComp;
  int		   numTrimComp, opposing_sides; /* opposing_sides is a flag */
						/* which is 1 if the two solids */
						/* lie on opposite sides of the */
						/* coincident plane - It is a */
						/* special case and is valid only */
						/* for UNIONs */
  TRIM_PATCH_INFO  tmp_opp_sides_sol1, tmp_opp_sides_sol2;
  TRIM_PATCH_INFO  *solid1_copy, *solid2_copy;
  int		   cNo;
  POINT4D	   first_plane, second_plane; /* return from Atul's code */
  double	   tmp_xval, tmp_yval, tmp_zval; /* X,Y,Z points evaluated for */
						 /* overlapping condition */
  double 	   plane_val_on_solid1, plane_val_on_solid2;
  int		   found_solidpt_outside_plane, dumb_flag;
  int		   patch_pair_overlap;
  int 		   first_inside, second_inside;
  POINT2D	   p2D, spt, ept;
  INTERSECTION_CURVE *fptr, *lptr;
  FiveTuple        *uv_arr, *st_arr;
  ParamCurve       *uv_crv, *st_crv;
  int              NumCrvPts, NumCrvs, nghbor, other_nghbor;
  UV_SPLINE_TYPE   *uv_crv1;
  XYZ_SPLINE_TYPE  *xyz_crv1;
  static int	   if_first = 0;
  ANOTHER_ARBIT_STRUCT bndry_info[50];
  int              bndry_info_size, should_transform, do_reject;
  double           transmat[16];
  
  dr = 1;
  sp = rf = 0;

#ifdef _INDIVIDUAL_STEP_SIZE
  printf("Input the Step Size: ");
  scanf("%lf", &Step_Size_Out);
  step_size = Step_Size_Out;
#else
  Step_Size_Out = step_size;
#endif

  if (!frep) {
    printf("findXsectionCurve: Solids not in spline reprn\n");
    exit(-1);
  }

  for (i=0; i<MAX_OVERLAP_NUM; i++)
    for (j=0; j<MAX_SPLINES_PER_TRIM_REGION; j++) 
      special_crvs_in_first_sol[i][j] = special_crvs_in_sec_sol[i][j] = -1;

#ifdef __GL

  if (!if_first) {
    setupWindows(0);
    if_first++;
  }

  switch (op) {
  case 1:
    strcpy(str,"Union");
    break;
  case 2:
    strcpy(str,"Intersection");
    break;
  case 3:
    strcpy(str,"Difference");
    break;
  }
  draw_the_patches("", *first_solid, first_set_patches,
		   *second_solid, second_set_patches, 0, 0, str);

#endif

  ALLOCN(solid1_copy, TRIM_PATCH_INFO, first_set_patches);
  ALLOCN(solid2_copy, TRIM_PATCH_INFO, second_set_patches);

#ifdef __PERFORM_TIME
  Time1 = GetTimer;
#endif
  for (i=0; i<first_set_patches; i++) {
    copy_trim_patch((*first_solid + i), &(solid1_copy[i])); 
    compute_bounding_box((*first_solid)[i].tp, &boxP0[i]);
  }

#ifndef _KEEP_OVRLAP
  do {
    printf("Pruning Xsections\n");
    do_reject = 1;
    should_transform = 0;
#endif
    for (i=0; i<second_set_patches; i++) {
      copy_trim_patch((*second_solid + i), &(solid2_copy[i]));
      compute_bounding_box((*second_solid)[i].tp, &boxP1[i]);
    }
    
    for (i=0; i<MAX_PATCH_NUM; i++) {
      planar_ovrlap_flag1[i] = 0;
      planar_ovrlap_flag2[i] = 0;
    }
    
    printf("DONE COMPUTING BOUNDING BOXES\n");
    for (k=i=0; i<first_set_patches; i++)
      for (j=0; j<second_set_patches; j++)
	if (boxes_overlap(boxP0[i], boxP1[j])) {
	  obox_store[k].patchNum0 = i;
	  obox_store[k].patchNum1 = j;
	  k++;
	}
    printf("Number of possibly overlapping patches = %d\n", k);
    Num_intersecting_patches = k;
    NumRejectPairs = NumOvrPatches = 0;
    
    for (i=0; i<Num_intersecting_patches; i++) {
      dumb_flag = ((*first_solid)[obox_store[i].patchNum0].tp.iDeg0 == 1 &&
		   (*first_solid)[obox_store[i].patchNum0].tp.iDeg1 == 1 &&
		   (*second_solid)[obox_store[i].patchNum1].tp.iDeg0 == 1 &&
		   (*second_solid)[obox_store[i].patchNum1].tp.iDeg1 == 1);
      if (dumb_flag)
	patch_pair_overlap = 
	  planar_patches_overlap(&((*first_solid)[obox_store[i].patchNum0].tp),
				 &((*second_solid)[obox_store[i].patchNum1].tp),
				 &first_plane, &second_plane); 
      else
	patch_pair_overlap = 0;
      
      if (patch_pair_overlap) {
	/* storing overlapping patches in different array */
	overlap_pairs[NumOvrPatches].ptches.patchNum0 =
	  obox_store[i].patchNum0;
	overlap_pairs[NumOvrPatches].ptches.patchNum1 =
	  obox_store[i].patchNum1;
	dot_prod = first_plane.dCoord[0]*second_plane.dCoord[0] +
	  first_plane.dCoord[1]*second_plane.dCoord[1] +
	  first_plane.dCoord[2]*second_plane.dCoord[2];
	if (!FP_EQ_EPS(dot_prod, 1.0, 1e-3) && !FP_EQ_EPS(dot_prod, -1.0, 1e-3)) {
	  printf("findXsectionCurve: Wrong return from planar_patches_overlap\n");
	  exit(-1);
	}

	if (FP_EQ_EPS(dot_prod,1.0,1e-3)) {
	  if (op == DIFFERENCE) {
	    overlap_pairs[NumOvrPatches++].same_side = 1;
	    do_reject = 1;
	  } else {
	    should_transform = 1;
	    do_reject = 0;
	    transmat[0]=1.0;transmat[1]=0.0;transmat[2]=0.0;transmat[3]=0.0;
	    transmat[4]=0.0;transmat[5]=1.0;transmat[6]=0.0;transmat[7]=0.0;
	    transmat[8]=0.0;transmat[9]=0.0;transmat[10]=1.0;transmat[11]=0.0;
	    
	    transmat[12]=0.05*first_plane.dCoord[0];
	    transmat[13]=0.05*first_plane.dCoord[1];
	    transmat[14]=0.05*first_plane.dCoord[2];
	    transmat[15]=1.0;

	    if (should_transform)
	      transform_trim((*second_solid), second_set_patches,transmat);
	  }
	} else {
#ifdef _KEEP_OVRLAP
	  overlap_pairs[NumOvrPatches++].same_side = 0;
	  do_reject = 1;
#else
	should_transform = 1;
	do_reject = 0;
	transmat[0]=1.0;transmat[1]=0.0;transmat[2]=0.0;transmat[3]=0.0;
	transmat[4]=0.0;transmat[5]=1.0;transmat[6]=0.0;transmat[7]=0.0;
	transmat[8]=0.0;transmat[9]=0.0;transmat[10]=1.0;transmat[11]=0.0;
	
	transmat[12]=0.01*first_plane.dCoord[0];
	transmat[13]=0.01*first_plane.dCoord[1];
	transmat[14]=0.01*first_plane.dCoord[2];
	transmat[15]=1.0;

	if (should_transform)
	  transform_trim((*second_solid), second_set_patches,transmat);
#endif	
	}

	if ((*first_solid)[obox_store[i].patchNum0].NumTrimPortions != 1 ||
	    (*second_solid)[obox_store[i].patchNum1].NumTrimPortions != 1) { 
	  printf("findXsectionCurve: Too many Trimportions\n");
	  exit(-1);
	}
	
	if (do_reject) {
	  /* Now I put all the neighbors of one of the overlapping patches */
	  /* and the other overlapping patch in a reject set and */
	  /* vice-versa to prevent doing this intersection */
	  for (j=0; j<(*first_solid)[obox_store[i].patchNum0].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
	    reject_pairs[NumRejectPairs].patchNum0 =
	      (*first_solid)[obox_store[i].patchNum0].TrimPortion[0].SplinePortion.AdjInfo[j]; 
	    reject_pairs[NumRejectPairs++].patchNum1 =
	      obox_store[i].patchNum1;
	  }
	  for (j=0; j<(*second_solid)[obox_store[i].patchNum1].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
	    reject_pairs[NumRejectPairs].patchNum0 =
	      obox_store[i].patchNum0;
	    reject_pairs[NumRejectPairs++].patchNum1 =
	      (*second_solid)[obox_store[i].patchNum1].TrimPortion[0].SplinePortion.AdjInfo[j]; 
	  }
	}
      }
    }
#ifndef _KEEP_OVRLAP
  } while (should_transform);
#endif

#ifdef PERF_CSG_TEMPFILES  
  write_solid_to_file(first_set_patches, *first_solid, "temp1.out");
  write_solid_to_file(second_set_patches, *second_solid, "temp2.out");
#endif
  
  for (k=i=0; i<Num_intersecting_patches; i++) {
    for (j=0; j<NumOvrPatches; j++)
      if (overlap_pairs[j].ptches.patchNum0 == obox_store[i].patchNum0 &&
	  overlap_pairs[j].ptches.patchNum1 == obox_store[i].patchNum1)
	break;
    if (j != NumOvrPatches) continue;
    for (j=0; j<NumRejectPairs; j++)
      if (reject_pairs[j].patchNum0 == obox_store[i].patchNum0 &&
	  reject_pairs[j].patchNum1 == obox_store[i].patchNum1)
	break;
    if (j != NumRejectPairs) continue;

    /* If I come here, it means that this is a possibly good candidate */
    /* to perform intersection */
    answer = do_lin_prog((*first_solid)[obox_store[i].patchNum0].tp,
			 (*second_solid)[obox_store[i].patchNum1].tp, -1);
    if (answer == INFEASIBLE) {
      ov_patch[k].patchNum0 = obox_store[i].patchNum0;    
      ov_patch[k].patchNum1 = obox_store[i].patchNum1;    
      k++;
    }
  }

  printf("Number of intersecting patches = %d\n", k);
  printf("Number of Overlapping patches = %d\n", NumOvrPatches);
  patch_cnt = k;
  if (!patch_cnt && !NumOvrPatches) return 0;

  ALLOCN(curveComp, INT_CURVE_STRUCT *, patch_cnt);
  for (i=0; i<patch_cnt; i++) {
    if (ov_patch[i].patchNum0 == -1 || ov_patch[i].patchNum1 == -1) continue;
    calc_curv_bbox((*first_solid)[ov_patch[i].patchNum0].tp, &(cboxP0[i][0]));
    calc_curv_bbox((*second_solid)[ov_patch[i].patchNum1].tp, &(cboxP1[i][0]));
  }
#ifdef __PERFORM_TIME
  Time2 = GetTimer;
  LPTime += TIMEDIFF(Time2, Time1);
#endif

  imp_time = cs_time = trace_time = 0;

#ifdef __GL
  if (sp) loadmatrix(Mp);
#endif
  for (i=0; i<patch_cnt; i++) {
    if (ov_patch[i].patchNum0 == -1 || ov_patch[i].patchNum1 == -1) continue;
    printf("Intersecting Patches are %d(red) and %d(blue)\n",
	   ov_patch[i].patchNum0, ov_patch[i].patchNum1);

#ifdef _LOOP
    do_complex = doLoopDetection(
				 (*first_solid)[ov_patch[i].patchNum0].tp,
				 (*second_solid)[ov_patch[i].patchNum1].tp);
    if (do_complex) {
      printf("This could be a loop case\n");
      HandleLoopCase(first_solid, ov_patch[i].patchNum0,
		     second_solid, ov_patch[i].patchNum1, op, step_size);
    } else { 
#else
      do_complex = 0;
#endif
      ALLOCN(curveComp[i], INT_CURVE_STRUCT, MAX_COMP_PER_PATCH);
      for (j=0; j<MAX_COMP_PER_PATCH; j++)
	curveComp[i][j].Curve = NULL;
      
#ifdef __PERFORM_TIME
      Time1 = GetTimer;
#endif
      GetInsideStPointsUV((*first_solid)[ov_patch[i].patchNum0].tp,
			  (*second_solid)[ov_patch[i].patchNum1].tp,
			  &second_imp[ov_patch[i].patchNum1], imp_scale, 
			  boxP1[ov_patch[i].patchNum1], &(cboxP0[i][0]),
			  &p4DSp);
      
      Param_1_in_Imp_0((*first_solid)[ov_patch[i].patchNum0].tp,
		       &first_imp[ov_patch[i].patchNum0], imp_scale, 
		       (*second_solid)[ov_patch[i].patchNum1].tp, 
		       boxP0[ov_patch[i].patchNum0], &(cboxP1[i][0]),
		       p4DSp, &(curveComp[i]), pfOutputFile, 
		       do_complex, dr, &cNo);
#ifdef __PERFORM_TIME
      Time2 = GetTimer;
      XsectTime += TIMEDIFF(Time2, Time1);
#endif
      
      for (j=0; j<cNo; j++) {
	/* I may have to rewrite obtain_correct_trimcurve again if it */
	/* doesnt work well SK - 08/07/95 */
	if (frep) {
#ifdef _ATUL_CORRECT_TRIMCURVE
	  numTrimComp = obtain_correct_trimcurve(
						 &((*first_solid)[ov_patch[i].patchNum0]),
						 &((*second_solid)[ov_patch[i].patchNum1]),
						 &(curveComp[i][j]), &trimComp);
	  for (l=0; l<numTrimComp; l++) {
	    convert_for_fitting(&trimComp[l],
				&(*second_solid)[ov_patch[i].patchNum1].tp,
				&(*first_solid)[ov_patch[i].patchNum0].tp,
				&uv_arr, &st_arr, &NumCrvPts);
/*
   This code was added just to provide ChunFa with the test case for
   his fitting routine.
	       
	     fp = fopen("fit.test","w");
	     for (x=0; x<NumCrvPts; x++)
	     fprintf(fp, "%g %g %g %g %g\n", uv_arr[x].s,
	     uv_arr[x].t, uv_arr[x].x, uv_arr[x].y,
	     uv_arr[x].z); 
	     fprintf(fp, "\n\n\n");
	     for (x=0; x<NumCrvPts; x++)
	     fprintf(fp, "%g %g %g %g %g\n", st_arr[x].s,
	     st_arr[x].t, st_arr[x].x, st_arr[x].y,
	     st_arr[x].z); 
	     fclose(fp);
*/

	    spec_fitting(NumCrvPts, uv_arr, st_arr, &NumCrvs, &uv_crv,
			 &st_crv, (char) -1, .001, 2);

	    /* now free the arrays uv_arr and st_arr to use later */
	    myfree(uv_arr); myfree(st_arr);
	    
	    /* For each of the curves generated merge it */
	    /* into a MERGED_CURVE */
	    for (p=0; p<NumCrvs; p++) {
	      UV_SPLINE_TYPE   tmp_uv_crv;
	      XYZ_SPLINE_TYPE  tmp_xyz_crv, *new_xyz_crv;
	      INTERSECTTION_CURVE *f_ptr, *s_ptr;
	      convert_spline_reprn(&st_crv[p], &tmp_uv_crv,
				   &tmp_xyz_crv);
	      copy_xyz_spline(&tmp_xyz_crv, new_xyz_crv);
	      f_ptr = MergeSplines(&(fs_spl_curves[ov_patch[i].patchNum0]),
				   &fs_nc[ov_patch[i].patchNum0],
				   &tmp_uv_crv, &tmp_xyz_crv,
				   ov_patch[i].patchNum1, 1, new_xyz_crv);
	      /* free cntrl pts inside tmp_uv and tmp_xyz */
	      myfree(tmp_uv_crv.CntrlPts); 
	      myfree(tmp_xyz_crv.CntrlPts); 
	      convert_spline_reprn(&uv_crv[p], &tmp_uv_crv,
				   &tmp_xyz_crv);
	      s_ptr = MergeSplines(&(ss_spl_curves[ov_patch[i].patchNum1]),
				   &ss_nc[ov_patch[i].patchNum1],
				   &tmp_uv_crv, new_xyz_crv,
				   ov_patch[i].patchNum0, 0, NULL);
	      f_ptr->other_solid_ptr = s_ptr;
	      s_ptr->other_solid_ptr = f_ptr;

	      /* free cntrl pts inside tmp_uv adn tmp_xyz */
	      myfree(tmp_uv_crv.CntrlPts); 
	      myfree(tmp_xyz_crv.CntrlPts); 
	      /* START HERE - JUST FINISHED MERGESPLINES */
	    }
	  }
#else
#ifdef __PERFORM_TIME
	  Time1 = GetTimer;
#endif
	  convert_for_fitting(&curveComp[i][j],
			      &(*second_solid)[ov_patch[i].patchNum1].tp,
			      &(*first_solid)[ov_patch[i].patchNum0].tp,
			      &uv_arr, &st_arr, &NumCrvPts);
	  
	  if (NumCrvPts > 0) {
	    spec_fitting(NumCrvPts, uv_arr, st_arr, &NumCrvs, &uv_crv,
			 &st_crv, (char) -1, .01, 2);
	    if (NumCrvs == 0) {
	      printf("Fitting sucks big time\n");
	      exit(-1);
	    }
	  }
	  
	  /* now free the arrays uv_arr and st_arr to use later */
	  myfree(uv_arr); myfree(st_arr);
	  
	  /* For each of the curves generated merge it */
	  /* into a MERGED_CURVE */
	  for (p=0; p<NumCrvs; p++) {
	    UV_SPLINE_TYPE   tmp_uv_crv, tmp_st_crv, *ret_uv_crv, *ret_st_crv;
	    XYZ_SPLINE_TYPE  tmp_xyz_crv, *ret_xyz_crv;
	    
	    convert_spline_reprn(&st_crv[p], &tmp_st_crv,
				 &tmp_xyz_crv);
	    convert_spline_reprn(&uv_crv[p], &tmp_uv_crv,
				 &tmp_xyz_crv);
	    numTrimComp = GetProperTrimCurve(&((*first_solid)[ov_patch[i].patchNum0]),
					     &((*second_solid)[ov_patch[i].patchNum1]),
					     &tmp_st_crv, &tmp_uv_crv, &tmp_xyz_crv, 
					     &ret_st_crv, &ret_uv_crv, &ret_xyz_crv);
	    
	    for (l=0; l<numTrimComp; l++) {
	      XYZ_SPLINE_TYPE  new_xyz_crv;
	      INTERSECTION_CURVE *f_ptr, *s_ptr;

	      copy_xyz_spline(&ret_xyz_crv[l], &new_xyz_crv);
	      f_ptr = MergeSplines(&(fs_spl_curves[ov_patch[i].patchNum0]),
				   &fs_nc[ov_patch[i].patchNum0],
				   &ret_st_crv[l], &ret_xyz_crv[l],
				   ov_patch[i].patchNum1, 1, &new_xyz_crv);
	      s_ptr = MergeSplines(&(ss_spl_curves[ov_patch[i].patchNum1]),
				   &ss_nc[ov_patch[i].patchNum1],
				   &ret_uv_crv[l], &ret_xyz_crv[l],
				   ov_patch[i].patchNum0, 0, NULL);
	      f_ptr->other_solid_ptr = s_ptr;
	      s_ptr->other_solid_ptr = f_ptr;
	      free_xyz_spline(&new_xyz_crv);
	    }
	    /* free cntrl pts inside tmp_uv adn tmp_xyz */
	    for (l=0; l<numTrimComp; l++) {
	      myfree(ret_uv_crv[l].CntrlPts);
	      ret_uv_crv[l].CntrlPts = NULL;
	      myfree(ret_st_crv[l].CntrlPts);
	      ret_st_crv[l].CntrlPts = NULL;
	      myfree(ret_xyz_crv[l].CntrlPts);
	      ret_xyz_crv[l].CntrlPts = NULL;
	    }
	    myfree(tmp_uv_crv.CntrlPts); 
	    tmp_uv_crv.CntrlPts = NULL;
	    myfree(tmp_st_crv.CntrlPts);
	    tmp_st_crv.CntrlPts = NULL;
	    myfree(tmp_xyz_crv.CntrlPts);
	    tmp_xyz_crv.CntrlPts = NULL;
	    /* START HERE - JUST FINISHED MERGESPLINES */
	  }
#ifdef __PERFORM_TIME
	  Time2 = GetTimer;
	  FittingTime += TIMEDIFF(Time2, Time1);
#endif
#endif /* _ATUL_CORRECT_TRIMCURVE */
	} else {
	  numTrimComp = obtain_correct_trimcurve(
						 &((*first_solid)[ov_patch[i].patchNum0]),
						 &((*second_solid)[ov_patch[i].patchNum1]),
						 &(curveComp[i][j]), &trimComp);
	  Merge2DLists(&(fs_pw_curves[ov_patch[i].patchNum0]), 
		       &fs_nc[ov_patch[i].patchNum0],
		       trimComp[l], 1);
	  Merge2DLists(&(ss_pw_curves[ov_patch[i].patchNum1]), 
		       &ss_nc[ov_patch[i].patchNum1],
		       trimComp[l], 2);
	}
      }
      p4DSp = NULL;
#ifdef _LOOP
    }
#endif
  }
  
  for (i=0; i<NumOvrPatches; i++)
    NumCurvesBeforeProjection[i] = ss_nc[overlap_pairs[i].ptches.patchNum1];

  for (i=0; i<NumOvrPatches; i++) {
    printf("Inside the overlapping code\n");
    printf("Intersecting Patches are %d(red) and %d(blue)\n",
	   overlap_pairs[i].ptches.patchNum0, overlap_pairs[i].ptches.patchNum1);
    if ((op == PLANE_CUT) || (!overlap_pairs[i].same_side && (op != DOUNION)))
      continue;

    /* Now I'll project each curve of the second patch boundary on the */
    /* domain of the first patch */
    bndry_info_size = 0;
    for (j=0; j<(*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
      if (overlap_pairs[i].same_side) {
	switch (op) {
	case DOUNION:
	  nghbor = overlap_pairs[i].ptches.patchNum1;
	  other_nghbor = -2;
	  break;
	case INTERSECTION:
	case DIFFERENCE:
	  nghbor =
	    (*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	  other_nghbor = overlap_pairs[i].ptches.patchNum1;
	  break;
	}
      } else {
	/* In this case, op can only be UNION */
	if (op != DOUNION) {
	  printf("Value of op is badly corrupted\n");
	  exit(-1);
	}
	nghbor = (*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	other_nghbor = overlap_pairs[i].ptches.patchNum1;
      }
      uv_crv1 = NULL;
      xyz_crv1 = NULL;
      l = ProjectPatch((*first_solid), overlap_pairs[i].ptches.patchNum0,
		       &(*second_solid)[overlap_pairs[i].ptches.patchNum1], j,
		       &uv_crv1, &xyz_crv1, !overlap_pairs[i].same_side,
		       &bndry_info[0], &bndry_info_size, &ssos[i][j]);
      if ((l > 0) && !overlap_pairs[i].same_side)
	special_crvs_in_sec_sol[overlap_pairs[i].ptches.patchNum1][j]
	  = overlap_pairs[i].ptches.patchNum0; 

      if (l > 0)
	first_complete_ovrlap[i] = 1;

      for (k=0; k<l; k++) {
	INTERSECTION_CURVE *fptr;

	fptr = MergeSplines(&(fs_spl_curves[overlap_pairs[i].ptches.patchNum0]),
			    &fs_nc[overlap_pairs[i].ptches.patchNum0],
			    &uv_crv1[k], &xyz_crv1[k], nghbor, 0,
			    NULL);
	fptr->other_solid_ptr = NULL;
	ovr_crv_loops_f[overlap_pairs[i].ptches.patchNum0][fs_nc[overlap_pairs[i].ptches.patchNum0]-1] = other_nghbor;
      }
      for (k=0; k<l; k++) {
	free_uv_spline(&uv_crv1[k]);
	free_xyz_spline(&xyz_crv1[k]);
      }
      myfree(uv_crv1); 
      myfree(xyz_crv1);
    }

    if (!overlap_pairs[i].same_side) 
      /* Now I fill in the fXsc array using the bndry_info array */
      for (j=0; j<fs_nc[overlap_pairs[i].ptches.patchNum0]; j++) {
	return_st_end_in_p2d(&fs_spl_curves[overlap_pairs[i].ptches.patchNum0][j],
			     &spt, &ept, &fptr, &lptr);
	FillXscArray(fptr, lptr, overlap_pairs[i].ptches.patchNum0, j, 1,
		     &bndry_info[0], bndry_info_size); 
      }

    /* Now I'll project each curve of the first patch boundary on the */
    /* domain of the second patch */
    bndry_info_size = 0;
    for (j=0; j<(*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
      if (overlap_pairs[i].same_side) {
	switch (op) {
	case DOUNION:
	  nghbor = overlap_pairs[i].ptches.patchNum1;
	  other_nghbor = -2;
	  break;
	case INTERSECTION:
	case DIFFERENCE:
	  nghbor =
	    (*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	  other_nghbor = overlap_pairs[i].ptches.patchNum0;
	  break;
	}
      } else {
	/* In this case, op can only be UNION */
	if (op != DOUNION) {
	  printf("Value of op is badly corrupted\n");
	  exit(-1);
	}
	nghbor = (*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	other_nghbor = overlap_pairs[i].ptches.patchNum0;
      }
      uv_crv1 = NULL;
      xyz_crv1 = NULL;
      l = ProjectPatch((*second_solid), overlap_pairs[i].ptches.patchNum1,
		       &(*first_solid)[overlap_pairs[i].ptches.patchNum0], j,
		       &uv_crv1, &xyz_crv1, !overlap_pairs[i].same_side, 
		       &bndry_info[0], &bndry_info_size, NULL);
      if ((l > 0) && !overlap_pairs[i].same_side)
	special_crvs_in_first_sol[overlap_pairs[i].ptches.patchNum0][j] =
	  overlap_pairs[i].ptches.patchNum1; 

      if (l > 0)
	second_complete_ovrlap[i] = 1;

      for (k=0; k<l; k++) {
	INTERSECTION_CURVE *sptr;

	sptr = MergeSplines(&(ss_spl_curves[overlap_pairs[i].ptches.patchNum1]),
			    &ss_nc[overlap_pairs[i].ptches.patchNum1],
			    &uv_crv1[k], &xyz_crv1[k], nghbor, 0, NULL);
	sptr->other_solid_ptr = NULL;
	ovr_crv_loops_s[overlap_pairs[i].ptches.patchNum1][ss_nc[overlap_pairs[i].ptches.patchNum1]-1] = other_nghbor;
      }
      for (k=0; k<l; k++) {
	free_uv_spline(&uv_crv1[k]);
	free_xyz_spline(&xyz_crv1[k]);
      }
      myfree(uv_crv1); 
      myfree(xyz_crv1);
    }

    if (!overlap_pairs[i].same_side)
      /* Now I fill in the sXsc array using the bndry_info array */
      for (j=0; j<ss_nc[overlap_pairs[i].ptches.patchNum1]; j++) {
	return_st_end_in_p2d(&ss_spl_curves[overlap_pairs[i].ptches.patchNum1][j],
			     &spt, &ept, &fptr, &lptr);
	FillXscArray(fptr, lptr, overlap_pairs[i].ptches.patchNum1, j, 0,
		     &bndry_info[0], bndry_info_size); 
      }

/*
    planar_ovrlap_flag1[ov_patch[i].patchNum0] = 1;
    planar_ovrlap_flag2[ov_patch[i].patchNum1] = 1;
    
    copy_trim_patch(&((*first_solid)[ov_patch[i].patchNum0]), &tmp_opp_sides_sol1);
    copy_trim_patch(&((*second_solid)[ov_patch[i].patchNum1]), &tmp_opp_sides_sol2);
    handle_planar_patch_operations(op, &tmp_opp_sides_sol1, &tmp_opp_sides_sol2, &p2D);
    EvalPointOnPatch((*first_solid)[ov_patch[i].patchNum0].tp, p2D.u, p2D.v,
		     &tmp_xval, &tmp_yval, &tmp_zval);
    tmp_xval += first_plane.dCoord[0] * SMALL_DISPLACEMENT;
    tmp_yval += first_plane.dCoord[1] * SMALL_DISPLACEMENT;
    tmp_zval += first_plane.dCoord[2] * SMALL_DISPLACEMENT;
    
    first_inside = RobustInOutTest(tmp_xval, tmp_yval, tmp_zval, 
				   solid1_copy, first_set_patches);
    second_inside = RobustInOutTest(tmp_xval, tmp_yval, tmp_zval, 
				    solid2_copy, second_set_patches);
    if (first_inside != second_inside) 
      opposing_sides = 1;
    else opposing_sides = 0;
    free_trim_patch(&tmp_opp_sides_sol1);
    free_trim_patch(&tmp_opp_sides_sol2);   
    
    if (!opposing_sides)
      handle_planar_patch_operations(op, 
				     &((*first_solid)[ov_patch[i].patchNum0]),
				     &((*second_solid)[ov_patch[i].patchNum1]), &p2D);
    else {
      if (op == INTERSECTION) {
	printf("Fatal Error:: Illegal CSG operation on solids\n");
	exit(-1); 
      }
      
      if (op == DIFFERENCE) {
	(*second_solid)[ov_patch[i].patchNum1].NumTrimPortions = 0;
      }  else {  
	copy_trim_patch(&((*first_solid)[ov_patch[i].patchNum0]), &tmp_opp_sides_sol1);
	copy_trim_patch(&((*second_solid)[ov_patch[i].patchNum1]), &tmp_opp_sides_sol2);
	handle_planar_patch_operations(op, 
				       &((*first_solid)[ov_patch[i].patchNum0]),
				       &tmp_opp_sides_sol2, &p2D);
	handle_planar_patch_operations(op, 
				       &((*second_solid)[ov_patch[i].patchNum1]),
				       &tmp_opp_sides_sol1, &p2D);
	
	free_trim_patch(&tmp_opp_sides_sol1);
	free_trim_patch(&tmp_opp_sides_sol2);   
      } 
    }
*/
  }

  for (i=0; i<patch_cnt; i++)
    for (j=0; j<MAX_COMP_PER_PATCH; j++)
      free_int_curve_struct(&curveComp[i][j]);

  for (i=0; i<patch_cnt; i++) {
    myfree(curveComp[i]);
    curveComp[i] = NULL;
  }
  myfree(curveComp);
  curveComp = NULL;

  for (i=0; i<first_set_patches; i++)
    free_trim_patch(&solid1_copy[i]);
  solid1_copy = NULL;
  for (i=0; i<second_set_patches; i++)
    free_trim_patch(&solid2_copy[i]);
  solid2_copy = NULL;

  printf("Finished computing intersection curve\n");
#ifdef __GL
  /* input_and_loop(); */
#endif
  return patch_cnt;
}
#else
int findXsectionCurve(first_solid, first_set_patches, second_solid,
		      second_set_patches, frep, op, step_size,
		      ovr_crv_loops_f, ovr_crv_loops_s)
TRIM_PATCH_INFO **first_solid, **second_solid;
int first_set_patches, second_set_patches, op, frep; /* frep is for */
						     /* representation */
						     /* 0 - piecewise */
						     /* 1 - splines */
double step_size;
short **ovr_crv_loops_f, **ovr_crv_loops_s;
{ 

  TENSOR_PATCH	   tpPatch0[MAX_PATCH_NUM], tpPatch1[MAX_PATCH_NUM];
  TENSOR_PATCH     ptch0, ptch1;
  double	   local_scale, temporary_store, dot_prod;
  char             str[15];
  BOUNDINGBOX	   boxP0[MAX_PATCH_NUM], boxP1[MAX_PATCH_NUM];
  BOUNDINGBOX	   box0, box1;
  BOUNDINGBOX	   cboxP0[MAX_PATCH_NUM][4], cboxP1[MAX_PATCH_NUM][4];
  BOUNDINGBOX	   cbox0[4], cbox1[4];
  OVERLAPBOXES     obox_store[MAX_OVERLAP_NUM],
                   ov_patch[MAX_OVERLAP_NUM], reject_pairs[MAX_OVERLAP_NUM]; 
  LIST_POINT4D	   *p4DSp;
  unsigned int	   i, j, k, l, x, y, z;
  int		   p, q, r, s;
  FILE             *pfOutputFile, *fp;
  int              cFine, Num_intersecting_patches, patch_cnt;
  int              NumRejectPairs;
  int		   do_complex, inp_ret, answer, dr, rf;
  time_t	   time1, time2;
  INT_CURVE_STRUCT **curveComp, *trimComp;
  int		   numTrimComp, opposing_sides; /* opposing_sides is a flag */
						/* which is 1 if the two solids */
						/* lie on opposite sides of the */
						/* coincident plane - It is a */
						/* special case and is valid only */
						/* for DOUNIONs */
  TRIM_PATCH_INFO  tmp_opp_sides_sol1, tmp_opp_sides_sol2;
  TRIM_PATCH_INFO  *solid1_copy, *solid2_copy;
  int		   cNo;
  POINT4D	   first_plane, second_plane; /* return from Atul's code */
  double	   tmp_xval, tmp_yval, tmp_zval; /* X,Y,Z points evaluated for */
						 /* overlapping condition */
  double 	   plane_val_on_solid1, plane_val_on_solid2;
  int		   found_solidpt_outside_plane, dumb_flag;
  int		   patch_pair_overlap;
  int 		   first_inside, second_inside;
  POINT2D	   p2D, spt, ept;
  INTERSECTION_CURVE *fptr, *lptr;
  FiveTuple        *uv_arr, *st_arr;
  ParamCurve       *uv_crv, *st_crv;
  int              NumCrvPts, NumCrvs, nghbor, other_nghbor;
  UV_SPLINE_TYPE   *uv_crv1;
  XYZ_SPLINE_TYPE  *xyz_crv1;
  static int	   if_first = 0;
  ANOTHER_ARBIT_STRUCT bndry_info[50];
  int              bndry_info_size, should_transform, do_reject;
  double           transmat[16];

  dr = 1;
  sp = rf = 0;

#ifdef _INDIVIDUAL_STEP_SIZE
  printf("Input the Step Size: ");
  scanf("%lf", &Step_Size_Out);
  step_size = Step_Size_Out;
#else
  Step_Size_Out = step_size;
#endif

  if (!frep) {
    printf("findXsectionCurve: Solids not in spline reprn\n");
    exit(-1);
  }

  for (i=0; i<MAX_OVERLAP_NUM; i++)
    for (j=0; j<MAX_SPLINES_PER_TRIM_REGION; j++) 
      special_crvs_in_first_sol[i][j] = special_crvs_in_sec_sol[i][j] = -1;

#ifdef __GL

  if (!if_first) {
    setupWindows(0);
    if_first++;
  }

  switch (op) {
  case 1:
    strcpy(str,"Union");
    break;
  case 2:
    strcpy(str,"Intersection");
    break;
  case 3:
    strcpy(str,"Difference");
    break;
  }
  draw_the_patches("", *first_solid, first_set_patches,
		   *second_solid, second_set_patches, 0, 0, str);

#endif

  ALLOCN(solid1_copy, TRIM_PATCH_INFO, first_set_patches);
  ALLOCN(solid2_copy, TRIM_PATCH_INFO, second_set_patches);

  for (i=0; i<first_set_patches; i++) {
    copy_trim_patch((*first_solid + i), &(solid1_copy[i])); 
    compute_bounding_box((*first_solid)[i].tp, &boxP0[i]);
  }

#ifndef _KEEP_OVRLAP
  do {
    printf("Pruning Xsections\n");
    do_reject = 1;
    should_transform = 0;
#endif
    for (i=0; i<second_set_patches; i++) {
      copy_trim_patch((*second_solid + i), &(solid2_copy[i]));
      compute_bounding_box((*second_solid)[i].tp, &boxP1[i]);
    }
    
    for (i=0; i<MAX_PATCH_NUM; i++) {
      planar_ovrlap_flag1[i] = 0;
      planar_ovrlap_flag2[i] = 0;
    }
    
    printf("DONE COMPUTING BOUNDING BOXES\n");
    for (k=i=0; i<first_set_patches; i++)
      for (j=0; j<second_set_patches; j++)
	if (boxes_overlap(boxP0[i], boxP1[j])) {
	  obox_store[k].patchNum0 = i;
	  obox_store[k].patchNum1 = j;
	  k++;
	}
    printf("Number of possibly overlapping patches = %d\n", k);
    Num_intersecting_patches = k;
    NumRejectPairs = NumOvrPatches = 0;
    
    for (i=0; i<Num_intersecting_patches; i++) {
      dumb_flag = ((*first_solid)[obox_store[i].patchNum0].tp.iDeg0 == 1 &&
		   (*first_solid)[obox_store[i].patchNum0].tp.iDeg1 == 1 &&
		   (*second_solid)[obox_store[i].patchNum1].tp.iDeg0 == 1 &&
		   (*second_solid)[obox_store[i].patchNum1].tp.iDeg1 == 1);
      if (dumb_flag)
	patch_pair_overlap = 
	  planar_patches_overlap(&((*first_solid)[obox_store[i].patchNum0].tp),
				 &((*second_solid)[obox_store[i].patchNum1].tp),
				 &first_plane, &second_plane); 
      else
	patch_pair_overlap = 0;
      
      if (patch_pair_overlap) {
	/* storing overlapping patches in different array */
	overlap_pairs[NumOvrPatches].ptches.patchNum0 =
	  obox_store[i].patchNum0;
	overlap_pairs[NumOvrPatches].ptches.patchNum1 =
	  obox_store[i].patchNum1;
	dot_prod = first_plane.dCoord[0]*second_plane.dCoord[0] +
	  first_plane.dCoord[1]*second_plane.dCoord[1] +
	  first_plane.dCoord[2]*second_plane.dCoord[2];
	if (!FP_EQ_EPS(dot_prod, 1.0, 1e-3) && !FP_EQ_EPS(dot_prod, -1.0, 1e-3)) {
	  printf("findXsectionCurve: Wrong return from planar_patches_overlap\n");
	  exit(-1);
	}

	if (FP_EQ_EPS(dot_prod,1.0,1e-3)) {
	  if (op == DIFFERENCE) {
	    overlap_pairs[NumOvrPatches++].same_side = 1;
	    do_reject = 1;
	  } else {
	    should_transform = 1;
	    do_reject = 0;
	    transmat[0]=1.0;transmat[1]=0.0;transmat[2]=0.0;transmat[3]=0.0;
	    transmat[4]=0.0;transmat[5]=1.0;transmat[6]=0.0;transmat[7]=0.0;
	    transmat[8]=0.0;transmat[9]=0.0;transmat[10]=1.0;transmat[11]=0.0;
	    
	    transmat[12]=0.05*first_plane.dCoord[0];
	    transmat[13]=0.05*first_plane.dCoord[1];
	    transmat[14]=0.05*first_plane.dCoord[2];
	    transmat[15]=1.0;

	    if (should_transform)
	      transform_trim((*second_solid), second_set_patches,transmat);
	  }
	} else {
#ifdef _KEEP_OVRLAP
	  overlap_pairs[NumOvrPatches++].same_side = 0;
	  do_reject = 1;
#else
	should_transform = 1;
	do_reject = 0;
	transmat[0]=1.0;transmat[1]=0.0;transmat[2]=0.0;transmat[3]=0.0;
	transmat[4]=0.0;transmat[5]=1.0;transmat[6]=0.0;transmat[7]=0.0;
	transmat[8]=0.0;transmat[9]=0.0;transmat[10]=1.0;transmat[11]=0.0;
	
	transmat[12]=0.01*first_plane.dCoord[0];
	transmat[13]=0.01*first_plane.dCoord[1];
	transmat[14]=0.01*first_plane.dCoord[2];
	transmat[15]=1.0;

	if (should_transform)
	  transform_trim((*second_solid), second_set_patches,transmat);
#endif	
	}

	if ((*first_solid)[obox_store[i].patchNum0].NumTrimPortions != 1 ||
	    (*second_solid)[obox_store[i].patchNum1].NumTrimPortions != 1) { 
	  printf("findXsectionCurve: Too many Trimportions\n");
	  exit(-1);
	}
	
	if (do_reject) {
	  /* Now I put all the neighbors of one of the overlapping patches */
	  /* and the other overlapping patch in a reject set and */
	  /* vice-versa to prevent doing this intersection */
	  for (j=0; j<(*first_solid)[obox_store[i].patchNum0].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
	    reject_pairs[NumRejectPairs].patchNum0 =
	      (*first_solid)[obox_store[i].patchNum0].TrimPortion[0].SplinePortion.AdjInfo[j]; 
	    reject_pairs[NumRejectPairs++].patchNum1 =
	      obox_store[i].patchNum1;
	  }
	  for (j=0; j<(*second_solid)[obox_store[i].patchNum1].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
	    reject_pairs[NumRejectPairs].patchNum0 =
	      obox_store[i].patchNum0;
	    reject_pairs[NumRejectPairs++].patchNum1 =
	      (*second_solid)[obox_store[i].patchNum1].TrimPortion[0].SplinePortion.AdjInfo[j]; 
	  }
	}
      }
    }
#ifndef _KEEP_OVRLAP
  } while (should_transform);
#endif
    
  for (k=i=0; i<Num_intersecting_patches; i++) {
    for (j=0; j<NumOvrPatches; j++)
      if (overlap_pairs[j].ptches.patchNum0 == obox_store[i].patchNum0 &&
	  overlap_pairs[j].ptches.patchNum1 == obox_store[i].patchNum1)
	break;
    if (j != NumOvrPatches) continue;
    for (j=0; j<NumRejectPairs; j++)
      if (reject_pairs[j].patchNum0 == obox_store[i].patchNum0 &&
	  reject_pairs[j].patchNum1 == obox_store[i].patchNum1)
	break;
    if (j != NumRejectPairs) continue;
    
    /* If I come here, it means that this is a possibly good candidate */
    /* to perform intersection */
    answer = do_lin_prog((*first_solid)[obox_store[i].patchNum0].tp,
			 (*second_solid)[obox_store[i].patchNum1].tp, -1);
    if (answer == INFEASIBLE) {
      ov_patch[k].patchNum0 = obox_store[i].patchNum0;    
      ov_patch[k].patchNum1 = obox_store[i].patchNum1;    
      k++;
    }
  }

  printf("Number of intersecting patches = %d\n", k);
  printf("Number of Overlapping patches = %d\n", NumOvrPatches);
  patch_cnt = k;
  if (!patch_cnt && !NumOvrPatches) return 0;

  ALLOCN(curveComp, INT_CURVE_STRUCT *, patch_cnt);
  for (i=0; i<patch_cnt; i++) {
    if (ov_patch[i].patchNum0 == -1 || ov_patch[i].patchNum1 == -1) continue;
    calc_curv_bbox((*first_solid)[ov_patch[i].patchNum0].tp, &(cboxP0[i][0]));
    calc_curv_bbox((*second_solid)[ov_patch[i].patchNum1].tp, &(cboxP1[i][0]));
  }

  imp_time = cs_time = trace_time = 0;

#ifdef __GL
  if (sp) loadmatrix(Mp);
#endif
  for (i=0; i<patch_cnt; i++) {
    if (ov_patch[i].patchNum0 == -1 || ov_patch[i].patchNum1 == -1) continue;
    printf("Intersecting Patches are %d(red) and %d(blue)\n",
	   ov_patch[i].patchNum0, ov_patch[i].patchNum1);

#ifdef _LOOP
    do_complex = doLoopDetection(
				 (*first_solid)[ov_patch[i].patchNum0].tp,
				 (*second_solid)[ov_patch[i].patchNum1].tp);
    if (do_complex) {
      printf("This could be a loop case\n");
      HandleLoopCase(first_solid, ov_patch[i].patchNum0,
		     second_solid, ov_patch[i].patchNum1, op, step_size);
    } else { 
#else
      do_complex = 0;
#endif
      ALLOCN(curveComp[i], INT_CURVE_STRUCT, MAX_COMP_PER_PATCH);
      for (j=0; j<MAX_COMP_PER_PATCH; j++)
	curveComp[i][j].Curve = NULL;
      
      GetInsideStPointsUV((*first_solid)[ov_patch[i].patchNum0].tp,
			  (*second_solid)[ov_patch[i].patchNum1].tp,
			  boxP1[ov_patch[i].patchNum1], &(cboxP0[i][0]),
			  &p4DSp);
      
      Param_1_in_Imp_0((*first_solid)[ov_patch[i].patchNum0].tp, 
		       (*second_solid)[ov_patch[i].patchNum1].tp, 
		       boxP0[ov_patch[i].patchNum0], &(cboxP1[i][0]),
		       p4DSp, &(curveComp[i]), pfOutputFile, 
		       do_complex, dr, &cNo);
      
      for (j=0; j<cNo; j++) {
/*
	POINT3D starting_p3d[10], ending_p3d[10];

	for (l=0; l<10; l++) {
	  starting_p3d[l].dCoord[0] = ending_p3d[l].dCoord[0] = HUGE_VAL;
	  starting_p3d[l].dCoord[1] = ending_p3d[l].dCoord[1] = HUGE_VAL;
	  starting_p3d[l].dCoord[2] = ending_p3d[l].dCoord[2] = HUGE_VAL;
	}
*/
	/* I may have to rewrite obtain_correct_trimcurve again if it */
	/* doesnt work well SK - 08/07/95 */
	if (frep) {
#ifdef _ATUL_CORRECT_TRIMCURVE
	  numTrimComp = obtain_correct_trimcurve(
						 &((*first_solid)[ov_patch[i].patchNum0]),
						 &((*second_solid)[ov_patch[i].patchNum1]),
						 &(curveComp[i][j]), &trimComp);
	  for (l=0; l<numTrimComp; l++) {
	    convert_for_fitting(&trimComp[l],
				&(*second_solid)[ov_patch[i].patchNum1].tp,
				&(*first_solid)[ov_patch[i].patchNum0].tp,
				&uv_arr, &st_arr, &NumCrvPts);
/*
   This code was added just to provide ChunFa with the test case for
   his fitting routine.
	       
	     fp = fopen("fit.test","w");
	     for (x=0; x<NumCrvPts; x++)
	     fprintf(fp, "%g %g %g %g %g\n", uv_arr[x].s,
	     uv_arr[x].t, uv_arr[x].x, uv_arr[x].y,
	     uv_arr[x].z); 
	     fprintf(fp, "\n\n\n");
	     for (x=0; x<NumCrvPts; x++)
	     fprintf(fp, "%g %g %g %g %g\n", st_arr[x].s,
	     st_arr[x].t, st_arr[x].x, st_arr[x].y,
	     st_arr[x].z); 
	     fclose(fp);
*/

	    spec_fitting(NumCrvPts, uv_arr, st_arr, &NumCrvs, &uv_crv,
			 &st_crv, (char) -1, .01, 2);

	    /* now free the arrays uv_arr and st_arr to use later */
	    myfree(uv_arr); myfree(st_arr);
	    
	    /* For each of the curves generated merge it */
	    /* into a MERGED_CURVE */
	    for (p=0; p<NumCrvs; p++) {
	      UV_SPLINE_TYPE   tmp_uv_crv;
	      XYZ_SPLINE_TYPE  tmp_xyz_crv;

	      convert_spline_reprn(&st_crv[p], &tmp_uv_crv,
				   &tmp_xyz_crv);
	      MergeSplines(&(fs_spl_curves[ov_patch[i].patchNum0]),
			   &fs_nc[ov_patch[i].patchNum0],
			   &tmp_uv_crv, &tmp_xyz_crv,
			   ov_patch[i].patchNum1);
	      /* free cntrl pts inside tmp_uv and tmp_xyz */
	      myfree(tmp_uv_crv.CntrlPts); 
	      myfree(tmp_xyz_crv.CntrlPts); 
	      convert_spline_reprn(&uv_crv[p], &tmp_uv_crv,
				   &tmp_xyz_crv);
	      MergeSplines(&(ss_spl_curves[ov_patch[i].patchNum1]),
			   &ss_nc[ov_patch[i].patchNum1],
			   &tmp_uv_crv, &tmp_xyz_crv,
			   ov_patch[i].patchNum0);
	      /* free cntrl pts inside tmp_uv adn tmp_xyz */
	      myfree(tmp_uv_crv.CntrlPts); 
	      myfree(tmp_xyz_crv.CntrlPts); 
	      /* START HERE - JUST FINISHED MERGESPLINES */
	    }
#else
	    convert_for_fitting(&curveComp[i][j],
				&(*second_solid)[ov_patch[i].patchNum1].tp,
				&(*first_solid)[ov_patch[i].patchNum0].tp,
				&uv_arr, &st_arr, &NumCrvPts);

	    spec_fitting(NumCrvPts, uv_arr, st_arr, &NumCrvs, &uv_crv,
			 &st_crv, (char) -1, .01, 2);

	    /* now free the arrays uv_arr and st_arr to use later */
	    myfree(uv_arr); myfree(st_arr);
	    
	    /* For each of the curves generated merge it */
	    /* into a MERGED_CURVE */
	    for (p=0; p<NumCrvs; p++) {
	      UV_SPLINE_TYPE   tmp_uv_crv, tmp_st_crv, *ret_uv_crv, *ret_st_crv;
	      XYZ_SPLINE_TYPE  tmp_xyz_crv, *ret_xyz_crv;

	      convert_spline_reprn(&st_crv[p], &tmp_st_crv,
				   &tmp_xyz_crv);
	      convert_spline_reprn(&uv_crv[p], &tmp_uv_crv,
				   &tmp_xyz_crv);
	      numTrimComp = GetProperTrimCurve(&((*first_solid)[ov_patch[i].patchNum0]),
					       &((*second_solid)[ov_patch[i].patchNum1]),
					       &tmp_st_crv, &tmp_uv_crv, &tmp_xyz_crv, 
					       &ret_st_crv, &ret_uv_crv, &ret_xyz_crv);

	      for (l=0; l<numTrimComp; l++) {
		MergeSplines(&(fs_spl_curves[ov_patch[i].patchNum0]),
			     &fs_nc[ov_patch[i].patchNum0],
			     &ret_st_crv[l], &ret_xyz_crv[l],
			     ov_patch[i].patchNum1);
		MergeSplines(&(ss_spl_curves[ov_patch[i].patchNum1]),
			     &ss_nc[ov_patch[i].patchNum1],
			     &ret_uv_crv[l], &ret_xyz_crv[l],
			     ov_patch[i].patchNum0);
	      }
	      /* free cntrl pts inside tmp_uv adn tmp_xyz */
	      for (l=0; l<numTrimComp; l++) {
		myfree(ret_uv_crv[l].CntrlPts);
		myfree(ret_st_crv[l].CntrlPts);
		myfree(ret_xyz_crv[l].CntrlPts);
	      }
	      myfree(tmp_uv_crv.CntrlPts); 
	      myfree(tmp_st_crv.CntrlPts); 
	      myfree(tmp_xyz_crv.CntrlPts); 
	      /* START HERE - JUST FINISHED MERGESPLINES */
	    }
#endif /* _ATUL_CORRECT_TRIMCURVE */
	  } else {
	    numTrimComp = obtain_correct_trimcurve(
					       &((*first_solid)[ov_patch[i].patchNum0]),
					       &((*second_solid)[ov_patch[i].patchNum1]),
					       &(curveComp[i][j]), &trimComp);
	    Merge2DLists(&(fs_pw_curves[ov_patch[i].patchNum0]), 
			 &fs_nc[ov_patch[i].patchNum0],
			 trimComp[l], 1);
	    Merge2DLists(&(ss_pw_curves[ov_patch[i].patchNum1]), 
			 &ss_nc[ov_patch[i].patchNum1],
			 trimComp[l], 2);
	  }
	}
      }
      p4DSp = NULL;
#ifdef _LOOP
    }
#endif
  }

  for (i=0; i<NumOvrPatches; i++)
    NumCurvesBeforeProjection[i] = ss_nc[overlap_pairs[i].ptches.patchNum1];

  for (i=0; i<NumOvrPatches; i++) {
    printf("Inside the overlapping code\n");
    printf("Intersecting Patches are %d(red) and %d(blue)\n",
	   overlap_pairs[i].ptches.patchNum0, overlap_pairs[i].ptches.patchNum1);
    if ((op == PLANE_CUT) || (!overlap_pairs[i].same_side && (op != DOUNION)))
      continue;

    /* Now I'll project each curve of the second patch boundary on the */
    /* domain of the first patch */
    bndry_info_size = 0;
    for (j=0; j<(*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
      if (overlap_pairs[i].same_side) {
	switch (op) {
	case DOUNION:
	  nghbor = overlap_pairs[i].ptches.patchNum1;
	  other_nghbor = -2;
	  break;
	case INTERSECTION:
	case DIFFERENCE:
	  nghbor =
	    (*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	  other_nghbor = overlap_pairs[i].ptches.patchNum1;
	  break;
	}
      } else {
	/* In this case, op can only be DOUNION */
	if (op != DOUNION) {
	  printf("Value of op is badly corrupted\n");
	  exit(-1);
	}
	nghbor = (*second_solid)[overlap_pairs[i].ptches.patchNum1].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	other_nghbor = overlap_pairs[i].ptches.patchNum1;
      }
      uv_crv1 = NULL;
      xyz_crv1 = NULL;
      l = ProjectPatch((*first_solid), overlap_pairs[i].ptches.patchNum0,
		       &(*second_solid)[overlap_pairs[i].ptches.patchNum1], j,
		       &uv_crv1, &xyz_crv1, !overlap_pairs[i].same_side,
		       &bndry_info[0], &bndry_info_size, &ssos[i][j]);
      if ((l > 0) && !overlap_pairs[i].same_side)
	special_crvs_in_sec_sol[overlap_pairs[i].ptches.patchNum1][j]
	  = overlap_pairs[i].ptches.patchNum0; 

      if (l > 0)
	first_complete_ovrlap[i] = 1;

      for (k=0; k<l; k++) {
	MergeSplines(&(fs_spl_curves[overlap_pairs[i].ptches.patchNum0]),
		     &fs_nc[overlap_pairs[i].ptches.patchNum0],
		     &uv_crv1[k], &xyz_crv1[k], nghbor);
	ovr_crv_loops_f[overlap_pairs[i].ptches.patchNum0][fs_nc[overlap_pairs[i].ptches.patchNum0]-1] = other_nghbor;
      }
      for (k=0; k<l; k++) {
	free_uv_spline(&uv_crv1[k]);
	free_xyz_spline(&xyz_crv1[k]);
      }
      myfree(uv_crv1); 
      myfree(xyz_crv1);
    }

    if (!overlap_pairs[i].same_side) 
      /* Now I fill in the fXsc array using the bndry_info array */
      for (j=0; j<fs_nc[overlap_pairs[i].ptches.patchNum0]; j++) {
	return_st_end_in_p2d(&fs_spl_curves[overlap_pairs[i].ptches.patchNum0][j],
			     &spt, &ept, &fptr, &lptr);
	FillXscArray(fptr, lptr, overlap_pairs[i].ptches.patchNum0, j, 1,
		     &bndry_info[0], bndry_info_size); 
      }

    /* Now I'll project each curve of the first patch boundary on the */
    /* domain of the second patch */
    bndry_info_size = 0;
    for (j=0; j<(*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].SplinePortion.NumOfCurves; j++) {
      if (overlap_pairs[i].same_side) {
	switch (op) {
	case DOUNION:
	  nghbor = overlap_pairs[i].ptches.patchNum1;
	  other_nghbor = -2;
	  break;
	case INTERSECTION:
	case DIFFERENCE:
	  nghbor =
	    (*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	  other_nghbor = overlap_pairs[i].ptches.patchNum0;
	  break;
	}
      } else {
	/* In this case, op can only be DOUNION */
	if (op != DOUNION) {
	  printf("Value of op is badly corrupted\n");
	  exit(-1);
	}
	nghbor = (*first_solid)[overlap_pairs[i].ptches.patchNum0].TrimPortion[0].
                                                         SplinePortion.AdjInfo[j];
	other_nghbor = overlap_pairs[i].ptches.patchNum0;
      }
      uv_crv1 = NULL;
      xyz_crv1 = NULL;
      l = ProjectPatch((*second_solid), overlap_pairs[i].ptches.patchNum1,
		       &(*first_solid)[overlap_pairs[i].ptches.patchNum0], j,
		       &uv_crv1, &xyz_crv1, !overlap_pairs[i].same_side, 
		       &bndry_info[0], &bndry_info_size, NULL);
      if ((l > 0) && !overlap_pairs[i].same_side)
	special_crvs_in_first_sol[overlap_pairs[i].ptches.patchNum0][j] =
	  overlap_pairs[i].ptches.patchNum1; 

      if (l > 0)
	second_complete_ovrlap[i] = 1;

      for (k=0; k<l; k++) {
	MergeSplines(&(ss_spl_curves[overlap_pairs[i].ptches.patchNum1]),
		     &ss_nc[overlap_pairs[i].ptches.patchNum1],
		     &uv_crv1[k], &xyz_crv1[k], nghbor);
	ovr_crv_loops_s[overlap_pairs[i].ptches.patchNum1][ss_nc[overlap_pairs[i].ptches.patchNum1]-1] = other_nghbor;
      }
      for (k=0; k<l; k++) {
	free_uv_spline(&uv_crv1[k]);
	free_xyz_spline(&xyz_crv1[k]);
      }
      myfree(uv_crv1); 
      myfree(xyz_crv1);
    }

    if (!overlap_pairs[i].same_side)
      /* Now I fill in the sXsc array using the bndry_info array */
      for (j=0; j<ss_nc[overlap_pairs[i].ptches.patchNum1]; j++) {
	return_st_end_in_p2d(&ss_spl_curves[overlap_pairs[i].ptches.patchNum1][j],
			     &spt, &ept, &fptr, &lptr);
	FillXscArray(fptr, lptr, overlap_pairs[i].ptches.patchNum1, j, 0,
		     &bndry_info[0], bndry_info_size); 
      }

/*
    planar_ovrlap_flag1[ov_patch[i].patchNum0] = 1;
    planar_ovrlap_flag2[ov_patch[i].patchNum1] = 1;
    
    copy_trim_patch(&((*first_solid)[ov_patch[i].patchNum0]), &tmp_opp_sides_sol1);
    copy_trim_patch(&((*second_solid)[ov_patch[i].patchNum1]), &tmp_opp_sides_sol2);
    handle_planar_patch_operations(op, &tmp_opp_sides_sol1, &tmp_opp_sides_sol2, &p2D);
    EvalPointOnPatch((*first_solid)[ov_patch[i].patchNum0].tp, p2D.u, p2D.v,
		     &tmp_xval, &tmp_yval, &tmp_zval);
    tmp_xval += first_plane.dCoord[0] * SMALL_DISPLACEMENT;
    tmp_yval += first_plane.dCoord[1] * SMALL_DISPLACEMENT;
    tmp_zval += first_plane.dCoord[2] * SMALL_DISPLACEMENT;
    
    first_inside = RobustInOutTest(tmp_xval, tmp_yval, tmp_zval, 
				   solid1_copy, first_set_patches);
    second_inside = RobustInOutTest(tmp_xval, tmp_yval, tmp_zval, 
				    solid2_copy, second_set_patches);
    if (first_inside != second_inside) 
      opposing_sides = 1;
    else opposing_sides = 0;
    free_trim_patch(&tmp_opp_sides_sol1);
    free_trim_patch(&tmp_opp_sides_sol2);   
    
    if (!opposing_sides)
      handle_planar_patch_operations(op, 
				     &((*first_solid)[ov_patch[i].patchNum0]),
				     &((*second_solid)[ov_patch[i].patchNum1]), &p2D);
    else {
      if (op == INTERSECTION) {
	printf("Fatal Error:: Illegal CSG operation on solids\n");
	exit(-1); 
      }
      
      if (op == DIFFERENCE) {
	(*second_solid)[ov_patch[i].patchNum1].NumTrimPortions = 0;
      }  else {  
	copy_trim_patch(&((*first_solid)[ov_patch[i].patchNum0]), &tmp_opp_sides_sol1);
	copy_trim_patch(&((*second_solid)[ov_patch[i].patchNum1]), &tmp_opp_sides_sol2);
	handle_planar_patch_operations(op, 
				       &((*first_solid)[ov_patch[i].patchNum0]),
				       &tmp_opp_sides_sol2, &p2D);
	handle_planar_patch_operations(op, 
				       &((*second_solid)[ov_patch[i].patchNum1]),
				       &tmp_opp_sides_sol1, &p2D);
	
	free_trim_patch(&tmp_opp_sides_sol1);
	free_trim_patch(&tmp_opp_sides_sol2);   
      } 
    }
*/
  }

  printf("Finished computing intersection curve\n");
#ifdef __GL
  /* input_and_loop(); */
#endif
  return patch_cnt;
}
#endif

#ifdef MYMAIN
main(argc, argv)
int argc;
char *argv[];
{
  int i, j, solid1_cnt, solid2_cnt;
  FILE *fp;
  TRIM_PATCH_INFO *solid1, *solid2, *psolid;
  TENSOR_PATCH tpPatch0[MAX_PATCH_NUM], tpPatch1[MAX_PATCH_NUM];
  double ssize;
  int parent_patch_num;
  int frep, srep;

/*
  setupWindows(0);
  draw_the_patches(argv[1], NULL, 0, NULL, 0, 0, 0);
*/

  sscanf(argv[2], "%lf", &ssize);
  sscanf(argv[3], "%lf", &Sphradius);
  read_input(tpPatch0, tpPatch1, &solid1_cnt, &solid2_cnt, argv[1]);
  ALLOCN(solid1, TRIM_PATCH_INFO, solid1_cnt);
  ALLOCN(solid2, TRIM_PATCH_INFO, solid2_cnt);
  ALLOCN(psolid, TRIM_PATCH_INFO, solid1_cnt+solid2_cnt);
  for (i=0; i<solid1_cnt; i++) {
    solid1[i].tp = tpPatch0[i];
    solid1[i].orig_tp = tpPatch0[i];
    solid1[i].csg_status = IN_CSG;
    solid1[i].SplineReprn = 0;
    solid1[i].NumTrimPortions = 1;
    solid1[i].TrimPortion[0].PiecePortion.NumOfVertices = 4;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[0][0] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[0][1] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[1][1] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[3][0] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[4][0] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.poly_vertex[4][1] = 0.0;
    solid1[i].TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
    solid1[i].TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;
  }
  for (i=0; i<solid2_cnt; i++) {
    solid2[i].tp = tpPatch1[i];
    solid2[i].orig_tp = tpPatch1[i];
    solid2[i].NumTrimCurves = 0;
    solid2[i].csg_status = IN_CSG;
    solid2[i].SplineReprn = 0;
    solid2[i].NumTrimPortions = 1;
    solid2[i].TrimPortion[0].PiecePortion.NumOfVertices = 4;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[0][0] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[0][1] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[1][0] = 1.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[1][1] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[2][0] = 1.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[2][1] = 1.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[3][0] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[3][1] = 1.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[4][0] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.poly_vertex[4][1] = 0.0;
    solid2[i].TrimPortion[0].PiecePortion.ptInsideRegion.u = 0.5;
    solid2[i].TrimPortion[0].PiecePortion.ptInsideRegion.v = 0.5;
  }

  printf("main: This version runs with piecewise represention\n");
  sleep(3);

  frep = srep = 0;
  /* changing call to perform_CSG to accomodate splines */
  parent_patch_num = perform_CSG(&solid1, solid1_cnt, &solid2, solid2_cnt,
				 frep, srep, DOUNION, 0, NULL, ssize, &psolid);
  printf("Finished Tracing\n");
#ifdef __GL
  input_and_loop();
#endif
}
#endif

int check_for_base_points(ndet, Nr, Nc, rank, row_p, col_p)
NDETERMINANT	ndet;
int		Nr, Nc, *rank;
int		**row_p, **col_p;
{
  double	**rgrgMd;
  double	Xval, Yval, Zval;
  int		i, j, k, ret;

  if (Nr != Nc) {
     printf("Something wrong here! Not a SQUARE matrix\n");
     exit(-1);
  }

  ALLOCMAT(rgrgMd, Nr, Nc, double);
  
  /* Assigning arbitrary values for Xval, Yval and Zval */
  Xval = drand48()*100;
  Yval = drand48()*100;
  Zval = drand48()*100;

  for (i = 0; i < Nr; i++)
  for (j = 0; j < Nc; j++)
     rgrgMd[i][j] = ndet.coeffs[i][j].dCoord[X] * Xval +
                    ndet.coeffs[i][j].dCoord[Y] * Yval +
                    ndet.coeffs[i][j].dCoord[Z] * Zval +
                    ndet.coeffs[i][j].dCoord[W];

  ret = forw_elim(rgrgMd, Nr, rank, row_p, col_p);
  qsort(*row_p,*rank,sizeof(int),compar);
  qsort(*col_p,*rank,sizeof(int),compar);
 
  FREEMAT(rgrgMd, Nr, Nc);

  if (*rank != Nr)
    return 1;
  return 0;
}

/* Solves the system Ax = B  and returns x. It does gauss elimination
   with complete pivoting. A is a (cDim x cDim) matrix. A and B are
   left unchanged. If A is singular, it returns 0.
*/
int
forw_elim(rgrgA, cDim, row_val, row_perm, col_perm)
double **rgrgA;
int    cDim;
int    *row_val;
int    **row_perm, **col_perm;
{

    double **rgrgL;
    int    iTemp;
    int    iOrder=cDim;
    int    i;
    int    currow, iRow, iCol, irgPivotCol, irgPivotRow;
    double dPivot, dTemp, dMaxPivot, dPrevMaxPivot;

    /* initialize the permutation array */
    ALLOCN(*row_perm, int, cDim);
    ALLOCN(*col_perm, int, cDim);
    for(iRow=0; iRow < cDim; iRow++) {
        (*row_perm)[iRow] = iRow;
        (*col_perm)[iRow] = iRow;
    }

    /* copy matrix A into L */
    ALLOCMAT(rgrgL, cDim, cDim, double)
    MATASSIGN(rgrgL, rgrgA, cDim, cDim);

    /* start at row one and loop through rows */
    /* until we have completed gaussian elimination portion */
    /* note: we continue to examine rows even after we have found */
    /* a pivot for each location, because one of the later rows */
    /* may render the matrix system inconsistent */

    dPrevMaxPivot = 0.0;
    for (currow=0;currow<iOrder;currow++) {
        /* find largest remaining value for complete pivoting */
        for (dMaxPivot=0.0,irgPivotRow= -1,irgPivotCol=-1,iRow=currow;iRow<iOrder;iRow++){
            for (iCol=currow;iCol<iOrder;iCol++) {
                dTemp = rgrgL[iRow][iCol];
                if (dTemp<0.0)
                    dTemp = -dTemp;

                if (dTemp > dMaxPivot ) {
                    irgPivotRow=iRow;
                    irgPivotCol=iCol;
                    dMaxPivot=dTemp;
                }
            }

        }

        /* if no pivot was found, there is a row of zeros or linearly
           dependent vectors */

        if ( !currow && FP_EQ_EPS(dMaxPivot, 0.0, 1e-5) ) {
           *row_val = currow;
	   FREEMAT(rgrgL, cDim, cDim);
           return(0);
        } else if (FP_EQ_EPS(dMaxPivot/dPrevMaxPivot,0.0,1e-4)) {
	   *row_val = currow;
	   FREEMAT(rgrgL, cDim, cDim);
	   return 0;
	}

	dPrevMaxPivot = dMaxPivot;
        /* otherwise, we have the largest pivot in irgPivotRow and
           irgPivotCol */

        if (currow!=irgPivotRow) {
            /* exchange rows of rgrgL */
            for (iCol=0;iCol<iOrder;iCol++)  {
                dTemp = rgrgL[currow][iCol];
                rgrgL[currow][iCol]=rgrgL[irgPivotRow][iCol];
                rgrgL[irgPivotRow][iCol]=dTemp;
            }
            /* swap the entries of the row permutation array */
            iTemp = (*row_perm)[irgPivotRow];
            (*row_perm)[irgPivotRow] = (*row_perm)[currow];
            (*row_perm)[currow] = iTemp;
        }

        if (currow!=irgPivotCol) {
            /* exchange cols */
            for (iRow=0;iRow<iOrder;iRow++)  {
                dTemp = rgrgL[iRow][currow];
                rgrgL[iRow][currow]=rgrgL[iRow][irgPivotCol];
                rgrgL[iRow][irgPivotCol]=dTemp;
            }
            /* swap the entries of the col permutation array */
            iTemp = (*col_perm)[irgPivotCol];
            (*col_perm)[irgPivotCol] = (*col_perm)[currow];
            (*col_perm)[currow] = iTemp;
        }


        /* pivot is now at currow,currow */

        dPivot = rgrgL[currow][currow];

        /* subtract this row from all of the others */
        for (iRow=currow+1;iRow<iOrder;iRow++){
            dTemp = rgrgL[iRow][currow]/dPivot;

            /* no need to zero the entry, but I will do it for
               easy error checking */

            rgrgL[iRow][currow]=0.0;

            /* for determinant portion, we must start earlier in
               row */
            for (iCol = currow+1;iCol<iOrder;iCol++)
              rgrgL[iRow][iCol] -= dTemp * rgrgL[currow][iCol];
        }

    }

    *row_val = iOrder;
    FREEMAT(rgrgL, cDim, cDim);

    return 0;
}

#ifdef _CALC_IMPLICIT
Param_1_in_Imp_0(tp0, ndetImplicit, imp_scale, tp1, box0, cbox1, p4DSp, cComp, pfOutputFile, c_trace, dr, cNo)
TENSOR_PATCH tp0, tp1;
double imp_scale;
NDETERMINANT *ndetImplicit;
BOUNDINGBOX box0, *cbox1;
LIST_POINT4D *p4DSp;
INT_CURVE_STRUCT **cComp;
FILE *pfOutputFile;
int c_trace, dr, *cNo;
{
  NDETERMINANT     fdetImplicit;
   /* NDETERMINANT_1D  ndet1D_duu, ndet1D_duv, ndet1D_dvv; */
  LIST_POINT4D     *plp4DStartingPoints, *plp4DComponentPoints;
  LIST_POINT4CPX   *plp4CPXStartingPoints;
  LIST_POINT4CPX   *plp4CPX;
  int              cBlue, ret_val;
  int		   NumBoundaryXsections;
  char             rgchStruct[50];
  int		   Nrows, Ncols;
  int 		   *row_perm, *col_perm;
  int		   base_case, rank_row_val;
  int		   i, j, k, matched_pt;
  POINT4D	   final_pt, *addedList, turn_pt4D;
  int		   addedList_idx;
  TENSOR_PATCH     tpPatch0, tpPatch1;

  /* I am going to scale the patches down so that my matrix entries */
  /* dont get too big. */
  tpPatch0.iDeg0 = tp0.iDeg0;
  tpPatch0.iDeg1 = tp0.iDeg1;
  tpPatch1.iDeg0 = tp1.iDeg0;
  tpPatch1.iDeg1 = tp1.iDeg1;
  allocate_tensor_patch(&tpPatch0);
  allocate_tensor_patch(&tpPatch1);
  /* Now calculating the exact Scale factor */
  Scale = imp_scale;
  for (i=0; i<=tp0.iDeg0; i++)
    for (j=0; j<=tp0.iDeg1; j++)
      for (k=X; k<=W; k++)
	tpPatch0.coeffs[i][j].dCoord[k] = tp0.coeffs[i][j].dCoord[k];
  for (i=0; i<=tp1.iDeg0; i++)
    for (j=0; j<=tp1.iDeg1; j++)
      for (k=X; k<=W; k++)
	tpPatch1.coeffs[i][j].dCoord[k] = tp1.coeffs[i][j].dCoord[k];

  /* Scale the tensor patch */
  scale_tensor_patch(&tpPatch0, 1);
  scale_tensor_patch(&tpPatch1, 1);

  Nrows = ndetImplicit->cRows;
  Ncols = ndetImplicit->cCols;
  base_case = check_for_base_points(*ndetImplicit, Nrows, Ncols, &rank_row_val,
				    &row_perm, &col_perm);

  if (base_case) {
    fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
    ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
    for (i=0; i<fdetImplicit.cCols; i++)
      ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

    for (i=0; i<rank_row_val; i++)
    {
      for (j=0; j<rank_row_val; j++)
      {
        fdetImplicit.coeffs[i][j].dCoord[X] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[X];
        fdetImplicit.coeffs[i][j].dCoord[Y] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
        fdetImplicit.coeffs[i][j].dCoord[Z] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
        fdetImplicit.coeffs[i][j].dCoord[W] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[W];
      }
    }
  }

  /* at this point we have the implicit form in a matrix form,
     where each element of the array signifies a vector (X,Y,Z,W)
     and the vector means that the elements are linear in the
     corresponding variables, so each entry is actually
           x*X + y*Y + z*Z + W,

     The implicit equation of the surface is the zero set of
     the symbolic expansion of this determinant.  If the 
     expansion causes the determinant to SYMBOLICALLY become
     zero (ie, before even considering any values of x, y, and
     z), then there are base points -- ie, places where the
     original parametric surface has (0,0,0,0), and so the
     degree of the final equation is less than 2mn, so,
     effectively one of the rows or cols is identical to 
     another (equivalent to a row or col being 0), so the
     det is automatically zero.  In this case, we could
     still find the implicit equation by considering the
     leading minor of the matrix, but for now we will just
     assume that the surfaces are sufficiently complex.
  */

  if (base_case) {
    allocDeterminant_1D( &fdet1D_uv, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
       2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);

    allocDeterminant_1D(&ndet1D_uv, rank_row_val,
       		        rank_row_val, tpPatch1.iDeg0, tpPatch1.iDeg1);

    change_to_uv( *ndetImplicit, fdet1D_uv, tpPatch1 );
    change_to_uv( fdetImplicit, ndet1D_uv, tpPatch1 );

  /* now we have M(u,v) where each entry of M is a polynomial
     of u and v in the bernstein basis. This is stored in ndet1D_uv. */
   
  /* initialize partial derivative polynomials */
    derivative_uv(ndet1D_uv, &ndet1D_du, U);
    derivative_uv(ndet1D_uv, &ndet1D_dv, V);
/*
    derivative_uv(ndet1D_du, &ndet1D_duu, U);
    derivative_uv(ndet1D_du, &ndet1D_duv, V);
    derivative_uv(ndet1D_dv, &ndet1D_dvv, V);
*/
  }
  else {
    allocDeterminant_1D( &ndet1D_uv, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
       2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);

    change_to_uv( *ndetImplicit, ndet1D_uv, tpPatch1 );

  /* now we have M(u,v) where each entry of M is a polynomial
     of u and v in the bernstein basis. This is stored in ndet1D_uv. */
   
  /* initialize partial derivative polynomials */
    derivative_uv(ndet1D_uv, &ndet1D_du, U);
    derivative_uv(ndet1D_uv, &ndet1D_dv, V);
/*
    derivative_uv(ndet1D_du, &ndet1D_duu, U);
    derivative_uv(ndet1D_du, &ndet1D_duv, V);
    derivative_uv(ndet1D_dv, &ndet1D_dvv, V);
*/
  }
  
  /* I send all of the info that might be necessary to 
     the routines that search for starting points.
     (only tracing open components right now, need to think
     of good termination for closed components and singularities)
   */

   /* printf("Enter Curve-Surface Int\n"); */
#ifdef BEZ_CLIP
  GetBoundStPointsUV(tpPatch0, tpPatch1, &plp4DStartingPoints);
#else  
  if (getComponents(&plp4DStartingPoints,&plp4CPXStartingPoints, fdet1D_uv,
		     ndet1D_uv,tpPatch0,tpPatch1, box0, cbox1, 
                     row_perm, base_case, 0 /* c_trace */) == -1) {
    printf("Could not get Components. exiting ...\n");
    return;
  }
#endif
   /* printf("Exit Curve-Surface Int\n"); */

/*
  if (c_trace) {
    ALLOCN(addedList, POINT4D, 10);
    addedList_idx = 0;
    for( plp4CPX = plp4CPXStartingPoints; plp4CPX; plp4CPX=plp4CPXStartingPoints)
    {
      plp4CPXStartingPoints = plp4CPXStartingPoints->plp4CPXNext;
      CTrace(ndet1D_uv, ndet1D_dv, plp4CPX->p4CPX, &plp4DStartingPoints,
            tpPatch0.iDeg0, tpPatch0.iDeg1,
            tpPatch0, tpPatch1, dr, addedList, &addedList_idx);
    }
    for (i=0; i<addedList_idx; i++) {
      matched_pt = 0;
      j = 0;
      while (j < i) 
	if (EQ_4D_EPS(addedList[i], addedList[j], 1e-1)) {
	  matched_pt = 1;
	  break;
	} else j++;
      if (!matched_pt) {
        printf("CXTRACE:reached turning point (%lf,%lf,%lf,%lf)\n",
               addedList[i].dCoord[0], addedList[i].dCoord[1],
               addedList[i].dCoord[2], addedList[i].dCoord[3]);
	for (k=0; k<4; k++)
	  turn_pt4D.dCoord[k] = power_pt(addedList[i].dCoord[k]);
	addPointToList(turn_pt4D, &plp4DStartingPoints);
      }
    }
    myfree(addedList);
  }
*/

#ifdef _NEW_CTRACE
  CTrace(ndet1D_uv, fdet1D_uv, &tpPatch0, &tpPatch1,
	 &plp4DStartingPoints, base_case, row_perm);
#endif	 
  StartTracing(plp4DStartingPoints, p4DSp, fdet1D_uv,
	       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
	       tpPatch1, row_perm, base_case, cComp, cNo);

  /* Restore Scale */
  Scale = 1.0;

#ifdef __FREE_ALL
  free_ten_p(&tpPatch0);
  free_ten_p(&tpPatch1);
  if (base_case) {
    free_alloc_det_1D(&fdet1D_uv);
    free_alloc_determ(&fdetImplicit);
  }
  free_alloc_det_1D(&ndet1D_uv);
  free_list_point4D(plp4DStartingPoints);
  if (c_trace)
    free_list_point4CPX(plp4CPXStartingPoints);
  myfree(row_perm);
  myfree(col_perm);
#endif
}
#else
Param_1_in_Imp_0(tp0, tp1, box0, cbox1, p4DSp, cComp, pfOutputFile, c_trace, dr, cNo)
TENSOR_PATCH tp0, tp1;
BOUNDINGBOX box0, *cbox1;
LIST_POINT4D *p4DSp;
INT_CURVE_STRUCT **cComp;
FILE *pfOutputFile;
int c_trace, dr, *cNo;
{
  NDETERMINANT     ndetImplicit, fdetImplicit, ndet;
   /* NDETERMINANT_1D  ndet1D_duu, ndet1D_duv, ndet1D_dvv; */
  LIST_POINT4D     *plp4DStartingPoints, *plp4DComponentPoints;
  LIST_POINT4CPX   *plp4CPXStartingPoints;
  LIST_POINT4CPX   *plp4CPX;
  int              cBlue, ret_val;
  int		   NumBoundaryXsections;
  char             rgchStruct[50];
  int		   Nrows, Ncols;
  int 		   *row_perm, *col_perm;
  int		   base_case, rank_row_val;
  int		   i, j, k, matched_pt;
  POINT4D	   final_pt, *addedList, turn_pt4D;
  int		   addedList_idx;
  TENSOR_PATCH     tpPatch0, tpPatch1;

  /* I am going to scale the patches down so that my matrix entries */
  /* dont get too big. */
  tpPatch0.iDeg0 = tp0.iDeg0;
  tpPatch0.iDeg1 = tp0.iDeg1;
  tpPatch1.iDeg0 = tp1.iDeg0;
  tpPatch1.iDeg1 = tp1.iDeg1;
  allocate_tensor_patch(&tpPatch0);
  allocate_tensor_patch(&tpPatch1);
  /* Now calculating the exact Scale factor */
  Scale = -1.0;
  for (i=0; i<=tp0.iDeg0; i++)
    for (j=0; j<=tp0.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tp0.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tp0.coeffs[i][j].dCoord[k]);
	tpPatch0.coeffs[i][j].dCoord[k] = tp0.coeffs[i][j].dCoord[k];
      }
      tpPatch0.coeffs[i][j].dCoord[3] = tp0.coeffs[i][j].dCoord[3];
    }
  for (i=0; i<=tp1.iDeg0; i++)
    for (j=0; j<=tp1.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tp1.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tp1.coeffs[i][j].dCoord[k]);
	tpPatch1.coeffs[i][j].dCoord[k] = tp1.coeffs[i][j].dCoord[k];
      }
      tpPatch1.coeffs[i][j].dCoord[3] = tp1.coeffs[i][j].dCoord[3];
    }

  /* Scale the tensor patch */
  scale_tensor_patch(&tpPatch0, 1);
  scale_tensor_patch(&tpPatch1, 1);

  bernsteinToPowerBasis( tpPatch0 );

  /* tpPatch0 is now in the power basis under the reparametrization 
     of s=s/1-s, t=t/1-t, but tpPatch1 is still a bezier tensor prod
     patch */

  allocDeterminant( &ndet, tpPatch0.iDeg0, tpPatch0.iDeg1, UNFACTORED );

  implicitize( tpPatch0, ndet );

  powerToBernsteinBasis( tpPatch0 );

  allocDeterminant( &ndetImplicit, tpPatch0.iDeg0, tpPatch0.iDeg1, FINAL );
  moveFactoredToFinalDeterminant( ndetImplicit, ndet, 
				 tpPatch0.iDeg0, tpPatch0.iDeg1 );

  Nrows = ndetImplicit.cRows;
  Ncols = ndetImplicit.cCols;
  base_case = check_for_base_points(ndetImplicit, Nrows, Ncols, &rank_row_val,
				    &row_perm, &col_perm);

  if (base_case) {
    fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
    ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
    for (i=0; i<fdetImplicit.cCols; i++)
      ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

    for (i=0; i<rank_row_val; i++)
    {
      for (j=0; j<rank_row_val; j++)
      {
        fdetImplicit.coeffs[i][j].dCoord[X] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[X];
        fdetImplicit.coeffs[i][j].dCoord[Y] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
        fdetImplicit.coeffs[i][j].dCoord[Z] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
        fdetImplicit.coeffs[i][j].dCoord[W] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[W];
      }
    }
  }

  /* at this point we have the implicit form in a matrix form,
     where each element of the array signifies a vector (X,Y,Z,W)
     and the vector means that the elements are linear in the
     corresponding variables, so each entry is actually
           x*X + y*Y + z*Z + W,

     The implicit equation of the surface is the zero set of
     the symbolic expansion of this determinant.  If the 
     expansion causes the determinant to SYMBOLICALLY become
     zero (ie, before even considering any values of x, y, and
     z), then there are base points -- ie, places where the
     original parametric surface has (0,0,0,0), and so the
     degree of the final equation is less than 2mn, so,
     effectively one of the rows or cols is identical to 
     another (equivalent to a row or col being 0), so the
     det is automatically zero.  In this case, we could
     still find the implicit equation by considering the
     leading minor of the matrix, but for now we will just
     assume that the surfaces are sufficiently complex.
  */

  if (base_case) {
    allocDeterminant_1D( &fdet1D_uv, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
       2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);

    allocDeterminant_1D(&ndet1D_uv, rank_row_val,
       		        rank_row_val, tpPatch1.iDeg0, tpPatch1.iDeg1);

    change_to_uv( ndetImplicit, fdet1D_uv, tpPatch1 );
    change_to_uv( fdetImplicit, ndet1D_uv, tpPatch1 );

  /* now we have M(u,v) where each entry of M is a polynomial
     of u and v in the bernstein basis. This is stored in ndet1D_uv. */
   
  /* initialize partial derivative polynomials */
    derivative_uv(ndet1D_uv, &ndet1D_du, U);
    derivative_uv(ndet1D_uv, &ndet1D_dv, V);
/*
    derivative_uv(ndet1D_du, &ndet1D_duu, U);
    derivative_uv(ndet1D_du, &ndet1D_duv, V);
    derivative_uv(ndet1D_dv, &ndet1D_dvv, V);
*/
  }
  else {
    allocDeterminant_1D( &ndet1D_uv, 2*tpPatch0.iDeg0*tpPatch0.iDeg1,
       2*tpPatch0.iDeg0*tpPatch0.iDeg1, tpPatch1.iDeg0, tpPatch1.iDeg1);

    change_to_uv( ndetImplicit, ndet1D_uv, tpPatch1 );

  /* now we have M(u,v) where each entry of M is a polynomial
     of u and v in the bernstein basis. This is stored in ndet1D_uv. */
   
  /* initialize partial derivative polynomials */
    derivative_uv(ndet1D_uv, &ndet1D_du, U);
    derivative_uv(ndet1D_uv, &ndet1D_dv, V);
/*
    derivative_uv(ndet1D_du, &ndet1D_duu, U);
    derivative_uv(ndet1D_du, &ndet1D_duv, V);
    derivative_uv(ndet1D_dv, &ndet1D_dvv, V);
*/
  }
  
  /* I send all of the info that might be necessary to 
     the routines that search for starting points.
     (only tracing open components right now, need to think
     of good termination for closed components and singularities)
   */

   /* printf("Enter Curve-Surface Int\n"); */
#ifdef BEZ_CLIP
  GetBoundStPointsUV(tpPatch0, tpPatch1, &plp4DStartingPoints);
#else  
  if (getComponents(&plp4DStartingPoints,&plp4CPXStartingPoints, fdet1D_uv,
		     ndet1D_uv,tpPatch0,tpPatch1, box0, cbox1, 
                     row_perm, base_case, 0 /* c_trace */) == -1) {
    printf("Could not get Components. exiting ...\n");
    return;
  }
#endif
   /* printf("Exit Curve-Surface Int\n"); */

/*
  if (c_trace) {
    ALLOCN(addedList, POINT4D, 10);
    addedList_idx = 0;
    for( plp4CPX = plp4CPXStartingPoints; plp4CPX; plp4CPX=plp4CPXStartingPoints)
    {
      plp4CPXStartingPoints = plp4CPXStartingPoints->plp4CPXNext;
      CTrace(ndet1D_uv, ndet1D_dv, plp4CPX->p4CPX, &plp4DStartingPoints,
            tpPatch0.iDeg0, tpPatch0.iDeg1,
            tpPatch0, tpPatch1, dr, addedList, &addedList_idx);
    }
    for (i=0; i<addedList_idx; i++) {
      matched_pt = 0;
      j = 0;
      while (j < i) 
	if (EQ_4D_EPS(addedList[i], addedList[j], 1e-1)) {
	  matched_pt = 1;
	  break;
	} else j++;
      if (!matched_pt) {
        printf("CXTRACE:reached turning point (%lf,%lf,%lf,%lf)\n",
               addedList[i].dCoord[0], addedList[i].dCoord[1],
               addedList[i].dCoord[2], addedList[i].dCoord[3]);
	for (k=0; k<4; k++)
	  turn_pt4D.dCoord[k] = power_pt(addedList[i].dCoord[k]);
	addPointToList(turn_pt4D, &plp4DStartingPoints);
      }
    }
    myfree(addedList);
  }
*/

#ifdef _NEW_CTRACE
  CTrace(ndet1D_uv, fdet1D_uv, &tpPatch0, &tpPatch1,
	 &plp4DStartingPoints, base_case, row_perm);
#endif	 
  StartTracing(plp4DStartingPoints, p4DSp, fdet1D_uv,
	       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
	       tpPatch1, row_perm, base_case, cComp, cNo);

  /* Restore Scale */
  Scale = 1.0;

#ifdef __FREE_ALL
  free_ten_p(&tpPatch0);
  free_ten_p(&tpPatch1);
  free_alloc_determ(&ndet);
  free_alloc_determ(&ndetImplicit);
  if (base_case)
    free_alloc_determ(&fdetImplicit);
  free_list_point4D(plp4DStartingPoints);
  if (c_trace)
    free_list_point4CPX(plp4CPXStartingPoints);
  myfree(row_perm);
  myfree(col_perm);
#endif
}
#endif

StartTracing(plp4DStartingPoints, p4DSp, fdet1D_uv,
	     ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0, 
	     tpPatch1, row_perm, base_case, cComp, cNo)
LIST_POINT4D *plp4DStartingPoints, *p4DSp;
NDETERMINANT_1D fdet1D_uv, ndet1D_uv;
NDETERMINANT_1D ndet1D_du, ndet1D_dv;
TENSOR_PATCH tpPatch0, tpPatch1;
INT_CURVE_STRUCT **cComp;
int *row_perm;
int base_case, *cNo;
{
  LIST_POINT4D *tmp_ptr;
  POINT4D *uvstStartList;
  int Elmt1, Elmt2, NumStListElmts;
  RECT_BOX domain, sing_domain;
  NDETERMINANT_1D      ndet1D_Tmp, ndet1D_u, ndet1D_v;
  NDETERMINANT_1D      fdet1D_Tmp, fdet1D_u, fdet1D_v;
  int i, j, k, l;
  FILE *fp;

  Elmt1 = getNumofPoints(plp4DStartingPoints);
  Elmt2 = getNumofPoints(p4DSp);
  NumStListElmts = Elmt1 + Elmt2;

  ALLOCN(uvstStartList, POINT4D, NumStListElmts);
  tmp_ptr = plp4DStartingPoints;
  for (i=0; i<Elmt1; i++) {
    uvstStartList[i] = tmp_ptr->p4D;
    tmp_ptr = tmp_ptr->plp4DNext;
  }
  tmp_ptr = p4DSp;
  for (i=Elmt1; i<NumStListElmts; i++) {
    uvstStartList[i] = tmp_ptr->p4D;
    tmp_ptr = tmp_ptr->plp4DNext;
  }

  allocDeterminant_1D(&ndet1D_Tmp, ndet1D_uv.cRows, ndet1D_uv.cCols,
		      tpPatch1.iDeg0, tpPatch1.iDeg1);
  allocDeterminant_1D(&ndet1D_u, ndet1D_uv.cRows, ndet1D_uv.cCols,
                      0, tpPatch1.iDeg0);
  allocDeterminant_1D(&ndet1D_v, ndet1D_uv.cRows, ndet1D_uv.cCols,
                      0, tpPatch1.iDeg1);
  if (base_case) {
    allocDeterminant_1D(&fdet1D_Tmp, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        tpPatch1.iDeg0, tpPatch1.iDeg1);
    allocDeterminant_1D(&fdet1D_u, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg0);
    allocDeterminant_1D(&fdet1D_v, fdet1D_uv.cRows, fdet1D_uv.cCols,
                        0, tpPatch1.iDeg1);
  }
  /* copy ndet1D_uv to ndet1D_Tmp */
  copyDeterminant_1D(ndet1D_Tmp, ndet1D_uv);

  /* convert bernstein basis to power basis for each entry of Tmp */
  for(i = 0; i < ndet1D_Tmp.cRows; i++)
    for(j = 0; j < ndet1D_Tmp.cCols; j++)
      bernsteinToPowerBasis_1D(ndet1D_Tmp.poly[i][j]);

  if (base_case) {
    /* copy fdet1D_uv to fdet1D_Tmp */
    copyDeterminant_1D(fdet1D_Tmp, fdet1D_uv);

    /* convert bernstein basis to power basis for each entry of Tmp */
    for(i = 0; i < fdet1D_Tmp.cRows; i++)
      for(j = 0; j < fdet1D_Tmp.cCols; j++)
        bernsteinToPowerBasis_1D(fdet1D_Tmp.poly[i][j]);
  }

  domain.i2Dbox[0].low = 0.0;
  domain.i2Dbox[0].high = 1.0;
  domain.i2Dbox[1].low = 0.0;
  domain.i2Dbox[1].high = 1.0;

  sing_domain.i2Dbox[0].low = 0.0;
  sing_domain.i2Dbox[0].high = 0.0;
  sing_domain.i2Dbox[1].low = 0.0;
  sing_domain.i2Dbox[1].high = 0.0;

  *cNo = 0;
  doDomainDecomp(uvstStartList, NumStListElmts, domain, &sing_domain,
		 fdet1D_uv, ndet1D_uv, ndet1D_du, ndet1D_dv, tpPatch0, 
		 tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
		 fdet1D_Tmp, fdet1D_u, fdet1D_v,
		 row_perm, base_case, cComp, cNo, 0);

#ifdef __FREE_ALL
  free_alloc_det_1D(&ndet1D_Tmp);
  free_alloc_det_1D(&ndet1D_u);
  free_alloc_det_1D(&ndet1D_v);
  if (base_case) {
    free_alloc_det_1D(&fdet1D_Tmp);
    free_alloc_det_1D(&fdet1D_u);
    free_alloc_det_1D(&fdet1D_v);
  }
  myfree(uvstStartList);
#endif
}

int getNumUList(uvstList, NumPoints, endUVal, domain)
POINT4D *uvstList;
int NumPoints;
double **endUVal;
RECT_BOX *domain;
{
  int i, j, k;
  double previous_max;
  int numUIntervals;

  if (NumPoints == 0) return 0;

/* I am mallocing NumPoints+2 because I think we can prove that there
   cannot be more than NumPoints+2 v-intervals */
  ALLOCN(*endUVal, double, NumPoints+2);
   previous_max = (*endUVal)[0] = domain->i2Dbox[0].low;
   /* previous_max = (*endUVal)[0] = uvstList[0].dCoord[U]; */

  for (k=i=0; i<NumPoints; i++) 
    if (uvstList[i].dCoord[U] > previous_max) {
      k++;
      previous_max = (*endUVal)[k] = uvstList[i].dCoord[U];
    }
  if (previous_max < domain->i2Dbox[0].high-EPSILON)
    (*endUVal)[++k] = domain->i2Dbox[0].high;

  numUIntervals = k;
  return numUIntervals;
}
 
int getNumVList(uvstList, NumPoints, endVVal, domain)
POINT4D *uvstList;
int NumPoints;
double **endVVal;
RECT_BOX *domain;
{
  int i, j, k;
  double previous_max;
  int numVIntervals;

  if (NumPoints == 0) return 0;

/* I am mallocing NumPoints+2 because I think we can prove that there
   cannot be more than NumPoints+2 v-intervals */
  ALLOCN(*endVVal, double, NumPoints+2);
   previous_max = (*endVVal)[0] = domain->i2Dbox[1].low;
   /* previous_max = (*endVVal)[0] = uvstList[0].dCoord[V]; */

  for (k=i=0; i<NumPoints; i++) 
    if (uvstList[i].dCoord[V] > previous_max) {
      k++;
      previous_max = (*endVVal)[k] = uvstList[i].dCoord[V];
    }
  if (previous_max < domain->i2Dbox[1].high-EPSILON)
    (*endVVal)[++k] = domain->i2Dbox[1].high;

  numVIntervals = k;
  return numVIntervals;
}
 
doDomainDecomp(uvstList, NumPoints, domain, s_dom, fdet1D_uv,
               ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
               tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
               fdet1D_Tmp, fdet1D_u, fdet1D_v,
               row_perm, base_case, cComp, compNo, first_time)
POINT4D *uvstList;
int NumPoints;
RECT_BOX domain, *s_dom;
NDETERMINANT_1D fdet1D_uv, ndet1D_uv;
NDETERMINANT_1D ndet1D_du, ndet1D_dv;
TENSOR_PATCH tpPatch0, tpPatch1;
INT_CURVE_STRUCT **cComp;
NDETERMINANT_1D      ndet1D_Tmp, ndet1D_u, ndet1D_v;
NDETERMINANT_1D      fdet1D_Tmp, fdet1D_u, fdet1D_v;
int *row_perm;
int base_case, *compNo, first_time;
{
  static int NumEntryPoints = 0; /* this static variable keeps track
				    of the number of times domain
				    decomposition enters the 4 point
				    part of the code */
  POINT4D p4DEnd;
  LIST_POINT4D *pplp4DOutput;
  double min_dom_width, max_dom_width, udiff, vdiff;
  double s_dom_domain_dist, ul_diff, uh_diff, vl_diff, vh_diff;
  double trace_step_size;
  int numVList, numUList;
  TRACE_INFO *t_info;
  double *endVVal, *endUVal, RetVal;
  double *nDPt, **nnDirVec;
  double ftol, bisect_pt, d2D_distance;
  int ord, NumIter;
  int                  num_points, num_cpoints;
  POINT4D              *pdPoints, *newEndPt;
  POINT2D              tpt2_1, tpt2_2;
  POINT4CPX            *pcpxPoints = NULL;
  float vert[2];
  int i, j, k, m, n, cIndex, ptIndex;
  int AnyOtherIndex, retCode;
  int tmpCompNo, traceRet;

  udiff = domain.i2Dbox[0].high - domain.i2Dbox[0].low;
  vdiff = domain.i2Dbox[1].high - domain.i2Dbox[1].low;
  min_dom_width = (udiff < vdiff) ? udiff : vdiff;
  max_dom_width = (udiff > vdiff) ? udiff : vdiff;

  m = tpPatch0.iDeg0;
  n = tpPatch0.iDeg1;
  i = (tpPatch1.iDeg0 > tpPatch1.iDeg1)? tpPatch1.iDeg0: tpPatch1.iDeg1;
  ALLOCN(pdPoints, POINT4D, i*ndet1D_uv.cRows);

/* The following code just removes some startpoints which are very
   close to some others. This causes problems in domain decomposition.
*/
  if (NumPoints > 2) {
    j=0; 
    while (j<NumPoints) {
      k = j+1;
      while (k < NumPoints) {
	if (EQ_4D_EPS(uvstList[j], uvstList[k], 1e-2)) {
	  for (i=k+1; i<NumPoints; i++) 
	    uvstList[i-1] = uvstList[i];
	  NumPoints--;
	} else k++;
      }
      j++;
    }
  }

#ifdef __GL
  winset(giduv);
  color(WHITE);
  bgnline();
  for (i=0; i<=udiff*0.85*UVST_SIZE; i++) {
    vert[0] = i + (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=vdiff*0.85*UVST_SIZE; i++) {
    vert[0] = (0.1+domain.i2Dbox[0].high*0.85)*UVST_SIZE;
    vert[1] = i + (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=udiff*0.85*UVST_SIZE; i++) {
    vert[0] = i + (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = (0.1+domain.i2Dbox[1].high*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  bgnline();
  for (i=0; i<=vdiff*0.85*UVST_SIZE; i++) {
    vert[0] = (0.1+domain.i2Dbox[0].low*0.85)*UVST_SIZE;
    vert[1] = i + (0.1+domain.i2Dbox[1].low*0.85)*UVST_SIZE;
    v2f(vert);
  }
  endline();
  winset(gidxyz);
#endif

/* Actually FIRST_TIME is a wrong name. If it is 1, it actually
   meant that it was called inside doDomain.. itself */
  if (NumPoints == 0) return;
  if (NumPoints < 2 && first_time) return;
  if (NumPoints == 2) {
#ifdef TRIVTRACE
    if (max_dom_width <= 0.06) {
      TrivialTrace(&pplp4DOutput, uvstList[0], uvstList[1], tpPatch1);
      return;
    }
#endif
    if (FP_EQ_EPS(min_dom_width, 0.0, 1e-3)) return;
    if (min_dom_width < 0.002) trace_step_size = min_dom_width/2.0;
    else if (max_dom_width <= 0.2) trace_step_size = min_dom_width/4.0;
     /* else if (max_dom_width <= 0.2) trace_step_size = 0.008; */
    else trace_step_size = Step_Size_Out;

    ALLOCN(newEndPt, POINT4D, 2);
    cIndex = ptIndex = -1;
    tmpCompNo = 0;
    if (*compNo == 0) {
      cIndex = 0;
      if (uvstList[0].dCoord[V] < uvstList[1].dCoord[V])
        ptIndex = 0;
      else if (FP_EQ_EPS(uvstList[1].dCoord[V],uvstList[0].dCoord[V],1e-6)){
        if (uvstList[0].dCoord[U] < uvstList[1].dCoord[U])
          ptIndex = 0;
        else ptIndex = 1;
      }
      else ptIndex = 1;
      tmpCompNo = 1;
      AnyOtherIndex = -1;
      (*cComp)[cIndex].endPoints[0] = uvstList[ptIndex];
      (*cComp)[cIndex].endPoints[1] = uvstList[1-ptIndex];
    }
    else {
      setUpTraceOrder(cComp, uvstList, *compNo, &cIndex,
                      &ptIndex, &AnyOtherIndex, &retCode, &newEndPt);
      if ((cIndex == -1) && (ptIndex == -1)) {
        cIndex = *compNo;
        if (uvstList[0].dCoord[V] < uvstList[1].dCoord[V])
          ptIndex = 0;
        else if (uvstList[1].dCoord[V] == uvstList[0].dCoord[V]) {
          if (uvstList[0].dCoord[U] < uvstList[1].dCoord[U])
            ptIndex = 0;
          else ptIndex = 1;
        }
        else ptIndex = 1;
        tmpCompNo = 1;
        AnyOtherIndex = -1;
        if (cIndex == MAX_COMP_PER_PATCH) {
          printf("Too many components. Halting Intersection.\n");
          return;
        }
        (*cComp)[cIndex].endPoints[0] = uvstList[ptIndex];
        (*cComp)[cIndex].endPoints[1] = uvstList[1-ptIndex];
      }
      else tmpCompNo = 2;
    }
     
/* The following few lines of code is added to take care of
   tracing when the endpoints are closer than the step size */
    tpt2_1.u = uvstList[0].dCoord[0];
    tpt2_1.v = uvstList[0].dCoord[1];
    tpt2_2.u = uvstList[1].dCoord[0];
    tpt2_2.v = uvstList[1].dCoord[1];
    d2D_distance = dist2D(&tpt2_1, &tpt2_2);
    d2D_distance = sqrt(d2D_distance);
    if (d2D_distance < 2*trace_step_size) 
      /*
      add_list_point(&((*cComp)[cIndex].Curve), uvstList[ptIndex],
		     tpPatch1, 0, 0, 1);
      add_list_point(&((*cComp)[cIndex].Curve), uvstList[1-ptIndex],
		     tpPatch1, 0, 0, 1);
      */
      trace_step_size = 0.25 * d2D_distance; /* 0.7 was chosen because
					      sqrt(2) is minimum */

    tpt2_1.u = fabs(tpt2_1.u - tpt2_2.u);
    tpt2_1.v = fabs(tpt2_1.v - tpt2_2.v);
    if (!(traceRet = Trace(ndet1D_uv, ndet1D_du, ndet1D_dv,
			   uvstList[ptIndex], domain, &p4DEnd,
			   &((*cComp)[cIndex].Curve), tpPatch0,
			   tpPatch1, 0, 0, fdet1D_uv, row_perm,
			   base_case, 1, trace_step_size,
			   tmpCompNo, compNo, &tpt2_1)) ) { 
      if (!EQ_4D(p4DEnd, uvstList[1-ptIndex]))
        add_list_point(&((*cComp)[cIndex].Curve), uvstList[1-ptIndex],
                       tpPatch1, 0, 0, 1);
      if (tmpCompNo == 2) {
        (*cComp)[cIndex].endPoints[0] = newEndPt[0];
        (*cComp)[cIndex].endPoints[1] = newEndPt[1];
      }
    }
    myfree(newEndPt);

    if (AnyOtherIndex != -1 && !traceRet)
      MergeCurveLists(cComp, cIndex, AnyOtherIndex, retCode, compNo);
    return;
  }
#ifdef OLD_DOM_DECOMP
  if (NumPoints == 4) {
   NumEntryPoints++;
   if (max_dom_width <= 0.06) {
    printf("Looks like we are either approaching a singularity\n");
    printf("or have two very close components\n");
    /* TRACE SINGULARITY PROPERLY WITH VERY SMALL STEP SIZE*/
    pow_tp0 = tpPatch0;
    pow_tp1 = tpPatch1;
    ord = 4;
    ftol = 0.01;
    ALLOCN(nDPt, double, 4);
    ALLOCMAT(nnDirVec, 4, 4, DOUBLE);
    RetVal = minim_func_for_sing(nDPt);
    nDPt[0] = uvstList[0].dCoord[0];
    nDPt[1] = uvstList[0].dCoord[1];
    nDPt[2] = uvstList[0].dCoord[2];
    nDPt[3] = uvstList[0].dCoord[3];
/*
    nDPt[0] = (domain.i2Dbox[0].low + domain.i2Dbox[0].high) / 2.0;
    nDPt[1] = (domain.i2Dbox[1].low + domain.i2Dbox[1].high) / 2.0;
*/
    for (i=0; i<4; i++)
    for (j=0; j<4; j++)
      nnDirVec[i][j] = (i == j) ? 1.0 : 0.0;
    powell(nDPt, nnDirVec, ord, ftol, &NumIter, &RetVal, minim_func_for_sing);
   printf("Values returned from POWELL are %g  %g  %g\n", nDPt[0], nDPt[1], RetVal);
    if (FP_EQ_EPS(RetVal, 0.0, 1.0e-10) &&
        nDPt[0] > (domain.i2Dbox[0].low + 1e-3) &&
        nDPt[0] < (domain.i2Dbox[0].high - 1e-3) &&
        nDPt[1] > (domain.i2Dbox[1].low + 1e-3) &&
        nDPt[1] < (domain.i2Dbox[1].high - 1e-3)) {
      printf("Actually Located a SINGULARITY at U = %g and V = %g\n", nDPt[0], nDPt[1]);
      ALLOCN(t_info, TRACE_INFO, 4); /* 4 is because we create four
                                        new subdomains around the
                                        singular point */
      for (i=0; i<4; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }
      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[0].high = nDPt[0];
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[0].domain_def.i2Dbox[1].high = nDPt[1];
      t_info[1].domain_def.i2Dbox[0].low = nDPt[0];
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[1].high = nDPt[1];
      t_info[2].domain_def.i2Dbox[0].low = nDPt[0];
      t_info[2].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[2].domain_def.i2Dbox[1].low = nDPt[1];
      t_info[2].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
      t_info[3].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[3].domain_def.i2Dbox[0].high = nDPt[0];
      t_info[3].domain_def.i2Dbox[1].low = nDPt[1];
      t_info[3].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
      for (i=0; i<NumPoints; i++)
        if (uvstList[i].dCoord[0] < nDPt[0]) {
          if (uvstList[i].dCoord[1] < nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else if (uvstList[i].dCoord[1] > nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 3, &t_info, domain, 0);
        } else if (uvstList[i].dCoord[0] > nDPt[0]) {
          if (uvstList[i].dCoord[1] < nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
          else if (uvstList[i].dCoord[1] > nDPt[1])
            addPointsToTinfoArray(uvstList[i], -1, 2, &t_info, domain, 0);
        }
/* Now must do curve-surface Xsection with isolines passing thru sing. point
*/
      compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, nDPt[1]);
      if (base_case)
        compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, nDPt[1]);
      find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,U,power_pt(nDPt[1]),
                        fdet1D_u, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      for (i=0; i < num_points; i++)
        for (j=0; j<4; j++)
          addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                            t_info[j].domain_def, 1);

      compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, nDPt[0]);
      if (base_case)
        compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, nDPt[0]);
      find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,V,power_pt(nDPt[0]),
                        fdet1D_v, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      for (i = 0; i < num_points; i++) {
        if (!FP_EQ_EPS(inverse_power_pt(pdPoints[i].dCoord[1]), nDPt[1],
1e-5))
          for (j=0; j<4; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
      }

      s_dom->i2Dbox[0].low = s_dom->i2Dbox[0].high = 0.0;
      s_dom->i2Dbox[1].low = s_dom->i2Dbox[1].high = 0.0;
      for (i=0; i<4; i++)
        if (t_info[i].Npts > 1)
          doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                         t_info[i].domain_def, s_dom, fdet1D_uv,
                         ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                         tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                         fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm,
                         base_case, cComp, compNo, 1);

#ifdef __FREE_ALL
     for (i=0; i<4; i++)
       myfree(t_info[i].ptList);
     myfree(t_info);
     myfree(nDPt);
     FREEMAT(nnDirVec, 2, 2);
     myfree(pdPoints);
#endif
     return;
    }
#ifdef __FREE_ALL
    myfree(nDPt);
    FREEMAT(nnDirVec, 2, 2);
#endif
   } else {
    if (NumEntryPoints < 3) {
      ul_diff = (domain.i2Dbox[0].low - s_dom->i2Dbox[0].low);
      uh_diff = (domain.i2Dbox[0].high - s_dom->i2Dbox[0].high);
      vl_diff = (domain.i2Dbox[1].low - s_dom->i2Dbox[1].low);
      vh_diff = (domain.i2Dbox[1].high - s_dom->i2Dbox[1].high);
      s_dom_domain_dist = ul_diff*ul_diff + uh_diff*uh_diff +
                        vl_diff*vl_diff + vh_diff*vh_diff;
    }
    if ((NumEntryPoints >= 3) || FP_EQ_EPS(s_dom_domain_dist, 0.0,
					   0.005)) {
      NumEntryPoints = 0;
      printf("Doing Bisection\n");
      ALLOCN(t_info, TRACE_INFO, 2); /* 2 is because we create two
                                        new subdomains by
                                        bisection */
      for (i=0; i<2; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;

      if (udiff > vdiff) {
        bisect_pt = (domain.i2Dbox[0].low +
                     domain.i2Dbox[0].high) / 2.0;
        t_info[0].domain_def.i2Dbox[0].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
        t_info[1].domain_def.i2Dbox[0].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[0], bisect_pt)) continue;
          if (uvstList[i].dCoord[0] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, bisect_pt);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(bisect_pt),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      else {
        bisect_pt = (domain.i2Dbox[1].low +
                     domain.i2Dbox[1].high) / 2.0;
        t_info[0].domain_def.i2Dbox[1].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
        t_info[1].domain_def.i2Dbox[1].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[1], bisect_pt)) continue;
          if (uvstList[i].dCoord[1] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, bisect_pt);
        find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,U,power_pt(bisect_pt),
                          fdet1D_u, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      for (i=0; i<2; i++) {
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
      }

      for (i=0; i<2; i++)
        myfree(t_info[i].ptList);
      myfree(t_info);
      return ;
    } else (*s_dom) = domain;
   }
  }

  qsort(uvstList, NumPoints, sizeof(POINT4D), comparV);
  numVList = getNumVList(uvstList, NumPoints, &endVVal, &domain);
  if (numVList <= 0) {
    printf("Something Wrong in Domain Decomp.\n");
    return;
  }
  if (numVList > 1) {
    ALLOCN(t_info, TRACE_INFO, numVList);
    for (i=0; i<numVList; i++) {
      t_info[i].Npts = 0;
      ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
    }

    compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, endVVal[0]);
    if (base_case)
      compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, endVVal[0]);
    find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                      pcpxPoints,U,power_pt(endVVal[0]),
                      fdet1D_u, row_perm, m, n, base_case,
                      tpPatch0, tpPatch1, 0);

    for (i = 0; i < num_points; i++)
      addPointsToTinfoArray(pdPoints[i], -1, 0, &t_info, domain, 1);
    for (i=0; i<numVList; i++) {
      t_info[i].domain_def.i2Dbox[U].low = domain.i2Dbox[U].low;
      t_info[i].domain_def.i2Dbox[U].high = domain.i2Dbox[U].high;
      t_info[i].domain_def.i2Dbox[V].low = endVVal[i];
      t_info[i].domain_def.i2Dbox[V].high = endVVal[i+1];

      compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, endVVal[i+1]);
      if (base_case)
        compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, endVVal[i+1]);
      find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,U,power_pt(endVVal[i+1]),
                        fdet1D_u, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      if (i == numVList-1) {
        for (j = 0; j < num_points; j++)
          addPointsToTinfoArray(pdPoints[j], -1, i, &t_info, domain, 1);
      } else {
        for (j = 0; j < num_points; j++)
          addPointsToTinfoArray(pdPoints[j], 0, i, &t_info, domain, 1);
      }
      if (t_info[i].Npts >= 2)
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
    }
  } else { /* Case when NumVList = 1 and we are dividing w.r.t U */
    qsort(uvstList, NumPoints, sizeof(POINT4D), comparU);
    numUList = getNumUList(uvstList, NumPoints, &endUVal, &domain);
    if (numUList <= 0) {
      printf("Something Wrong in Domain Decomp.\n");
      return;
    }
    if (numUList > 1) {
      ALLOCN(t_info, TRACE_INFO, numUList);
      for (i=0; i<numUList; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, endUVal[0]);
      if (base_case)
        compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, endUVal[0]);
      find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                        pcpxPoints,V,power_pt(endUVal[0]),
                        fdet1D_v, row_perm, m, n, base_case,
                        tpPatch0, tpPatch1, 0);
      for (i = 0; i < num_points; i++)
        addPointsToTinfoArray(pdPoints[i], -1, 0, &t_info, domain, 1);
      for (i=0; i<numUList; i++) {
        t_info[i].domain_def.i2Dbox[U].low = endUVal[i];
        t_info[i].domain_def.i2Dbox[U].high = endUVal[i+1];
        t_info[i].domain_def.i2Dbox[V].low = domain.i2Dbox[V].low;
        t_info[i].domain_def.i2Dbox[V].high = domain.i2Dbox[V].high;

        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, endUVal[i+1]);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, endUVal[i+1]);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(endUVal[i+1]),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        if (i == numUList-1) {
          for (j = 0; j < num_points; j++)
            addPointsToTinfoArray(pdPoints[j], -1, i, &t_info, domain, 1);
        } else {
          for (j = 0; j < num_points; j++)
            addPointsToTinfoArray(pdPoints[j], 0, i, &t_info, domain, 1);
        }
        if (t_info[i].Npts >= 2)
          doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                         t_info[i].domain_def, s_dom, fdet1D_uv,
                         ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                         tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                         fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm,
                         base_case,cComp, compNo, 1);
      }
    }
    else { /* Case when numUList and numVList are 1 => Cannot divide
              the domain further */
#ifndef DEBUG_PRINT_POLY
      printf("numUList and numVList are 1: So doing bisection\n");
#endif
      ALLOCN(t_info, TRACE_INFO, 2); /* 2 is because we create two
                                        new subdomains by
                                        bisection */
      for (i=0; i<2; i++) {
        t_info[i].Npts = 0;
        ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
      }

      t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
      t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
      t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
      t_info[1].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;

      if (udiff > vdiff) {
        bisect_pt = (domain.i2Dbox[0].low +
                     domain.i2Dbox[0].high) / 2.0;
        t_info[0].domain_def.i2Dbox[0].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
        t_info[1].domain_def.i2Dbox[0].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[0], bisect_pt)) continue;
          if (uvstList[i].dCoord[0] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, bisect_pt);
        find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,V,power_pt(bisect_pt),
                          fdet1D_v, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      else {
        bisect_pt = (domain.i2Dbox[1].low +
                     domain.i2Dbox[1].high) / 2.0;
        t_info[0].domain_def.i2Dbox[1].high = bisect_pt;
        t_info[0].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
        t_info[1].domain_def.i2Dbox[1].low = bisect_pt;
        t_info[1].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;

        for (i=0; i<NumPoints; i++) {
          if (FP_EQ(uvstList[i].dCoord[1], bisect_pt)) continue;
          if (uvstList[i].dCoord[1] < bisect_pt)
            addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
          else
            addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
        }

        compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, bisect_pt);
        if (base_case)
          compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, bisect_pt);
        find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
                          pcpxPoints,U,power_pt(bisect_pt),
                          fdet1D_u, row_perm, m, n, base_case,
                          tpPatch0, tpPatch1, 0);
        for (i = 0; i < num_points; i++) {
          for (j=0; j<2; j++)
            addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
                                              t_info[j].domain_def, 1);
        }
      }
      for (i=0; i<2; i++) {
        doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
                       t_info[i].domain_def, s_dom, fdet1D_uv,
                       ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
                       tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
                       fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
                       cComp, compNo, 1);
      }

      for (i=0; i<2; i++)
        myfree(t_info[i].ptList);
      myfree(t_info);
    }
#ifdef __FREE_ALL
  myfree(endUVal);
#endif
  }
#ifdef __FREE_ALL
  myfree(endVVal);
  myfree(pdPoints);
#endif
  printf("Doing Bisection\n");
#else /* OLD_DOM_DECOMP */
  ALLOCN(t_info, TRACE_INFO, 2); /* 2 is because we create two
				    new subdomains by
				    bisection */
  for (i=0; i<2; i++) {
    t_info[i].Npts = 0;
    ALLOCN(t_info[i].ptList, POINT4D, TINFO_PT_LIST_SIZE);
  }
  
  t_info[0].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
  t_info[0].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
  t_info[1].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
  t_info[1].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
  
  if (udiff > vdiff) {
    bisect_pt = (domain.i2Dbox[0].low +
		 domain.i2Dbox[0].high) / 2.0;
    t_info[0].domain_def.i2Dbox[0].high = bisect_pt;
    t_info[0].domain_def.i2Dbox[1].high = domain.i2Dbox[1].high;
    t_info[1].domain_def.i2Dbox[0].low = bisect_pt;
    t_info[1].domain_def.i2Dbox[1].low = domain.i2Dbox[1].low;
    
    for (i=0; i<NumPoints; i++) {
      if (FP_EQ(uvstList[i].dCoord[0], bisect_pt)) continue;
      if (uvstList[i].dCoord[0] < bisect_pt)
	addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
      else
	addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
    }
    
    compute_curve_1D(ndet1D_v, ndet1D_Tmp, V, bisect_pt);
    if (base_case)
      compute_curve_1D(fdet1D_v, fdet1D_Tmp, V, bisect_pt);
    find_start_points(ndet1D_v,&num_points,pdPoints,&num_cpoints,
		      pcpxPoints,V,power_pt(bisect_pt),
		      fdet1D_v, row_perm, m, n, base_case,
		      tpPatch0, tpPatch1, 0);
    for (i = 0; i < num_points; i++) {
      for (j=0; j<2; j++)
	addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
			      t_info[j].domain_def, 1);
    }
  }
  else {
    bisect_pt = (domain.i2Dbox[1].low +
		 domain.i2Dbox[1].high) / 2.0;
    t_info[0].domain_def.i2Dbox[1].high = bisect_pt;
    t_info[0].domain_def.i2Dbox[0].high = domain.i2Dbox[0].high;
    t_info[1].domain_def.i2Dbox[1].low = bisect_pt;
    t_info[1].domain_def.i2Dbox[0].low = domain.i2Dbox[0].low;
    
    for (i=0; i<NumPoints; i++) {
      if (FP_EQ(uvstList[i].dCoord[1], bisect_pt)) continue;
      if (uvstList[i].dCoord[1] < bisect_pt)
	addPointsToTinfoArray(uvstList[i], -1, 0, &t_info, domain, 0);
      else
	addPointsToTinfoArray(uvstList[i], -1, 1, &t_info, domain, 0);
    }
    
    compute_curve_1D(ndet1D_u, ndet1D_Tmp, U, bisect_pt);
    if (base_case)
      compute_curve_1D(fdet1D_u, fdet1D_Tmp, U, bisect_pt);
    find_start_points(ndet1D_u,&num_points,pdPoints,&num_cpoints,
		      pcpxPoints,U,power_pt(bisect_pt),
		      fdet1D_u, row_perm, m, n, base_case,
		      tpPatch0, tpPatch1, 0);
    for (i = 0; i < num_points; i++) {
      for (j=0; j<2; j++)
	addPointsToTinfoArray(pdPoints[i], -1, j, &t_info,
			      t_info[j].domain_def, 1);
    }
  }
  for (i=0; i<2; i++) {
    doDomainDecomp(t_info[i].ptList, t_info[i].Npts,
		   t_info[i].domain_def, s_dom, fdet1D_uv,
		   ndet1D_uv,ndet1D_du, ndet1D_dv, tpPatch0,
		   tpPatch1, ndet1D_Tmp, ndet1D_u, ndet1D_v,
		   fdet1D_Tmp, fdet1D_u, fdet1D_v, row_perm, base_case,
		   cComp, compNo, 1);
  }
  
  for (i=0; i<2; i++)
    myfree(t_info[i].ptList);
  myfree(t_info);
#endif /* OLD_DOM_DECOMP */ 
}

addPointsToTinfoArray(point4D_val, flag, index, ppArray, domain, inv)
POINT4D point4D_val;
int flag, index, inv;
TRACE_INFO **ppArray;
RECT_BOX domain;
{
  int i;
  double ulow, uhigh, vlow, vhigh;
 
  if ((flag != 0) && (flag != -1)) {
    printf("Incorrect Value of FLAG\n");
    return;
  }

  ulow = domain.i2Dbox[U].low;
  uhigh = domain.i2Dbox[U].high;
  vlow = domain.i2Dbox[V].low;
  vhigh = domain.i2Dbox[V].high;

  if (inv)
    for(i=0; i < 4; i++)
      point4D_val.dCoord[i] = inverse_power_pt(point4D_val.dCoord[i]);

  if ((point4D_val.dCoord[U] >= (ulow-NEW_EPS)) && (point4D_val.dCoord[U] <= (uhigh+NEW_EPS)) &&
      (point4D_val.dCoord[V] >= (vlow-NEW_EPS)) && (point4D_val.dCoord[V] <= (vhigh+NEW_EPS)) &&
      (point4D_val.dCoord[S] >= (-NEW_EPS)) && (point4D_val.dCoord[S] <= (1.0+NEW_EPS)) &&
      (point4D_val.dCoord[T] >= (-NEW_EPS)) && (point4D_val.dCoord[T] <= (1.0+NEW_EPS)))
  {
    (*ppArray)[index].ptList[(*ppArray)[index].Npts] = point4D_val;
    (*ppArray)[index].Npts += 1;
    if ((*ppArray)[index].Npts == 8) 
      printf("OverFlow in TINFO array: Seg Fault Imminent\n");
    if (flag == 0) {
      (*ppArray)[index+1].ptList[(*ppArray)[index+1].Npts] = point4D_val;
      (*ppArray)[index+1].Npts += 1;
      if ((*ppArray)[index+1].Npts == 8) 
        printf("OverFlow in TINFO array: Seg Fault Imminent\n");
    }
  }
}

#ifdef BEZ_CLIP
GetBoundStPointsUV(tp0, tp1, plp4DStartingPoints)
TENSOR_PATCH tp0, tp1;
LIST_POINT4D **plp4DStartingPoints;
{
  int i, j;
  int numhits;
  LIST_POINT4D *new_point;
  XYZ_SPLINE_TYPE *spl;
  double uhits[20], shits[20], thits[20];

  *plp4DStartingPoints = NULL;
  /* I am going to get curves from tp1 and do clipping with tp0 */

  /* tp1 - u = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp1, 0, 0.0, spl);
  curve_surface_intersect(&tp0, spl, CLIP_TOL, CLIP_TOL, &numhits, uhits,
			  shits, thits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = 0.0;
    new_point->p4D.dCoord[V] = uhits[i];
    new_point->p4D.dCoord[S] = shits[i];
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp1 - u = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp1, 0, 1.0, spl);
  curve_surface_intersect(&tp0, spl, CLIP_TOL, CLIP_TOL, &numhits, uhits,
			  shits, thits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = 1.0;
    new_point->p4D.dCoord[V] = uhits[i];
    new_point->p4D.dCoord[S] = shits[i];
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp1 - v = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp1, 1, 0.0, spl);
  curve_surface_intersect(&tp0, spl, CLIP_TOL, CLIP_TOL, &numhits, uhits,
			  shits, thits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = 0.0;
    new_point->p4D.dCoord[S] = shits[i];
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp1 - v = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp1, 1, 1.0, spl);
  curve_surface_intersect(&tp0, spl, CLIP_TOL, CLIP_TOL, &numhits, uhits,
			  shits, thits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = 1.0;
    new_point->p4D.dCoord[S] = shits[i];
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);
}
#endif

#ifdef _CALC_IMPLICIT
GetInsideStPointsUV(tp0, tp1, ndetImplicit, imp_scale, box1, cbox0, plp4DStartingPoints)
TENSOR_PATCH tp0, tp1;
double imp_scale;
NDETERMINANT *ndetImplicit;
BOUNDINGBOX box1, *cbox0;
LIST_POINT4D **plp4DStartingPoints;
{
#ifdef BEZ_CLIP
  int i, j;
  int numhits;
  LIST_POINT4D *new_point;
  XYZ_SPLINE_TYPE *spl;
  double thits[20], uhits[20], vhits[20];

  *plp4DStartingPoints = NULL;
  /* I am going to get curves from tp0 and do clipping with tp1 */

  /* tp0 - s = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 0, 0.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = 0.0;
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - s = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 0, 1.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = 1.0;
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - t = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 1, 0.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = thits[i];
    new_point->p4D.dCoord[T] = 0.0;
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - t = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 1, 1.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = thits[i];
    new_point->p4D.dCoord[T] = 1.0;
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

#else
  NDETERMINANT     fdetImplicit;
  NDETERMINANT_1D  n_uv, f_uv;
  LIST_POINT4CPX   *plp4CPXStartingPoints;
  int              cBlue, ret_val;
  char             rgchStruct[50];
  int              Nrows, Ncols;
  int              *row_perm, *col_perm;
  int              base_case, rank_row_val;
  int              i, j, k;
  POINT4D	   final_pt;
  TENSOR_PATCH     tpPatch0, tpPatch1;

  /* I am going to scale the patches down so that my matrix entries */
  /* dont get too big. */
  tpPatch0.iDeg0 = tp0.iDeg0;
  tpPatch0.iDeg1 = tp0.iDeg1;
  tpPatch1.iDeg0 = tp1.iDeg0;
  tpPatch1.iDeg1 = tp1.iDeg1;
  allocate_tensor_patch(&tpPatch0);
  allocate_tensor_patch(&tpPatch1);
  /* Now calculating the exact Scale factor */
  Scale = imp_scale;
  for (i=0; i<=tp0.iDeg0; i++)
    for (j=0; j<=tp0.iDeg1; j++)
      for (k=X; k<=W; k++)
	tpPatch0.coeffs[i][j].dCoord[k] = tp0.coeffs[i][j].dCoord[k];
  for (i=0; i<=tp1.iDeg0; i++)
    for (j=0; j<=tp1.iDeg1; j++)
      for (k=X; k<=W; k++)
	tpPatch1.coeffs[i][j].dCoord[k] = tp1.coeffs[i][j].dCoord[k];

  /* Scale the tensor patch */
  scale_tensor_patch(&tpPatch0, 1);
  scale_tensor_patch(&tpPatch1, 1);

  Nrows = ndetImplicit->cRows;
  Ncols = ndetImplicit->cCols;
  base_case = check_for_base_points(*ndetImplicit, Nrows, Ncols, &rank_row_val,
                                    &row_perm, &col_perm);

  if (base_case) {
    fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
    ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
    for (i=0; i<fdetImplicit.cCols; i++)
      ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

    for (i=0; i<rank_row_val; i++)
    {
      for (j=0; j<rank_row_val; j++)
      {
        fdetImplicit.coeffs[i][j].dCoord[X] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[X];
        fdetImplicit.coeffs[i][j].dCoord[Y] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
        fdetImplicit.coeffs[i][j].dCoord[Z] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
        fdetImplicit.coeffs[i][j].dCoord[W] =
                  ndetImplicit->coeffs[row_perm[i]][col_perm[j]].dCoord[W];
      }
    }
  }

  /* at this point we have the implicit form in a matrix form,
     where each element of the array signifies a vector (X,Y,Z,W)
     and the vector means that the elements are linear in the
     corresponding variables, so each entry is actually
           x*X + y*Y + z*Z + W,

     The implicit equation of the surface is the zero set of
     the symbolic expansion of this determinant.  If the 
     expansion causes the determinant to SYMBOLICALLY become
     zero (ie, before even considering any values of x, y, and
     z), then there are base points -- ie, places where the
     original parametric surface has (0,0,0,0), and so the
     degree of the final equation is less than 2mn, so,
     effectively one of the rows or cols is identical to 
     another (equivalent to a row or col being 0), so the
     det is automatically zero.  In this case, we could
     still find the implicit equation by considering the
     leading minor of the matrix, but for now we will just
     assume that the surfaces are sufficiently complex.
  */

  if (base_case) {
    allocDeterminant_1D( &f_uv, 2*tpPatch1.iDeg0*tpPatch1.iDeg1,
       2*tpPatch1.iDeg0*tpPatch1.iDeg1, tpPatch0.iDeg0, tpPatch0.iDeg1);

    allocDeterminant_1D(&n_uv, rank_row_val,
                        rank_row_val, tpPatch0.iDeg0, tpPatch0.iDeg1);

    change_to_uv( *ndetImplicit, f_uv, tpPatch0 );
    change_to_uv( fdetImplicit, n_uv, tpPatch0 );
  }
  else {
    allocDeterminant_1D( &n_uv, 2*tpPatch1.iDeg0*tpPatch1.iDeg1,
       2*tpPatch1.iDeg0*tpPatch1.iDeg1, tpPatch0.iDeg0, tpPatch0.iDeg1);

    change_to_uv( *ndetImplicit, n_uv, tpPatch0 );
  }
  
  /* I send all of the info that might be necessary to 
     the routines that search for starting points.
     (only tracing open components right now, need to think
     of good termination for closed components and singularities)
   */

   /* printf("Enter Curve-Surface Int\n"); */
  if (getXtraComponents(plp4DStartingPoints,&plp4CPXStartingPoints, f_uv,
		     n_uv,tpPatch1,tpPatch0, box1, cbox0, 
                     row_perm, base_case) == -1) {
    printf("Could not get Components. exiting ...\n");
    return;
  }
   /* printf("Exit Curve-Surface Int\n"); */

  /* Restore Scale */
  Scale = 1.0;

#ifdef __FREE_ALL
  free_ten_p(&tpPatch0);
  free_ten_p(&tpPatch1);
  free_alloc_det_1D(&n_uv);
  if (base_case) {
    free_alloc_determ(&fdetImplicit);
    free_alloc_det_1D(&f_uv);
  }
  myfree(row_perm);
  myfree(col_perm);
#endif /* FREE_ALL */
#endif /* BEZ_CLIP */
}
#else
GetInsideStPointsUV(tp0, tp1, box1, cbox0, plp4DStartingPoints)
TENSOR_PATCH tp0, tp1;
BOUNDINGBOX box1, *cbox0;
LIST_POINT4D **plp4DStartingPoints;
{
#ifdef BEZ_CLIP
  int i, j;
  int numhits;
  LIST_POINT4D *new_point;
  XYZ_SPLINE_TYPE *spl;
  double thits[20], uhits[20], vhits[20];

  *plp4DStartingPoints = NULL;
  /* I am going to get curves from tp0 and do clipping with tp1 */

  /* tp0 - s = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 0, 0.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = 0.0;
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - s = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 0, 1.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = 1.0;
    new_point->p4D.dCoord[T] = thits[i];
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - t = 0 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 1, 0.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = thits[i];
    new_point->p4D.dCoord[T] = 0.0;
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

  /* tp0 - t = 1 */
  ALLOCN(spl, XYZ_SPLINE_TYPE, 1);
  GetSplineFromPatch(&tp0, 1, 1.0, spl);
  curve_surface_intersect(&tp1, spl, CLIP_TOL, CLIP_TOL, &numhits, thits,
			  uhits, vhits);
  for (i=0; i<numhits; i++) {
    ALLOCN(new_point, LIST_POINT4D, 1);
    new_point->plp4DNext = *plp4DStartingPoints;
    new_point->p4D.dCoord[U] = uhits[i];
    new_point->p4D.dCoord[V] = vhits[i];
    new_point->p4D.dCoord[S] = thits[i];
    new_point->p4D.dCoord[T] = 1.0;
    (*plp4DStartingPoints) = new_point;
  }
  free_xyz_spline(spl);

#else
  NDETERMINANT     ndetImplicit, fdetImplicit, ndet;
  NDETERMINANT_1D  n_uv, f_uv;
  LIST_POINT4CPX   *plp4CPXStartingPoints;
  int              cBlue, ret_val;
  char             rgchStruct[50];
  int              Nrows, Ncols;
  int              *row_perm, *col_perm;
  int              base_case, rank_row_val;
  int              i, j, k;
  POINT4D	   final_pt;
  TENSOR_PATCH     tpPatch0, tpPatch1;

  /* I am going to scale the patches down so that my matrix entries */
  /* dont get too big. */
  tpPatch0.iDeg0 = tp0.iDeg0;
  tpPatch0.iDeg1 = tp0.iDeg1;
  tpPatch1.iDeg0 = tp1.iDeg0;
  tpPatch1.iDeg1 = tp1.iDeg1;
  allocate_tensor_patch(&tpPatch0);
  allocate_tensor_patch(&tpPatch1);
  /* Now calculating the exact Scale factor */
  Scale = -1.0;
  for (i=0; i<=tp0.iDeg0; i++)
    for (j=0; j<=tp0.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tp0.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tp0.coeffs[i][j].dCoord[k]);
	tpPatch0.coeffs[i][j].dCoord[k] = tp0.coeffs[i][j].dCoord[k];
      }
      tpPatch0.coeffs[i][j].dCoord[3] = tp0.coeffs[i][j].dCoord[3];
    }
  for (i=0; i<=tp1.iDeg0; i++)
    for (j=0; j<=tp1.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tp1.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tp1.coeffs[i][j].dCoord[k]);
	tpPatch1.coeffs[i][j].dCoord[k] = tp1.coeffs[i][j].dCoord[k];
      }
      tpPatch1.coeffs[i][j].dCoord[3] = tp1.coeffs[i][j].dCoord[3];
    }

  /* Scale the tensor patch */
  scale_tensor_patch(&tpPatch0, 1);
  scale_tensor_patch(&tpPatch1, 1);

  bernsteinToPowerBasis( tpPatch1 );

  /* tpPatch0 is now in the power basis under the reparametrization 
     of s=s/1-s, t=t/1-t, but tpPatch1 is still a bezier tensor prod
     patch */

  allocDeterminant( &ndet, tpPatch1.iDeg0, tpPatch1.iDeg1, UNFACTORED );

  implicitize( tpPatch1, ndet );

  powerToBernsteinBasis( tpPatch1 );

  allocDeterminant( &ndetImplicit, tpPatch1.iDeg0, tpPatch1.iDeg1, FINAL );
  moveFactoredToFinalDeterminant( ndetImplicit, ndet, 
				 tpPatch1.iDeg0, tpPatch1.iDeg1 );

  Nrows = ndetImplicit.cRows;
  Ncols = ndetImplicit.cCols;
  base_case = check_for_base_points(ndetImplicit, Nrows, Ncols, &rank_row_val,
                                    &row_perm, &col_perm);

  if (base_case) {
    fdetImplicit.cRows = fdetImplicit.cCols = rank_row_val;
    ALLOCN(fdetImplicit.coeffs, POINT4D*, fdetImplicit.cRows);
    for (i=0; i<fdetImplicit.cCols; i++)
      ALLOCN(fdetImplicit.coeffs[i], POINT4D, fdetImplicit.cCols);

    for (i=0; i<rank_row_val; i++)
    {
      for (j=0; j<rank_row_val; j++)
      {
        fdetImplicit.coeffs[i][j].dCoord[X] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[X];
        fdetImplicit.coeffs[i][j].dCoord[Y] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Y];
        fdetImplicit.coeffs[i][j].dCoord[Z] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[Z];
        fdetImplicit.coeffs[i][j].dCoord[W] =
                  ndetImplicit.coeffs[row_perm[i]][col_perm[j]].dCoord[W];
      }
    }
  }

  /* at this point we have the implicit form in a matrix form,
     where each element of the array signifies a vector (X,Y,Z,W)
     and the vector means that the elements are linear in the
     corresponding variables, so each entry is actually
           x*X + y*Y + z*Z + W,

     The implicit equation of the surface is the zero set of
     the symbolic expansion of this determinant.  If the 
     expansion causes the determinant to SYMBOLICALLY become
     zero (ie, before even considering any values of x, y, and
     z), then there are base points -- ie, places where the
     original parametric surface has (0,0,0,0), and so the
     degree of the final equation is less than 2mn, so,
     effectively one of the rows or cols is identical to 
     another (equivalent to a row or col being 0), so the
     det is automatically zero.  In this case, we could
     still find the implicit equation by considering the
     leading minor of the matrix, but for now we will just
     assume that the surfaces are sufficiently complex.
  */

  if (base_case) {
    allocDeterminant_1D( &f_uv, 2*tpPatch1.iDeg0*tpPatch1.iDeg1,
       2*tpPatch1.iDeg0*tpPatch1.iDeg1, tpPatch0.iDeg0, tpPatch0.iDeg1);

    allocDeterminant_1D(&n_uv, rank_row_val,
                        rank_row_val, tpPatch0.iDeg0, tpPatch0.iDeg1);

    change_to_uv( ndetImplicit, f_uv, tpPatch0 );
    change_to_uv( fdetImplicit, n_uv, tpPatch0 );
  }
  else {
    allocDeterminant_1D( &n_uv, 2*tpPatch1.iDeg0*tpPatch1.iDeg1,
       2*tpPatch1.iDeg0*tpPatch1.iDeg1, tpPatch0.iDeg0, tpPatch0.iDeg1);

    change_to_uv( ndetImplicit, n_uv, tpPatch0 );
  }
  
  /* I send all of the info that might be necessary to 
     the routines that search for starting points.
     (only tracing open components right now, need to think
     of good termination for closed components and singularities)
   */

   /* printf("Enter Curve-Surface Int\n"); */
  if (getXtraComponents(plp4DStartingPoints,&plp4CPXStartingPoints, f_uv,
		     n_uv,tpPatch1,tpPatch0, box1, cbox0, 
                     row_perm, base_case) == -1) {
    printf("Could not get Components. exiting ...\n");
    return;
  }
   /* printf("Exit Curve-Surface Int\n"); */

  /* Restore Scale */
  Scale = 1.0;

#ifdef __FREE_ALL
  free_ten_p(&tpPatch0);
  free_ten_p(&tpPatch1);
  free_alloc_determ(&ndet);
  free_alloc_determ(&ndetImplicit);
  free_alloc_det_1D(&n_uv);
  if (base_case) {
    free_alloc_determ(&fdetImplicit);
    free_alloc_det_1D(&f_uv);
  }
  myfree(row_perm);
  myfree(col_perm);
#endif /* __FREE_ALL */
#endif /* BEZ_CLIP */
}
#endif

#ifdef __PERFORM_TIME
void PrintAllTimes()
{
  printf("Total Intersection time = %g\n", XsectTime);
  printf("Total Bounding Box + LP time = %g\n", LPTime); 
  printf("Total Implicitization time = %g\n", ImplicitTime);
  printf("Total Partitioning time = %g\n", PartitionTime);
  printf("Total Graph Computation + Ray Shooting + Final Solid Comp. time = %g\n", ClassifyTime); 
  printf("Total Fitting + Merging time = %g\n", FittingTime); 

}
#endif

