#include "defs.h"
#include "linalg.h"

typedef struct {
  double x, y;
} point_typ;

#ifdef _TESTING_ALONE
#include <stdlib.h>
#define new malloc
#define myfree free
#endif

#ifdef _TESTING_ALONE
LINALGINIT
#else
EXTERN_LINALG
#endif

#define DIST_BETWEEN_PTS 0.03 /* This constant gives the distance */
			      /* between 2 adjacent points in the */
			      /* curve */

/* These two global variables are 4d arrays whose [i][j][k][l]th */
/* element holds the curve number of the trimming polygon of the ith */
/* patch which is touched by the jth merged curve. k coord signifies */
/* if the point is a starting pt or ending pt. I have left 8 possible */
/* curve numbers because different curves of the trim curve can be */
/* adjacent to the same patch. I will use this in PartitionSplinegon */
/* routine. The values are filled in CheckIntersectionCurves routine */
short fXsc[100][10][2][8], sXsc[100][10][2][8]; 
static UV_SPLINE_TYPE uvst[2*MAX_SPLINES_PER_TRIM_REGION][10]; 
                                          /* 10 for maximum */
                                          /* intersections with one curve */
static XYZ_SPLINE_TYPE xyzst[2*MAX_SPLINES_PER_TRIM_REGION][10];

extern NEIGHBORING_CURVE ngh_crv[];
extern double Scale;
extern int NumNghCrvs, NumOvrPatches;
extern OVR_BOXES overlap_pairs[];
extern short special_crvs_in_first_sol[][MAX_SPLINES_PER_TRIM_REGION],
             special_crvs_in_sec_sol[][MAX_SPLINES_PER_TRIM_REGION]; 
extern short first_same_solid[MAX_PATCHES_PER_SOLID][1][MAX_PARTNS_PER_PATCH][MAX_SPLINES_PER_TRIM_REGION]; 
                                              /* this is the global */
					      /* variable that holds */
					      /* the boolean for */
					      /* [patchno][trimportion */
					      /* no][regionno][curveno] */
extern short second_same_solid[MAX_PATCHES_PER_SOLID][1][MAX_PARTNS_PER_PATCH][MAX_SPLINES_PER_TRIM_REGION];
extern short first_complete_ovrlap[], second_complete_ovrlap[];
extern double dist3D();

short GetAdjacentTrimPortion();
int EvaluateTangentVector(UV_SPLINE_TYPE* spl, double Param, POINT2D* tgt);

/* Evaluates a 2d point at a particular param of the curve */
void Eval2DCurve(curve, Value, pt)
UV_SPLINE_TYPE *curve;
double Value;
POINT2D *pt;
{
  int dg, i, j;
  double uval, vval, wval;
  double *SPow, *OmsPow;
  long *CombArr;
  double OneMinusVal;

  OneMinusVal = 1.0 - Value;
  dg = curve->Deg;

  SPow = (double *) new((dg + 1) *
			   sizeof(double));
  OmsPow = (double *) new((dg + 1) *
			   sizeof(double));
  CombArr = (long *) new((dg + 1) *
			 sizeof(long));

  SPow[0] = OmsPow[0] = 1.0;
  SPow[1] = Value; OmsPow[1] = OneMinusVal;
  CombArr[0] = 1; CombArr[1] = dg;
  for (i=2; i<=dg; i++) {
    SPow[i] = SPow[i-1] * Value;
    OmsPow[i] = OmsPow[i-1] * OneMinusVal;
    CombArr[i] = CombArr[i-1] * (dg + 1 - i) / i;
  }

  uval = vval = wval = 0.0;
  for (i=0; i<=dg; i++) {
    uval += curve->CntrlPts[i].dCoord[0] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
    vval += curve->CntrlPts[i].dCoord[1] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
    wval += curve->CntrlPts[i].dCoord[2] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
  }

  if (FP_EQ_EPS(wval, 0.0, 1e-3)) {
    printf("wval is very close to zero\n");
    exit(-1);
  }
  pt->u = uval / wval;
  pt->v = vval / wval;

  myfree(SPow);
  myfree(OmsPow);
  myfree(CombArr);
}

/* Evaluates a 3d point at a particular param of the curve */
void Eval3DCurve(curve, Value, pt)
XYZ_SPLINE_TYPE *curve;
double Value;
POINT3D *pt;
{
  int dg, i, j;
  double xval, yval, zval, wval;
  double *SPow, *OmsPow;
  long *CombArr;
  double OneMinusVal;

  OneMinusVal = 1.0 - Value;
  dg = curve->Deg;

  SPow = (double *) new((dg + 1) *
			   sizeof(double));
  OmsPow = (double *) new((dg + 1) *
			   sizeof(double));
  CombArr = (long *) new((dg + 1) *
			 sizeof(long));

  SPow[0] = OmsPow[0] = 1.0;
  SPow[1] = Value; OmsPow[1] = OneMinusVal;
  CombArr[0] = 1; CombArr[1] = dg;
  for (i=2; i<=dg; i++) {
    SPow[i] = SPow[i-1] * Value;
    OmsPow[i] = OmsPow[i-1] * OneMinusVal;
    CombArr[i] = CombArr[i-1] * (dg + 1 - i) / i;
  }

  xval = yval = zval = wval = 0.0;
  for (i=0; i<=dg; i++) {
    xval += curve->CntrlPts[i].dCoord[0] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
    yval += curve->CntrlPts[i].dCoord[1] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
    zval += curve->CntrlPts[i].dCoord[2] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
    wval += curve->CntrlPts[i].dCoord[3] * CombArr[i] *
	     SPow[i] * OmsPow[dg - i];
  }

  if (FP_EQ_EPS(wval, 0.0, 1e-3)) {
    printf("wval is very close to zero\n");
    exit(-1);
  }
  pt->dCoord[0] = xval / wval;
  pt->dCoord[1] = yval / wval;
  pt->dCoord[2] = zval / wval;

  myfree(SPow);
  myfree(OmsPow);
  myfree(CombArr);
}

point_typ GetPtInsideSplinegon(tpst)
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  point_typ pt;
  POINT2D p2D;
  int num_crvs, num_pts;
  double vert[MAX_POINTS_PER_TRIM_REGION][2], step_size;
  int i, j, k;

  num_crvs = tpst->NumOfCurves;
  for (k=0, i=0; i<num_crvs; i++) {
    num_pts = NumPointsOnCurve(&tpst->UV_Splines[i], 0.03);
    step_size = 1.0 / num_pts;
    for (j=0; j<num_pts; j++) {
      Eval2DCurve(&tpst->UV_Splines[i], j*step_size, &p2D);
      vert[k][0] = p2D.u;
      vert[k++][1] = p2D.v;
      if (k >= MAX_POINTS_PER_TRIM_REGION) {
	printf("GetPtInsideSplinegon: Too many points for one trim region\n");
	exit(-1);
      }
    }
  }
  vert[k][0] = vert[0][0];
  vert[k][1] = vert[0][1];

#ifdef DEBUG_PRINT_POLY
  printf("############\n");
  for (i=0; i<=k; i++)
    printf("%g %g\n", vert[i][0], vert[i][1]);
  printf("############\n");
#endif

  find_point_inside_polygon(k, vert, &pt);
  return pt;
}

int IsPtInsideSplinegon(tpst, pt)
TRIM_PORTION_SPLINE_TYPE *tpst;
point_typ *pt;
{
  POINT2D p2D;
  int num_crvs, num_pts;
  double vert[MAX_POINTS_PER_TRIM_REGION][2], step_size;
  int i, j, k;

  num_crvs = tpst->NumOfCurves;
  for (k=0, i=0; i<num_crvs; i++) {
    num_pts = NumPointsOnCurve(&tpst->UV_Splines[i], 0.03);
    step_size = 1.0 / num_pts;
    for (j=0; j<num_pts; j++) {
      Eval2DCurve(&tpst->UV_Splines[i], j*step_size, &p2D);
      vert[k][0] = p2D.u;
      vert[k++][1] = p2D.v;
      if (k >= MAX_POINTS_PER_TRIM_REGION) {
	printf("GetPtInsideSplinegon: Too many points for one trim region\n");
	exit(-1);
      }
    }
  }
  vert[k][0] = vert[0][0];
  vert[k][1] = vert[0][1];

  return is_point_inside_polygon(k, vert, pt);
}

int IsSplinegonAntiClockwise(tpst)
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  POINT2D p2D;
  int num_crvs, num_pts;
  double vert[MAX_POINTS_PER_TRIM_REGION][2], step_size;
  int i, j, k;

  num_crvs = tpst->NumOfCurves;
  for (k=0, i=0; i<num_crvs; i++) {
    num_pts = NumPointsOnCurve(&tpst->UV_Splines[i], 0.03);
    step_size = 1.0 / num_pts;
    for (j=0; j<num_pts; j++) {
      Eval2DCurve(&tpst->UV_Splines[i], j*step_size, &p2D);
      vert[k][0] = p2D.u;
      vert[k++][1] = p2D.v;
      if (k >= MAX_POINTS_PER_TRIM_REGION) {
	printf("IsSplinegonAntiClockwise: Too many points for one trim region\n");
	exit(-1);
      }
    }
  }
  vert[k][0] = vert[0][0];
  vert[k][1] = vert[0][1];

  return is_curve_anticlockwise(vert, k);
}

/* Reverses control points of a 2d spline */
void ReverseCntrlPtsOf2DSpline(Spl)
UV_SPLINE_TYPE *Spl;
{
  UV_SPLINE_TYPE TmpSpl;
  int i;

  TmpSpl.Deg = Spl->Deg;
  TmpSpl.CntrlPts = (POINT3D *) new((TmpSpl.Deg + 1)*sizeof(POINT3D));
  for (i=0; i<=TmpSpl.Deg; i++)
    TmpSpl.CntrlPts[i] = Spl->CntrlPts[i];

  for (i=0; i<=TmpSpl.Deg; i++)
    Spl->CntrlPts[i] = TmpSpl.CntrlPts[TmpSpl.Deg - i];

  myfree(TmpSpl.CntrlPts);
}

/* Reverses control points of a 3d spline */
void ReverseCntrlPtsOf3DSpline(Spl)
XYZ_SPLINE_TYPE *Spl;
{
  XYZ_SPLINE_TYPE TmpSpl;
  int i;

  TmpSpl.Deg = Spl->Deg;
  TmpSpl.CntrlPts = (POINT4D *) new((TmpSpl.Deg + 1)*sizeof(POINT4D));
  for (i=0; i<=TmpSpl.Deg; i++)
    TmpSpl.CntrlPts[i] = Spl->CntrlPts[i];

  for (i=0; i<=TmpSpl.Deg; i++)
    Spl->CntrlPts[i] = TmpSpl.CntrlPts[TmpSpl.Deg - i];

  myfree(TmpSpl.CntrlPts);
}

/* Return the 3d spline corresponding to the curve at a constant value */
/* of the patch */
int GetSplineFromPatch(tp, Var, Value, Spl)
TENSOR_PATCH *tp;
int Var; /* 0 means s is constant and 1 means t is constant */
double Value;
XYZ_SPLINE_TYPE *Spl;
{
  int i, j, k;
  int OtherDeg;
  double *SPow, *OmsPow;
  long *CombArr;
  double OneMinusVal;

  OneMinusVal = 1.0 - Value;

  switch (Var) {
  case 0: 
    Spl->Deg = tp->iDeg1;
    OtherDeg = tp->iDeg0;
    break;
  case 1:
    Spl->Deg = tp->iDeg0;
    OtherDeg = tp->iDeg1;
    break;
  default:
    printf("Var is not correct\n");
    exit(-1);
  }

  Spl->CntrlPts = (POINT4D *) new((Spl->Deg + 1) * 
				  sizeof(POINT4D));

  SPow = (double *) new((OtherDeg + 1) *
			   sizeof(double));
  OmsPow = (double *) new((OtherDeg + 1) *
			   sizeof(double));
  CombArr = (long *) new((OtherDeg + 1) *
			 sizeof(long));

  SPow[0] = OmsPow[0] = 1.0;
  SPow[1] = Value; OmsPow[1] = OneMinusVal;
  CombArr[0] = 1; CombArr[1] = OtherDeg;
  for (i=2; i<=OtherDeg; i++) {
    SPow[i] = SPow[i-1] * Value;
    OmsPow[i] = OmsPow[i-1] * OneMinusVal;
    CombArr[i] = CombArr[i-1] * (OtherDeg + 1 - i) / i;
  }

  for (i=0; i<=Spl->Deg; i++)
    for (k=0; k<4; k++) 
      Spl->CntrlPts[i].dCoord[k] = 0.0;

  for (i=0; i<=Spl->Deg; i++)
    for (j=0; j<=OtherDeg; j++)
      for (k=0; k<4; k++)
	switch (Var) {
	case 0:
	  Spl->CntrlPts[i].dCoord[k] += (tp->coeffs[j][i].dCoord[k] *
					 CombArr[j]*SPow[j]*OmsPow[OtherDeg-j]);
	  break;
	case 1:
	  Spl->CntrlPts[i].dCoord[k] += (tp->coeffs[i][j].dCoord[k] *
					 CombArr[j]*SPow[j]*OmsPow[OtherDeg-j]);
	  break;
	}

  myfree(SPow);
  myfree(OmsPow);
  myfree(CombArr);

  return 0;
}

/* 3d subdivision on one side */
int ThreeDBezSub(InC, par, OutC, flag, do_alloc)
XYZ_SPLINE_TYPE *InC, *OutC;
double par;
int flag, do_alloc;
{
  int i, j, k;
  double act_par;
  POINT4D *ptr, *tmp_ptr;

  if (InC->Deg == 0) {
    printf("BezSub : Bad curve degree\n");
    exit(-1);
  }

  OutC->Deg = InC->Deg;
  if (do_alloc)
    ALLOCN(OutC->CntrlPts, POINT4D, ((OutC->Deg)+1));

  ALLOCN(tmp_ptr, POINT4D, (InC->Deg + 1));

  if ((flag != 0) && (flag != 1)) {
    printf("BezSub: Illegal value of flag\n");
    exit(-1);
  }

  for (i=0; i<=InC->Deg; i++) 
    OutC->CntrlPts[i] = InC->CntrlPts[i];

  act_par = par;
  ptr = OutC->CntrlPts;

  if (!flag) {
    for (i=1; i<=InC->Deg; i++) {
      for (j=i-1; j<=InC->Deg; j++)
	tmp_ptr[j] = ptr[j];

      for (j=i; j<=InC->Deg; j++)
	for (k=0; k<4; k++)
	  ptr[j].dCoord[k] = (1 - act_par)*tmp_ptr[j-1].dCoord[k] + 
	                     act_par*tmp_ptr[j].dCoord[k];
    }
  } else {
    for (i=(InC->Deg - 1); i>=0; i--) {
      for (j=(i+1); j>=0; j--)
	tmp_ptr[j] = ptr[j];

      for (j=i; j>=0; j--)
	for (k=0; k<4; k++)
	  ptr[j].dCoord[k] = (1 - act_par)*tmp_ptr[j].dCoord[k] + 
	                     act_par*tmp_ptr[j+1].dCoord[k];
    }
  }

  myfree(tmp_ptr);
  return 0;
}

/* 3d bezier subdivision from c to d */
int ThreeDBezierSubdivision(InC, c, d, OutC)
XYZ_SPLINE_TYPE *InC, **OutC;
double c, d;
{
  int i;
  XYZ_SPLINE_TYPE *TmpC;
  double nc;

  (*OutC) = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));

  TmpC = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));

  if (!FP_EQ_EPS(d, 1.0, 5e-3))
    ThreeDBezSub(InC, d, TmpC, 0, 1);
  else {
    TmpC->Deg = InC->Deg;
    ALLOCN(TmpC->CntrlPts, POINT4D, ((InC->Deg)+1));
    for (i=0; i<=InC->Deg; i++)
      TmpC->CntrlPts[i] = InC->CntrlPts[i];
  }

  nc = c/d;
  if (!FP_EQ_EPS(nc, 0.0, 5e-3))
    ThreeDBezSub(TmpC, nc, (*OutC), 1, 1);
  else {
    (*OutC)->Deg = InC->Deg;
    ALLOCN((*OutC)->CntrlPts, POINT4D, ((InC->Deg)+1));
    for (i=0; i<=TmpC->Deg; i++)
      (*OutC)->CntrlPts[i] = TmpC->CntrlPts[i];
  }

  myfree(TmpC->CntrlPts);
  return 0;
}

/* This procedure gives the number of points required to be evaluated */
/* based on the bounding box. This is used in obtain_proper_trim_curve */
/* routine */
int NumPointsOnCurve(ic, dist)
UV_SPLINE_TYPE *ic;
double dist;
{
  int dg, i, j, num;
  double minu, maxu, minv, maxv;
  double tmpu, tmpv, bigdiff;

  dg = ic->Deg;
  minu = maxu = ic->CntrlPts[0].dCoord[0] / ic->CntrlPts[0].dCoord[2];
  minv = maxv = ic->CntrlPts[0].dCoord[1] / ic->CntrlPts[0].dCoord[2];
  
  for (i=1; i<=dg; i++) {
    tmpu = ic->CntrlPts[i].dCoord[0] / ic->CntrlPts[i].dCoord[2];
    tmpv = ic->CntrlPts[i].dCoord[1] / ic->CntrlPts[i].dCoord[2];
    if (tmpu < minu)
      minu = tmpu;
    if (tmpu > maxu)
      maxu = tmpu;
    if (tmpv < minv)
      minv = tmpv;
    if (tmpv > maxv)
      maxv = tmpv;
  }
  
  if ((maxu - minu) > (maxv - minv))
    bigdiff = maxu - minu;
  else
    bigdiff = maxv - minv;

  num = (int) ((bigdiff/dist) + 0.5);
  if (num < 1)
    num = 1;
  return num;
}

/* plane bezier being subdivided based on par. if flag == 0, curve */
/* returned is from [0,par]. if flag == 1, curve returned is from */
/* [par,1] */
int TwoDBezSub(InC, par, OutC, flag, do_alloc)
UV_SPLINE_TYPE *InC, *OutC;
double par;
int flag, do_alloc;
{
  int i, j, k;
  double act_par;
  POINT3D *ptr, *tmp_ptr;

  if (InC->Deg == 0) {
    printf("TwoDBezSub : Bad curve degree\n");
    exit(-1);
  }

  OutC->Deg = InC->Deg;
  if (do_alloc)
    ALLOCN(OutC->CntrlPts, POINT3D, ((OutC->Deg)+1));

  ALLOCN(tmp_ptr, POINT3D, (InC->Deg + 1));

  if ((flag != 0) && (flag != 1)) {
    printf("BezSub: Illegal value of flag\n");
    exit(-1);
  }

  for (i=0; i<=InC->Deg; i++) 
    OutC->CntrlPts[i] = InC->CntrlPts[i];

  act_par = par;
  ptr = OutC->CntrlPts;

  if (!flag) {
    for (i=1; i<=InC->Deg; i++) {
      for (j=i-1; j<=InC->Deg; j++)
	tmp_ptr[j] = ptr[j];

      for (j=i; j<=InC->Deg; j++)
	for (k=0; k<3; k++)
	  ptr[j].dCoord[k] = (1 - act_par)*tmp_ptr[j-1].dCoord[k] + 
	                     act_par*tmp_ptr[j].dCoord[k];
    }
  } else {
    for (i=(InC->Deg - 1); i>=0; i--) {
      for (j=(i+1); j>=0; j--)
	tmp_ptr[j] = ptr[j];

      for (j=i; j>=0; j--)
	for (k=0; k<3; k++)
	  ptr[j].dCoord[k] = (1 - act_par)*tmp_ptr[j].dCoord[k] + 
	                     act_par*tmp_ptr[j+1].dCoord[k];
    }
  }

  myfree(tmp_ptr);
  return 0;
}

/* This routine takes a UVspline and a XYZspline and splits them into */
/* (n+1) curves given n param values. All the Out curves are arrays */
/* which have been allocated. Actually, the n+1 value in param should */
/* be 1 ie. param[n] = 1 */
int SubdivideTrimCurve(In_uvc, In_xyzc, num, param, Out_uvc, Out_xyzc)
UV_SPLINE_TYPE *In_uvc, *Out_uvc;
XYZ_SPLINE_TYPE *In_xyzc, *Out_xyzc;
double *param;
short num;
{
  int i;
  UV_SPLINE_TYPE *Tmp_uv;
  XYZ_SPLINE_TYPE *Tmp_xyz;

  if (num < 0) {
    printf("SubdivideTrimCurve: Error in num input\n");
    exit(-1);
  }

  if (!num) {
    Out_uvc[0].Deg = In_uvc->Deg;
    ALLOCN(Out_uvc[0].CntrlPts, POINT3D, ((In_uvc->Deg)+1));
    for (i=0; i<=Out_uvc[0].Deg; i++)
      Out_uvc[0].CntrlPts[i] = In_uvc->CntrlPts[i];

    Out_xyzc[0].Deg = In_xyzc->Deg;
    ALLOCN(Out_xyzc[0].CntrlPts, POINT4D, ((In_xyzc->Deg)+1));
    for (i=0; i<=Out_xyzc[0].Deg; i++)
      Out_xyzc[0].CntrlPts[i] = In_xyzc->CntrlPts[i];

    return 0;
  }

  Tmp_uv = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  Tmp_xyz = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  Tmp_uv->Deg = In_uvc->Deg;
  Tmp_xyz->Deg = In_xyzc->Deg;
  ALLOCN(Tmp_uv->CntrlPts, POINT3D, ((In_uvc->Deg)+1));
  ALLOCN(Tmp_xyz->CntrlPts, POINT4D, ((In_xyzc->Deg)+1));
  for (i=0; i<=Tmp_uv->Deg; i++)
    Tmp_uv->CntrlPts[i] = In_uvc->CntrlPts[i];
  for (i=0; i<=Tmp_xyz->Deg; i++)
    Tmp_xyz->CntrlPts[i] = In_xyzc->CntrlPts[i];

  for (i=num-1; i>=0; i--) {
    TwoDBezSub(Tmp_uv, param[i]/param[i+1], &Out_uvc[i+1], 1, 1);
    ThreeDBezSub(Tmp_xyz, param[i]/param[i+1], &Out_xyzc[i+1], 1, 1);

    if (i == 0) {
      TwoDBezSub(Tmp_uv, param[i]/param[i+1], &Out_uvc[i], 0, 1);
      ThreeDBezSub(Tmp_xyz, param[i]/param[i+1], &Out_xyzc[i], 0, 1);
    } else {
      TwoDBezSub(Tmp_uv, param[i]/param[i+1], Tmp_uv, 0, 0);
      ThreeDBezSub(Tmp_xyz, param[i]/param[i+1], Tmp_xyz, 0, 0);
    }
  }

  myfree(Tmp_uv->CntrlPts);
  myfree(Tmp_xyz->CntrlPts);
  myfree(Tmp_uv); myfree(Tmp_xyz);
  return 0;
}

/* Planar bezier subdivision from c to d */
int TwoDBezierSubdivision(InC, c, d, OutC)
UV_SPLINE_TYPE *InC, **OutC;
double c, d;
{
  int i;
  UV_SPLINE_TYPE *TmpC;
  double nc;

  (*OutC) = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));

  TmpC = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));

  if (!FP_EQ_EPS(d, 1.0, 5e-3))
    TwoDBezSub(InC, d, TmpC, 0, 1);
  else {
    TmpC->Deg = InC->Deg;
    ALLOCN(TmpC->CntrlPts, POINT3D, ((InC->Deg)+1));
    for (i=0; i<=InC->Deg; i++)
      TmpC->CntrlPts[i] = InC->CntrlPts[i];
  }

  nc = c/d;
  if (!FP_EQ_EPS(nc, 0.0, 5e-3))
    TwoDBezSub(TmpC, nc, (*OutC), 1, 1);
  else {
    (*OutC)->Deg = InC->Deg;
    ALLOCN((*OutC)->CntrlPts, POINT3D, ((InC->Deg)+1));
    for (i=0; i<=TmpC->Deg; i++)
      (*OutC)->CntrlPts[i] = TmpC->CntrlPts[i];
  }

  myfree(TmpC->CntrlPts);
  return 0;
}

/* Reverses both the domain and space curve in an INTERSECTION_CURVE */
void reverse_int_curve(new_crv, l_ptr)
INTERSECTION_CURVE **new_crv, **l_ptr;
{
  INTERSECTION_CURVE *ptr1, *ptr2, *ptr3;

  if ((*new_crv)->next == NULL) {
    ReverseCntrlPtsOf2DSpline(&(*new_crv)->uv_spline);
    ReverseCntrlPtsOf3DSpline(&(*new_crv)->xyz_spline);
    (*l_ptr) = (*new_crv);
    return;
  }

  ptr1 = (*new_crv);
  ptr2 = ptr1->next;
  ReverseCntrlPtsOf2DSpline(&ptr1->uv_spline);
  ReverseCntrlPtsOf3DSpline(&ptr1->xyz_spline);
  ptr1->next = NULL;

  reverse_int_curve(&ptr2, &ptr3);
  ptr3->next = ptr1;
  (*l_ptr) = ptr1;
  (*new_crv) = ptr2;
  return;
}

/* This routine reverses a trim_portion_spline_type */
void ReverseTPST(tpst)
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  int i, j, num;
  TRIM_PORTION_SPLINE_TYPE tmp_tpst;

  copy_trim_portion_spline_type(&tmp_tpst, tpst, 0);
  num = tpst->NumOfCurves;

  for (i=0; i<num; i++) {
    myfree(tpst->UV_Splines[i].CntrlPts);
    myfree(tpst->XYZ_Splines[i].CntrlPts);
  }

  for (i=0; i<num; i++) {
    ReverseCntrlPtsOf2DSpline(&tmp_tpst.UV_Splines[i]);
    ReverseCntrlPtsOf3DSpline(&tmp_tpst.XYZ_Splines[i]);
    tpst->UV_Splines[num-i-1].Deg = tmp_tpst.UV_Splines[i].Deg;
    ALLOCN(tpst->UV_Splines[num-i-1].CntrlPts, POINT3D,
	   (tpst->UV_Splines[num-i-1].Deg+1)); 
    for (j=0; j<=tpst->UV_Splines[num-i-1].Deg; j++)
      tpst->UV_Splines[num-i-1].CntrlPts[j] = tmp_tpst.UV_Splines[i].CntrlPts[j];
    tpst->XYZ_Splines[num-i-1].Deg = tmp_tpst.XYZ_Splines[i].Deg;
    ALLOCN(tpst->XYZ_Splines[num-i-1].CntrlPts, POINT4D,
	   (tpst->XYZ_Splines[num-i-1].Deg+1)); 
    for (j=0; j<=tpst->XYZ_Splines[num-i-1].Deg; j++)
      tpst->XYZ_Splines[num-i-1].CntrlPts[j] = tmp_tpst.XYZ_Splines[i].CntrlPts[j];
    tpst->AdjInfo[num-i-1] = tmp_tpst.AdjInfo[i];
  }
}

/* Takes uvc, xyzc and ngh and puts it in INTERSECTION_CURVE */
void allocate_one_int_curve_struct(new_crv, uvc, xyzc, ngh)
INTERSECTION_CURVE **new_crv;
UV_SPLINE_TYPE *uvc;
XYZ_SPLINE_TYPE *xyzc;
int ngh;
{
  int i, uvdeg, xyzdeg;

  ALLOCN((*new_crv), INTERSECTION_CURVE, 1);
  (*new_crv)->next = NULL;
  (*new_crv)->neighbor = ngh;
  uvdeg = (*new_crv)->uv_spline.Deg = uvc->Deg;
  xyzdeg = (*new_crv)->xyz_spline.Deg = xyzc->Deg;

  ALLOCN((*new_crv)->uv_spline.CntrlPts, POINT3D, uvdeg+1);
  ALLOCN((*new_crv)->xyz_spline.CntrlPts, POINT4D, xyzdeg+1);
  for (i=0; i<=uvdeg; i++) 
    (*new_crv)->uv_spline.CntrlPts[i] = uvc->CntrlPts[i];
  for (i=0; i<=xyzdeg; i++) 
    (*new_crv)->xyz_spline.CntrlPts[i] = xyzc->CntrlPts[i];
}

/* Given a MERGED_CURVE, it returns two POINT2Ds, one the starting */
/* point and other the ending point in the list. It also returns the */
/* ptrs to the first and last elements in the list */
void return_st_end_in_p2d(mc, sp2D, ep2D, fptr, lptr)
MERGED_CURVE *mc;  
POINT2D *sp2D, *ep2D;
INTERSECTION_CURVE **fptr, **lptr;
{
  INTERSECTION_CURVE *tptr;
  double w_rec;
  int deg;

  tptr = mc->int_curve;
  if (tptr == NULL) {
    printf("return_st_end_in_p2d: Empty curve\n");
    sp2D->u = sp2D->v = ep2D->u = ep2D->v = -1.0;
    return;
  }

  (*fptr) = tptr;
  w_rec = 1 / tptr->uv_spline.CntrlPts[0].dCoord[2];
  sp2D->u = tptr->uv_spline.CntrlPts[0].dCoord[0] * w_rec;
  sp2D->v = tptr->uv_spline.CntrlPts[0].dCoord[1] * w_rec;

  /* Go to the last element of the list */
  while (tptr->next != NULL)
    tptr = tptr->next;

  (*lptr) = tptr;
  deg = tptr->uv_spline.Deg;
  w_rec = 1 / tptr->uv_spline.CntrlPts[deg].dCoord[2];
  ep2D->u = tptr->uv_spline.CntrlPts[deg].dCoord[0] * w_rec;
  ep2D->v = tptr->uv_spline.CntrlPts[deg].dCoord[1] * w_rec;
}

/* Given a spline, it merges it with a set of curves. This routine is */
/* similar to Merge2DLists except that it is for splines */
INTERSECTION_CURVE *MergeSplines(m_curve, num_curv, uvc, xyzc, ngh, error_correct, f_xyzc)
MERGED_CURVE **m_curve;
int *num_curv;
UV_SPLINE_TYPE *uvc;
XYZ_SPLINE_TYPE *xyzc, *f_xyzc;
int ngh, error_correct;
{
  int p, q;
  int i, j, k;
  INTERSECTION_CURVE *new_crv;
  INTERSECTION_CURVE *tptr, *f_ptr, *l_ptr, *f_ptr2, *l_ptr2;
  POINT2D comp_ep0, comp_ep1;
  POINT2D st_pt, end_pt, st_pt2, end_pt2;
  double dist0, dist1;
  double wcrd;
  int close0, close1, tmp_deg;
  int degenerate_case; /* = 1 when the two endpts of comp are
			  within 2e-5 dist. apart */

  allocate_one_int_curve_struct(&new_crv, uvc, xyzc, ngh);

/*
  if (!FP_EQ(uvc->CntrlPts[0].dCoord[2], 1.0) ||
      !FP_EQ(uvc->CntrlPts[uvc->Deg].dCoord[2], 1.0)) {
    printf("MergeSplines: Curve is not polynomial\n");
    exit(-1);
  }
*/

  comp_ep0.u = uvc->CntrlPts[0].dCoord[0] / uvc->CntrlPts[0].dCoord[2];
  comp_ep0.v = uvc->CntrlPts[0].dCoord[1] / uvc->CntrlPts[0].dCoord[2];
  comp_ep1.u = uvc->CntrlPts[uvc->Deg].dCoord[0] / uvc->CntrlPts[uvc->Deg].dCoord[2];
  comp_ep1.v = uvc->CntrlPts[uvc->Deg].dCoord[1] / uvc->CntrlPts[uvc->Deg].dCoord[2];

  if (EQUAL2D_EPS(comp_ep0, comp_ep1, 1e-8))
    degenerate_case = 1;
  else degenerate_case = 0;

  if (degenerate_case) {
    printf("MergeSplines: Curve in question is a loop\n");
    exit(-1);
  }

  for (i=0; i<(*num_curv); i++) {
    return_st_end_in_p2d(&((*m_curve)[i]), &st_pt, &end_pt, &f_ptr, &l_ptr);
    close0 = closer_pt(&st_pt, &comp_ep0, &comp_ep1, &dist0);
    close1 = closer_pt(&end_pt, &comp_ep0, &comp_ep1, &dist1);

    if (EQUAL2D_EPS(comp_ep0, st_pt, 5e-6) && close0 == 0
	&& dist0 <= dist1) { 
      reverse_int_curve(&new_crv, &tptr);

      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* satement can be removed if it leads to major problems */
      tmp_deg = new_crv->uv_spline.Deg;
      wcrd = new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[2];
      new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
      new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = new_crv->xyz_spline.Deg;
	new_crv->xyz_spline.CntrlPts[tmp_deg] = f_ptr->xyz_spline.CntrlPts[0];
	f_xyzc->CntrlPts[0] = new_crv->xyz_spline.CntrlPts[tmp_deg];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the front of the list */
      (*m_curve)[i].int_curve = new_crv;
      new_crv->next = f_ptr;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      st_pt = comp_ep1;
      f_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(st_pt, st_pt2, 5e-6) && close0 == 0 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = st_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    f_ptr->xyz_spline.CntrlPts[0] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_xyzc->CntrlPts[f_xyzc->Deg] = f_ptr->xyz_spline.CntrlPts[0];
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  tptr->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(st_pt, end_pt2, 5e-6) && close0 == 1 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    f_ptr->xyz_spline.CntrlPts[0] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	    f_xyzc->CntrlPts[f_xyzc->Deg] = f_ptr->xyz_spline.CntrlPts[0];
	  }

	  l_ptr2->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep0, end_pt, 5e-6) && close1 == 0
	       && dist1 <= dist0) {
      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      wcrd = new_crv->uv_spline.CntrlPts[0].dCoord[2];
      new_crv->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
      new_crv->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = l_ptr->xyz_spline.Deg;
	new_crv->xyz_spline.CntrlPts[0] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	f_xyzc->CntrlPts[0] = new_crv->xyz_spline.CntrlPts[0];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the end of the list */
      l_ptr->next = new_crv;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      end_pt = comp_ep1;
      l_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(end_pt, st_pt2, 5e-6) && close1 == 0 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr->xyz_spline.Deg;
	    l_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_xyzc->CntrlPts[tmp_deg] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	  }

	  l_ptr->next = f_ptr2;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(end_pt, end_pt2, 5e-6) && close1 == 1 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = end_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg] =
	      l_ptr2->xyz_spline.CntrlPts[tmp_deg];   
	    f_xyzc->CntrlPts[l_ptr->xyz_spline.Deg] =
	      l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg]; 
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  l_ptr->next = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep1, st_pt, 5e-6) && close0 == 1
	       && dist0 <= dist1) {  
      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      tmp_deg = new_crv->uv_spline.Deg;
      wcrd = new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[2];
      new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
      new_crv->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = new_crv->xyz_spline.Deg;
	new_crv->xyz_spline.CntrlPts[tmp_deg] = f_ptr->xyz_spline.CntrlPts[0];
	f_xyzc->CntrlPts[tmp_deg] = new_crv->xyz_spline.CntrlPts[tmp_deg];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the beginning of the list */
      new_crv->next = f_ptr;
      (*m_curve)[i].int_curve = new_crv;

      /* reassign st_pt to the new starting pt - it was the first point */
      /* of the original curve */
      st_pt = comp_ep0;
      f_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(st_pt, st_pt2, 5e-6) && close0 == 0 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = st_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    f_ptr->xyz_spline.CntrlPts[0] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_xyzc->CntrlPts[0] = f_ptr->xyz_spline.CntrlPts[0];
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  tptr->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(st_pt, end_pt2, 5e-6) && close0 == 1 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    f_ptr->xyz_spline.CntrlPts[0] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	    f_xyzc->CntrlPts[0] = f_ptr->xyz_spline.CntrlPts[0];
	  }

	  l_ptr2->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	}
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep1, end_pt, 5e-6) && close1 == 1
	       && dist1 <= dist0) {
      reverse_int_curve(&new_crv, &tptr);

      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      wcrd = new_crv->uv_spline.CntrlPts[0].dCoord[2];
      new_crv->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
      new_crv->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = l_ptr->xyz_spline.Deg;
	new_crv->xyz_spline.CntrlPts[0] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	f_xyzc->CntrlPts[f_xyzc->Deg] = new_crv->xyz_spline.CntrlPts[0];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the end of the list */
      l_ptr->next = new_crv;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      end_pt = comp_ep0;
      l_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(end_pt, st_pt2, 5e-6) && close1 == 0 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr->xyz_spline.Deg;
	    l_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_xyzc->CntrlPts[0] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	  }

	  l_ptr->next = f_ptr2;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(end_pt, end_pt2, 5e-6) && close1 == 1 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = end_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg] =
	      l_ptr2->xyz_spline.CntrlPts[tmp_deg];  
	    f_xyzc->CntrlPts[0] =
	      l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg]; 
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  l_ptr->next = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    }
  }

  if (i == (*num_curv)) {
    if (MAX_TRIM_CURVES_PER_PATCH == (*num_curv)) {
      printf("MergeSplines: Too many trim curves in patch\n");
      exit(-1);
    }
    (*m_curve)[(*num_curv)].int_curve = new_crv;
    (*num_curv)++;
  } 
  return new_crv;
}

int SplineSense(spl1, spl2)
XYZ_SPLINE_TYPE *spl1, *spl2;
{
  POINT3D p1, p2, p3, p4;
  int deg1, deg2;
  
  deg1 = spl1->Deg;
  deg2 = spl2->Deg;

  p1.dCoord[0] = spl1->CntrlPts[0].dCoord[0] / spl1->CntrlPts[0].dCoord[3];
  p1.dCoord[1] = spl1->CntrlPts[0].dCoord[1] / spl1->CntrlPts[0].dCoord[3];
  p1.dCoord[2] = spl1->CntrlPts[0].dCoord[2] / spl1->CntrlPts[0].dCoord[3];
  p2.dCoord[0] = spl1->CntrlPts[deg1].dCoord[0] / spl1->CntrlPts[deg1].dCoord[3];
  p2.dCoord[1] = spl1->CntrlPts[deg1].dCoord[1] / spl1->CntrlPts[deg1].dCoord[3];
  p2.dCoord[2] = spl1->CntrlPts[deg1].dCoord[2] / spl1->CntrlPts[deg1].dCoord[3];

  p3.dCoord[0] = spl2->CntrlPts[0].dCoord[0] / spl2->CntrlPts[0].dCoord[3];
  p3.dCoord[1] = spl2->CntrlPts[0].dCoord[1] / spl2->CntrlPts[0].dCoord[3];
  p3.dCoord[2] = spl2->CntrlPts[0].dCoord[2] / spl2->CntrlPts[0].dCoord[3];
  p4.dCoord[0] = spl2->CntrlPts[deg2].dCoord[0] / spl2->CntrlPts[deg2].dCoord[3];
  p4.dCoord[1] = spl2->CntrlPts[deg2].dCoord[1] / spl2->CntrlPts[deg2].dCoord[3];
  p4.dCoord[2] = spl2->CntrlPts[deg2].dCoord[2] / spl2->CntrlPts[deg2].dCoord[3];

  if (EQUAL3D_ABS(p1, p3, 5e-4) && EQUAL3D_ABS(p2, p4, 5e-4))
    return 1;

  if (EQUAL3D_ABS(p1, p4, 5e-4) && EQUAL3D_ABS(p2, p3, 5e-4))
    return 0;
  
  printf("SplineSense: Endpoints should have matched\n");
  exit(-1);
}

/* Given a merged curve, it merges it with a set of merged curves. */
INTERSECTION_CURVE *MergeSplinesFurther(m_curve, num_curv, new_m_curve, error_correct)
MERGED_CURVE **m_curve, *new_m_curve;
int *num_curv;
int error_correct;
{
  int p, q;
  int i, j, k;
  INTERSECTION_CURVE *new_crv, *nf_ptr, *nl_ptr;
  INTERSECTION_CURVE *tptr, *f_ptr, *l_ptr, *f_ptr2, *l_ptr2;
  POINT2D comp_ep0, comp_ep1;
  POINT2D st_pt, end_pt, st_pt2, end_pt2;
  double dist0, dist1;
  double wcrd;
  int close0, close1, tmp_deg, dir;
  int degenerate_case; /* = 1 when the two endpts of comp are
			  within 2e-5 dist. apart */

  new_crv = new_m_curve->int_curve;
  return_st_end_in_p2d(new_m_curve, &comp_ep0, &comp_ep1, &nf_ptr, &nl_ptr);

  if (EQUAL2D_EPS(comp_ep0, comp_ep1, 1e-8))
    return new_m_curve->int_curve;

  for (i=0; i<(*num_curv); i++) {
    return_st_end_in_p2d(&((*m_curve)[i]), &st_pt, &end_pt, &f_ptr, &l_ptr);
    close0 = closer_pt(&st_pt, &comp_ep0, &comp_ep1, &dist0);
    close1 = closer_pt(&end_pt, &comp_ep0, &comp_ep1, &dist1);

    if (EQUAL2D_EPS(comp_ep0, st_pt, 5e-4) && close0 == 0
	&& dist0 <= dist1) { 
      reverse_int_curve(&new_crv, &tptr);

      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* satement can be removed if it leads to major problems */
      tmp_deg = nl_ptr->uv_spline.Deg;
      wcrd = nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[2];
      nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
      nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = nl_ptr->xyz_spline.Deg;
	dir = SplineSense(&nl_ptr->xyz_spline, &nl_ptr->other_solid_ptr->xyz_spline);
	nl_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr->xyz_spline.CntrlPts[0];
	nl_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*nl_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr->xyz_spline.CntrlPts[0]; 
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the front of the list */
      (*m_curve)[i].int_curve = new_crv;
      nl_ptr->next = f_ptr;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      st_pt = comp_ep1;
      f_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(st_pt, st_pt2, 5e-4) && close0 == 0 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = st_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    dir = SplineSense(&f_ptr->xyz_spline, &f_ptr->other_solid_ptr->xyz_spline);
	    f_ptr->xyz_spline.CntrlPts[0] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*f_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr2->xyz_spline.CntrlPts[0];
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  tptr->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(st_pt, end_pt2, 5e-4) && close0 == 1 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    dir = SplineSense(&f_ptr->xyz_spline, &f_ptr->other_solid_ptr->xyz_spline);
	    f_ptr->xyz_spline.CntrlPts[0] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	    f_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*f_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	  }

	  l_ptr2->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep0, end_pt, 5e-4) && close1 == 0
	       && dist1 <= dist0) {
      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      wcrd = nf_ptr->uv_spline.CntrlPts[0].dCoord[2];
      nf_ptr->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
      nf_ptr->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = l_ptr->xyz_spline.Deg;
	dir = SplineSense(&nf_ptr->xyz_spline, &nf_ptr->other_solid_ptr->xyz_spline);
	nf_ptr->xyz_spline.CntrlPts[0] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	nf_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*nf_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the end of the list */
      l_ptr->next = new_crv;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      end_pt = comp_ep1;
      l_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(end_pt, st_pt2, 5e-4) && close1 == 0 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr->xyz_spline.Deg;
	    dir = SplineSense(&l_ptr->xyz_spline, &l_ptr->other_solid_ptr->xyz_spline);
	    l_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr2->xyz_spline.CntrlPts[0];
	    l_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*l_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr2->xyz_spline.CntrlPts[0];
	  }

	  l_ptr->next = f_ptr2;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(end_pt, end_pt2, 5e-4) && close1 == 1 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = end_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    dir = SplineSense(&l_ptr->xyz_spline, &l_ptr->other_solid_ptr->xyz_spline);
	    l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg] =
	      l_ptr2->xyz_spline.CntrlPts[tmp_deg];   
	    l_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*l_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr2->xyz_spline.CntrlPts[tmp_deg]; 
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  l_ptr->next = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep1, st_pt, 5e-4) && close0 == 1
	       && dist0 <= dist1) {  
      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      tmp_deg = nl_ptr->uv_spline.Deg;
      wcrd = nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[2];
      nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
      nl_ptr->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = nl_ptr->xyz_spline.Deg;
	dir = SplineSense(&nl_ptr->xyz_spline, &nl_ptr->other_solid_ptr->xyz_spline);
	nl_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr->xyz_spline.CntrlPts[0];
	nl_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*nl_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr->xyz_spline.CntrlPts[0];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the beginning of the list */
      new_crv->next = f_ptr;
      (*m_curve)[i].int_curve = new_crv;

      /* reassign st_pt to the new starting pt - it was the first point */
      /* of the original curve */
      st_pt = comp_ep0;
      f_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(st_pt, st_pt2, 5e-4) && close0 == 0 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = st_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    dir = SplineSense(&f_ptr->xyz_spline, &f_ptr->other_solid_ptr->xyz_spline);
	    f_ptr->xyz_spline.CntrlPts[0] = f_ptr2->xyz_spline.CntrlPts[0];
	    f_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*f_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr2->xyz_spline.CntrlPts[0];
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  tptr->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(st_pt, end_pt2, 5e-4) && close0 == 1 && 
	    dist0 <= dist1) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = st_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = st_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    dir = SplineSense(&f_ptr->xyz_spline, &f_ptr->other_solid_ptr->xyz_spline);
	    f_ptr->xyz_spline.CntrlPts[0] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	    f_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*f_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr2->xyz_spline.CntrlPts[tmp_deg];
	  }

	  l_ptr2->next = (*m_curve)[i].int_curve;
	  (*m_curve)[i].int_curve = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	}
      }
      break;
    } else if (EQUAL2D_EPS(comp_ep1, end_pt, 5e-4) && close1 == 1
	       && dist1 <= dist0) {
      reverse_int_curve(&new_crv, &tptr);

      /* I am going to make the control point in the new_crv equal to */
      /* the bigger curve to prevent error problems later. This */
      /* statement can be removed if it leads to major problems */
      wcrd = nf_ptr->uv_spline.CntrlPts[0].dCoord[2];
      nf_ptr->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
      nf_ptr->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
      if (error_correct) {
	tmp_deg = l_ptr->xyz_spline.Deg;
	dir = SplineSense(&nf_ptr->xyz_spline, &nf_ptr->other_solid_ptr->xyz_spline);
	nf_ptr->xyz_spline.CntrlPts[0] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
	nf_ptr->other_solid_ptr->xyz_spline.CntrlPts[(1-dir)*nf_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr->xyz_spline.CntrlPts[tmp_deg];
      }
      /* -------------  ENDS HERE ---------------------- */

      /* Putting new curve at the end of the list */
      l_ptr->next = new_crv;

      /* reassign st_pt to the new starting pt - it was the last point */
      /* of the original curve */
      end_pt = comp_ep0;
      l_ptr = new_crv;

      for (j=i+1; j<(*num_curv); j++) {
	return_st_end_in_p2d(&((*m_curve)[j]), &st_pt2, &end_pt2,
			     &f_ptr2, &l_ptr2); 
	close0 = closer_pt(&st_pt, &st_pt2, &end_pt2, &dist0);
	close1 = closer_pt(&end_pt, &st_pt2, &end_pt2, &dist1);

	if (EQUAL2D_EPS(end_pt, st_pt2, 5e-4) && close1 == 0 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  wcrd = f_ptr2->uv_spline.CntrlPts[0].dCoord[2];
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[0] = end_pt.u * wcrd;
	  f_ptr2->uv_spline.CntrlPts[0].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr->xyz_spline.Deg;
	    dir = SplineSense(&l_ptr->xyz_spline, &l_ptr->other_solid_ptr->xyz_spline);
	    l_ptr->xyz_spline.CntrlPts[tmp_deg] = f_ptr2->xyz_spline.CntrlPts[0];
	    l_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*l_ptr->other_solid_ptr->xyz_spline.Deg] = f_ptr2->xyz_spline.CntrlPts[0];
	  }

	  l_ptr->next = f_ptr2;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} else
	if (EQUAL2D_EPS(end_pt, end_pt2, 5e-4) && close1 == 1 && 
	    dist1 <= dist0) {
	  printf("2nd Merge: \n");
	  /* Making close points equal */
	  tmp_deg = l_ptr2->uv_spline.Deg;
	  wcrd = l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[2];
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[0] = end_pt.u * wcrd;
	  l_ptr2->uv_spline.CntrlPts[tmp_deg].dCoord[1] = end_pt.v * wcrd;
	  if (error_correct) {
	    tmp_deg = l_ptr2->xyz_spline.Deg;
	    dir = SplineSense(&l_ptr->xyz_spline, &l_ptr->other_solid_ptr->xyz_spline);
	    l_ptr->xyz_spline.CntrlPts[l_ptr->xyz_spline.Deg] =
	      l_ptr2->xyz_spline.CntrlPts[tmp_deg];  
	    l_ptr->other_solid_ptr->xyz_spline.CntrlPts[dir*l_ptr->other_solid_ptr->xyz_spline.Deg] = l_ptr2->xyz_spline.CntrlPts[tmp_deg]; 
	  }

	  reverse_int_curve(&(*m_curve)[j].int_curve, &tptr);
	  l_ptr->next = (*m_curve)[j].int_curve;
	  for (k=j+1; k<(*num_curv); k++)
	    (*m_curve)[k-1].int_curve = (*m_curve)[k].int_curve; 
	  (*num_curv)--;
	  break;
	} 
      }
      break;
    }
  }

  if (i == (*num_curv)) {
    if (MAX_TRIM_CURVES_PER_PATCH == (*num_curv)) {
      printf("MergeSplinesFurther: Too many trim curves in patch\n");
      exit(-1);
    }
    (*m_curve)[(*num_curv)].int_curve = new_crv;
    (*num_curv)++;
  } 
  return new_crv;
}

/* The next set of three routines is called by qsort for sorting a */
/* bunch 3d points by x. y and z are of lesser priority */
static int comparZ(f1, f2)
CURVE_MATCHING_TYPE *f1, *f2;
{
  if (FP_EQ_EPS(f1->pt3d.dCoord[2], f2->pt3d.dCoord[2], 1e-2)) return 0;
  if (f1->pt3d.dCoord[2] < f2->pt3d.dCoord[2]) return -1;
  if (f1->pt3d.dCoord[2] > f2->pt3d.dCoord[2]) return 1;
}

static int comparY(f1, f2)
CURVE_MATCHING_TYPE *f1, *f2;
{
  if (FP_EQ_EPS(f1->pt3d.dCoord[1], f2->pt3d.dCoord[1], 1e-2)) 
    return comparZ(f1, f2);
  if (f1->pt3d.dCoord[1] < f2->pt3d.dCoord[1]) return -1;
  if (f1->pt3d.dCoord[1] > f2->pt3d.dCoord[1]) return 1;
}

static int comparX(f1, f2)
CURVE_MATCHING_TYPE *f1, *f2;
{
  if (FP_EQ_EPS(f1->pt3d.dCoord[0], f2->pt3d.dCoord[0], 1e-2)) 
    return comparY(f1, f2);
  if (f1->pt3d.dCoord[0] < f2->pt3d.dCoord[0]) return -1;
  if (f1->pt3d.dCoord[0] > f2->pt3d.dCoord[0]) return 1;
}

/* This routine returns the index which holds the other endpoint of */
/* the same curve */
short GetIndexofCorrespondingPatch(cmt, num, index)
CURVE_MATCHING_TYPE *cmt;
int num;
short index;
{
  int i, j;
  short pNo, cNo, ending, ret_idx;

  pNo = cmt[index].patchNo;
  cNo = cmt[index].curve_num;
  ending = cmt[index].s_or_e;

  ret_idx = -1;
  for (i=0; i<num; i++) {
    if (i == index) continue;
    if ((cmt[i].patchNo == pNo) && (cmt[i].curve_num == cNo) &&
	(cmt[i].s_or_e != ending)) {
      ret_idx = i;
      break;
    }
  }
  if (ret_idx == -1) {
    printf("GetIndexofCorrespondingPatch: cmt array is corrupted\n");
    exit(-1);
  }
  return ret_idx;
}

/* Returns the next uncovered index (basically 0 element) in an array. */
/* returns -1 if none is 0 */
short NextUncoveredIndex(arr, num)
short *arr;
int num;
{
  short i;

  for (i=0; i<num; i++) 
    if (!arr[i]) break;

  if (i < (short) num)
    return i;
  return -1;
}

/* This routine checks if the intersection curves when put together */
/* form a closed loop or not, and also returns the number of */
/* components in each solid. They must be the same in both solids. */
/* If (flag == 1) then I actually do not check if the curves are */
/* closed in space. I am just calling this function to fill in the */
/* fXsc and sXsc arrays. flag is 1 if the solids are overlapping in */
/* the opposite direction */
int CheckIntersectionCurves(fs, fc, fnum, fs_nc, fcmt, fcmt_num,
			    ss, sc, snum, ss_nc, scmt, scmt_num,
			    flag)
TRIM_PATCH_INFO **fs, **ss;
MERGED_CURVE **fc, **sc;
int fnum, snum, *fs_nc, *ss_nc, *fcmt_num, *scmt_num, flag;
CURVE_MATCHING_TYPE *fcmt, *scmt; /* these arrays are already */
				  /* allocated, 100 elements only */
{
  int i, j, k, l, deg, chck_comp, cnt, another_flag, new_flag;
  POINT2D spt, ept;
  double wcrd; /* reciprocal of w coord */
  INTERSECTION_CURVE *fptr, *lptr, *fptr1, *lptr1;
  short covered_index[100]; /* temporary array to store if I have taken */
			  /* care of that index or not */
  short s_index, other_idx, indx1, indx2, neighb_ptch;
  TRIM_PORTION_SPLINE_TYPE *tmp_tpst;
  short did_reversal = 0;
  int min_ind;
  double min_d, tmp_d;

  another_flag = 0;
  if ((fnum >= 100) || (snum >= 100)) {
    printf("CheckIntersectionCurves:not enough space in fXsect.. arrays\n");
    exit(-1);
  }

  for (k=0,i=0; i<fnum; i++)
    for (j=0; j<fs_nc[i]; j++) {
      if (k == 100) {
	printf("CheckIntersectionCurves: too many points to merge\n");
	exit(-1);
      }
      return_st_end_in_p2d(&fc[i][j], &spt, &ept, &fptr, &lptr);
      wcrd = 1 / fptr->xyz_spline.CntrlPts[0].dCoord[3];
      fcmt[k].pt3d.dCoord[0] = fptr->xyz_spline.CntrlPts[0].dCoord[0] * wcrd;
      fcmt[k].pt3d.dCoord[1] = fptr->xyz_spline.CntrlPts[0].dCoord[1] * wcrd;
      fcmt[k].pt3d.dCoord[2] = fptr->xyz_spline.CntrlPts[0].dCoord[2] * wcrd;
      fcmt[k].patchNo = (short) i;
      fcmt[k].s_or_e = 0;
      fcmt[k].curve_num = (short) j;
      fcmt[k++].marked = -1;
	
      if (k == 100) {
	printf("CheckIntersectionCurves: too many points to merge\n");
	exit(-1);
      }
      deg = lptr->xyz_spline.Deg;
      wcrd = 1 / lptr->xyz_spline.CntrlPts[deg].dCoord[3];
      fcmt[k].pt3d.dCoord[0] = lptr->xyz_spline.CntrlPts[deg].dCoord[0] * wcrd;
      fcmt[k].pt3d.dCoord[1] = lptr->xyz_spline.CntrlPts[deg].dCoord[1] * wcrd;
      fcmt[k].pt3d.dCoord[2] = lptr->xyz_spline.CntrlPts[deg].dCoord[2] * wcrd;
      fcmt[k].patchNo = (short) i;
      fcmt[k].s_or_e = 1;
      fcmt[k].curve_num = (short) j;
      fcmt[k++].marked = -1;  
    }
    
  if (k > 0) {
    /* k is the number of curves in the whole solid. now sort the x */
    /* coordinates */ 
    *fcmt_num = k;
    qsort(&fcmt[0], k, sizeof(CURVE_MATCHING_TYPE), comparX);
    i = 0;
    while (i < k) {
      if (fcmt[i].marked != -1) {
	i++;
	continue;
      }
      new_flag = 0;
      if (flag)
	for (l=0; l<NumOvrPatches; l++)
	  if (overlap_pairs[l].ptches.patchNum0 == fcmt[i].patchNo) {
	    new_flag = 1;
	    break;
	  }

      j = i+1;
      min_d = 1e+20;
      min_ind = -1;
      while (j<k) {
	if (fcmt[j].marked == -1) {
	  tmp_d = dist3D(&fcmt[i].pt3d, &fcmt[j].pt3d);
	  if (tmp_d < min_d) {
	    min_d = tmp_d;
	    min_ind = j;
	  }
	}
	j++;
      }
      if (min_ind != -1)
	j = min_ind;
      else {
	printf("CheckIntersectionCurves: Endpoints dont match in space\n");
	exit(-1);
      }

#if 0
      if ((j != i+1) && !new_flag) { /* condition was j == k, but this is stronger */
	printf("CheckIntersectionCurves: j==k: Not properly closed\n");
	return 0;
      }
#endif
      while (!EQUAL3D_EPS(fcmt[i].pt3d, fcmt[j].pt3d, 2e-5) &&
	     !new_flag && (j < k)) 
	j++;
      if (j == k) {
	printf("CheckIntersectionCurves: Not properly closed\n");
	return 0;
      }
      /* fcmt[i] and fcmt[j] are neighbors */
      fcmt[i].marked = j;
      fcmt[j].marked = i;
      i = i+1;
    }
    if (i != k) {
      printf("CheckIntersectionCurves: Something wrong in while loop\n");
      exit(-1);
    }
    
    /* go from the start of the fcmt array in a proper */
    /* fashion and reverse curves that have two endpts or two startpts */
    for (i=0; i<k; i++)
      covered_index[i] = 0;
    s_index = 0;
    while (s_index != -1) {
      /* error correction step */
      INTERSECTION_CURVE *fptr2, *lptr2;
      /* end - error correction step */

      if (fcmt[s_index].s_or_e)
	s_index = GetIndexofCorrespondingPatch(&fcmt[0], k, s_index);

      /* error correction step */
      return_st_end_in_p2d(&fc[fcmt[s_index].patchNo][fcmt[s_index].curve_num], &spt, &ept, &fptr2, &lptr2);
      /* end - error correction step */

      covered_index[s_index] = 1;
      indx2 = GetIndexofCorrespondingPatch(&fcmt[0], k, s_index);
      covered_index[indx2] = 1;

      /* error correction step */
      return_st_end_in_p2d(&fc[fcmt[indx2].patchNo][fcmt[indx2].curve_num], &spt, &ept, &fptr, &lptr);
      /* end - error correction step */

      indx1 = fcmt[indx2].marked;
      while (indx1 != s_index) {
	/* From now on, indx2 always will be the index of the ending */
	/* point */
	if (fcmt[indx2].s_or_e == fcmt[indx1].s_or_e) {
	  reverse_int_curve(
			    &fc[fcmt[indx1].patchNo][fcmt[indx1].curve_num].int_curve, 
			    &lptr1);
	  did_reversal = 1;
	  indx2 = GetIndexofCorrespondingPatch(&fcmt[0], k, indx1);
	  if (fcmt[indx1].s_or_e) {
	    fcmt[indx1].s_or_e = 0;
	    fcmt[indx2].s_or_e = 1;
	  } else {
	    fcmt[indx1].s_or_e = 1;
	    fcmt[indx2].s_or_e = 0;
	  }
	}
	covered_index[indx1] = 1;

	/* error correction step */
	return_st_end_in_p2d(&fc[fcmt[indx1].patchNo][fcmt[indx1].curve_num], &spt, &ept, &fptr1, &lptr1);
	if (!fcmt[indx1].s_or_e) {
	  int direct = -1;
	  POINT3D p1, p2, p3;

	  if (fptr1->other_solid_ptr != NULL) {
	    p1.dCoord[0] = fptr1->xyz_spline.CntrlPts[0].dCoord[0] /
	      fptr1->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p1.dCoord[1] = fptr1->xyz_spline.CntrlPts[0].dCoord[1] /
	      fptr1->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p1.dCoord[2] = fptr1->xyz_spline.CntrlPts[0].dCoord[2] /
	      fptr1->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p2.dCoord[0] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[0] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p2.dCoord[1] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[1] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p2.dCoord[2] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[2] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    deg = fptr1->other_solid_ptr->xyz_spline.Deg;
	    p3.dCoord[0] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[0] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p3.dCoord[1] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[1] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p3.dCoord[2] = fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[2] /
	      fptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    if ((direct == -1) && EQUAL3D_ABS(p1, p2, 5e-4))
	      direct = 1;
	    if ((direct == -1) && EQUAL3D_ABS(p1, p3, 5e-4))
	      direct = 0;
	    if (direct == -1) {
	      printf("I give up!! Should have matched\n");
	      exit(-1);
	    }
	  }
	  /* match lptr and fptr1 */
	  deg = lptr->xyz_spline.Deg;
	  if (fptr1->other_solid_ptr != NULL)
	    fptr1->other_solid_ptr->xyz_spline.CntrlPts[(1-direct)*fptr1->other_solid_ptr->xyz_spline.Deg] = lptr->xyz_spline.CntrlPts[deg];
	  fptr1->xyz_spline.CntrlPts[0] = lptr->xyz_spline.CntrlPts[deg];
	} else {
	  int direct = -1;
	  POINT3D p1, p2, p3;

	  /* match fptr and lptr1 */
	  if (lptr1->other_solid_ptr != NULL) {
	    deg = lptr1->xyz_spline.Deg;
	    p1.dCoord[0] = lptr1->xyz_spline.CntrlPts[deg].dCoord[0] /
	      lptr1->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p1.dCoord[1] = lptr1->xyz_spline.CntrlPts[deg].dCoord[1] /
	      lptr1->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p1.dCoord[2] = lptr1->xyz_spline.CntrlPts[deg].dCoord[2] /
	      lptr1->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p3.dCoord[0] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[0] /
	      lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p3.dCoord[1] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[1] /
	      lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    p3.dCoord[2] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[2] /
	    lptr1->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	    deg = lptr1->other_solid_ptr->xyz_spline.Deg;
	    p2.dCoord[0] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[0] /
	      lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p2.dCoord[1] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[1] /
	      lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    p2.dCoord[2] = lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[2] /
	    lptr1->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	    if ((direct == -1) && EQUAL3D_ABS(p1, p2, 5e-4))
	      direct = 1;
	    if ((direct == -1) && EQUAL3D_ABS(p1, p3, 5e-4))
	      direct = 0;
	    if (direct == -1) {
	      printf("I give up!! Should have matched\n");
	      exit(-1);
	    }
	    lptr1->other_solid_ptr->xyz_spline.CntrlPts[direct*deg] =
	      fptr->xyz_spline.CntrlPts[0]; 
	  }
	  deg = lptr1->xyz_spline.Deg;
	  lptr1->xyz_spline.CntrlPts[deg] = fptr->xyz_spline.CntrlPts[0];
	}
	fptr = fptr1;
	lptr = lptr1;
	/* end - error correction step */
	
	if (!did_reversal)
	  indx2 = GetIndexofCorrespondingPatch(&fcmt[0], k, indx1);
	else
	  did_reversal = 0;
	covered_index[indx2] = 1;
	indx1 = fcmt[indx2].marked;
      }

      /* error correction step */
      return_st_end_in_p2d(&fc[fcmt[indx2].patchNo][fcmt[indx2].curve_num], &spt, &ept, &fptr, &lptr);
      if (fcmt[indx2].s_or_e) {
	int direct = -1;
	POINT3D p1, p2, p3;
	
	if (fptr2->other_solid_ptr != NULL) {
	  p1.dCoord[0] = fptr2->xyz_spline.CntrlPts[0].dCoord[0] /
	    fptr2->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p1.dCoord[1] = fptr2->xyz_spline.CntrlPts[0].dCoord[1] /
	    fptr2->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p1.dCoord[2] = fptr2->xyz_spline.CntrlPts[0].dCoord[2] /
	    fptr2->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p2.dCoord[0] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[0] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p2.dCoord[1] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[1] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p2.dCoord[2] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[2] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  deg = fptr2->other_solid_ptr->xyz_spline.Deg;
	  p3.dCoord[0] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[0] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p3.dCoord[1] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[1] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p3.dCoord[2] = fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[2] /
	    fptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  if ((direct == -1) && EQUAL3D_ABS(p1, p2, 5e-4))
	    direct = 1;
	  if ((direct == -1) && EQUAL3D_ABS(p1, p3, 5e-4))
	    direct = 0;
	  if (direct == -1) {
	    printf("I give up!! Should have matched\n");
	    exit(-1);
	  }
	}
	/* match lptr and fptr2 */
	deg = lptr->xyz_spline.Deg;
	if (fptr2->other_solid_ptr != NULL)
	  fptr2->other_solid_ptr->xyz_spline.CntrlPts[(1-direct)*fptr2->other_solid_ptr->xyz_spline.Deg] = lptr->xyz_spline.CntrlPts[deg];
	fptr2->xyz_spline.CntrlPts[0] = lptr->xyz_spline.CntrlPts[deg];
      } else {
	int direct = -1;
	POINT3D p1, p2, p3;
	
	/* match fptr and lptr2 */
	if (lptr2->other_solid_ptr != NULL) {
	  deg = lptr2->xyz_spline.Deg;
	  p1.dCoord[0] = lptr2->xyz_spline.CntrlPts[deg].dCoord[0] /
	    lptr2->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p1.dCoord[1] = lptr2->xyz_spline.CntrlPts[deg].dCoord[1] /
	    lptr2->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p1.dCoord[2] = lptr2->xyz_spline.CntrlPts[deg].dCoord[2] /
	    lptr2->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p3.dCoord[0] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[0] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p3.dCoord[1] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[1] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  p3.dCoord[2] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[2] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[0].dCoord[3]; 
	  deg = lptr2->other_solid_ptr->xyz_spline.Deg;
	  p2.dCoord[0] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[0] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p2.dCoord[1] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[1] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  p2.dCoord[2] = lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[2] /
	    lptr2->other_solid_ptr->xyz_spline.CntrlPts[deg].dCoord[3]; 
	  if ((direct == -1) && EQUAL3D_ABS(p1, p2, 5e-4))
	    direct = 1;
	  if ((direct == -1) && EQUAL3D_ABS(p1, p3, 5e-4))
	    direct = 0;
	  if (direct == -1) {
	    printf("I give up!! Should have matched\n");
	    exit(-1);
	  }
	  lptr2->other_solid_ptr->xyz_spline.CntrlPts[direct*deg] =
	    fptr->xyz_spline.CntrlPts[0]; 
	}
	/* match fptr and lptr2 */
	deg = lptr2->xyz_spline.Deg;
	lptr2->xyz_spline.CntrlPts[deg] = fptr->xyz_spline.CntrlPts[0];
      }
      /* end - error correction step */

      s_index = NextUncoveredIndex(&covered_index[0], k);
    }
    /* Now fill the fXsc array */
    for (i=0; i<k; i++) {
      neighb_ptch = fcmt[fcmt[i].marked].patchNo;
      for (j=0; j<(*fs)[fcmt[i].patchNo].NumTrimPortions; j++) {
	tmp_tpst = &(*fs)[fcmt[i].patchNo].TrimPortion[j].SplinePortion;
	for (cnt=0,l=0; l<tmp_tpst->NumOfCurves; l++)
	  if (tmp_tpst->AdjInfo[l] == neighb_ptch)
	    if (cnt < 8)
	      fXsc[fcmt[i].patchNo][fcmt[i].curve_num][fcmt[i].s_or_e][cnt++] = l;
	    else {
	      printf("CheckIntersectionCurves: 8 not enough for last dimension of array\n");
	      exit(-1);
	    }
      }
    }
  }

  for (i=0; i<NumOvrPatches; i++) {
    if (first_complete_ovrlap[i] && !second_complete_ovrlap[i]) {
      another_flag = 1;
      break;
    }
    if (!first_complete_ovrlap[i] && second_complete_ovrlap[i]) {
      another_flag = 1;
      break;
    }
  }

  for (k=0,i=0; i<snum; i++)
    for (j=0; j<ss_nc[i]; j++) {
      if (k == 100) {
	printf("CheckIntersectionCurves: too many points to merge\n");
	exit(-1);
      }
      return_st_end_in_p2d(&sc[i][j], &spt, &ept, &fptr, &lptr);
      wcrd = 1 / fptr->xyz_spline.CntrlPts[0].dCoord[3];
      scmt[k].pt3d.dCoord[0] = fptr->xyz_spline.CntrlPts[0].dCoord[0] * wcrd;
      scmt[k].pt3d.dCoord[1] = fptr->xyz_spline.CntrlPts[0].dCoord[1] * wcrd;
      scmt[k].pt3d.dCoord[2] = fptr->xyz_spline.CntrlPts[0].dCoord[2] * wcrd;
      scmt[k].patchNo = (short) i;
      scmt[k].s_or_e = 0;
      scmt[k].curve_num = (short) j;
      scmt[k++].marked = -1;
      
      if (k == 100) {
	printf("CheckIntersectionCurves: too many points to merge\n");
	exit(-1);
      }
      deg = lptr->xyz_spline.Deg;
      wcrd = 1 / lptr->xyz_spline.CntrlPts[deg].dCoord[3];
      scmt[k].pt3d.dCoord[0] = lptr->xyz_spline.CntrlPts[deg].dCoord[0] * wcrd;
      scmt[k].pt3d.dCoord[1] = lptr->xyz_spline.CntrlPts[deg].dCoord[1] * wcrd;
      scmt[k].pt3d.dCoord[2] = lptr->xyz_spline.CntrlPts[deg].dCoord[2] * wcrd;
      scmt[k].patchNo = (short) i;
      scmt[k].s_or_e = 1;
      scmt[k].curve_num = (short) j;
      scmt[k++].marked = -1;  
    }
  
  if (k > 0) {
    /* k is the number of curves in the whole solid now sort the x */
    /* coordinates */ 
    *scmt_num = k;
    qsort(&scmt[0], k, sizeof(CURVE_MATCHING_TYPE), comparX);
    i = 0;
    while (i < k) {
      if (scmt[i].marked != -1) {
	i++;
	continue;
      }
      new_flag = 0;
      if (flag)
	for (l=0; l<NumOvrPatches; l++)
	  if (overlap_pairs[l].ptches.patchNum1 == scmt[i].patchNo) {
	    new_flag = 1;
	    break;
	  }

      j = i+1;
      min_d = 1e+20;
      min_ind = -1;
      while (j<k) {
	if (scmt[j].marked == -1) {
	  tmp_d = dist3D(&scmt[i].pt3d, &scmt[j].pt3d);
	  if (tmp_d < min_d) {
	    min_d = tmp_d;
	    min_ind = j;
	  }
	}
	j++;
      }
      if (min_ind != -1)
	j = min_ind;
      else {
	printf("CheckIntersectionCurves: Endpoints dont match in space\n");
	exit(-1);
      }

#if 0
      if ((j != i+1) && !new_flag) { /* condition was j == k, but this is stronger */
	printf("CheckIntersectionCurves: j==k: Not properly closed\n");
	return 0;
      }
#endif
      while (!EQUAL3D_EPS(scmt[i].pt3d, scmt[j].pt3d, 2e-5) &&
	     !new_flag && (j < k)) 
	j++;
      if (j == k) {
	printf("CheckIntersectionCurves: Not properly closed\n");
	return 0;
      }
      /* scmt[i] and scmt[j] are neighbors */
      scmt[i].marked = j;
      scmt[j].marked = i;
      i = i+1;
    }
    if (i != k) {
      printf("CheckIntersectionCurves: Something wrong in while loop\n");
      exit(-1);
    }
    
    /* go from the start of the scmt array in a proper */
    /* fashion and reverse curves that have two endpts or two startpts */
    for (i=0; i<k; i++)
      covered_index[i] = 0;
    s_index = 0;
    while (s_index != -1) {
      if (scmt[s_index].s_or_e)
	s_index = GetIndexofCorrespondingPatch(&scmt[0], k, s_index);

      covered_index[s_index] = 1;
      indx2 = GetIndexofCorrespondingPatch(&scmt[0], k, s_index);
      covered_index[indx2] = 1;

      indx1 = scmt[indx2].marked;
      while (indx1 != s_index) {
	/* From now on, indx2 always will be the index of the ending */
	/* point */
	if (scmt[indx2].s_or_e == scmt[indx1].s_or_e) {
	  reverse_int_curve(
			    &sc[scmt[indx1].patchNo][scmt[indx1].curve_num].int_curve, 
			    &lptr1);
	  did_reversal = 1;
	  indx2 = GetIndexofCorrespondingPatch(&scmt[0], k, indx1);
	  if (scmt[indx1].s_or_e) {
	    scmt[indx1].s_or_e = 0;
	    scmt[indx2].s_or_e = 1;
	  } else {
	    scmt[indx1].s_or_e = 1;
	    scmt[indx2].s_or_e = 0;
	  }
	}
	covered_index[indx1] = 1;

	if (!did_reversal)
	  indx2 = GetIndexofCorrespondingPatch(&scmt[0], k, indx1);
	else
	  did_reversal = 0;
	covered_index[indx2] = 1;
	indx1 = scmt[indx2].marked;
      }

      s_index = NextUncoveredIndex(&covered_index[0], k);
    }
    /* Now fill the sXsc array */
    for (i=0; i<k; i++) {
      neighb_ptch = scmt[scmt[i].marked].patchNo;
      for (j=0; j<(*ss)[scmt[i].patchNo].NumTrimPortions; j++) {
	tmp_tpst = &(*ss)[scmt[i].patchNo].TrimPortion[j].SplinePortion;
	for (cnt=0,l=0; l<tmp_tpst->NumOfCurves; l++)
	  if (tmp_tpst->AdjInfo[l] == neighb_ptch)
	    if (cnt < 3)
	      sXsc[scmt[i].patchNo][scmt[i].curve_num][scmt[i].s_or_e][cnt++] = l;
	    else {
	      printf("CheckIntersectionCurves: 3 not enough for last dimension of array\n");
	      exit(-1);
	    }
      }
    }
  }

  return 1;
}

/* This routines returns the value of parameter of the curve that */
/* evaluates to pt2d */
int EvaluateParam(uvc, pt2d, param)
UV_SPLINE_TYPE *uvc;
POINT2D *pt2d;
double *param;
{
  int i, j, k;
  int deg, act_deg;
  long nci;
  double coeffs[6]; /* 5 is max degree of curve allowed */
  double **mat, rec;
  double *Re, *Im, **Vec;
  POINT2D tpt2d;
  
  deg = uvc->Deg;
  if (deg > 5) {
    printf("EvaluateParam: Curve has high degree. Cannot handle\n");
    exit(-1);
  }
  for (i=0; i<=deg; i++) {
    nci = combChoose(deg, i);
    coeffs[i] = nci * (uvc->CntrlPts[i].dCoord[0] - 
		       pt2d->u * uvc->CntrlPts[i].dCoord[2]);
    /* Basically converting to power reprn using t/(1-t) */
  }
  /* To prevent any problems make sure that the leading coeffs is */
  /* non-zero */
  act_deg = deg;
  while ((act_deg >= 0) && FP_EQ_EPS(coeffs[act_deg], 0.0, 1e-4))
    act_deg--;

  if (act_deg < 0) { /* cannot get using u, trying v */
    for (i=0; i<=deg; i++) {
      nci = combChoose(deg, i);
      coeffs[i] = nci * (uvc->CntrlPts[i].dCoord[1] - 
			 pt2d->v * uvc->CntrlPts[i].dCoord[2]);
      /* Basically converting to power reprn using t/(1-t) */
    }
    /* To prevent any problems make sure that the leading coeffs is */
    /* non-zero */
    act_deg = deg;
    while ((act_deg >= 0) && FP_EQ_EPS(coeffs[act_deg], 0.0, 1e-4))
      act_deg--;
  }

  if (act_deg < 0) {
    printf("EvaluateParam: Unable to set up equation. Quitting ..\n");
    exit(-1);
  } else if (act_deg == 0) {
    *param = 1.0; /* this would mean that the equation would actually */
		  /* be like coeffs[0]*(1-t)^deg = 0 => t = 1.0 */
    return 1;
  }

  ALLOCMAT(mat, act_deg, act_deg, double);
  ALLOCMAT(Vec, act_deg, act_deg, double);
  ALLOCN(Re, double, act_deg);
  ALLOCN(Im, double, act_deg);

  for (i=0; i<(act_deg-1); i++)
    for (j=0; j<act_deg; j++) {
      if (j == (i+1)) mat[i][j] = 1.0;
      else mat[i][j] = 0.0;
    }
  rec = 1.0 / coeffs[act_deg];
  for (j=0; j<act_deg; j++)
    mat[act_deg-1][j] = -coeffs[j] * rec;

  call_rg(mat, act_deg, act_deg, Re, Im, Vec, 0);
  for (i=0; i<act_deg; i++) 
    if (FP_EQ_EPS(Im[i],0.0,1e-6) && (Re[i] > -1e-4)) {
      (*param) = Re[i]/(1+Re[i]);
      Eval2DCurve(uvc, (*param), &tpt2d);
      if (EQUAL2D_EPS((*pt2d), tpt2d, 1.0e-5))
	break;
    }

  myfree(Re); myfree(Im);
  FREEMAT(mat, act_deg, act_deg); FREEMAT(Vec, act_deg, act_deg);

  if (i == act_deg) {
    *param = -1.0;
    return 0;
  } else 
    return 1;
}

static int comparCv(p1, p2)
PART_SPLINE_TYPE *p1, *p2;
{
  if (p1->spl_c_num < p2->spl_c_num) return -1;
  if (p1->spl_c_num > p2->spl_c_num) return 1;

  /* p1->spl_c_num == p2->spl_c_num */
  if (FP_EQ_EPS(p1->param, p2->param, 1e-4)) return 0;
  if (p1->param < p2->param) return -1;
  if (p1->param > p2->param) return 1;
}

/* This routine adds a merged curve in TRIM_PORTION_SPLINE_TYPE format */
/* if flag is 0, mc[indx] is added directly, otherwise, it is */
/* reversed. same_solid is an array which tells if the ith curve's */
/* neighbor belongs to the same solid or not. If it is part of an */
/* intersection curve, it is not. same_solid is filled inside the routine */
void AddMergedCurvetoRegion(mc, indx, tpst, flag, same_solid)
MERGED_CURVE *mc;
short indx, flag, *same_solid;
TRIM_PORTION_SPLINE_TYPE *tpst;
{
  INTERSECTION_CURVE *tptr;

  if (flag)
    reverse_int_curve(&mc[indx].int_curve, &tptr);
  
  tptr = mc[indx].int_curve;
  while (tptr != NULL) {
    if (tpst->NumOfCurves == MAX_SPLINES_PER_TRIM_REGION) {
      printf("AddMergedCurvetoRegion: too many curves in region\n");
      exit(-1);
    }
    copy_uv_spline(&tptr->uv_spline,
		   &tpst->UV_Splines[tpst->NumOfCurves]);
    copy_xyz_spline(&tptr->xyz_spline,
		    &tpst->XYZ_Splines[tpst->NumOfCurves]);
    same_solid[tpst->NumOfCurves] = 0;
    tpst->AdjInfo[(tpst->NumOfCurves)++] = tptr->neighbor;
    tptr = tptr->next;
  }
  if (flag)
    reverse_int_curve(&mc[indx].int_curve, &tptr);
}

int Spurious2DSpline(spl)
UV_SPLINE_TYPE *spl;
{
  if (EQUAL3D_ABS(spl->CntrlPts[0], spl->CntrlPts[spl->Deg], 2e-8))
    return 1;
  return 0;
}
    
/* This routine outputs the next region from the stack. for same_solid */
/* look at the comments of the AddMergedCurvetoRegion routine */
void ReadOutNextRegion(pst, mc, pst_or_st, c_index, sub_c_index,
		       stack_ptr, mc_num, in, out, same_solid,
		       spec_crvs)
PART_SPLINE_TYPE *pst;
MERGED_CURVE *mc;
short *pst_or_st, *c_index, *sub_c_index, *stack_ptr, *same_solid, mc_num;
TRIM_PORTION_SPLINE_TYPE *in, *out;
short *spec_crvs;
{
  short old_sp, new_sp, i;
  int spec_case;
  int deg, j;

  spec_case = 1;
  new_sp = old_sp = *stack_ptr;
  if (mc_num != -1) {
    while (new_sp >= 0) 
      if ((pst_or_st[new_sp] == 1) && 
	  (pst[c_index[new_sp]].merge_c_num == mc_num))
	break;
      else
	new_sp--;

    if (new_sp < 0) {
      printf("ReadOutNextRegion: could not find corresponding curve in stack\n");
      exit(-1);
    }
    *stack_ptr = new_sp; /* assign new stack top */
  } else 
    new_sp = -1;
  /* Add the intersection curve (mergedcurve) first */
  out->NumOfCurves = 0;
  out->UV_Splines = (UV_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					   * sizeof(UV_SPLINE_TYPE));
  out->XYZ_Splines = (XYZ_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					     * sizeof(XYZ_SPLINE_TYPE));
  out->AdjInfo = (int *) new(MAX_SPLINES_PER_TRIM_REGION
			     * sizeof(int));
  if (new_sp != -1)
    if (pst[c_index[new_sp]].ending)
      AddMergedCurvetoRegion(mc, mc_num, out, 0, same_solid);
    else
      AddMergedCurvetoRegion(mc, mc_num, out, 1, same_solid);

  /* Now determine if this region is a special case when all the */
  /* same_solids should made 0 */
  for (i=new_sp+1; i<=old_sp; i++)
    if (!pst_or_st[i] && (spec_crvs[c_index[i]] == -1)) {
      spec_case = 0;
      break;
    }

  for (i=new_sp+1; i<=old_sp; i++)
    if (!pst_or_st[i]) {
      if (out->NumOfCurves == MAX_SPLINES_PER_TRIM_REGION) {
	printf("ReadOutNextRegion: too many curves in region\n");
	exit(-1);
      }
      if (Spurious2DSpline(&uvst[c_index[i]][sub_c_index[i]]))
	continue;
      copy_uv_spline(&uvst[c_index[i]][sub_c_index[i]],
		     &out->UV_Splines[out->NumOfCurves]);
      copy_xyz_spline(&xyzst[c_index[i]][sub_c_index[i]],
		      &out->XYZ_Splines[out->NumOfCurves]);
      if (spec_case) {
	same_solid[out->NumOfCurves] = 0;
	out->AdjInfo[(out->NumOfCurves)++] = spec_crvs[c_index[i]];
	
	/* Now I am filling in the ngh_crv structure so that I can */
	/* change the same_solid and adjacent patch number after the */
	/* call to PartitionSplinegon */
	deg = out->XYZ_Splines[out->NumOfCurves - 1].Deg;
	ngh_crv[NumNghCrvs].spt = out->XYZ_Splines[out->NumOfCurves - 1].CntrlPts[0];
	ngh_crv[NumNghCrvs].ept = out->XYZ_Splines[out->NumOfCurves - 1].CntrlPts[deg];
	ngh_crv[NumNghCrvs].patchNo = in->AdjInfo[c_index[i]];
	ngh_crv[NumNghCrvs++].adj_pnum = spec_crvs[c_index[i]];
      } else {
	same_solid[out->NumOfCurves] = 1;
	out->AdjInfo[(out->NumOfCurves)++] = in->AdjInfo[c_index[i]];
      }
    } else {
      /* Future intersection curves are read out in the same order if */
      /* they are starting from that point. If they are ending, they */
      /* must be reversed */
      if (!pst[c_index[i]].ending)
	AddMergedCurvetoRegion(mc, pst[c_index[i]].merge_c_num, 
			       out, 0, same_solid);
      else
	AddMergedCurvetoRegion(mc, pst[c_index[i]].merge_c_num, 
			       out, 1, same_solid);
    }
}

/* This procedure takes a 2d spline and finds the parameters at which */
/* this curve reaches const_val for the constIn parameter (usually V). */
/* flag is used to indicate whether it should return the leftmost and */
/* rightmost values (flag == 0), or whether it should return the */
/* rightmost value left of left_limit and leftmost value right of */
/* right_limit. It also returns the corresponding U (if constIn == V) */
/* values in left_line_param and right_line_param */
void FindParamAtVal(uvs, const_val, constIn, left_line_param,
		    left_curve_param, right_line_param,
		    right_curve_param, flag, left_limit, right_limit)
UV_SPLINE_TYPE *uvs;
double const_val;
int constIn;
double *left_line_param, *left_curve_param, *right_line_param,
       *right_curve_param, *left_limit, *right_limit;
int flag;
{
  int i, j;
  double k;
  int curve_deg, index;
  double coeff_arr[MAX_CURVE_DEG];
  double **mat;
  double *wr, *wi, **vec;
  POINT2D pt;

  if (!flag) {
    *left_line_param = 1.0+EPSILON;
    *right_line_param = -EPSILON;
  } else {
    *left_line_param = -EPSILON;
    *right_line_param = 1.0+EPSILON;
  }

  curve_deg = uvs->Deg;
  if (curve_deg >= MAX_CURVE_DEG) {
    printf("FindParamAtVal: Curve degree too high\n");
    exit(-1);
  }

  if (constIn == U) 
    index = 0;
  else {
    if (constIn == V)
      index = 1;
    else {
      printf("FindParamAtVal: constIn has bad value\n");
      exit(-1);
    }
  }

  for (i=0; i<=curve_deg; i++) 
    coeff_arr[i] = combChoose(curve_deg, i) *
                   (uvs->CntrlPts[i].dCoord[index] - 
		    const_val*uvs->CntrlPts[i].dCoord[2]);
  j = curve_deg;
  while (j >= 1) {
    if (FP_EQ_EPS(coeff_arr[j], 0.0, 1e-6))
      j--;
    else
      break;
  }
  curve_deg = j;
  if (curve_deg == 0) {
    /*printf("FindParamAtVal: Extremely degenerate curve. Returning ...\n");*/
    return;
  }

  k = 1.0 / coeff_arr[curve_deg];
  for (i=0; i<curve_deg; i++)
    coeff_arr[i] *= k;

  ALLOCMAT(mat, curve_deg, curve_deg, double);
  ALLOCN(wr, double, curve_deg);
  ALLOCN(wi, double, curve_deg);
  ALLOCMAT(vec, curve_deg, curve_deg, double);

  for (i=0; i<curve_deg-1; i++)
    for (j=0; j<curve_deg; j++) 
      if (j == i+1)
	mat[i][j] = 1.0;
      else
	mat[i][j] = 0.0;

  for (j=0; j<curve_deg; j++)
    mat[curve_deg-1][j] = -coeff_arr[j];

  call_rg(mat, curve_deg, curve_deg, wr, wi, vec, 0);

  for (i=0; i<curve_deg; i++) {
    if (FP_EQ_EPS(wi[i], 0.0, 1e-3) && (wr[i] > -EPSILON)) {
      k = wr[i] / (1.0 + wr[i]);
      Eval2DCurve(uvs, k, &pt);
      if ((constIn == U) && !FP_EQ_EPS(pt.u, const_val, 1e-3)) {
	printf("FindParamAtVal: U: wrong result from eigenvalue\n");
	exit(-1);
      }
      if ((constIn == V) && !FP_EQ_EPS(pt.v, const_val, 1e-3)) {
	printf("FindParamAtVal: V: wrong result from eigenvalue\n");
	exit(-1);
      }
      if (constIn != V) {
	printf("FindParamAtVal: U case not handled because not necessary. Call must be wrong\n");
	exit(-1);
      }
      if (!flag) {
	if (pt.u < *left_line_param) {
	  *left_line_param = pt.u;
	  *left_curve_param = k;
	}
	if (pt.u > *right_line_param) {
	  *right_line_param = pt.u;
	  *right_curve_param = k;
	}
      } else {
	if ((pt.u > *left_line_param) && (pt.u < *left_limit)) {
	  *left_line_param = pt.u;
	  *left_curve_param = k;
	}
	if ((pt.u < *right_line_param) && (pt.u > *right_limit)) {
	  *right_line_param = pt.u;
	  *right_curve_param = k;
	}
      }
    }
  }

  FREEMAT(mat, curve_deg, curve_deg);
  FREEMAT(vec, curve_deg, curve_deg);
  myfree(wr); myfree(wi);
}

/* This routine is called when the intersection curve results in a */
/* loop and we have to divide the outermost toroidal structure into */
/* two regions. So we take the loop and a line (v = line_val) and find */
/* the extremal Xsection points between them. flag is used to indicate */
/* if this is called for the outer or inner loop. For the inner loop */
/* (flag = 0) we have to return the smallest and largest param as */
/* answers, but for the outer loop, we should return the rightmost */
/* param left of uval_l and leftmost param right of uval_r */
int IntersectTPSTwithLine(tpst, line_val, l_param, r_param, 
			  l_indx, cl_param, r_indx, cr_param, flag,
			  uval_l, uval_r)
TRIM_PORTION_SPLINE_TYPE *tpst;
double *line_val; /* Yvalue of the line used to divide the curve */
double *l_param, *r_param, *cl_param, *cr_param;
int *l_indx, *r_indx;
int flag; 
double *uval_l, *uval_r;
{
  int i;
  double tmp_cl_param, tmp_cr_param;
  double tmp_l_param, tmp_r_param;

  if (!flag) {
    *l_param = 1.0+EPSILON;
    *r_param = -EPSILON;
  } else {
    *l_param = -EPSILON;
    *r_param = 1.0+EPSILON;
  }

  for (i=0; i<tpst->NumOfCurves; i++) {
    FindParamAtVal(&tpst->UV_Splines[i], *line_val, V, &tmp_l_param,
		   &tmp_cl_param, &tmp_r_param, &tmp_cr_param, flag,
		   uval_l, uval_r);
    if (!flag) {
      if (tmp_l_param < *l_param) {
	*l_indx = i;
	*l_param = tmp_l_param;
	*cl_param = tmp_cl_param;
      }
      if (tmp_r_param > *r_param) {
	*r_indx = i;
	*r_param = tmp_r_param;
	*cr_param = tmp_cr_param;
      }
    } else {
      if (tmp_l_param > *l_param) {
	*l_indx = i;
	*l_param = tmp_l_param;
	*cl_param = tmp_cl_param;
      }
      if (tmp_r_param < *r_param) {
	*r_indx = i;
	*r_param = tmp_r_param;
	*cr_param = tmp_cr_param;
      }
    }
  }
}

/* This routine takes one TRIM_PORTION_SPLINE_TYPE tpst1 and another one */
/* tpst2 which lies completely inside the first one and produces two */
/* regions resulting in the difference of the two params. tpst1 is */
/* replaced by one of the output regions. The other is tpst3. The */
/* toroidal region tpst1-tpst2 is divided into two halves by a */
/* dividing line. tpst3 contains the bottom half and tpst1 is replaced */
/* with the top half. */
void SplitAndCreateNewRegion(tpst1, tpst2, tpst3, tp, pnum, first_sol,
			     same_solid1, same_solid3, sts_arr,
			     sts_arr_len, additional_patch, solid)
TRIM_PORTION_SPLINE_TYPE *tpst1, *tpst2, *tpst3;
TENSOR_PATCH *tp;
int pnum, first_sol;
short *same_solid1, *same_solid3;
SPLIT_TPST_STRUCT *sts_arr;
int *sts_arr_len;
short additional_patch;
TRIM_PATCH_INFO *solid;
{
  TRIM_PORTION_SPLINE_TYPE tmp_tpst, tmp_tpst1;
  UV_SPLINE_TYPE *uv_il_bot, *uv_il_top, *uv_ir_bot, *uv_ir_top;
  UV_SPLINE_TYPE *uv_ol_bot, *uv_ol_top, *uv_or_bot, *uv_or_top;
  XYZ_SPLINE_TYPE *xyz_il_bot, *xyz_il_top, *xyz_ir_bot, *xyz_ir_top;
  XYZ_SPLINE_TYPE *xyz_ol_bot, *xyz_ol_top, *xyz_or_bot, *xyz_or_top;
  UV_SPLINE_TYPE *uv_left_part, *uv_right_part;
  XYZ_SPLINE_TYPE *tmp_xyz_spl, *xyz_left_part, *xyz_right_part;
  POINT2D p2d;
  int i, j, k, iflag, oflag;
  double tmp_y;
  short tmp_same_solid[MAX_SPLINES_PER_TRIM_REGION];
  double param_arr[3]; /* I dont need more than 3 because this is used */
		       /* to split the curves il_indx and ir_indx of */
		       /* tpst2, so there are atmost 3 elements */
  short par_len;
  point_typ pt;
  int ol_indx, or_indx; /* ol and or refers to the left and right */
			/* curve indices of the outer loop */
  double ol_param, or_param; /* Refers to the parameter values for the */
			     /* curves bounding the left and right */
			     /* side of the outer loop */
  int il_indx, ir_indx; /* il and ir refers to the left and right */
			/* curve indices of the inner loop */
  double il_param, ir_param; /* Refers to the parameter values for the */
			     /* curves bounding the left and right */
			     /* side of the inner loop */
  double col_param, cor_param; /* Refers to the extreme param values for the */
                               /* curve used to cut the torus */
			       /* structure at which it intersects the */
			       /* the outer loop */
  double cil_param, cir_param; /* Refers to the extreme param values for the */
                               /* curve used to cut the torus */
			       /* structure at which it intersects the */
			       /* the inner loop */
  int sm_indx, lg_indx;
  double sm_par, lg_par;
  short partn_num, dir;
  int cnum;

  /* Allocating the tmp splines */
  uv_il_bot = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_il_top = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_ol_bot = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_ol_top = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_ir_bot = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_ir_top = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_or_bot = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_or_top = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_left_part = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  uv_right_part = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));

  xyz_il_bot = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_il_top = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_ol_bot = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_ol_top = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_ir_bot = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_ir_top = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_or_bot = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  xyz_or_top = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  tmp_xyz_spl = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));

  copy_trim_portion_spline_type(&tmp_tpst, tpst2, 0);
  copy_trim_portion_spline_type(&tmp_tpst1, tpst1, 0);

  /* I shall copy the array same_solid1 into tmp_same_solid so that I */
  /* can fill same_solid1 and same_solid3 with the appropriate info */
  for (i=0; i<MAX_SPLINES_PER_TRIM_REGION; i++)
    tmp_same_solid[i] = same_solid1[i];

  free_trim_portion_spline_type(tpst1);
  ReverseTPST(&tmp_tpst);

  if (tmp_tpst.NumOfCurves <= 1) {
    printf("SplitAndCreateNewRegion: Too few curves to do splitting\n");
    exit(-1);
  }

  pt = GetPtInsideSplinegon(tpst2);
  IntersectTPSTwithLine(&tmp_tpst, &pt.y, &il_param, &ir_param, 
			&il_indx, &cil_param, &ir_indx, &cir_param, 
			0, NULL, NULL);

  IntersectTPSTwithLine(&tmp_tpst1, &pt.y, &ol_param, &or_param, 
			&ol_indx, &col_param, &or_indx, &cor_param, 
			1, &il_param, &ir_param);

  while (FP_EQ_EPS(cil_param, 0.0, 1e-2) || FP_EQ_EPS(cil_param, 1.0, 1e-2) ||
	 FP_EQ_EPS(cir_param, 0.0, 1e-2) || FP_EQ_EPS(cir_param, 1.0, 1e-2) || 
	 FP_EQ_EPS(col_param, 0.0, 1e-2) || FP_EQ_EPS(col_param, 1.0, 1e-2) ||
	 FP_EQ_EPS(cor_param, 0.0, 1e-2) || FP_EQ_EPS(cor_param, 1.0, 1e-2)) {
    printf("SplitAndCreateNewRegion: Unexpected param value: cil_p, cir_p, col_p and cor_p = %lf, %lf, %lf and %lf\n", cil_param, cir_param, col_param, cor_param);
    pt.y -= 0.05;
    while (!IsPtInsideSplinegon(tpst2, &pt))
      pt.y += 0.01;
    IntersectTPSTwithLine(&tmp_tpst, &pt.y, &il_param, &ir_param, 
			  &il_indx, &cil_param, &ir_indx, &cir_param, 
			  0, NULL, NULL);
    IntersectTPSTwithLine(&tmp_tpst1, &pt.y, &ol_param, &or_param, 
			  &ol_indx, &col_param, &or_indx, &cor_param, 
			  1, &il_param, &ir_param);
  }

  /* Now to split curves il_indx and ir_indx of tpst2 */
  if (tpst2->NumOfCurves >= MAX_SPLINES_PER_TRIM_REGION-2) {
    printf("Too many curves in one region. Quitting ...\n");
    exit(-1);
  }
  if (il_indx != ir_indx) {
    if (il_indx < ir_indx) {
      sm_indx = tpst2->NumOfCurves-ir_indx-1;
      sm_par = 1.0 - cir_param;
      lg_indx = tpst2->NumOfCurves-il_indx-1;
      lg_par = 1.0 - cil_param;
    } else {
      sm_indx = tpst2->NumOfCurves-il_indx-1;
      sm_par = 1.0 - cil_param;
      lg_indx = tpst2->NumOfCurves-ir_indx-1;
      lg_par = 1.0 - cir_param;
    }
  } else {
    sm_indx = lg_indx = tpst2->NumOfCurves-ir_indx-1;
    if (cil_param < cir_param) {
      sm_par = 1.0 - cir_param;
      lg_par = 1.0 - cil_param;
    } else {
      sm_par = 1.0 - cil_param;
      lg_par = 1.0 - cir_param;
    }
  }
  if (il_indx != ir_indx) {
    for (i=tpst2->NumOfCurves-1; i>lg_indx; i--) {
      copy_uv_spline(&tpst2->UV_Splines[i], &tpst2->UV_Splines[i+2]);
      copy_xyz_spline(&tpst2->XYZ_Splines[i], &tpst2->XYZ_Splines[i+2]);
      tpst2->AdjInfo[i+2] = tpst2->AdjInfo[i];
    }
    for (i=lg_indx; i>sm_indx; i--) {
      copy_uv_spline(&tpst2->UV_Splines[i], &tpst2->UV_Splines[i+1]);
      copy_xyz_spline(&tpst2->XYZ_Splines[i], &tpst2->XYZ_Splines[i+1]);
      tpst2->AdjInfo[i+1] = tpst2->AdjInfo[i];
    }
    tpst2->NumOfCurves += 2;

    par_len = 1;
    param_arr[0] = sm_par;
    param_arr[1] = 1.0;
    SubdivideTrimCurve(&tpst2->UV_Splines[sm_indx],
		       &tpst2->XYZ_Splines[sm_indx], par_len,
		       param_arr, &tpst2->UV_Splines[sm_indx],
		       &tpst2->XYZ_Splines[sm_indx]);
    tpst2->AdjInfo[sm_indx+1] = tpst2->AdjInfo[sm_indx];

    par_len = 1;
    param_arr[0] = lg_par;
    param_arr[1] = 1.0;
    SubdivideTrimCurve(&tpst2->UV_Splines[lg_indx+1],
		       &tpst2->XYZ_Splines[lg_indx+1], par_len,
		       param_arr, &tpst2->UV_Splines[lg_indx+1],
		       &tpst2->XYZ_Splines[lg_indx+1]);
    tpst2->AdjInfo[lg_indx+2] = tpst2->AdjInfo[lg_indx+1];
  } else {
    for (i=tpst2->NumOfCurves-1; i>lg_indx; i--) {
      copy_uv_spline(&tpst2->UV_Splines[i], &tpst2->UV_Splines[i+2]);
      copy_xyz_spline(&tpst2->XYZ_Splines[i], &tpst2->XYZ_Splines[i+2]);
      tpst2->AdjInfo[i+2] = tpst2->AdjInfo[i];
    }
    tpst2->NumOfCurves += 2;

    par_len = 2;
    param_arr[0] = sm_par;
    param_arr[1] = lg_par;
    param_arr[2] = 1.0;
    SubdivideTrimCurve(&tpst2->UV_Splines[lg_indx],
		       &tpst2->XYZ_Splines[lg_indx], par_len,
		       param_arr, &tpst2->UV_Splines[lg_indx],
		       &tpst2->XYZ_Splines[lg_indx]);
    tpst2->AdjInfo[lg_indx+2] = tpst2->AdjInfo[lg_indx+1] =
      tpst2->AdjInfo[lg_indx]; 
  }
  /* finished splitting il_indx and ir_indx curves of tpst2 */

/* ######
  IntersectTPSTwithLine(&tmp_tpst1, &pt.y, &ol_param, &or_param, 
			&ol_indx, &col_param, &or_indx, &cor_param, 
			1, &il_param, &ir_param);

  if (FP_EQ_EPS(col_param, 0.0, 1e-2) || FP_EQ_EPS(col_param, 1.0, 1e-2) ||
      FP_EQ_EPS(cor_param, 0.0, 1e-2) || FP_EQ_EPS(cor_param, 1.0, 1e-2)) {
    printf("SplitAndCreateNewRegion: Unexpected param value: col_p and cor_p = %lf and %lf\n", col_param, cor_param);
    pt.y -= 0.05;
    while (!IsPtInsideSplinegon(tpst2, &pt))
      pt.y += 0.01;
    IntersectTPSTwithLine(&tmp_tpst1, &pt.y, &ol_param, &or_param, 
			  &ol_indx, &col_param, &or_indx, &cor_param, 
			  1, &il_param, &ir_param);
  }
###### */

  /* Now I shall split the curve at the corresponding param value and */
  /* store them in temporary variables */
  if (il_indx != ir_indx) {
    TwoDBezSub(&tmp_tpst.UV_Splines[il_indx], cil_param, uv_il_bot, 0, 1);
    ThreeDBezSub(&tmp_tpst.XYZ_Splines[il_indx], cil_param,
		 xyz_il_bot, 0, 1);
    TwoDBezSub(&tmp_tpst.UV_Splines[il_indx], cil_param, uv_il_top, 1, 1);
    ThreeDBezSub(&tmp_tpst.XYZ_Splines[il_indx], cil_param,
		 xyz_il_top, 1, 1);
    sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
    sts_arr[(*sts_arr_len)].same_solid = 0;
    sts_arr[(*sts_arr_len)].do_twice = 1;
    sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst.AdjInfo[il_indx];
    ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
    sts_arr[(*sts_arr_len)].param[0] = 1.0 - cil_param;
    sts_arr[(*sts_arr_len)].param[1] = 1.0;
    sts_arr[(*sts_arr_len)].param_length = 1;
    sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].st_pt) = 
      tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
    sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].end_pt) = 
      tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
    sts_arr[(*sts_arr_len)].other_patch_num = pnum;
    (*sts_arr_len)++;

    if (additional_patch != -1) {
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
      sts_arr[(*sts_arr_len)].same_solid = 0;
      sts_arr[(*sts_arr_len)-1].do_twice = 0;
      sts_arr[(*sts_arr_len)].do_twice = 0;
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - cil_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 1;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      if (additional_patch != -2)
	sts_arr[(*sts_arr_len)].patch_num = additional_patch;
      else {
	partn_num = GetAdjacentTrimPortion(&solid[tmp_tpst.AdjInfo[il_indx]],
					   sts_arr[(*sts_arr_len)].st_pt,
					   sts_arr[(*sts_arr_len)].end_pt, &cnum,
					   -1, &dir);
	sts_arr[(*sts_arr_len)].patch_num = 
	  solid[tmp_tpst.AdjInfo[il_indx]].TrimPortion[partn_num].SplinePortion.
	                                                            AdjInfo[cnum]; 
      }
      (*sts_arr_len)++;
    }
    
    TwoDBezSub(&tmp_tpst.UV_Splines[ir_indx], cir_param, uv_ir_bot, 1, 1);
    ThreeDBezSub(&tmp_tpst.XYZ_Splines[ir_indx], cir_param,
		 xyz_ir_bot, 1, 1);
    TwoDBezSub(&tmp_tpst.UV_Splines[ir_indx], cir_param, uv_ir_top, 0, 1);
    ThreeDBezSub(&tmp_tpst.XYZ_Splines[ir_indx], cir_param,
		 xyz_ir_top, 0, 1);
    sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
    sts_arr[(*sts_arr_len)].same_solid = 0;
    sts_arr[(*sts_arr_len)].do_twice = 1;
    sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst.AdjInfo[ir_indx];
    ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
    sts_arr[(*sts_arr_len)].param[0] = 1.0 - cir_param;
    sts_arr[(*sts_arr_len)].param[1] = 1.0;
    sts_arr[(*sts_arr_len)].param_length = 1;
    sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].st_pt) = 
      tmp_tpst.XYZ_Splines[ir_indx].CntrlPts[0];
    sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].end_pt) = 
      tmp_tpst.XYZ_Splines[ir_indx].CntrlPts[tmp_tpst.XYZ_Splines[ir_indx].Deg];
    sts_arr[(*sts_arr_len)].other_patch_num = pnum;
    (*sts_arr_len)++;
    
    if (additional_patch != -1) {
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
      sts_arr[(*sts_arr_len)].same_solid = 0;
      sts_arr[(*sts_arr_len)-1].do_twice = 0;
      sts_arr[(*sts_arr_len)].do_twice = 0;
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - cir_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 1;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst.XYZ_Splines[ir_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst.XYZ_Splines[ir_indx].CntrlPts[tmp_tpst.XYZ_Splines[ir_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      if (additional_patch != -2)
	sts_arr[(*sts_arr_len)].patch_num = additional_patch;
      else {
	partn_num = GetAdjacentTrimPortion(&solid[tmp_tpst.AdjInfo[ir_indx]],
					   sts_arr[(*sts_arr_len)].st_pt,
					   sts_arr[(*sts_arr_len)].end_pt, &cnum,
					   -1, &dir);
	sts_arr[(*sts_arr_len)].patch_num = 
	  solid[tmp_tpst.AdjInfo[ir_indx]].TrimPortion[partn_num].SplinePortion.
	                                                            AdjInfo[cnum]; 
      }
      (*sts_arr_len)++;
    }
    
  } else {
    iflag = -1;
    tmp_y = tmp_tpst.UV_Splines[il_indx].CntrlPts[0].dCoord[1] / 
            tmp_tpst.UV_Splines[il_indx].CntrlPts[0].dCoord[2];
    if (tmp_y < pt.y)
      iflag = 1;
    else
      iflag = 2;
    switch (iflag) {
    case 1: 
      if (cil_param > cir_param) {
	printf("SplitAndCreateNewRegion: cil_param > cir_param\n");
	sleep(3);
	exit(-1);
      }
      TwoDBezSub(&tmp_tpst.UV_Splines[il_indx], cil_param, uv_il_bot,
		 0, 1);
      ThreeDBezSub(&tmp_tpst.XYZ_Splines[il_indx], cil_param,
		   xyz_il_bot, 0, 1);
      
      TwoDBezSub(&tmp_tpst.UV_Splines[ir_indx], cir_param, uv_ir_bot,
		 1, 1);
      ThreeDBezSub(&tmp_tpst.XYZ_Splines[ir_indx], cir_param,
		   xyz_ir_bot, 1, 1);

      TwoDBezierSubdivision(&tmp_tpst.UV_Splines[il_indx], cil_param,
			    cir_param, &uv_ir_top);
      ThreeDBezierSubdivision(&tmp_tpst.XYZ_Splines[il_indx], cil_param,
			      cir_param, &xyz_ir_top);
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
      sts_arr[(*sts_arr_len)].same_solid = 0;
      sts_arr[(*sts_arr_len)].do_twice = 1;
      sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst.AdjInfo[il_indx];
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - cir_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0 - cil_param;
      sts_arr[(*sts_arr_len)].param[2] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 2;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      (*sts_arr_len)++;

      if (additional_patch != -1) {
	sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
	sts_arr[(*sts_arr_len)].same_solid = 0;
	sts_arr[(*sts_arr_len)-1].do_twice = 0;
	sts_arr[(*sts_arr_len)].do_twice = 0;
	ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
	sts_arr[(*sts_arr_len)].param[0] = 1.0 - cir_param;
	sts_arr[(*sts_arr_len)].param[1] = 1.0 - cil_param;
	sts_arr[(*sts_arr_len)].param[2] = 1.0;
	sts_arr[(*sts_arr_len)].param_length = 2;
	sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
	(*sts_arr[(*sts_arr_len)].st_pt) = 
	  tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
	sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
	(*sts_arr[(*sts_arr_len)].end_pt) = 
	  tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
	sts_arr[(*sts_arr_len)].other_patch_num = pnum;
	if (additional_patch != -2)
	  sts_arr[(*sts_arr_len)].patch_num = additional_patch;
	else {
	  partn_num = GetAdjacentTrimPortion(&solid[tmp_tpst.AdjInfo[il_indx]],
					     sts_arr[(*sts_arr_len)].st_pt,
					     sts_arr[(*sts_arr_len)].end_pt, &cnum,
					     -1, &dir);
	  sts_arr[(*sts_arr_len)].patch_num = 
	    solid[tmp_tpst.AdjInfo[il_indx]].TrimPortion[partn_num].SplinePortion.
	      AdjInfo[cnum]; 
	}
	(*sts_arr_len)++;
      }
    
      break;
    case 2:
      if (cir_param > cil_param) {
	printf("SplitAndCreateNewRegion: cir_param > cil_param\n");
	sleep(3);
	exit(-1);
      }
      TwoDBezSub(&tmp_tpst.UV_Splines[il_indx], cil_param, uv_il_top,
		 1, 1);
      ThreeDBezSub(&tmp_tpst.XYZ_Splines[il_indx], cil_param,
		   xyz_il_top, 1, 1);
      
      TwoDBezSub(&tmp_tpst.UV_Splines[ir_indx], cir_param, uv_ir_top,
		 0, 1);
      ThreeDBezSub(&tmp_tpst.XYZ_Splines[ir_indx], cir_param,
		   xyz_ir_top, 0, 1);

      TwoDBezierSubdivision(&tmp_tpst.UV_Splines[il_indx], cir_param,
			    cil_param, &uv_ir_bot);
      ThreeDBezierSubdivision(&tmp_tpst.XYZ_Splines[il_indx], cir_param,
			      cil_param, &xyz_ir_bot);
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
      sts_arr[(*sts_arr_len)].same_solid = 0;
      sts_arr[(*sts_arr_len)].do_twice = 1;
      sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst.AdjInfo[il_indx];
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - cil_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0 - cir_param;
      sts_arr[(*sts_arr_len)].param[2] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 2;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      (*sts_arr_len)++;

      if (additional_patch != -1) {
	sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 2 : 1;
	sts_arr[(*sts_arr_len)].same_solid = 0;
	sts_arr[(*sts_arr_len)-1].do_twice = 0;
	sts_arr[(*sts_arr_len)].do_twice = 0;
	ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
	sts_arr[(*sts_arr_len)].param[0] = 1.0 - cil_param;
	sts_arr[(*sts_arr_len)].param[1] = 1.0 - cir_param;
	sts_arr[(*sts_arr_len)].param[2] = 1.0;
	sts_arr[(*sts_arr_len)].param_length = 2;
	sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
	(*sts_arr[(*sts_arr_len)].st_pt) = 
	  tmp_tpst.XYZ_Splines[il_indx].CntrlPts[0];
	sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
	(*sts_arr[(*sts_arr_len)].end_pt) = 
	  tmp_tpst.XYZ_Splines[il_indx].CntrlPts[tmp_tpst.XYZ_Splines[il_indx].Deg];
	sts_arr[(*sts_arr_len)].other_patch_num = pnum;
	if (additional_patch != -2)
	  sts_arr[(*sts_arr_len)].patch_num = additional_patch;
	else {
	  partn_num = GetAdjacentTrimPortion(&solid[tmp_tpst.AdjInfo[il_indx]],
					     sts_arr[(*sts_arr_len)].st_pt,
					     sts_arr[(*sts_arr_len)].end_pt, &cnum,
					     -1, &dir);
	  sts_arr[(*sts_arr_len)].patch_num = 
	    solid[tmp_tpst.AdjInfo[il_indx]].TrimPortion[partn_num].SplinePortion.
	                                                            AdjInfo[cnum]; 
	}
	(*sts_arr_len)++;
      }

      break;
    }
  }

  if (ol_indx != or_indx) {
    TwoDBezSub(&tmp_tpst1.UV_Splines[ol_indx], col_param, uv_ol_bot, 1, 1);
    ThreeDBezSub(&tmp_tpst1.XYZ_Splines[ol_indx], col_param,
		 xyz_ol_bot, 1, 1);
    TwoDBezSub(&tmp_tpst1.UV_Splines[ol_indx], col_param, uv_ol_top, 0, 1);
    ThreeDBezSub(&tmp_tpst1.XYZ_Splines[ol_indx], col_param,
		 xyz_ol_top, 0, 1);
    sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 
                                          (2-tmp_same_solid[ol_indx]) :
					  (1+tmp_same_solid[ol_indx]);
    sts_arr[(*sts_arr_len)].same_solid = tmp_same_solid[ol_indx];
    sts_arr[(*sts_arr_len)].do_twice = 1;
    sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst1.AdjInfo[ol_indx];
    ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
    sts_arr[(*sts_arr_len)].param[0] = 1.0 - col_param;
    sts_arr[(*sts_arr_len)].param[1] = 1.0;
    sts_arr[(*sts_arr_len)].param_length = 1;
    sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].st_pt) = 
      tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[0];
    sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].end_pt) = 
      tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[tmp_tpst1.XYZ_Splines[ol_indx].Deg];
    sts_arr[(*sts_arr_len)].other_patch_num = pnum;
    (*sts_arr_len)++;
    
    
    TwoDBezSub(&tmp_tpst1.UV_Splines[or_indx], cor_param, uv_or_bot, 0, 1);
    ThreeDBezSub(&tmp_tpst1.XYZ_Splines[or_indx], cor_param,
		 xyz_or_bot, 0, 1);
    TwoDBezSub(&tmp_tpst1.UV_Splines[or_indx], cor_param, uv_or_top, 1, 1);
    ThreeDBezSub(&tmp_tpst1.XYZ_Splines[or_indx], cor_param,
		 xyz_or_top, 1, 1);
    sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 
                                          (2-tmp_same_solid[or_indx]) :
					  (1+tmp_same_solid[or_indx]);
    sts_arr[(*sts_arr_len)].same_solid = tmp_same_solid[or_indx];
    sts_arr[(*sts_arr_len)].do_twice = 1;
    sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst1.AdjInfo[or_indx];
    ALLOCN(sts_arr[(*sts_arr_len)].param, double, 2);
    sts_arr[(*sts_arr_len)].param[0] = 1.0 - cor_param;
    sts_arr[(*sts_arr_len)].param[1] = 1.0;
    sts_arr[(*sts_arr_len)].param_length = 1;
    sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].st_pt) = 
      tmp_tpst1.XYZ_Splines[or_indx].CntrlPts[0];
    sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
    (*sts_arr[(*sts_arr_len)].end_pt) = 
      tmp_tpst1.XYZ_Splines[or_indx].CntrlPts[tmp_tpst1.XYZ_Splines[or_indx].Deg];
    sts_arr[(*sts_arr_len)].other_patch_num = pnum;
    (*sts_arr_len)++;
    
  } else {
    oflag = -1;
    tmp_y = tmp_tpst1.UV_Splines[ol_indx].CntrlPts[0].dCoord[1] / 
            tmp_tpst1.UV_Splines[ol_indx].CntrlPts[0].dCoord[2];
    if (tmp_y < pt.y)
      oflag = 1;
    else
      oflag = 2;
    switch (oflag) {
    case 1: 
      if (cor_param > col_param) {
	printf("SplitAndCreateNewRegion: cor_param > col_param\n");
	sleep(3);
	exit(-1);
      }
      TwoDBezSub(&tmp_tpst1.UV_Splines[ol_indx], col_param, uv_ol_bot,
		 1, 1);
      ThreeDBezSub(&tmp_tpst1.XYZ_Splines[ol_indx], col_param,
		   xyz_ol_bot, 1, 1);
      
      TwoDBezSub(&tmp_tpst1.UV_Splines[or_indx], cor_param, uv_or_bot,
		 0, 1);
      ThreeDBezSub(&tmp_tpst1.XYZ_Splines[or_indx], cor_param,
		   xyz_or_bot, 0, 1);

      TwoDBezierSubdivision(&tmp_tpst1.UV_Splines[ol_indx], cor_param,
			    col_param, &uv_or_top);
      ThreeDBezierSubdivision(&tmp_tpst1.XYZ_Splines[ol_indx], cor_param,
			      col_param, &xyz_or_top);
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 
	                                    (2-tmp_same_solid[ol_indx]) :
	                                    (1+tmp_same_solid[ol_indx]);
      sts_arr[(*sts_arr_len)].same_solid = tmp_same_solid[ol_indx];
      sts_arr[(*sts_arr_len)].do_twice = 1;
      sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst1.AdjInfo[ol_indx];
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - col_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0 - cor_param;
      sts_arr[(*sts_arr_len)].param[2] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 2;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[tmp_tpst1.XYZ_Splines[ol_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      (*sts_arr_len)++;
    
      break;
    case 2:
      if (col_param > cor_param) {
	printf("SplitAndCreateNewRegion: col_param > cor_param\n");
	sleep(3);
	exit(-1);
      }
      TwoDBezSub(&tmp_tpst1.UV_Splines[ol_indx], col_param, uv_ol_top,
		 0, 1);
      ThreeDBezSub(&tmp_tpst1.XYZ_Splines[ol_indx], col_param,
		   xyz_ol_top, 0, 1);
      
      TwoDBezSub(&tmp_tpst1.UV_Splines[or_indx], cor_param, uv_or_top,
		 1, 1);
      ThreeDBezSub(&tmp_tpst1.XYZ_Splines[or_indx], cor_param,
		   xyz_or_top, 1, 1);

      TwoDBezierSubdivision(&tmp_tpst1.UV_Splines[ol_indx], col_param,
			    cor_param, &uv_or_bot);
      ThreeDBezierSubdivision(&tmp_tpst1.XYZ_Splines[ol_indx], col_param,
			      cor_param, &xyz_or_bot);
      sts_arr[(*sts_arr_len)].which_solid = (first_sol) ? 
	                                    (2-tmp_same_solid[ol_indx]) :
	                                    (1+tmp_same_solid[ol_indx]);
      sts_arr[(*sts_arr_len)].same_solid = tmp_same_solid[ol_indx];
      sts_arr[(*sts_arr_len)].do_twice = 1;
      sts_arr[(*sts_arr_len)].patch_num = (short) tmp_tpst1.AdjInfo[ol_indx];
      ALLOCN(sts_arr[(*sts_arr_len)].param, double, 3);
      sts_arr[(*sts_arr_len)].param[0] = 1.0 - cor_param;
      sts_arr[(*sts_arr_len)].param[1] = 1.0 - col_param;
      sts_arr[(*sts_arr_len)].param[2] = 1.0;
      sts_arr[(*sts_arr_len)].param_length = 2;
      sts_arr[(*sts_arr_len)].st_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].st_pt) = 
	tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[0];
      sts_arr[(*sts_arr_len)].end_pt = (POINT4D *) new(sizeof(POINT4D));
      (*sts_arr[(*sts_arr_len)].end_pt) = 
	tmp_tpst1.XYZ_Splines[ol_indx].CntrlPts[tmp_tpst1.XYZ_Splines[ol_indx].Deg];
      sts_arr[(*sts_arr_len)].other_patch_num = pnum;
      (*sts_arr_len)++;
    
      break;
    }
  }

  uv_left_part->Deg = uv_right_part->Deg = 1;
  ALLOCN(uv_left_part->CntrlPts, POINT3D, uv_left_part->Deg+1);
  ALLOCN(uv_right_part->CntrlPts, POINT3D, uv_right_part->Deg+1);
  for (i=0; i<=1; i++) {
    uv_left_part->CntrlPts[i].dCoord[1] =
      uv_right_part->CntrlPts[i].dCoord[1] = pt.y;
    uv_left_part->CntrlPts[i].dCoord[2] =
      uv_right_part->CntrlPts[i].dCoord[2] = 1.0;
  }
  uv_left_part->CntrlPts[0].dCoord[0] = ol_param;
  uv_left_part->CntrlPts[1].dCoord[0] = il_param;
  uv_right_part->CntrlPts[0].dCoord[0] = ir_param;
  uv_right_part->CntrlPts[1].dCoord[0] = or_param;
  GetSplineFromPatch(tp, 1, pt.y, tmp_xyz_spl);
  ThreeDBezierSubdivision(tmp_xyz_spl, ol_param, il_param,
			  &xyz_left_part);
  ThreeDBezierSubdivision(tmp_xyz_spl, ir_param, or_param,
			  &xyz_right_part); 

  tpst1->NumOfCurves = tpst3->NumOfCurves = 0;
  tpst1->UV_Splines = (UV_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					   * sizeof(UV_SPLINE_TYPE));
  tpst3->UV_Splines = (UV_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					   * sizeof(UV_SPLINE_TYPE));
  tpst1->XYZ_Splines = (XYZ_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					    * sizeof(XYZ_SPLINE_TYPE));
  tpst3->XYZ_Splines = (XYZ_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION
					    * sizeof(XYZ_SPLINE_TYPE));
  tpst1->AdjInfo = (int *) new(MAX_SPLINES_PER_TRIM_REGION
				* sizeof(int));
  tpst3->AdjInfo = (int *) new(MAX_SPLINES_PER_TRIM_REGION
				* sizeof(int));
  /* Now generating tpst1 */
  copy_uv_spline(uv_left_part,
		 &tpst1->UV_Splines[tpst1->NumOfCurves]);
  copy_xyz_spline(xyz_left_part,
		  &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
  same_solid1[tpst1->NumOfCurves] = 1;
  tpst1->AdjInfo[tpst1->NumOfCurves++] = pnum;
  

  if ((il_indx != ir_indx) || (iflag == 2)) {
    copy_uv_spline(uv_il_top,
		   &tpst1->UV_Splines[tpst1->NumOfCurves]);
    copy_xyz_spline(xyz_il_top,
		    &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
    same_solid1[tpst1->NumOfCurves] = 0;
    tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst.AdjInfo[il_indx];
    
    i = (il_indx+1) % tmp_tpst.NumOfCurves;
    while (i != ir_indx) {
      copy_uv_spline(&tmp_tpst.UV_Splines[i],
		     &tpst1->UV_Splines[tpst1->NumOfCurves]);
      copy_xyz_spline(&tmp_tpst.XYZ_Splines[i],
		      &tpst1->XYZ_Splines[tpst1->NumOfCurves]);
      same_solid1[tpst1->NumOfCurves] = 0;
      tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst.AdjInfo[i];
      i = (i+1) % tmp_tpst.NumOfCurves;
    }
  }

  copy_uv_spline(uv_ir_top,
		 &tpst1->UV_Splines[tpst1->NumOfCurves]);
  copy_xyz_spline(xyz_ir_top,
		  &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
  same_solid1[tpst1->NumOfCurves] = 0;
  tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst.AdjInfo[ir_indx];

  copy_uv_spline(uv_right_part,
		 &tpst1->UV_Splines[tpst1->NumOfCurves]);
  copy_xyz_spline(xyz_right_part,
		  &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
  same_solid1[tpst1->NumOfCurves] = 1;
  tpst1->AdjInfo[tpst1->NumOfCurves++] = pnum;
  
  copy_uv_spline(uv_or_top,
		 &tpst1->UV_Splines[tpst1->NumOfCurves]);
  copy_xyz_spline(xyz_or_top,
		  &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
  same_solid1[tpst1->NumOfCurves] = tmp_same_solid[or_indx];
  tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst1.AdjInfo[or_indx];

  if ((ol_indx != or_indx) || (oflag == 2)) {
    i = (or_indx+1) % tmp_tpst1.NumOfCurves;
    while (i != ol_indx) {
      copy_uv_spline(&tmp_tpst1.UV_Splines[i],
		     &tpst1->UV_Splines[tpst1->NumOfCurves]);
      copy_xyz_spline(&tmp_tpst1.XYZ_Splines[i],
		      &tpst1->XYZ_Splines[tpst1->NumOfCurves]);
      same_solid1[tpst1->NumOfCurves] = tmp_same_solid[i];
      tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst1.AdjInfo[i];
      i = (i+1) % tmp_tpst1.NumOfCurves;
    }
    
    copy_uv_spline(uv_ol_top,
		   &tpst1->UV_Splines[tpst1->NumOfCurves]);
    copy_xyz_spline(xyz_ol_top,
		    &tpst1->XYZ_Splines[tpst1->NumOfCurves]); 
    same_solid1[tpst1->NumOfCurves] = tmp_same_solid[ol_indx];
    tpst1->AdjInfo[tpst1->NumOfCurves++] = tmp_tpst1.AdjInfo[ol_indx];
  }

  /* Now generating tpst3 */
  copy_uv_spline(uv_right_part,
		 &tpst3->UV_Splines[tpst3->NumOfCurves]);
  ReverseCntrlPtsOf2DSpline(&tpst3->UV_Splines[tpst3->NumOfCurves]);
  copy_xyz_spline(xyz_right_part,
		  &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
  ReverseCntrlPtsOf3DSpline(&tpst3->XYZ_Splines[tpst3->NumOfCurves]);
  same_solid3[tpst3->NumOfCurves] = 1;
  tpst3->AdjInfo[tpst3->NumOfCurves++] = pnum;
  
  copy_uv_spline(uv_ir_bot,
		 &tpst3->UV_Splines[tpst3->NumOfCurves]);
  copy_xyz_spline(xyz_ir_bot,
		  &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
  same_solid3[tpst3->NumOfCurves] = 0;
  tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst.AdjInfo[ir_indx];

  if ((il_indx != ir_indx) || (iflag == 1)) {
    i = (ir_indx+1) % tmp_tpst.NumOfCurves;
    while (i != il_indx) {
      copy_uv_spline(&tmp_tpst.UV_Splines[i],
		     &tpst3->UV_Splines[tpst3->NumOfCurves]);
      copy_xyz_spline(&tmp_tpst.XYZ_Splines[i],
		      &tpst3->XYZ_Splines[tpst3->NumOfCurves]);
      same_solid3[tpst3->NumOfCurves] = 0;
      tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst.AdjInfo[i];
      i = (i+1) % tmp_tpst.NumOfCurves;
    }

    copy_uv_spline(uv_il_bot,
		   &tpst3->UV_Splines[tpst3->NumOfCurves]);
    copy_xyz_spline(xyz_il_bot,
		    &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
    same_solid3[tpst3->NumOfCurves] = 0;
    tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst.AdjInfo[il_indx];
  }

  copy_uv_spline(uv_left_part,
		 &tpst3->UV_Splines[tpst3->NumOfCurves]);
  ReverseCntrlPtsOf2DSpline(&tpst3->UV_Splines[tpst3->NumOfCurves]);
  copy_xyz_spline(xyz_left_part,
		  &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
  ReverseCntrlPtsOf3DSpline(&tpst3->XYZ_Splines[tpst3->NumOfCurves]);
  same_solid3[tpst3->NumOfCurves] = 1;
  tpst3->AdjInfo[tpst3->NumOfCurves++] = pnum;

  if ((ol_indx != or_indx) || (oflag == 1)) {
    copy_uv_spline(uv_ol_bot,
		   &tpst3->UV_Splines[tpst3->NumOfCurves]);
    copy_xyz_spline(xyz_ol_bot,
		    &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
    same_solid3[tpst3->NumOfCurves] = tmp_same_solid[ol_indx];
    tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst1.AdjInfo[ol_indx];
    
    i = (ol_indx+1) % tmp_tpst1.NumOfCurves;
    while (i != or_indx) {
      copy_uv_spline(&tmp_tpst1.UV_Splines[i],
		     &tpst3->UV_Splines[tpst3->NumOfCurves]);
      copy_xyz_spline(&tmp_tpst1.XYZ_Splines[i],
		      &tpst3->XYZ_Splines[tpst3->NumOfCurves]);
      same_solid3[tpst3->NumOfCurves] = tmp_same_solid[i];
      tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst1.AdjInfo[i];
      i = (i+1) % tmp_tpst1.NumOfCurves;
    }
  }

  copy_uv_spline(uv_or_bot,
		 &tpst3->UV_Splines[tpst3->NumOfCurves]);
  copy_xyz_spline(xyz_or_bot,
		  &tpst3->XYZ_Splines[tpst3->NumOfCurves]); 
  same_solid3[tpst3->NumOfCurves] = tmp_same_solid[or_indx];
  tpst3->AdjInfo[tpst3->NumOfCurves++] = tmp_tpst1.AdjInfo[or_indx];

  free_uv_spline(uv_il_bot);
  free_uv_spline(uv_il_top);
  free_uv_spline(uv_ir_bot);
  free_uv_spline(uv_ir_top);
  free_uv_spline(uv_ol_bot);
  free_uv_spline(uv_ol_top);
  free_uv_spline(uv_or_bot);
  free_uv_spline(uv_or_top);
  free_xyz_spline(xyz_il_bot);
  free_xyz_spline(xyz_il_top);
  free_xyz_spline(xyz_ir_bot);
  free_xyz_spline(xyz_ir_top);
  free_xyz_spline(xyz_ol_bot);
  free_xyz_spline(xyz_ol_top);
  free_xyz_spline(xyz_or_bot);
  free_xyz_spline(xyz_or_top);
  free_uv_spline(uv_left_part);
  free_uv_spline(uv_right_part);
  free_xyz_spline(xyz_left_part);
  free_xyz_spline(xyz_right_part);
  free_xyz_spline(tmp_xyz_spl);
}
  
/* This routine takes a set of curves and partitions the trimming */
/* splinegon with it. Right now, loops lying completely inside the */
/* splinegon are treated in the old way ie., keep resulting splits */
/* devoid of holes. pnum indicates patchnum and tpnum indicates the */
/* trim portion number */
int PartitionSplinegon(tpst, first_sol, pnum, tpnum, mc, num_mc, op_tpst,  
		       n_info, solid, sts_arr, sts_arr_len, ovr_loop,
		       other_solid) 
TRIM_PORTION_SPLINE_TYPE *tpst, **op_tpst;
MERGED_CURVE *mc;
int num_mc, pnum, tpnum, first_sol;
neighbour_t *n_info;
TRIM_PATCH_INFO **solid;
SPLIT_TPST_STRUCT *sts_arr;
int *sts_arr_len;
short *ovr_loop;
TRIM_PATCH_INFO *other_solid;
{
  register int found;
  int i, j, k, l, c_num;
  short cntr[MAX_SPLINES_PER_TRIM_REGION]; /* 30 indicates maximum number of curves in a */
                                          /* splinegon */
  POINT2D p2d[20]; /* allocated for 2*num_mc */
#ifdef RECTIFY_SPACE_CURVE
  POINT4D p4d[20];
#endif

  INTERSECTION_CURVE *sptr, *lptr;
  PART_SPLINE_TYPE pst[20];
  UV_SPLINE_TYPE *tuvptr;
  XYZ_SPLINE_TYPE *txyzptr;
  double param[10];
  TRIM_PORTION_SPLINE_TYPE new_tpst;
  short pst_or_st[100]; /* if 0 then original trim curve, else */
			/* intersection curve */
  short mc_top, stack_ptr; /* maintains top of stack */
  short c_index[100], sub_c_index[100], mc_index[100];
  int push, NumRegions, num_non_loops, num_loops, non_loop_mc, comps, deg;
  short loop_curve_nums[MAX_LOOPS_PER_PATCH]; /* Gives indices in */
					      /* mc which are loops */
  short *same_solid, *same_solid1, *same_solid2;
  short *spec_crvs_arr;
  point_typ pt;

  if (num_mc > 10) {
    printf("PartitionSplinegon: Too many merged curves\n");
    exit(-1);
  }

  if (first_sol) 
    spec_crvs_arr = &special_crvs_in_first_sol[pnum][0];
  else
    spec_crvs_arr = &special_crvs_in_sec_sol[pnum][0];

  if (num_mc == 0) {
    if (first_sol) 
      same_solid = &first_same_solid[pnum][tpnum][0][0];
    else
      same_solid = &second_same_solid[pnum][tpnum][0][0];
    copy_trim_portion_spline_type(&(*op_tpst)[0], tpst, 0);

    for (i=0; i<tpst->NumOfCurves; i++)
      if (spec_crvs_arr[i] != -1) {
	same_solid[i] = 0;
	(*op_tpst)[0].AdjInfo[i] = spec_crvs_arr[i];

	/* Now I am filling in the ngh_crv structure so that I can */
	/* change the same_solid and adjacent patch number after the */
	/* call to PartitionSplinegon */
	deg = (*op_tpst)[0].XYZ_Splines[i].Deg;
	ngh_crv[NumNghCrvs].spt = (*op_tpst)[0].XYZ_Splines[i].CntrlPts[0];
	ngh_crv[NumNghCrvs].ept = (*op_tpst)[0].XYZ_Splines[i].CntrlPts[deg];
	ngh_crv[NumNghCrvs].patchNo = tpst->AdjInfo[i];
	ngh_crv[NumNghCrvs++].adj_pnum = spec_crvs_arr[i];
      } else
	same_solid[i] = 1;
    return 1;
  }

  /* check if there are any closed loops. */
  num_loops = 0; 
  for (i=0; i<num_mc; i++) {
    return_st_end_in_p2d(&mc[i], &p2d[2*i], &p2d[2*i+1], &sptr, &lptr);
#ifdef RECTIFY_SPACE_CURVE
    p4d[2*i] = sptr->xyz_spline.CntrlPts[0];
    p4d[2*i+1] = lptr->xyz_spline.CntrlPts[lptr->xyz_spline.Deg];
#endif
    if (EQUAL2D_EPS(p2d[2*i], p2d[2*i+1], 1e-5)) {
      printf("PartitionSplinegon:loops as merged curves. \n");
      if (num_loops == MAX_LOOPS_PER_PATCH-1) {
	printf("PartitionSplinegon: Too many loops\n");
	exit(-1);
      }
      loop_curve_nums[num_loops++] = i;
    }
  }

  num_non_loops = num_mc - num_loops;

  if (num_loops > 1) {
    printf("PartitionSplinegon Warning: Too many loops to handle. Might result in incorrect partitioning\n");
  }

  /* Resolve all curves that are not loops */
  for (k=0,i=0; i<num_mc; i++) {
    non_loop_mc = 1;
    for (j=0; j<num_loops; j++)
      if (loop_curve_nums[j] == i) {
	non_loop_mc = 0;
	break;
      }
    if (!non_loop_mc)
      continue;
    for (l=0; l<2; l++) {
      j = 0;
      found = 0;
      if (first_sol) 
	c_num = fXsc[pnum][i][l][j];
      else
	c_num = sXsc[pnum][i][l][j];
      while ((c_num != -1) && (j <= 7))
	if (!EvaluateParam(&tpst->UV_Splines[c_num], &p2d[2*i+l],
			   &pst[k].param)) {
	  j++;
	  if (first_sol) 
	    c_num = fXsc[pnum][i][l][j];
	  else
	    c_num = sXsc[pnum][i][l][j];
	} else {
	  found = 1;
	  pst[k].spl_c_num = c_num;
	  pst[k].ending = l;
#ifdef RECTIFY_SPACE_CURVE
	  pst[k].p4d = p4d[2*i+l];
#endif
	  pst[k++].merge_c_num = i;
	  break;
	}
      if (!found) {
	printf("PartitionSplinegon: Could not find Xsection point for curve %d\n", i); 
	exit(-1);
      }
    }
  }

  if (k > 0) {
#ifdef RECTIFY_SPACE_CURVE
    POINT4D tmpp4d[20];
    int n;
#endif

    /* 2*num_mc because there are so many endpts for num_mc curves */
    qsort(&pst[0], k /*2*num_mc*/, sizeof(PART_SPLINE_TYPE), comparCv);
    i = j = 0;
    while (i < tpst->NumOfCurves) {
      if ((pst[j].spl_c_num < i) && (j < k /*2*num_mc*/)) {
	printf("PartitionSplinegon: something wrong in double loop\n");
	exit(-1);
      }
      cntr[i] = 0;
      while ((pst[j].spl_c_num == i) && (j < k /*2*num_mc*/)) {
	if (cntr[i] > 7) {
	  printf("PartitionSplinegon: Too many intersection pts for one curve\n");
	  exit(-1);
	}
#ifdef RECTIFY_SPACE_CURVE
	tmpp4d[cntr[i]] = pst[j].p4d;
#endif
	param[cntr[i]++] = pst[j++].param;
      }
      param[cntr[i]] = 1.0;
      tuvptr = (UV_SPLINE_TYPE *) uvst[i];
      txyzptr = (XYZ_SPLINE_TYPE *) xyzst[i];
      SubdivideTrimCurve(&tpst->UV_Splines[i], &tpst->XYZ_Splines[i],
			 cntr[i], param, tuvptr, txyzptr);
#ifdef RECTIFY_SPACE_CURVE
      if (cntr[i] > 0) {
	txyzptr[0].CntrlPts[txyzptr[0].Deg] = tmpp4d[0];
	txyzptr[cntr[i]].CntrlPts[0] = tmpp4d[cntr[i]-1];
	for (n=1; n<cntr[i]; n++) {
	  txyzptr[n].CntrlPts[0] = txyzptr[n-1].CntrlPts[txyzptr[n-1].Deg];
	  txyzptr[n].CntrlPts[txyzptr[n].Deg] = tmpp4d[n];
	}
      }
#endif
      i++;
    }
    
    NumRegions = 0;
    mc_top = stack_ptr = -1;
    i = j = 0;
    while (i < tpst->NumOfCurves) {
      pst_or_st[++stack_ptr] = 0; /* original trim curve */
      c_index[stack_ptr] = i;
      sub_c_index[stack_ptr] = 0;
      
      for (k=0; k<cntr[i]; k++) {
	if (pst[j].spl_c_num != i) {
	  printf("PartitionSplinegon: cntr[i] estimate wrong\n");
	  exit(-1);
	}
	/* determine if you have to push or pop */    
	if ((mc_top != -1) && (pst[j].merge_c_num == mc_index[mc_top]))
	  push = 0;
	else
	  push = 1;
	if (push) {
	  pst_or_st[++stack_ptr] = 1; /* intersection curve */
	  c_index[stack_ptr] = j;
	  mc_index[++mc_top] = pst[j].merge_c_num;
	} else { /* pop */
	  /* decrement mc_top to maintain consistency, then */
	  /* go through the stack popping out the curves until the */
	  /* appropriate Xsection pt. is reached. Then push the */
	  /* intersection curve back into the stack. THE REGION NUMBER */
	  /* IS THE SAME AS THE MERGED CURVE NO. IT SIMPLIFIES N_INFO */
	  /* DETERMINATION */
	  mc_top--;
	  /* update n_info */
	  if (mc_top != -1) {
	    n_info[pst[j].merge_c_num].adj[n_info[pst[j].merge_c_num].nnb] = 
	      mc_index[mc_top]; 
	    n_info[pst[j].merge_c_num].nnb++;
	    n_info[mc_index[mc_top]].adj[n_info[mc_index[mc_top]].nnb] =
	      pst[j].merge_c_num;
	    n_info[mc_index[mc_top]].nnb++;
	  } else {
	    n_info[pst[j].merge_c_num].adj[n_info[pst[j].merge_c_num].nnb] = 
	      num_mc; 
	    n_info[pst[j].merge_c_num].nnb++;
	    n_info[num_mc].adj[n_info[num_mc].nnb] = pst[j].merge_c_num;
	    n_info[num_mc].nnb++;
	  }
	  if (first_sol) 
	    same_solid = &first_same_solid[pnum][tpnum][pst[j].merge_c_num][0];
	  else
	    same_solid = &second_same_solid[pnum][tpnum][pst[j].merge_c_num][0];
	  ReadOutNextRegion(pst, mc, pst_or_st, c_index, sub_c_index,
			    &stack_ptr, pst[j].merge_c_num,
			    tpst, &(*op_tpst)[pst[j].merge_c_num],
			    same_solid, spec_crvs_arr);
	}
	j++;
	/* when we encounter a piece of original trim curve, we must */
	/* push always */
	pst_or_st[++stack_ptr] = 0; /* original trim curve */
	c_index[stack_ptr] = i;
	sub_c_index[stack_ptr] = k+1;
      }
      i++;
    }
    /* Now there is one last region left which must be read out */
    if (first_sol) 
      same_solid = &first_same_solid[pnum][tpnum][num_mc][0];
    else
      same_solid = &second_same_solid[pnum][tpnum][num_mc][0];
    ReadOutNextRegion(pst, mc, pst_or_st, c_index, sub_c_index,
		      &stack_ptr, -1, tpst, 
		      &(*op_tpst)[num_mc], same_solid, spec_crvs_arr);
  } else {
    /* This case arises when there are no non loop cases. In that */
    /* case, I have to simply copy the whole tpst into the region 1 of */
    /* op_tpst because that would mean that curve 0 is a loop. */
    /* Further, I have the restriction that I cannot handle more than */
    /* one loop inside a region. */
    copy_trim_portion_spline_type(&(*op_tpst)[1], tpst, 0);
    n_info[1].nnb = 0;
    if (first_sol) 
      same_solid = &first_same_solid[pnum][tpnum][1][0];
    else
      same_solid = &second_same_solid[pnum][tpnum][1][0];
    for (i=0; i<tpst->NumOfCurves; i++)
      same_solid[i] = 1;
  }

  /* Resolve all the loops */
  comps = num_mc + 1;
  /* Each loop has an merged curve number which is not used in the */
  /* stack algorithm. So I am using that to save space in op_tpst */
  /* array */
  for (i=0; i<num_loops; i++) {
    pt.x = p2d[2*loop_curve_nums[i]].u;
    pt.y = p2d[2*loop_curve_nums[i]].v;
    for (j=0; j<comps; j++) {
      if (j == loop_curve_nums[i])
	continue;
      if (IsPtInsideSplinegon(&(*op_tpst)[j], &pt))
	break;
    }
    if (j == comps) {
      printf("PartitionSplinegon: mistake - loop not inside patch\n");
      exit(-1);
    }

    /* assigning neighboring info for the loop region */
    n_info[loop_curve_nums[i]].adj[n_info[loop_curve_nums[i]].nnb++] = j;
    n_info[j].adj[n_info[j].nnb++] = loop_curve_nums[i];

    if (first_sol) 
      same_solid = &first_same_solid[pnum][tpnum][loop_curve_nums[i]][0];
    else
      same_solid = &second_same_solid[pnum][tpnum][loop_curve_nums[i]][0];
    (*op_tpst)[loop_curve_nums[i]].NumOfCurves = 0;
    (*op_tpst)[loop_curve_nums[i]].UV_Splines = (UV_SPLINE_TYPE *) 
      new(MAX_SPLINES_PER_TRIM_REGION * sizeof(UV_SPLINE_TYPE));
    (*op_tpst)[loop_curve_nums[i]].XYZ_Splines = (XYZ_SPLINE_TYPE *)
      new(MAX_SPLINES_PER_TRIM_REGION * sizeof(XYZ_SPLINE_TYPE));
    (*op_tpst)[loop_curve_nums[i]].AdjInfo = (int *)
      new(MAX_SPLINES_PER_TRIM_REGION * sizeof(int));
    AddMergedCurvetoRegion(mc, loop_curve_nums[i],
			   &(*op_tpst)[loop_curve_nums[i]], 0,
			   same_solid);
    /* I am going to make the two entries after the last one also 0 */
    /* because anyway this tpst has to be split (done in */
    /* SplitAndCreateNewRegion) and I get two more curves */
    same_solid[(*op_tpst)[loop_curve_nums[i]].NumOfCurves] = 
      same_solid[(*op_tpst)[loop_curve_nums[i]].NumOfCurves+1] = 0;

    if (!IsSplinegonAntiClockwise(&(*op_tpst)[loop_curve_nums[i]]))
      ReverseTPST(&(*op_tpst)[loop_curve_nums[i]],
		  &(*op_tpst)[loop_curve_nums[i]]); 

    /* Now I have to split the jth splinegon and create one more */
    /* region. I have to also create neighboring information for the */
    /* newly created region with number comps. Since I am splitting */
    /* region j to create regions j and comps, all I do in the graph */
    /* is to make all the neighbors of j also neighbors of comps */
    /* (symmetrically). This is because even though j and comps are */
    /* adjacent, they belong to the same region and should not have an */
    /* edge between them */
    for (l=0; l<n_info[j].nnb; l++) {
      n_info[comps].adj[n_info[comps].nnb++] = n_info[j].adj[l];
      n_info[n_info[j].adj[l]].adj[n_info[n_info[j].adj[l]].nnb++] = comps;
    }

    if (first_sol) {
      same_solid1 = &first_same_solid[pnum][tpnum][j][0];
      same_solid2 = &first_same_solid[pnum][tpnum][comps][0];
    }
    else {
      same_solid1 = &second_same_solid[pnum][tpnum][j][0];
      same_solid2 = &second_same_solid[pnum][tpnum][comps][0];
    }
    SplitAndCreateNewRegion(&(*op_tpst)[j], &(*op_tpst)[loop_curve_nums[i]],
			    &(*op_tpst)[comps], &(*solid)[pnum].tp,
			    pnum, first_sol, 
			    same_solid1, same_solid2, sts_arr,
			    sts_arr_len, ovr_loop[loop_curve_nums[i]],
			    other_solid); 
    comps++;
  }

  return num_mc+num_loops+1;
}

int GetNextUmarkedRegion(comp_assgn_arr, num_patches, num_partns_arr,
			 pNo, rNo)
int *comp_assgn_arr, num_patches, *num_partns_arr, *pNo, *rNo;
/* pNo and rNo are patchno and region number resply - to be returned */
{
  int i, j;

  *pNo = *rNo = -1;
  for (i=0; i<num_patches; i++) {
    for (j=0; j<num_partns_arr[i]; j++) 
      if (comp_assgn_arr[i*MAX_PARTNS_PER_PATCH+j] == -1) {
	*pNo = i;
	*rNo = j;
	break;
      }
    if (*pNo != -1) break;
  }
  
  if (*pNo == -1)
    return -1;

  return 0;
}

/* This determines the partition number one of whose bounding curves */
/* have their endpts being fpt and spt. rnum is -1 usually except when */
/* I am finding adjacent partition from the same patch in which case */
/* it is the region number I should not return. This happens when */
/* there are loops in the merged curve, and I have to partition the */
/* toroidal region. It also returns the curve number in cnum. dir */
/* returns if the match occured with the curve in the same direction */
/* (dir == 1) or not (dir == 0) */
short GetAdjacentPartn(tpst, pnum, num_partns, fpt, spt, rnum, cnum,
		       dir, ptch)
TRIM_PORTION_SPLINE_TYPE **tpst;
int pnum, num_partns;
POINT4D *fpt, *spt;
short *cnum, *dir;
int ptch;
{
  short i, j, ret;
  int deg;
  POINT3D p1, p2, p3, p4;
  TRIM_PORTION_SPLINE_TYPE *tptr;
  
  p1.dCoord[0] = fpt->dCoord[0] / fpt->dCoord[3];
  p1.dCoord[1] = fpt->dCoord[1] / fpt->dCoord[3];
  p1.dCoord[2] = fpt->dCoord[2] / fpt->dCoord[3];

  p2.dCoord[0] = spt->dCoord[0] / spt->dCoord[3];
  p2.dCoord[1] = spt->dCoord[1] / spt->dCoord[3];
  p2.dCoord[2] = spt->dCoord[2] / spt->dCoord[3];
  tptr = &tpst[pnum][0];
  ret = -1;

  for (i=0; i<num_partns; i++) {
    /* This is to test if the current region number is one given as */
    /* parameter. If so I have to not check for matching points. This */
    /* is again a loop case if i == rnum. Usually, rnum is -1 so */
    /* checking works even then. */
    if (i == rnum)
      continue;
    for (j=0; j<tptr[i].NumOfCurves; j++) {
      deg = tptr[i].XYZ_Splines[j].Deg;

      p3.dCoord[0] = tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[0] / 
	tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[3];
      p3.dCoord[1] = tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[1] / 
	tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[3];
      p3.dCoord[2] = tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[2] / 
	tptr[i].XYZ_Splines[j].CntrlPts[0].dCoord[3];

      p4.dCoord[0] = tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[0] / 
	tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[3];
      p4.dCoord[1] = tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[1] / 
	tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[3];
      p4.dCoord[2] = tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[2] / 
	tptr[i].XYZ_Splines[j].CntrlPts[deg].dCoord[3];

      if (EQUAL3D_ABS(p1, p3, 5e-4) && EQUAL3D_ABS(p2, p4, 5e-4)) {
	ret = i;
	*cnum = j;
	*dir = 1;
	break;
      }
      if (EQUAL3D_ABS(p1, p4, 5e-4) && EQUAL3D_ABS(p2, p3, 5e-4)) {
	ret = i;
	*cnum = j;
	*dir = 0;
	break;
      }
    }
    if (ret != -1) 
      break;
  }

  if (ret == -1) {
    printf("GetAdjacentPartn: couldn't find adjacent partition for patch %d\n", ptch);
    exit(-1);
  }

  return ret;
}

/* This determines the trim portion number one of whose bounding curves */
/* have their endpts being fpt and spt. cnum is returned. It is the */
/* number of the curve in tpi which matches the endpts */
short GetAdjacentTrimPortion(tpi, fpt, spt, cnum, rnum, dir)
TRIM_PATCH_INFO *tpi;
POINT4D *fpt, *spt;
int *cnum, rnum;
short *dir;
{
  short i, j, ret;
  int deg, num_partns;
  POINT3D p1, p2, p3, p4;
  TRIM_PORTION_SPLINE_TYPE *tptr;
  
  p1.dCoord[0] = fpt->dCoord[0] / fpt->dCoord[3];
  p1.dCoord[1] = fpt->dCoord[1] / fpt->dCoord[3];
  p1.dCoord[2] = fpt->dCoord[2] / fpt->dCoord[3];

  p2.dCoord[0] = spt->dCoord[0] / spt->dCoord[3];
  p2.dCoord[1] = spt->dCoord[1] / spt->dCoord[3];
  p2.dCoord[2] = spt->dCoord[2] / spt->dCoord[3];
  ret = -1;

  num_partns = tpi->NumTrimPortions;
  for (i=0; i<num_partns; i++) {
    if (i == rnum) 
      continue;
    tptr = &tpi->TrimPortion[i].SplinePortion;
    for (j=0; j<tptr->NumOfCurves; j++) {
      deg = tptr->XYZ_Splines[j].Deg;

      p3.dCoord[0] = tptr->XYZ_Splines[j].CntrlPts[0].dCoord[0] / 
	tptr->XYZ_Splines[j].CntrlPts[0].dCoord[3];
      p3.dCoord[1] = tptr->XYZ_Splines[j].CntrlPts[0].dCoord[1] / 
	tptr->XYZ_Splines[j].CntrlPts[0].dCoord[3];
      p3.dCoord[2] = tptr->XYZ_Splines[j].CntrlPts[0].dCoord[2] / 
	tptr->XYZ_Splines[j].CntrlPts[0].dCoord[3];

      p4.dCoord[0] = tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[0] / 
	tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[3];
      p4.dCoord[1] = tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[1] / 
	tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[3];
      p4.dCoord[2] = tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[2] / 
	tptr->XYZ_Splines[j].CntrlPts[deg].dCoord[3];

      if (EQUAL3D_ABS(p1, p3, 5e-4) && EQUAL3D_ABS(p2, p4, 5e-4)) {
	ret = i;
	*cnum = j;
	*dir = 1;
	break;
      }
      if (EQUAL3D_ABS(p1, p4, 5e-4) && EQUAL3D_ABS(p2, p3, 5e-4)) {
	ret = i;
	*cnum = j;
	*dir = 0;
	break;
      }
    }
    if (ret != -1) 
      break;
  }

  if (ret == -1) {
    printf("GetAdjacentTrimPortion: could not find adjacent trim portion\n");
    exit(-1);
  }

  return ret;
}

/* This routine goes through all the bounding curves and assigns the */
/* same component number recursively */
void AssignAllNeighbors(tpst, pnum, rnum, first, comp_assign_arr, 
			num_partns, comp_num)
TRIM_PORTION_SPLINE_TYPE **tpst;
int pnum, rnum, comp_num, *comp_assign_arr, first, *num_partns;
{
  int i, j, k, n_curves, deg;
  short *same_solid, partn_num, cnum, dir;
  POINT4D first_pt, second_pt;

  if (first) 
    same_solid = &first_same_solid[pnum][0][rnum][0];
  else
    same_solid = &second_same_solid[pnum][0][rnum][0];
  n_curves = tpst[pnum][rnum].NumOfCurves;

  i = 0;
  while (i < n_curves) {
    if (!same_solid[i]) {
      i++;
      continue;
    }
    j = tpst[pnum][rnum].AdjInfo[i];
    /* we now want to find the neighbor of curve i of pnum in patch j */
    first_pt = tpst[pnum][rnum].XYZ_Splines[i].CntrlPts[0];
    deg = tpst[pnum][rnum].XYZ_Splines[i].Deg;
    second_pt = tpst[pnum][rnum].XYZ_Splines[i].CntrlPts[deg];
    if (j == pnum)
      partn_num = GetAdjacentPartn(tpst, j, num_partns[j], &first_pt,
				   &second_pt, rnum, &cnum, &dir,
				   pnum);
    else
      partn_num = GetAdjacentPartn(tpst, j, num_partns[j], &first_pt,
				   &second_pt, -1, &cnum, &dir, pnum);

    if (comp_assign_arr[j*MAX_PARTNS_PER_PATCH+partn_num] != -1) {
      i++;
      continue;
    }
    comp_assign_arr[j*MAX_PARTNS_PER_PATCH+partn_num] = comp_num;
    AssignAllNeighbors(tpst, j, partn_num, first, comp_assign_arr,
		       num_partns, comp_num);
  }
}

/* This routine takes the new partitioned region of a solid and */
/* assigns each region a unique global comp. number. first is just to */
/* indicate if it is the first solid or not. num_comp gives the total */
/* number of global components this solid will have */
/* int comp_assgn_arr[MAX_PATCHES_PER_SOLID*MAX_PARTNS_PER_PATCH]; */
                                        /* this is 100 times 10, so comp */
                                        /* number for patch i region j */
				        /* is in index 10*i+j */
void ObtainNewGlobalGraph(tpst, num_patches, num_partns_arr, n_info,
			  first, comp_assgn_arr, num_comp, g_graph)
TRIM_PORTION_SPLINE_TYPE **tpst;
int num_patches, *num_partns_arr, first, *num_comp;
neighbour_t n_info[][MAX_PARTNS_PER_PATCH], **g_graph;
int *comp_assgn_arr;
{
  int i, j, k, l;
  int next_comp_num, glob_comp_num1, glob_comp_num2, not_assigned;
  short *same_solid;

  /* initialize each element of comp_assgn_arr to -1 first except the */
  /* first partition of the first patch */
  for (i=0; i<MAX_PATCHES_PER_SOLID*MAX_PARTNS_PER_PATCH; i++)
    comp_assgn_arr[i] = -1;
  next_comp_num = 0;

  while (GetNextUmarkedRegion(&comp_assgn_arr[0], num_patches,
			      num_partns_arr, &i, &j) != -1) {
    comp_assgn_arr[i*MAX_PARTNS_PER_PATCH+j] = next_comp_num++;
    /* to obtain global comp number */
    AssignAllNeighbors(tpst, i, j, first, comp_assgn_arr, 
		       num_partns_arr, next_comp_num-1);
  }
	
  (*num_comp) = next_comp_num;
  ALLOCN((*g_graph), neighbour_t, (*num_comp));

  /* I have computed global region numbers of every */
  /* region. Now I have to get the global graph */
  for (k=0; k<*num_comp; k++)
    (*g_graph)[k].nnb = 0;

  for (i=0; i<num_patches; i++)
    for (j=0; j<num_partns_arr[i]; j++) {
      glob_comp_num1 = comp_assgn_arr[i*MAX_PARTNS_PER_PATCH+j];
      for (k=0; k<n_info[i][j].nnb; k++) {
	glob_comp_num2 = 
	  comp_assgn_arr[i*MAX_PARTNS_PER_PATCH+n_info[i][j].adj[k]];
	not_assigned = 1;
	for (l=0; l<(*g_graph)[glob_comp_num1].nnb; l++)
	  if ((*g_graph)[glob_comp_num1].adj[l] == glob_comp_num2) {
	    not_assigned = 0;
	    break;
	  }
	if (not_assigned) {
	  (*g_graph)[glob_comp_num1].adj[(*g_graph)[glob_comp_num1].nnb++] = 
	    glob_comp_num2;
	  (*g_graph)[glob_comp_num2].adj[(*g_graph)[glob_comp_num2].nnb++] = 
	    glob_comp_num1;
	}
      }
    }
}

int comparParam(v1, v2)
double *v1, *v2;
{
  if (FP_EQ_EPS((*v1), (*v2), 1e-3)) return 0;
  return (int)100.0*((*v1) - (*v2));
}

int comparArbStruct(v1, v2)
ARBIT_STRUCT *v1, *v2;
{
  if (FP_EQ_EPS((*v1).param, (*v2).param, 1e-2)) return 0;
  return ((*v1).param - (*v2).param);
}

/* This routine is called from ProjectPatch. Here I just subdivide the */
/* crv_num-th curve using the param_arr and displace the other curves */
/* appropriately. */
void AdjustTrimCurves(tpst, crv_num, param_arr, param_length)
TRIM_PORTION_SPLINE_TYPE *tpst;
int crv_num, param_length;
double *param_arr;
{
  int i, j;
  int num_curves;

/*  
  if (tpi->NumTrimPortions != 1) {
    printf("AdjustTrimCurves: Too many trim portions\n");
    exit(-1);
  }
  tpst = &tpi->TrimPortion[0].SplinePortion;
*/
  if (param_length <= 0) return;

  num_curves = tpst->NumOfCurves;

  if (crv_num >= num_curves) {
    printf("AdjustTrimCurves: Wrong crv_num parameter\n");
    exit(-1);
  }
  
  if (num_curves + param_length >= MAX_SPLINES_PER_TRIM_REGION) {
    printf("AdjustTrimCurves: Too many trim curves in one region\n");
    exit(-1);
  }

  for (i=(num_curves-1); i>crv_num; i--) {
    copy_uv_spline(&tpst->UV_Splines[i], &tpst->UV_Splines[i+param_length]);
    copy_xyz_spline(&tpst->XYZ_Splines[i], &tpst->XYZ_Splines[i+param_length]);
    tpst->AdjInfo[i+param_length] = tpst->AdjInfo[i];
  }
  tpst->NumOfCurves += param_length;
  for (i=crv_num+1; i<=crv_num+param_length; i++)
    tpst->AdjInfo[i] = tpst->AdjInfo[crv_num];

  SubdivideTrimCurve(&tpst->UV_Splines[crv_num],
		     &tpst->XYZ_Splines[crv_num], param_length,
		     param_arr, &tpst->UV_Splines[crv_num],
		     &tpst->XYZ_Splines[crv_num]); 
}

/* This routine takes the "curve_num"th curve of trimpatch tpi2 and
   projects it onto the domain of tpi1. This is called only when the
   two patches overlap. uv_crv and xyz_crv are the portions of the
   original curve that lie inside the trimmed region of tpi1. */
int ProjectPatch(sol, indx, tpi2, curve_num, uv_crv, xyz_crv,
		 do_partn, bndry_crv_info, bci_size, ssos)
TRIM_PATCH_INFO *sol, *tpi2;
int curve_num, indx;
UV_SPLINE_TYPE **uv_crv;
XYZ_SPLINE_TYPE **xyz_crv;
int do_partn;
ANOTHER_ARBIT_STRUCT *bndry_crv_info;
int *bci_size;
SAME_SIDE_OVR_STRUCT *ssos;
{
  int i, j, k;
  int neigh;
  point_typ pt;
  int deg, NumXsections, NumCurves, tmpXns;
  int initial_cond; /* This tells if the first point of the curve was */
		    /* inside (== 1) or outside (== 0) the trimmed */
		    /* region */
  ARBIT_STRUCT XsectionArray[20]; /* I am assuming a max of 20 Xns between */
                                  /* curve and trimmed region */
  TRIM_PATCH_INFO *tpi1;
  double uval, vval, wt;
  double x, y, z;
  double *wt1, *wt2;
#ifdef _OLD
  double *s, *t;
#else
  double s[10], t[10];
#endif
  POINT4D fpt, ept;
  UV_SPLINE_TYPE transformed_uv, *tmp_uv, *tmp_uv1;
  UV_SPLINE_TYPE *tmp_uv_crv;
  XYZ_SPLINE_TYPE *tmp_xyz;
  XYZ_SPLINE_TYPE *tmp_xyz_crv;
  TENSOR_PATCH tp1, tp2;
  ParamCurve curve1, curve2;
  double param_arr[10], *param_arr2;
  int cnum, par_arr2_indx;
  short dir;

  if (ssos != NULL)
    param_arr2 = &ssos->param_arr[0];

  tpi1 = &sol[indx];

#ifdef _OLD
  curve1.control = curve2.control = NULL;
  wt1 = wt2 = NULL;
  s = t = NULL;
#endif

  tmp_uv = &tpi2->TrimPortion[0].SplinePortion.UV_Splines[curve_num];
  tmp_xyz = &tpi2->TrimPortion[0].SplinePortion.XYZ_Splines[curve_num];

  deg = tmp_uv->Deg;
  transformed_uv.Deg = deg;
  ALLOCN(transformed_uv.CntrlPts, POINT3D, (deg+1));
  NumXsections = 0;

  /* I am going to copy tpi1->tp into tp1 and tpi2->tp into tp2 and */
  /* scale them so that I can get accurate values from */
  /* special_plane_case */
  tp1.iDeg0 = tpi1->tp.iDeg0;
  tp1.iDeg1 = tpi1->tp.iDeg1;
  allocate_tensor_patch(&tp1);
  tp2.iDeg0 = tpi2->tp.iDeg0;
  tp2.iDeg1 = tpi2->tp.iDeg1;
  allocate_tensor_patch(&tp2);
  Scale = -1.0;
  for (i=0; i<=tp1.iDeg0; i++)
    for (j=0; j<=tp1.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tpi1->tp.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tpi1->tp.coeffs[i][j].dCoord[k]);
	tp1.coeffs[i][j].dCoord[k] = tpi1->tp.coeffs[i][j].dCoord[k];
      }
      tp1.coeffs[i][j].dCoord[3] = tpi1->tp.coeffs[i][j].dCoord[3];
    }
  for (i=0; i<=tp2.iDeg0; i++)
    for (j=0; j<=tp2.iDeg1; j++) {
      for (k=X; k<=Z; k++) {
	if (Scale < fabs(tpi2->tp.coeffs[i][j].dCoord[k]))
	  Scale = fabs(tpi2->tp.coeffs[i][j].dCoord[k]);
	tp2.coeffs[i][j].dCoord[k] = tpi2->tp.coeffs[i][j].dCoord[k];
      }
      tp2.coeffs[i][j].dCoord[3] = tpi2->tp.coeffs[i][j].dCoord[3];
    }
  /* Scale the tensor patch */
  scale_tensor_patch(&tp1, 1);
  scale_tensor_patch(&tp2, 1);
      
  /* restore Scale */
  Scale = 1.0;

  for (i=0; i<=deg; i++) {
    uval = tmp_uv->CntrlPts[i].dCoord[0] / tmp_uv->CntrlPts[i].dCoord[2];
    vval = tmp_uv->CntrlPts[i].dCoord[1] / tmp_uv->CntrlPts[i].dCoord[2];
    wt = tmp_uv->CntrlPts[i].dCoord[2];
    EvalPointOnPatch(tp2, uval, vval, &x, &y, &z);
    special_plane_case(&tp1, x, y, z, &uval, &vval);
    transformed_uv.CntrlPts[i].dCoord[0] = uval * wt;
    transformed_uv.CntrlPts[i].dCoord[1] = vval * wt;
    transformed_uv.CntrlPts[i].dCoord[2] = wt;
  }

  /* Set the initial condition by looking at the first control point */
  pt.x = transformed_uv.CntrlPts[0].dCoord[0] / transformed_uv.CntrlPts[0].dCoord[2];
  pt.y = transformed_uv.CntrlPts[0].dCoord[1] / transformed_uv.CntrlPts[0].dCoord[2];
  initial_cond = IsPtInsideSplinegon(&tpi1->TrimPortion[0].SplinePortion, &pt);
  /* if the first point was inside the polygon, add 0.0 to the */
  /* intersection list */
  if (initial_cond) {
    XsectionArray[NumXsections].param = 0.0;
    XsectionArray[NumXsections++].bndry_crv_num = -1;
  }

  par_arr2_indx = 0;
  for (j=0; j<tpi1->TrimPortion[0].SplinePortion.NumOfCurves; j++) {
    tmp_uv1 = &tpi1->TrimPortion[0].SplinePortion.UV_Splines[j];
    neigh = tpi1->TrimPortion[0].SplinePortion.AdjInfo[j];
    deg = tpi1->TrimPortion[0].SplinePortion.XYZ_Splines[j].Deg;
    fpt = tpi1->TrimPortion[0].SplinePortion.XYZ_Splines[j].CntrlPts[0];
    ept = tpi1->TrimPortion[0].SplinePortion.XYZ_Splines[j].CntrlPts[deg];
    if (neigh == indx) {
      printf("ProjectPatch: neighboring patch of one solid is itself\n");
      exit(-1);
    }

#ifdef _OLD
    curve1.degree = tmp_uv1->Deg;
    curve2.degree = transformed_uv.Deg;
    curve1.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
					 (curve1.degree+1));
    curve2.control = (FiveTuple*) malloc(sizeof(FiveTuple) *
					 (curve2.degree+1));
    wt1 = (double *) malloc(sizeof(double) * (curve1.degree+1));
    wt2 = (double *) malloc(sizeof(double) * (curve2.degree+1));
    
    for (k=0; k<=curve1.degree; k++) {
      curve1.control[k].s = tmp_uv1->CntrlPts[k].dCoord[0];
      curve1.control[k].t = tmp_uv1->CntrlPts[k].dCoord[1];
      wt1[k] = tmp_uv1->CntrlPts[k].dCoord[2];
    }
    for (k=0; k<=curve2.degree; k++) {
      curve2.control[k].s = transformed_uv.CntrlPts[k].dCoord[0];
      curve2.control[k].t = transformed_uv.CntrlPts[k].dCoord[1];
      wt2[k] = transformed_uv.CntrlPts[k].dCoord[2];
    }
    if (curve2.degree < curve1.degree) {
      check_self_intersect(curve2, wt2, curve1, wt1, &tmpXns, &s, &t);

      /* To maintain proper adjacencies, I split the curve which lies */
      /* on the adjacent patch. do_partn is TRUE only when the solids */
      /* lie on opposite sides of the overlapping patch */
      if ((tmpXns > 0) && do_partn) {
	GetAdjacentTrimPortion(&sol[neigh], &fpt, &ept, &cnum, -1, &dir);
	if (tmpXns > 9) {
	  printf("ProjectPatch: Too many intersections to handle\n");
	  exit(-1);
	}
	for (i=0; i<tmpXns; i++)
	  param_arr[i] = (1-dir) + (2*dir-1)*t[i]; /* this is only to */
						   /* replace t[i] by */
						   /* (1-t[i]) based */
						   /* on dir */
	param_arr[tmpXns] = 1.0;
	qsort(&param_arr[0], tmpXns, sizeof(double), comparParam);
	if (sol[neigh].NumTrimPortions != 1) {
	  printf("ProjectPatch: Too many trim portions\n");
	  exit(-1);
	}
	AdjustTrimCurves(&(sol[neigh].TrimPortion[0].SplinePortion),
			 cnum, &param_arr[0], tmpXns); 
      }

      for (i=0; i<tmpXns; i++) {
	if (NumXsections > 19) 
	  printf("Too many intersection points\n");
	XsectionArray[NumXsections].param = s[i];
	XsectionArray[NumXsections++].bndry_crv_num = j;
	if (ssos != NULL)
	  param_arr2[par_arr2_indx++] = s[i];
      }
    } else {
      check_self_intersect(curve1, wt1, curve2, wt2, &tmpXns, &s, &t);
      if ((tmpXns > 0) && do_partn) {
	GetAdjacentTrimPortion(&sol[neigh], &fpt, &ept, &cnum, -1, &dir);
	if (tmpXns > 9) {
	  printf("ProjectPatch: Too many intersections to handle\n");
	  exit(-1);
	}
	for (i=0; i<tmpXns; i++)
	  param_arr[i] = (1-dir) + (2*dir-1)*s[i]; /* this is only to */
						   /* replace s[i] by */
						   /* (1-s[i]) based */
						   /* on dir */
	param_arr[tmpXns] = 1.0;
	qsort(&param_arr[0], tmpXns, sizeof(double), comparParam);
	if (sol[neigh].NumTrimPortions != 1) {
	  printf("ProjectPatch: Too many trim portions\n");
	  exit(-1);
	}
	AdjustTrimCurves(&(sol[neigh].TrimPortion[0].SplinePortion),
			 cnum, &param_arr[0], tmpXns); 
      }

      for (i=0; i<tmpXns; i++) {
	if (NumXsections > 19) 
	  printf("Too many intersection points\n");
	XsectionArray[NumXsections].param = t[i];
	XsectionArray[NumXsections++].bndry_crv_num = j;
	if (ssos != NULL)
	  param_arr2[par_arr2_indx++] = t[i];
      }
    }
    myfree(curve1.control); curve1.control = NULL;
    myfree(curve2.control); curve2.control = NULL;
    myfree(wt1); myfree(wt2); wt1 = wt2 = NULL;
    myfree(s); myfree(t); s = t = NULL;
#else
    intersect_plane_curves(tmp_uv1, &transformed_uv, &s[0], &t[0], &tmpXns);
    if ((tmpXns > 0) && do_partn) {
      GetAdjacentTrimPortion(&sol[neigh], &fpt, &ept, &cnum, -1, &dir);
      if (tmpXns > 9) {
	printf("ProjectPatch: Too many intersections to handle\n");
	exit(-1);
      }
      for (i=0; i<tmpXns; i++)
	param_arr[i] = (1-dir) + (2*dir-1)*s[i]; /* this is only to */
                                                 /* replace s[i] by */
                                                 /* (1-s[i]) based */
                                                 /* on dir */
      param_arr[tmpXns] = 1.0;
      qsort(&param_arr[0], tmpXns, sizeof(double), comparParam);
      if (sol[neigh].NumTrimPortions != 1) {
	printf("ProjectPatch: Too many trim portions\n");
	exit(-1);
      }
      AdjustTrimCurves(&(sol[neigh].TrimPortion[0].SplinePortion),
		       cnum, &param_arr[0], tmpXns); 
    }
    
    for (i=0; i<tmpXns; i++) {
      if (NumXsections > 19) 
	printf("Too many intersection points\n");
      XsectionArray[NumXsections].param = t[i];
      XsectionArray[NumXsections++].bndry_crv_num = j;
      if (ssos != NULL)
	param_arr2[par_arr2_indx++] = t[i];
    }
#endif
  }

  /* if the NumXsections variable is ODD, then I have to add 1.0 as */
  /* the last parameter value */
  if (NumXsections % 2) {
    XsectionArray[NumXsections].param = 1.0;
    XsectionArray[NumXsections++].bndry_crv_num = -1;
  }
  if (ssos != NULL) {
    param_arr2[par_arr2_indx] = 1.0;
    qsort(&param_arr2[0], par_arr2_indx+1, sizeof(double), comparParam);
  }
  /* I am breaking the boundary of the projected patch so that in the */
  /* case of same_side overlap I dont have to project the first_solid */
  /* patch on the overlapping second_solid patch */
  if (!do_partn && (ssos != NULL)) {
    ssos->curve_num = curve_num;
    ssos->param_length = par_arr2_indx;
    ssos->initial = initial_cond;
    ssos->adj_patch = indx;
  }

  /* Now I have all the intersection parameters for the curve in */
  /* XsectionArray. All I have to do is sort the array and use the */
  /* initial_cond to determine which curves to return */
  qsort(&XsectionArray[0], NumXsections, sizeof(ARBIT_STRUCT), comparArbStruct);
  i = 0;
  tmp_uv_crv = (*uv_crv) = NULL;
  tmp_xyz_crv = (*xyz_crv) = NULL;
  (*uv_crv) = (UV_SPLINE_TYPE *) new((NumXsections/2) * sizeof(UV_SPLINE_TYPE));
  tmp_uv_crv = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  (*xyz_crv) = (XYZ_SPLINE_TYPE *) new((NumXsections/2) * sizeof(XYZ_SPLINE_TYPE));
  tmp_xyz_crv = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));
  while (i < NumXsections) {
    TwoDBezierSubdivision(&transformed_uv, XsectionArray[i].param,
			  XsectionArray[i+1].param, &tmp_uv_crv);
    copy_uv_spline(tmp_uv_crv, &(*uv_crv)[i/2]);
    ThreeDBezierSubdivision(tmp_xyz, XsectionArray[i].param,
			    XsectionArray[i+1].param, &tmp_xyz_crv);
    if (do_partn) {
      bndry_crv_info[(*bci_size)].p4d = tmp_xyz_crv->CntrlPts[0];
      bndry_crv_info[(*bci_size)++].bndry_crv_num = XsectionArray[i].bndry_crv_num;
      bndry_crv_info[(*bci_size)].p4d = tmp_xyz_crv->CntrlPts[tmp_xyz_crv->Deg];
      bndry_crv_info[(*bci_size)++].bndry_crv_num = XsectionArray[i+1].bndry_crv_num;
    }
    copy_xyz_spline(tmp_xyz_crv, &(*xyz_crv)[i/2]);
    i += 2;
  }
  return (NumXsections/2);
}

void AnotherInverseOp(tp0, st_crv, param, tp1, uv_crv, ret_param)
TENSOR_PATCH *tp0, *tp1;
UV_SPLINE_TYPE *st_crv, *uv_crv;
double param, *ret_param;
{
  int i;
  POINT2D p2d;
  double s, t;

  Eval2DCurve(st_crv, param, &p2d);
  GeneralInverseOp(*tp1, *tp0, p2d.u, U, p2d.v, &s, &t);
  p2d.u = s; p2d.v = t;
  EvaluateParam(uv_crv, &p2d, ret_param);
}

/* This procedure takes in an intersection curve, and returns portions
   of that curve that lies inside both the trim patches. This
   procedure replaces Atul's obtain_proper_trimcurve because of
   inaccuracy problems. */
int GetProperTrimCurve(tp0, tp1, st_crv, uv_crv, xyz_crv, 
		       ret_st_crv, ret_uv_crv, ret_xyz_crv)
TRIM_PATCH_INFO *tp0, *tp1;
UV_SPLINE_TYPE *st_crv, *uv_crv, **ret_st_crv, **ret_uv_crv;
XYZ_SPLINE_TYPE *xyz_crv, **ret_xyz_crv;
{
  int i, j, k;
  double s[10], t[10], Xns0[100], Xns1[100], finXns[100], corresp_finXns[100];
  int totXns0, totXns1, fin_numXns;
  int num_crvs0, num_crvs1, numXns;
  int is_inside0, is_inside1;
  short low_flag, high_flag;
  int change;
  point_typ pt;
  TRIM_PORTION_SPLINE_TYPE *tpst0, *tpst1;
  INTERVAL int1, int2, ret_int;
  UV_SPLINE_TYPE *tmp_2d_crv;
  XYZ_SPLINE_TYPE *tmp_3d_crv;
  /* added by tim */
  POINT2D xn_tgt, trim_tgt;
  int rc1, rc2;
  int xn_enters[100];
  double min_xn;

  if ((tp0->NumTrimPortions != 1) || (tp1->NumTrimPortions != 1)) {
    printf("GetProperTrimCurve: Too many trim portions\n");
    exit(-1);
  }
  tpst0 = &tp0->TrimPortion[0].SplinePortion;
  tpst1 = &tp1->TrimPortion[0].SplinePortion;

  num_crvs0 = tpst0->NumOfCurves;
  num_crvs1 = tpst1->NumOfCurves;

  totXns0 = totXns1 = 0;

  /* patch 0 */

  for (i=0; i<num_crvs0; i++) {
    intersect_plane_curves(&tpst0->UV_Splines[i], st_crv, &s[0], &t[0], &numXns);
    if (numXns > 10) {
      printf("GetProperTrimCurve: Too many curve-curve intersections\n");
      exit(-1);
    }
    for (j=0; j<numXns; j++) {
      /*
      if ((FP_EQ_EPS(s[j], 0.0, 1e-3) || FP_EQ_EPS(t[j], 0.0, 1e-4) || 
	   FP_EQ_EPS(t[j], 1.0, 1e-4))) 
	continue;
	*/
      Xns0[totXns0] = t[j];
      /* find out whether it's an entry or an exit to the trimmed portion */
      rc1 = EvaluateTangentVector(&tpst0->UV_Splines[i], s[j], &trim_tgt);
      rc2 = EvaluateTangentVector(st_crv, t[j], &xn_tgt);
      if (rc1 || rc2) {
	printf("GetProperTrimCurve: tangent vector troubles\n");
	exit(-1);
      }
      xn_enters[totXns0] = (xn_tgt.u * trim_tgt.v - xn_tgt.v * trim_tgt.u) < 0;
      ++totXns0;
    }
  }

  /* find first intersection point */
  for (i=0,min_xn=1.0; i<totXns0; i++)
    if (Xns0[i] < min_xn) {
      min_xn = Xns0[i];
      is_inside0 = !xn_enters[i];
    }
  if (!is_inside0)
    Xns0[totXns0++] = 0.0;
  if (totXns0 % 2) 
    Xns0[totXns0++] = 1.0;
  qsort(&Xns0[0], totXns0, sizeof(double), comparParam);

  /* patch 1 */
  /* xn_enters is reused */

  for (i=0; i<num_crvs1; i++) {
    intersect_plane_curves(&tpst1->UV_Splines[i], uv_crv, &s[0], &t[0], &numXns);
    if (numXns > 10) {
      printf("GetProperTrimCurve: Too many curve-curve intersections\n");
      exit(-1);
    }
    for (j=0; j<numXns; j++) {
      /*
      if ((FP_EQ_EPS(s[j], 0.0, 1e-3) || FP_EQ_EPS(t[j], 0.0, 1e-4) || 
	   FP_EQ_EPS(t[j], 1.0, 1e-4))) 
	continue;
	*/
      Xns1[totXns1] = t[j];
      /* find out whether it's an entry or an exit to the trimmed portion */
      rc1 = EvaluateTangentVector(&tpst1->UV_Splines[i], s[j], &trim_tgt);
      rc2 = EvaluateTangentVector(uv_crv, t[j], &xn_tgt);
      if (rc1 || rc2) {
	printf("GetProperTrimCurve: tangent vector troubles\n");
	exit(-1);
      }
      xn_enters[totXns1] = (xn_tgt.u * trim_tgt.v - xn_tgt.v * trim_tgt.u) < 0;
      ++totXns1;
    }
  }

  /* find first intersection point */
  for (i=0,min_xn=1.0; i<totXns1; i++)
    if (Xns1[i] < min_xn) {
      min_xn = Xns1[i];
      is_inside1 = !xn_enters[i];
    }
  if (!is_inside1)
    Xns1[totXns1++] = 0.0;
  if (totXns1 % 2) 
    Xns1[totXns1++] = 1.0;
  qsort(&Xns1[0], totXns1, sizeof(double), comparParam);

  /* Now I have intersection points between the intersection curve and all the curves
     that make up the trimmed region over both domains. Now to figure out the exact
     parameter values such that the intersection curve lies inside both domains */
  fin_numXns = 0;
  i = j = 0;
  change = 3;
  while ((i < totXns0) && (j < totXns1)) {
    if ((change == 1) || (change == 3)) {
      int1.low = Xns0[i];
      int1.high = Xns0[i+1];
    }
    if ((change == 2) || (change == 3)) {
      int2.low = Xns1[j];
      int2.high = Xns1[j+1];
    }
    if (!open_intervals_overlap(int1, int2)) {
      if (int1.high < int2.high) { change = 1; i += 2; }
      if (int2.high < int1.high) { change = 2; j += 2; }
      continue;
    }
    change = common_interval_overlap(int1, int2, &ret_int);
    finXns[fin_numXns++] = ret_int.low;
    finXns[fin_numXns++] = ret_int.high;

    switch (change) {
    case 1: 
      i += 2;
      break;
    case 2: 
      j += 2;
      break;
    case 3:
      i += 2;
      j += 2;
      break;
    }
  }

  i = 0;
  tmp_2d_crv = (*ret_uv_crv) = (*ret_st_crv) = NULL;
  tmp_3d_crv = (*ret_xyz_crv) = NULL;
  (*ret_uv_crv) = (UV_SPLINE_TYPE *) new((fin_numXns/2) * sizeof(UV_SPLINE_TYPE));
  (*ret_st_crv) = (UV_SPLINE_TYPE *) new((fin_numXns/2) * sizeof(UV_SPLINE_TYPE));
  tmp_2d_crv = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  (*ret_xyz_crv) = (XYZ_SPLINE_TYPE *) new((fin_numXns/2) * sizeof(XYZ_SPLINE_TYPE));
  tmp_3d_crv = (XYZ_SPLINE_TYPE *) new(sizeof(XYZ_SPLINE_TYPE));

  while (i < fin_numXns) {
    TwoDBezierSubdivision(st_crv, finXns[i], finXns[i+1], &tmp_2d_crv);
    copy_uv_spline(tmp_2d_crv, &(*ret_st_crv)[i/2]);
    TwoDBezierSubdivision(uv_crv, finXns[i], finXns[i+1], &tmp_2d_crv);
    copy_uv_spline(tmp_2d_crv, &(*ret_uv_crv)[i/2]);
    ThreeDBezierSubdivision(xyz_crv, finXns[i], finXns[i+1], &tmp_3d_crv);
    copy_xyz_spline(tmp_3d_crv, &(*ret_xyz_crv)[i/2]);
    i += 2;
  }
  return (fin_numXns/2);
}

#ifdef _TESTING_ALONE
main(argc, argv)
int argc;
char *argv[];
{
  FILE *fp;
  double t, par;
  int i, j, k;
  UV_SPLINE_TYPE *ic, *oc;
  POINT2D ap, bp;

  t = 0.2 + 0.25*drand48();
  printf("t is %g \n", t);

  fp = fopen(argv[1], "r");
  ic = (UV_SPLINE_TYPE *) new(sizeof(UV_SPLINE_TYPE));
  fscanf(fp, "%d", &(ic->Deg));
  ic->CntrlPts = (POINT3D *) new((ic->Deg + 1)*sizeof(POINT3D));

  for (i=0; i<=ic->Deg; i++) 
    for (j=0; j<4; j++)
      fscanf(fp, "%lf", &(ic->CntrlPts[i].dCoord[j]));

  fclose(fp);

  Eval2DCurve(ic, t, &ap);
  if (!EvaluateParam(ic, &ap, &par))
    printf("Could not find param\n");
  else
    printf("param is %g\n", par);
/*
  BezierSubdivision(ic, c, d, &oc);

  printf("Output degree is %d\n", oc->Deg);
  for (i=0; i<=oc->Deg; i++) {
    printf("cntrl point [%d] = ", i);
    for (j=0; j<4; j++)
      printf("%g ", oc->CntrlPts[i].dCoord[j]);
    printf("\n");
  }
*/
}
#endif

/* Shankar's tangent vector stuff */
#if 0
#include "defs.h"
#include "linalg.h"

#define new malloc
#define myfree free

void allocate_uv_spline(spline)
UV_SPLINE_TYPE *spline;
{
  ALLOCN(spline->CntrlPts, POINT3D, spline->Deg+1);
}
#endif

/* Evaluates an (x,y,w) triple at a particular param of the curve */
void ObtainProjCoordAtCurve(curve, Value, xval, yval, wval, flag)
UV_SPLINE_TYPE *curve;
double Value;
double *xval, *yval, *wval;
int flag;
{
  int dg, i, j;
  double *SPow, *OmsPow;
  long *CombArr;
  double OneMinusVal;

  OneMinusVal = 1.0 - Value;
  dg = curve->Deg;

  SPow = (double *) new((dg + 1) *
			   sizeof(double));
  OmsPow = (double *) new((dg + 1) *
			   sizeof(double));
  CombArr = (long *) new((dg + 1) *
			 sizeof(long));

  SPow[0] = OmsPow[0] = 1.0;
  SPow[1] = Value; OmsPow[1] = OneMinusVal;
  CombArr[0] = 1; CombArr[1] = dg;
  for (i=2; i<=dg; i++) {
    SPow[i] = SPow[i-1] * Value;
    OmsPow[i] = OmsPow[i-1] * OneMinusVal;
    CombArr[i] = CombArr[i-1] * (dg + 1 - i) / i;
  }

  *xval = *yval = *wval = 0.0;
  for (i=0; i<=dg; i++) {
    (*xval) += curve->CntrlPts[i].dCoord[0] * CombArr[i] *
               SPow[i] * OmsPow[dg - i];
    (*yval) += curve->CntrlPts[i].dCoord[1] * CombArr[i] *
	       SPow[i] * OmsPow[dg - i];
    (*wval) += curve->CntrlPts[i].dCoord[2] * CombArr[i] *
	       SPow[i] * OmsPow[dg - i];
  }

  if (flag && FP_EQ_EPS(*wval, 0.0, 1e-3)) {
    printf("wval is very close to zero\n");
    exit(-1);
  }

  myfree(SPow);
  myfree(OmsPow);
  myfree(CombArr);
}

/* This function takes a Bezier curve F(t) = < X(t), Y(t), Z(t) > and
   generates F'(t) = < X'(t), Y'(t), Z'(t) > */
void get_partials(spl, d_spl)
UV_SPLINE_TYPE *spl, *d_spl;
{
  int i, j;

  d_spl->Deg = spl->Deg - 1;
  allocate_uv_spline(d_spl);

  for (i=0; i<=d_spl->Deg; i++)
    for (j=0; j<3; j++) 
      d_spl->CntrlPts[i].dCoord[j] =
	spl->Deg*(spl->CntrlPts[i+1].dCoord[j] -
		  spl->CntrlPts[i].dCoord[j]);
}

/* Given a Bezier curve spl, it evaluates the tangent vector to the
   curve at Param in tgt */
int EvaluateTangentVector(spl, Param, tgt)
UV_SPLINE_TYPE *spl;
double Param;
POINT2D *tgt;
{
  int i, j, k;
  UV_SPLINE_TYPE spl_dash;
  double xv, yv, wv, xv_dash, yv_dash, wv_dash;
  double one_over_w_squared;

  get_partials(spl, &spl_dash);
  ObtainProjCoordAtCurve(spl, Param, &xv, &yv, &wv, 1);
  ObtainProjCoordAtCurve(&spl_dash, Param, &xv_dash, &yv_dash,
			 &wv_dash, 0);
  
  one_over_w_squared = 1.0/(wv * wv);
  tgt->u = one_over_w_squared * (wv * xv_dash - xv * wv_dash);
  tgt->v = one_over_w_squared * (wv * yv_dash - yv * wv_dash);

  return 0;
}

#if 0
void main(argc, argv)
int argc;
char *argv[];
{
  FILE *fp;
  UV_SPLINE_TYPE spl;
  int i, j, k;
  double Param;
  POINT2D tgt;

  if (argc != 2) {
    printf("Usage: deriv <filename>\n");
    exit(-1);
  }

  if ((fp = fopen(argv[1], "r")) == NULL) {
    printf("Couldn't open %s\n", argv[1]);
    exit(-1);
  }

  fscanf(fp, "%d", &spl.Deg);
  allocate_uv_spline(&spl);

  for (i=0; i<=spl.Deg; i++)
    for (j=0; j<3; j++)
      fscanf(fp, "%lf", &(spl.CntrlPts[i].dCoord[j]));

  fclose(fp);

  printf("Enter parameter value at which tangent vector is to be found: ");
  scanf("%lf", &Param);

  if (!EvaluateTangentVector(&spl, Param, &tgt))
    printf("Tangent vector is (%lf, %lf)\n", tgt.u, tgt.v);
  else
    printf("Tangent vector could not be found\n");
}

#endif
